
#include "stdafx.h"
#include "standard_crypto_primitives.h"
#include "polymorphic_medley.h"


namespace polymorphic_medley_cipher
{

		// congruential KISS sub-functions
	#define CNG0 ( pkctx->w+=pkctx->z, pkctx->z=6906969069LL*pkctx->z+(pkctx->c>>((pkctx->w & 0x2f))), pkctx->w )	/* + + */
	#define CNG1 ( pkctx->w^=pkctx->z, pkctx->z=6906969069LL*pkctx->z+(pkctx->w>>((pkctx->c & 0x1f))), pkctx->w ) /* ^ + */
	#define CNG2 ( pkctx->w^=pkctx->z, pkctx->z=6906969069LL*pkctx->z+(pkctx->c>>((pkctx->w & 0x2f))), pkctx->w ) /* ^ + */
	#define CNG3 ( pkctx->w+=pkctx->z, pkctx->z=6906969069LL*pkctx->z+(pkctx->w>>((pkctx->c & 0x1f))), pkctx->w ) /* + + */

		// XOR Shift KISS sub-functions
	#define XSH0 ( pkctx->jsr^=(pkctx->jsr<<13), pkctx->jsr^=(pkctx->jsr>>7), pkctx->jsr^=(pkctx->jsr<<17) )	/* 13 7 17 */
	#define XSH1 ( pkctx->jsr^=(pkctx->jsr<<13), pkctx->jsr^=(pkctx->jsr>>9), pkctx->jsr^=(pkctx->jsr<<15) )	/* 13 9 15 */
	#define XSH2 ( pkctx->jsr^=(pkctx->jsr<<13), pkctx->jsr^=(pkctx->jsr>>9), pkctx->jsr^=(pkctx->jsr<<50) )	/* 13 9 50 */
	#define XSH3 ( pkctx->jsr^=(pkctx->jsr<<13), pkctx->jsr^=(pkctx->jsr>>13), pkctx->jsr^=(pkctx->jsr<<19) )	/* 13 13 19 */
	#define XSH4 ( pkctx->jsr^=(pkctx->jsr<<13), pkctx->jsr^=(pkctx->jsr>>17), pkctx->jsr^=(pkctx->jsr<<43) )	/* 13 17 43 */
	#define XSH5 ( pkctx->jsr^=(pkctx->jsr<<13), pkctx->jsr^=(pkctx->jsr>>19), pkctx->jsr^=(pkctx->jsr<<28) )	/* 13 19 28 */
	#define XSH6 ( pkctx->jsr^=(pkctx->jsr<<13), pkctx->jsr^=(pkctx->jsr>>19), pkctx->jsr^=(pkctx->jsr<<47) )	/* 13 19 47 */
	#define XSH7 ( pkctx->jsr^=(pkctx->jsr<<13), pkctx->jsr^=(pkctx->jsr>>21), pkctx->jsr^=(pkctx->jsr<<18) )	/* 13 21 18 */

		// Multiply-With-Carry KISS sub-functions
	#define MWC0 ( t=(pkctx->x<<58)+pkctx->c, pkctx->c=(pkctx->x>>6), pkctx->x+=t, pkctx->c+=(pkctx->x<t), pkctx->x ) /* 6 + */
	#define MWC1 ( t=(pkctx->x<<58)+pkctx->c, pkctx->c=(pkctx->x>>6), pkctx->x-=t, pkctx->c+=(pkctx->x<t), pkctx->x ) /* 6 - */
	#define MWC2 ( t=(pkctx->x<<58)+pkctx->c, pkctx->c=(pkctx->x>>5), pkctx->x+=t, pkctx->c+=(pkctx->x<t), pkctx->x ) /* 5 + */
	#define MWC3 ( t=(pkctx->x<<58)+pkctx->c, pkctx->c=(pkctx->x>>5), pkctx->x-=t, pkctx->c+=(pkctx->x<t), pkctx->x ) /* 5 - */ 



	// KISSes:

	word64 kiss0(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH0+MWC0);
	}
	word64 kiss1(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH0+MWC1);
	}
	word64 kiss2(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH0+MWC2);
	}
	word64 kiss3(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH0+MWC3);
	}
	word64 kiss4(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH1+MWC0);
	}
	word64 kiss5(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH1+MWC1);
	}
	word64 kiss6(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH1+MWC2);
	}
	word64 kiss7(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH1+MWC3);
	}
	word64 kiss8(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH2+MWC0);
	}
	word64 kiss9(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH2+MWC1);
	}
	word64 kiss10(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH2+MWC2);
	}
	word64 kiss11(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH2+MWC3);
	}
	word64 kiss12(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH3+MWC0);
	}
	word64 kiss13(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH3+MWC1);
	}
	word64 kiss14(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH3+MWC2);
	}
	word64 kiss15(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH3+MWC3);
	}
	word64 kiss16(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH4+MWC0);
	}
	word64 kiss17(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH4+MWC1);
	}
	word64 kiss18(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH4+MWC2);
	}
	word64 kiss19(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH4+MWC3);
	}
	word64 kiss20(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH5+MWC0);
	}
	word64 kiss21(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH5+MWC1);
	}
	word64 kiss22(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH5+MWC2);
	}
	word64 kiss23(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH5+MWC3);
	}
	word64 kiss24(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH6+MWC0);
	}
	word64 kiss25(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH6+MWC1);
	}
	word64 kiss26(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH6+MWC2);
	}
	word64 kiss27(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH6+MWC3);
	}
	word64 kiss28(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH7+MWC0);
	}
	word64 kiss29(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH7+MWC1);
	}
	word64 kiss30(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH7+MWC2);
	}
	word64 kiss31(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG0+XSH7+MWC3);
	}
	word64 kiss32(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH0+MWC0);
	}
	word64 kiss33(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH0+MWC1);
	}
	word64 kiss34(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH0+MWC2);
	}
	word64 kiss35(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH0+MWC3);
	}
	word64 kiss36(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH1+MWC0);
	}
	word64 kiss37(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH1+MWC1);
	}
	word64 kiss38(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH1+MWC2);
	}
	word64 kiss39(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH1+MWC3);
	}
	word64 kiss40(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH2+MWC0);
	}
	word64 kiss41(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH2+MWC1);
	}
	word64 kiss42(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH2+MWC2);
	}
	word64 kiss43(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH2+MWC3);
	}
	word64 kiss44(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH3+MWC0);
	}
	word64 kiss45(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH3+MWC1);
	}
	word64 kiss46(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH3+MWC2);
	}
	word64 kiss47(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH3+MWC3);
	}
	word64 kiss48(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH4+MWC0);
	}
	word64 kiss49(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH4+MWC1);
	}
	word64 kiss50(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH4+MWC2);
	}
	word64 kiss51(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH4+MWC3);
	}
	word64 kiss52(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH5+MWC0);
	}
	word64 kiss53(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH5+MWC1);
	}
	word64 kiss54(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH5+MWC2);
	}
	word64 kiss55(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH5+MWC3);
	}
	word64 kiss56(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH6+MWC0);
	}
	word64 kiss57(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH6+MWC1);
	}
	word64 kiss58(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH6+MWC2);
	}
	word64 kiss59(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH6+MWC3);
	}
	word64 kiss60(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH7+MWC0);
	}
	word64 kiss61(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH7+MWC1);
	}
	word64 kiss62(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH7+MWC2);
	}
	word64 kiss63(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG1+XSH7+MWC3);
	}
	word64 kiss64(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH0+MWC0);
	}
	word64 kiss65(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH0+MWC1);
	}
	word64 kiss66(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH0+MWC2);
	}
	word64 kiss67(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH0+MWC3);
	}
	word64 kiss68(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH1+MWC0);
	}
	word64 kiss69(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH1+MWC1);
	}
	word64 kiss70(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH1+MWC2);
	}
	word64 kiss71(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH1+MWC3);
	}
	word64 kiss72(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH2+MWC0);
	}
	word64 kiss73(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH2+MWC1);
	}
	word64 kiss74(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH2+MWC2);
	}
	word64 kiss75(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH2+MWC3);
	}
	word64 kiss76(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH3+MWC0);
	}
	word64 kiss77(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH3+MWC1);
	}
	word64 kiss78(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH3+MWC2);
	}
	word64 kiss79(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH3+MWC3);
	}
	word64 kiss80(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH4+MWC0);
	}
	word64 kiss81(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH4+MWC1);
	}
	word64 kiss82(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH4+MWC2);
	}
	word64 kiss83(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH4+MWC3);
	}
	word64 kiss84(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH5+MWC0);
	}
	word64 kiss85(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH5+MWC1);
	}
	word64 kiss86(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH5+MWC2);
	}
	word64 kiss87(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH5+MWC3);
	}
	word64 kiss88(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH6+MWC0);
	}
	word64 kiss89(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH6+MWC1);
	}
	word64 kiss90(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH6+MWC2);
	}
	word64 kiss91(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH6+MWC3);
	}
	word64 kiss92(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH7+MWC0);
	}
	word64 kiss93(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH7+MWC1);
	}
	word64 kiss94(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH7+MWC2);
	}
	word64 kiss95(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG2+XSH7+MWC3);
	}
	word64 kiss96(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH0+MWC0);
	}
	word64 kiss97(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH0+MWC1);
	}
	word64 kiss98(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH0+MWC2);
	}
	word64 kiss99(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH0+MWC3);
	}
	word64 kiss100(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH1+MWC0);
	}
	word64 kiss101(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH1+MWC1);
	}
	word64 kiss102(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH1+MWC2);
	}
	word64 kiss103(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH1+MWC3);
	}
	word64 kiss104(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH2+MWC0);
	}
	word64 kiss105(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH2+MWC1);
	}
	word64 kiss106(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH2+MWC2);
	}
	word64 kiss107(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH2+MWC3);
	}
	word64 kiss108(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH3+MWC0);
	}
	word64 kiss109(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH3+MWC1);
	}
	word64 kiss110(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH3+MWC2);
	}
	word64 kiss111(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH3+MWC3);
	}
	word64 kiss112(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH4+MWC0);
	}
	word64 kiss113(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH4+MWC1);
	}
	word64 kiss114(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH4+MWC2);
	}
	word64 kiss115(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH4+MWC3);
	}
	word64 kiss116(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH5+MWC0);
	}
	word64 kiss117(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH5+MWC1);
	}
	word64 kiss118(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH5+MWC2);
	}
	word64 kiss119(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH5+MWC3);
	}
	word64 kiss120(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH6+MWC0);
	}
	word64 kiss121(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH6+MWC1);
	}
	word64 kiss122(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH6+MWC2);
	}
	word64 kiss123(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH6+MWC3);
	}
	word64 kiss124(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH7+MWC0);
	}
	word64 kiss125(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH7+MWC1);
	}
	word64 kiss126(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH7+MWC2);
	}
	word64 kiss127(gbpmc_kiss_context * pkctx)
	{
		word64 t;
		return (CNG3+XSH7+MWC3);
	}


	void init_gbpmc_kiss_delegate_array(KISSDELEGATE	*	pgbpmc_k[])
	{
		pgbpmc_k[0]=(KISSDELEGATE*) &kiss0;
		pgbpmc_k[1]=(KISSDELEGATE*) &kiss1;
		pgbpmc_k[2]=(KISSDELEGATE*) &kiss2;
		pgbpmc_k[3]=(KISSDELEGATE*) &kiss3;
		pgbpmc_k[4]=(KISSDELEGATE*) &kiss4;
		pgbpmc_k[5]=(KISSDELEGATE*) &kiss5;
		pgbpmc_k[6]=(KISSDELEGATE*) &kiss6;
		pgbpmc_k[7]=(KISSDELEGATE*) &kiss7;
		pgbpmc_k[8]=(KISSDELEGATE*) &kiss8;
		pgbpmc_k[9]=(KISSDELEGATE*) &kiss9;
		pgbpmc_k[10]=(KISSDELEGATE*) &kiss10;
		pgbpmc_k[11]=(KISSDELEGATE*) &kiss11;
		pgbpmc_k[12]=(KISSDELEGATE*) &kiss12;
		pgbpmc_k[13]=(KISSDELEGATE*) &kiss13;
		pgbpmc_k[14]=(KISSDELEGATE*) &kiss14;
		pgbpmc_k[15]=(KISSDELEGATE*) &kiss15;
		pgbpmc_k[16]=(KISSDELEGATE*) &kiss16;
		pgbpmc_k[17]=(KISSDELEGATE*) &kiss17;
		pgbpmc_k[18]=(KISSDELEGATE*) &kiss18;
		pgbpmc_k[19]=(KISSDELEGATE*) &kiss19;
		pgbpmc_k[20]=(KISSDELEGATE*) &kiss20;
		pgbpmc_k[21]=(KISSDELEGATE*) &kiss21;
		pgbpmc_k[22]=(KISSDELEGATE*) &kiss22;
		pgbpmc_k[23]=(KISSDELEGATE*) &kiss23;
		pgbpmc_k[24]=(KISSDELEGATE*) &kiss24;
		pgbpmc_k[25]=(KISSDELEGATE*) &kiss25;
		pgbpmc_k[26]=(KISSDELEGATE*) &kiss26;
		pgbpmc_k[27]=(KISSDELEGATE*) &kiss27;
		pgbpmc_k[28]=(KISSDELEGATE*) &kiss28;
		pgbpmc_k[29]=(KISSDELEGATE*) &kiss29;
		pgbpmc_k[30]=(KISSDELEGATE*) &kiss30;
		pgbpmc_k[31]=(KISSDELEGATE*) &kiss31;
		pgbpmc_k[32]=(KISSDELEGATE*) &kiss32;
		pgbpmc_k[33]=(KISSDELEGATE*) &kiss33;
		pgbpmc_k[34]=(KISSDELEGATE*) &kiss34;
		pgbpmc_k[35]=(KISSDELEGATE*) &kiss35;
		pgbpmc_k[36]=(KISSDELEGATE*) &kiss36;
		pgbpmc_k[37]=(KISSDELEGATE*) &kiss37;
		pgbpmc_k[38]=(KISSDELEGATE*) &kiss38;
		pgbpmc_k[39]=(KISSDELEGATE*) &kiss39;
		pgbpmc_k[40]=(KISSDELEGATE*) &kiss40;
		pgbpmc_k[41]=(KISSDELEGATE*) &kiss41;
		pgbpmc_k[42]=(KISSDELEGATE*) &kiss42;
		pgbpmc_k[43]=(KISSDELEGATE*) &kiss43;
		pgbpmc_k[44]=(KISSDELEGATE*) &kiss44;
		pgbpmc_k[45]=(KISSDELEGATE*) &kiss45;
		pgbpmc_k[46]=(KISSDELEGATE*) &kiss46;
		pgbpmc_k[47]=(KISSDELEGATE*) &kiss47;
		pgbpmc_k[48]=(KISSDELEGATE*) &kiss48;
		pgbpmc_k[49]=(KISSDELEGATE*) &kiss49;
		pgbpmc_k[50]=(KISSDELEGATE*) &kiss50;
		pgbpmc_k[51]=(KISSDELEGATE*) &kiss51;
		pgbpmc_k[52]=(KISSDELEGATE*) &kiss52;
		pgbpmc_k[53]=(KISSDELEGATE*) &kiss53;
		pgbpmc_k[54]=(KISSDELEGATE*) &kiss54;
		pgbpmc_k[55]=(KISSDELEGATE*) &kiss55;
		pgbpmc_k[56]=(KISSDELEGATE*) &kiss56;
		pgbpmc_k[57]=(KISSDELEGATE*) &kiss57;
		pgbpmc_k[58]=(KISSDELEGATE*) &kiss58;
		pgbpmc_k[59]=(KISSDELEGATE*) &kiss59;
		pgbpmc_k[60]=(KISSDELEGATE*) &kiss60;
		pgbpmc_k[61]=(KISSDELEGATE*) &kiss61;
		pgbpmc_k[62]=(KISSDELEGATE*) &kiss62;
		pgbpmc_k[63]=(KISSDELEGATE*) &kiss63;
		pgbpmc_k[64]=(KISSDELEGATE*) &kiss64;
		pgbpmc_k[65]=(KISSDELEGATE*) &kiss65;
		pgbpmc_k[66]=(KISSDELEGATE*) &kiss66;
		pgbpmc_k[67]=(KISSDELEGATE*) &kiss67;
		pgbpmc_k[68]=(KISSDELEGATE*) &kiss68;
		pgbpmc_k[69]=(KISSDELEGATE*) &kiss69;
		pgbpmc_k[70]=(KISSDELEGATE*) &kiss70;
		pgbpmc_k[71]=(KISSDELEGATE*) &kiss71;
		pgbpmc_k[72]=(KISSDELEGATE*) &kiss72;
		pgbpmc_k[73]=(KISSDELEGATE*) &kiss73;
		pgbpmc_k[74]=(KISSDELEGATE*) &kiss74;
		pgbpmc_k[75]=(KISSDELEGATE*) &kiss75;
		pgbpmc_k[76]=(KISSDELEGATE*) &kiss76;
		pgbpmc_k[77]=(KISSDELEGATE*) &kiss77;
		pgbpmc_k[78]=(KISSDELEGATE*) &kiss78;
		pgbpmc_k[79]=(KISSDELEGATE*) &kiss79;
		pgbpmc_k[80]=(KISSDELEGATE*) &kiss80;
		pgbpmc_k[81]=(KISSDELEGATE*) &kiss81;
		pgbpmc_k[82]=(KISSDELEGATE*) &kiss82;
		pgbpmc_k[83]=(KISSDELEGATE*) &kiss83;
		pgbpmc_k[84]=(KISSDELEGATE*) &kiss84;
		pgbpmc_k[85]=(KISSDELEGATE*) &kiss85;
		pgbpmc_k[86]=(KISSDELEGATE*) &kiss86;
		pgbpmc_k[87]=(KISSDELEGATE*) &kiss87;
		pgbpmc_k[88]=(KISSDELEGATE*) &kiss88;
		pgbpmc_k[89]=(KISSDELEGATE*) &kiss89;
		pgbpmc_k[90]=(KISSDELEGATE*) &kiss90;
		pgbpmc_k[91]=(KISSDELEGATE*) &kiss91;
		pgbpmc_k[92]=(KISSDELEGATE*) &kiss92;
		pgbpmc_k[93]=(KISSDELEGATE*) &kiss93;
		pgbpmc_k[94]=(KISSDELEGATE*) &kiss94;
		pgbpmc_k[95]=(KISSDELEGATE*) &kiss95;
		pgbpmc_k[96]=(KISSDELEGATE*) &kiss96;
		pgbpmc_k[97]=(KISSDELEGATE*) &kiss97;
		pgbpmc_k[98]=(KISSDELEGATE*) &kiss98;
		pgbpmc_k[99]=(KISSDELEGATE*) &kiss99;
		pgbpmc_k[100]=(KISSDELEGATE*) &kiss100;
		pgbpmc_k[101]=(KISSDELEGATE*) &kiss101;
		pgbpmc_k[102]=(KISSDELEGATE*) &kiss102;
		pgbpmc_k[103]=(KISSDELEGATE*) &kiss103;
		pgbpmc_k[104]=(KISSDELEGATE*) &kiss104;
		pgbpmc_k[105]=(KISSDELEGATE*) &kiss105;
		pgbpmc_k[106]=(KISSDELEGATE*) &kiss106;
		pgbpmc_k[107]=(KISSDELEGATE*) &kiss107;
		pgbpmc_k[108]=(KISSDELEGATE*) &kiss108;
		pgbpmc_k[109]=(KISSDELEGATE*) &kiss109;
		pgbpmc_k[110]=(KISSDELEGATE*) &kiss110;
		pgbpmc_k[111]=(KISSDELEGATE*) &kiss111;
		pgbpmc_k[112]=(KISSDELEGATE*) &kiss112;
		pgbpmc_k[113]=(KISSDELEGATE*) &kiss113;
		pgbpmc_k[114]=(KISSDELEGATE*) &kiss114;
		pgbpmc_k[115]=(KISSDELEGATE*) &kiss115;
		pgbpmc_k[116]=(KISSDELEGATE*) &kiss116;
		pgbpmc_k[117]=(KISSDELEGATE*) &kiss117;
		pgbpmc_k[118]=(KISSDELEGATE*) &kiss118;
		pgbpmc_k[119]=(KISSDELEGATE*) &kiss119;
		pgbpmc_k[120]=(KISSDELEGATE*) &kiss120;
		pgbpmc_k[121]=(KISSDELEGATE*) &kiss121;
		pgbpmc_k[122]=(KISSDELEGATE*) &kiss122;
		pgbpmc_k[123]=(KISSDELEGATE*) &kiss123;
		pgbpmc_k[124]=(KISSDELEGATE*) &kiss124;
		pgbpmc_k[125]=(KISSDELEGATE*) &kiss125;
		pgbpmc_k[126]=(KISSDELEGATE*) &kiss126;
		pgbpmc_k[127]=(KISSDELEGATE*) &kiss127;
	}

	void init_gbpmc_kiss_context(gbpmc_kiss_context * kc,word64 param1,word64 param2)
	{
		kc->z			=	param1;
		kc->jsr		=	(param1^param2)| 0x1000000000000001;
		kc->w			=	param2;    
		kc->j			=	param1-param2;
		kc->x			=	param1;
		kc->c			=	kc->jsr;
	}


		// PMCMED_keysetup returns 0 if the key could be set up correctly.
		// These values are returned in case of an error:
		// -1: insufficient memory
		// -2: MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES does NOT equal the return value of CIPHER_PRIMITIVE_ComputeMaxSizeOfCryptoContext() => MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES must be corrected and then the code recompiled
	int PMCMED_keysetup(uint8 * pKey,void * pPMCMED_cc,uint32 key_length_in_bits,uint32 complexity)			// key_length_in_bits: 128 .. 1024 bit, complexity: 0: very fast, 2^16-1: huge set of keyed operations to compute internal state
	{
		int															retval=0;																														// 0: no error, any other value indicates an error
		struct PMCMED_cipher_context	*	pPMCMED_cipher_ctx;
		int															i,j=key_length_in_bits;
		int															k;
		int															key_length_in_bytes=key_length_in_bits>>3;
		uint8														intermediate_hash_result[512];
		HASH_FUNCTION_DELEGATE				*	pHash_Func_Delegate_buf;
		uint32													complexity_count;
		uint32													rnd32=0;
		int															size_of_cc;
		uint8														local_cc[MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES];
		uint8													*	pCc;
		KEY_SETUP_FUNCTION_DELEGATE		*	pKey_Setup_Func_delegate_buf;
		ENCRYPTION_FUNCTION_DELEGATE	*	pEncryption_Func_delegate_buf;
		DECRYPTION_FUNCTION_DELEGATE	*	pDecryption_Func_delegate_buf;
		uint8														Ciphertext[16];
		uint32												*	pUi32;
		word64													(*pMemory_bias64)[];
		int															memory_bias_cnt=0;																									// counter for filling our memory bias array slowly but surely
		word64													rnd64=0;
		word64													w64_1,w64_2;
		KISSDELEGATE									*	kiss_func=NULL;


		pPMCMED_cipher_ctx=(struct PMCMED_cipher_context *)pPMCMED_cc;

		init_gbpmc_kiss_delegate_array(pPMCMED_cipher_ctx->gbpmc_kiss);
		for (i=0;i<NUM_OF_KISS_CONTEXTS;i++)
		{
			pPMCMED_cipher_ctx->initial_ctx_array[i]=new gbpmc_kiss_context;
		}

		pPMCMED_cipher_ctx->memory_bias=new uint8[SIZE_OF_MEMORY_BIAS_ARRAY];
		if (!pPMCMED_cipher_ctx->memory_bias) return -1;																										// ERROR: insufficient memory

		pMemory_bias64=(word64 (*)[]) pPMCMED_cipher_ctx->memory_bias;																			// the data array of bytes can thus be addressed as an array of unsigned __int64 from now on

		pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results=new uint8[16*(NUM_OF_HASH_RESULTS+4)];						// +3: we must take care of different hash lengths
		if (!pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results) return -1;																	// ERROR: insufficient memory

		if (CIPHER_PRIMITIVE_ComputeMaxSizeOfCryptoContext()>MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES) return -2;// ERROR: MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES must be set to a bigger value. If this is done, then the software can be compiled again.
		pPMCMED_cipher_ctx->pcc=new uint8[MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES*NUM_OF_CIPHER_FUNCTION_DELEGATES*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE];
		if (!pPMCMED_cipher_ctx->pcc) return -1;																														// ERROR: insufficient memory

		size_of_cc=MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES*NUM_OF_CIPHER_FUNCTION_DELEGATES*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE;

		if (complexity>65535) complexity=65535;																															// 2^16-1 is the absolute maximum. Otherwise we risk crashes.
		if (complexity<256) complexity=256;																																	// 256 is the absolute minimum

		for (i=0;i<512;i++) intermediate_hash_result[i]=(uint8)((i+key_length_in_bits+complexity) & 0xff);

			// step #1: setup pointers to hash functions as well as to our cipher primitives
		pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[0]=(HASH_FUNCTION_DELEGATE*) &HASH_FUNCTION_Whirlpool512;
		pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[1]=(HASH_FUNCTION_DELEGATE*) &HASH_FUNCTION_Haval256;
		pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[2]=(HASH_FUNCTION_DELEGATE*) &HASH_FUNCTION_Tiger192;
		pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[3]=(HASH_FUNCTION_DELEGATE*) &HASH_FUNCTION_Ripemd320;
		pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[4]=(HASH_FUNCTION_DELEGATE*) &HASH_FUNCTION_Sha256;

		pPMCMED_cipher_ctx->key_setup_func_delegates[0]=(KEY_SETUP_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_KEY_SETUP_Anubis128;
		pPMCMED_cipher_ctx->key_setup_func_delegates[1]=(KEY_SETUP_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_KEY_SETUP_Seed128;
		pPMCMED_cipher_ctx->key_setup_func_delegates[2]=(KEY_SETUP_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_KEY_SETUP_RC6_128;
		pPMCMED_cipher_ctx->key_setup_func_delegates[3]=(KEY_SETUP_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_KEY_SETUP_Camellia128;
		pPMCMED_cipher_ctx->key_setup_func_delegates[4]=(KEY_SETUP_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_KEY_SETUP_Cast256_128;
		pPMCMED_cipher_ctx->key_setup_func_delegates[5]=(KEY_SETUP_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_KEY_SETUP_Serpent128;
		pPMCMED_cipher_ctx->key_setup_func_delegates[6]=(KEY_SETUP_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_KEY_SETUP_Rijndael128;
		pPMCMED_cipher_ctx->key_setup_func_delegates[7]=(KEY_SETUP_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_KEY_SETUP_Twofish128;

		pPMCMED_cipher_ctx->encryption_func_delegates[0]=(ENCRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_ENCRYPT_Anubis128;
		pPMCMED_cipher_ctx->encryption_func_delegates[1]=(ENCRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_ENCRYPT_Seed128;
		pPMCMED_cipher_ctx->encryption_func_delegates[2]=(ENCRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_ENCRYPT_RC6_128;
		pPMCMED_cipher_ctx->encryption_func_delegates[3]=(ENCRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_ENCRYPT_Camellia128;
		pPMCMED_cipher_ctx->encryption_func_delegates[4]=(ENCRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_ENCRYPT_Cast256_128;
		pPMCMED_cipher_ctx->encryption_func_delegates[5]=(ENCRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_ENCRYPT_Serpent128;
		pPMCMED_cipher_ctx->encryption_func_delegates[6]=(ENCRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_ENCRYPT_Rijndael128;
		pPMCMED_cipher_ctx->encryption_func_delegates[7]=(ENCRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_ENCRYPT_Twofish128;

		pPMCMED_cipher_ctx->decryption_func_delegates[0]=(DECRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_DECRYPT_Anubis128;
		pPMCMED_cipher_ctx->decryption_func_delegates[1]=(DECRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_DECRYPT_Seed128;
		pPMCMED_cipher_ctx->decryption_func_delegates[2]=(DECRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_DECRYPT_RC6_128;
		pPMCMED_cipher_ctx->decryption_func_delegates[3]=(DECRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_DECRYPT_Camellia128;
		pPMCMED_cipher_ctx->decryption_func_delegates[4]=(DECRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_DECRYPT_Cast256_128;
		pPMCMED_cipher_ctx->decryption_func_delegates[5]=(DECRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_DECRYPT_Serpent128;
		pPMCMED_cipher_ctx->decryption_func_delegates[6]=(DECRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_DECRYPT_Rijndael128;
		pPMCMED_cipher_ctx->decryption_func_delegates[7]=(DECRYPTION_FUNCTION_DELEGATE*) &CIPHER_PRIMITIVE_DECRYPT_Twofish128;

		for (i=0;i<512;i++)
		{
			do {
				intermediate_hash_result[i]=(uint8)(((intermediate_hash_result[i]+j)*7) & 0xff);
				j=((intermediate_hash_result[i]*7)>>3) & 0x07;																									// & 0x07: needs to be adapted once there are more than 5 hash function primitives available
			} while (j>=NO_OF_HASH_FUNCTION_DELEGATES);

				// swap pointers (introduces almost no pseudorandomness, but this is ok at this stage)
			pHash_Func_Delegate_buf=pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[0];
			pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[0]=pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[j];
			pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[j]=pHash_Func_Delegate_buf;
		}

			// step #2: perform the first keyed operation in order to generate a few bytes that fully depend on the key
		pHash_Func_Delegate_buf(pKey,key_length_in_bytes,&intermediate_hash_result[128]);										// compute between 20 to 64 bytes that all depend on the key
		for (i=0;i<key_length_in_bytes;i++) intermediate_hash_result[i]+=pKey[i];														// put up to 128 bytes of the key in front of the computed values
		rnd32=(((uint32)intermediate_hash_result[129])<<24) + (((uint32)intermediate_hash_result[135])<<16) + (((uint32)intermediate_hash_result[137])<<8) + ((uint32)intermediate_hash_result[143]);

			// let's init our 64 bit PRNGs
		w64_1=(((word64)intermediate_hash_result[128])<<(32+24)) + (((word64)intermediate_hash_result[130])<<(32+16)) + (((word64)intermediate_hash_result[129])<<(32+8)) + ((word64)intermediate_hash_result[131]<<32)+(((word64)intermediate_hash_result[134])<<24) + (((word64)intermediate_hash_result[132])<<16) + (((word64)intermediate_hash_result[133])<<8) + ((word64)intermediate_hash_result[135]);
		w64_2=(((word64)intermediate_hash_result[136])<<(32+24)) + (((word64)intermediate_hash_result[138])<<(32+16)) + (((word64)intermediate_hash_result[139])<<(32+8)) + ((word64)intermediate_hash_result[137]<<32)+(((word64)intermediate_hash_result[140])<<24) + (((word64)intermediate_hash_result[141])<<16) + (((word64)intermediate_hash_result[143])<<8) + ((word64)intermediate_hash_result[142]);

		kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(((w64_1>>4)*13)-((w64_2)>>2)*3) & (NUM_OF_KISSDELEGATES-1)];
		for (i=0;i<NUM_OF_KISS_CONTEXTS;i++)
		{
			init_gbpmc_kiss_context(pPMCMED_cipher_ctx->initial_ctx_array[i],w64_1,w64_2);
				// now the context is actually ready for use
			w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[i]);
			w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[i]);
			kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[w64_1 & (NUM_OF_KISSDELEGATES-1)];
		} // for (i=0;i<NUM_OF_KISS_CONTEXTS;i++)
		j=(int)(w64_2 & (NUM_OF_KISS_CONTEXTS-1));
		for (i=0;i<(SIZE_OF_MEMORY_BIAS_ARRAY>>3);i=i+4)
		{
			(*pMemory_bias64)[i]=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[j])+3*w64_1;
			w64_1^=(*pMemory_bias64)[i+1]=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_2 & (NUM_OF_KISS_CONTEXTS-1)]);
			w64_2^=(*pMemory_bias64)[i+2]=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[j]);
			w64_1=w64_1-3*w64_2;
			w64_2+=(*pMemory_bias64)[i+3]=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_1 & (NUM_OF_KISS_CONTEXTS-1)]);
		}


			// step 3: compute preliminary hash results
		j=0;
		for (complexity_count=rnd32 & 0x000f;complexity_count<complexity;complexity_count=complexity_count+((rnd32>>5) & 0x001f))
		{
			if (memory_bias_cnt+NUM_OF_HASH_RESULTS*11>=SIZE_OF_MEMORY_BIAS_ARRAY>>3) 
			{
				memory_bias_cnt=0;
			}

			for (i=0;i<NUM_OF_HASH_RESULTS;i++)
			{
				pHash_Func_Delegate_buf=pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[j];
				j++;
				if (j>=NO_OF_HASH_FUNCTION_DELEGATES) j=0;
					// each hash function returns the # of bytes of the sum. Depending on rnd32, we sometimes add 0 .. 3 to i, so that  sometimes 16 .. 64 bytes of the sum are stored (depending on the capabilities of the hash function)
				k=pHash_Func_Delegate_buf(intermediate_hash_result,192+i+((complexity_count>>3) & 0x00ff),&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[i<<4])/(17+(rnd32 & 0x3f));	// i<<4: 16 bytes increment
				rnd32^=intermediate_hash_result[192+i+((complexity_count>>3) & 0x00ff)]+=pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[(i<<4)+((rnd32>>3) & 0x0f)];	// we constantly modify intermediate_hash_result

				w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_1 & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
				memory_bias_cnt++;

					// we sometimes interrupt the sequence and swap function pointers to our hash functions
				if ((rnd32 & 0x1f) < NO_OF_HASH_FUNCTION_DELEGATES) j=rnd32 & 0x1f;
					// swap pointers (introduces almost no pseudorandomness, but this is ok at this stage)
				pHash_Func_Delegate_buf=pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[0];
				pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[0]=pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[j];
				pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[j]=pHash_Func_Delegate_buf;

				w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_1 & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_2;
				memory_bias_cnt++;

					// let's also swap pointers to our cipher primitives
				pKey_Setup_Func_delegate_buf=pPMCMED_cipher_ctx->key_setup_func_delegates[0];
				pPMCMED_cipher_ctx->key_setup_func_delegates[0]=pPMCMED_cipher_ctx->key_setup_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
				pPMCMED_cipher_ctx->key_setup_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]=pKey_Setup_Func_delegate_buf;
				pEncryption_Func_delegate_buf=pPMCMED_cipher_ctx->encryption_func_delegates[0];
				pPMCMED_cipher_ctx->encryption_func_delegates[0]=pPMCMED_cipher_ctx->encryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
				pPMCMED_cipher_ctx->encryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]=pEncryption_Func_delegate_buf;
				pDecryption_Func_delegate_buf=pPMCMED_cipher_ctx->decryption_func_delegates[0];
				pPMCMED_cipher_ctx->decryption_func_delegates[0]=pPMCMED_cipher_ctx->decryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
				pPMCMED_cipher_ctx->decryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]=pDecryption_Func_delegate_buf;

				w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_2 & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
				memory_bias_cnt++;

					// .. and let's use the freshly exchanged cipher in order to generate more pseudorandomness
				pPMCMED_cipher_ctx->key_setup_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)](&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[i<<4],(void *)&local_cc);
				if ((rnd32>>27) & 0x01) pPMCMED_cipher_ctx->encryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]((void *)&local_cc,&intermediate_hash_result[128],Ciphertext);
				else pPMCMED_cipher_ctx->decryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]((void *)&local_cc,&intermediate_hash_result[129],Ciphertext);;

				w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_1 & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
				memory_bias_cnt++;

				intermediate_hash_result[(rnd32>>22) & 0x01ff]^=0x55;
				rnd32=rnd32<<1;
				rnd32+=(((uint32)Ciphertext[Ciphertext[Ciphertext[10] & 0x000f] & 0x000f]) << 24);
				rnd32^=(((uint32)Ciphertext[Ciphertext[Ciphertext[1] & 0x000f] & 0x000f]) << 16);
				rnd32-=(((uint32)Ciphertext[Ciphertext[Ciphertext[3] & 0x000f] & 0x000f]) << 8);
				rnd32+=((uint32)Ciphertext[Ciphertext[Ciphertext[13] & 0x000f] & 0x000f]);

				w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_1 & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_2;
				memory_bias_cnt++;

				rnd64-=rnd32;
				w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_2 & (NUM_OF_KISS_CONTEXTS-1)]);
				w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_1 & (NUM_OF_KISS_CONTEXTS-1)]);
				kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(w64_1-5LL*w64_2) & (NUM_OF_KISSDELEGATES-1)];
				rnd64+=w64_2;
				(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
				memory_bias_cnt++;


				pPMCMED_cipher_ctx->key_setup_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)](&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[i<<4],(void *)&local_cc);
				if ((rnd32>>17) & 0x01) pPMCMED_cipher_ctx->encryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]((void *)&local_cc,&intermediate_hash_result[128],Ciphertext);
				else pPMCMED_cipher_ctx->decryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]((void *)&local_cc,&intermediate_hash_result[129],Ciphertext);;

				w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_2 & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_2;
				memory_bias_cnt++;

				rnd32^=(((uint32)Ciphertext[Ciphertext[Ciphertext[9] & 0x000f] & 0x000f]) << 24);
				rnd32+=(((uint32)Ciphertext[Ciphertext[Ciphertext[0] & 0x000f] & 0x000f]) << 16);
				rnd32-=(((uint32)Ciphertext[Ciphertext[Ciphertext[4] & 0x000f] & 0x000f]) << 8);
				rnd32+=((uint32)Ciphertext[Ciphertext[Ciphertext[7] & 0x000f] & 0x000f]);

				rnd64-=rnd32;
				w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_2 & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
				memory_bias_cnt++;

				w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_1 & (NUM_OF_KISS_CONTEXTS-1)]);
				kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(w64_1-3LL*w64_2) & (NUM_OF_KISSDELEGATES-1)];
				rnd64^=w64_2;
				(*pMemory_bias64)[memory_bias_cnt]^=w64_2;
				memory_bias_cnt++;


				pPMCMED_cipher_ctx->key_setup_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)](&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[i<<4],(void *)&local_cc);
				if ((rnd32>>11) & 0x01) pPMCMED_cipher_ctx->encryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]((void *)&local_cc,&intermediate_hash_result[128],Ciphertext);
				else pPMCMED_cipher_ctx->decryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]((void *)&local_cc,&intermediate_hash_result[129],Ciphertext);;
				rnd32+=(((uint32)Ciphertext[Ciphertext[Ciphertext[11] & 0x000f] & 0x000f]) << 24);
				rnd32^=(((uint32)Ciphertext[Ciphertext[Ciphertext[2] & 0x000f] & 0x000f]) << 16);
				rnd32-=(((uint32)Ciphertext[Ciphertext[Ciphertext[5] & 0x000f] & 0x000f]) << 8);
				rnd32+=((uint32)Ciphertext[Ciphertext[Ciphertext[15] & 0x000f] & 0x000f]);

				(*pMemory_bias64)[memory_bias_cnt]=rnd64;
				memory_bias_cnt++;

				w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[rnd32 & (NUM_OF_KISS_CONTEXTS-1)]);
				w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[rnd64 & (NUM_OF_KISS_CONTEXTS-1)]);
				kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(w64_1-5LL*w64_2) & (NUM_OF_KISSDELEGATES-1)];
				rnd64-=w64_2;
				(*pMemory_bias64)[memory_bias_cnt]^=w64_2;
				memory_bias_cnt++;

				i+=k;
			} // for (i=0;i<NUM_OF_HASH_RESULTS;i++)
		} // for (complexity_count=rnd32 & 0x000f;complexity_count<complexity;complexity_count=complexity_count+((rnd32>>5) & 0x00ff))


		// step 4: now that we have pretty useful hash results, we can initialize our cipher contexts and function pointers
		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES<<7;i++)
		{
			if (memory_bias_cnt+11>=SIZE_OF_MEMORY_BIAS_ARRAY>>3) memory_bias_cnt=0;

				// we sometimes swap function pointers to our hash functions
			if ((rnd32 & 0x07) < NO_OF_HASH_FUNCTION_DELEGATES)
			{
				j=rnd32 & 0x07;
					// swap pointers (introduces almost no pseudorandomness, but this is ok at this stage)
				pHash_Func_Delegate_buf=pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[0];
				pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[0]=pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[j];
				pPMCMED_cipher_ctx->PMCMED_hash_ctx.hash_func_delegates[j]=pHash_Func_Delegate_buf;

				w64_1^=(*pMemory_bias64)[rnd32 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)];
				w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[rnd32 & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_2;
				memory_bias_cnt++;
			}

			pPMCMED_cipher_ctx->key_setup_func_delegates[(rnd32>>15) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)](&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[((i & NUM_OF_HASH_RESULTS-1))<<4],(void *)&local_cc);
			pPMCMED_cipher_ctx->decryption_func_delegates[(rnd32>>15) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]((void *)&local_cc,&intermediate_hash_result[128],Ciphertext);

			w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(w64_2^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
			(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
			memory_bias_cnt++;

			rnd32+=(((uint32)Ciphertext[Ciphertext[Ciphertext[3] & 0x000f] & 0x000f]) << 24);
			rnd32+=(((uint32)Ciphertext[Ciphertext[Ciphertext[7] & 0x000f] & 0x000f]) << 16);
			rnd32-=(((uint32)Ciphertext[Ciphertext[Ciphertext[9] & 0x000f] & 0x000f]) << 8)+(word32)(w64_2^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
			rnd32-=((uint32)Ciphertext[Ciphertext[Ciphertext[14] & 0x000f] & 0x000f]);
			memcpy(&intermediate_hash_result[147+((rnd32>>7)&0x7f)],Ciphertext,16);														// we need to alter intermediate_hash_result[]

			w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[w64_1 & (NUM_OF_KISS_CONTEXTS-1)]);
			(*pMemory_bias64)[memory_bias_cnt]^=w64_2;
			memory_bias_cnt++;

			w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(rnd32^(*pMemory_bias64)[w64_2 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
			w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(rnd64^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
			kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(w64_1-5LL*w64_2) & (NUM_OF_KISSDELEGATES-1)];
			rnd64^=w64_2;
			(*pMemory_bias64)[memory_bias_cnt]^=rnd64;
			memory_bias_cnt++;

			k=pHash_Func_Delegate_buf(intermediate_hash_result,256+(rnd32 & 0x00ff),&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[((i & NUM_OF_HASH_RESULTS-1))<<4])/(27+(rnd32 & 0x3f));

			w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(w64_2^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
			(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
			memory_bias_cnt++;

				// let's further swap pointers to our cipher primitives
			pKey_Setup_Func_delegate_buf=pPMCMED_cipher_ctx->key_setup_func_delegates[0];
			pPMCMED_cipher_ctx->key_setup_func_delegates[0]=pPMCMED_cipher_ctx->key_setup_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
			pPMCMED_cipher_ctx->key_setup_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]=pKey_Setup_Func_delegate_buf;
			pEncryption_Func_delegate_buf=pPMCMED_cipher_ctx->encryption_func_delegates[0];
			pPMCMED_cipher_ctx->encryption_func_delegates[0]=pPMCMED_cipher_ctx->encryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
			pPMCMED_cipher_ctx->encryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]=pEncryption_Func_delegate_buf;
			pDecryption_Func_delegate_buf=pPMCMED_cipher_ctx->decryption_func_delegates[0];
			pPMCMED_cipher_ctx->decryption_func_delegates[0]=pPMCMED_cipher_ctx->decryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
			pPMCMED_cipher_ctx->decryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]=pDecryption_Func_delegate_buf;

			w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(w64_1^(*pMemory_bias64)[w64_2 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
			(*pMemory_bias64)[memory_bias_cnt]^=w64_2;
			memory_bias_cnt++;

				// .. and let's use the freshly exchanged cipher in order to generate more pseudorandomness
			pPMCMED_cipher_ctx->key_setup_func_delegates[((rnd32^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)])>>25) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)](&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[((i & NUM_OF_HASH_RESULTS-1))<<4],(void *)&local_cc);
			if ((rnd32>>31) & 0x01) pPMCMED_cipher_ctx->encryption_func_delegates[((rnd32^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)])>>25) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]((void *)&local_cc,&intermediate_hash_result[128],Ciphertext);
			else pPMCMED_cipher_ctx->decryption_func_delegates[((rnd32^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)])>>25) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]((void *)&local_cc,&intermediate_hash_result[129],Ciphertext);

			w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(w64_2^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
			(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
			memory_bias_cnt++;

			intermediate_hash_result[(rnd32>>21) & 0x01ff]^=0xa5;
			rnd32-=(((uint32)Ciphertext[Ciphertext[(Ciphertext[2]>>4) & 0x000f] & 0x000f]) << 24);
			rnd32-=(((uint32)Ciphertext[Ciphertext[(Ciphertext[6]>>3) & 0x000f] & 0x000f]) << 16)+(word32)(w64_2^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
			rnd32+=(((uint32)Ciphertext[Ciphertext[(Ciphertext[8]>>1) & 0x000f] & 0x000f]) << 8);
			rnd32+=((uint32)Ciphertext[Ciphertext[(Ciphertext[11]>>2) & 0x000f] & 0x000f]);

			w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(w64_2^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
			(*pMemory_bias64)[memory_bias_cnt]^=w64_2;
			memory_bias_cnt++;

			w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(rnd32^(*pMemory_bias64)[w64_2 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
			w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(rnd64^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
			kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(w64_1-3LL*w64_2) & (NUM_OF_KISSDELEGATES-1)];
			rnd64^=w64_2;
			(*pMemory_bias64)[memory_bias_cnt]^=rnd64+rnd32-(w64_2^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
			memory_bias_cnt++;

			rnd32^=(((uint32)Ciphertext[Ciphertext[(Ciphertext[0]>>1) & 0x000f] & 0x000f]) << 24)+(word32)(rnd64^(*pMemory_bias64)[w64_2 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
			rnd32-=(((uint32)Ciphertext[Ciphertext[(Ciphertext[4]>>2) & 0x000f] & 0x000f]) << 16);
			rnd32+=(((uint32)Ciphertext[Ciphertext[(Ciphertext[10]>>3) & 0x000f] & 0x000f]) << 8)+(word32)(rnd32^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
			rnd32-=((uint32)Ciphertext[Ciphertext[(Ciphertext[13]>>2) & 0x000f] & 0x000f]);

			w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(rnd32-(*pMemory_bias64)[rnd32 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
			(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
			memory_bias_cnt++;

			w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[rnd64 & (NUM_OF_KISS_CONTEXTS-1)]);
			kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(w64_1-3*(w64_2^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)])) & (NUM_OF_KISSDELEGATES-1)];
			rnd64^=w64_2;
			(*pMemory_bias64)[memory_bias_cnt]+=(rnd64^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)])+((word64)rnd32)*17LL;
			memory_bias_cnt++;

			i+=k;
		} // for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES<<7;i++)
		memset(&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[NUM_OF_HASH_RESULTS<<4],0x00,48);					// we need to reset the extra bytes as they would allow a little peek into the sequence of compression functions

		
			// step 5: let's now init our scheduler. That function is nothing but one of our cipher primitives that "eats" it's ciphertext again and again in order to generate a pseudorandom number sequence
		pPMCMED_cipher_ctx->key_setup_func_of_scheduler=pPMCMED_cipher_ctx->key_setup_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
		for (i=0;i<16;i++) Ciphertext[i]+=intermediate_hash_result[((rnd32>>9) & 0x0ff)+i];
		pPMCMED_cipher_ctx->key_setup_func_of_scheduler(Ciphertext,pPMCMED_cipher_ctx->crypto_context_of_scheduler);
		pPMCMED_cipher_ctx->encryption_func_of_scheduler=pPMCMED_cipher_ctx->encryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
		pPMCMED_cipher_ctx->decryption_func_of_scheduler=pPMCMED_cipher_ctx->decryption_func_delegates[rnd32 & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
		for (i=0;i<16;i++) pPMCMED_cipher_ctx->plaintext_scheduler[i]=intermediate_hash_result[((rnd32>>17) & 0x0ff)+i];

		if (rnd32 & 0x0001) pPMCMED_cipher_ctx->encryption_func_of_scheduler(pPMCMED_cipher_ctx->crypto_context_of_scheduler,pPMCMED_cipher_ctx->plaintext_scheduler,pPMCMED_cipher_ctx->ciphertext_scheduler);
		else pPMCMED_cipher_ctx->decryption_func_of_scheduler(pPMCMED_cipher_ctx->crypto_context_of_scheduler,pPMCMED_cipher_ctx->plaintext_scheduler,pPMCMED_cipher_ctx->ciphertext_scheduler);

		memcpy(pPMCMED_cipher_ctx->plaintext_scheduler,pPMCMED_cipher_ctx->ciphertext_scheduler,16);				// from now on will the scheduler be able to generate a pseudorandom number sequence

		rnd32-=(((uint32)pPMCMED_cipher_ctx->plaintext_scheduler[pPMCMED_cipher_ctx->ciphertext_scheduler[(Ciphertext[1]>>4) & 0x000f] & 0x000f]) << 24)+(word32)(rnd64^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
		rnd32-=(((uint32)pPMCMED_cipher_ctx->plaintext_scheduler[pPMCMED_cipher_ctx->ciphertext_scheduler[(Ciphertext[5]>>3) & 0x000f] & 0x000f]) << 16)+(word32)(w64_1^(*pMemory_bias64)[w64_2 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
		rnd32+=(((uint32)pPMCMED_cipher_ctx->plaintext_scheduler[pPMCMED_cipher_ctx->ciphertext_scheduler[(Ciphertext[9]>>2) & 0x000f] & 0x000f]) << 8)+(word32)(w64_2^(*pMemory_bias64)[rnd32 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
		rnd32+=((uint32)pPMCMED_cipher_ctx->plaintext_scheduler[pPMCMED_cipher_ctx->ciphertext_scheduler[(Ciphertext[13]>>3) & 0x000f] & 0x000f])^(word32)(rnd64+(*pMemory_bias64)[rnd32 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
		
	
			// step 6: now we can finally create all cipher contexts
		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES;i++)
		{
			if (memory_bias_cnt+5>=SIZE_OF_MEMORY_BIAS_ARRAY>>3) memory_bias_cnt=0;
			for (j=0;j<NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE;j++)
			{
				pCc=&pPMCMED_cipher_ctx->pcc[((i*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES];
				pPMCMED_cipher_ctx->key_setup_func_delegates[i](&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[((rnd32 & NUM_OF_HASH_RESULTS-1))<<4],(void *)pCc);

				w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(w64_2^(*pMemory_bias64)[rnd32 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
				memory_bias_cnt++;

					// we should alter the hash result so that next time the same one is used, we don't create the same result
				pUi32=(uint32 *)&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[((rnd32 & NUM_OF_HASH_RESULTS-1))<<4];
				*pUi32^=rnd32;

				w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(w64_1^(*pMemory_bias64)[w64_2 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_2;
				memory_bias_cnt++;

				pPMCMED_cipher_ctx->encryption_func_of_scheduler(pPMCMED_cipher_ctx->crypto_context_of_scheduler,pPMCMED_cipher_ctx->plaintext_scheduler,pPMCMED_cipher_ctx->ciphertext_scheduler);
				memcpy(pPMCMED_cipher_ctx->plaintext_scheduler,pPMCMED_cipher_ctx->ciphertext_scheduler,16);				// from now on will the scheduler be able to generate a pseudorandom number sequence
				rnd32+=(((uint32)pPMCMED_cipher_ctx->ciphertext_scheduler[pPMCMED_cipher_ctx->plaintext_scheduler[(pPMCMED_cipher_ctx->ciphertext_scheduler[1]>>4) & 0x000f] & 0x000f]) << 24)+(word32)(rnd64^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
				rnd32+=(((uint32)pPMCMED_cipher_ctx->ciphertext_scheduler[pPMCMED_cipher_ctx->plaintext_scheduler[(pPMCMED_cipher_ctx->ciphertext_scheduler[5]>>3) & 0x000f] & 0x000f]) << 16)+(word32)(w64_1^(*pMemory_bias64)[rnd32 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
				rnd32+=(((uint32)pPMCMED_cipher_ctx->ciphertext_scheduler[pPMCMED_cipher_ctx->plaintext_scheduler[(pPMCMED_cipher_ctx->ciphertext_scheduler[9]>>2) & 0x000f] & 0x000f]) << 8)+(word32)(rnd32^(*pMemory_bias64)[w64_2 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
				rnd32+=((uint32)pPMCMED_cipher_ctx->ciphertext_scheduler[pPMCMED_cipher_ctx->plaintext_scheduler[(pPMCMED_cipher_ctx->ciphertext_scheduler[13]>>3) & 0x000f] & 0x000f])+(word32)((rnd64+w64_1)^(*pMemory_bias64)[rnd32 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);

				k=memory_bias_cnt-(i*j)-1;
				if (k<0) k=memory_bias_cnt-j-1;

				w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(w64_2^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
				memory_bias_cnt++;

				rnd64+=(*pMemory_bias64)[k];																																		// let's use our history from now on
				w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(w64_2^(*pMemory_bias64)[rnd64 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]) & (NUM_OF_KISS_CONTEXTS-1)]);
				(*pMemory_bias64)[memory_bias_cnt]^=w64_1;
				memory_bias_cnt++;

				w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[rnd32 & (NUM_OF_KISS_CONTEXTS-1)]);
				kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(w64_1-17*w64_2) & (NUM_OF_KISSDELEGATES-1)];
				rnd64^=w64_2;
				(*pMemory_bias64)[memory_bias_cnt]-=rnd64+((word64)rnd32)*17LL;
				memory_bias_cnt++;
			} // for (j=0;j<NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE;j++)
		} // for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES;i++)

			// let's now generate the IV for CBC operations
		pPMCMED_cipher_ctx->encryption_func_of_scheduler(pPMCMED_cipher_ctx->crypto_context_of_scheduler,pPMCMED_cipher_ctx->plaintext_scheduler,pPMCMED_cipher_ctx->ciphertext_scheduler);
		pPMCMED_cipher_ctx->key_setup_func_delegates[(rnd32>>23) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)](&pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results[(((rnd32>>7) & NUM_OF_HASH_RESULTS-1))<<4],(void *)&local_cc);
		pPMCMED_cipher_ctx->encryption_func_delegates[(rnd32>>23) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]((void *)&local_cc,pPMCMED_cipher_ctx->ciphertext_scheduler,pPMCMED_cipher_ctx->IV_for_CBC);
		
		

			// step 7: fill (most of) the memory bias array with pseudorandomness
		rnd32^=((word32)pPMCMED_cipher_ctx->IV_for_CBC[0]*(word32)pPMCMED_cipher_ctx->IV_for_CBC[1])^(word32)pPMCMED_cipher_ctx->IV_for_CBC[3]+(word32)(rnd64^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
		rnd64^=((word64)pPMCMED_cipher_ctx->IV_for_CBC[5]*(word32)pPMCMED_cipher_ctx->IV_for_CBC[7])^(word32)pPMCMED_cipher_ctx->IV_for_CBC[11]+(w64_2^(*pMemory_bias64)[rnd32 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)]);
		k=memory_bias_cnt-1;
		j=k;
		while (memory_bias_cnt<((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-4))
		{
			(*pMemory_bias64)[memory_bias_cnt]^=w64_1^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[((rnd64^(*pMemory_bias64)[w64_1 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)])>>2) & (NUM_OF_KISS_CONTEXTS-1)]);
			memory_bias_cnt++;
			(*pMemory_bias64)[memory_bias_cnt]^=w64_2^=kiss_func(pPMCMED_cipher_ctx->initial_ctx_array[(*pMemory_bias64)[j] & (NUM_OF_KISS_CONTEXTS-1)]);// let's use our history extensively
			memory_bias_cnt++;
			j--;
			if (j<0) j=k;
			if (((memory_bias_cnt+(rnd64 & 0x007f)) & 0xff)==0) kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(w64_1-3*w64_2) & (NUM_OF_KISSDELEGATES-1)];// let's only change the PRNG function from time to time
			rnd64^=w64_2+(*pMemory_bias64)[j]+(w64_1*7LL);
			(*pMemory_bias64)[memory_bias_cnt]^=rnd64^(5LL*w64_1);
			memory_bias_cnt++;
			(*pMemory_bias64)[memory_bias_cnt]^=((*pMemory_bias64)[j]^w64_1)+(w64_2>>5)-(w64_2<<49);
			memory_bias_cnt++;
		}

			// we can already delete the hash results
		if (pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results)
		{
			memset(pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results,0xaa,16*(NUM_OF_HASH_RESULTS+4));
			delete pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results;
			pPMCMED_cipher_ctx->PMCMED_hash_ctx.pHash_results=NULL;
		}

		memcpy(pPMCMED_cipher_ctx->initial_plaintext_scheduler,pPMCMED_cipher_ctx->plaintext_scheduler,16);

		return retval;
	} // int PMCMED_keysetup(uint8 * pKey,void * pPMCMED_cc,int key_length_in_bits)



		// PMCMED_free_memory returns 0 if memory could be freed
		// In case of an error: -1
	int PMCMED_free_memory(void * pPMCMED_cc)
	{
		int															i;
		int															retval=0;																														// 0: no error, any other value indicates an error
		struct PMCMED_cipher_context	*	pPMCMED_cipher_ctx;

		if (!pPMCMED_cc) return -1;																																					// Error
		pPMCMED_cipher_ctx=(struct PMCMED_cipher_context *)pPMCMED_cc;
		
		for (i=0;i<NUM_OF_KISS_CONTEXTS;i++)
		{
			if (pPMCMED_cipher_ctx->initial_ctx_array[i])
			{
				delete pPMCMED_cipher_ctx->initial_ctx_array[i];
				pPMCMED_cipher_ctx->initial_ctx_array[i]=NULL;
			}
		}

		if (pPMCMED_cipher_ctx->memory_bias)
		{
			delete pPMCMED_cipher_ctx->memory_bias;
			pPMCMED_cipher_ctx->memory_bias=NULL;
		}
		else return -1;																																											// Error

		if (pPMCMED_cipher_ctx->pcc)
		{
			delete pPMCMED_cipher_ctx->pcc;
			pPMCMED_cipher_ctx->pcc=NULL;
		}
		else return -1;																																											// Error

		return retval;
	}



		// encryption / decryption with 8! = 8*7*6*5*4*3*2*1 = 40320 possibilities (each base cipher is used exactly once)
	void PMCMED_encrypt_cascade(void * pPMCMED_cc,uint8 * pPlaintext,uint8 * pCiphertext)									// encryption in ECB mode: all 8 cipher primitives are used to encrypt a single block of data
	{
		int															i,j;
		struct PMCMED_cipher_context	*	pPMCMED_cipher_ctx;
		KISSDELEGATE									*	kiss_func=NULL;
		gbpmc_kiss_context							kiss_ctx;
		word64													w64;


		pPMCMED_cipher_ctx=(struct PMCMED_cipher_context *)pPMCMED_cc;

		j=pPMCMED_cipher_ctx->ciphertext_scheduler[0] & (NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE-1);			// let's select a certain set of contexts
		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES;i++)
		{
			pPMCMED_cipher_ctx->encryption_func_delegates[i]((void *)&pPMCMED_cipher_ctx->pcc[((i*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pPlaintext,pCiphertext);
			i++;
			pPMCMED_cipher_ctx->encryption_func_delegates[i]((void *)&pPMCMED_cipher_ctx->pcc[((i*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pCiphertext,pPlaintext);
			if (i==3) {
				w64=((word64*)pPlaintext)[1];
				w64^=((word64*)pPMCMED_cipher_ctx->memory_bias)[w64 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)];

				kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[w64 & (NUM_OF_KISSDELEGATES-1)];
				kiss_ctx=*(pPMCMED_cipher_ctx->initial_ctx_array[(w64>>30) & (NUM_OF_KISS_CONTEXTS-1)]);
				(*((word64*)pPlaintext))^=kiss_func(&kiss_ctx);
			}
		}
		memcpy(pCiphertext,pPlaintext,16);																																	// we need to invest a little more time into copying for the last one
		memset(pPlaintext,0xaa,16);																																					// let's disguise our last intermediate result
	}

	void PMCMED_decrypt_cascade(void * pPMCMED_cc,uint8 * pCiphertext,uint8 * pPlaintext)									// decryption in ECB mode: all 8 cipher primitives are used in reverse order to decrypt a single block of data
	{
		int															i,j;
		struct PMCMED_cipher_context	*	pPMCMED_cipher_ctx;
		KISSDELEGATE									*	kiss_func=NULL;
		gbpmc_kiss_context							kiss_ctx;
		word64													w64;

		pPMCMED_cipher_ctx=(struct PMCMED_cipher_context *)pPMCMED_cc;
		kiss_ctx=*(pPMCMED_cipher_ctx->initial_ctx_array[0]);

		j=pPMCMED_cipher_ctx->ciphertext_scheduler[0] & (NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE-1);			// let's select a certain set of contexts

		for (i=NUM_OF_CIPHER_FUNCTION_DELEGATES-1;i>=0;i--)
		{
			if (i==3) {
				w64=((word64*)pCiphertext)[1];
				w64^=((word64*)pPMCMED_cipher_ctx->memory_bias)[w64 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)];

				kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[w64 & (NUM_OF_KISSDELEGATES-1)];
				kiss_ctx=*(pPMCMED_cipher_ctx->initial_ctx_array[(w64>>30) & (NUM_OF_KISS_CONTEXTS-1)]);
				(*((word64*)pCiphertext))^=kiss_func(&kiss_ctx);
			}
			pPMCMED_cipher_ctx->decryption_func_delegates[i]((void *)&pPMCMED_cipher_ctx->pcc[((i*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pCiphertext,pPlaintext);
			i--;
			pPMCMED_cipher_ctx->decryption_func_delegates[i]((void *)&pPMCMED_cipher_ctx->pcc[((i*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pPlaintext,pCiphertext);
		}
		memcpy(pPlaintext,pCiphertext,16);																																	// we need to invest a little more time into copying for the last one
		memset(pCiphertext,0xaa,16);																																				// let's disguise our last intermediate result
	}



		// encryption / decryption with 8^8 = 2^24 = 16777216 possibilities (free choice of the base ciphers)
	void PMCMED_encrypt_max_var_cascade(void * pPMCMED_cc,uint8 * pPlaintext,uint8 * pCiphertext)					// encryption in ECB mode: all 8 cipher primitives are used to encrypt a single block of data
	{
		int															i,j;
		struct PMCMED_cipher_context	*	pPMCMED_cipher_ctx;
		int															index_arr[NUM_OF_CIPHER_FUNCTION_DELEGATES];
		KISSDELEGATE									*	kiss_func=NULL;
		gbpmc_kiss_context							kiss_ctx;
		word64													w64;

		pPMCMED_cipher_ctx=(struct PMCMED_cipher_context *)pPMCMED_cc;

		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES;i++) index_arr[i]=(pPMCMED_cipher_ctx->ciphertext_scheduler[i]>>4) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1);

		j=pPMCMED_cipher_ctx->ciphertext_scheduler[pPMCMED_cipher_ctx->ciphertext_scheduler[0] & 0x0f] & (NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE-1);			// let's select a certain set of contexts

		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES;i++)
		{
			pPMCMED_cipher_ctx->encryption_func_delegates[index_arr[i]]((void *)&pPMCMED_cipher_ctx->pcc[((index_arr[i]*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pPlaintext,pCiphertext);
			i++;
			pPMCMED_cipher_ctx->encryption_func_delegates[index_arr[i]]((void *)&pPMCMED_cipher_ctx->pcc[((index_arr[i]*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pCiphertext,pPlaintext);
			if (i==3) {
				w64=((word64*)pPlaintext)[1];
				w64^=((word64*)pPMCMED_cipher_ctx->memory_bias)[w64 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)];

				kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[w64 & (NUM_OF_KISSDELEGATES-1)];
				kiss_ctx=*(pPMCMED_cipher_ctx->initial_ctx_array[(w64>>30) & (NUM_OF_KISS_CONTEXTS-1)]);
				(*((word64*)pPlaintext))^=kiss_func(&kiss_ctx);
			}
		}
		memcpy(pCiphertext,pPlaintext,16);																																	// we need to invest a little more time into copying for the last one
		memset(pPlaintext,0xaa,16);																																					// let's disguise our last intermediate result
	}

	void PMCMED_decrypt_max_var_cascade(void * pPMCMED_cc,uint8 * pCiphertext,uint8 * pPlaintext)					// decryption in ECB mode: all 8 cipher primitives are used in reverse order to decrypt a single block of data
	{
		int															i,j;
		struct PMCMED_cipher_context	*	pPMCMED_cipher_ctx;
		int															index_arr[NUM_OF_CIPHER_FUNCTION_DELEGATES];
		KISSDELEGATE									*	kiss_func=NULL;
		gbpmc_kiss_context							kiss_ctx;
		word64													w64;

		pPMCMED_cipher_ctx=(struct PMCMED_cipher_context *)pPMCMED_cc;

		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES;i++) index_arr[i]=(pPMCMED_cipher_ctx->ciphertext_scheduler[i]>>4) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1);

		j=pPMCMED_cipher_ctx->ciphertext_scheduler[pPMCMED_cipher_ctx->ciphertext_scheduler[0] & 0x0f] & (NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE-1);			// let's select a certain set of contexts
		for (i=NUM_OF_CIPHER_FUNCTION_DELEGATES-1;i>=0;i--)
		{
			if (i==3) {
				w64=((word64*)pCiphertext)[1];
				w64^=((word64*)pPMCMED_cipher_ctx->memory_bias)[w64 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)];

				kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[w64 & (NUM_OF_KISSDELEGATES-1)];
				kiss_ctx=*(pPMCMED_cipher_ctx->initial_ctx_array[(w64>>30) & (NUM_OF_KISS_CONTEXTS-1)]);
				(*((word64*)pCiphertext))^=kiss_func(&kiss_ctx);
			}
			pPMCMED_cipher_ctx->decryption_func_delegates[index_arr[i]]((void *)&pPMCMED_cipher_ctx->pcc[((index_arr[i]*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pCiphertext,pPlaintext);
			i--;
			pPMCMED_cipher_ctx->decryption_func_delegates[index_arr[i]]((void *)&pPMCMED_cipher_ctx->pcc[((index_arr[i]*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pPlaintext,pCiphertext);
		}
		memcpy(pPlaintext,pCiphertext,16);																																	// we need to invest a little more time into copying for the last one
		memset(pCiphertext,0xaa,16);																																				// let's disguise our last intermediate result
	}




	void PMCMED_init_CBC_mode(void * pPMCMED_cc,word64 CBC_block_counter_start_value=0LL)
	{
		struct PMCMED_cipher_context	*	pPMCMED_cipher_ctx;

		if (!pPMCMED_cc) return;																																						// Error
		pPMCMED_cipher_ctx=(struct PMCMED_cipher_context *)pPMCMED_cc;

		pPMCMED_cipher_ctx->CBC_block_counter=CBC_block_counter_start_value;																// the cipher allows to set a start value, which might be practical
		memcpy(pPMCMED_cipher_ctx->last_block_CBC,pPMCMED_cipher_ctx->IV_for_CBC,16);												// set IV
	}

	void PMCMED_CBC_encrypt_cascade(void * pPMCMED_cc,uint8 * pPlaintext,uint8 * pCiphertext)
	{
		struct PMCMED_cipher_context	*	pPMCMED_cipher_ctx;
		word64													w64buf;
		int															i,j;
		int															index_arr[NUM_OF_CIPHER_FUNCTION_DELEGATES];
		KISSDELEGATE									*	kiss_func=NULL;
		gbpmc_kiss_context							kiss_ctx;
		word64													w64;

		if (!pPMCMED_cc) return;																																						// Error
		pPMCMED_cipher_ctx=(struct PMCMED_cipher_context *)pPMCMED_cc;

		w64buf=pPMCMED_cipher_ctx->CBC_block_counter;
		pPMCMED_cipher_ctx->CBC_block_counter++;																														// each block is different. That's what the counter is good for.

		memcpy(pPMCMED_cipher_ctx->plaintext_scheduler,pPMCMED_cipher_ctx->initial_plaintext_scheduler,16);
		pPMCMED_cipher_ctx->plaintext_scheduler[7]^=(uint8)(w64buf & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[1]^=(uint8)((w64buf>>8) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[5]^=(uint8)((w64buf>>16) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[4]^=(uint8)((w64buf>>24) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[3]^=(uint8)((w64buf>>32) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[2]^=(uint8)((w64buf>>40) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[6]^=(uint8)((w64buf>>48) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[0]^=(uint8)((w64buf>>56) & 0xff);

			// let's now generate 128 bit that are impossible to guess. We'll derive from that data the sequence of the ciphers
		pPMCMED_cipher_ctx->encryption_func_of_scheduler(pPMCMED_cipher_ctx->crypto_context_of_scheduler,pPMCMED_cipher_ctx->plaintext_scheduler,pPMCMED_cipher_ctx->ciphertext_scheduler);

		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES;i++) index_arr[i]=(pPMCMED_cipher_ctx->ciphertext_scheduler[i]>>4) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1);
			// let's exchange the sequence now
		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES+(pPMCMED_cipher_ctx->ciphertext_scheduler[0] & 0x0007);i++) 
		{
			j=index_arr[0];
			index_arr[0]=index_arr[(i+pPMCMED_cipher_ctx->ciphertext_scheduler[i & 0x0f]) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
			index_arr[(i+pPMCMED_cipher_ctx->ciphertext_scheduler[i & 0x0f]) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]=j;
		}

		j=pPMCMED_cipher_ctx->ciphertext_scheduler[0] & (NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE-1);			// let's select a certain set of contexts

			// perform CBC now
		for (i=0;i<16;i++) pPlaintext[i]^=pPMCMED_cipher_ctx->last_block_CBC[i];

			// encrypt with unknown sequence of ciphers
		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES;i++)
		{
			pPMCMED_cipher_ctx->encryption_func_delegates[index_arr[i]]((void *)&pPMCMED_cipher_ctx->pcc[((index_arr[i]*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pPlaintext,pCiphertext);
			i++;
			pPMCMED_cipher_ctx->encryption_func_delegates[index_arr[i]]((void *)&pPMCMED_cipher_ctx->pcc[((index_arr[i]*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pCiphertext,pPlaintext);
			if (i==3) {
				w64=((word64*)pPlaintext)[1];
				w64^=((word64*)pPMCMED_cipher_ctx->memory_bias)[w64 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)];

				kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(w64+index_arr[4]) & (NUM_OF_KISSDELEGATES-1)];
				kiss_ctx=*(pPMCMED_cipher_ctx->initial_ctx_array[((w64>>30)+index_arr[2]) & (NUM_OF_KISS_CONTEXTS-1)]);
				(*((word64*)pPlaintext))^=kiss_func(&kiss_ctx);
			}
		}
		memcpy(pCiphertext,pPlaintext,16);																																	// we need to invest a little more time into copying for the last one

		memcpy(pPMCMED_cipher_ctx->last_block_CBC,pCiphertext,16);																					// preparation for CBC for next block

		memset(pPlaintext,0xaa,16);																																					// let's disguise our last intermediate result
	}

	void PMCMED_CBC_decrypt_cascade(void * pPMCMED_cc,uint8 * pCiphertext,uint8 * pPlaintext)
	{
		struct PMCMED_cipher_context	*	pPMCMED_cipher_ctx;
		word64													w64buf;
		int															i,j;
		int															index_arr[NUM_OF_CIPHER_FUNCTION_DELEGATES];
		uint8														block_buffer[16];
		KISSDELEGATE									*	kiss_func=NULL;
		gbpmc_kiss_context							kiss_ctx;
		word64													w64;

		if (!pPMCMED_cc) return;																																						// Error
		pPMCMED_cipher_ctx=(struct PMCMED_cipher_context *)pPMCMED_cc;

		w64buf=pPMCMED_cipher_ctx->CBC_block_counter;
		pPMCMED_cipher_ctx->CBC_block_counter++;																														// each block is different. That's what the counter is good for.

		memcpy(pPMCMED_cipher_ctx->plaintext_scheduler,pPMCMED_cipher_ctx->initial_plaintext_scheduler,16);
		pPMCMED_cipher_ctx->plaintext_scheduler[7]^=(uint8)(w64buf & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[1]^=(uint8)((w64buf>>8) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[5]^=(uint8)((w64buf>>16) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[4]^=(uint8)((w64buf>>24) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[3]^=(uint8)((w64buf>>32) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[2]^=(uint8)((w64buf>>40) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[6]^=(uint8)((w64buf>>48) & 0xff);
		pPMCMED_cipher_ctx->plaintext_scheduler[0]^=(uint8)((w64buf>>56) & 0xff);

			// let's now generate 128 bit that are impossible to guess. We'll derive from that data the sequence of the ciphers
		pPMCMED_cipher_ctx->encryption_func_of_scheduler(pPMCMED_cipher_ctx->crypto_context_of_scheduler,pPMCMED_cipher_ctx->plaintext_scheduler,pPMCMED_cipher_ctx->ciphertext_scheduler);

		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES;i++) index_arr[i]=(pPMCMED_cipher_ctx->ciphertext_scheduler[i]>>4) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1);
			// let's exchange the sequence now
		for (i=0;i<NUM_OF_CIPHER_FUNCTION_DELEGATES+(pPMCMED_cipher_ctx->ciphertext_scheduler[0] & 0x0007);i++) 
		{
			j=index_arr[0];
			index_arr[0]=index_arr[(i+pPMCMED_cipher_ctx->ciphertext_scheduler[i & 0x0f]) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)];
			index_arr[(i+pPMCMED_cipher_ctx->ciphertext_scheduler[i & 0x0f]) & (NUM_OF_CIPHER_FUNCTION_DELEGATES-1)]=j;
		}

		j=pPMCMED_cipher_ctx->ciphertext_scheduler[0] & (NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE-1);			// let's select a certain set of contexts

		memcpy(block_buffer,pCiphertext,16);																																// CBC

			// decrypt with unknown sequence of ciphers
		for (i=NUM_OF_CIPHER_FUNCTION_DELEGATES-1;i>=0;i--)
		{
			if (i==3) {
				w64=((word64*)pCiphertext)[1];
				w64^=((word64*)pPMCMED_cipher_ctx->memory_bias)[w64 & ((SIZE_OF_MEMORY_BIAS_ARRAY>>3)-1)];

				kiss_func=pPMCMED_cipher_ctx->gbpmc_kiss[(w64+index_arr[4]) & (NUM_OF_KISSDELEGATES-1)];
				kiss_ctx=*(pPMCMED_cipher_ctx->initial_ctx_array[((w64>>30)+index_arr[2]) & (NUM_OF_KISS_CONTEXTS-1)]);
				(*((word64*)pCiphertext))^=kiss_func(&kiss_ctx);
			}
			pPMCMED_cipher_ctx->decryption_func_delegates[index_arr[i]]((void *)&pPMCMED_cipher_ctx->pcc[((index_arr[i]*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pCiphertext,pPlaintext);
			i--;
			pPMCMED_cipher_ctx->decryption_func_delegates[index_arr[i]]((void *)&pPMCMED_cipher_ctx->pcc[((index_arr[i]*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE)+j)*MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES],pPlaintext,pCiphertext);
		}
		memcpy(pPlaintext,pCiphertext,16);																																	// we need to invest a little more time into copying for the last one

			// perform CBC now
		for (i=0;i<16;i++) pPlaintext[i]^=pPMCMED_cipher_ctx->last_block_CBC[i];
		memcpy(pPMCMED_cipher_ctx->last_block_CBC,block_buffer,16);																					// preparation for CBC for next block

		memset(pCiphertext,0xaa,16);																																				// let's disguise our last intermediate result
	}





		// those are the standard test vectors
	static unsigned char PMC_MED_KEYs[3][17] =
	{
			"\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
			"\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
			"\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e"
	};

	static unsigned char PMC_MED_PTs[3][17] =
	{
			"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
			"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
			"\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e"
	};

	static unsigned char PMC_MED_CTs[3][17] =
	{
		"\x6a\xe5\x59\x5b\x31\x9e\x2e\x30\x2d\x5c\x7b\x90\x50\xd7\x77\xbb",
		"\x62\x8a\xf4\x3c\x27\xbb\x5e\x14\x1e\x26\x73\x47\x7e\x7b\x1e\x4b",
		"\xc3\x89\x5d\xf9\x8c\x1c\x70\x37\x0e\xd5\x4f\x91\x2c\x4a\x7b\x6d"
	};

	static unsigned char PMC_MED_CTs_max_var[3][17] =
	{
		"\x0b\xa5\x8d\x99\xc8\x75\x16\x2f\x94\x1c\x39\x45\x1a\xb9\xe2\x9b",
		"\x9e\x82\x78\x1d\xb5\x32\xcc\xca\xfb\xb0\xd9\x7c\x85\x96\x30\xb5",
		"\x08\xc9\x5f\xa7\xf4\x15\x51\xef\x87\xa7\x2f\x87\x38\x13\x77\x77"
	};

	static unsigned char PMC_MED_CTs_3times_CBC[3][17] =
	{
		"\x9a\xf9\x1f\x8d\x6a\x2a\x6d\xec\x4e\x73\x99\x6c\x63\x26\x22\x0b",
		"\x1b\x71\xdb\x9b\x7f\x7c\xbf\x78\x40\xef\x72\x55\x4e\xb4\xac\x61",
		"\x78\xb3\x25\x8a\x93\xce\x68\xe7\xe8\x59\xe4\x30\xc4\x6d\xf4\xf3"
	};

	bool test_PMC_Medley(void)																																						// tests our cipher and returns true if everything is ok
	{
		bool result=true; // false=fail, true=pass
		int i;
		unsigned char data[16];
		unsigned char ciphertext[16];
		unsigned char CBC_ciphertext[3][16];
		PMCMED_cipher_context		pmc_med_cc;

			// test vectors for 3 iterations of cascades with all 8 cipher primitives being used
		for( i = 0; i < 3; i++ )
		{
			memcpy( data, PMC_MED_PTs[i], 16 );
				
			if (PMCMED_keysetup(PMC_MED_KEYs[i],(void *)&pmc_med_cc,128,i*1024)) return false;								// key_length_in_bits: 128 .. 1024 bit, complexity: 0: very fast, 2^16-1: huge set of keyed operations to compute internal state
			PMCMED_encrypt_cascade((void *)&pmc_med_cc,data,ciphertext);																			// encryption in ECB mode: all 8 cipher primitives are used to encrypt a single block of data
				
			if (memcmp( ciphertext, PMC_MED_CTs[i], 16 ) )
			{
				return false;
			}
			PMCMED_decrypt_cascade((void *)&pmc_med_cc,ciphertext,data);																			// encryption in ECB mode: all 8 cipher primitives are used to encrypt a single block of data
			if (memcmp( data, PMC_MED_PTs[i], 16 ) )
			{
				return false;
			}
			PMCMED_free_memory((void *)&pmc_med_cc);
		}

		for( i = 0; i < 3; i++ )
		{
			memcpy( ciphertext, PMC_MED_CTs[i], 16 );
				
			if (PMCMED_keysetup(PMC_MED_KEYs[i],(void *)&pmc_med_cc,128,i*1024)) return false;								// key_length_in_bits: 128 .. 1024 bit, complexity: 0: very fast, 2^16-1: huge set of keyed operations to compute internal state
			PMCMED_decrypt_cascade((void *)&pmc_med_cc,ciphertext,data);																			// decryption in ECB mode: all 8 cipher primitives are used to encrypt a single block of data
				
			if (memcmp( data, PMC_MED_PTs[i], 16 ) )
			{
				return false;
			}
			PMCMED_free_memory((void *)&pmc_med_cc);
		}






			// test vectors for 3 iterations of cascades with maximum variability in ECB mode
		for( i = 0; i < 3; i++ )
		{
			memcpy( data, PMC_MED_PTs[i], 16 );
				
			if (PMCMED_keysetup(PMC_MED_KEYs[i],(void *)&pmc_med_cc,128,i*1024)) return false;								// key_length_in_bits: 128 .. 1024 bit, complexity: 0: very fast, 2^16-1: huge set of keyed operations to compute internal state
			PMCMED_encrypt_max_var_cascade((void *)&pmc_med_cc,data,ciphertext);															// encryption in ECB mode:  free choice of the base ciphers => 2^24 possible combinations
				
			if (memcmp( ciphertext, PMC_MED_CTs_max_var[i], 16 ) )
			{
				return false;
			}
			PMCMED_decrypt_max_var_cascade((void *)&pmc_med_cc,ciphertext,data);															// encryption in ECB mode: free choice of the base ciphers => 2^24 possible combinations
			if (memcmp( data, PMC_MED_PTs[i], 16 ) )
			{
				return false;
			}
			PMCMED_free_memory((void *)&pmc_med_cc);
		}

		for( i = 0; i < 3; i++ )
		{
			memcpy( ciphertext, PMC_MED_CTs_max_var[i], 16 );
				
			if (PMCMED_keysetup(PMC_MED_KEYs[i],(void *)&pmc_med_cc,128,i*1024)) return false;								// key_length_in_bits: 128 .. 1024 bit, complexity: 0: very fast, 2^16-1: huge set of keyed operations to compute internal state
			PMCMED_decrypt_max_var_cascade((void *)&pmc_med_cc,ciphertext,data);															// decryption in ECB mode:  free choice of the base ciphers => 2^24 possible combinations
				
			if (memcmp( data, PMC_MED_PTs[i], 16 ) )
			{
				return false;
			}
			PMCMED_free_memory((void *)&pmc_med_cc);
		}





			// test vectors for 3 iterations in CBC mode
		if (PMCMED_keysetup(PMC_MED_KEYs[2],(void *)&pmc_med_cc,128,3*1024)) return false;								// key_length_in_bits: 128 .. 1024 bit, complexity: 0: very fast, 2^16-1: huge set of keyed operations to compute internal state
		
		PMCMED_init_CBC_mode((void *)&pmc_med_cc,127);																										// 127: CBC_block_counter_start_value must be the same for decryption. It is good e.g. for encrypting data streams and it is automatically incremented by PMCMED_CBC_encrypt_cascade()
		memcpy( data, PMC_MED_PTs[0], 16 );
		PMCMED_CBC_encrypt_cascade((void *)&pmc_med_cc,data,CBC_ciphertext[0]);														// encryption in CBC mode: all 8 cipher primitives are used to encrypt a single block of data
		if (memcmp( CBC_ciphertext[0], PMC_MED_CTs_3times_CBC[0], 16 ) )
		{
			return false;
		}
		memcpy( data, PMC_MED_PTs[1], 16 );
		PMCMED_CBC_encrypt_cascade((void *)&pmc_med_cc,data,CBC_ciphertext[1]);
		if (memcmp( CBC_ciphertext[1], PMC_MED_CTs_3times_CBC[1], 16 ) )
		{
			return false;
		}
		memcpy( data, PMC_MED_PTs[2], 16 );
		PMCMED_CBC_encrypt_cascade((void *)&pmc_med_cc,data,CBC_ciphertext[2]);
		if (memcmp( CBC_ciphertext[2], PMC_MED_CTs_3times_CBC[2], 16 ) )
		{
			return false;
		}


		PMCMED_init_CBC_mode((void *)&pmc_med_cc,127);																										// 127: CBC_block_counter_start_value must be the same for decryption.  It is good e.g. for decrypting data streams and it is automatically incremented by PMCMED_CBC_decrypt_cascade()

		PMCMED_CBC_decrypt_cascade((void *)&pmc_med_cc,CBC_ciphertext[0],data);														// decryption in CBC mode: all 8 cipher primitives are used to encrypt a single block of data
		if (memcmp( data, PMC_MED_PTs[0], 16 ) )
		{
			return false;
		}
		PMCMED_CBC_decrypt_cascade((void *)&pmc_med_cc,CBC_ciphertext[1],data);														// decryption in CBC mode: all 8 cipher primitives are used to encrypt a single block of data
		if (memcmp( data, PMC_MED_PTs[1], 16 ) )
		{
			return false;
		}
		PMCMED_CBC_decrypt_cascade((void *)&pmc_med_cc,CBC_ciphertext[2],data);														// decryption in CBC mode: all 8 cipher primitives are used to encrypt a single block of data
		if (memcmp( data, PMC_MED_PTs[2], 16 ) )
		{
			return false;
		}

		PMCMED_free_memory((void *)&pmc_med_cc);

		return result;
	}


		// -----------------------------------------------------------------------------------------------------------------------------------------------
		// wrappers for our hash functions

	uint32 HASH_FUNCTION_Whirlpool512(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum)								// retval is the number of bytes of the sum
	{
		int retval=64;																																											// 512 bit=64 bytes
		struct crypto_primitives::NESSIEstruct w;
			
		crypto_primitives::NESSIEinit(&w);
		crypto_primitives::NESSIEadd((const unsigned char *) pInputBytes,InputLength<<3,&w);
		crypto_primitives::NESSIEfinalize(&w,pSum);
		memset(&w,0x00,sizeof(w));

		return retval;
	}

	uint32 HASH_FUNCTION_Haval256(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum)										// retval is the number of bytes of the sum
	{
		int retval=32;																																											// 256 bit=32 bytes
		crypto_primitives::havalContext ctx;
			
		crypto_primitives::havalInit256(&ctx);
		crypto_primitives::havalUpdate(&ctx,pInputBytes,(size_t)InputLength);
		crypto_primitives::havalFinal(&ctx,pSum);
		memset(&ctx,0x00,sizeof(ctx));

		return retval;
	}

	uint32 HASH_FUNCTION_Tiger192(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum)										// retval is the number of bytes of the sum
	{
		int retval=24;																																											// 192 bit=24 bytes
		crypto_primitives::tiger_ctx ctx;
		
		crypto_primitives::tiger_init(&ctx);
		crypto_primitives::tiger_update(&ctx,pInputBytes,InputLength);
		crypto_primitives::tiger_final(&ctx);
		crypto_primitives::tiger192_digest(&ctx,pSum);
		memset(&ctx,0x00,sizeof(ctx));

		return retval;
	}

	uint32 HASH_FUNCTION_Ripemd320(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum)										// retval is the number of bytes of the sum
	{
		int retval=40;																																											// 320 bit=40 bytes
		crypto_primitives::ripemd_ctx ctx;
				
		crypto_primitives::ripemd320_init(&ctx);
		crypto_primitives::ripemd_update(&ctx,pInputBytes,InputLength);
		crypto_primitives::ripemd_final(&ctx);
		crypto_primitives::ripemd_digest(&ctx,pSum);
		memset(&ctx,0x00,sizeof(ctx));

		return retval;
	}
	
	uint32 HASH_FUNCTION_Sha256(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum)											// retval is the number of bytes of the sum
	{
		int retval=32;																																											// 256 bit=32 bytes
		crypto_primitives::sha256_context ctx;
		
		crypto_primitives::sha256_starts(&ctx);
		crypto_primitives::sha256_update(&ctx,pInputBytes,InputLength);
		crypto_primitives::sha256_finish(&ctx,pSum);
		memset(&ctx,0x00,sizeof(ctx));

		return retval;
	}





		// -----------------------------------------------------------------------------------------------------------------------------------------------
		// wrapper functions for our cipher primitives
	void CIPHER_PRIMITIVE_KEY_SETUP_Anubis128(uint8 * p128bitKey,void  * pCC)															// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		struct crypto_primitives::NESSIEstruct_anubis * pAnubis_cc;
	
		pAnubis_cc=(crypto_primitives::NESSIEstruct_anubis *)pCC;
		pAnubis_cc->keyBits=128;																																						// MUST be initialized prior to doing anything else
		crypto_primitives::NESSIEkeysetup(p128bitKey,pAnubis_cc);
	}
	void CIPHER_PRIMITIVE_ENCRYPT_Anubis128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		struct crypto_primitives::NESSIEstruct_anubis * pAnubis_cc;

		pAnubis_cc=(crypto_primitives::NESSIEstruct_anubis *)pCC;
		crypto_primitives::NESSIEencrypt(pAnubis_cc,p128bit_Plaintext,p128bit_Ciphertext);
	}
	void CIPHER_PRIMITIVE_DECRYPT_Anubis128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		struct crypto_primitives::NESSIEstruct_anubis * pAnubis_cc;

		pAnubis_cc=(crypto_primitives::NESSIEstruct_anubis *)pCC;
		crypto_primitives::NESSIEdecrypt(pAnubis_cc,p128bit_Ciphertext,p128bit_Plaintext);
	}

	void CIPHER_PRIMITIVE_KEY_SETUP_Seed128(uint8 * p128bitKey,void  * pCC)																// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		DWORD * pDwRoundKey;
	
		pDwRoundKey=(DWORD *)pCC;
		crypto_primitives::SeedRoundKey(pDwRoundKey,p128bitKey);
	}
	void CIPHER_PRIMITIVE_ENCRYPT_Seed128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		DWORD * pDwRoundKey;
	
		pDwRoundKey=(DWORD *)pCC;
		crypto_primitives::SeedEncrypt(p128bit_Plaintext,pDwRoundKey);
		memcpy(p128bit_Ciphertext,p128bit_Plaintext,16);																										// some cipher primitives destroy the plaintext. We thus have to copy the ciphertext.
	}
	void CIPHER_PRIMITIVE_DECRYPT_Seed128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		DWORD * pDwRoundKey;
	
		pDwRoundKey=(DWORD *)pCC;
		crypto_primitives::SeedDecrypt(p128bit_Ciphertext,pDwRoundKey);
		memcpy(p128bit_Plaintext,p128bit_Ciphertext,16);																										// some cipher primitives destroy the plaintext. We thus have to copy the ciphertext.
	}

	void CIPHER_PRIMITIVE_KEY_SETUP_RC6_128(uint8 * p128bitKey,void  * pCC)																// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		word32 * pL_key;
	
		pL_key=(word32 *)pCC;
		crypto_primitives::rc6_set_key(pL_key,p128bitKey,128);
	}
	void CIPHER_PRIMITIVE_ENCRYPT_RC6_128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		word32 * pL_key;
	
		pL_key=(word32 *)pCC;
		crypto_primitives::rc6_encrypt(pL_key,p128bit_Plaintext,p128bit_Ciphertext);
	}
	void CIPHER_PRIMITIVE_DECRYPT_RC6_128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		word32 * pL_key;
	
		pL_key=(word32 *)pCC;
		crypto_primitives::rc6_decrypt(pL_key,p128bit_Ciphertext,p128bit_Plaintext);
	}

	void CIPHER_PRIMITIVE_KEY_SETUP_Camellia128(uint8 * p128bitKey,void  * pCC)													// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		word32 * pKkeyTable;
	
		pKkeyTable=(word32 *)pCC;
		crypto_primitives::Camellia_Ekeygen(128,p128bitKey,pKkeyTable);
	}
	void CIPHER_PRIMITIVE_ENCRYPT_Camellia128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		word32 * pKkeyTable;
	
		pKkeyTable=(word32 *)pCC;
		crypto_primitives::Camellia_EncryptBlock(128,p128bit_Plaintext,pKkeyTable,p128bit_Ciphertext);
	}
	void CIPHER_PRIMITIVE_DECRYPT_Camellia128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		word32 * pKkeyTable;
	
		pKkeyTable=(word32 *)pCC;
		crypto_primitives::Camellia_DecryptBlock(128,p128bit_Ciphertext,pKkeyTable,p128bit_Plaintext);
	}

	void CIPHER_PRIMITIVE_KEY_SETUP_Cast256_128(uint8 * p128bitKey,void  * pCC)														// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		crypto_primitives::cast256_instance * pKey;
	
		pKey=(crypto_primitives::cast256_instance *)pCC;
		crypto_primitives::cast256_set_key(pKey,(const unsigned __int32 *)p128bitKey,16);
	}
	void CIPHER_PRIMITIVE_ENCRYPT_Cast256_128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		crypto_primitives::cast256_instance * pKey;
	
		pKey=(crypto_primitives::cast256_instance *)pCC;
		crypto_primitives::cast256_encrypt(pKey,(unsigned __int32 *)p128bit_Plaintext);
		memcpy(p128bit_Ciphertext,p128bit_Plaintext,16);																										// some cipher primitives destroy the plaintext. We thus have to copy the ciphertext.
	}
	void CIPHER_PRIMITIVE_DECRYPT_Cast256_128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		crypto_primitives::cast256_instance * pKey;
	
		pKey=(crypto_primitives::cast256_instance *)pCC;
		crypto_primitives::cast256_decrypt(pKey,(unsigned __int32 *)p128bit_Ciphertext);
		memcpy(p128bit_Plaintext,p128bit_Ciphertext,16);																										// some cipher primitives destroy the plaintext. We thus have to copy the ciphertext.
	}

	void CIPHER_PRIMITIVE_KEY_SETUP_Serpent128(uint8 * p128bitKey,void  * pCC)														// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		crypto_primitives::SERPENT_KEY * pKey;
	
		pKey=(crypto_primitives::SERPENT_KEY *)pCC;
		crypto_primitives::serpent_set_key(pKey,(const unsigned __int32 *)p128bitKey,16);
	}
	void CIPHER_PRIMITIVE_ENCRYPT_Serpent128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		crypto_primitives::SERPENT_KEY * pKey;
	
		pKey=(crypto_primitives::SERPENT_KEY *)pCC;
		crypto_primitives::serpent_encrypt(pKey,(unsigned __int32 *)p128bit_Plaintext);
		memcpy(p128bit_Ciphertext,p128bit_Plaintext,16);																										// some cipher primitives destroy the plaintext. We thus have to copy the ciphertext.
	}
	void CIPHER_PRIMITIVE_DECRYPT_Serpent128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		crypto_primitives::SERPENT_KEY * pKey;
	
		pKey=(crypto_primitives::SERPENT_KEY *)pCC;
		crypto_primitives::serpent_decrypt(pKey,(unsigned __int32 *)p128bit_Ciphertext);
		memcpy(p128bit_Plaintext,p128bit_Ciphertext,16);																										// some cipher primitives destroy the plaintext. We thus have to copy the ciphertext.
	}

	void CIPHER_PRIMITIVE_KEY_SETUP_Rijndael128(uint8 * p128bitKey,void  * pCC)														// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		struct crypto_primitives::aes_context * pCtx;
	
		pCtx=(crypto_primitives::aes_context *)pCC;
		crypto_primitives::aes_set_key(pCtx,p128bitKey,128);
	}
	void CIPHER_PRIMITIVE_ENCRYPT_Rijndael128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		struct crypto_primitives::aes_context * pCtx;
	
		pCtx=(crypto_primitives::aes_context *)pCC;
		crypto_primitives::aes_encrypt(pCtx,p128bit_Plaintext);
		memcpy(p128bit_Ciphertext,p128bit_Plaintext,16);																										// some cipher primitives destroy the plaintext. We thus have to copy the ciphertext.
	}
	void CIPHER_PRIMITIVE_DECRYPT_Rijndael128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		struct crypto_primitives::aes_context * pCtx;
	
		pCtx=(crypto_primitives::aes_context *)pCC;
		crypto_primitives::aes_decrypt(pCtx,p128bit_Ciphertext);
		memcpy(p128bit_Plaintext,p128bit_Ciphertext,16);																										// some cipher primitives destroy the plaintext. We thus have to copy the ciphertext.
	}

	void CIPHER_PRIMITIVE_KEY_SETUP_Twofish128(uint8 * p128bitKey,void  * pCC)														// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		struct crypto_primitives::twofish_context * pCtx;
	
		pCtx=(crypto_primitives::twofish_context *)pCC;
		crypto_primitives::twofish_set_key(pCtx,(const unsigned __int32 *)p128bitKey,16);
	}
	void CIPHER_PRIMITIVE_ENCRYPT_Twofish128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		struct crypto_primitives::twofish_context * pCtx;
	
		pCtx=(crypto_primitives::twofish_context *)pCC;
		crypto_primitives::twofish_encrypt(pCtx,(unsigned __int32 *)p128bit_Plaintext);
		memcpy(p128bit_Ciphertext,p128bit_Plaintext,16);																										// some cipher primitives destroy the plaintext. We thus have to copy the ciphertext.
	}
	void CIPHER_PRIMITIVE_DECRYPT_Twofish128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext)// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	{
		struct crypto_primitives::twofish_context * pCtx;
	
		pCtx=(crypto_primitives::twofish_context *)pCC;
		crypto_primitives::twofish_decrypt(pCtx,(unsigned __int32 *)p128bit_Ciphertext);
		memcpy(p128bit_Plaintext,p128bit_Ciphertext,16);																										// some cipher primitives destroy the plaintext. We thus have to copy the ciphertext.
	}

		// in order to allocate sufficient RAM for all our cipher contexts, we need to know which one is the biggest. The value in bytes is returned by this function
	int CIPHER_PRIMITIVE_ComputeMaxSizeOfCryptoContext()
	{
		int retval=0;
		int i;
		word32 l_key[44];																																										// RC6
		DWORD dwRoundKey[32];																																								// SEED
			
			// let's use the variables in order to satisfy the compiler
		l_key[0]=0;
		dwRoundKey[0]=0;

		i= sizeof(struct crypto_primitives::twofish_context);
		if (i>retval) retval=i;
		i= sizeof(struct crypto_primitives::aes_context);																										// AES Rijndael
		if (i>retval) retval=i;
		i= sizeof(crypto_primitives::SERPENT_KEY);
		if (i>retval) retval=i;
		i= sizeof(crypto_primitives::KEY_TABLE_TYPE);																												// Camellia
		if (i>retval) retval=i;
		i= sizeof(l_key);																																										// RC6
		if (i>retval) retval=i;
		i= sizeof(dwRoundKey);																																							// SEED
		if (i>retval) retval=i;
		i= sizeof(struct crypto_primitives::NESSIEstruct_anubis);
		if (i>retval) retval=i;

		return retval;
	}



			// AES speed test: Encryption of 256Mbyte with only one time key setup
	void AES_Speed_Test()
	{
		byte key[32];
		struct crypto_primitives::aes_context ctx;

		int i, byte_count=0;
		uint8 data_in[16];

			// init key ONLY ONCE
		for (i=0;i<32;i++) key[i]=i;
		crypto_primitives::aes_set_key(&ctx,key,256); // 256 bit key

		while (byte_count<268435456) // 2^28
		{
			memset(data_in,0,16);
			(*(int *)data_in)=byte_count;
			crypto_primitives::aes_encrypt(&ctx,data_in);
			byte_count+=16;
		} // while (byte_count<268435456)
	}


	void AES_10million_key_setups_speed_Test()
	{
		byte key[32];
		struct crypto_primitives::aes_context ctx;

		int i;

			// init key ONLY ONCE
		for (i=0;i<32;i++) key[i]=i;

		for (i=0;i<10000000;i++) crypto_primitives::aes_set_key(&ctx,key,256); // 256 bit key
	}


		// return values:
		// 0: no errors
		// 1: Could not open plaintext file.
		// 2: Could not create ciphertext file.
	int EncryptFileWithPolymorphicMedley(WCHAR * pNameOfPlaintextFile, WCHAR * pNameOfCiphertextFile,uint8 * pKey,int key_length_in_bits,int complexity)
	{
		__int64						i;
		uint8							data_in[65536+16];	// 16 bytes reserve (in order to have no trouble with padding)
		uint8							ciphertext[16];	
		__int64						ciphertext_filesize;
		__int64						no_of_bytes_remaining;
		__int64						position_in_buffer;
		__int64						no_of_bytes_in_buffer,no_of_bytes_remaining_in_buffer;
		__int64						no_of_bytes_written;
		FILE							*in_stream, *out_stream;
		errno_t						err;
		WCHAR							in_fname[MAX_PATH];
		WCHAR							out_fname[MAX_PATH];
		struct _stati64		stat_buffer; 
		__int64						plaintext_filesize=0LL;
		PMCMED_cipher_context			PMCMED_cc;

		wcscpy_s(in_fname,MAX_PATH,pNameOfPlaintextFile);
		wcscpy_s(out_fname,MAX_PATH,pNameOfCiphertextFile);


		// let's check if the file exists
		if (_wstati64(pNameOfPlaintextFile,&stat_buffer)==0)
		{
			plaintext_filesize=stat_buffer.st_size;
		}
		else return 1; // Could not open plaintext file.


		no_of_bytes_remaining=plaintext_filesize;
		ciphertext_filesize=0LL;

			// open plaintext file
		if ((err=_wfopen_s(&in_stream,in_fname,L"rb"))!=0)
		{
			return 1; // Could not open plaintext file.
		}

			// create ciphertext file
		if ((err=_wfopen_s(&out_stream,out_fname,L"w+b"))!=0)
		{
			return 2; // Could not create ciphertext file.
		}

			// write size of plaintext file so that we know how many bytes of the stuff that we decrypt is real data and not padded data
		no_of_bytes_written=(__int64)fwrite((void*)&plaintext_filesize,sizeof(char),8,out_stream); // let's suppose that 8 bytes have been written

		PMCMED_keysetup((uint8*)pKey,(void*)&PMCMED_cc,key_length_in_bits,complexity);											// key_length_in_bits: 128 .. 1024 bit, complexity: 0: very fast, 2^16-1: huge set of keyed operations to compute internal state
		PMCMED_init_CBC_mode((void *)&PMCMED_cc,0);																													// 0: CBC_block_counter_start_value must be the same for decryption. It is good e.g. for encrypting data streams and it is automatically incremented by PMCMED_CBC_encrypt_cascade()

		while (no_of_bytes_remaining>0)
		{
			no_of_bytes_in_buffer=no_of_bytes_remaining_in_buffer=(int)fread((void*)data_in,sizeof(char),65536,in_stream);
			position_in_buffer=0;
			while (no_of_bytes_remaining_in_buffer>0)
			{
					// we need to pad the last block
				if (no_of_bytes_remaining<16)
				{
					for (i=position_in_buffer+no_of_bytes_remaining;i<position_in_buffer+no_of_bytes_remaining+16;i++) data_in[i]=0x00;
				}

				PMCMED_CBC_encrypt_cascade((void *)&PMCMED_cc,(unsigned char *)&(data_in[position_in_buffer]),ciphertext);	
				for (i=0;i<16;i++) data_in[i+position_in_buffer]=ciphertext[i];

				position_in_buffer+=16;
				no_of_bytes_remaining_in_buffer-=16;
				no_of_bytes_remaining-=16;
			} // while (no_of_bytes_remaining_in_buffer>0)
			no_of_bytes_written=(__int64)fwrite((void*)data_in,sizeof(char),(size_t)position_in_buffer,out_stream);
		}
		fflush(out_stream);

		fclose(in_stream);
		fclose(out_stream);

		PMCMED_free_memory((void *)&PMCMED_cc);

		return 0; // 0: no error
	}


		// return values:
		// 0: no errors
		// 1: Could not open ciphertext file.
		// 2: Could not create plaintext file.
	int DecryptFileWithPolymorphicMedley(WCHAR * pNameOfCiphertextFile, WCHAR * pNameOfPlaintextFile,uint8 * pKey,int key_length_in_bits,int complexity)
	{
		__int64						i;
		uint8							data_in[65536+16];	// 16 bytes reserve (in order to have no trouble with padding)
		uint8							plaintext[16];
		__int64						no_of_bytes_remaining;
		__int64						position_in_buffer;
		__int64						no_of_bytes_in_buffer,no_of_bytes_remaining_in_buffer;
		__int64						no_of_bytes_written;
		FILE							*in_stream, *out_stream;
		errno_t						err;
		WCHAR							in_fname[MAX_PATH];
		WCHAR							out_fname[MAX_PATH];
		struct _stati64		stat_buffer; 
		__int64						ciphertext_filesize=0LL;
		__int64						plaintext_filesize=0LL;
		__int64						no_of_plaintext_bytes_remaining;
		PMCMED_cipher_context			PMCMED_cc;

		wcscpy_s(in_fname,MAX_PATH,pNameOfCiphertextFile);
		wcscpy_s(out_fname,MAX_PATH,pNameOfPlaintextFile);


		// let's check if the file exists
		if (_wstati64(pNameOfCiphertextFile,&stat_buffer)==0)
		{
			ciphertext_filesize=stat_buffer.st_size;
		}
		else return 1; // Could not open plaintext file.

			// open plaintext file
		if ((err=_wfopen_s(&in_stream,in_fname,L"rb"))!=0)
		{
			return 1; // Could not open ciphertext file.
		}

			// read size of plaintext file (due to padding to 16 bytes, the plaintext file is usually up to 16 bytes shorter than the raw plaintext file)
		i=(__int64)fread((void*)&plaintext_filesize,sizeof(char),8,in_stream);
		no_of_bytes_remaining=plaintext_filesize;
		no_of_plaintext_bytes_remaining=plaintext_filesize;


			// create ciphertext file
		if ((err=_wfopen_s(&out_stream,out_fname,L"w+b"))!=0)
		{
			return 2; // Could not create plaintext file.
		}

		PMCMED_keysetup((uint8*)pKey,(void*)&PMCMED_cc,key_length_in_bits,complexity);											// key_length_in_bits: 128 .. 1024 bit, complexity: 0: very fast, 2^16-1: huge set of keyed operations to compute internal state
		PMCMED_init_CBC_mode((void *)&PMCMED_cc,0);																													// 0: CBC_block_counter_start_value must be the same for decryption. It is good e.g. for encrypting data streams and it is automatically incremented by PMCMED_CBC_encrypt_cascade()


		while (no_of_bytes_remaining>0)
		{
			no_of_bytes_in_buffer=no_of_bytes_remaining_in_buffer=(__int64)fread((void*)data_in,sizeof(char),65536,in_stream);
			position_in_buffer=0;
			while (no_of_bytes_remaining_in_buffer>0)
			{
				PMCMED_CBC_decrypt_cascade((void *)&PMCMED_cc,(unsigned char *)&(data_in[position_in_buffer]),plaintext);	

				for (i=0;i<16;i++) data_in[i+position_in_buffer]=plaintext[i];

				no_of_bytes_remaining_in_buffer-=16;
				no_of_bytes_remaining-=16;
				no_of_plaintext_bytes_remaining-=16;
				position_in_buffer+=16;
			} // while (no_of_bytes_remaining_in_buffer>0)
			if (no_of_bytes_remaining>0) no_of_bytes_written=(__int64)fwrite((void*)data_in,sizeof(char),(size_t)position_in_buffer,out_stream);
			else no_of_bytes_written=(__int64)fwrite((void*)data_in,sizeof(char),(size_t)(position_in_buffer+no_of_plaintext_bytes_remaining),out_stream); // write last bunch of blocks MINUS the padded bytes 
		}
		fflush(out_stream);

		fclose(in_stream);
		fclose(out_stream);

		PMCMED_free_memory((void *)&PMCMED_cc);

		return 0; // 0: no error
	}




} // namespace polymorphic_medley_cipher
