
#pragma once

namespace polymorphic_medley_cipher
{
	#ifndef uint8
		#define uint8  unsigned __int8
	#endif
	#ifndef uint32
		#define uint32 unsigned __int32
	#endif
	#ifndef word8
		#define word8 unsigned __int8
	#endif
	#ifndef word16
		#define word16 unsigned __int16
	#endif
	#ifndef word32
		#define word32 unsigned __int32
	#endif
	#ifndef word64
		#define word64 unsigned __int64
	#endif



	// need for massive RAM usage during key setup: http://www.intel.com/content/www/us/en/high-performance-computing/high-performance-xeon-phi-coprocessor-brief.html
	// Xeaon Phi: 60 cores, 32 KB L1 I/D cache, 512 KB L2 cache (per core)

	#define NUM_OF_HASH_FUNCTION_DELEGATES					5																											// we've only got 5 base functions
	#define MAX_NUMBER_OF_CONTEXTS_PER_BASE_CIPHER	16																										// MUST be a power of 2
	#define NO_OF_HASH_FUNCTION_DELEGATES						5																											// currently we have five hash functions (can be any number greater than 0)
	#define NUM_OF_HASH_RESULTS											32																										// 32: 32 possible worker keys available for our ciphers (32 is the absolute minimum). Value MUST be a power of 2
	#define NUM_OF_CIPHER_FUNCTION_DELEGATES				8																											// currently we have 8 cipher primitives (MUST be a power of 2 like 8, 16 or 32)
	#define NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE	16																								// 16: for Twofish, Rijndael, Anubis, etc. exist 16 different cipher contexts
	#define MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES			1204 //8892																						// this constant MUST be the same as the value returned by CIPHER_PRIMITIVE_ComputeMaxSizeOfCryptoContext(). Can be up to 8892 bytes because of Twofish if the tables are held in the RAM.
	#define	SIZE_OF_MEMORY_BIAS_ARRAY								16*1024*1024																					// # of bytes that are key depended (if SIZE_OF_MEMORY_BIAS_ARRAY == 1048576, then LB_SIZE_OF_MEMORY_BIAS_ARRAY must equal 20 (2^20=1048576)
	#define LB_SIZE_OF_MEMORY_BIAS_ARRAY						24																										// 2^20 = 1048576, 2^24=16*1024*1024=16777216
	#define NUM_OF_KISSDELEGATES										128																										// we have 128 different static KISS RNGs that feature identical excecution times and identical current consumption
	#define LB_OF_NUM_OF_KISSDELEGATES							7																											// 2^7 = 128
	// switching between KISS contexts is easy, but these contexts need to be reset prior to encryption or decryption of each packet
	
	#define NUM_OF_KISS_CONTEXTS										16																										// minimum permissible number: 8
	#define LB_OF_NUM_OF_KISS_CONTEXTS							4																											// 16=2^4  => LB_OF_NUM_OF_KISS_CONTEXTS is 4

	#define MINIMUM_COMPLEXITY_KEY_SETUP						0																											// PMCMED_keysetup() executes fast. As a matter of consequence, brute force attack security is comparably low.
	#define MEDIUM_COMPLEXITY_KEY_SETUP							16384																									// PMCMED_keysetup() executes relatively slowly and brute force attack security is high
	#define MAXIMUM_COMPLEXITY_KEY_SETUP						65535																									// PMCMED_keysetup() executes slowly, but brute force attack security is highest


		// polymorphic KISS RNG
	typedef struct 
	{
			word64  z;
			word64  w;
			word64  jsr;
			word64  j;
			word64	x;
			word64	c;
	} gbpmc_kiss_context;


	typedef word64	(KISSDELEGATE)(gbpmc_kiss_context * pctx);																						// kiss0() .. kiss127()

	typedef uint32 (HASH_FUNCTION_DELEGATE)(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum);					// retval is the number of bytes of the sum

	typedef void (KEY_SETUP_FUNCTION_DELEGATE)(uint8 * p128bitKey,void  * pCC);
	typedef void (ENCRYPTION_FUNCTION_DELEGATE)(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext);
	typedef void (DECRYPTION_FUNCTION_DELEGATE)(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext);

	uint32 HASH_FUNCTION_Whirlpool512(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum);								// retval is the number of bytes of the sum
	uint32 HASH_FUNCTION_Haval256(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum);										// retval is the number of bytes of the sum
	uint32 HASH_FUNCTION_Tiger192(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum);										// retval is the number of bytes of the sum
	uint32 HASH_FUNCTION_Ripemd320(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum);									// retval is the number of bytes of the sum
	uint32 HASH_FUNCTION_Sha256(uint8 * pInputBytes,uint32 InputLength,uint8 * pSum);											// retval is the number of bytes of the sum

	void CIPHER_PRIMITIVE_KEY_SETUP_Anubis128(uint8 * p128bitKey,void  * pCC);														// keysize: 128 bit ONLY, pCC must be sufficiently big
	void CIPHER_PRIMITIVE_ENCRYPT_Anubis128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_DECRYPT_Anubis128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_KEY_SETUP_Seed128(uint8 * p128bitKey,void  * pCC);															// keysize: 128 bit ONLY, pCC must be sufficiently big
	void CIPHER_PRIMITIVE_ENCRYPT_Seed128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_DECRYPT_Seed128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_KEY_SETUP_RC6_128(uint8 * p128bitKey,void  * pCC);															// keysize: 128 bit ONLY, pCC must be sufficiently big
	void CIPHER_PRIMITIVE_ENCRYPT_RC6_128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_DECRYPT_RC6_128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_KEY_SETUP_Camellia128(uint8 * p128bitKey,void  * pCC);													// keysize: 128 bit ONLY, pCC must be sufficiently big
	void CIPHER_PRIMITIVE_ENCRYPT_Camellia128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_DECRYPT_Camellia128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_KEY_SETUP_Cast256_128(uint8 * p128bitKey,void  * pCC);													// keysize: 128 bit ONLY, pCC must be sufficiently big
	void CIPHER_PRIMITIVE_ENCRYPT_Cast256_128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_DECRYPT_Cast256_128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_KEY_SETUP_Serpent128(uint8 * p128bitKey,void  * pCC);														// keysize: 128 bit ONLY, pCC must be sufficiently big
	void CIPHER_PRIMITIVE_ENCRYPT_Serpent128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_DECRYPT_Serpent128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_KEY_SETUP_Rijndael128(uint8 * p128bitKey,void  * pCC);													// keysize: 128 bit ONLY, pCC must be sufficiently big
	void CIPHER_PRIMITIVE_ENCRYPT_Rijndael128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_DECRYPT_Rijndael128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_KEY_SETUP_Twofish128(uint8 * p128bitKey,void  * pCC);														// keysize: 128 bit ONLY, pCC must be sufficiently big
	void CIPHER_PRIMITIVE_ENCRYPT_Twofish128(void  * pCC,uint8 * p128bit_Plaintext,uint8 * p128bit_Ciphertext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
	void CIPHER_PRIMITIVE_DECRYPT_Twofish128(void  * pCC,uint8 * p128bit_Ciphertext,uint8 * p128bit_Plaintext);// all our 128 bit base ciphers feature 128 bit block size and 128 bit key length
			// in order to allocate sufficient RAM for all our cipher contexts, we need to know which one is the biggest. The value in bytes is returned by this function
	int CIPHER_PRIMITIVE_ComputeMaxSizeOfCryptoContext();


	typedef struct PMCMED_hash_context { 
		HASH_FUNCTION_DELEGATE	* hash_func_delegates[NUM_OF_HASH_FUNCTION_DELEGATES];
		uint8										*	pHash_results;																														// e.g. [16*(NUM_OF_HASH_RESULTS+3)]  :  +3: we must take care of different hash lengths (pointer will be NULL again after key setup)
	} _PMCMED_hash_context;

	typedef struct PMCMED_cipher_context { 
		struct PMCMED_hash_context	PMCMED_hash_ctx;

		KEY_SETUP_FUNCTION_DELEGATE		* key_setup_func_delegates[NUM_OF_CIPHER_FUNCTION_DELEGATES];
		ENCRYPTION_FUNCTION_DELEGATE	* encryption_func_delegates[NUM_OF_CIPHER_FUNCTION_DELEGATES];
		DECRYPTION_FUNCTION_DELEGATE	* decryption_func_delegates[NUM_OF_CIPHER_FUNCTION_DELEGATES];

		uint8 * pcc;																																												// crypto context. Size=MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES*NUM_OF_CIPHER_FUNCTION_DELEGATES*NUM_OF_CRYPTO_CONTEXTS_PER_CIPHER_DELEGATE

		KEY_SETUP_FUNCTION_DELEGATE		* key_setup_func_of_scheduler;																				// one cipher of our 8 cipher primitives will be used as scheduler
		ENCRYPTION_FUNCTION_DELEGATE	* encryption_func_of_scheduler;
		DECRYPTION_FUNCTION_DELEGATE	* decryption_func_of_scheduler;
		uint8														IV_for_CBC[16];																											// 128 bit initialization vector for our CBC functions
		uint8													*	memory_bias;																												// huge array that is needed to buffer lots of history during key setup
		uint8														last_block_CBC[16];																									// used for our CBC operations to buffer the last ciphertext
		word64													CBC_block_counter;																									// counts from 0 during CBC encryption/decryption. Increments automatically with each call to PMCMED_CBC_encrypt_cascade() or PMCMED_CBC_decrypt_cascade()
		uint8													* crypto_context_of_scheduler[MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES];		// is initialized by PMCMED_keysetup()
		uint8														initial_plaintext_scheduler[16];																		// 128 bit (as our scheduler is nothing but a normal 128 bit cipher primitive)
		uint8														plaintext_scheduler[16];																						// 128 bit (as our scheduler is nothing but a normal 128 bit cipher primitive)
		uint8														ciphertext_scheduler[16];																						// 128 bit (as our scheduler is nothing but a normal 128 bit cipher primitive)
		KISSDELEGATE									*	gbpmc_kiss[NUM_OF_KISSDELEGATES];
		gbpmc_kiss_context						*	initial_ctx_array[NUM_OF_KISS_CONTEXTS];

	} _PMCMED_cipher_context; 


		// interface to the Polymorphic Medley cipher

		// PMCMED_keysetup returns 0 if the key could be set up correctly.
		// These values are returned in case of an error:
		// -1: insufficient memory
		// -2: MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES does NOT equal the return value of CIPHER_PRIMITIVE_ComputeMaxSizeOfCryptoContext() => MAX_SIZE_OF_CRYPTO_CONTEXT_IN_BYTES must be corrected and then the code recompiled
	int PMCMED_keysetup(uint8 * pKey,void * pPMCMED_cc,uint32 key_length_in_bits,uint32 complexity);			// key_length_in_bits: 128 .. 1024 bit, complexity: 0: very fast, 2^16-1: huge set of keyed operations to compute internal state
	void PMCMED_encrypt_cascade(void * pPMCMED_cc,uint8 * pPlaintext,uint8 * pCiphertext);								// encryption in ECB mode: all 8 cipher primitives are used to encrypt a single block of data (only 40k combinations)
	void PMCMED_decrypt_cascade(void * pPMCMED_cc,uint8 * pCiphertext,uint8 * pPlaintext);								// decryption in ECB mode: all 8 cipher primitives are used in reverse order to decrypt a single block of data (only 40k combinations)
	void PMCMED_encrypt_max_var_cascade(void * pPMCMED_cc,uint8 * pPlaintext,uint8 * pCiphertext);				// encryption in ECB mode: free choice of the base ciphers => 2^24 possible combinations
	void PMCMED_decrypt_max_var_cascade(void * pPMCMED_cc,uint8 * pCiphertext,uint8 * pPlaintext);				// decryption in ECB mode: free choice of the base ciphers => 2^24 possible combinations
	int PMCMED_free_memory(void * pPMCMED_cc);

	void PMCMED_init_CBC_mode(void * pPMCMED_cc,word64 CBC_block_counter_start_value);
	void PMCMED_CBC_encrypt_cascade(void * pPMCMED_cc,uint8 * pPlaintext,uint8 * pCiphertext);
	void PMCMED_CBC_decrypt_cascade(void * pPMCMED_cc,uint8 * pCiphertext,uint8 * pPlaintext);

	bool test_PMC_Medley(void);																																						// tests our cipher and returns true if everything is ok


	void AES_Speed_Test(void);																																						// Encryption of 256Mbyte with only one time key setup
	void AES_10million_key_setups_speed_Test();

	int EncryptFileWithPolymorphicMedley(WCHAR * pNameOfPlaintextFile, WCHAR * pNameOfCiphertextFile,uint8 * pKey,int key_length_in_bits,int complexity);
	int DecryptFileWithPolymorphicMedley(WCHAR * pNameOfCiphertextFile, WCHAR * pNameOfPlaintextFile,uint8 * pKey,int key_length_in_bits,int complexity);

	
		// 64 bit PRNGs
	void init_gbpmc_kiss_delegate_array(KISSDELEGATE	*	pgbpmc_k[]);
	void init_gbpmc_kiss_context(gbpmc_kiss_context * kc,word64 param1,word64 param2);

	word64 kiss0(gbpmc_kiss_context * pkctx);
	word64 kiss1(gbpmc_kiss_context * pkctx);
	word64 kiss2(gbpmc_kiss_context * pkctx);
	word64 kiss3(gbpmc_kiss_context * pkctx);
	word64 kiss4(gbpmc_kiss_context * pkctx);
	word64 kiss5(gbpmc_kiss_context * pkctx);
	word64 kiss6(gbpmc_kiss_context * pkctx);
	word64 kiss7(gbpmc_kiss_context * pkctx);
	word64 kiss8(gbpmc_kiss_context * pkctx);
	word64 kiss9(gbpmc_kiss_context * pkctx);
	word64 kiss10(gbpmc_kiss_context * pkctx);
	word64 kiss11(gbpmc_kiss_context * pkctx);
	word64 kiss12(gbpmc_kiss_context * pkctx);
	word64 kiss13(gbpmc_kiss_context * pkctx);
	word64 kiss14(gbpmc_kiss_context * pkctx);
	word64 kiss15(gbpmc_kiss_context * pkctx);
	word64 kiss16(gbpmc_kiss_context * pkctx);
	word64 kiss17(gbpmc_kiss_context * pkctx);
	word64 kiss18(gbpmc_kiss_context * pkctx);
	word64 kiss19(gbpmc_kiss_context * pkctx);
	word64 kiss20(gbpmc_kiss_context * pkctx);
	word64 kiss21(gbpmc_kiss_context * pkctx);
	word64 kiss22(gbpmc_kiss_context * pkctx);
	word64 kiss23(gbpmc_kiss_context * pkctx);
	word64 kiss24(gbpmc_kiss_context * pkctx);
	word64 kiss25(gbpmc_kiss_context * pkctx);
	word64 kiss26(gbpmc_kiss_context * pkctx);
	word64 kiss27(gbpmc_kiss_context * pkctx);
	word64 kiss28(gbpmc_kiss_context * pkctx);
	word64 kiss29(gbpmc_kiss_context * pkctx);
	word64 kiss30(gbpmc_kiss_context * pkctx);
	word64 kiss31(gbpmc_kiss_context * pkctx);
	word64 kiss32(gbpmc_kiss_context * pkctx);
	word64 kiss33(gbpmc_kiss_context * pkctx);
	word64 kiss34(gbpmc_kiss_context * pkctx);
	word64 kiss35(gbpmc_kiss_context * pkctx);
	word64 kiss36(gbpmc_kiss_context * pkctx);
	word64 kiss37(gbpmc_kiss_context * pkctx);
	word64 kiss38(gbpmc_kiss_context * pkctx);
	word64 kiss39(gbpmc_kiss_context * pkctx);
	word64 kiss40(gbpmc_kiss_context * pkctx);
	word64 kiss41(gbpmc_kiss_context * pkctx);
	word64 kiss42(gbpmc_kiss_context * pkctx);
	word64 kiss43(gbpmc_kiss_context * pkctx);
	word64 kiss44(gbpmc_kiss_context * pkctx);
	word64 kiss45(gbpmc_kiss_context * pkctx);
	word64 kiss46(gbpmc_kiss_context * pkctx);
	word64 kiss47(gbpmc_kiss_context * pkctx);
	word64 kiss48(gbpmc_kiss_context * pkctx);
	word64 kiss49(gbpmc_kiss_context * pkctx);
	word64 kiss50(gbpmc_kiss_context * pkctx);
	word64 kiss51(gbpmc_kiss_context * pkctx);
	word64 kiss52(gbpmc_kiss_context * pkctx);
	word64 kiss53(gbpmc_kiss_context * pkctx);
	word64 kiss54(gbpmc_kiss_context * pkctx);
	word64 kiss55(gbpmc_kiss_context * pkctx);
	word64 kiss56(gbpmc_kiss_context * pkctx);
	word64 kiss57(gbpmc_kiss_context * pkctx);
	word64 kiss58(gbpmc_kiss_context * pkctx);
	word64 kiss59(gbpmc_kiss_context * pkctx);
	word64 kiss60(gbpmc_kiss_context * pkctx);
	word64 kiss61(gbpmc_kiss_context * pkctx);
	word64 kiss62(gbpmc_kiss_context * pkctx);
	word64 kiss63(gbpmc_kiss_context * pkctx);
	word64 kiss64(gbpmc_kiss_context * pkctx);
	word64 kiss65(gbpmc_kiss_context * pkctx);
	word64 kiss66(gbpmc_kiss_context * pkctx);
	word64 kiss67(gbpmc_kiss_context * pkctx);
	word64 kiss68(gbpmc_kiss_context * pkctx);
	word64 kiss69(gbpmc_kiss_context * pkctx);
	word64 kiss70(gbpmc_kiss_context * pkctx);
	word64 kiss71(gbpmc_kiss_context * pkctx);
	word64 kiss72(gbpmc_kiss_context * pkctx);
	word64 kiss73(gbpmc_kiss_context * pkctx);
	word64 kiss74(gbpmc_kiss_context * pkctx);
	word64 kiss75(gbpmc_kiss_context * pkctx);
	word64 kiss76(gbpmc_kiss_context * pkctx);
	word64 kiss77(gbpmc_kiss_context * pkctx);
	word64 kiss78(gbpmc_kiss_context * pkctx);
	word64 kiss79(gbpmc_kiss_context * pkctx);
	word64 kiss80(gbpmc_kiss_context * pkctx);
	word64 kiss81(gbpmc_kiss_context * pkctx);
	word64 kiss82(gbpmc_kiss_context * pkctx);
	word64 kiss83(gbpmc_kiss_context * pkctx);
	word64 kiss84(gbpmc_kiss_context * pkctx);
	word64 kiss85(gbpmc_kiss_context * pkctx);
	word64 kiss86(gbpmc_kiss_context * pkctx);
	word64 kiss87(gbpmc_kiss_context * pkctx);
	word64 kiss88(gbpmc_kiss_context * pkctx);
	word64 kiss89(gbpmc_kiss_context * pkctx);
	word64 kiss90(gbpmc_kiss_context * pkctx);
	word64 kiss91(gbpmc_kiss_context * pkctx);
	word64 kiss92(gbpmc_kiss_context * pkctx);
	word64 kiss93(gbpmc_kiss_context * pkctx);
	word64 kiss94(gbpmc_kiss_context * pkctx);
	word64 kiss95(gbpmc_kiss_context * pkctx);
	word64 kiss96(gbpmc_kiss_context * pkctx);
	word64 kiss97(gbpmc_kiss_context * pkctx);
	word64 kiss98(gbpmc_kiss_context * pkctx);
	word64 kiss99(gbpmc_kiss_context * pkctx);
	word64 kiss100(gbpmc_kiss_context * pkctx);
	word64 kiss101(gbpmc_kiss_context * pkctx);
	word64 kiss102(gbpmc_kiss_context * pkctx);
	word64 kiss103(gbpmc_kiss_context * pkctx);
	word64 kiss104(gbpmc_kiss_context * pkctx);
	word64 kiss105(gbpmc_kiss_context * pkctx);
	word64 kiss106(gbpmc_kiss_context * pkctx);
	word64 kiss107(gbpmc_kiss_context * pkctx);
	word64 kiss108(gbpmc_kiss_context * pkctx);
	word64 kiss109(gbpmc_kiss_context * pkctx);
	word64 kiss110(gbpmc_kiss_context * pkctx);
	word64 kiss111(gbpmc_kiss_context * pkctx);
	word64 kiss112(gbpmc_kiss_context * pkctx);
	word64 kiss113(gbpmc_kiss_context * pkctx);
	word64 kiss114(gbpmc_kiss_context * pkctx);
	word64 kiss115(gbpmc_kiss_context * pkctx);
	word64 kiss116(gbpmc_kiss_context * pkctx);
	word64 kiss117(gbpmc_kiss_context * pkctx);
	word64 kiss118(gbpmc_kiss_context * pkctx);
	word64 kiss119(gbpmc_kiss_context * pkctx);
	word64 kiss120(gbpmc_kiss_context * pkctx);
	word64 kiss121(gbpmc_kiss_context * pkctx);
	word64 kiss122(gbpmc_kiss_context * pkctx);
	word64 kiss123(gbpmc_kiss_context * pkctx);
	word64 kiss124(gbpmc_kiss_context * pkctx);
	word64 kiss125(gbpmc_kiss_context * pkctx);
	word64 kiss126(gbpmc_kiss_context * pkctx);
	word64 kiss127(gbpmc_kiss_context * pkctx);



} // namespace polymorphic_medley_cipher
