
#include "stdafx.h"
#include "standard_crypto_primitives.h"



namespace crypto_primitives
{

/**
 * <b>The Anubis block cipher.</b>
 *
 * <P>
 * <b>References</b>
 *
 * <P>
 * The Anubis algorithm was developed by
 * <a href="mailto:pbarreto@scopus.com.br">Paulo S. L. M. Barreto</a> and
 * <a href="mailto:vincent.rijmen@esat.kuleuven.ac.be">Vincent Rijmen</a>.
 *
 * See
 *		P.S.L.M. Barreto, V. Rijmen,
 *		``The Anubis block cipher,''
 *		NESSIE submission, 2000.
 * 
 * @author	Paulo S.L.M. Barreto
 * @author	Vincent Rijmen.
 *
 * @version 1.0 (2000.09.29)
 * 
 * =============================================================================
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>

//#include "nessie.h"

	/*	128 bit test vectors
	Set 1, vector#  0:
                      key=80000000000000000000000000000000
                    plain=00000000000000000000000000000000
                   cipher=F06860FC6730E818F132C78AF4132AFE
                decrypted=00000000000000000000000000000000
       Iterated 100 times=0DD6A9856746E6381196C1C3D0A9A678
      Iterated 1000 times=51BECE79FC4DB0B4F2EAB992A0F32943

Set 1, vector# 17:
                      key=00004000000000000000000000000000
                    plain=00000000000000000000000000000000
                   cipher=D74C1B10C7D891F4B4591B2CDB939910
                decrypted=00000000000000000000000000000000
       Iterated 100 times=852C1C139A0C20430C2EC7B93563B82B
      Iterated 1000 times=D316D5A204727ABDAA25ED47A642FD76

Set 3, vector#126:
                      key=7E7E7E7E7E7E7E7E7E7E7E7E7E7E7E7E
                    plain=7E7E7E7E7E7E7E7E7E7E7E7E7E7E7E7E
                   cipher=CBD856A8F17A04D38F469B2A3C339E2C
                decrypted=7E7E7E7E7E7E7E7E7E7E7E7E7E7E7E7E
       Iterated 100 times=F994E3AC38C142E1CD295C16377038D3
      Iterated 1000 times=75A163E2764C25CFEF1D898FEC8EFC1B

	*/

		// those are the standard test vectors
	static unsigned char ANUBIS_KEYs[3][17] =
	{
			"\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
			"\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
			"\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e"
	};

	static unsigned char ANUBIS_PTs[3][17] =
	{
			"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
			"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
			"\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e\x7e"
	};

	static unsigned char ANUBIS_CTs[3][17] =
	{
			"\xf0\x68\x60\xfc\x67\x30\xe8\x18\xf1\x32\xc7\x8a\xf4\x13\x2a\xfe",
			"\xd7\x4c\x1b\x10\xc7\xd8\x91\xf4\xb4\x59\x1b\x2c\xdb\x93\x99\x10",
			"\xcb\xd8\x56\xa8\xf1\x7a\x04\xd3\x8f\x46\x9b\x2a\x3c\x33\x9e\x2c"
	};

	static unsigned char ANUBIS_CTs_100times[3][17] =
	{
			"\x0d\xd6\xa9\x85\x67\x46\xe6\x38\x11\x96\xc1\xc3\xd0\xa9\xa6\x78",
			"\x85\x2c\x1c\x13\x9a\x0c\x20\x43\x0c\x2e\xc7\xb9\x35\x63\xb8\x2b",
			"\xf9\x94\xe3\xac\x38\xc1\x42\xe1\xcd\x29\x5c\x16\x37\x70\x38\xd3"
	};


		// returns true for pass
	bool test_Anubis(void) {
			bool result=true; // false=fail, true=pass
			int i,j;
			struct NESSIEstruct_anubis anubis_cc;
			unsigned char data[16];
			unsigned char ciphertext[16];

			for( i = 0; i < 3; i++ )
			{
				memcpy( data, ANUBIS_PTs[i], 16 );
				
				anubis_cc.keyBits=128;																													// MUST be initialized prior to doing anything else
				NESSIEkeysetup(ANUBIS_KEYs[i],&anubis_cc);
				NESSIEencrypt(&anubis_cc, data, ciphertext);
				
				if (memcmp( ciphertext, ANUBIS_CTs[i], 16 ) )
				{
					result=0;
					return(result);
				}
			}
			
			for( i = 0; i < 3; i++ )
			{
				memcpy( ciphertext, ANUBIS_CTs[i], 16 );
				
				NESSIEkeysetup(ANUBIS_KEYs[i],&anubis_cc);
				NESSIEdecrypt(&anubis_cc, ciphertext, data);
				
				if (memcmp( data, ANUBIS_PTs[i], 16 ) )
				{
					result=false;
					return(result);
				}
			}

				// very nice: there exist test vectors for 100 and 1000 iterations. Let's test 100 times only...
			for( i = 0; i < 3; i++ )
			{
				memcpy( data, ANUBIS_PTs[i], 16 );
				
				anubis_cc.keyBits=128;																													// MUST be initialized prior to doing anything else
				NESSIEkeysetup(ANUBIS_KEYs[i],&anubis_cc);

				for (j=0;j<100;j++)
				{
					NESSIEencrypt(&anubis_cc, data, ciphertext);
					memcpy(data,ciphertext,16);
				}
				
				if (memcmp( ciphertext, ANUBIS_CTs_100times[i], 16 ) )
				{
					result=0;
					return(result);
				}
			}

			return(result);
	}






/*
 * Though Anubis is endianness-neutral, the encryption tables are listed
 * in BIG-ENDIAN format, which is adopted throughout this implementation
 * (but little-endian notation would be equally suitable if consistently
 * employed).
 */

static const u32 T0[256] = {
    0xa753a6f5U, 0xd3bb6bd0U, 0xe6d1bf6eU, 0x71e2d93bU,
    0xd0bd67daU, 0xac458acfU, 0x4d9a29b3U, 0x79f2f90bU,
    0x3a74e89cU, 0xc98f038cU, 0x913f7e41U, 0xfce5d732U,
    0x1e3c7844U, 0x478e018fU, 0x54a84de5U, 0xbd67cea9U,
    0x8c050a0fU, 0xa557aef9U, 0x7af4f501U, 0xfbebcb20U,
    0x63c69157U, 0xb86ddab7U, 0xdda753f4U, 0xd4b577c2U,
    0xe5d7b364U, 0xb37bf68dU, 0xc59733a4U, 0xbe61c2a3U,
    0xa94f9ed1U, 0x880d1a17U, 0x0c183028U, 0xa259b2ebU,
    0x3972e496U, 0xdfa35bf8U, 0x2952a4f6U, 0xdaa94fe6U,
    0x2b56acfaU, 0xa84d9ad7U, 0xcb8b0b80U, 0x4c982db5U,
    0x4b9631a7U, 0x224488ccU, 0xaa4992dbU, 0x244890d8U,
    0x4182199bU, 0x70e0dd3dU, 0xa651a2f3U, 0xf9efc32cU,
    0x5ab475c1U, 0xe2d9af76U, 0xb07dfa87U, 0x366cd8b4U,
    0x7dfae913U, 0xe4d5b762U, 0x3366ccaaU, 0xffe3db38U,
    0x60c09d5dU, 0x204080c0U, 0x08102030U, 0x8b0b161dU,
    0x5ebc65d9U, 0xab4b96ddU, 0x7ffee11fU, 0x78f0fd0dU,
    0x7cf8ed15U, 0x2c58b0e8U, 0x57ae41efU, 0xd2b96fd6U,
    0xdca557f2U, 0x6ddaa973U, 0x7efce519U, 0x0d1a342eU,
    0x53a651f7U, 0x94356a5fU, 0xc39b2bb0U, 0x2850a0f0U,
    0x274e9cd2U, 0x060c1814U, 0x5fbe61dfU, 0xad478ec9U,
    0x67ce814fU, 0x5cb86dd5U, 0x55aa49e3U, 0x48903dadU,
    0x0e1c3824U, 0x52a455f1U, 0xeac98f46U, 0x42841591U,
    0x5bb671c7U, 0x5dba69d3U, 0x3060c0a0U, 0x58b07dcdU,
    0x51a259fbU, 0x59b279cbU, 0x3c78f088U, 0x4e9c25b9U,
    0x3870e090U, 0x8a09121bU, 0x72e4d531U, 0x14285078U,
    0xe7d3bb68U, 0xc6913faeU, 0xdea15ffeU, 0x50a05dfdU,
    0x8e010203U, 0x9239724bU, 0xd1bf63dcU, 0x77eec12fU,
    0x933b764dU, 0x458a0983U, 0x9a29527bU, 0xce811f9eU,
    0x2d5ab4eeU, 0x03060c0aU, 0x62c49551U, 0xb671e293U,
    0xb96fdeb1U, 0xbf63c6a5U, 0x96316253U, 0x6bd6b167U,
    0x3f7efc82U, 0x070e1c12U, 0x1224486cU, 0xae4182c3U,
    0x40801d9dU, 0x3468d0b8U, 0x468c0589U, 0x3e7cf884U,
    0xdbab4be0U, 0xcf831b98U, 0xecc59752U, 0xcc851792U,
    0xc19f23bcU, 0xa15fbee1U, 0xc09d27baU, 0xd6b17fceU,
    0x1d3a744eU, 0xf4f5f702U, 0x61c2995bU, 0x3b76ec9aU,
    0x10204060U, 0xd8ad47eaU, 0x68d0bd6dU, 0xa05dbae7U,
    0xb17ffe81U, 0x0a14283cU, 0x69d2b96bU, 0x6cd8ad75U,
    0x499239abU, 0xfae9cf26U, 0x76ecc529U, 0xc49537a2U,
    0x9e214263U, 0x9b2b567dU, 0x6edca579U, 0x992f5e71U,
    0xc2992fb6U, 0xb773e695U, 0x982d5a77U, 0xbc65caafU,
    0x8f030605U, 0x85172e39U, 0x1f3e7c42U, 0xb475ea9fU,
    0xf8edc72aU, 0x11224466U, 0x2e5cb8e4U, 0x00000000U,
    0x254a94deU, 0x1c387048U, 0x2a54a8fcU, 0x3d7af48eU,
    0x050a141eU, 0x4f9e21bfU, 0x7bf6f107U, 0xb279f28bU,
    0x3264c8acU, 0x903d7a47U, 0xaf4386c5U, 0x19326456U,
    0xa35bb6edU, 0xf7f3fb08U, 0x73e6d137U, 0x9d274e69U,
    0x152a547eU, 0x74e8cd25U, 0xeec19f5eU, 0xca890f86U,
    0x9f234665U, 0x0f1e3c22U, 0x1b366c5aU, 0x75eac923U,
    0x86112233U, 0x84152a3fU, 0x9c254a6fU, 0x4a9435a1U,
    0x97336655U, 0x1a34685cU, 0x65ca8943U, 0xf6f1ff0eU,
    0xedc79354U, 0x09122436U, 0xbb6bd6bdU, 0x264c98d4U,
    0x831b362dU, 0xebcb8b40U, 0x6fdea17fU, 0x811f3e21U,
    0x04081018U, 0x6ad4b561U, 0x43861197U, 0x01020406U,
    0x172e5c72U, 0xe1dfa37cU, 0x87132635U, 0xf5f7f304U,
    0x8d070e09U, 0xe3dbab70U, 0x23468ccaU, 0x801d3a27U,
    0x44880d85U, 0x162c5874U, 0x66cc8549U, 0x214284c6U,
    0xfee1df3eU, 0xd5b773c4U, 0x3162c4a6U, 0xd9af43ecU,
    0x356ad4beU, 0x18306050U, 0x0204080cU, 0x64c88d45U,
    0xf2f9ef16U, 0xf1ffe31cU, 0x56ac45e9U, 0xcd871394U,
    0x8219322bU, 0xc88d078aU, 0xba69d2bbU, 0xf0fde71aU,
    0xefc39b58U, 0xe9cf834cU, 0xe8cd874aU, 0xfde7d334U,
    0x890f1e11U, 0xd7b37bc8U, 0xc7933ba8U, 0xb577ee99U,
    0xa455aaffU, 0x2f5ebce2U, 0x95376e59U, 0x13264c6aU,
    0x0b162c3aU, 0xf3fbeb10U, 0xe0dda77aU, 0x376edcb2U,
};

static const u32 T1[256] = {
    0x53a7f5a6U, 0xbbd3d06bU, 0xd1e66ebfU, 0xe2713bd9U,
    0xbdd0da67U, 0x45accf8aU, 0x9a4db329U, 0xf2790bf9U,
    0x743a9ce8U, 0x8fc98c03U, 0x3f91417eU, 0xe5fc32d7U,
    0x3c1e4478U, 0x8e478f01U, 0xa854e54dU, 0x67bda9ceU,
    0x058c0f0aU, 0x57a5f9aeU, 0xf47a01f5U, 0xebfb20cbU,
    0xc6635791U, 0x6db8b7daU, 0xa7ddf453U, 0xb5d4c277U,
    0xd7e564b3U, 0x7bb38df6U, 0x97c5a433U, 0x61bea3c2U,
    0x4fa9d19eU, 0x0d88171aU, 0x180c2830U, 0x59a2ebb2U,
    0x723996e4U, 0xa3dff85bU, 0x5229f6a4U, 0xa9dae64fU,
    0x562bfaacU, 0x4da8d79aU, 0x8bcb800bU, 0x984cb52dU,
    0x964ba731U, 0x4422cc88U, 0x49aadb92U, 0x4824d890U,
    0x82419b19U, 0xe0703dddU, 0x51a6f3a2U, 0xeff92cc3U,
    0xb45ac175U, 0xd9e276afU, 0x7db087faU, 0x6c36b4d8U,
    0xfa7d13e9U, 0xd5e462b7U, 0x6633aaccU, 0xe3ff38dbU,
    0xc0605d9dU, 0x4020c080U, 0x10083020U, 0x0b8b1d16U,
    0xbc5ed965U, 0x4babdd96U, 0xfe7f1fe1U, 0xf0780dfdU,
    0xf87c15edU, 0x582ce8b0U, 0xae57ef41U, 0xb9d2d66fU,
    0xa5dcf257U, 0xda6d73a9U, 0xfc7e19e5U, 0x1a0d2e34U,
    0xa653f751U, 0x35945f6aU, 0x9bc3b02bU, 0x5028f0a0U,
    0x4e27d29cU, 0x0c061418U, 0xbe5fdf61U, 0x47adc98eU,
    0xce674f81U, 0xb85cd56dU, 0xaa55e349U, 0x9048ad3dU,
    0x1c0e2438U, 0xa452f155U, 0xc9ea468fU, 0x84429115U,
    0xb65bc771U, 0xba5dd369U, 0x6030a0c0U, 0xb058cd7dU,
    0xa251fb59U, 0xb259cb79U, 0x783c88f0U, 0x9c4eb925U,
    0x703890e0U, 0x098a1b12U, 0xe47231d5U, 0x28147850U,
    0xd3e768bbU, 0x91c6ae3fU, 0xa1defe5fU, 0xa050fd5dU,
    0x018e0302U, 0x39924b72U, 0xbfd1dc63U, 0xee772fc1U,
    0x3b934d76U, 0x8a458309U, 0x299a7b52U, 0x81ce9e1fU,
    0x5a2deeb4U, 0x06030a0cU, 0xc4625195U, 0x71b693e2U,
    0x6fb9b1deU, 0x63bfa5c6U, 0x31965362U, 0xd66b67b1U,
    0x7e3f82fcU, 0x0e07121cU, 0x24126c48U, 0x41aec382U,
    0x80409d1dU, 0x6834b8d0U, 0x8c468905U, 0x7c3e84f8U,
    0xabdbe04bU, 0x83cf981bU, 0xc5ec5297U, 0x85cc9217U,
    0x9fc1bc23U, 0x5fa1e1beU, 0x9dc0ba27U, 0xb1d6ce7fU,
    0x3a1d4e74U, 0xf5f402f7U, 0xc2615b99U, 0x763b9aecU,
    0x20106040U, 0xadd8ea47U, 0xd0686dbdU, 0x5da0e7baU,
    0x7fb181feU, 0x140a3c28U, 0xd2696bb9U, 0xd86c75adU,
    0x9249ab39U, 0xe9fa26cfU, 0xec7629c5U, 0x95c4a237U,
    0x219e6342U, 0x2b9b7d56U, 0xdc6e79a5U, 0x2f99715eU,
    0x99c2b62fU, 0x73b795e6U, 0x2d98775aU, 0x65bcafcaU,
    0x038f0506U, 0x1785392eU, 0x3e1f427cU, 0x75b49feaU,
    0xedf82ac7U, 0x22116644U, 0x5c2ee4b8U, 0x00000000U,
    0x4a25de94U, 0x381c4870U, 0x542afca8U, 0x7a3d8ef4U,
    0x0a051e14U, 0x9e4fbf21U, 0xf67b07f1U, 0x79b28bf2U,
    0x6432acc8U, 0x3d90477aU, 0x43afc586U, 0x32195664U,
    0x5ba3edb6U, 0xf3f708fbU, 0xe67337d1U, 0x279d694eU,
    0x2a157e54U, 0xe87425cdU, 0xc1ee5e9fU, 0x89ca860fU,
    0x239f6546U, 0x1e0f223cU, 0x361b5a6cU, 0xea7523c9U,
    0x11863322U, 0x15843f2aU, 0x259c6f4aU, 0x944aa135U,
    0x33975566U, 0x341a5c68U, 0xca654389U, 0xf1f60effU,
    0xc7ed5493U, 0x12093624U, 0x6bbbbdd6U, 0x4c26d498U,
    0x1b832d36U, 0xcbeb408bU, 0xde6f7fa1U, 0x1f81213eU,
    0x08041810U, 0xd46a61b5U, 0x86439711U, 0x02010604U,
    0x2e17725cU, 0xdfe17ca3U, 0x13873526U, 0xf7f504f3U,
    0x078d090eU, 0xdbe370abU, 0x4623ca8cU, 0x1d80273aU,
    0x8844850dU, 0x2c167458U, 0xcc664985U, 0x4221c684U,
    0xe1fe3edfU, 0xb7d5c473U, 0x6231a6c4U, 0xafd9ec43U,
    0x6a35bed4U, 0x30185060U, 0x04020c08U, 0xc864458dU,
    0xf9f216efU, 0xfff11ce3U, 0xac56e945U, 0x87cd9413U,
    0x19822b32U, 0x8dc88a07U, 0x69babbd2U, 0xfdf01ae7U,
    0xc3ef589bU, 0xcfe94c83U, 0xcde84a87U, 0xe7fd34d3U,
    0x0f89111eU, 0xb3d7c87bU, 0x93c7a83bU, 0x77b599eeU,
    0x55a4ffaaU, 0x5e2fe2bcU, 0x3795596eU, 0x26136a4cU,
    0x160b3a2cU, 0xfbf310ebU, 0xdde07aa7U, 0x6e37b2dcU,
};

static const u32 T2[256] = {
    0xa6f5a753U, 0x6bd0d3bbU, 0xbf6ee6d1U, 0xd93b71e2U,
    0x67dad0bdU, 0x8acfac45U, 0x29b34d9aU, 0xf90b79f2U,
    0xe89c3a74U, 0x038cc98fU, 0x7e41913fU, 0xd732fce5U,
    0x78441e3cU, 0x018f478eU, 0x4de554a8U, 0xcea9bd67U,
    0x0a0f8c05U, 0xaef9a557U, 0xf5017af4U, 0xcb20fbebU,
    0x915763c6U, 0xdab7b86dU, 0x53f4dda7U, 0x77c2d4b5U,
    0xb364e5d7U, 0xf68db37bU, 0x33a4c597U, 0xc2a3be61U,
    0x9ed1a94fU, 0x1a17880dU, 0x30280c18U, 0xb2eba259U,
    0xe4963972U, 0x5bf8dfa3U, 0xa4f62952U, 0x4fe6daa9U,
    0xacfa2b56U, 0x9ad7a84dU, 0x0b80cb8bU, 0x2db54c98U,
    0x31a74b96U, 0x88cc2244U, 0x92dbaa49U, 0x90d82448U,
    0x199b4182U, 0xdd3d70e0U, 0xa2f3a651U, 0xc32cf9efU,
    0x75c15ab4U, 0xaf76e2d9U, 0xfa87b07dU, 0xd8b4366cU,
    0xe9137dfaU, 0xb762e4d5U, 0xccaa3366U, 0xdb38ffe3U,
    0x9d5d60c0U, 0x80c02040U, 0x20300810U, 0x161d8b0bU,
    0x65d95ebcU, 0x96ddab4bU, 0xe11f7ffeU, 0xfd0d78f0U,
    0xed157cf8U, 0xb0e82c58U, 0x41ef57aeU, 0x6fd6d2b9U,
    0x57f2dca5U, 0xa9736ddaU, 0xe5197efcU, 0x342e0d1aU,
    0x51f753a6U, 0x6a5f9435U, 0x2bb0c39bU, 0xa0f02850U,
    0x9cd2274eU, 0x1814060cU, 0x61df5fbeU, 0x8ec9ad47U,
    0x814f67ceU, 0x6dd55cb8U, 0x49e355aaU, 0x3dad4890U,
    0x38240e1cU, 0x55f152a4U, 0x8f46eac9U, 0x15914284U,
    0x71c75bb6U, 0x69d35dbaU, 0xc0a03060U, 0x7dcd58b0U,
    0x59fb51a2U, 0x79cb59b2U, 0xf0883c78U, 0x25b94e9cU,
    0xe0903870U, 0x121b8a09U, 0xd53172e4U, 0x50781428U,
    0xbb68e7d3U, 0x3faec691U, 0x5ffedea1U, 0x5dfd50a0U,
    0x02038e01U, 0x724b9239U, 0x63dcd1bfU, 0xc12f77eeU,
    0x764d933bU, 0x0983458aU, 0x527b9a29U, 0x1f9ece81U,
    0xb4ee2d5aU, 0x0c0a0306U, 0x955162c4U, 0xe293b671U,
    0xdeb1b96fU, 0xc6a5bf63U, 0x62539631U, 0xb1676bd6U,
    0xfc823f7eU, 0x1c12070eU, 0x486c1224U, 0x82c3ae41U,
    0x1d9d4080U, 0xd0b83468U, 0x0589468cU, 0xf8843e7cU,
    0x4be0dbabU, 0x1b98cf83U, 0x9752ecc5U, 0x1792cc85U,
    0x23bcc19fU, 0xbee1a15fU, 0x27bac09dU, 0x7fced6b1U,
    0x744e1d3aU, 0xf702f4f5U, 0x995b61c2U, 0xec9a3b76U,
    0x40601020U, 0x47ead8adU, 0xbd6d68d0U, 0xbae7a05dU,
    0xfe81b17fU, 0x283c0a14U, 0xb96b69d2U, 0xad756cd8U,
    0x39ab4992U, 0xcf26fae9U, 0xc52976ecU, 0x37a2c495U,
    0x42639e21U, 0x567d9b2bU, 0xa5796edcU, 0x5e71992fU,
    0x2fb6c299U, 0xe695b773U, 0x5a77982dU, 0xcaafbc65U,
    0x06058f03U, 0x2e398517U, 0x7c421f3eU, 0xea9fb475U,
    0xc72af8edU, 0x44661122U, 0xb8e42e5cU, 0x00000000U,
    0x94de254aU, 0x70481c38U, 0xa8fc2a54U, 0xf48e3d7aU,
    0x141e050aU, 0x21bf4f9eU, 0xf1077bf6U, 0xf28bb279U,
    0xc8ac3264U, 0x7a47903dU, 0x86c5af43U, 0x64561932U,
    0xb6eda35bU, 0xfb08f7f3U, 0xd13773e6U, 0x4e699d27U,
    0x547e152aU, 0xcd2574e8U, 0x9f5eeec1U, 0x0f86ca89U,
    0x46659f23U, 0x3c220f1eU, 0x6c5a1b36U, 0xc92375eaU,
    0x22338611U, 0x2a3f8415U, 0x4a6f9c25U, 0x35a14a94U,
    0x66559733U, 0x685c1a34U, 0x894365caU, 0xff0ef6f1U,
    0x9354edc7U, 0x24360912U, 0xd6bdbb6bU, 0x98d4264cU,
    0x362d831bU, 0x8b40ebcbU, 0xa17f6fdeU, 0x3e21811fU,
    0x10180408U, 0xb5616ad4U, 0x11974386U, 0x04060102U,
    0x5c72172eU, 0xa37ce1dfU, 0x26358713U, 0xf304f5f7U,
    0x0e098d07U, 0xab70e3dbU, 0x8cca2346U, 0x3a27801dU,
    0x0d854488U, 0x5874162cU, 0x854966ccU, 0x84c62142U,
    0xdf3efee1U, 0x73c4d5b7U, 0xc4a63162U, 0x43ecd9afU,
    0xd4be356aU, 0x60501830U, 0x080c0204U, 0x8d4564c8U,
    0xef16f2f9U, 0xe31cf1ffU, 0x45e956acU, 0x1394cd87U,
    0x322b8219U, 0x078ac88dU, 0xd2bbba69U, 0xe71af0fdU,
    0x9b58efc3U, 0x834ce9cfU, 0x874ae8cdU, 0xd334fde7U,
    0x1e11890fU, 0x7bc8d7b3U, 0x3ba8c793U, 0xee99b577U,
    0xaaffa455U, 0xbce22f5eU, 0x6e599537U, 0x4c6a1326U,
    0x2c3a0b16U, 0xeb10f3fbU, 0xa77ae0ddU, 0xdcb2376eU,
};

static const u32 T3[256] = {
    0xf5a653a7U, 0xd06bbbd3U, 0x6ebfd1e6U, 0x3bd9e271U,
    0xda67bdd0U, 0xcf8a45acU, 0xb3299a4dU, 0x0bf9f279U,
    0x9ce8743aU, 0x8c038fc9U, 0x417e3f91U, 0x32d7e5fcU,
    0x44783c1eU, 0x8f018e47U, 0xe54da854U, 0xa9ce67bdU,
    0x0f0a058cU, 0xf9ae57a5U, 0x01f5f47aU, 0x20cbebfbU,
    0x5791c663U, 0xb7da6db8U, 0xf453a7ddU, 0xc277b5d4U,
    0x64b3d7e5U, 0x8df67bb3U, 0xa43397c5U, 0xa3c261beU,
    0xd19e4fa9U, 0x171a0d88U, 0x2830180cU, 0xebb259a2U,
    0x96e47239U, 0xf85ba3dfU, 0xf6a45229U, 0xe64fa9daU,
    0xfaac562bU, 0xd79a4da8U, 0x800b8bcbU, 0xb52d984cU,
    0xa731964bU, 0xcc884422U, 0xdb9249aaU, 0xd8904824U,
    0x9b198241U, 0x3ddde070U, 0xf3a251a6U, 0x2cc3eff9U,
    0xc175b45aU, 0x76afd9e2U, 0x87fa7db0U, 0xb4d86c36U,
    0x13e9fa7dU, 0x62b7d5e4U, 0xaacc6633U, 0x38dbe3ffU,
    0x5d9dc060U, 0xc0804020U, 0x30201008U, 0x1d160b8bU,
    0xd965bc5eU, 0xdd964babU, 0x1fe1fe7fU, 0x0dfdf078U,
    0x15edf87cU, 0xe8b0582cU, 0xef41ae57U, 0xd66fb9d2U,
    0xf257a5dcU, 0x73a9da6dU, 0x19e5fc7eU, 0x2e341a0dU,
    0xf751a653U, 0x5f6a3594U, 0xb02b9bc3U, 0xf0a05028U,
    0xd29c4e27U, 0x14180c06U, 0xdf61be5fU, 0xc98e47adU,
    0x4f81ce67U, 0xd56db85cU, 0xe349aa55U, 0xad3d9048U,
    0x24381c0eU, 0xf155a452U, 0x468fc9eaU, 0x91158442U,
    0xc771b65bU, 0xd369ba5dU, 0xa0c06030U, 0xcd7db058U,
    0xfb59a251U, 0xcb79b259U, 0x88f0783cU, 0xb9259c4eU,
    0x90e07038U, 0x1b12098aU, 0x31d5e472U, 0x78502814U,
    0x68bbd3e7U, 0xae3f91c6U, 0xfe5fa1deU, 0xfd5da050U,
    0x0302018eU, 0x4b723992U, 0xdc63bfd1U, 0x2fc1ee77U,
    0x4d763b93U, 0x83098a45U, 0x7b52299aU, 0x9e1f81ceU,
    0xeeb45a2dU, 0x0a0c0603U, 0x5195c462U, 0x93e271b6U,
    0xb1de6fb9U, 0xa5c663bfU, 0x53623196U, 0x67b1d66bU,
    0x82fc7e3fU, 0x121c0e07U, 0x6c482412U, 0xc38241aeU,
    0x9d1d8040U, 0xb8d06834U, 0x89058c46U, 0x84f87c3eU,
    0xe04babdbU, 0x981b83cfU, 0x5297c5ecU, 0x921785ccU,
    0xbc239fc1U, 0xe1be5fa1U, 0xba279dc0U, 0xce7fb1d6U,
    0x4e743a1dU, 0x02f7f5f4U, 0x5b99c261U, 0x9aec763bU,
    0x60402010U, 0xea47add8U, 0x6dbdd068U, 0xe7ba5da0U,
    0x81fe7fb1U, 0x3c28140aU, 0x6bb9d269U, 0x75add86cU,
    0xab399249U, 0x26cfe9faU, 0x29c5ec76U, 0xa23795c4U,
    0x6342219eU, 0x7d562b9bU, 0x79a5dc6eU, 0x715e2f99U,
    0xb62f99c2U, 0x95e673b7U, 0x775a2d98U, 0xafca65bcU,
    0x0506038fU, 0x392e1785U, 0x427c3e1fU, 0x9fea75b4U,
    0x2ac7edf8U, 0x66442211U, 0xe4b85c2eU, 0x00000000U,
    0xde944a25U, 0x4870381cU, 0xfca8542aU, 0x8ef47a3dU,
    0x1e140a05U, 0xbf219e4fU, 0x07f1f67bU, 0x8bf279b2U,
    0xacc86432U, 0x477a3d90U, 0xc58643afU, 0x56643219U,
    0xedb65ba3U, 0x08fbf3f7U, 0x37d1e673U, 0x694e279dU,
    0x7e542a15U, 0x25cde874U, 0x5e9fc1eeU, 0x860f89caU,
    0x6546239fU, 0x223c1e0fU, 0x5a6c361bU, 0x23c9ea75U,
    0x33221186U, 0x3f2a1584U, 0x6f4a259cU, 0xa135944aU,
    0x55663397U, 0x5c68341aU, 0x4389ca65U, 0x0efff1f6U,
    0x5493c7edU, 0x36241209U, 0xbdd66bbbU, 0xd4984c26U,
    0x2d361b83U, 0x408bcbebU, 0x7fa1de6fU, 0x213e1f81U,
    0x18100804U, 0x61b5d46aU, 0x97118643U, 0x06040201U,
    0x725c2e17U, 0x7ca3dfe1U, 0x35261387U, 0x04f3f7f5U,
    0x090e078dU, 0x70abdbe3U, 0xca8c4623U, 0x273a1d80U,
    0x850d8844U, 0x74582c16U, 0x4985cc66U, 0xc6844221U,
    0x3edfe1feU, 0xc473b7d5U, 0xa6c46231U, 0xec43afd9U,
    0xbed46a35U, 0x50603018U, 0x0c080402U, 0x458dc864U,
    0x16eff9f2U, 0x1ce3fff1U, 0xe945ac56U, 0x941387cdU,
    0x2b321982U, 0x8a078dc8U, 0xbbd269baU, 0x1ae7fdf0U,
    0x589bc3efU, 0x4c83cfe9U, 0x4a87cde8U, 0x34d3e7fdU,
    0x111e0f89U, 0xc87bb3d7U, 0xa83b93c7U, 0x99ee77b5U,
    0xffaa55a4U, 0xe2bc5e2fU, 0x596e3795U, 0x6a4c2613U,
    0x3a2c160bU, 0x10ebfbf3U, 0x7aa7dde0U, 0xb2dc6e37U,
};

static const u32 T4[256] = {
    0xa7a7a7a7U, 0xd3d3d3d3U, 0xe6e6e6e6U, 0x71717171U,
    0xd0d0d0d0U, 0xacacacacU, 0x4d4d4d4dU, 0x79797979U,
    0x3a3a3a3aU, 0xc9c9c9c9U, 0x91919191U, 0xfcfcfcfcU,
    0x1e1e1e1eU, 0x47474747U, 0x54545454U, 0xbdbdbdbdU,
    0x8c8c8c8cU, 0xa5a5a5a5U, 0x7a7a7a7aU, 0xfbfbfbfbU,
    0x63636363U, 0xb8b8b8b8U, 0xddddddddU, 0xd4d4d4d4U,
    0xe5e5e5e5U, 0xb3b3b3b3U, 0xc5c5c5c5U, 0xbebebebeU,
    0xa9a9a9a9U, 0x88888888U, 0x0c0c0c0cU, 0xa2a2a2a2U,
    0x39393939U, 0xdfdfdfdfU, 0x29292929U, 0xdadadadaU,
    0x2b2b2b2bU, 0xa8a8a8a8U, 0xcbcbcbcbU, 0x4c4c4c4cU,
    0x4b4b4b4bU, 0x22222222U, 0xaaaaaaaaU, 0x24242424U,
    0x41414141U, 0x70707070U, 0xa6a6a6a6U, 0xf9f9f9f9U,
    0x5a5a5a5aU, 0xe2e2e2e2U, 0xb0b0b0b0U, 0x36363636U,
    0x7d7d7d7dU, 0xe4e4e4e4U, 0x33333333U, 0xffffffffU,
    0x60606060U, 0x20202020U, 0x08080808U, 0x8b8b8b8bU,
    0x5e5e5e5eU, 0xababababU, 0x7f7f7f7fU, 0x78787878U,
    0x7c7c7c7cU, 0x2c2c2c2cU, 0x57575757U, 0xd2d2d2d2U,
    0xdcdcdcdcU, 0x6d6d6d6dU, 0x7e7e7e7eU, 0x0d0d0d0dU,
    0x53535353U, 0x94949494U, 0xc3c3c3c3U, 0x28282828U,
    0x27272727U, 0x06060606U, 0x5f5f5f5fU, 0xadadadadU,
    0x67676767U, 0x5c5c5c5cU, 0x55555555U, 0x48484848U,
    0x0e0e0e0eU, 0x52525252U, 0xeaeaeaeaU, 0x42424242U,
    0x5b5b5b5bU, 0x5d5d5d5dU, 0x30303030U, 0x58585858U,
    0x51515151U, 0x59595959U, 0x3c3c3c3cU, 0x4e4e4e4eU,
    0x38383838U, 0x8a8a8a8aU, 0x72727272U, 0x14141414U,
    0xe7e7e7e7U, 0xc6c6c6c6U, 0xdedededeU, 0x50505050U,
    0x8e8e8e8eU, 0x92929292U, 0xd1d1d1d1U, 0x77777777U,
    0x93939393U, 0x45454545U, 0x9a9a9a9aU, 0xcecececeU,
    0x2d2d2d2dU, 0x03030303U, 0x62626262U, 0xb6b6b6b6U,
    0xb9b9b9b9U, 0xbfbfbfbfU, 0x96969696U, 0x6b6b6b6bU,
    0x3f3f3f3fU, 0x07070707U, 0x12121212U, 0xaeaeaeaeU,
    0x40404040U, 0x34343434U, 0x46464646U, 0x3e3e3e3eU,
    0xdbdbdbdbU, 0xcfcfcfcfU, 0xececececU, 0xccccccccU,
    0xc1c1c1c1U, 0xa1a1a1a1U, 0xc0c0c0c0U, 0xd6d6d6d6U,
    0x1d1d1d1dU, 0xf4f4f4f4U, 0x61616161U, 0x3b3b3b3bU,
    0x10101010U, 0xd8d8d8d8U, 0x68686868U, 0xa0a0a0a0U,
    0xb1b1b1b1U, 0x0a0a0a0aU, 0x69696969U, 0x6c6c6c6cU,
    0x49494949U, 0xfafafafaU, 0x76767676U, 0xc4c4c4c4U,
    0x9e9e9e9eU, 0x9b9b9b9bU, 0x6e6e6e6eU, 0x99999999U,
    0xc2c2c2c2U, 0xb7b7b7b7U, 0x98989898U, 0xbcbcbcbcU,
    0x8f8f8f8fU, 0x85858585U, 0x1f1f1f1fU, 0xb4b4b4b4U,
    0xf8f8f8f8U, 0x11111111U, 0x2e2e2e2eU, 0x00000000U,
    0x25252525U, 0x1c1c1c1cU, 0x2a2a2a2aU, 0x3d3d3d3dU,
    0x05050505U, 0x4f4f4f4fU, 0x7b7b7b7bU, 0xb2b2b2b2U,
    0x32323232U, 0x90909090U, 0xafafafafU, 0x19191919U,
    0xa3a3a3a3U, 0xf7f7f7f7U, 0x73737373U, 0x9d9d9d9dU,
    0x15151515U, 0x74747474U, 0xeeeeeeeeU, 0xcacacacaU,
    0x9f9f9f9fU, 0x0f0f0f0fU, 0x1b1b1b1bU, 0x75757575U,
    0x86868686U, 0x84848484U, 0x9c9c9c9cU, 0x4a4a4a4aU,
    0x97979797U, 0x1a1a1a1aU, 0x65656565U, 0xf6f6f6f6U,
    0xededededU, 0x09090909U, 0xbbbbbbbbU, 0x26262626U,
    0x83838383U, 0xebebebebU, 0x6f6f6f6fU, 0x81818181U,
    0x04040404U, 0x6a6a6a6aU, 0x43434343U, 0x01010101U,
    0x17171717U, 0xe1e1e1e1U, 0x87878787U, 0xf5f5f5f5U,
    0x8d8d8d8dU, 0xe3e3e3e3U, 0x23232323U, 0x80808080U,
    0x44444444U, 0x16161616U, 0x66666666U, 0x21212121U,
    0xfefefefeU, 0xd5d5d5d5U, 0x31313131U, 0xd9d9d9d9U,
    0x35353535U, 0x18181818U, 0x02020202U, 0x64646464U,
    0xf2f2f2f2U, 0xf1f1f1f1U, 0x56565656U, 0xcdcdcdcdU,
    0x82828282U, 0xc8c8c8c8U, 0xbabababaU, 0xf0f0f0f0U,
    0xefefefefU, 0xe9e9e9e9U, 0xe8e8e8e8U, 0xfdfdfdfdU,
    0x89898989U, 0xd7d7d7d7U, 0xc7c7c7c7U, 0xb5b5b5b5U,
    0xa4a4a4a4U, 0x2f2f2f2fU, 0x95959595U, 0x13131313U,
    0x0b0b0b0bU, 0xf3f3f3f3U, 0xe0e0e0e0U, 0x37373737U,
};

static const u32 T5[256] = {
    0x00000000U, 0x01020608U, 0x02040c10U, 0x03060a18U,
    0x04081820U, 0x050a1e28U, 0x060c1430U, 0x070e1238U,
    0x08103040U, 0x09123648U, 0x0a143c50U, 0x0b163a58U,
    0x0c182860U, 0x0d1a2e68U, 0x0e1c2470U, 0x0f1e2278U,
    0x10206080U, 0x11226688U, 0x12246c90U, 0x13266a98U,
    0x142878a0U, 0x152a7ea8U, 0x162c74b0U, 0x172e72b8U,
    0x183050c0U, 0x193256c8U, 0x1a345cd0U, 0x1b365ad8U,
    0x1c3848e0U, 0x1d3a4ee8U, 0x1e3c44f0U, 0x1f3e42f8U,
    0x2040c01dU, 0x2142c615U, 0x2244cc0dU, 0x2346ca05U,
    0x2448d83dU, 0x254ade35U, 0x264cd42dU, 0x274ed225U,
    0x2850f05dU, 0x2952f655U, 0x2a54fc4dU, 0x2b56fa45U,
    0x2c58e87dU, 0x2d5aee75U, 0x2e5ce46dU, 0x2f5ee265U,
    0x3060a09dU, 0x3162a695U, 0x3264ac8dU, 0x3366aa85U,
    0x3468b8bdU, 0x356abeb5U, 0x366cb4adU, 0x376eb2a5U,
    0x387090ddU, 0x397296d5U, 0x3a749ccdU, 0x3b769ac5U,
    0x3c7888fdU, 0x3d7a8ef5U, 0x3e7c84edU, 0x3f7e82e5U,
    0x40809d3aU, 0x41829b32U, 0x4284912aU, 0x43869722U,
    0x4488851aU, 0x458a8312U, 0x468c890aU, 0x478e8f02U,
    0x4890ad7aU, 0x4992ab72U, 0x4a94a16aU, 0x4b96a762U,
    0x4c98b55aU, 0x4d9ab352U, 0x4e9cb94aU, 0x4f9ebf42U,
    0x50a0fdbaU, 0x51a2fbb2U, 0x52a4f1aaU, 0x53a6f7a2U,
    0x54a8e59aU, 0x55aae392U, 0x56ace98aU, 0x57aeef82U,
    0x58b0cdfaU, 0x59b2cbf2U, 0x5ab4c1eaU, 0x5bb6c7e2U,
    0x5cb8d5daU, 0x5dbad3d2U, 0x5ebcd9caU, 0x5fbedfc2U,
    0x60c05d27U, 0x61c25b2fU, 0x62c45137U, 0x63c6573fU,
    0x64c84507U, 0x65ca430fU, 0x66cc4917U, 0x67ce4f1fU,
    0x68d06d67U, 0x69d26b6fU, 0x6ad46177U, 0x6bd6677fU,
    0x6cd87547U, 0x6dda734fU, 0x6edc7957U, 0x6fde7f5fU,
    0x70e03da7U, 0x71e23bafU, 0x72e431b7U, 0x73e637bfU,
    0x74e82587U, 0x75ea238fU, 0x76ec2997U, 0x77ee2f9fU,
    0x78f00de7U, 0x79f20befU, 0x7af401f7U, 0x7bf607ffU,
    0x7cf815c7U, 0x7dfa13cfU, 0x7efc19d7U, 0x7ffe1fdfU,
    0x801d2774U, 0x811f217cU, 0x82192b64U, 0x831b2d6cU,
    0x84153f54U, 0x8517395cU, 0x86113344U, 0x8713354cU,
    0x880d1734U, 0x890f113cU, 0x8a091b24U, 0x8b0b1d2cU,
    0x8c050f14U, 0x8d07091cU, 0x8e010304U, 0x8f03050cU,
    0x903d47f4U, 0x913f41fcU, 0x92394be4U, 0x933b4decU,
    0x94355fd4U, 0x953759dcU, 0x963153c4U, 0x973355ccU,
    0x982d77b4U, 0x992f71bcU, 0x9a297ba4U, 0x9b2b7dacU,
    0x9c256f94U, 0x9d27699cU, 0x9e216384U, 0x9f23658cU,
    0xa05de769U, 0xa15fe161U, 0xa259eb79U, 0xa35bed71U,
    0xa455ff49U, 0xa557f941U, 0xa651f359U, 0xa753f551U,
    0xa84dd729U, 0xa94fd121U, 0xaa49db39U, 0xab4bdd31U,
    0xac45cf09U, 0xad47c901U, 0xae41c319U, 0xaf43c511U,
    0xb07d87e9U, 0xb17f81e1U, 0xb2798bf9U, 0xb37b8df1U,
    0xb4759fc9U, 0xb57799c1U, 0xb67193d9U, 0xb77395d1U,
    0xb86db7a9U, 0xb96fb1a1U, 0xba69bbb9U, 0xbb6bbdb1U,
    0xbc65af89U, 0xbd67a981U, 0xbe61a399U, 0xbf63a591U,
    0xc09dba4eU, 0xc19fbc46U, 0xc299b65eU, 0xc39bb056U,
    0xc495a26eU, 0xc597a466U, 0xc691ae7eU, 0xc793a876U,
    0xc88d8a0eU, 0xc98f8c06U, 0xca89861eU, 0xcb8b8016U,
    0xcc85922eU, 0xcd879426U, 0xce819e3eU, 0xcf839836U,
    0xd0bddaceU, 0xd1bfdcc6U, 0xd2b9d6deU, 0xd3bbd0d6U,
    0xd4b5c2eeU, 0xd5b7c4e6U, 0xd6b1cefeU, 0xd7b3c8f6U,
    0xd8adea8eU, 0xd9afec86U, 0xdaa9e69eU, 0xdbabe096U,
    0xdca5f2aeU, 0xdda7f4a6U, 0xdea1febeU, 0xdfa3f8b6U,
    0xe0dd7a53U, 0xe1df7c5bU, 0xe2d97643U, 0xe3db704bU,
    0xe4d56273U, 0xe5d7647bU, 0xe6d16e63U, 0xe7d3686bU,
    0xe8cd4a13U, 0xe9cf4c1bU, 0xeac94603U, 0xebcb400bU,
    0xecc55233U, 0xedc7543bU, 0xeec15e23U, 0xefc3582bU,
    0xf0fd1ad3U, 0xf1ff1cdbU, 0xf2f916c3U, 0xf3fb10cbU,
    0xf4f502f3U, 0xf5f704fbU, 0xf6f10ee3U, 0xf7f308ebU,
    0xf8ed2a93U, 0xf9ef2c9bU, 0xfae92683U, 0xfbeb208bU,
    0xfce532b3U, 0xfde734bbU, 0xfee13ea3U, 0xffe338abU,
};

/**
 * The round constants.
 */
static const u32 rc_anubis[] = {
    0xa7d3e671U, 0xd0ac4d79U, 0x3ac991fcU, 0x1e4754bdU,
	0x8ca57afbU, 0x63b8ddd4U, 0xe5b3c5beU, 0xa9880ca2U,
	0x39df29daU, 0x2ba8cb4cU, 0x4b22aa24U, 0x4170a6f9U,
	0x5ae2b036U, 0x7de433ffU, 0x6020088bU, 0x5eab7f78U,
	0x7c2c57d2U, 0xdc6d7e0dU, 0x5394c328U,
};


/**
 * Create the Anubis key schedule for a given cipher key.
 * Both encryption and decryption key schedules are generated.
 * 
 * @param key			The 32N-bit cipher key.
 * @param structpointer	Pointer to the structure that will hold the expanded key.
 */
void NESSIEkeysetup(const unsigned char * const key,
					struct NESSIEstruct_anubis * const structpointer) {

	int N, R, i, pos, r;
	u32 kappa[MAX_N];
	u32 inter[MAX_N];

	structpointer->keyBits = KEYSIZEB*8;

	/*
	 * determine the N length parameter:
	 * (N.B. it is assumed that the key length is valid!)
	 */
	N = structpointer->keyBits >> 5;

	/*
	 * determine number of rounds from key size:
	 */
	structpointer->R = R = 8 + N;

    /*
	 * map cipher key to initial key state (mu):
	 */
    for (i = 0, pos = 0; i < N; i++, pos += 4) {
		kappa[i] =
			(key[pos    ] << 24) ^
			(key[pos + 1] << 16) ^
			(key[pos + 2] <<  8) ^
			(key[pos + 3]      );
    }

	/*
	 * generate R + 1 round keys:
	 */
	for (r = 0; r <= R; r++) {
		u32 K0, K1, K2, K3;
		/*
		 * generate r-th round key K^r:
		 */
		K0 = T4[(kappa[N - 1] >> 24)       ];
		K1 = T4[(kappa[N - 1] >> 16) & 0xff];
		K2 = T4[(kappa[N - 1] >>  8) & 0xff];
		K3 = T4[(kappa[N - 1]      ) & 0xff];
		for (i = N - 2; i >= 0; i--) {
			K0 = T4[(kappa[i] >> 24)       ] ^
				(T5[(K0 >> 24)       ] & 0xff000000U) ^
				(T5[(K0 >> 16) & 0xff] & 0x00ff0000U) ^
				(T5[(K0 >>  8) & 0xff] & 0x0000ff00U) ^
				(T5[(K0      ) & 0xff] & 0x000000ffU);
			K1 = T4[(kappa[i] >> 16) & 0xff] ^
				(T5[(K1 >> 24)       ] & 0xff000000U) ^
				(T5[(K1 >> 16) & 0xff] & 0x00ff0000U) ^
				(T5[(K1 >>  8) & 0xff] & 0x0000ff00U) ^
				(T5[(K1      ) & 0xff] & 0x000000ffU);
			K2 = T4[(kappa[i] >>  8) & 0xff] ^
				(T5[(K2 >> 24)       ] & 0xff000000U) ^
				(T5[(K2 >> 16) & 0xff] & 0x00ff0000U) ^
				(T5[(K2 >>  8) & 0xff] & 0x0000ff00U) ^
				(T5[(K2      ) & 0xff] & 0x000000ffU);
			K3 = T4[(kappa[i]      ) & 0xff] ^
				(T5[(K3 >> 24)       ] & 0xff000000U) ^
				(T5[(K3 >> 16) & 0xff] & 0x00ff0000U) ^
				(T5[(K3 >>  8) & 0xff] & 0x0000ff00U) ^
				(T5[(K3      ) & 0xff] & 0x000000ffU);
		}
		/*
		-- this is the code to use with the large U tables:
		K0 = K1 = K2 = K3 = 0;
		for (i = 0; i < N; i++) {
			K0 ^= U[i][(kappa[i] >> 24)       ];
			K1 ^= U[i][(kappa[i] >> 16) & 0xff];
			K2 ^= U[i][(kappa[i] >>  8) & 0xff];
			K3 ^= U[i][(kappa[i]      ) & 0xff];
		}
		*/
		structpointer->roundKeyEnc[r][0] = K0;
		structpointer->roundKeyEnc[r][1] = K1;
		structpointer->roundKeyEnc[r][2] = K2;
		structpointer->roundKeyEnc[r][3] = K3;

		/*
		 * compute kappa^{r+1} from kappa^r:
		 */
		if (r == R) {
			break;
		}
		for (i = 0; i < N; i++) {
			int j = i;
			inter[i]  = T0[(kappa[j--] >> 24)       ]; if (j < 0) j = N - 1;
			inter[i] ^= T1[(kappa[j--] >> 16) & 0xff]; if (j < 0) j = N - 1;
			inter[i] ^= T2[(kappa[j--] >>  8) & 0xff]; if (j < 0) j = N - 1;
			inter[i] ^= T3[(kappa[j  ]      ) & 0xff];
		}
		kappa[0] = inter[0] ^ rc_anubis[r];
		for (i = 1; i < N; i++) {
			kappa[i] = inter[i];
		}
	}

	/*
	 * generate inverse key schedule: K'^0 = K^R, K'^R = K^0, K'^r = theta(K^{R-r}):
	 */
	for (i = 0; i < 4; i++) {
		structpointer->roundKeyDec[0][i] = structpointer->roundKeyEnc[R][i];
		structpointer->roundKeyDec[R][i] = structpointer->roundKeyEnc[0][i];
	}
	for (r = 1; r < R; r++) {
		for (i = 0; i < 4; i++) {
			u32 v = structpointer->roundKeyEnc[R - r][i];
			structpointer->roundKeyDec[r][i] =
				T0[T4[(v >> 24)       ] & 0xff] ^
				T1[T4[(v >> 16) & 0xff] & 0xff] ^
				T2[T4[(v >>  8) & 0xff] & 0xff] ^
				T3[T4[(v      ) & 0xff] & 0xff];
		}
	}
}

/**
 * Either encrypt or decrypt a data block, according to the key schedule.
 * 
 * @param	block		the data block to be encrypted/decrypted.
 * @param	roundKey	the key schedule to be used.
 * @param	R			number of rounds.
 */
static void crypt(const u8 plaintext[/*16*/], u8 ciphertext[/*16*/],
				  const u32 roundKey[MAX_ROUNDS + 1][4], int R) {
	int i, pos, r;
	u32 state[4];
	u32 inter[4];

    /*
	 * map plaintext block to cipher state (mu)
	 * and add initial round key (sigma[K^0]):
	 */
    for (i = 0, pos = 0; i < 4; i++, pos += 4) {
		state[i] =
			(plaintext[pos    ] << 24) ^
			(plaintext[pos + 1] << 16) ^
			(plaintext[pos + 2] <<  8) ^
			(plaintext[pos + 3]      ) ^
			roundKey[0][i];
    }

    /*
     * R - 1 full rounds:
     */
    for (r = 1; r < R; r++) {
		inter[0] =
			T0[(state[0] >> 24)       ] ^
			T1[(state[1] >> 24)       ] ^
			T2[(state[2] >> 24)       ] ^
			T3[(state[3] >> 24)       ] ^
			roundKey[r][0];
		inter[1] =
			T0[(state[0] >> 16) & 0xff] ^
			T1[(state[1] >> 16) & 0xff] ^
			T2[(state[2] >> 16) & 0xff] ^
			T3[(state[3] >> 16) & 0xff] ^
			roundKey[r][1];
		inter[2] =
			T0[(state[0] >>  8) & 0xff] ^
			T1[(state[1] >>  8) & 0xff] ^
			T2[(state[2] >>  8) & 0xff] ^
			T3[(state[3] >>  8) & 0xff] ^
			roundKey[r][2];
		inter[3] =
			T0[(state[0]      ) & 0xff] ^
			T1[(state[1]      ) & 0xff] ^
			T2[(state[2]      ) & 0xff] ^
			T3[(state[3]      ) & 0xff] ^
			roundKey[r][3];
		state[0] = inter[0];
		state[1] = inter[1];
		state[2] = inter[2];
		state[3] = inter[3];
    }

    /*
	 * last round:
	 */
	inter[0] =
		(T0[(state[0] >> 24)       ] & 0xff000000U) ^
		(T1[(state[1] >> 24)       ] & 0x00ff0000U) ^
		(T2[(state[2] >> 24)       ] & 0x0000ff00U) ^
		(T3[(state[3] >> 24)       ] & 0x000000ffU) ^
		roundKey[R][0];
	inter[1] =
		(T0[(state[0] >> 16) & 0xff] & 0xff000000U) ^
		(T1[(state[1] >> 16) & 0xff] & 0x00ff0000U) ^
		(T2[(state[2] >> 16) & 0xff] & 0x0000ff00U) ^
		(T3[(state[3] >> 16) & 0xff] & 0x000000ffU) ^
		roundKey[R][1];
	inter[2] =
		(T0[(state[0] >>  8) & 0xff] & 0xff000000U) ^
		(T1[(state[1] >>  8) & 0xff] & 0x00ff0000U) ^
		(T2[(state[2] >>  8) & 0xff] & 0x0000ff00U) ^
		(T3[(state[3] >>  8) & 0xff] & 0x000000ffU) ^
		roundKey[R][2];
	inter[3] =
		(T0[(state[0]      ) & 0xff] & 0xff000000U) ^
		(T1[(state[1]      ) & 0xff] & 0x00ff0000U) ^
		(T2[(state[2]      ) & 0xff] & 0x0000ff00U) ^
		(T3[(state[3]      ) & 0xff] & 0x000000ffU) ^
		roundKey[R][3];

	/*
	 * map cipher state to ciphertext block (mu^{-1}):
	 */
    for (i = 0, pos = 0; i < 4; i++, pos += 4) {
        u32 w = inter[i];
        ciphertext[pos    ] = (u8)(w >> 24);
        ciphertext[pos + 1] = (u8)(w >> 16);
        ciphertext[pos + 2] = (u8)(w >>  8);
        ciphertext[pos + 3] = (u8)(w      );
    }
}


/**
 * Encrypt a data block.
 * 
 * @param	structpointer	the expanded key.
 * @param	plaintext		the data block to be encrypted.
 * @param	ciphertext		the encrypted data block.
 */
void NESSIEencrypt(const struct NESSIEstruct_anubis * const structpointer,
				   const unsigned char * const plaintext,
				         unsigned char * const ciphertext) {
	crypt(plaintext, ciphertext, structpointer->roundKeyEnc, structpointer->R);
}

/**
 * Decrypt a data block.
 * 
 * @param	structpointer	the expanded key.
 * @param	ciphertext		the data block to be decrypted.
 * @param	plaintext		the decrypted data block.
 */
void NESSIEdecrypt(const struct NESSIEstruct_anubis * const structpointer,
				   const unsigned char * const ciphertext,
				         unsigned char * const plaintext) {
	crypt(ciphertext, plaintext, structpointer->roundKeyDec, structpointer->R);
}












	// SEED
		// SEED test routine:
/*
AlgorithmType: SymmetricCipher
Name: SEED/ECB
Source: RFC 4269
Key:        00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
Plaintext:  00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F
Ciphertext: 5E BA C6 E0 05 4E 16 68 19 AF F1 CC 6D 34 6C DB
Test: Encrypt
Key:        00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F
Plaintext:  00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
Ciphertext: C1 1F 22 F2 01 40 50 50 84 48 35 97 E4 37 0F 43
Test: Encrypt
Key:        47 06 48 08 51 E6 1B E8 5D 74 BF B3 FD 95 61 85
Plaintext:  83 A2 F8 A2 88 64 1F B9 A4 E9 A5 CC 2F 13 1C 7D
Ciphertext: EE 54 D1 3E BC AE 70 6D 22 6B C3 14 2C D4 0D 4A
Test: Encrypt
Key:        28 DB C3 BC 49 FF D8 7D CF A5 09 B1 1D 42 2B E7
Plaintext:  B4 1E 6B E2 EB A8 4A 14 8E 2E ED 84 59 3C 5E C7
Ciphertext: 9B 9B 7B FC D1 81 3C B9 5D 0B 36 18 F4 0F 51 22
Test: Encrypt*/

		// those are the standard test vectors
	static unsigned char SEED_KEYs[3][17] =
	{
			"\x00\x00\x00\x00\x0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
			"\x00\x01\x02\x03\x04\x5\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f",
			"\x47\x06\x48\x08\x51\xe6\x1b\xe8\x5d\x74\xbf\xb3\xfd\x95\x61\x85"
	};

	static unsigned char SEED_PTs[3][17] =
	{
			"\x00\x01\x02\x03\x04\x5\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f",
			"\x00\x00\x00\x00\x0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
			"\x83\xa2\xf8\xa2\x88\x64\x1f\xb9\xa4\xe9\xa5\xcc\x2f\x13\x1c\x7d"
	};

	static unsigned char SEED_CTs[3][17] =
	{
			"\x5e\xba\xc6\xe0\x05\x4e\x16\x68\x19\xaf\xf1\xcc\x6d\x34\x6c\xdb",
			"\xc1\x1f\x22\xf2\x01\x40\x50\x50\x84\x48\x35\x97\xe4\x37\x0f\x43",
			"\xee\x54\xd1\x3e\xbc\xae\x70\x6d\x22\x6b\xc3\x14\x2c\xd4\x0d\x4a"
	};


		// returns true for pass
	bool test_SEED(void) {
			bool result=true; // false=fail, true=pass
			int i;
			DWORD dwRoundKey[32];
			unsigned char data[16];


			for( i = 0; i < 3; i++ )
			{
				memcpy( data, SEED_PTs[i], 16 );
				
				SeedRoundKey(dwRoundKey,SEED_KEYs[i]);																					// Only 128 bit keys are supported
				SeedEncrypt(data, dwRoundKey);
				
				if (memcmp( data, SEED_CTs[i], 16 ) )
				{
					result=0;
					return(result);
				}
			}
			
			for( i = 0; i < 3; i++ )
			{
				memcpy( data, SEED_CTs[i], 16 );
				
				SeedRoundKey(dwRoundKey,SEED_KEYs[i]);																				// Only 128 bit keys are supported
				SeedDecrypt(data, dwRoundKey);
				
				if (memcmp( data, SEED_PTs[i], 16 ) )
				{
					result=false;
					return(result);
				}
			}
			
			return(result);
	}



/*******************************************************************************
* Created 29, June, 1999
* Modified 4, June, 2008
* FILE:         SEED_KISA.c
*
* DESCRIPTION: Core routines for the enhanced SEED
* 
*******************************************************************************/

//#include <stdint.h>
//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>
//#ifdef __unix__
//#include <endian.h>
//#endif

/******************************* Type Definitions *****************************/
/*
// If your machine uses different data types, you must change them to proper ones.
#if 0
#ifndef TYPE_DEFINITION
    #define TYPE_DEFINITION
    #if defined(__alpha)
        typedef unsigned int        DWORD;		// unsigned 4-byte data type
        typedef unsigned short      WORD;		// unsigned 2-byte data type
    #else
        typedef unsigned long int   DWORD;		// unsigned 4-byte data type
        typedef unsigned short int  WORD;		// unsigned 2-byte data type
    #endif
    typedef unsigned char           BYTE;		// unsigned 1-byte data type
#endif
#else
#define BYTE word8
#define WORD word16
#define DWORD word32
#endif
*/

/***************************** Endianness Define ******************************/
/*
#if __BYTE_ORDER == __BIG_ENDIAN
#define SEED_BIG_ENDIAN
#elif __BYTE_ORDER == __LITTLE_ENDIAN
#define SEED_LITTLE_ENDIAN
#else
#error "Endian not defined"
#endif
*/	
	// VC++
#if BYTE_ORDER == LITTLE_ENDIAN
#define SEED_LITTLE_ENDIAN
#else
#if  BYTE_ORDER == BIG_ENDIAN
#define SEED_BIG_ENDIAN
#endif
#endif

// If endianness is not defined correctly, you must modify here.
// SEED uses the Little endian as a defalut order

#if !defined(SEED_BIG_ENDIAN) && !defined(SEED_LITTLE_ENDIAN)
#if __alpha__	||	__alpha	||	__i386__	||	i386	||	_M_I86	||	_M_IX86	||	\
	__OS2__		||	sun386	||	__TURBOC__	||	vax		||	vms		||	VMS		||	__VMS 
#define SEED_LITTLE_ENDIAN
#else
#define SEED_BIG_ENDIAN
#endif
#endif


/**************************** Constant Definitions ****************************/

#define NoRounds         16						// the number of rounds
#define NoRoundKeys      (NoRounds*2)			// the number of round-keys
#define SeedBlockSize    16    					// block length in bytes
#define SeedBlockLen     128   					// block length in bits


/******************************** Common Macros *******************************/

// macroses for left or right rotations
#if defined(_MSC_VER)
    #define ROTL(x, n)     (_lrotl((x), (n)))		// left rotation
    #define ROTR(x, n)     (_lrotr((x), (n)))		// right rotation
#else
    #define ROTL(x, n)     (((x) << (n)) | ((x) >> (32-(n))))		// left rotation
    #define ROTR(x, n)     (((x) >> (n)) | ((x) << (32-(n))))		// right rotation
#endif

// macroses for converting endianess
#define EndianChange(dwS)                       \
    ( (ROTL((dwS),  8) & (DWORD)0x00ff00ff) |   \
      (ROTL((dwS), 24) & (DWORD)0xff00ff00) )


static DWORD SS0[256] = {
0x2989a1a8, 0x05858184, 0x16c6d2d4, 0x13c3d3d0, 0x14445054, 0x1d0d111c, 0x2c8ca0ac, 0x25052124,
0x1d4d515c, 0x03434340, 0x18081018, 0x1e0e121c, 0x11415150, 0x3cccf0fc, 0x0acac2c8, 0x23436360,
0x28082028, 0x04444044, 0x20002020, 0x1d8d919c, 0x20c0e0e0, 0x22c2e2e0, 0x08c8c0c8, 0x17071314,
0x2585a1a4, 0x0f8f838c, 0x03030300, 0x3b4b7378, 0x3b8bb3b8, 0x13031310, 0x12c2d2d0, 0x2ecee2ec,
0x30407070, 0x0c8c808c, 0x3f0f333c, 0x2888a0a8, 0x32023230, 0x1dcdd1dc, 0x36c6f2f4, 0x34447074,
0x2ccce0ec, 0x15859194, 0x0b0b0308, 0x17475354, 0x1c4c505c, 0x1b4b5358, 0x3d8db1bc, 0x01010100,
0x24042024, 0x1c0c101c, 0x33437370, 0x18889098, 0x10001010, 0x0cccc0cc, 0x32c2f2f0, 0x19c9d1d8,
0x2c0c202c, 0x27c7e3e4, 0x32427270, 0x03838380, 0x1b8b9398, 0x11c1d1d0, 0x06868284, 0x09c9c1c8,
0x20406060, 0x10405050, 0x2383a3a0, 0x2bcbe3e8, 0x0d0d010c, 0x3686b2b4, 0x1e8e929c, 0x0f4f434c,
0x3787b3b4, 0x1a4a5258, 0x06c6c2c4, 0x38487078, 0x2686a2a4, 0x12021210, 0x2f8fa3ac, 0x15c5d1d4,
0x21416160, 0x03c3c3c0, 0x3484b0b4, 0x01414140, 0x12425250, 0x3d4d717c, 0x0d8d818c, 0x08080008,
0x1f0f131c, 0x19899198, 0x00000000, 0x19091118, 0x04040004, 0x13435350, 0x37c7f3f4, 0x21c1e1e0,
0x3dcdf1fc, 0x36467274, 0x2f0f232c, 0x27072324, 0x3080b0b0, 0x0b8b8388, 0x0e0e020c, 0x2b8ba3a8,
0x2282a2a0, 0x2e4e626c, 0x13839390, 0x0d4d414c, 0x29496168, 0x3c4c707c, 0x09090108, 0x0a0a0208,
0x3f8fb3bc, 0x2fcfe3ec, 0x33c3f3f0, 0x05c5c1c4, 0x07878384, 0x14041014, 0x3ecef2fc, 0x24446064,
0x1eced2dc, 0x2e0e222c, 0x0b4b4348, 0x1a0a1218, 0x06060204, 0x21012120, 0x2b4b6368, 0x26466264,
0x02020200, 0x35c5f1f4, 0x12829290, 0x0a8a8288, 0x0c0c000c, 0x3383b3b0, 0x3e4e727c, 0x10c0d0d0,
0x3a4a7278, 0x07474344, 0x16869294, 0x25c5e1e4, 0x26062224, 0x00808080, 0x2d8da1ac, 0x1fcfd3dc,
0x2181a1a0, 0x30003030, 0x37073334, 0x2e8ea2ac, 0x36063234, 0x15051114, 0x22022220, 0x38083038,
0x34c4f0f4, 0x2787a3a4, 0x05454144, 0x0c4c404c, 0x01818180, 0x29c9e1e8, 0x04848084, 0x17879394,
0x35053134, 0x0bcbc3c8, 0x0ecec2cc, 0x3c0c303c, 0x31417170, 0x11011110, 0x07c7c3c4, 0x09898188,
0x35457174, 0x3bcbf3f8, 0x1acad2d8, 0x38c8f0f8, 0x14849094, 0x19495158, 0x02828280, 0x04c4c0c4,
0x3fcff3fc, 0x09494148, 0x39093138, 0x27476364, 0x00c0c0c0, 0x0fcfc3cc, 0x17c7d3d4, 0x3888b0b8,
0x0f0f030c, 0x0e8e828c, 0x02424240, 0x23032320, 0x11819190, 0x2c4c606c, 0x1bcbd3d8, 0x2484a0a4,
0x34043034, 0x31c1f1f0, 0x08484048, 0x02c2c2c0, 0x2f4f636c, 0x3d0d313c, 0x2d0d212c, 0x00404040,
0x3e8eb2bc, 0x3e0e323c, 0x3c8cb0bc, 0x01c1c1c0, 0x2a8aa2a8, 0x3a8ab2b8, 0x0e4e424c, 0x15455154,
0x3b0b3338, 0x1cccd0dc, 0x28486068, 0x3f4f737c, 0x1c8c909c, 0x18c8d0d8, 0x0a4a4248, 0x16465254,
0x37477374, 0x2080a0a0, 0x2dcde1ec, 0x06464244, 0x3585b1b4, 0x2b0b2328, 0x25456164, 0x3acaf2f8,
0x23c3e3e0, 0x3989b1b8, 0x3181b1b0, 0x1f8f939c, 0x1e4e525c, 0x39c9f1f8, 0x26c6e2e4, 0x3282b2b0,
0x31013130, 0x2acae2e8, 0x2d4d616c, 0x1f4f535c, 0x24c4e0e4, 0x30c0f0f0, 0x0dcdc1cc, 0x08888088,
0x16061214, 0x3a0a3238, 0x18485058, 0x14c4d0d4, 0x22426260, 0x29092128, 0x07070304, 0x33033330,
0x28c8e0e8, 0x1b0b1318, 0x05050104, 0x39497178, 0x10809090, 0x2a4a6268, 0x2a0a2228, 0x1a8a9298
};

static DWORD SS1[256] = {
0x38380830, 0xe828c8e0, 0x2c2d0d21, 0xa42686a2, 0xcc0fcfc3, 0xdc1eced2, 0xb03383b3, 0xb83888b0,
0xac2f8fa3, 0x60204060, 0x54154551, 0xc407c7c3, 0x44044440, 0x6c2f4f63, 0x682b4b63, 0x581b4b53,
0xc003c3c3, 0x60224262, 0x30330333, 0xb43585b1, 0x28290921, 0xa02080a0, 0xe022c2e2, 0xa42787a3,
0xd013c3d3, 0x90118191, 0x10110111, 0x04060602, 0x1c1c0c10, 0xbc3c8cb0, 0x34360632, 0x480b4b43,
0xec2fcfe3, 0x88088880, 0x6c2c4c60, 0xa82888a0, 0x14170713, 0xc404c4c0, 0x14160612, 0xf434c4f0,
0xc002c2c2, 0x44054541, 0xe021c1e1, 0xd416c6d2, 0x3c3f0f33, 0x3c3d0d31, 0x8c0e8e82, 0x98188890,
0x28280820, 0x4c0e4e42, 0xf436c6f2, 0x3c3e0e32, 0xa42585a1, 0xf839c9f1, 0x0c0d0d01, 0xdc1fcfd3,
0xd818c8d0, 0x282b0b23, 0x64264662, 0x783a4a72, 0x24270723, 0x2c2f0f23, 0xf031c1f1, 0x70324272,
0x40024242, 0xd414c4d0, 0x40014141, 0xc000c0c0, 0x70334373, 0x64274763, 0xac2c8ca0, 0x880b8b83,
0xf437c7f3, 0xac2d8da1, 0x80008080, 0x1c1f0f13, 0xc80acac2, 0x2c2c0c20, 0xa82a8aa2, 0x34340430,
0xd012c2d2, 0x080b0b03, 0xec2ecee2, 0xe829c9e1, 0x5c1d4d51, 0x94148490, 0x18180810, 0xf838c8f0,
0x54174753, 0xac2e8ea2, 0x08080800, 0xc405c5c1, 0x10130313, 0xcc0dcdc1, 0x84068682, 0xb83989b1,
0xfc3fcff3, 0x7c3d4d71, 0xc001c1c1, 0x30310131, 0xf435c5f1, 0x880a8a82, 0x682a4a62, 0xb03181b1,
0xd011c1d1, 0x20200020, 0xd417c7d3, 0x00020202, 0x20220222, 0x04040400, 0x68284860, 0x70314171,
0x04070703, 0xd81bcbd3, 0x9c1d8d91, 0x98198991, 0x60214161, 0xbc3e8eb2, 0xe426c6e2, 0x58194951,
0xdc1dcdd1, 0x50114151, 0x90108090, 0xdc1cccd0, 0x981a8a92, 0xa02383a3, 0xa82b8ba3, 0xd010c0d0,
0x80018181, 0x0c0f0f03, 0x44074743, 0x181a0a12, 0xe023c3e3, 0xec2ccce0, 0x8c0d8d81, 0xbc3f8fb3,
0x94168692, 0x783b4b73, 0x5c1c4c50, 0xa02282a2, 0xa02181a1, 0x60234363, 0x20230323, 0x4c0d4d41,
0xc808c8c0, 0x9c1e8e92, 0x9c1c8c90, 0x383a0a32, 0x0c0c0c00, 0x2c2e0e22, 0xb83a8ab2, 0x6c2e4e62,
0x9c1f8f93, 0x581a4a52, 0xf032c2f2, 0x90128292, 0xf033c3f3, 0x48094941, 0x78384870, 0xcc0cccc0,
0x14150511, 0xf83bcbf3, 0x70304070, 0x74354571, 0x7c3f4f73, 0x34350531, 0x10100010, 0x00030303,
0x64244460, 0x6c2d4d61, 0xc406c6c2, 0x74344470, 0xd415c5d1, 0xb43484b0, 0xe82acae2, 0x08090901,
0x74364672, 0x18190911, 0xfc3ecef2, 0x40004040, 0x10120212, 0xe020c0e0, 0xbc3d8db1, 0x04050501,
0xf83acaf2, 0x00010101, 0xf030c0f0, 0x282a0a22, 0x5c1e4e52, 0xa82989a1, 0x54164652, 0x40034343,
0x84058581, 0x14140410, 0x88098981, 0x981b8b93, 0xb03080b0, 0xe425c5e1, 0x48084840, 0x78394971,
0x94178793, 0xfc3cccf0, 0x1c1e0e12, 0x80028282, 0x20210121, 0x8c0c8c80, 0x181b0b13, 0x5c1f4f53,
0x74374773, 0x54144450, 0xb03282b2, 0x1c1d0d11, 0x24250521, 0x4c0f4f43, 0x00000000, 0x44064642,
0xec2dcde1, 0x58184850, 0x50124252, 0xe82bcbe3, 0x7c3e4e72, 0xd81acad2, 0xc809c9c1, 0xfc3dcdf1,
0x30300030, 0x94158591, 0x64254561, 0x3c3c0c30, 0xb43686b2, 0xe424c4e0, 0xb83b8bb3, 0x7c3c4c70,
0x0c0e0e02, 0x50104050, 0x38390931, 0x24260622, 0x30320232, 0x84048480, 0x68294961, 0x90138393,
0x34370733, 0xe427c7e3, 0x24240420, 0xa42484a0, 0xc80bcbc3, 0x50134353, 0x080a0a02, 0x84078783,
0xd819c9d1, 0x4c0c4c40, 0x80038383, 0x8c0f8f83, 0xcc0ecec2, 0x383b0b33, 0x480a4a42, 0xb43787b3
};

static DWORD SS2[256] = {
0xa1a82989, 0x81840585, 0xd2d416c6, 0xd3d013c3, 0x50541444, 0x111c1d0d, 0xa0ac2c8c, 0x21242505,
0x515c1d4d, 0x43400343, 0x10181808, 0x121c1e0e, 0x51501141, 0xf0fc3ccc, 0xc2c80aca, 0x63602343,
0x20282808, 0x40440444, 0x20202000, 0x919c1d8d, 0xe0e020c0, 0xe2e022c2, 0xc0c808c8, 0x13141707,
0xa1a42585, 0x838c0f8f, 0x03000303, 0x73783b4b, 0xb3b83b8b, 0x13101303, 0xd2d012c2, 0xe2ec2ece,
0x70703040, 0x808c0c8c, 0x333c3f0f, 0xa0a82888, 0x32303202, 0xd1dc1dcd, 0xf2f436c6, 0x70743444,
0xe0ec2ccc, 0x91941585, 0x03080b0b, 0x53541747, 0x505c1c4c, 0x53581b4b, 0xb1bc3d8d, 0x01000101,
0x20242404, 0x101c1c0c, 0x73703343, 0x90981888, 0x10101000, 0xc0cc0ccc, 0xf2f032c2, 0xd1d819c9,
0x202c2c0c, 0xe3e427c7, 0x72703242, 0x83800383, 0x93981b8b, 0xd1d011c1, 0x82840686, 0xc1c809c9,
0x60602040, 0x50501040, 0xa3a02383, 0xe3e82bcb, 0x010c0d0d, 0xb2b43686, 0x929c1e8e, 0x434c0f4f,
0xb3b43787, 0x52581a4a, 0xc2c406c6, 0x70783848, 0xa2a42686, 0x12101202, 0xa3ac2f8f, 0xd1d415c5,
0x61602141, 0xc3c003c3, 0xb0b43484, 0x41400141, 0x52501242, 0x717c3d4d, 0x818c0d8d, 0x00080808,
0x131c1f0f, 0x91981989, 0x00000000, 0x11181909, 0x00040404, 0x53501343, 0xf3f437c7, 0xe1e021c1,
0xf1fc3dcd, 0x72743646, 0x232c2f0f, 0x23242707, 0xb0b03080, 0x83880b8b, 0x020c0e0e, 0xa3a82b8b,
0xa2a02282, 0x626c2e4e, 0x93901383, 0x414c0d4d, 0x61682949, 0x707c3c4c, 0x01080909, 0x02080a0a,
0xb3bc3f8f, 0xe3ec2fcf, 0xf3f033c3, 0xc1c405c5, 0x83840787, 0x10141404, 0xf2fc3ece, 0x60642444,
0xd2dc1ece, 0x222c2e0e, 0x43480b4b, 0x12181a0a, 0x02040606, 0x21202101, 0x63682b4b, 0x62642646,
0x02000202, 0xf1f435c5, 0x92901282, 0x82880a8a, 0x000c0c0c, 0xb3b03383, 0x727c3e4e, 0xd0d010c0,
0x72783a4a, 0x43440747, 0x92941686, 0xe1e425c5, 0x22242606, 0x80800080, 0xa1ac2d8d, 0xd3dc1fcf,
0xa1a02181, 0x30303000, 0x33343707, 0xa2ac2e8e, 0x32343606, 0x11141505, 0x22202202, 0x30383808,
0xf0f434c4, 0xa3a42787, 0x41440545, 0x404c0c4c, 0x81800181, 0xe1e829c9, 0x80840484, 0x93941787,
0x31343505, 0xc3c80bcb, 0xc2cc0ece, 0x303c3c0c, 0x71703141, 0x11101101, 0xc3c407c7, 0x81880989,
0x71743545, 0xf3f83bcb, 0xd2d81aca, 0xf0f838c8, 0x90941484, 0x51581949, 0x82800282, 0xc0c404c4,
0xf3fc3fcf, 0x41480949, 0x31383909, 0x63642747, 0xc0c000c0, 0xc3cc0fcf, 0xd3d417c7, 0xb0b83888,
0x030c0f0f, 0x828c0e8e, 0x42400242, 0x23202303, 0x91901181, 0x606c2c4c, 0xd3d81bcb, 0xa0a42484,
0x30343404, 0xf1f031c1, 0x40480848, 0xc2c002c2, 0x636c2f4f, 0x313c3d0d, 0x212c2d0d, 0x40400040,
0xb2bc3e8e, 0x323c3e0e, 0xb0bc3c8c, 0xc1c001c1, 0xa2a82a8a, 0xb2b83a8a, 0x424c0e4e, 0x51541545,
0x33383b0b, 0xd0dc1ccc, 0x60682848, 0x737c3f4f, 0x909c1c8c, 0xd0d818c8, 0x42480a4a, 0x52541646,
0x73743747, 0xa0a02080, 0xe1ec2dcd, 0x42440646, 0xb1b43585, 0x23282b0b, 0x61642545, 0xf2f83aca,
0xe3e023c3, 0xb1b83989, 0xb1b03181, 0x939c1f8f, 0x525c1e4e, 0xf1f839c9, 0xe2e426c6, 0xb2b03282,
0x31303101, 0xe2e82aca, 0x616c2d4d, 0x535c1f4f, 0xe0e424c4, 0xf0f030c0, 0xc1cc0dcd, 0x80880888,
0x12141606, 0x32383a0a, 0x50581848, 0xd0d414c4, 0x62602242, 0x21282909, 0x03040707, 0x33303303,
0xe0e828c8, 0x13181b0b, 0x01040505, 0x71783949, 0x90901080, 0x62682a4a, 0x22282a0a, 0x92981a8a
};

static DWORD SS3[256] = {
0x08303838, 0xc8e0e828, 0x0d212c2d, 0x86a2a426, 0xcfc3cc0f, 0xced2dc1e, 0x83b3b033, 0x88b0b838,
0x8fa3ac2f, 0x40606020, 0x45515415, 0xc7c3c407, 0x44404404, 0x4f636c2f, 0x4b63682b, 0x4b53581b,
0xc3c3c003, 0x42626022, 0x03333033, 0x85b1b435, 0x09212829, 0x80a0a020, 0xc2e2e022, 0x87a3a427,
0xc3d3d013, 0x81919011, 0x01111011, 0x06020406, 0x0c101c1c, 0x8cb0bc3c, 0x06323436, 0x4b43480b,
0xcfe3ec2f, 0x88808808, 0x4c606c2c, 0x88a0a828, 0x07131417, 0xc4c0c404, 0x06121416, 0xc4f0f434,
0xc2c2c002, 0x45414405, 0xc1e1e021, 0xc6d2d416, 0x0f333c3f, 0x0d313c3d, 0x8e828c0e, 0x88909818,
0x08202828, 0x4e424c0e, 0xc6f2f436, 0x0e323c3e, 0x85a1a425, 0xc9f1f839, 0x0d010c0d, 0xcfd3dc1f,
0xc8d0d818, 0x0b23282b, 0x46626426, 0x4a72783a, 0x07232427, 0x0f232c2f, 0xc1f1f031, 0x42727032,
0x42424002, 0xc4d0d414, 0x41414001, 0xc0c0c000, 0x43737033, 0x47636427, 0x8ca0ac2c, 0x8b83880b,
0xc7f3f437, 0x8da1ac2d, 0x80808000, 0x0f131c1f, 0xcac2c80a, 0x0c202c2c, 0x8aa2a82a, 0x04303434,
0xc2d2d012, 0x0b03080b, 0xcee2ec2e, 0xc9e1e829, 0x4d515c1d, 0x84909414, 0x08101818, 0xc8f0f838,
0x47535417, 0x8ea2ac2e, 0x08000808, 0xc5c1c405, 0x03131013, 0xcdc1cc0d, 0x86828406, 0x89b1b839,
0xcff3fc3f, 0x4d717c3d, 0xc1c1c001, 0x01313031, 0xc5f1f435, 0x8a82880a, 0x4a62682a, 0x81b1b031,
0xc1d1d011, 0x00202020, 0xc7d3d417, 0x02020002, 0x02222022, 0x04000404, 0x48606828, 0x41717031,
0x07030407, 0xcbd3d81b, 0x8d919c1d, 0x89919819, 0x41616021, 0x8eb2bc3e, 0xc6e2e426, 0x49515819,
0xcdd1dc1d, 0x41515011, 0x80909010, 0xccd0dc1c, 0x8a92981a, 0x83a3a023, 0x8ba3a82b, 0xc0d0d010,
0x81818001, 0x0f030c0f, 0x47434407, 0x0a12181a, 0xc3e3e023, 0xcce0ec2c, 0x8d818c0d, 0x8fb3bc3f,
0x86929416, 0x4b73783b, 0x4c505c1c, 0x82a2a022, 0x81a1a021, 0x43636023, 0x03232023, 0x4d414c0d,
0xc8c0c808, 0x8e929c1e, 0x8c909c1c, 0x0a32383a, 0x0c000c0c, 0x0e222c2e, 0x8ab2b83a, 0x4e626c2e,
0x8f939c1f, 0x4a52581a, 0xc2f2f032, 0x82929012, 0xc3f3f033, 0x49414809, 0x48707838, 0xccc0cc0c,
0x05111415, 0xcbf3f83b, 0x40707030, 0x45717435, 0x4f737c3f, 0x05313435, 0x00101010, 0x03030003,
0x44606424, 0x4d616c2d, 0xc6c2c406, 0x44707434, 0xc5d1d415, 0x84b0b434, 0xcae2e82a, 0x09010809,
0x46727436, 0x09111819, 0xcef2fc3e, 0x40404000, 0x02121012, 0xc0e0e020, 0x8db1bc3d, 0x05010405,
0xcaf2f83a, 0x01010001, 0xc0f0f030, 0x0a22282a, 0x4e525c1e, 0x89a1a829, 0x46525416, 0x43434003,
0x85818405, 0x04101414, 0x89818809, 0x8b93981b, 0x80b0b030, 0xc5e1e425, 0x48404808, 0x49717839,
0x87939417, 0xccf0fc3c, 0x0e121c1e, 0x82828002, 0x01212021, 0x8c808c0c, 0x0b13181b, 0x4f535c1f,
0x47737437, 0x44505414, 0x82b2b032, 0x0d111c1d, 0x05212425, 0x4f434c0f, 0x00000000, 0x46424406,
0xcde1ec2d, 0x48505818, 0x42525012, 0xcbe3e82b, 0x4e727c3e, 0xcad2d81a, 0xc9c1c809, 0xcdf1fc3d,
0x00303030, 0x85919415, 0x45616425, 0x0c303c3c, 0x86b2b436, 0xc4e0e424, 0x8bb3b83b, 0x4c707c3c,
0x0e020c0e, 0x40505010, 0x09313839, 0x06222426, 0x02323032, 0x84808404, 0x49616829, 0x83939013,
0x07333437, 0xc7e3e427, 0x04202424, 0x84a0a424, 0xcbc3c80b, 0x43535013, 0x0a02080a, 0x87838407,
0xc9d1d819, 0x4c404c0c, 0x83838003, 0x8f838c0f, 0xcec2cc0e, 0x0b33383b, 0x4a42480a, 0x87b3b437
};




/******************** Macros for Encryption and Decryption ********************/

#define GetB0(A)  ( (BYTE)((A)    ) )
#define GetB1(A)  ( (BYTE)((A)>> 8) )
#define GetB2(A)  ( (BYTE)((A)>>16) )
#define GetB3(A)  ( (BYTE)((A)>>24) )

// Round function F and adding output of F to L.
// L0, L1 : left input values at each round
// R0, R1 : right input values at each round
// K : round keys at each round
#define SeedRound(L0, L1, R0, R1, K) {             \
    T0 = R0 ^ (K)[0];                              \
    T1 = R1 ^ (K)[1];                              \
    T1 ^= T0;                                      \
    T1 = SS0[GetB0(T1)] ^ SS1[GetB1(T1)] ^         \
         SS2[GetB2(T1)] ^ SS3[GetB3(T1)];          \
    T0 += T1;                                      \
    T0 = SS0[GetB0(T0)] ^ SS1[GetB1(T0)] ^         \
         SS2[GetB2(T0)] ^ SS3[GetB3(T0)];          \
    T1 += T0;                                      \
    T1 = SS0[GetB0(T1)] ^ SS1[GetB1(T1)] ^         \
         SS2[GetB2(T1)] ^ SS3[GetB3(T1)];          \
    T0 += T1;                                      \
    L0 ^= T0; L1 ^= T1;                            \
}

/********************************* Encryption *********************************/

void SeedEncrypt (
		BYTE *pbData, 				// [in,out]	data to be encrypted
		DWORD *pdwRoundKey)			// [in]			round keys for encryption
{
	DWORD L0, L1, R0, R1;		// Iuput/output values at each rounds
	DWORD T0, T1;				// Temporary variables for round function F
	DWORD *K = pdwRoundKey;		// Pointer of round keys

// Set up input values for first round
    L0 = ((DWORD *)pbData)[0];
    L1 = ((DWORD *)pbData)[1];
	R0 = ((DWORD *)pbData)[2];
    R1 = ((DWORD *)pbData)[3];
    
// Reorder for big endian 
// Because SEED use little endian order in default
#ifndef SEED_BIG_ENDIAN
    L0 = EndianChange(L0);
    L1 = EndianChange(L1);
    R0 = EndianChange(R0);
    R1 = EndianChange(R1);
#endif

    SeedRound(L0, L1, R0, R1, K   ); 	// Round 1
    SeedRound(R0, R1, L0, L1, K+ 2); 	// Round 2
    SeedRound(L0, L1, R0, R1, K+ 4); 	// Round 3
    SeedRound(R0, R1, L0, L1, K+ 6); 	// Round 4
    SeedRound(L0, L1, R0, R1, K+ 8); 	// Round 5
    SeedRound(R0, R1, L0, L1, K+10); 	// Round 6
    SeedRound(L0, L1, R0, R1, K+12); 	// Round 7
    SeedRound(R0, R1, L0, L1, K+14); 	// Round 8
    SeedRound(L0, L1, R0, R1, K+16); 	// Round 9
    SeedRound(R0, R1, L0, L1, K+18); 	// Round 10
    SeedRound(L0, L1, R0, R1, K+20); 	// Round 11
    SeedRound(R0, R1, L0, L1, K+22); 	// Round 12
    SeedRound(L0, L1, R0, R1, K+24); 	// Round 13
    SeedRound(R0, R1, L0, L1, K+26); 	// Round 14
    SeedRound(L0, L1, R0, R1, K+28); 	// Round 15
    SeedRound(R0, R1, L0, L1, K+30); 	// Round 16

#ifndef SEED_BIG_ENDIAN
    L0 = EndianChange(L0);
    L1 = EndianChange(L1);
    R0 = EndianChange(R0);
    R1 = EndianChange(R1);
#endif

// Copying output values from last round to pbData
    ((DWORD *)pbData)[0] = R0;
    ((DWORD *)pbData)[1] = R1;
    ((DWORD *)pbData)[2] = L0;
    ((DWORD *)pbData)[3] = L1;
}


/********************************* Decryption *********************************/

// Same as encrypt, except that round keys are applied in reverse order
void SeedDecrypt (
		BYTE *pbData, 				// [in,out]	data to be decrypted
		DWORD *pdwRoundKey)			// [in]			round keys for decryption
{
	DWORD L0, L1, R0, R1;		// Iuput/output values at each rounds
	DWORD T0, T1;				// Temporary variables for round function F
	DWORD *K = pdwRoundKey;		// Pointer of round keys

// Set up input values for first round
    L0 = ((DWORD *)pbData)[0];
    L1 = ((DWORD *)pbData)[1];
    R0 = ((DWORD *)pbData)[2];
    R1 = ((DWORD *)pbData)[3];

// Reorder for big endian 
#ifndef SEED_BIG_ENDIAN
    L0 = EndianChange(L0);
    L1 = EndianChange(L1);
    R0 = EndianChange(R0);
    R1 = EndianChange(R1);
#endif

    SeedRound(L0, L1, R0, R1, K+30); 	// Round 1
    SeedRound(R0, R1, L0, L1, K+28); 	// Round 2 
    SeedRound(L0, L1, R0, R1, K+26); 	// Round 3 
    SeedRound(R0, R1, L0, L1, K+24); 	// Round 4 
    SeedRound(L0, L1, R0, R1, K+22); 	// Round 5 
    SeedRound(R0, R1, L0, L1, K+20); 	// Round 6 
    SeedRound(L0, L1, R0, R1, K+18); 	// Round 7 
    SeedRound(R0, R1, L0, L1, K+16); 	// Round 8 
    SeedRound(L0, L1, R0, R1, K+14); 	// Round 9 
    SeedRound(R0, R1, L0, L1, K+12); 	// Round 10
    SeedRound(L0, L1, R0, R1, K+10); 	// Round 11
    SeedRound(R0, R1, L0, L1, K+ 8); 	// Round 12
    SeedRound(L0, L1, R0, R1, K+ 6); 	// Round 13
    SeedRound(R0, R1, L0, L1, K+ 4); 	// Round 14
    SeedRound(L0, L1, R0, R1, K+ 2); 	// Round 15
    SeedRound(R0, R1, L0, L1, K+ 0); 	// Round 16

#ifndef SEED_BIG_ENDIAN
    L0 = EndianChange(L0);
    L1 = EndianChange(L1);
    R0 = EndianChange(R0);
    R1 = EndianChange(R1);
#endif

// Copy output values from last round to pbData
    ((DWORD *)pbData)[0] = R0;
    ((DWORD *)pbData)[1] = R1;
    ((DWORD *)pbData)[2] = L0;
    ((DWORD *)pbData)[3] = L1;
}


/************************ Constants for Key schedule **************************/

//		KC0 = golden ratio; KCi = ROTL(KCi-1, 1) 
#define KC0     0x9e3779b9UL
#define KC1     0x3c6ef373UL
#define KC2     0x78dde6e6UL
#define KC3     0xf1bbcdccUL
#define KC4     0xe3779b99UL
#define KC5     0xc6ef3733UL
#define KC6     0x8dde6e67UL
#define KC7     0x1bbcdccfUL
#define KC8     0x3779b99eUL
#define KC9     0x6ef3733cUL
#define KC10    0xdde6e678UL
#define KC11    0xbbcdccf1UL
#define KC12    0x779b99e3UL
#define KC13    0xef3733c6UL
#define KC14    0xde6e678dUL
#define KC15    0xbcdccf1bUL


/************************** Macros for Key schedule ***************************/

#define RoundKeyUpdate0(K, A, B, C, D, KC) {	\
    T0 = A + C - KC;                            \
    T1 = B + KC - D;                            \
    (K)[0] = SS0[GetB0(T0)] ^ SS1[GetB1(T0)] ^  \
             SS2[GetB2(T0)] ^ SS3[GetB3(T0)];   \
    (K)[1] = SS0[GetB0(T1)] ^ SS1[GetB1(T1)] ^  \
             SS2[GetB2(T1)] ^ SS3[GetB3(T1)];   \
    T0 = A;                                     \
    A = (A>>8) ^ (B<<24);                       \
    B = (B>>8) ^ (T0<<24);                      \
}

#define RoundKeyUpdate1(K, A, B, C, D, KC) {	\
    T0 = A + C - KC;                            \
    T1 = B + KC - D;                            \
    (K)[0] = SS0[GetB0(T0)] ^ SS1[GetB1(T0)] ^  \
             SS2[GetB2(T0)] ^ SS3[GetB3(T0)];   \
    (K)[1] = SS0[GetB0(T1)] ^ SS1[GetB1(T1)] ^  \
             SS2[GetB2(T1)] ^ SS3[GetB3(T1)];   \
    T0 = C;                                     \
    C = (C<<8) ^ (D>>24);                       \
    D = (D<<8) ^ (T0>>24);                      \
}


/******************************** Key Schedule ********************************/

void SeedRoundKey(
		DWORD *pdwRoundKey,			// [out]	round keys for encryption or decryption
		BYTE *pbUserKey)			// [in]		secret user key
{
	DWORD A, B, C, D;				// Iuput/output values at each rounds
	DWORD T0, T1;					// Temporary variable
	DWORD *K = pdwRoundKey;			// Pointer of round keys

// Set up input values for Key Schedule	
	A = ((DWORD *)pbUserKey)[0];
	B = ((DWORD *)pbUserKey)[1];
	C = ((DWORD *)pbUserKey)[2];
	D = ((DWORD *)pbUserKey)[3];
	 
// Reorder for big endian 
#ifndef SEED_BIG_ENDIAN
	A = EndianChange(A);
	B = EndianChange(B);
	C = EndianChange(C);
	D = EndianChange(D);
#endif

// i-th round keys( K_i,0 and K_i,1 ) are denoted as K[2*(i-1)] and K[2*i-1], respectively
    RoundKeyUpdate0(K   , A, B, C, D, KC0 );	// K_1,0 and K_1,1
    RoundKeyUpdate1(K+ 2, A, B, C, D, KC1 );	// K_2,0 and K_2,1
    RoundKeyUpdate0(K+ 4, A, B, C, D, KC2 );	// K_3,0 and K_3,1
    RoundKeyUpdate1(K+ 6, A, B, C, D, KC3 );	// K_4,0 and K_4,1
    RoundKeyUpdate0(K+ 8, A, B, C, D, KC4 );	// K_5,0 and K_5,1
    RoundKeyUpdate1(K+10, A, B, C, D, KC5 );	// K_6,0 and K_6,1
    RoundKeyUpdate0(K+12, A, B, C, D, KC6 );	// K_7,0 and K_7,1
    RoundKeyUpdate1(K+14, A, B, C, D, KC7 );	// K_8,0 and K_8,1
    RoundKeyUpdate0(K+16, A, B, C, D, KC8 );	// K_9,0 and K_9,1
    RoundKeyUpdate1(K+18, A, B, C, D, KC9 );	// K_10,0 and K_10,1
    RoundKeyUpdate0(K+20, A, B, C, D, KC10);	// K_11,0 and K_11,1
    RoundKeyUpdate1(K+22, A, B, C, D, KC11);	// K_12,0 and K_12,1
    RoundKeyUpdate0(K+24, A, B, C, D, KC12);	// K_13,0 and K_13,1
    RoundKeyUpdate1(K+26, A, B, C, D, KC13);	// K_14,0 and K_14,1
    RoundKeyUpdate0(K+28, A, B, C, D, KC14);	// K_15,0 and K_15,1

    T0 = A + C - KC15;
    T1 = B - D + KC15;
    K[30] = SS0[GetB0(T0)] ^ SS1[GetB1(T0)] ^	// K_16,0
           SS2[GetB2(T0)] ^ SS3[GetB3(T0)];
    K[31] = SS0[GetB0(T1)] ^ SS1[GetB1(T1)] ^	// K_16,1
           SS2[GetB2(T1)] ^ SS3[GetB3(T1)];

}

#undef ROTR
#undef ROTL

/*********************************** END **************************************/

























		// RC6 test routine:

		// those are the standard test vectors
	static unsigned char RC6_KEYs[3][33] =
	{
			"\x01\x23\x45\x67\x89\xab\xcd\xef\x01\x12\x23\x34\x45\x56\x67\x78",

			"\x01\x23\x45\x67\x89\xab\xcd\xef\x01\x12\x23\x34\x45\x56\x67\x78" \
					"\x89\x9a\xab\xbc\xcd\xde\xef\xf0",

			"\x01\x23\x45\x67\x89\xab\xcd\xef\x01\x12\x23\x34\x45\x56\x67\x78" \
					"\x89\x9a\xab\xbc\xcd\xde\xef\xf0\x10\x32\x54\x76\x98\xba\xdc\xfe"
	};

	static unsigned char RC6_PT[17] =
			"\x02\x13\x24\x35\x46\x57\x68\x79\x8a\x9b\xac\xbd\xce\xdf\xe0\xf1";

	static unsigned char RC6_CTs[3][17] =
	{
			"\x52\x4e\x19\x2f\x47\x15\xc6\x23\x1f\x51\xf6\x36\x7e\xa4\x3f\x18",
			"\x68\x83\x29\xd0\x19\xe5\x05\x04\x1e\x52\xe9\x2a\xf9\x52\x91\xd4",
			"\xc8\x24\x18\x16\xf0\xd7\xe4\x89\x20\xad\x16\xa1\x67\x4e\x5d\x48"
	};



		// returns true for pass
	bool test_RC6(void) {
			bool result=true; // false=fail, true=pass
			int i;
			word32 l_key[44];
			unsigned char data[16];
			unsigned char ciphertext[16];
			
			for( i = 0; i < 3; i++ )
			{
				memcpy( data, RC6_PT, 16 );
				
				rc6_set_key( &l_key[0], RC6_KEYs[i], 128 + i * 64 );
				rc6_encrypt( &l_key[0], data, ciphertext );
				
				if (memcmp( ciphertext, RC6_CTs[i], 16 ) )
				{
					result=0;
					return(result);
				}
			}
			
			for( i = 0; i < 3; i++ )
			{
				memcpy( ciphertext, RC6_CTs[i], 16 );
				
				rc6_set_key( &l_key[0], RC6_KEYs[i], 128 + i * 64 );
				rc6_decrypt( &l_key[0], ciphertext, data );
				
				if (memcmp( data, RC6_PT, 16 ) )
				{
					result=false;
					return(result);
				}
			}
			
			return(result);
	}



	/* NOTE: This implementation has been changed from the original
		 source.  See ChangeLog for more information.
		 Maintained by Alexander Kjeldaas <astor@fast.no>
	 */

	/* This is an independent implementation of the RC6 algorithm that	*/
	/* Ron Rivest and RSA Labs have submitted as a candidate for the	*/
	/* NIST AES activity.  Refer to RSA Labs and Ron Rivest for any 	*/
	/* copyright, patent or license issues for the RC6 algorithm.		*/
	/*																	*/
	/* Copyright in this implementation is held by Dr B R Gladman but	*/
	/* I hereby give permission for its free direct or derivative use	*/
	/* subject to acknowledgment of its origin and compliance with any	*/
	/* constraints that are placed on the exploitation of RC6 by its	*/
	/* designers.														*/
	/*																	*/
	/* Dr Brian Gladman (gladman@seven77.demon.co.uk) 18th July 1998	*/
	/*
		 Timing data:

	Algorithm: rc6 (rc62.c)
	128 bit key:
	Key Setup:    1580 cycles
	Encrypt:       286 cycles =    89.6 mbits/sec
	Decrypt:       236 cycles =   108.6 mbits/sec
	Mean:          261 cycles =    98.2 mbits/sec
	192 bit key:
	Key Setup:    1882 cycles
	Encrypt:       286 cycles =    89.5 mbits/sec
	Decrypt:       235 cycles =   108.9 mbits/sec
	Mean:          261 cycles =    98.3 mbits/sec
	256 bit key:
	Key Setup:    1774 cycles
	Encrypt:       285 cycles =    89.7 mbits/sec
	Decrypt:       236 cycles =   108.3 mbits/sec
	Mean:          261 cycles =    98.1 mbits/sec

	*/

//	#if 0
	#define rotl rotl32
	#define rotr rotr32
//	#else
//	#define rotl generic_rotl32
//	#define rotr generic_rotr32
//	#endif

	#define f_rnd(i,a,b,c,d)				   \
			u = rotl(d * (d + d + 1), 5);	   \
			t = rotl(b * (b + b + 1), 5);	   \
			a = rotl(a ^ t, u) + l_key[i];	   \
			c = rotl(c ^ u, t) + l_key[i + 1]

	#define i_rnd(i,a,b,c,d)				   \
			u = rotl(d * (d + d + 1), 5);	   \
			t = rotl(b * (b + b + 1), 5);	   \
			c = rotr(c - l_key[i + 1], t) ^ u; \
			a = rotr(a - l_key[i], u) ^ t


//	word32  l_key[44];  /* storage for the key schedule         */

	/* initialise the key schedule from the user supplied key	*/

	int rc6_set_key(word32 *l_key, const unsigned char *key, int key_len)					/* l_key - storage for the key schedule */
	{
		word32 *in_key = (u32 *)key;
		word32	i, j, k, a, b, l[8], t;

		key_len=key_len >> 3;
		if (key_len != 16 && key_len != 24 && key_len != 32)
			return -EINVAL; /* unsupported key length */

		key_len *= 8;

		l_key[0] = 0xb7e15163;

		for(k = 1; k < 44; ++k)
		
			l_key[k] = l_key[k - 1] + 0x9e3779b9;

		for(k = 0; k < (word32)key_len / 32; ++k)

			l[k] = in_key[k];

		t = (key_len / 32) - 1;

		a = b = i = j = 0;

		for(k = 0; k < 132; ++k)
		{	a = rotl(l_key[i] + a + b, 3); b += a;
			b = rotl(l[j] + b, b);
			l_key[i] = a; l[j] = b;
			i = (i == 43 ? 0 : i + 1);
			j = (j == t ? 0 : j + 1);
		}

		return 0;
	};

	/* encrypt a block of text	*/

	int rc6_encrypt(word32 *l_key, const word8 *in, word8 *out)
	{
		word32 *in_blk = (u32 *)in;
		word32 *out_blk = (u32 *)out;
		word32	a,b,c,d,t,u;

		a = in_blk[0]; b = in_blk[1] + l_key[0];
		c = in_blk[2]; d = in_blk[3] + l_key[1];

		f_rnd( 2,a,b,c,d); f_rnd( 4,b,c,d,a);
		f_rnd( 6,c,d,a,b); f_rnd( 8,d,a,b,c);
		f_rnd(10,a,b,c,d); f_rnd(12,b,c,d,a);
		f_rnd(14,c,d,a,b); f_rnd(16,d,a,b,c);
		f_rnd(18,a,b,c,d); f_rnd(20,b,c,d,a);
		f_rnd(22,c,d,a,b); f_rnd(24,d,a,b,c);
		f_rnd(26,a,b,c,d); f_rnd(28,b,c,d,a);
		f_rnd(30,c,d,a,b); f_rnd(32,d,a,b,c);
		f_rnd(34,a,b,c,d); f_rnd(36,b,c,d,a);
		f_rnd(38,c,d,a,b); f_rnd(40,d,a,b,c);

		out_blk[0] = a + l_key[42]; out_blk[1] = b;
		out_blk[2] = c + l_key[43]; out_blk[3] = d;
		return 0;
	};

	/* decrypt a block of text	*/

	int rc6_decrypt(word32 *l_key, const word8 *in, word8 *out)
	{ 
		word32 *in_blk = (u32 *)in;
		word32 *out_blk = (u32 *)out;
		word32 a,b,c,d,t,u;

		d = in_blk[3]; c = in_blk[2] - l_key[43]; 
		b = in_blk[1]; a = in_blk[0] - l_key[42];

		i_rnd(40,d,a,b,c); i_rnd(38,c,d,a,b);
		i_rnd(36,b,c,d,a); i_rnd(34,a,b,c,d);
		i_rnd(32,d,a,b,c); i_rnd(30,c,d,a,b);
		i_rnd(28,b,c,d,a); i_rnd(26,a,b,c,d);
		i_rnd(24,d,a,b,c); i_rnd(22,c,d,a,b);
		i_rnd(20,b,c,d,a); i_rnd(18,a,b,c,d);
		i_rnd(16,d,a,b,c); i_rnd(14,c,d,a,b);
		i_rnd(12,b,c,d,a); i_rnd(10,a,b,c,d);
		i_rnd( 8,d,a,b,c); i_rnd( 6,c,d,a,b);
		i_rnd( 4,b,c,d,a); i_rnd( 2,a,b,c,d);

		out_blk[3] = d - l_key[1]; out_blk[2] = c; 
		out_blk[1] = b - l_key[0]; out_blk[0] = a; 
		return 0;
	};
	
	#undef rotl
	#undef rotr
	#undef f_rnd
	#undef i_rnd










/* camellia.c ver 1.2.0
 *
 * Copyright (c) 2006,2007
 * NTT (Nippon Telegraph and Telephone Corporation) . All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer as
 *   the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NTT ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL NTT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Algorithm Specification
 *  http://info.isl.ntt.co.jp/crypt/eng/camellia/specifications.html
 */

//#include "config.h"

//#include <string.h>
//#include <stdlib.h>

//#include <krb5-types.h>
//#include "camellia-ntt.h"

//#include <roken.h>

/* key constants */

#define CAMELLIA_SIGMA1L (0xA09E667FL)
#define CAMELLIA_SIGMA1R (0x3BCC908BL)
#define CAMELLIA_SIGMA2L (0xB67AE858L)
#define CAMELLIA_SIGMA2R (0x4CAA73B2L)
#define CAMELLIA_SIGMA3L (0xC6EF372FL)
#define CAMELLIA_SIGMA3R (0xE94F82BEL)
#define CAMELLIA_SIGMA4L (0x54FF53A5L)
#define CAMELLIA_SIGMA4R (0xF1D36F1CL)
#define CAMELLIA_SIGMA5L (0x10E527FAL)
#define CAMELLIA_SIGMA5R (0xDE682D1DL)
#define CAMELLIA_SIGMA6L (0xB05688C2L)
#define CAMELLIA_SIGMA6R (0xB3E6C1FDL)

/*
 *  macros
 */


#if defined(_MSC_VER)

# define SWAP(x) (_lrotl(x, 8) & 0x00ff00ff | _lrotr(x, 8) & 0xff00ff00)
# define GETU32(p) SWAP(*((u32 *)(p)))
# define PUTU32(ct, st) {*((u32 *)(ct)) = SWAP((st));}

#else /* not MS-VC */

# define GETU32(pt)				\
    (((u32)(pt)[0] << 24)			\
     ^ ((u32)(pt)[1] << 16)			\
     ^ ((u32)(pt)[2] <<  8)			\
     ^ ((u32)(pt)[3]))

# define PUTU32(ct, st)  {			\
	(ct)[0] = (u8)((st) >> 24);		\
	(ct)[1] = (u8)((st) >> 16);		\
	(ct)[2] = (u8)((st) >>  8);		\
	(ct)[3] = (u8)(st); }

#endif

#define CamelliaSubkeyL(INDEX) (subkey[(INDEX)*2])
#define CamelliaSubkeyR(INDEX) (subkey[(INDEX)*2 + 1])

/* rotation right shift 1byte */
#define CAMELLIA_RR8(x) (((x) >> 8) + ((x) << 24))
/* rotation left shift 1bit */
#define CAMELLIA_RL1(x) (((x) << 1) + ((x) >> 31))
/* rotation left shift 1byte */
#define CAMELLIA_RL8(x) (((x) << 8) + ((x) >> 24))

#define CAMELLIA_ROLDQ(ll, lr, rl, rr, w0, w1, bits)	\
    do {						\
	w0 = ll;					\
	ll = (ll << bits) + (lr >> (32 - bits));	\
	lr = (lr << bits) + (rl >> (32 - bits));	\
	rl = (rl << bits) + (rr >> (32 - bits));	\
	rr = (rr << bits) + (w0 >> (32 - bits));	\
    } while(0)

#define CAMELLIA_ROLDQo32(ll, lr, rl, rr, w0, w1, bits)	\
    do {						\
	w0 = ll;					\
	w1 = lr;					\
	ll = (lr << (bits - 32)) + (rl >> (64 - bits));	\
	lr = (rl << (bits - 32)) + (rr >> (64 - bits));	\
	rl = (rr << (bits - 32)) + (w0 >> (64 - bits));	\
	rr = (w0 << (bits - 32)) + (w1 >> (64 - bits));	\
    } while(0)

#define CAMELLIA_SP1110(INDEX) (camellia_sp1110[(INDEX)])
#define CAMELLIA_SP0222(INDEX) (camellia_sp0222[(INDEX)])
#define CAMELLIA_SP3033(INDEX) (camellia_sp3033[(INDEX)])
#define CAMELLIA_SP4404(INDEX) (camellia_sp4404[(INDEX)])

#define CAMELLIA_F(xl, xr, kl, kr, yl, yr, il, ir, t0, t1)	\
    do {							\
	il = xl ^ kl;						\
	ir = xr ^ kr;						\
	t0 = il >> 16;						\
	t1 = ir >> 16;						\
	yl = CAMELLIA_SP1110(ir & 0xff)				\
	    ^ CAMELLIA_SP0222((t1 >> 8) & 0xff)			\
	    ^ CAMELLIA_SP3033(t1 & 0xff)			\
	    ^ CAMELLIA_SP4404((ir >> 8) & 0xff);		\
	yr = CAMELLIA_SP1110((t0 >> 8) & 0xff)			\
	    ^ CAMELLIA_SP0222(t0 & 0xff)			\
	    ^ CAMELLIA_SP3033((il >> 8) & 0xff)			\
	    ^ CAMELLIA_SP4404(il & 0xff);			\
	yl ^= yr;						\
	yr = CAMELLIA_RR8(yr);					\
	yr ^= yl;						\
    } while(0)


/*
 * for speed up
 *
 */
#define CAMELLIA_FLS(ll, lr, rl, rr, kll, klr, krl, krr, t0, t1, t2, t3) \
    do {								\
	t0 = kll;							\
	t0 &= ll;							\
	lr ^= CAMELLIA_RL1(t0);						\
	t1 = klr;							\
	t1 |= lr;							\
	ll ^= t1;							\
									\
	t2 = krr;							\
	t2 |= rr;							\
	rl ^= t2;							\
	t3 = krl;							\
	t3 &= rl;							\
	rr ^= CAMELLIA_RL1(t3);						\
    } while(0)

#define CAMELLIA_ROUNDSM(xl, xr, kl, kr, yl, yr, il, ir, t0, t1)	\
    do {								\
	ir = CAMELLIA_SP1110(xr & 0xff)					\
	    ^ CAMELLIA_SP0222((xr >> 24) & 0xff)			\
	    ^ CAMELLIA_SP3033((xr >> 16) & 0xff)			\
	    ^ CAMELLIA_SP4404((xr >> 8) & 0xff);			\
	il = CAMELLIA_SP1110((xl >> 24) & 0xff)				\
	    ^ CAMELLIA_SP0222((xl >> 16) & 0xff)			\
	    ^ CAMELLIA_SP3033((xl >> 8) & 0xff)				\
	    ^ CAMELLIA_SP4404(xl & 0xff);				\
	il ^= kl;							\
	ir ^= kr;							\
	ir ^= il;							\
	il = CAMELLIA_RR8(il);						\
	il ^= ir;							\
	yl ^= ir;							\
	yr ^= il;							\
    } while(0)


static const u32 camellia_sp1110[256] = {
    0x70707000,0x82828200,0x2c2c2c00,0xececec00,
    0xb3b3b300,0x27272700,0xc0c0c000,0xe5e5e500,
    0xe4e4e400,0x85858500,0x57575700,0x35353500,
    0xeaeaea00,0x0c0c0c00,0xaeaeae00,0x41414100,
    0x23232300,0xefefef00,0x6b6b6b00,0x93939300,
    0x45454500,0x19191900,0xa5a5a500,0x21212100,
    0xededed00,0x0e0e0e00,0x4f4f4f00,0x4e4e4e00,
    0x1d1d1d00,0x65656500,0x92929200,0xbdbdbd00,
    0x86868600,0xb8b8b800,0xafafaf00,0x8f8f8f00,
    0x7c7c7c00,0xebebeb00,0x1f1f1f00,0xcecece00,
    0x3e3e3e00,0x30303000,0xdcdcdc00,0x5f5f5f00,
    0x5e5e5e00,0xc5c5c500,0x0b0b0b00,0x1a1a1a00,
    0xa6a6a600,0xe1e1e100,0x39393900,0xcacaca00,
    0xd5d5d500,0x47474700,0x5d5d5d00,0x3d3d3d00,
    0xd9d9d900,0x01010100,0x5a5a5a00,0xd6d6d600,
    0x51515100,0x56565600,0x6c6c6c00,0x4d4d4d00,
    0x8b8b8b00,0x0d0d0d00,0x9a9a9a00,0x66666600,
    0xfbfbfb00,0xcccccc00,0xb0b0b000,0x2d2d2d00,
    0x74747400,0x12121200,0x2b2b2b00,0x20202000,
    0xf0f0f000,0xb1b1b100,0x84848400,0x99999900,
    0xdfdfdf00,0x4c4c4c00,0xcbcbcb00,0xc2c2c200,
    0x34343400,0x7e7e7e00,0x76767600,0x05050500,
    0x6d6d6d00,0xb7b7b700,0xa9a9a900,0x31313100,
    0xd1d1d100,0x17171700,0x04040400,0xd7d7d700,
    0x14141400,0x58585800,0x3a3a3a00,0x61616100,
    0xdedede00,0x1b1b1b00,0x11111100,0x1c1c1c00,
    0x32323200,0x0f0f0f00,0x9c9c9c00,0x16161600,
    0x53535300,0x18181800,0xf2f2f200,0x22222200,
    0xfefefe00,0x44444400,0xcfcfcf00,0xb2b2b200,
    0xc3c3c300,0xb5b5b500,0x7a7a7a00,0x91919100,
    0x24242400,0x08080800,0xe8e8e800,0xa8a8a800,
    0x60606000,0xfcfcfc00,0x69696900,0x50505000,
    0xaaaaaa00,0xd0d0d000,0xa0a0a000,0x7d7d7d00,
    0xa1a1a100,0x89898900,0x62626200,0x97979700,
    0x54545400,0x5b5b5b00,0x1e1e1e00,0x95959500,
    0xe0e0e000,0xffffff00,0x64646400,0xd2d2d200,
    0x10101000,0xc4c4c400,0x00000000,0x48484800,
    0xa3a3a300,0xf7f7f700,0x75757500,0xdbdbdb00,
    0x8a8a8a00,0x03030300,0xe6e6e600,0xdadada00,
    0x09090900,0x3f3f3f00,0xdddddd00,0x94949400,
    0x87878700,0x5c5c5c00,0x83838300,0x02020200,
    0xcdcdcd00,0x4a4a4a00,0x90909000,0x33333300,
    0x73737300,0x67676700,0xf6f6f600,0xf3f3f300,
    0x9d9d9d00,0x7f7f7f00,0xbfbfbf00,0xe2e2e200,
    0x52525200,0x9b9b9b00,0xd8d8d800,0x26262600,
    0xc8c8c800,0x37373700,0xc6c6c600,0x3b3b3b00,
    0x81818100,0x96969600,0x6f6f6f00,0x4b4b4b00,
    0x13131300,0xbebebe00,0x63636300,0x2e2e2e00,
    0xe9e9e900,0x79797900,0xa7a7a700,0x8c8c8c00,
    0x9f9f9f00,0x6e6e6e00,0xbcbcbc00,0x8e8e8e00,
    0x29292900,0xf5f5f500,0xf9f9f900,0xb6b6b600,
    0x2f2f2f00,0xfdfdfd00,0xb4b4b400,0x59595900,
    0x78787800,0x98989800,0x06060600,0x6a6a6a00,
    0xe7e7e700,0x46464600,0x71717100,0xbababa00,
    0xd4d4d400,0x25252500,0xababab00,0x42424200,
    0x88888800,0xa2a2a200,0x8d8d8d00,0xfafafa00,
    0x72727200,0x07070700,0xb9b9b900,0x55555500,
    0xf8f8f800,0xeeeeee00,0xacacac00,0x0a0a0a00,
    0x36363600,0x49494900,0x2a2a2a00,0x68686800,
    0x3c3c3c00,0x38383800,0xf1f1f100,0xa4a4a400,
    0x40404000,0x28282800,0xd3d3d300,0x7b7b7b00,
    0xbbbbbb00,0xc9c9c900,0x43434300,0xc1c1c100,
    0x15151500,0xe3e3e300,0xadadad00,0xf4f4f400,
    0x77777700,0xc7c7c700,0x80808000,0x9e9e9e00,
};

static const u32 camellia_sp0222[256] = {
    0x00e0e0e0,0x00050505,0x00585858,0x00d9d9d9,
    0x00676767,0x004e4e4e,0x00818181,0x00cbcbcb,
    0x00c9c9c9,0x000b0b0b,0x00aeaeae,0x006a6a6a,
    0x00d5d5d5,0x00181818,0x005d5d5d,0x00828282,
    0x00464646,0x00dfdfdf,0x00d6d6d6,0x00272727,
    0x008a8a8a,0x00323232,0x004b4b4b,0x00424242,
    0x00dbdbdb,0x001c1c1c,0x009e9e9e,0x009c9c9c,
    0x003a3a3a,0x00cacaca,0x00252525,0x007b7b7b,
    0x000d0d0d,0x00717171,0x005f5f5f,0x001f1f1f,
    0x00f8f8f8,0x00d7d7d7,0x003e3e3e,0x009d9d9d,
    0x007c7c7c,0x00606060,0x00b9b9b9,0x00bebebe,
    0x00bcbcbc,0x008b8b8b,0x00161616,0x00343434,
    0x004d4d4d,0x00c3c3c3,0x00727272,0x00959595,
    0x00ababab,0x008e8e8e,0x00bababa,0x007a7a7a,
    0x00b3b3b3,0x00020202,0x00b4b4b4,0x00adadad,
    0x00a2a2a2,0x00acacac,0x00d8d8d8,0x009a9a9a,
    0x00171717,0x001a1a1a,0x00353535,0x00cccccc,
    0x00f7f7f7,0x00999999,0x00616161,0x005a5a5a,
    0x00e8e8e8,0x00242424,0x00565656,0x00404040,
    0x00e1e1e1,0x00636363,0x00090909,0x00333333,
    0x00bfbfbf,0x00989898,0x00979797,0x00858585,
    0x00686868,0x00fcfcfc,0x00ececec,0x000a0a0a,
    0x00dadada,0x006f6f6f,0x00535353,0x00626262,
    0x00a3a3a3,0x002e2e2e,0x00080808,0x00afafaf,
    0x00282828,0x00b0b0b0,0x00747474,0x00c2c2c2,
    0x00bdbdbd,0x00363636,0x00222222,0x00383838,
    0x00646464,0x001e1e1e,0x00393939,0x002c2c2c,
    0x00a6a6a6,0x00303030,0x00e5e5e5,0x00444444,
    0x00fdfdfd,0x00888888,0x009f9f9f,0x00656565,
    0x00878787,0x006b6b6b,0x00f4f4f4,0x00232323,
    0x00484848,0x00101010,0x00d1d1d1,0x00515151,
    0x00c0c0c0,0x00f9f9f9,0x00d2d2d2,0x00a0a0a0,
    0x00555555,0x00a1a1a1,0x00414141,0x00fafafa,
    0x00434343,0x00131313,0x00c4c4c4,0x002f2f2f,
    0x00a8a8a8,0x00b6b6b6,0x003c3c3c,0x002b2b2b,
    0x00c1c1c1,0x00ffffff,0x00c8c8c8,0x00a5a5a5,
    0x00202020,0x00898989,0x00000000,0x00909090,
    0x00474747,0x00efefef,0x00eaeaea,0x00b7b7b7,
    0x00151515,0x00060606,0x00cdcdcd,0x00b5b5b5,
    0x00121212,0x007e7e7e,0x00bbbbbb,0x00292929,
    0x000f0f0f,0x00b8b8b8,0x00070707,0x00040404,
    0x009b9b9b,0x00949494,0x00212121,0x00666666,
    0x00e6e6e6,0x00cecece,0x00ededed,0x00e7e7e7,
    0x003b3b3b,0x00fefefe,0x007f7f7f,0x00c5c5c5,
    0x00a4a4a4,0x00373737,0x00b1b1b1,0x004c4c4c,
    0x00919191,0x006e6e6e,0x008d8d8d,0x00767676,
    0x00030303,0x002d2d2d,0x00dedede,0x00969696,
    0x00262626,0x007d7d7d,0x00c6c6c6,0x005c5c5c,
    0x00d3d3d3,0x00f2f2f2,0x004f4f4f,0x00191919,
    0x003f3f3f,0x00dcdcdc,0x00797979,0x001d1d1d,
    0x00525252,0x00ebebeb,0x00f3f3f3,0x006d6d6d,
    0x005e5e5e,0x00fbfbfb,0x00696969,0x00b2b2b2,
    0x00f0f0f0,0x00313131,0x000c0c0c,0x00d4d4d4,
    0x00cfcfcf,0x008c8c8c,0x00e2e2e2,0x00757575,
    0x00a9a9a9,0x004a4a4a,0x00575757,0x00848484,
    0x00111111,0x00454545,0x001b1b1b,0x00f5f5f5,
    0x00e4e4e4,0x000e0e0e,0x00737373,0x00aaaaaa,
    0x00f1f1f1,0x00dddddd,0x00595959,0x00141414,
    0x006c6c6c,0x00929292,0x00545454,0x00d0d0d0,
    0x00787878,0x00707070,0x00e3e3e3,0x00494949,
    0x00808080,0x00505050,0x00a7a7a7,0x00f6f6f6,
    0x00777777,0x00939393,0x00868686,0x00838383,
    0x002a2a2a,0x00c7c7c7,0x005b5b5b,0x00e9e9e9,
    0x00eeeeee,0x008f8f8f,0x00010101,0x003d3d3d,
};

static const u32 camellia_sp3033[256] = {
    0x38003838,0x41004141,0x16001616,0x76007676,
    0xd900d9d9,0x93009393,0x60006060,0xf200f2f2,
    0x72007272,0xc200c2c2,0xab00abab,0x9a009a9a,
    0x75007575,0x06000606,0x57005757,0xa000a0a0,
    0x91009191,0xf700f7f7,0xb500b5b5,0xc900c9c9,
    0xa200a2a2,0x8c008c8c,0xd200d2d2,0x90009090,
    0xf600f6f6,0x07000707,0xa700a7a7,0x27002727,
    0x8e008e8e,0xb200b2b2,0x49004949,0xde00dede,
    0x43004343,0x5c005c5c,0xd700d7d7,0xc700c7c7,
    0x3e003e3e,0xf500f5f5,0x8f008f8f,0x67006767,
    0x1f001f1f,0x18001818,0x6e006e6e,0xaf00afaf,
    0x2f002f2f,0xe200e2e2,0x85008585,0x0d000d0d,
    0x53005353,0xf000f0f0,0x9c009c9c,0x65006565,
    0xea00eaea,0xa300a3a3,0xae00aeae,0x9e009e9e,
    0xec00ecec,0x80008080,0x2d002d2d,0x6b006b6b,
    0xa800a8a8,0x2b002b2b,0x36003636,0xa600a6a6,
    0xc500c5c5,0x86008686,0x4d004d4d,0x33003333,
    0xfd00fdfd,0x66006666,0x58005858,0x96009696,
    0x3a003a3a,0x09000909,0x95009595,0x10001010,
    0x78007878,0xd800d8d8,0x42004242,0xcc00cccc,
    0xef00efef,0x26002626,0xe500e5e5,0x61006161,
    0x1a001a1a,0x3f003f3f,0x3b003b3b,0x82008282,
    0xb600b6b6,0xdb00dbdb,0xd400d4d4,0x98009898,
    0xe800e8e8,0x8b008b8b,0x02000202,0xeb00ebeb,
    0x0a000a0a,0x2c002c2c,0x1d001d1d,0xb000b0b0,
    0x6f006f6f,0x8d008d8d,0x88008888,0x0e000e0e,
    0x19001919,0x87008787,0x4e004e4e,0x0b000b0b,
    0xa900a9a9,0x0c000c0c,0x79007979,0x11001111,
    0x7f007f7f,0x22002222,0xe700e7e7,0x59005959,
    0xe100e1e1,0xda00dada,0x3d003d3d,0xc800c8c8,
    0x12001212,0x04000404,0x74007474,0x54005454,
    0x30003030,0x7e007e7e,0xb400b4b4,0x28002828,
    0x55005555,0x68006868,0x50005050,0xbe00bebe,
    0xd000d0d0,0xc400c4c4,0x31003131,0xcb00cbcb,
    0x2a002a2a,0xad00adad,0x0f000f0f,0xca00caca,
    0x70007070,0xff00ffff,0x32003232,0x69006969,
    0x08000808,0x62006262,0x00000000,0x24002424,
    0xd100d1d1,0xfb00fbfb,0xba00baba,0xed00eded,
    0x45004545,0x81008181,0x73007373,0x6d006d6d,
    0x84008484,0x9f009f9f,0xee00eeee,0x4a004a4a,
    0xc300c3c3,0x2e002e2e,0xc100c1c1,0x01000101,
    0xe600e6e6,0x25002525,0x48004848,0x99009999,
    0xb900b9b9,0xb300b3b3,0x7b007b7b,0xf900f9f9,
    0xce00cece,0xbf00bfbf,0xdf00dfdf,0x71007171,
    0x29002929,0xcd00cdcd,0x6c006c6c,0x13001313,
    0x64006464,0x9b009b9b,0x63006363,0x9d009d9d,
    0xc000c0c0,0x4b004b4b,0xb700b7b7,0xa500a5a5,
    0x89008989,0x5f005f5f,0xb100b1b1,0x17001717,
    0xf400f4f4,0xbc00bcbc,0xd300d3d3,0x46004646,
    0xcf00cfcf,0x37003737,0x5e005e5e,0x47004747,
    0x94009494,0xfa00fafa,0xfc00fcfc,0x5b005b5b,
    0x97009797,0xfe00fefe,0x5a005a5a,0xac00acac,
    0x3c003c3c,0x4c004c4c,0x03000303,0x35003535,
    0xf300f3f3,0x23002323,0xb800b8b8,0x5d005d5d,
    0x6a006a6a,0x92009292,0xd500d5d5,0x21002121,
    0x44004444,0x51005151,0xc600c6c6,0x7d007d7d,
    0x39003939,0x83008383,0xdc00dcdc,0xaa00aaaa,
    0x7c007c7c,0x77007777,0x56005656,0x05000505,
    0x1b001b1b,0xa400a4a4,0x15001515,0x34003434,
    0x1e001e1e,0x1c001c1c,0xf800f8f8,0x52005252,
    0x20002020,0x14001414,0xe900e9e9,0xbd00bdbd,
    0xdd00dddd,0xe400e4e4,0xa100a1a1,0xe000e0e0,
    0x8a008a8a,0xf100f1f1,0xd600d6d6,0x7a007a7a,
    0xbb00bbbb,0xe300e3e3,0x40004040,0x4f004f4f,
};

static const u32 camellia_sp4404[256] = {
    0x70700070,0x2c2c002c,0xb3b300b3,0xc0c000c0,
    0xe4e400e4,0x57570057,0xeaea00ea,0xaeae00ae,
    0x23230023,0x6b6b006b,0x45450045,0xa5a500a5,
    0xeded00ed,0x4f4f004f,0x1d1d001d,0x92920092,
    0x86860086,0xafaf00af,0x7c7c007c,0x1f1f001f,
    0x3e3e003e,0xdcdc00dc,0x5e5e005e,0x0b0b000b,
    0xa6a600a6,0x39390039,0xd5d500d5,0x5d5d005d,
    0xd9d900d9,0x5a5a005a,0x51510051,0x6c6c006c,
    0x8b8b008b,0x9a9a009a,0xfbfb00fb,0xb0b000b0,
    0x74740074,0x2b2b002b,0xf0f000f0,0x84840084,
    0xdfdf00df,0xcbcb00cb,0x34340034,0x76760076,
    0x6d6d006d,0xa9a900a9,0xd1d100d1,0x04040004,
    0x14140014,0x3a3a003a,0xdede00de,0x11110011,
    0x32320032,0x9c9c009c,0x53530053,0xf2f200f2,
    0xfefe00fe,0xcfcf00cf,0xc3c300c3,0x7a7a007a,
    0x24240024,0xe8e800e8,0x60600060,0x69690069,
    0xaaaa00aa,0xa0a000a0,0xa1a100a1,0x62620062,
    0x54540054,0x1e1e001e,0xe0e000e0,0x64640064,
    0x10100010,0x00000000,0xa3a300a3,0x75750075,
    0x8a8a008a,0xe6e600e6,0x09090009,0xdddd00dd,
    0x87870087,0x83830083,0xcdcd00cd,0x90900090,
    0x73730073,0xf6f600f6,0x9d9d009d,0xbfbf00bf,
    0x52520052,0xd8d800d8,0xc8c800c8,0xc6c600c6,
    0x81810081,0x6f6f006f,0x13130013,0x63630063,
    0xe9e900e9,0xa7a700a7,0x9f9f009f,0xbcbc00bc,
    0x29290029,0xf9f900f9,0x2f2f002f,0xb4b400b4,
    0x78780078,0x06060006,0xe7e700e7,0x71710071,
    0xd4d400d4,0xabab00ab,0x88880088,0x8d8d008d,
    0x72720072,0xb9b900b9,0xf8f800f8,0xacac00ac,
    0x36360036,0x2a2a002a,0x3c3c003c,0xf1f100f1,
    0x40400040,0xd3d300d3,0xbbbb00bb,0x43430043,
    0x15150015,0xadad00ad,0x77770077,0x80800080,
    0x82820082,0xecec00ec,0x27270027,0xe5e500e5,
    0x85850085,0x35350035,0x0c0c000c,0x41410041,
    0xefef00ef,0x93930093,0x19190019,0x21210021,
    0x0e0e000e,0x4e4e004e,0x65650065,0xbdbd00bd,
    0xb8b800b8,0x8f8f008f,0xebeb00eb,0xcece00ce,
    0x30300030,0x5f5f005f,0xc5c500c5,0x1a1a001a,
    0xe1e100e1,0xcaca00ca,0x47470047,0x3d3d003d,
    0x01010001,0xd6d600d6,0x56560056,0x4d4d004d,
    0x0d0d000d,0x66660066,0xcccc00cc,0x2d2d002d,
    0x12120012,0x20200020,0xb1b100b1,0x99990099,
    0x4c4c004c,0xc2c200c2,0x7e7e007e,0x05050005,
    0xb7b700b7,0x31310031,0x17170017,0xd7d700d7,
    0x58580058,0x61610061,0x1b1b001b,0x1c1c001c,
    0x0f0f000f,0x16160016,0x18180018,0x22220022,
    0x44440044,0xb2b200b2,0xb5b500b5,0x91910091,
    0x08080008,0xa8a800a8,0xfcfc00fc,0x50500050,
    0xd0d000d0,0x7d7d007d,0x89890089,0x97970097,
    0x5b5b005b,0x95950095,0xffff00ff,0xd2d200d2,
    0xc4c400c4,0x48480048,0xf7f700f7,0xdbdb00db,
    0x03030003,0xdada00da,0x3f3f003f,0x94940094,
    0x5c5c005c,0x02020002,0x4a4a004a,0x33330033,
    0x67670067,0xf3f300f3,0x7f7f007f,0xe2e200e2,
    0x9b9b009b,0x26260026,0x37370037,0x3b3b003b,
    0x96960096,0x4b4b004b,0xbebe00be,0x2e2e002e,
    0x79790079,0x8c8c008c,0x6e6e006e,0x8e8e008e,
    0xf5f500f5,0xb6b600b6,0xfdfd00fd,0x59590059,
    0x98980098,0x6a6a006a,0x46460046,0xbaba00ba,
    0x25250025,0x42420042,0xa2a200a2,0xfafa00fa,
    0x07070007,0x55550055,0xeeee00ee,0x0a0a000a,
    0x49490049,0x68680068,0x38380038,0xa4a400a4,
    0x28280028,0x7b7b007b,0xc9c900c9,0xc1c100c1,
    0xe3e300e3,0xf4f400f4,0xc7c700c7,0x9e9e009e,
};


/**
 * Stuff related to the Camellia key schedule
 */
#define subl(x) subL[(x)]
#define subr(x) subR[(x)]

static void camellia_setup128(const unsigned char *key, u32 *subkey)
{
    u32 kll, klr, krl, krr;
    u32 il, ir, t0, t1, w0, w1;
    u32 kw4l, kw4r, dw, tl, tr;
    u32 subL[26];
    u32 subR[26];

    /**
     *  k == kll || klr || krl || krr (|| is concatination)
     */
    kll = GETU32(key     );
    klr = GETU32(key +  4);
    krl = GETU32(key +  8);
    krr = GETU32(key + 12);
    /**
     * generate KL dependent subkeys
     */
    subl(0) = kll; subr(0) = klr;
    subl(1) = krl; subr(1) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 15);
    subl(4) = kll; subr(4) = klr;
    subl(5) = krl; subr(5) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 30);
    subl(10) = kll; subr(10) = klr;
    subl(11) = krl; subr(11) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 15);
    subl(13) = krl; subr(13) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 17);
    subl(16) = kll; subr(16) = klr;
    subl(17) = krl; subr(17) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 17);
    subl(18) = kll; subr(18) = klr;
    subl(19) = krl; subr(19) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 17);
    subl(22) = kll; subr(22) = klr;
    subl(23) = krl; subr(23) = krr;

    /* generate KA */
    kll = subl(0); klr = subr(0);
    krl = subl(1); krr = subr(1);
    CAMELLIA_F(kll, klr,
	       CAMELLIA_SIGMA1L, CAMELLIA_SIGMA1R,
	       w0, w1, il, ir, t0, t1);
    krl ^= w0; krr ^= w1;
    CAMELLIA_F(krl, krr,
	       CAMELLIA_SIGMA2L, CAMELLIA_SIGMA2R,
	       kll, klr, il, ir, t0, t1);
    CAMELLIA_F(kll, klr,
	       CAMELLIA_SIGMA3L, CAMELLIA_SIGMA3R,
	       krl, krr, il, ir, t0, t1);
    krl ^= w0; krr ^= w1;
    CAMELLIA_F(krl, krr,
	       CAMELLIA_SIGMA4L, CAMELLIA_SIGMA4R,
	       w0, w1, il, ir, t0, t1);
    kll ^= w0; klr ^= w1;

    /* generate KA dependent subkeys */
    subl(2) = kll; subr(2) = klr;
    subl(3) = krl; subr(3) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 15);
    subl(6) = kll; subr(6) = klr;
    subl(7) = krl; subr(7) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 15);
    subl(8) = kll; subr(8) = klr;
    subl(9) = krl; subr(9) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 15);
    subl(12) = kll; subr(12) = klr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 15);
    subl(14) = kll; subr(14) = klr;
    subl(15) = krl; subr(15) = krr;
    CAMELLIA_ROLDQo32(kll, klr, krl, krr, w0, w1, 34);
    subl(20) = kll; subr(20) = klr;
    subl(21) = krl; subr(21) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 17);
    subl(24) = kll; subr(24) = klr;
    subl(25) = krl; subr(25) = krr;


    /* absorb kw2 to other subkeys */
    subl(3) ^= subl(1); subr(3) ^= subr(1);
    subl(5) ^= subl(1); subr(5) ^= subr(1);
    subl(7) ^= subl(1); subr(7) ^= subr(1);
    subl(1) ^= subr(1) & ~subr(9);
    dw = subl(1) & subl(9), subr(1) ^= CAMELLIA_RL1(dw);
    subl(11) ^= subl(1); subr(11) ^= subr(1);
    subl(13) ^= subl(1); subr(13) ^= subr(1);
    subl(15) ^= subl(1); subr(15) ^= subr(1);
    subl(1) ^= subr(1) & ~subr(17);
    dw = subl(1) & subl(17), subr(1) ^= CAMELLIA_RL1(dw);
    subl(19) ^= subl(1); subr(19) ^= subr(1);
    subl(21) ^= subl(1); subr(21) ^= subr(1);
    subl(23) ^= subl(1); subr(23) ^= subr(1);
    subl(24) ^= subl(1); subr(24) ^= subr(1);

    /* absorb kw4 to other subkeys */
    kw4l = subl(25); kw4r = subr(25);
    subl(22) ^= kw4l; subr(22) ^= kw4r;
    subl(20) ^= kw4l; subr(20) ^= kw4r;
    subl(18) ^= kw4l; subr(18) ^= kw4r;
    kw4l ^= kw4r & ~subr(16);
    dw = kw4l & subl(16), kw4r ^= CAMELLIA_RL1(dw);
    subl(14) ^= kw4l; subr(14) ^= kw4r;
    subl(12) ^= kw4l; subr(12) ^= kw4r;
    subl(10) ^= kw4l; subr(10) ^= kw4r;
    kw4l ^= kw4r & ~subr(8);
    dw = kw4l & subl(8), kw4r ^= CAMELLIA_RL1(dw);
    subl(6) ^= kw4l; subr(6) ^= kw4r;
    subl(4) ^= kw4l; subr(4) ^= kw4r;
    subl(2) ^= kw4l; subr(2) ^= kw4r;
    subl(0) ^= kw4l; subr(0) ^= kw4r;

    /* key XOR is end of F-function */
    CamelliaSubkeyL(0) = subl(0) ^ subl(2);
    CamelliaSubkeyR(0) = subr(0) ^ subr(2);
    CamelliaSubkeyL(2) = subl(3);
    CamelliaSubkeyR(2) = subr(3);
    CamelliaSubkeyL(3) = subl(2) ^ subl(4);
    CamelliaSubkeyR(3) = subr(2) ^ subr(4);
    CamelliaSubkeyL(4) = subl(3) ^ subl(5);
    CamelliaSubkeyR(4) = subr(3) ^ subr(5);
    CamelliaSubkeyL(5) = subl(4) ^ subl(6);
    CamelliaSubkeyR(5) = subr(4) ^ subr(6);
    CamelliaSubkeyL(6) = subl(5) ^ subl(7);
    CamelliaSubkeyR(6) = subr(5) ^ subr(7);
    tl = subl(10) ^ (subr(10) & ~subr(8));
    dw = tl & subl(8), tr = subr(10) ^ CAMELLIA_RL1(dw);
    CamelliaSubkeyL(7) = subl(6) ^ tl;
    CamelliaSubkeyR(7) = subr(6) ^ tr;
    CamelliaSubkeyL(8) = subl(8);
    CamelliaSubkeyR(8) = subr(8);
    CamelliaSubkeyL(9) = subl(9);
    CamelliaSubkeyR(9) = subr(9);
    tl = subl(7) ^ (subr(7) & ~subr(9));
    dw = tl & subl(9), tr = subr(7) ^ CAMELLIA_RL1(dw);
    CamelliaSubkeyL(10) = tl ^ subl(11);
    CamelliaSubkeyR(10) = tr ^ subr(11);
    CamelliaSubkeyL(11) = subl(10) ^ subl(12);
    CamelliaSubkeyR(11) = subr(10) ^ subr(12);
    CamelliaSubkeyL(12) = subl(11) ^ subl(13);
    CamelliaSubkeyR(12) = subr(11) ^ subr(13);
    CamelliaSubkeyL(13) = subl(12) ^ subl(14);
    CamelliaSubkeyR(13) = subr(12) ^ subr(14);
    CamelliaSubkeyL(14) = subl(13) ^ subl(15);
    CamelliaSubkeyR(14) = subr(13) ^ subr(15);
    tl = subl(18) ^ (subr(18) & ~subr(16));
    dw = tl & subl(16),	tr = subr(18) ^ CAMELLIA_RL1(dw);
    CamelliaSubkeyL(15) = subl(14) ^ tl;
    CamelliaSubkeyR(15) = subr(14) ^ tr;
    CamelliaSubkeyL(16) = subl(16);
    CamelliaSubkeyR(16) = subr(16);
    CamelliaSubkeyL(17) = subl(17);
    CamelliaSubkeyR(17) = subr(17);
    tl = subl(15) ^ (subr(15) & ~subr(17));
    dw = tl & subl(17),	tr = subr(15) ^ CAMELLIA_RL1(dw);
    CamelliaSubkeyL(18) = tl ^ subl(19);
    CamelliaSubkeyR(18) = tr ^ subr(19);
    CamelliaSubkeyL(19) = subl(18) ^ subl(20);
    CamelliaSubkeyR(19) = subr(18) ^ subr(20);
    CamelliaSubkeyL(20) = subl(19) ^ subl(21);
    CamelliaSubkeyR(20) = subr(19) ^ subr(21);
    CamelliaSubkeyL(21) = subl(20) ^ subl(22);
    CamelliaSubkeyR(21) = subr(20) ^ subr(22);
    CamelliaSubkeyL(22) = subl(21) ^ subl(23);
    CamelliaSubkeyR(22) = subr(21) ^ subr(23);
    CamelliaSubkeyL(23) = subl(22);
    CamelliaSubkeyR(23) = subr(22);
    CamelliaSubkeyL(24) = subl(24) ^ subl(23);
    CamelliaSubkeyR(24) = subr(24) ^ subr(23);

    /* apply the inverse of the last half of P-function */
    dw = CamelliaSubkeyL(2) ^ CamelliaSubkeyR(2), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(2) = CamelliaSubkeyL(2) ^ dw, CamelliaSubkeyL(2) = dw;
    dw = CamelliaSubkeyL(3) ^ CamelliaSubkeyR(3), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(3) = CamelliaSubkeyL(3) ^ dw, CamelliaSubkeyL(3) = dw;
    dw = CamelliaSubkeyL(4) ^ CamelliaSubkeyR(4), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(4) = CamelliaSubkeyL(4) ^ dw, CamelliaSubkeyL(4) = dw;
    dw = CamelliaSubkeyL(5) ^ CamelliaSubkeyR(5), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(5) = CamelliaSubkeyL(5) ^ dw, CamelliaSubkeyL(5) = dw;
    dw = CamelliaSubkeyL(6) ^ CamelliaSubkeyR(6), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(6) = CamelliaSubkeyL(6) ^ dw, CamelliaSubkeyL(6) = dw;
    dw = CamelliaSubkeyL(7) ^ CamelliaSubkeyR(7), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(7) = CamelliaSubkeyL(7) ^ dw, CamelliaSubkeyL(7) = dw;
    dw = CamelliaSubkeyL(10) ^ CamelliaSubkeyR(10), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(10) = CamelliaSubkeyL(10) ^ dw, CamelliaSubkeyL(10) = dw;
    dw = CamelliaSubkeyL(11) ^ CamelliaSubkeyR(11), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(11) = CamelliaSubkeyL(11) ^ dw, CamelliaSubkeyL(11) = dw;
    dw = CamelliaSubkeyL(12) ^ CamelliaSubkeyR(12), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(12) = CamelliaSubkeyL(12) ^ dw, CamelliaSubkeyL(12) = dw;
    dw = CamelliaSubkeyL(13) ^ CamelliaSubkeyR(13), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(13) = CamelliaSubkeyL(13) ^ dw, CamelliaSubkeyL(13) = dw;
    dw = CamelliaSubkeyL(14) ^ CamelliaSubkeyR(14), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(14) = CamelliaSubkeyL(14) ^ dw, CamelliaSubkeyL(14) = dw;
    dw = CamelliaSubkeyL(15) ^ CamelliaSubkeyR(15), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(15) = CamelliaSubkeyL(15) ^ dw, CamelliaSubkeyL(15) = dw;
    dw = CamelliaSubkeyL(18) ^ CamelliaSubkeyR(18), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(18) = CamelliaSubkeyL(18) ^ dw, CamelliaSubkeyL(18) = dw;
    dw = CamelliaSubkeyL(19) ^ CamelliaSubkeyR(19), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(19) = CamelliaSubkeyL(19) ^ dw, CamelliaSubkeyL(19) = dw;
    dw = CamelliaSubkeyL(20) ^ CamelliaSubkeyR(20), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(20) = CamelliaSubkeyL(20) ^ dw, CamelliaSubkeyL(20) = dw;
    dw = CamelliaSubkeyL(21) ^ CamelliaSubkeyR(21), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(21) = CamelliaSubkeyL(21) ^ dw, CamelliaSubkeyL(21) = dw;
    dw = CamelliaSubkeyL(22) ^ CamelliaSubkeyR(22), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(22) = CamelliaSubkeyL(22) ^ dw, CamelliaSubkeyL(22) = dw;
    dw = CamelliaSubkeyL(23) ^ CamelliaSubkeyR(23), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(23) = CamelliaSubkeyL(23) ^ dw, CamelliaSubkeyL(23) = dw;

    return;
}

static void camellia_setup256(const unsigned char *key, u32 *subkey)
{
    u32 kll,klr,krl,krr;           /* left half of key */
    u32 krll,krlr,krrl,krrr;       /* right half of key */
    u32 il, ir, t0, t1, w0, w1;    /* temporary variables */
    u32 kw4l, kw4r, dw, tl, tr;
    u32 subL[34];
    u32 subR[34];

    /**
     *  key = (kll || klr || krl || krr || krll || krlr || krrl || krrr)
     *  (|| is concatination)
     */

    kll  = GETU32(key     );
    klr  = GETU32(key +  4);
    krl  = GETU32(key +  8);
    krr  = GETU32(key + 12);
    krll = GETU32(key + 16);
    krlr = GETU32(key + 20);
    krrl = GETU32(key + 24);
    krrr = GETU32(key + 28);

    /* generate KL dependent subkeys */
    subl(0) = kll; subr(0) = klr;
    subl(1) = krl; subr(1) = krr;
    CAMELLIA_ROLDQo32(kll, klr, krl, krr, w0, w1, 45);
    subl(12) = kll; subr(12) = klr;
    subl(13) = krl; subr(13) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 15);
    subl(16) = kll; subr(16) = klr;
    subl(17) = krl; subr(17) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 17);
    subl(22) = kll; subr(22) = klr;
    subl(23) = krl; subr(23) = krr;
    CAMELLIA_ROLDQo32(kll, klr, krl, krr, w0, w1, 34);
    subl(30) = kll; subr(30) = klr;
    subl(31) = krl; subr(31) = krr;

    /* generate KR dependent subkeys */
    CAMELLIA_ROLDQ(krll, krlr, krrl, krrr, w0, w1, 15);
    subl(4) = krll; subr(4) = krlr;
    subl(5) = krrl; subr(5) = krrr;
    CAMELLIA_ROLDQ(krll, krlr, krrl, krrr, w0, w1, 15);
    subl(8) = krll; subr(8) = krlr;
    subl(9) = krrl; subr(9) = krrr;
    CAMELLIA_ROLDQ(krll, krlr, krrl, krrr, w0, w1, 30);
    subl(18) = krll; subr(18) = krlr;
    subl(19) = krrl; subr(19) = krrr;
    CAMELLIA_ROLDQo32(krll, krlr, krrl, krrr, w0, w1, 34);
    subl(26) = krll; subr(26) = krlr;
    subl(27) = krrl; subr(27) = krrr;
    CAMELLIA_ROLDQo32(krll, krlr, krrl, krrr, w0, w1, 34);

    /* generate KA */
    kll = subl(0) ^ krll; klr = subr(0) ^ krlr;
    krl = subl(1) ^ krrl; krr = subr(1) ^ krrr;
    CAMELLIA_F(kll, klr,
	       CAMELLIA_SIGMA1L, CAMELLIA_SIGMA1R,
	       w0, w1, il, ir, t0, t1);
    krl ^= w0; krr ^= w1;
    CAMELLIA_F(krl, krr,
	       CAMELLIA_SIGMA2L, CAMELLIA_SIGMA2R,
	       kll, klr, il, ir, t0, t1);
    kll ^= krll; klr ^= krlr;
    CAMELLIA_F(kll, klr,
	       CAMELLIA_SIGMA3L, CAMELLIA_SIGMA3R,
	       krl, krr, il, ir, t0, t1);
    krl ^= w0 ^ krrl; krr ^= w1 ^ krrr;
    CAMELLIA_F(krl, krr,
	       CAMELLIA_SIGMA4L, CAMELLIA_SIGMA4R,
	       w0, w1, il, ir, t0, t1);
    kll ^= w0; klr ^= w1;

    /* generate KB */
    krll ^= kll; krlr ^= klr;
    krrl ^= krl; krrr ^= krr;
    CAMELLIA_F(krll, krlr,
	       CAMELLIA_SIGMA5L, CAMELLIA_SIGMA5R,
	       w0, w1, il, ir, t0, t1);
    krrl ^= w0; krrr ^= w1;
    CAMELLIA_F(krrl, krrr,
	       CAMELLIA_SIGMA6L, CAMELLIA_SIGMA6R,
	       w0, w1, il, ir, t0, t1);
    krll ^= w0; krlr ^= w1;

    /* generate KA dependent subkeys */
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 15);
    subl(6) = kll; subr(6) = klr;
    subl(7) = krl; subr(7) = krr;
    CAMELLIA_ROLDQ(kll, klr, krl, krr, w0, w1, 30);
    subl(14) = kll; subr(14) = klr;
    subl(15) = krl; subr(15) = krr;
    subl(24) = klr; subr(24) = krl;
    subl(25) = krr; subr(25) = kll;
    CAMELLIA_ROLDQo32(kll, klr, krl, krr, w0, w1, 49);
    subl(28) = kll; subr(28) = klr;
    subl(29) = krl; subr(29) = krr;

    /* generate KB dependent subkeys */
    subl(2) = krll; subr(2) = krlr;
    subl(3) = krrl; subr(3) = krrr;
    CAMELLIA_ROLDQ(krll, krlr, krrl, krrr, w0, w1, 30);
    subl(10) = krll; subr(10) = krlr;
    subl(11) = krrl; subr(11) = krrr;
    CAMELLIA_ROLDQ(krll, krlr, krrl, krrr, w0, w1, 30);
    subl(20) = krll; subr(20) = krlr;
    subl(21) = krrl; subr(21) = krrr;
    CAMELLIA_ROLDQo32(krll, krlr, krrl, krrr, w0, w1, 51);
    subl(32) = krll; subr(32) = krlr;
    subl(33) = krrl; subr(33) = krrr;

    /* absorb kw2 to other subkeys */
    subl(3) ^= subl(1); subr(3) ^= subr(1);
    subl(5) ^= subl(1); subr(5) ^= subr(1);
    subl(7) ^= subl(1); subr(7) ^= subr(1);
    subl(1) ^= subr(1) & ~subr(9);
    dw = subl(1) & subl(9), subr(1) ^= CAMELLIA_RL1(dw);
    subl(11) ^= subl(1); subr(11) ^= subr(1);
    subl(13) ^= subl(1); subr(13) ^= subr(1);
    subl(15) ^= subl(1); subr(15) ^= subr(1);
    subl(1) ^= subr(1) & ~subr(17);
    dw = subl(1) & subl(17), subr(1) ^= CAMELLIA_RL1(dw);
    subl(19) ^= subl(1); subr(19) ^= subr(1);
    subl(21) ^= subl(1); subr(21) ^= subr(1);
    subl(23) ^= subl(1); subr(23) ^= subr(1);
    subl(1) ^= subr(1) & ~subr(25);
    dw = subl(1) & subl(25), subr(1) ^= CAMELLIA_RL1(dw);
    subl(27) ^= subl(1); subr(27) ^= subr(1);
    subl(29) ^= subl(1); subr(29) ^= subr(1);
    subl(31) ^= subl(1); subr(31) ^= subr(1);
    subl(32) ^= subl(1); subr(32) ^= subr(1);

    /* absorb kw4 to other subkeys */
    kw4l = subl(33); kw4r = subr(33);
    subl(30) ^= kw4l; subr(30) ^= kw4r;
    subl(28) ^= kw4l; subr(28) ^= kw4r;
    subl(26) ^= kw4l; subr(26) ^= kw4r;
    kw4l ^= kw4r & ~subr(24);
    dw = kw4l & subl(24), kw4r ^= CAMELLIA_RL1(dw);
    subl(22) ^= kw4l; subr(22) ^= kw4r;
    subl(20) ^= kw4l; subr(20) ^= kw4r;
    subl(18) ^= kw4l; subr(18) ^= kw4r;
    kw4l ^= kw4r & ~subr(16);
    dw = kw4l & subl(16), kw4r ^= CAMELLIA_RL1(dw);
    subl(14) ^= kw4l; subr(14) ^= kw4r;
    subl(12) ^= kw4l; subr(12) ^= kw4r;
    subl(10) ^= kw4l; subr(10) ^= kw4r;
    kw4l ^= kw4r & ~subr(8);
    dw = kw4l & subl(8), kw4r ^= CAMELLIA_RL1(dw);
    subl(6) ^= kw4l; subr(6) ^= kw4r;
    subl(4) ^= kw4l; subr(4) ^= kw4r;
    subl(2) ^= kw4l; subr(2) ^= kw4r;
    subl(0) ^= kw4l; subr(0) ^= kw4r;

    /* key XOR is end of F-function */
    CamelliaSubkeyL(0) = subl(0) ^ subl(2);
    CamelliaSubkeyR(0) = subr(0) ^ subr(2);
    CamelliaSubkeyL(2) = subl(3);
    CamelliaSubkeyR(2) = subr(3);
    CamelliaSubkeyL(3) = subl(2) ^ subl(4);
    CamelliaSubkeyR(3) = subr(2) ^ subr(4);
    CamelliaSubkeyL(4) = subl(3) ^ subl(5);
    CamelliaSubkeyR(4) = subr(3) ^ subr(5);
    CamelliaSubkeyL(5) = subl(4) ^ subl(6);
    CamelliaSubkeyR(5) = subr(4) ^ subr(6);
    CamelliaSubkeyL(6) = subl(5) ^ subl(7);
    CamelliaSubkeyR(6) = subr(5) ^ subr(7);
    tl = subl(10) ^ (subr(10) & ~subr(8));
    dw = tl & subl(8), tr = subr(10) ^ CAMELLIA_RL1(dw);
    CamelliaSubkeyL(7) = subl(6) ^ tl;
    CamelliaSubkeyR(7) = subr(6) ^ tr;
    CamelliaSubkeyL(8) = subl(8);
    CamelliaSubkeyR(8) = subr(8);
    CamelliaSubkeyL(9) = subl(9);
    CamelliaSubkeyR(9) = subr(9);
    tl = subl(7) ^ (subr(7) & ~subr(9));
    dw = tl & subl(9), tr = subr(7) ^ CAMELLIA_RL1(dw);
    CamelliaSubkeyL(10) = tl ^ subl(11);
    CamelliaSubkeyR(10) = tr ^ subr(11);
    CamelliaSubkeyL(11) = subl(10) ^ subl(12);
    CamelliaSubkeyR(11) = subr(10) ^ subr(12);
    CamelliaSubkeyL(12) = subl(11) ^ subl(13);
    CamelliaSubkeyR(12) = subr(11) ^ subr(13);
    CamelliaSubkeyL(13) = subl(12) ^ subl(14);
    CamelliaSubkeyR(13) = subr(12) ^ subr(14);
    CamelliaSubkeyL(14) = subl(13) ^ subl(15);
    CamelliaSubkeyR(14) = subr(13) ^ subr(15);
    tl = subl(18) ^ (subr(18) & ~subr(16));
    dw = tl & subl(16), tr = subr(18) ^ CAMELLIA_RL1(dw);
    CamelliaSubkeyL(15) = subl(14) ^ tl;
    CamelliaSubkeyR(15) = subr(14) ^ tr;
    CamelliaSubkeyL(16) = subl(16);
    CamelliaSubkeyR(16) = subr(16);
    CamelliaSubkeyL(17) = subl(17);
    CamelliaSubkeyR(17) = subr(17);
    tl = subl(15) ^ (subr(15) & ~subr(17));
    dw = tl & subl(17), tr = subr(15) ^ CAMELLIA_RL1(dw);
    CamelliaSubkeyL(18) = tl ^ subl(19);
    CamelliaSubkeyR(18) = tr ^ subr(19);
    CamelliaSubkeyL(19) = subl(18) ^ subl(20);
    CamelliaSubkeyR(19) = subr(18) ^ subr(20);
    CamelliaSubkeyL(20) = subl(19) ^ subl(21);
    CamelliaSubkeyR(20) = subr(19) ^ subr(21);
    CamelliaSubkeyL(21) = subl(20) ^ subl(22);
    CamelliaSubkeyR(21) = subr(20) ^ subr(22);
    CamelliaSubkeyL(22) = subl(21) ^ subl(23);
    CamelliaSubkeyR(22) = subr(21) ^ subr(23);
    tl = subl(26) ^ (subr(26) & ~subr(24));
    dw = tl & subl(24), tr = subr(26) ^ CAMELLIA_RL1(dw);
    CamelliaSubkeyL(23) = subl(22) ^ tl;
    CamelliaSubkeyR(23) = subr(22) ^ tr;
    CamelliaSubkeyL(24) = subl(24);
    CamelliaSubkeyR(24) = subr(24);
    CamelliaSubkeyL(25) = subl(25);
    CamelliaSubkeyR(25) = subr(25);
    tl = subl(23) ^ (subr(23) &  ~subr(25));
    dw = tl & subl(25), tr = subr(23) ^ CAMELLIA_RL1(dw);
    CamelliaSubkeyL(26) = tl ^ subl(27);
    CamelliaSubkeyR(26) = tr ^ subr(27);
    CamelliaSubkeyL(27) = subl(26) ^ subl(28);
    CamelliaSubkeyR(27) = subr(26) ^ subr(28);
    CamelliaSubkeyL(28) = subl(27) ^ subl(29);
    CamelliaSubkeyR(28) = subr(27) ^ subr(29);
    CamelliaSubkeyL(29) = subl(28) ^ subl(30);
    CamelliaSubkeyR(29) = subr(28) ^ subr(30);
    CamelliaSubkeyL(30) = subl(29) ^ subl(31);
    CamelliaSubkeyR(30) = subr(29) ^ subr(31);
    CamelliaSubkeyL(31) = subl(30);
    CamelliaSubkeyR(31) = subr(30);
    CamelliaSubkeyL(32) = subl(32) ^ subl(31);
    CamelliaSubkeyR(32) = subr(32) ^ subr(31);

    /* apply the inverse of the last half of P-function */
    dw = CamelliaSubkeyL(2) ^ CamelliaSubkeyR(2), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(2) = CamelliaSubkeyL(2) ^ dw, CamelliaSubkeyL(2) = dw;
    dw = CamelliaSubkeyL(3) ^ CamelliaSubkeyR(3), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(3) = CamelliaSubkeyL(3) ^ dw, CamelliaSubkeyL(3) = dw;
    dw = CamelliaSubkeyL(4) ^ CamelliaSubkeyR(4), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(4) = CamelliaSubkeyL(4) ^ dw, CamelliaSubkeyL(4) = dw;
    dw = CamelliaSubkeyL(5) ^ CamelliaSubkeyR(5), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(5) = CamelliaSubkeyL(5) ^ dw, CamelliaSubkeyL(5) = dw;
    dw = CamelliaSubkeyL(6) ^ CamelliaSubkeyR(6), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(6) = CamelliaSubkeyL(6) ^ dw, CamelliaSubkeyL(6) = dw;
    dw = CamelliaSubkeyL(7) ^ CamelliaSubkeyR(7), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(7) = CamelliaSubkeyL(7) ^ dw, CamelliaSubkeyL(7) = dw;
    dw = CamelliaSubkeyL(10) ^ CamelliaSubkeyR(10), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(10) = CamelliaSubkeyL(10) ^ dw, CamelliaSubkeyL(10) = dw;
    dw = CamelliaSubkeyL(11) ^ CamelliaSubkeyR(11), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(11) = CamelliaSubkeyL(11) ^ dw, CamelliaSubkeyL(11) = dw;
    dw = CamelliaSubkeyL(12) ^ CamelliaSubkeyR(12), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(12) = CamelliaSubkeyL(12) ^ dw, CamelliaSubkeyL(12) = dw;
    dw = CamelliaSubkeyL(13) ^ CamelliaSubkeyR(13), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(13) = CamelliaSubkeyL(13) ^ dw, CamelliaSubkeyL(13) = dw;
    dw = CamelliaSubkeyL(14) ^ CamelliaSubkeyR(14), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(14) = CamelliaSubkeyL(14) ^ dw, CamelliaSubkeyL(14) = dw;
    dw = CamelliaSubkeyL(15) ^ CamelliaSubkeyR(15), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(15) = CamelliaSubkeyL(15) ^ dw, CamelliaSubkeyL(15) = dw;
    dw = CamelliaSubkeyL(18) ^ CamelliaSubkeyR(18), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(18) = CamelliaSubkeyL(18) ^ dw, CamelliaSubkeyL(18) = dw;
    dw = CamelliaSubkeyL(19) ^ CamelliaSubkeyR(19), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(19) = CamelliaSubkeyL(19) ^ dw, CamelliaSubkeyL(19) = dw;
    dw = CamelliaSubkeyL(20) ^ CamelliaSubkeyR(20), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(20) = CamelliaSubkeyL(20) ^ dw, CamelliaSubkeyL(20) = dw;
    dw = CamelliaSubkeyL(21) ^ CamelliaSubkeyR(21), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(21) = CamelliaSubkeyL(21) ^ dw, CamelliaSubkeyL(21) = dw;
    dw = CamelliaSubkeyL(22) ^ CamelliaSubkeyR(22), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(22) = CamelliaSubkeyL(22) ^ dw, CamelliaSubkeyL(22) = dw;
    dw = CamelliaSubkeyL(23) ^ CamelliaSubkeyR(23), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(23) = CamelliaSubkeyL(23) ^ dw, CamelliaSubkeyL(23) = dw;
    dw = CamelliaSubkeyL(26) ^ CamelliaSubkeyR(26), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(26) = CamelliaSubkeyL(26) ^ dw, CamelliaSubkeyL(26) = dw;
    dw = CamelliaSubkeyL(27) ^ CamelliaSubkeyR(27), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(27) = CamelliaSubkeyL(27) ^ dw, CamelliaSubkeyL(27) = dw;
    dw = CamelliaSubkeyL(28) ^ CamelliaSubkeyR(28), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(28) = CamelliaSubkeyL(28) ^ dw, CamelliaSubkeyL(28) = dw;
    dw = CamelliaSubkeyL(29) ^ CamelliaSubkeyR(29), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(29) = CamelliaSubkeyL(29) ^ dw, CamelliaSubkeyL(29) = dw;
    dw = CamelliaSubkeyL(30) ^ CamelliaSubkeyR(30), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(30) = CamelliaSubkeyL(30) ^ dw, CamelliaSubkeyL(30) = dw;
    dw = CamelliaSubkeyL(31) ^ CamelliaSubkeyR(31), dw = CAMELLIA_RL8(dw);
    CamelliaSubkeyR(31) = CamelliaSubkeyL(31) ^ dw,CamelliaSubkeyL(31) = dw;

    return;
}

static void camellia_setup192(const unsigned char *key, u32 *subkey)
{
    unsigned char kk[32];
    u32 krll, krlr, krrl,krrr;

    memcpy(kk, key, 24);
    memcpy((unsigned char *)&krll, key+16,4);
    memcpy((unsigned char *)&krlr, key+20,4);
    krrl = ~krll;
    krrr = ~krlr;
    memcpy(kk+24, (unsigned char *)&krrl, 4);
    memcpy(kk+28, (unsigned char *)&krrr, 4);
    camellia_setup256(kk, subkey);
    return;
}


/**
 * Stuff related to camellia encryption/decryption
 *
 * "io" must be 4byte aligned and big-endian data.
 */
static void camellia_encrypt128(const u32 *subkey, u32 *io)
{
    u32 il, ir, t0, t1;

    /* pre whitening but absorb kw2*/
    io[0] ^= CamelliaSubkeyL(0);
    io[1] ^= CamelliaSubkeyR(0);
    /* main iteration */

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(2),CamelliaSubkeyR(2),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(3),CamelliaSubkeyR(3),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(4),CamelliaSubkeyR(4),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(5),CamelliaSubkeyR(5),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(6),CamelliaSubkeyR(6),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(7),CamelliaSubkeyR(7),
		     io[0],io[1],il,ir,t0,t1);

    CAMELLIA_FLS(io[0],io[1],io[2],io[3],
		 CamelliaSubkeyL(8),CamelliaSubkeyR(8),
		 CamelliaSubkeyL(9),CamelliaSubkeyR(9),
		 t0,t1,il,ir);

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(10),CamelliaSubkeyR(10),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(11),CamelliaSubkeyR(11),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(12),CamelliaSubkeyR(12),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(13),CamelliaSubkeyR(13),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(14),CamelliaSubkeyR(14),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(15),CamelliaSubkeyR(15),
		     io[0],io[1],il,ir,t0,t1);

    CAMELLIA_FLS(io[0],io[1],io[2],io[3],
		 CamelliaSubkeyL(16),CamelliaSubkeyR(16),
		 CamelliaSubkeyL(17),CamelliaSubkeyR(17),
		 t0,t1,il,ir);

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(18),CamelliaSubkeyR(18),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(19),CamelliaSubkeyR(19),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(20),CamelliaSubkeyR(20),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(21),CamelliaSubkeyR(21),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(22),CamelliaSubkeyR(22),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(23),CamelliaSubkeyR(23),
		     io[0],io[1],il,ir,t0,t1);

    /* post whitening but kw4 */
    io[2] ^= CamelliaSubkeyL(24);
    io[3] ^= CamelliaSubkeyR(24);

    t0 = io[0];
    t1 = io[1];
    io[0] = io[2];
    io[1] = io[3];
    io[2] = t0;
    io[3] = t1;
	
    return;
}

static void camellia_decrypt128(const u32 *subkey, u32 *io)
{
    u32 il,ir,t0,t1;               /* temporary valiables */

    /* pre whitening but absorb kw2*/
    io[0] ^= CamelliaSubkeyL(24);
    io[1] ^= CamelliaSubkeyR(24);

    /* main iteration */
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(23),CamelliaSubkeyR(23),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(22),CamelliaSubkeyR(22),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(21),CamelliaSubkeyR(21),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(20),CamelliaSubkeyR(20),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(19),CamelliaSubkeyR(19),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(18),CamelliaSubkeyR(18),
		     io[0],io[1],il,ir,t0,t1);

    CAMELLIA_FLS(io[0],io[1],io[2],io[3],
		 CamelliaSubkeyL(17),CamelliaSubkeyR(17),
		 CamelliaSubkeyL(16),CamelliaSubkeyR(16),
		 t0,t1,il,ir);

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(15),CamelliaSubkeyR(15),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(14),CamelliaSubkeyR(14),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(13),CamelliaSubkeyR(13),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(12),CamelliaSubkeyR(12),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(11),CamelliaSubkeyR(11),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(10),CamelliaSubkeyR(10),
		     io[0],io[1],il,ir,t0,t1);

    CAMELLIA_FLS(io[0],io[1],io[2],io[3],
		 CamelliaSubkeyL(9),CamelliaSubkeyR(9),
		 CamelliaSubkeyL(8),CamelliaSubkeyR(8),
		 t0,t1,il,ir);

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(7),CamelliaSubkeyR(7),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(6),CamelliaSubkeyR(6),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(5),CamelliaSubkeyR(5),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(4),CamelliaSubkeyR(4),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(3),CamelliaSubkeyR(3),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(2),CamelliaSubkeyR(2),
		     io[0],io[1],il,ir,t0,t1);

    /* post whitening but kw4 */
    io[2] ^= CamelliaSubkeyL(0);
    io[3] ^= CamelliaSubkeyR(0);

    t0 = io[0];
    t1 = io[1];
    io[0] = io[2];
    io[1] = io[3];
    io[2] = t0;
    io[3] = t1;

    return;
}

/**
 * stuff for 192 and 256bit encryption/decryption
 */
static void camellia_encrypt256(const u32 *subkey, u32 *io)
{
    u32 il,ir,t0,t1;           /* temporary valiables */

    /* pre whitening but absorb kw2*/
    io[0] ^= CamelliaSubkeyL(0);
    io[1] ^= CamelliaSubkeyR(0);

    /* main iteration */
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(2),CamelliaSubkeyR(2),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(3),CamelliaSubkeyR(3),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(4),CamelliaSubkeyR(4),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(5),CamelliaSubkeyR(5),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(6),CamelliaSubkeyR(6),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(7),CamelliaSubkeyR(7),
		     io[0],io[1],il,ir,t0,t1);

    CAMELLIA_FLS(io[0],io[1],io[2],io[3],
		 CamelliaSubkeyL(8),CamelliaSubkeyR(8),
		 CamelliaSubkeyL(9),CamelliaSubkeyR(9),
		 t0,t1,il,ir);

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(10),CamelliaSubkeyR(10),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(11),CamelliaSubkeyR(11),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(12),CamelliaSubkeyR(12),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(13),CamelliaSubkeyR(13),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(14),CamelliaSubkeyR(14),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(15),CamelliaSubkeyR(15),
		     io[0],io[1],il,ir,t0,t1);

    CAMELLIA_FLS(io[0],io[1],io[2],io[3],
		 CamelliaSubkeyL(16),CamelliaSubkeyR(16),
		 CamelliaSubkeyL(17),CamelliaSubkeyR(17),
		 t0,t1,il,ir);

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(18),CamelliaSubkeyR(18),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(19),CamelliaSubkeyR(19),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(20),CamelliaSubkeyR(20),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(21),CamelliaSubkeyR(21),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(22),CamelliaSubkeyR(22),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(23),CamelliaSubkeyR(23),
		     io[0],io[1],il,ir,t0,t1);

    CAMELLIA_FLS(io[0],io[1],io[2],io[3],
		 CamelliaSubkeyL(24),CamelliaSubkeyR(24),
		 CamelliaSubkeyL(25),CamelliaSubkeyR(25),
		 t0,t1,il,ir);

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(26),CamelliaSubkeyR(26),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(27),CamelliaSubkeyR(27),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(28),CamelliaSubkeyR(28),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(29),CamelliaSubkeyR(29),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(30),CamelliaSubkeyR(30),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(31),CamelliaSubkeyR(31),
		     io[0],io[1],il,ir,t0,t1);

    /* post whitening but kw4 */
    io[2] ^= CamelliaSubkeyL(32);
    io[3] ^= CamelliaSubkeyR(32);

    t0 = io[0];
    t1 = io[1];
    io[0] = io[2];
    io[1] = io[3];
    io[2] = t0;
    io[3] = t1;

    return;
}

static void camellia_decrypt256(const u32 *subkey, u32 *io)
{
    u32 il,ir,t0,t1;           /* temporary valiables */

    /* pre whitening but absorb kw2*/
    io[0] ^= CamelliaSubkeyL(32);
    io[1] ^= CamelliaSubkeyR(32);
	
    /* main iteration */
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(31),CamelliaSubkeyR(31),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(30),CamelliaSubkeyR(30),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(29),CamelliaSubkeyR(29),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(28),CamelliaSubkeyR(28),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(27),CamelliaSubkeyR(27),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(26),CamelliaSubkeyR(26),
		     io[0],io[1],il,ir,t0,t1);

    CAMELLIA_FLS(io[0],io[1],io[2],io[3],
		 CamelliaSubkeyL(25),CamelliaSubkeyR(25),
		 CamelliaSubkeyL(24),CamelliaSubkeyR(24),
		 t0,t1,il,ir);

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(23),CamelliaSubkeyR(23),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(22),CamelliaSubkeyR(22),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(21),CamelliaSubkeyR(21),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(20),CamelliaSubkeyR(20),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(19),CamelliaSubkeyR(19),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(18),CamelliaSubkeyR(18),
		     io[0],io[1],il,ir,t0,t1);

    CAMELLIA_FLS(io[0],io[1],io[2],io[3],
		 CamelliaSubkeyL(17),CamelliaSubkeyR(17),
		 CamelliaSubkeyL(16),CamelliaSubkeyR(16),
		 t0,t1,il,ir);

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(15),CamelliaSubkeyR(15),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(14),CamelliaSubkeyR(14),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(13),CamelliaSubkeyR(13),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(12),CamelliaSubkeyR(12),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(11),CamelliaSubkeyR(11),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(10),CamelliaSubkeyR(10),
		     io[0],io[1],il,ir,t0,t1);

    CAMELLIA_FLS(io[0],io[1],io[2],io[3],
		 CamelliaSubkeyL(9),CamelliaSubkeyR(9),
		 CamelliaSubkeyL(8),CamelliaSubkeyR(8),
		 t0,t1,il,ir);

    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(7),CamelliaSubkeyR(7),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(6),CamelliaSubkeyR(6),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(5),CamelliaSubkeyR(5),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(4),CamelliaSubkeyR(4),
		     io[0],io[1],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[0],io[1],
		     CamelliaSubkeyL(3),CamelliaSubkeyR(3),
		     io[2],io[3],il,ir,t0,t1);
    CAMELLIA_ROUNDSM(io[2],io[3],
		     CamelliaSubkeyL(2),CamelliaSubkeyR(2),
		     io[0],io[1],il,ir,t0,t1);

    /* post whitening but kw4 */
    io[2] ^= CamelliaSubkeyL(0);
    io[3] ^= CamelliaSubkeyR(0);

    t0 = io[0];
    t1 = io[1];
    io[0] = io[2];
    io[1] = io[3];
    io[2] = t0;
    io[3] = t1;

    return;
}

/***
 *
 * API for compatibility
 */

void Camellia_Ekeygen(const int keyBitLength,
		      const unsigned char *rawKey,
		      KEY_TABLE_TYPE keyTable)
{
    switch(keyBitLength) {
    case 128:
	camellia_setup128(rawKey, keyTable);
	break;
    case 192:
	camellia_setup192(rawKey, keyTable);
	break;
    case 256:
	camellia_setup256(rawKey, keyTable);
	break;
    default:
	break;
    }
}


void Camellia_EncryptBlock(const int keyBitLength,
			   const unsigned char *plaintext,
			   const KEY_TABLE_TYPE keyTable,
			   unsigned char *ciphertext)
{
    u32 tmp[4];

    tmp[0] = GETU32(plaintext);
    tmp[1] = GETU32(plaintext + 4);
    tmp[2] = GETU32(plaintext + 8);
    tmp[3] = GETU32(plaintext + 12);

    switch (keyBitLength) {
    case 128:
	camellia_encrypt128(keyTable, tmp);
	break;
    case 192:
	/* fall through */
    case 256:
	camellia_encrypt256(keyTable, tmp);
	break;
    default:
	break;
    }

    PUTU32(ciphertext, tmp[0]);
    PUTU32(ciphertext + 4, tmp[1]);
    PUTU32(ciphertext + 8, tmp[2]);
    PUTU32(ciphertext + 12, tmp[3]);
}

void Camellia_DecryptBlock(const int keyBitLength,
			   const unsigned char *ciphertext,
			   const KEY_TABLE_TYPE keyTable,
			   unsigned char *plaintext)
{
    u32 tmp[4];

    tmp[0] = GETU32(ciphertext);
    tmp[1] = GETU32(ciphertext + 4);
    tmp[2] = GETU32(ciphertext + 8);
    tmp[3] = GETU32(ciphertext + 12);

    switch (keyBitLength) {
    case 128:
	camellia_decrypt128(keyTable, tmp);
	break;
    case 192:
	/* fall through */
    case 256:
	camellia_decrypt256(keyTable, tmp);
	break;
    default:
	break;
    }
    PUTU32(plaintext, tmp[0]);
    PUTU32(plaintext + 4, tmp[1]);
    PUTU32(plaintext + 8, tmp[2]);
    PUTU32(plaintext + 12, tmp[3]);
}





		// Camellia test routine:
/*	 Here are test data for Camellia in hexadecimal form.

   128-bit key
       Key       : 01 23 45 67 89 ab cd ef fe dc ba 98 76 54 32 10
       Plaintext : 01 23 45 67 89 ab cd ef fe dc ba 98 76 54 32 10
       Ciphertext: 67 67 31 38 54 96 69 73 08 57 06 56 48 ea be 43

   192-bit key
       Key       : 01 23 45 67 89 ab cd ef fe dc ba 98 76 54 32 10
                 : 00 11 22 33 44 55 66 77
       Plaintext : 01 23 45 67 89 ab cd ef fe dc ba 98 76 54 32 10
       Ciphertext: b4 99 34 01 b3 e9 96 f8 4e e5 ce e7 d7 9b 09 b9

   256-bit key
       Key       : 01 23 45 67 89 ab cd ef fe dc ba 98 76 54 32 10
                 : 00 11 22 33 44 55 66 77 88 99 aa bb cc dd ee ff
       Plaintext : 01 23 45 67 89 ab cd ef fe dc ba 98 76 54 32 10
       Ciphertext: 9a cc 23 7d ff 16 d7 6c 20 ef 7c 91 9e 3a 75 09
*/

		// those are the standard test vectors
	static unsigned char camellia_KEYs[3][33] =
	{
			"\x01\x23\x45\x67\x89\xab\xcd\xef\xfe\xdc\xba\x98\x76\x54\x32\x10",

			"\x01\x23\x45\x67\x89\xab\xcd\xef\xfe\xdc\xba\x98\x76\x54\x32\x10" \
					"\x00\x11\x22\x33\x44\x55\x66\x77",

			"\x01\x23\x45\x67\x89\xab\xcd\xef\xfe\xdc\xba\x98\x76\x54\x32\x10" \
					"\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xAA\xBB\xCC\xDD\xEE\xFF"
	};

	static unsigned char camellia_PT[17] =
			"\x01\x23\x45\x67\x89\xab\xcd\xef\xfe\xdc\xba\x98\x76\x54\x32\x10";

	static unsigned char camellia_CTs[3][17] =
	{
			"\x67\x67\x31\x38\x54\x96\x69\x73\x08\x57\x06\x56\x48\xea\xbe\x43",
			"\xb4\x99\x34\x01\xb3\xe9\x96\xf8\x4e\xe5\xce\xe7\xd7\x9b\x09\xb9",
			"\x9a\xcc\x23\x7d\xff\x16\xd7\x6c\x20\xef\x7c\x91\x9e\x3a\x75\x09"
	};



		// returns true for pass
	bool test_camellia(void) {
			bool result=true; // false=fail, true=pass
			int i;
			KEY_TABLE_TYPE keyTable;
			unsigned char data[16];
			unsigned char ciphertext[16];
			
			for( i = 0; i < 3; i++ )
			{
				memcpy( data, camellia_PT, 16 );
				
				Camellia_Ekeygen( 128 + i * 64, camellia_KEYs[i], keyTable );
				Camellia_EncryptBlock( 128 + i * 64, data, keyTable, ciphertext );
				
				if (memcmp( ciphertext, camellia_CTs[i], 16 ) )
				{
					result=0;
					return(result);
				}
			}
			
			for( i = 0; i < 3; i++ )
			{
				memcpy( ciphertext, camellia_CTs[i], 16 );
				
				Camellia_Ekeygen( 128 + i * 64, camellia_KEYs[i], keyTable );
				Camellia_DecryptBlock( 128 + i * 64, ciphertext, keyTable, data );
				
				if (memcmp( data, camellia_PT, 16 ) )
				{
					result=false;
					return(result);
				}
			}
			
			return(result);
	}











//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------

/**
 * The Whirlpool hashing function.
 *
 * <P>
 * <b>References</b>
 *
 * <P>
 * The Whirlpool algorithm was developed by
 * <a href="mailto:pbarreto@scopus.com.br">Paulo S. L. M. Barreto</a> and
 * <a href="mailto:vincent.rijmen@cryptomathic.com">Vincent Rijmen</a>.
 *
 * See
 *      P.S.L.M. Barreto, V. Rijmen,
 *      ``The Whirlpool hashing function,''
 *      NESSIE submission, 2000 (tweaked version, 2001),
 *      <https://www.cosic.esat.kuleuven.ac.be/nessie/workshop/submissions/whirlpool.zip>
 * 
 * @author  Paulo S.L.M. Barreto
 * @author  Vincent Rijmen.
 *
 * @version 3.0 (2003.03.12)
 *
 * =============================================================================
 *
 * Differences from version 2.1:
 *
 * - Suboptimal diffusion matrix replaced by cir(1, 1, 4, 1, 8, 5, 2, 9).
 *
 * =============================================================================
 *
 * Differences from version 2.0:
 *
 * - Generation of ISO/IEC 10118-3 test vectors.
 * - Bug fix: nonzero carry was ignored when tallying the data length
 *      (this bug apparently only manifested itself when feeding data
 *      in pieces rather than in a single chunk at once).
 * - Support for MS Visual C++ 64-bit integer arithmetic.
 *
 * Differences from version 1.0:
 *
 * - Original S-box replaced by the tweaked, hardware-efficient version.
 *
 * =============================================================================
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>
//#include <time.h>


/* #define TRACE_INTERMEDIATE_VALUES */

/*
 * The number of rounds of the internal dedicated block cipher.
 */
#define R 10

/*
 * Though Whirlpool is endianness-neutral, the encryption tables are listed
 * in BIG-ENDIAN format, which is adopted throughout this implementation
 * (but little-endian notation would be equally suitable if consistently
 * employed).
 */

static const u64 C0[256] = {
    LL(0x18186018c07830d8), LL(0x23238c2305af4626), LL(0xc6c63fc67ef991b8), LL(0xe8e887e8136fcdfb),
    LL(0x878726874ca113cb), LL(0xb8b8dab8a9626d11), LL(0x0101040108050209), LL(0x4f4f214f426e9e0d),
    LL(0x3636d836adee6c9b), LL(0xa6a6a2a6590451ff), LL(0xd2d26fd2debdb90c), LL(0xf5f5f3f5fb06f70e),
    LL(0x7979f979ef80f296), LL(0x6f6fa16f5fcede30), LL(0x91917e91fcef3f6d), LL(0x52525552aa07a4f8),
    LL(0x60609d6027fdc047), LL(0xbcbccabc89766535), LL(0x9b9b569baccd2b37), LL(0x8e8e028e048c018a),
    LL(0xa3a3b6a371155bd2), LL(0x0c0c300c603c186c), LL(0x7b7bf17bff8af684), LL(0x3535d435b5e16a80),
    LL(0x1d1d741de8693af5), LL(0xe0e0a7e05347ddb3), LL(0xd7d77bd7f6acb321), LL(0xc2c22fc25eed999c),
    LL(0x2e2eb82e6d965c43), LL(0x4b4b314b627a9629), LL(0xfefedffea321e15d), LL(0x575741578216aed5),
    LL(0x15155415a8412abd), LL(0x7777c1779fb6eee8), LL(0x3737dc37a5eb6e92), LL(0xe5e5b3e57b56d79e),
    LL(0x9f9f469f8cd92313), LL(0xf0f0e7f0d317fd23), LL(0x4a4a354a6a7f9420), LL(0xdada4fda9e95a944),
    LL(0x58587d58fa25b0a2), LL(0xc9c903c906ca8fcf), LL(0x2929a429558d527c), LL(0x0a0a280a5022145a),
    LL(0xb1b1feb1e14f7f50), LL(0xa0a0baa0691a5dc9), LL(0x6b6bb16b7fdad614), LL(0x85852e855cab17d9),
    LL(0xbdbdcebd8173673c), LL(0x5d5d695dd234ba8f), LL(0x1010401080502090), LL(0xf4f4f7f4f303f507),
    LL(0xcbcb0bcb16c08bdd), LL(0x3e3ef83eedc67cd3), LL(0x0505140528110a2d), LL(0x676781671fe6ce78),
    LL(0xe4e4b7e47353d597), LL(0x27279c2725bb4e02), LL(0x4141194132588273), LL(0x8b8b168b2c9d0ba7),
    LL(0xa7a7a6a7510153f6), LL(0x7d7de97dcf94fab2), LL(0x95956e95dcfb3749), LL(0xd8d847d88e9fad56),
    LL(0xfbfbcbfb8b30eb70), LL(0xeeee9fee2371c1cd), LL(0x7c7ced7cc791f8bb), LL(0x6666856617e3cc71),
    LL(0xdddd53dda68ea77b), LL(0x17175c17b84b2eaf), LL(0x4747014702468e45), LL(0x9e9e429e84dc211a),
    LL(0xcaca0fca1ec589d4), LL(0x2d2db42d75995a58), LL(0xbfbfc6bf9179632e), LL(0x07071c07381b0e3f),
    LL(0xadad8ead012347ac), LL(0x5a5a755aea2fb4b0), LL(0x838336836cb51bef), LL(0x3333cc3385ff66b6),
    LL(0x636391633ff2c65c), LL(0x02020802100a0412), LL(0xaaaa92aa39384993), LL(0x7171d971afa8e2de),
    LL(0xc8c807c80ecf8dc6), LL(0x19196419c87d32d1), LL(0x494939497270923b), LL(0xd9d943d9869aaf5f),
    LL(0xf2f2eff2c31df931), LL(0xe3e3abe34b48dba8), LL(0x5b5b715be22ab6b9), LL(0x88881a8834920dbc),
    LL(0x9a9a529aa4c8293e), LL(0x262698262dbe4c0b), LL(0x3232c8328dfa64bf), LL(0xb0b0fab0e94a7d59),
    LL(0xe9e983e91b6acff2), LL(0x0f0f3c0f78331e77), LL(0xd5d573d5e6a6b733), LL(0x80803a8074ba1df4),
    LL(0xbebec2be997c6127), LL(0xcdcd13cd26de87eb), LL(0x3434d034bde46889), LL(0x48483d487a759032),
    LL(0xffffdbffab24e354), LL(0x7a7af57af78ff48d), LL(0x90907a90f4ea3d64), LL(0x5f5f615fc23ebe9d),
    LL(0x202080201da0403d), LL(0x6868bd6867d5d00f), LL(0x1a1a681ad07234ca), LL(0xaeae82ae192c41b7),
    LL(0xb4b4eab4c95e757d), LL(0x54544d549a19a8ce), LL(0x93937693ece53b7f), LL(0x222288220daa442f),
    LL(0x64648d6407e9c863), LL(0xf1f1e3f1db12ff2a), LL(0x7373d173bfa2e6cc), LL(0x12124812905a2482),
    LL(0x40401d403a5d807a), LL(0x0808200840281048), LL(0xc3c32bc356e89b95), LL(0xecec97ec337bc5df),
    LL(0xdbdb4bdb9690ab4d), LL(0xa1a1bea1611f5fc0), LL(0x8d8d0e8d1c830791), LL(0x3d3df43df5c97ac8),
    LL(0x97976697ccf1335b), LL(0x0000000000000000), LL(0xcfcf1bcf36d483f9), LL(0x2b2bac2b4587566e),
    LL(0x7676c57697b3ece1), LL(0x8282328264b019e6), LL(0xd6d67fd6fea9b128), LL(0x1b1b6c1bd87736c3),
    LL(0xb5b5eeb5c15b7774), LL(0xafaf86af112943be), LL(0x6a6ab56a77dfd41d), LL(0x50505d50ba0da0ea),
    LL(0x45450945124c8a57), LL(0xf3f3ebf3cb18fb38), LL(0x3030c0309df060ad), LL(0xefef9bef2b74c3c4),
    LL(0x3f3ffc3fe5c37eda), LL(0x55554955921caac7), LL(0xa2a2b2a2791059db), LL(0xeaea8fea0365c9e9),
    LL(0x656589650fecca6a), LL(0xbabad2bab9686903), LL(0x2f2fbc2f65935e4a), LL(0xc0c027c04ee79d8e),
    LL(0xdede5fdebe81a160), LL(0x1c1c701ce06c38fc), LL(0xfdfdd3fdbb2ee746), LL(0x4d4d294d52649a1f),
    LL(0x92927292e4e03976), LL(0x7575c9758fbceafa), LL(0x06061806301e0c36), LL(0x8a8a128a249809ae),
    LL(0xb2b2f2b2f940794b), LL(0xe6e6bfe66359d185), LL(0x0e0e380e70361c7e), LL(0x1f1f7c1ff8633ee7),
    LL(0x6262956237f7c455), LL(0xd4d477d4eea3b53a), LL(0xa8a89aa829324d81), LL(0x96966296c4f43152),
    LL(0xf9f9c3f99b3aef62), LL(0xc5c533c566f697a3), LL(0x2525942535b14a10), LL(0x59597959f220b2ab),
    LL(0x84842a8454ae15d0), LL(0x7272d572b7a7e4c5), LL(0x3939e439d5dd72ec), LL(0x4c4c2d4c5a619816),
    LL(0x5e5e655eca3bbc94), LL(0x7878fd78e785f09f), LL(0x3838e038ddd870e5), LL(0x8c8c0a8c14860598),
    LL(0xd1d163d1c6b2bf17), LL(0xa5a5aea5410b57e4), LL(0xe2e2afe2434dd9a1), LL(0x616199612ff8c24e),
    LL(0xb3b3f6b3f1457b42), LL(0x2121842115a54234), LL(0x9c9c4a9c94d62508), LL(0x1e1e781ef0663cee),
    LL(0x4343114322528661), LL(0xc7c73bc776fc93b1), LL(0xfcfcd7fcb32be54f), LL(0x0404100420140824),
    LL(0x51515951b208a2e3), LL(0x99995e99bcc72f25), LL(0x6d6da96d4fc4da22), LL(0x0d0d340d68391a65),
    LL(0xfafacffa8335e979), LL(0xdfdf5bdfb684a369), LL(0x7e7ee57ed79bfca9), LL(0x242490243db44819),
    LL(0x3b3bec3bc5d776fe), LL(0xabab96ab313d4b9a), LL(0xcece1fce3ed181f0), LL(0x1111441188552299),
    LL(0x8f8f068f0c890383), LL(0x4e4e254e4a6b9c04), LL(0xb7b7e6b7d1517366), LL(0xebeb8beb0b60cbe0),
    LL(0x3c3cf03cfdcc78c1), LL(0x81813e817cbf1ffd), LL(0x94946a94d4fe3540), LL(0xf7f7fbf7eb0cf31c),
    LL(0xb9b9deb9a1676f18), LL(0x13134c13985f268b), LL(0x2c2cb02c7d9c5851), LL(0xd3d36bd3d6b8bb05),
    LL(0xe7e7bbe76b5cd38c), LL(0x6e6ea56e57cbdc39), LL(0xc4c437c46ef395aa), LL(0x03030c03180f061b),
    LL(0x565645568a13acdc), LL(0x44440d441a49885e), LL(0x7f7fe17fdf9efea0), LL(0xa9a99ea921374f88),
    LL(0x2a2aa82a4d825467), LL(0xbbbbd6bbb16d6b0a), LL(0xc1c123c146e29f87), LL(0x53535153a202a6f1),
    LL(0xdcdc57dcae8ba572), LL(0x0b0b2c0b58271653), LL(0x9d9d4e9d9cd32701), LL(0x6c6cad6c47c1d82b),
    LL(0x3131c43195f562a4), LL(0x7474cd7487b9e8f3), LL(0xf6f6fff6e309f115), LL(0x464605460a438c4c),
    LL(0xacac8aac092645a5), LL(0x89891e893c970fb5), LL(0x14145014a04428b4), LL(0xe1e1a3e15b42dfba),
    LL(0x16165816b04e2ca6), LL(0x3a3ae83acdd274f7), LL(0x6969b9696fd0d206), LL(0x09092409482d1241),
    LL(0x7070dd70a7ade0d7), LL(0xb6b6e2b6d954716f), LL(0xd0d067d0ceb7bd1e), LL(0xeded93ed3b7ec7d6),
    LL(0xcccc17cc2edb85e2), LL(0x424215422a578468), LL(0x98985a98b4c22d2c), LL(0xa4a4aaa4490e55ed),
    LL(0x2828a0285d885075), LL(0x5c5c6d5cda31b886), LL(0xf8f8c7f8933fed6b), LL(0x8686228644a411c2),
};

static const u64 C1[256] = {
    LL(0xd818186018c07830), LL(0x2623238c2305af46), LL(0xb8c6c63fc67ef991), LL(0xfbe8e887e8136fcd),
    LL(0xcb878726874ca113), LL(0x11b8b8dab8a9626d), LL(0x0901010401080502), LL(0x0d4f4f214f426e9e),
    LL(0x9b3636d836adee6c), LL(0xffa6a6a2a6590451), LL(0x0cd2d26fd2debdb9), LL(0x0ef5f5f3f5fb06f7),
    LL(0x967979f979ef80f2), LL(0x306f6fa16f5fcede), LL(0x6d91917e91fcef3f), LL(0xf852525552aa07a4),
    LL(0x4760609d6027fdc0), LL(0x35bcbccabc897665), LL(0x379b9b569baccd2b), LL(0x8a8e8e028e048c01),
    LL(0xd2a3a3b6a371155b), LL(0x6c0c0c300c603c18), LL(0x847b7bf17bff8af6), LL(0x803535d435b5e16a),
    LL(0xf51d1d741de8693a), LL(0xb3e0e0a7e05347dd), LL(0x21d7d77bd7f6acb3), LL(0x9cc2c22fc25eed99),
    LL(0x432e2eb82e6d965c), LL(0x294b4b314b627a96), LL(0x5dfefedffea321e1), LL(0xd5575741578216ae),
    LL(0xbd15155415a8412a), LL(0xe87777c1779fb6ee), LL(0x923737dc37a5eb6e), LL(0x9ee5e5b3e57b56d7),
    LL(0x139f9f469f8cd923), LL(0x23f0f0e7f0d317fd), LL(0x204a4a354a6a7f94), LL(0x44dada4fda9e95a9),
    LL(0xa258587d58fa25b0), LL(0xcfc9c903c906ca8f), LL(0x7c2929a429558d52), LL(0x5a0a0a280a502214),
    LL(0x50b1b1feb1e14f7f), LL(0xc9a0a0baa0691a5d), LL(0x146b6bb16b7fdad6), LL(0xd985852e855cab17),
    LL(0x3cbdbdcebd817367), LL(0x8f5d5d695dd234ba), LL(0x9010104010805020), LL(0x07f4f4f7f4f303f5),
    LL(0xddcbcb0bcb16c08b), LL(0xd33e3ef83eedc67c), LL(0x2d0505140528110a), LL(0x78676781671fe6ce),
    LL(0x97e4e4b7e47353d5), LL(0x0227279c2725bb4e), LL(0x7341411941325882), LL(0xa78b8b168b2c9d0b),
    LL(0xf6a7a7a6a7510153), LL(0xb27d7de97dcf94fa), LL(0x4995956e95dcfb37), LL(0x56d8d847d88e9fad),
    LL(0x70fbfbcbfb8b30eb), LL(0xcdeeee9fee2371c1), LL(0xbb7c7ced7cc791f8), LL(0x716666856617e3cc),
    LL(0x7bdddd53dda68ea7), LL(0xaf17175c17b84b2e), LL(0x454747014702468e), LL(0x1a9e9e429e84dc21),
    LL(0xd4caca0fca1ec589), LL(0x582d2db42d75995a), LL(0x2ebfbfc6bf917963), LL(0x3f07071c07381b0e),
    LL(0xacadad8ead012347), LL(0xb05a5a755aea2fb4), LL(0xef838336836cb51b), LL(0xb63333cc3385ff66),
    LL(0x5c636391633ff2c6), LL(0x1202020802100a04), LL(0x93aaaa92aa393849), LL(0xde7171d971afa8e2),
    LL(0xc6c8c807c80ecf8d), LL(0xd119196419c87d32), LL(0x3b49493949727092), LL(0x5fd9d943d9869aaf),
    LL(0x31f2f2eff2c31df9), LL(0xa8e3e3abe34b48db), LL(0xb95b5b715be22ab6), LL(0xbc88881a8834920d),
    LL(0x3e9a9a529aa4c829), LL(0x0b262698262dbe4c), LL(0xbf3232c8328dfa64), LL(0x59b0b0fab0e94a7d),
    LL(0xf2e9e983e91b6acf), LL(0x770f0f3c0f78331e), LL(0x33d5d573d5e6a6b7), LL(0xf480803a8074ba1d),
    LL(0x27bebec2be997c61), LL(0xebcdcd13cd26de87), LL(0x893434d034bde468), LL(0x3248483d487a7590),
    LL(0x54ffffdbffab24e3), LL(0x8d7a7af57af78ff4), LL(0x6490907a90f4ea3d), LL(0x9d5f5f615fc23ebe),
    LL(0x3d202080201da040), LL(0x0f6868bd6867d5d0), LL(0xca1a1a681ad07234), LL(0xb7aeae82ae192c41),
    LL(0x7db4b4eab4c95e75), LL(0xce54544d549a19a8), LL(0x7f93937693ece53b), LL(0x2f222288220daa44),
    LL(0x6364648d6407e9c8), LL(0x2af1f1e3f1db12ff), LL(0xcc7373d173bfa2e6), LL(0x8212124812905a24),
    LL(0x7a40401d403a5d80), LL(0x4808082008402810), LL(0x95c3c32bc356e89b), LL(0xdfecec97ec337bc5),
    LL(0x4ddbdb4bdb9690ab), LL(0xc0a1a1bea1611f5f), LL(0x918d8d0e8d1c8307), LL(0xc83d3df43df5c97a),
    LL(0x5b97976697ccf133), LL(0x0000000000000000), LL(0xf9cfcf1bcf36d483), LL(0x6e2b2bac2b458756),
    LL(0xe17676c57697b3ec), LL(0xe68282328264b019), LL(0x28d6d67fd6fea9b1), LL(0xc31b1b6c1bd87736),
    LL(0x74b5b5eeb5c15b77), LL(0xbeafaf86af112943), LL(0x1d6a6ab56a77dfd4), LL(0xea50505d50ba0da0),
    LL(0x5745450945124c8a), LL(0x38f3f3ebf3cb18fb), LL(0xad3030c0309df060), LL(0xc4efef9bef2b74c3),
    LL(0xda3f3ffc3fe5c37e), LL(0xc755554955921caa), LL(0xdba2a2b2a2791059), LL(0xe9eaea8fea0365c9),
    LL(0x6a656589650fecca), LL(0x03babad2bab96869), LL(0x4a2f2fbc2f65935e), LL(0x8ec0c027c04ee79d),
    LL(0x60dede5fdebe81a1), LL(0xfc1c1c701ce06c38), LL(0x46fdfdd3fdbb2ee7), LL(0x1f4d4d294d52649a),
    LL(0x7692927292e4e039), LL(0xfa7575c9758fbcea), LL(0x3606061806301e0c), LL(0xae8a8a128a249809),
    LL(0x4bb2b2f2b2f94079), LL(0x85e6e6bfe66359d1), LL(0x7e0e0e380e70361c), LL(0xe71f1f7c1ff8633e),
    LL(0x556262956237f7c4), LL(0x3ad4d477d4eea3b5), LL(0x81a8a89aa829324d), LL(0x5296966296c4f431),
    LL(0x62f9f9c3f99b3aef), LL(0xa3c5c533c566f697), LL(0x102525942535b14a), LL(0xab59597959f220b2),
    LL(0xd084842a8454ae15), LL(0xc57272d572b7a7e4), LL(0xec3939e439d5dd72), LL(0x164c4c2d4c5a6198),
    LL(0x945e5e655eca3bbc), LL(0x9f7878fd78e785f0), LL(0xe53838e038ddd870), LL(0x988c8c0a8c148605),
    LL(0x17d1d163d1c6b2bf), LL(0xe4a5a5aea5410b57), LL(0xa1e2e2afe2434dd9), LL(0x4e616199612ff8c2),
    LL(0x42b3b3f6b3f1457b), LL(0x342121842115a542), LL(0x089c9c4a9c94d625), LL(0xee1e1e781ef0663c),
    LL(0x6143431143225286), LL(0xb1c7c73bc776fc93), LL(0x4ffcfcd7fcb32be5), LL(0x2404041004201408),
    LL(0xe351515951b208a2), LL(0x2599995e99bcc72f), LL(0x226d6da96d4fc4da), LL(0x650d0d340d68391a),
    LL(0x79fafacffa8335e9), LL(0x69dfdf5bdfb684a3), LL(0xa97e7ee57ed79bfc), LL(0x19242490243db448),
    LL(0xfe3b3bec3bc5d776), LL(0x9aabab96ab313d4b), LL(0xf0cece1fce3ed181), LL(0x9911114411885522),
    LL(0x838f8f068f0c8903), LL(0x044e4e254e4a6b9c), LL(0x66b7b7e6b7d15173), LL(0xe0ebeb8beb0b60cb),
    LL(0xc13c3cf03cfdcc78), LL(0xfd81813e817cbf1f), LL(0x4094946a94d4fe35), LL(0x1cf7f7fbf7eb0cf3),
    LL(0x18b9b9deb9a1676f), LL(0x8b13134c13985f26), LL(0x512c2cb02c7d9c58), LL(0x05d3d36bd3d6b8bb),
    LL(0x8ce7e7bbe76b5cd3), LL(0x396e6ea56e57cbdc), LL(0xaac4c437c46ef395), LL(0x1b03030c03180f06),
    LL(0xdc565645568a13ac), LL(0x5e44440d441a4988), LL(0xa07f7fe17fdf9efe), LL(0x88a9a99ea921374f),
    LL(0x672a2aa82a4d8254), LL(0x0abbbbd6bbb16d6b), LL(0x87c1c123c146e29f), LL(0xf153535153a202a6),
    LL(0x72dcdc57dcae8ba5), LL(0x530b0b2c0b582716), LL(0x019d9d4e9d9cd327), LL(0x2b6c6cad6c47c1d8),
    LL(0xa43131c43195f562), LL(0xf37474cd7487b9e8), LL(0x15f6f6fff6e309f1), LL(0x4c464605460a438c),
    LL(0xa5acac8aac092645), LL(0xb589891e893c970f), LL(0xb414145014a04428), LL(0xbae1e1a3e15b42df),
    LL(0xa616165816b04e2c), LL(0xf73a3ae83acdd274), LL(0x066969b9696fd0d2), LL(0x4109092409482d12),
    LL(0xd77070dd70a7ade0), LL(0x6fb6b6e2b6d95471), LL(0x1ed0d067d0ceb7bd), LL(0xd6eded93ed3b7ec7),
    LL(0xe2cccc17cc2edb85), LL(0x68424215422a5784), LL(0x2c98985a98b4c22d), LL(0xeda4a4aaa4490e55),
    LL(0x752828a0285d8850), LL(0x865c5c6d5cda31b8), LL(0x6bf8f8c7f8933fed), LL(0xc28686228644a411),
};

static const u64 C2[256] = {
    LL(0x30d818186018c078), LL(0x462623238c2305af), LL(0x91b8c6c63fc67ef9), LL(0xcdfbe8e887e8136f),
    LL(0x13cb878726874ca1), LL(0x6d11b8b8dab8a962), LL(0x0209010104010805), LL(0x9e0d4f4f214f426e),
    LL(0x6c9b3636d836adee), LL(0x51ffa6a6a2a65904), LL(0xb90cd2d26fd2debd), LL(0xf70ef5f5f3f5fb06),
    LL(0xf2967979f979ef80), LL(0xde306f6fa16f5fce), LL(0x3f6d91917e91fcef), LL(0xa4f852525552aa07),
    LL(0xc04760609d6027fd), LL(0x6535bcbccabc8976), LL(0x2b379b9b569baccd), LL(0x018a8e8e028e048c),
    LL(0x5bd2a3a3b6a37115), LL(0x186c0c0c300c603c), LL(0xf6847b7bf17bff8a), LL(0x6a803535d435b5e1),
    LL(0x3af51d1d741de869), LL(0xddb3e0e0a7e05347), LL(0xb321d7d77bd7f6ac), LL(0x999cc2c22fc25eed),
    LL(0x5c432e2eb82e6d96), LL(0x96294b4b314b627a), LL(0xe15dfefedffea321), LL(0xaed5575741578216),
    LL(0x2abd15155415a841), LL(0xeee87777c1779fb6), LL(0x6e923737dc37a5eb), LL(0xd79ee5e5b3e57b56),
    LL(0x23139f9f469f8cd9), LL(0xfd23f0f0e7f0d317), LL(0x94204a4a354a6a7f), LL(0xa944dada4fda9e95),
    LL(0xb0a258587d58fa25), LL(0x8fcfc9c903c906ca), LL(0x527c2929a429558d), LL(0x145a0a0a280a5022),
    LL(0x7f50b1b1feb1e14f), LL(0x5dc9a0a0baa0691a), LL(0xd6146b6bb16b7fda), LL(0x17d985852e855cab),
    LL(0x673cbdbdcebd8173), LL(0xba8f5d5d695dd234), LL(0x2090101040108050), LL(0xf507f4f4f7f4f303),
    LL(0x8bddcbcb0bcb16c0), LL(0x7cd33e3ef83eedc6), LL(0x0a2d050514052811), LL(0xce78676781671fe6),
    LL(0xd597e4e4b7e47353), LL(0x4e0227279c2725bb), LL(0x8273414119413258), LL(0x0ba78b8b168b2c9d),
    LL(0x53f6a7a7a6a75101), LL(0xfab27d7de97dcf94), LL(0x374995956e95dcfb), LL(0xad56d8d847d88e9f),
    LL(0xeb70fbfbcbfb8b30), LL(0xc1cdeeee9fee2371), LL(0xf8bb7c7ced7cc791), LL(0xcc716666856617e3),
    LL(0xa77bdddd53dda68e), LL(0x2eaf17175c17b84b), LL(0x8e45474701470246), LL(0x211a9e9e429e84dc),
    LL(0x89d4caca0fca1ec5), LL(0x5a582d2db42d7599), LL(0x632ebfbfc6bf9179), LL(0x0e3f07071c07381b),
    LL(0x47acadad8ead0123), LL(0xb4b05a5a755aea2f), LL(0x1bef838336836cb5), LL(0x66b63333cc3385ff),
    LL(0xc65c636391633ff2), LL(0x041202020802100a), LL(0x4993aaaa92aa3938), LL(0xe2de7171d971afa8),
    LL(0x8dc6c8c807c80ecf), LL(0x32d119196419c87d), LL(0x923b494939497270), LL(0xaf5fd9d943d9869a),
    LL(0xf931f2f2eff2c31d), LL(0xdba8e3e3abe34b48), LL(0xb6b95b5b715be22a), LL(0x0dbc88881a883492),
    LL(0x293e9a9a529aa4c8), LL(0x4c0b262698262dbe), LL(0x64bf3232c8328dfa), LL(0x7d59b0b0fab0e94a),
    LL(0xcff2e9e983e91b6a), LL(0x1e770f0f3c0f7833), LL(0xb733d5d573d5e6a6), LL(0x1df480803a8074ba),
    LL(0x6127bebec2be997c), LL(0x87ebcdcd13cd26de), LL(0x68893434d034bde4), LL(0x903248483d487a75),
    LL(0xe354ffffdbffab24), LL(0xf48d7a7af57af78f), LL(0x3d6490907a90f4ea), LL(0xbe9d5f5f615fc23e),
    LL(0x403d202080201da0), LL(0xd00f6868bd6867d5), LL(0x34ca1a1a681ad072), LL(0x41b7aeae82ae192c),
    LL(0x757db4b4eab4c95e), LL(0xa8ce54544d549a19), LL(0x3b7f93937693ece5), LL(0x442f222288220daa),
    LL(0xc86364648d6407e9), LL(0xff2af1f1e3f1db12), LL(0xe6cc7373d173bfa2), LL(0x248212124812905a),
    LL(0x807a40401d403a5d), LL(0x1048080820084028), LL(0x9b95c3c32bc356e8), LL(0xc5dfecec97ec337b),
    LL(0xab4ddbdb4bdb9690), LL(0x5fc0a1a1bea1611f), LL(0x07918d8d0e8d1c83), LL(0x7ac83d3df43df5c9),
    LL(0x335b97976697ccf1), LL(0x0000000000000000), LL(0x83f9cfcf1bcf36d4), LL(0x566e2b2bac2b4587),
    LL(0xece17676c57697b3), LL(0x19e68282328264b0), LL(0xb128d6d67fd6fea9), LL(0x36c31b1b6c1bd877),
    LL(0x7774b5b5eeb5c15b), LL(0x43beafaf86af1129), LL(0xd41d6a6ab56a77df), LL(0xa0ea50505d50ba0d),
    LL(0x8a5745450945124c), LL(0xfb38f3f3ebf3cb18), LL(0x60ad3030c0309df0), LL(0xc3c4efef9bef2b74),
    LL(0x7eda3f3ffc3fe5c3), LL(0xaac755554955921c), LL(0x59dba2a2b2a27910), LL(0xc9e9eaea8fea0365),
    LL(0xca6a656589650fec), LL(0x6903babad2bab968), LL(0x5e4a2f2fbc2f6593), LL(0x9d8ec0c027c04ee7),
    LL(0xa160dede5fdebe81), LL(0x38fc1c1c701ce06c), LL(0xe746fdfdd3fdbb2e), LL(0x9a1f4d4d294d5264),
    LL(0x397692927292e4e0), LL(0xeafa7575c9758fbc), LL(0x0c3606061806301e), LL(0x09ae8a8a128a2498),
    LL(0x794bb2b2f2b2f940), LL(0xd185e6e6bfe66359), LL(0x1c7e0e0e380e7036), LL(0x3ee71f1f7c1ff863),
    LL(0xc4556262956237f7), LL(0xb53ad4d477d4eea3), LL(0x4d81a8a89aa82932), LL(0x315296966296c4f4),
    LL(0xef62f9f9c3f99b3a), LL(0x97a3c5c533c566f6), LL(0x4a102525942535b1), LL(0xb2ab59597959f220),
    LL(0x15d084842a8454ae), LL(0xe4c57272d572b7a7), LL(0x72ec3939e439d5dd), LL(0x98164c4c2d4c5a61),
    LL(0xbc945e5e655eca3b), LL(0xf09f7878fd78e785), LL(0x70e53838e038ddd8), LL(0x05988c8c0a8c1486),
    LL(0xbf17d1d163d1c6b2), LL(0x57e4a5a5aea5410b), LL(0xd9a1e2e2afe2434d), LL(0xc24e616199612ff8),
    LL(0x7b42b3b3f6b3f145), LL(0x42342121842115a5), LL(0x25089c9c4a9c94d6), LL(0x3cee1e1e781ef066),
    LL(0x8661434311432252), LL(0x93b1c7c73bc776fc), LL(0xe54ffcfcd7fcb32b), LL(0x0824040410042014),
    LL(0xa2e351515951b208), LL(0x2f2599995e99bcc7), LL(0xda226d6da96d4fc4), LL(0x1a650d0d340d6839),
    LL(0xe979fafacffa8335), LL(0xa369dfdf5bdfb684), LL(0xfca97e7ee57ed79b), LL(0x4819242490243db4),
    LL(0x76fe3b3bec3bc5d7), LL(0x4b9aabab96ab313d), LL(0x81f0cece1fce3ed1), LL(0x2299111144118855),
    LL(0x03838f8f068f0c89), LL(0x9c044e4e254e4a6b), LL(0x7366b7b7e6b7d151), LL(0xcbe0ebeb8beb0b60),
    LL(0x78c13c3cf03cfdcc), LL(0x1ffd81813e817cbf), LL(0x354094946a94d4fe), LL(0xf31cf7f7fbf7eb0c),
    LL(0x6f18b9b9deb9a167), LL(0x268b13134c13985f), LL(0x58512c2cb02c7d9c), LL(0xbb05d3d36bd3d6b8),
    LL(0xd38ce7e7bbe76b5c), LL(0xdc396e6ea56e57cb), LL(0x95aac4c437c46ef3), LL(0x061b03030c03180f),
    LL(0xacdc565645568a13), LL(0x885e44440d441a49), LL(0xfea07f7fe17fdf9e), LL(0x4f88a9a99ea92137),
    LL(0x54672a2aa82a4d82), LL(0x6b0abbbbd6bbb16d), LL(0x9f87c1c123c146e2), LL(0xa6f153535153a202),
    LL(0xa572dcdc57dcae8b), LL(0x16530b0b2c0b5827), LL(0x27019d9d4e9d9cd3), LL(0xd82b6c6cad6c47c1),
    LL(0x62a43131c43195f5), LL(0xe8f37474cd7487b9), LL(0xf115f6f6fff6e309), LL(0x8c4c464605460a43),
    LL(0x45a5acac8aac0926), LL(0x0fb589891e893c97), LL(0x28b414145014a044), LL(0xdfbae1e1a3e15b42),
    LL(0x2ca616165816b04e), LL(0x74f73a3ae83acdd2), LL(0xd2066969b9696fd0), LL(0x124109092409482d),
    LL(0xe0d77070dd70a7ad), LL(0x716fb6b6e2b6d954), LL(0xbd1ed0d067d0ceb7), LL(0xc7d6eded93ed3b7e),
    LL(0x85e2cccc17cc2edb), LL(0x8468424215422a57), LL(0x2d2c98985a98b4c2), LL(0x55eda4a4aaa4490e),
    LL(0x50752828a0285d88), LL(0xb8865c5c6d5cda31), LL(0xed6bf8f8c7f8933f), LL(0x11c28686228644a4),
};

static const u64 C3[256] = {
    LL(0x7830d818186018c0), LL(0xaf462623238c2305), LL(0xf991b8c6c63fc67e), LL(0x6fcdfbe8e887e813),
    LL(0xa113cb878726874c), LL(0x626d11b8b8dab8a9), LL(0x0502090101040108), LL(0x6e9e0d4f4f214f42),
    LL(0xee6c9b3636d836ad), LL(0x0451ffa6a6a2a659), LL(0xbdb90cd2d26fd2de), LL(0x06f70ef5f5f3f5fb),
    LL(0x80f2967979f979ef), LL(0xcede306f6fa16f5f), LL(0xef3f6d91917e91fc), LL(0x07a4f852525552aa),
    LL(0xfdc04760609d6027), LL(0x766535bcbccabc89), LL(0xcd2b379b9b569bac), LL(0x8c018a8e8e028e04),
    LL(0x155bd2a3a3b6a371), LL(0x3c186c0c0c300c60), LL(0x8af6847b7bf17bff), LL(0xe16a803535d435b5),
    LL(0x693af51d1d741de8), LL(0x47ddb3e0e0a7e053), LL(0xacb321d7d77bd7f6), LL(0xed999cc2c22fc25e),
    LL(0x965c432e2eb82e6d), LL(0x7a96294b4b314b62), LL(0x21e15dfefedffea3), LL(0x16aed55757415782),
    LL(0x412abd15155415a8), LL(0xb6eee87777c1779f), LL(0xeb6e923737dc37a5), LL(0x56d79ee5e5b3e57b),
    LL(0xd923139f9f469f8c), LL(0x17fd23f0f0e7f0d3), LL(0x7f94204a4a354a6a), LL(0x95a944dada4fda9e),
    LL(0x25b0a258587d58fa), LL(0xca8fcfc9c903c906), LL(0x8d527c2929a42955), LL(0x22145a0a0a280a50),
    LL(0x4f7f50b1b1feb1e1), LL(0x1a5dc9a0a0baa069), LL(0xdad6146b6bb16b7f), LL(0xab17d985852e855c),
    LL(0x73673cbdbdcebd81), LL(0x34ba8f5d5d695dd2), LL(0x5020901010401080), LL(0x03f507f4f4f7f4f3),
    LL(0xc08bddcbcb0bcb16), LL(0xc67cd33e3ef83eed), LL(0x110a2d0505140528), LL(0xe6ce78676781671f),
    LL(0x53d597e4e4b7e473), LL(0xbb4e0227279c2725), LL(0x5882734141194132), LL(0x9d0ba78b8b168b2c),
    LL(0x0153f6a7a7a6a751), LL(0x94fab27d7de97dcf), LL(0xfb374995956e95dc), LL(0x9fad56d8d847d88e),
    LL(0x30eb70fbfbcbfb8b), LL(0x71c1cdeeee9fee23), LL(0x91f8bb7c7ced7cc7), LL(0xe3cc716666856617),
    LL(0x8ea77bdddd53dda6), LL(0x4b2eaf17175c17b8), LL(0x468e454747014702), LL(0xdc211a9e9e429e84),
    LL(0xc589d4caca0fca1e), LL(0x995a582d2db42d75), LL(0x79632ebfbfc6bf91), LL(0x1b0e3f07071c0738),
    LL(0x2347acadad8ead01), LL(0x2fb4b05a5a755aea), LL(0xb51bef838336836c), LL(0xff66b63333cc3385),
    LL(0xf2c65c636391633f), LL(0x0a04120202080210), LL(0x384993aaaa92aa39), LL(0xa8e2de7171d971af),
    LL(0xcf8dc6c8c807c80e), LL(0x7d32d119196419c8), LL(0x70923b4949394972), LL(0x9aaf5fd9d943d986),
    LL(0x1df931f2f2eff2c3), LL(0x48dba8e3e3abe34b), LL(0x2ab6b95b5b715be2), LL(0x920dbc88881a8834),
    LL(0xc8293e9a9a529aa4), LL(0xbe4c0b262698262d), LL(0xfa64bf3232c8328d), LL(0x4a7d59b0b0fab0e9),
    LL(0x6acff2e9e983e91b), LL(0x331e770f0f3c0f78), LL(0xa6b733d5d573d5e6), LL(0xba1df480803a8074),
    LL(0x7c6127bebec2be99), LL(0xde87ebcdcd13cd26), LL(0xe468893434d034bd), LL(0x75903248483d487a),
    LL(0x24e354ffffdbffab), LL(0x8ff48d7a7af57af7), LL(0xea3d6490907a90f4), LL(0x3ebe9d5f5f615fc2),
    LL(0xa0403d202080201d), LL(0xd5d00f6868bd6867), LL(0x7234ca1a1a681ad0), LL(0x2c41b7aeae82ae19),
    LL(0x5e757db4b4eab4c9), LL(0x19a8ce54544d549a), LL(0xe53b7f93937693ec), LL(0xaa442f222288220d),
    LL(0xe9c86364648d6407), LL(0x12ff2af1f1e3f1db), LL(0xa2e6cc7373d173bf), LL(0x5a24821212481290),
    LL(0x5d807a40401d403a), LL(0x2810480808200840), LL(0xe89b95c3c32bc356), LL(0x7bc5dfecec97ec33),
    LL(0x90ab4ddbdb4bdb96), LL(0x1f5fc0a1a1bea161), LL(0x8307918d8d0e8d1c), LL(0xc97ac83d3df43df5),
    LL(0xf1335b97976697cc), LL(0x0000000000000000), LL(0xd483f9cfcf1bcf36), LL(0x87566e2b2bac2b45),
    LL(0xb3ece17676c57697), LL(0xb019e68282328264), LL(0xa9b128d6d67fd6fe), LL(0x7736c31b1b6c1bd8),
    LL(0x5b7774b5b5eeb5c1), LL(0x2943beafaf86af11), LL(0xdfd41d6a6ab56a77), LL(0x0da0ea50505d50ba),
    LL(0x4c8a574545094512), LL(0x18fb38f3f3ebf3cb), LL(0xf060ad3030c0309d), LL(0x74c3c4efef9bef2b),
    LL(0xc37eda3f3ffc3fe5), LL(0x1caac75555495592), LL(0x1059dba2a2b2a279), LL(0x65c9e9eaea8fea03),
    LL(0xecca6a656589650f), LL(0x686903babad2bab9), LL(0x935e4a2f2fbc2f65), LL(0xe79d8ec0c027c04e),
    LL(0x81a160dede5fdebe), LL(0x6c38fc1c1c701ce0), LL(0x2ee746fdfdd3fdbb), LL(0x649a1f4d4d294d52),
    LL(0xe0397692927292e4), LL(0xbceafa7575c9758f), LL(0x1e0c360606180630), LL(0x9809ae8a8a128a24),
    LL(0x40794bb2b2f2b2f9), LL(0x59d185e6e6bfe663), LL(0x361c7e0e0e380e70), LL(0x633ee71f1f7c1ff8),
    LL(0xf7c4556262956237), LL(0xa3b53ad4d477d4ee), LL(0x324d81a8a89aa829), LL(0xf4315296966296c4),
    LL(0x3aef62f9f9c3f99b), LL(0xf697a3c5c533c566), LL(0xb14a102525942535), LL(0x20b2ab59597959f2),
    LL(0xae15d084842a8454), LL(0xa7e4c57272d572b7), LL(0xdd72ec3939e439d5), LL(0x6198164c4c2d4c5a),
    LL(0x3bbc945e5e655eca), LL(0x85f09f7878fd78e7), LL(0xd870e53838e038dd), LL(0x8605988c8c0a8c14),
    LL(0xb2bf17d1d163d1c6), LL(0x0b57e4a5a5aea541), LL(0x4dd9a1e2e2afe243), LL(0xf8c24e616199612f),
    LL(0x457b42b3b3f6b3f1), LL(0xa542342121842115), LL(0xd625089c9c4a9c94), LL(0x663cee1e1e781ef0),
    LL(0x5286614343114322), LL(0xfc93b1c7c73bc776), LL(0x2be54ffcfcd7fcb3), LL(0x1408240404100420),
    LL(0x08a2e351515951b2), LL(0xc72f2599995e99bc), LL(0xc4da226d6da96d4f), LL(0x391a650d0d340d68),
    LL(0x35e979fafacffa83), LL(0x84a369dfdf5bdfb6), LL(0x9bfca97e7ee57ed7), LL(0xb44819242490243d),
    LL(0xd776fe3b3bec3bc5), LL(0x3d4b9aabab96ab31), LL(0xd181f0cece1fce3e), LL(0x5522991111441188),
    LL(0x8903838f8f068f0c), LL(0x6b9c044e4e254e4a), LL(0x517366b7b7e6b7d1), LL(0x60cbe0ebeb8beb0b),
    LL(0xcc78c13c3cf03cfd), LL(0xbf1ffd81813e817c), LL(0xfe354094946a94d4), LL(0x0cf31cf7f7fbf7eb),
    LL(0x676f18b9b9deb9a1), LL(0x5f268b13134c1398), LL(0x9c58512c2cb02c7d), LL(0xb8bb05d3d36bd3d6),
    LL(0x5cd38ce7e7bbe76b), LL(0xcbdc396e6ea56e57), LL(0xf395aac4c437c46e), LL(0x0f061b03030c0318),
    LL(0x13acdc565645568a), LL(0x49885e44440d441a), LL(0x9efea07f7fe17fdf), LL(0x374f88a9a99ea921),
    LL(0x8254672a2aa82a4d), LL(0x6d6b0abbbbd6bbb1), LL(0xe29f87c1c123c146), LL(0x02a6f153535153a2),
    LL(0x8ba572dcdc57dcae), LL(0x2716530b0b2c0b58), LL(0xd327019d9d4e9d9c), LL(0xc1d82b6c6cad6c47),
    LL(0xf562a43131c43195), LL(0xb9e8f37474cd7487), LL(0x09f115f6f6fff6e3), LL(0x438c4c464605460a),
    LL(0x2645a5acac8aac09), LL(0x970fb589891e893c), LL(0x4428b414145014a0), LL(0x42dfbae1e1a3e15b),
    LL(0x4e2ca616165816b0), LL(0xd274f73a3ae83acd), LL(0xd0d2066969b9696f), LL(0x2d12410909240948),
    LL(0xade0d77070dd70a7), LL(0x54716fb6b6e2b6d9), LL(0xb7bd1ed0d067d0ce), LL(0x7ec7d6eded93ed3b),
    LL(0xdb85e2cccc17cc2e), LL(0x578468424215422a), LL(0xc22d2c98985a98b4), LL(0x0e55eda4a4aaa449),
    LL(0x8850752828a0285d), LL(0x31b8865c5c6d5cda), LL(0x3fed6bf8f8c7f893), LL(0xa411c28686228644),
};

static const u64 C4[256] = {
    LL(0xc07830d818186018), LL(0x05af462623238c23), LL(0x7ef991b8c6c63fc6), LL(0x136fcdfbe8e887e8),
    LL(0x4ca113cb87872687), LL(0xa9626d11b8b8dab8), LL(0x0805020901010401), LL(0x426e9e0d4f4f214f),
    LL(0xadee6c9b3636d836), LL(0x590451ffa6a6a2a6), LL(0xdebdb90cd2d26fd2), LL(0xfb06f70ef5f5f3f5),
    LL(0xef80f2967979f979), LL(0x5fcede306f6fa16f), LL(0xfcef3f6d91917e91), LL(0xaa07a4f852525552),
    LL(0x27fdc04760609d60), LL(0x89766535bcbccabc), LL(0xaccd2b379b9b569b), LL(0x048c018a8e8e028e),
    LL(0x71155bd2a3a3b6a3), LL(0x603c186c0c0c300c), LL(0xff8af6847b7bf17b), LL(0xb5e16a803535d435),
    LL(0xe8693af51d1d741d), LL(0x5347ddb3e0e0a7e0), LL(0xf6acb321d7d77bd7), LL(0x5eed999cc2c22fc2),
    LL(0x6d965c432e2eb82e), LL(0x627a96294b4b314b), LL(0xa321e15dfefedffe), LL(0x8216aed557574157),
    LL(0xa8412abd15155415), LL(0x9fb6eee87777c177), LL(0xa5eb6e923737dc37), LL(0x7b56d79ee5e5b3e5),
    LL(0x8cd923139f9f469f), LL(0xd317fd23f0f0e7f0), LL(0x6a7f94204a4a354a), LL(0x9e95a944dada4fda),
    LL(0xfa25b0a258587d58), LL(0x06ca8fcfc9c903c9), LL(0x558d527c2929a429), LL(0x5022145a0a0a280a),
    LL(0xe14f7f50b1b1feb1), LL(0x691a5dc9a0a0baa0), LL(0x7fdad6146b6bb16b), LL(0x5cab17d985852e85),
    LL(0x8173673cbdbdcebd), LL(0xd234ba8f5d5d695d), LL(0x8050209010104010), LL(0xf303f507f4f4f7f4),
    LL(0x16c08bddcbcb0bcb), LL(0xedc67cd33e3ef83e), LL(0x28110a2d05051405), LL(0x1fe6ce7867678167),
    LL(0x7353d597e4e4b7e4), LL(0x25bb4e0227279c27), LL(0x3258827341411941), LL(0x2c9d0ba78b8b168b),
    LL(0x510153f6a7a7a6a7), LL(0xcf94fab27d7de97d), LL(0xdcfb374995956e95), LL(0x8e9fad56d8d847d8),
    LL(0x8b30eb70fbfbcbfb), LL(0x2371c1cdeeee9fee), LL(0xc791f8bb7c7ced7c), LL(0x17e3cc7166668566),
    LL(0xa68ea77bdddd53dd), LL(0xb84b2eaf17175c17), LL(0x02468e4547470147), LL(0x84dc211a9e9e429e),
    LL(0x1ec589d4caca0fca), LL(0x75995a582d2db42d), LL(0x9179632ebfbfc6bf), LL(0x381b0e3f07071c07),
    LL(0x012347acadad8ead), LL(0xea2fb4b05a5a755a), LL(0x6cb51bef83833683), LL(0x85ff66b63333cc33),
    LL(0x3ff2c65c63639163), LL(0x100a041202020802), LL(0x39384993aaaa92aa), LL(0xafa8e2de7171d971),
    LL(0x0ecf8dc6c8c807c8), LL(0xc87d32d119196419), LL(0x7270923b49493949), LL(0x869aaf5fd9d943d9),
    LL(0xc31df931f2f2eff2), LL(0x4b48dba8e3e3abe3), LL(0xe22ab6b95b5b715b), LL(0x34920dbc88881a88),
    LL(0xa4c8293e9a9a529a), LL(0x2dbe4c0b26269826), LL(0x8dfa64bf3232c832), LL(0xe94a7d59b0b0fab0),
    LL(0x1b6acff2e9e983e9), LL(0x78331e770f0f3c0f), LL(0xe6a6b733d5d573d5), LL(0x74ba1df480803a80),
    LL(0x997c6127bebec2be), LL(0x26de87ebcdcd13cd), LL(0xbde468893434d034), LL(0x7a75903248483d48),
    LL(0xab24e354ffffdbff), LL(0xf78ff48d7a7af57a), LL(0xf4ea3d6490907a90), LL(0xc23ebe9d5f5f615f),
    LL(0x1da0403d20208020), LL(0x67d5d00f6868bd68), LL(0xd07234ca1a1a681a), LL(0x192c41b7aeae82ae),
    LL(0xc95e757db4b4eab4), LL(0x9a19a8ce54544d54), LL(0xece53b7f93937693), LL(0x0daa442f22228822),
    LL(0x07e9c86364648d64), LL(0xdb12ff2af1f1e3f1), LL(0xbfa2e6cc7373d173), LL(0x905a248212124812),
    LL(0x3a5d807a40401d40), LL(0x4028104808082008), LL(0x56e89b95c3c32bc3), LL(0x337bc5dfecec97ec),
    LL(0x9690ab4ddbdb4bdb), LL(0x611f5fc0a1a1bea1), LL(0x1c8307918d8d0e8d), LL(0xf5c97ac83d3df43d),
    LL(0xccf1335b97976697), LL(0x0000000000000000), LL(0x36d483f9cfcf1bcf), LL(0x4587566e2b2bac2b),
    LL(0x97b3ece17676c576), LL(0x64b019e682823282), LL(0xfea9b128d6d67fd6), LL(0xd87736c31b1b6c1b),
    LL(0xc15b7774b5b5eeb5), LL(0x112943beafaf86af), LL(0x77dfd41d6a6ab56a), LL(0xba0da0ea50505d50),
    LL(0x124c8a5745450945), LL(0xcb18fb38f3f3ebf3), LL(0x9df060ad3030c030), LL(0x2b74c3c4efef9bef),
    LL(0xe5c37eda3f3ffc3f), LL(0x921caac755554955), LL(0x791059dba2a2b2a2), LL(0x0365c9e9eaea8fea),
    LL(0x0fecca6a65658965), LL(0xb9686903babad2ba), LL(0x65935e4a2f2fbc2f), LL(0x4ee79d8ec0c027c0),
    LL(0xbe81a160dede5fde), LL(0xe06c38fc1c1c701c), LL(0xbb2ee746fdfdd3fd), LL(0x52649a1f4d4d294d),
    LL(0xe4e0397692927292), LL(0x8fbceafa7575c975), LL(0x301e0c3606061806), LL(0x249809ae8a8a128a),
    LL(0xf940794bb2b2f2b2), LL(0x6359d185e6e6bfe6), LL(0x70361c7e0e0e380e), LL(0xf8633ee71f1f7c1f),
    LL(0x37f7c45562629562), LL(0xeea3b53ad4d477d4), LL(0x29324d81a8a89aa8), LL(0xc4f4315296966296),
    LL(0x9b3aef62f9f9c3f9), LL(0x66f697a3c5c533c5), LL(0x35b14a1025259425), LL(0xf220b2ab59597959),
    LL(0x54ae15d084842a84), LL(0xb7a7e4c57272d572), LL(0xd5dd72ec3939e439), LL(0x5a6198164c4c2d4c),
    LL(0xca3bbc945e5e655e), LL(0xe785f09f7878fd78), LL(0xddd870e53838e038), LL(0x148605988c8c0a8c),
    LL(0xc6b2bf17d1d163d1), LL(0x410b57e4a5a5aea5), LL(0x434dd9a1e2e2afe2), LL(0x2ff8c24e61619961),
    LL(0xf1457b42b3b3f6b3), LL(0x15a5423421218421), LL(0x94d625089c9c4a9c), LL(0xf0663cee1e1e781e),
    LL(0x2252866143431143), LL(0x76fc93b1c7c73bc7), LL(0xb32be54ffcfcd7fc), LL(0x2014082404041004),
    LL(0xb208a2e351515951), LL(0xbcc72f2599995e99), LL(0x4fc4da226d6da96d), LL(0x68391a650d0d340d),
    LL(0x8335e979fafacffa), LL(0xb684a369dfdf5bdf), LL(0xd79bfca97e7ee57e), LL(0x3db4481924249024),
    LL(0xc5d776fe3b3bec3b), LL(0x313d4b9aabab96ab), LL(0x3ed181f0cece1fce), LL(0x8855229911114411),
    LL(0x0c8903838f8f068f), LL(0x4a6b9c044e4e254e), LL(0xd1517366b7b7e6b7), LL(0x0b60cbe0ebeb8beb),
    LL(0xfdcc78c13c3cf03c), LL(0x7cbf1ffd81813e81), LL(0xd4fe354094946a94), LL(0xeb0cf31cf7f7fbf7),
    LL(0xa1676f18b9b9deb9), LL(0x985f268b13134c13), LL(0x7d9c58512c2cb02c), LL(0xd6b8bb05d3d36bd3),
    LL(0x6b5cd38ce7e7bbe7), LL(0x57cbdc396e6ea56e), LL(0x6ef395aac4c437c4), LL(0x180f061b03030c03),
    LL(0x8a13acdc56564556), LL(0x1a49885e44440d44), LL(0xdf9efea07f7fe17f), LL(0x21374f88a9a99ea9),
    LL(0x4d8254672a2aa82a), LL(0xb16d6b0abbbbd6bb), LL(0x46e29f87c1c123c1), LL(0xa202a6f153535153),
    LL(0xae8ba572dcdc57dc), LL(0x582716530b0b2c0b), LL(0x9cd327019d9d4e9d), LL(0x47c1d82b6c6cad6c),
    LL(0x95f562a43131c431), LL(0x87b9e8f37474cd74), LL(0xe309f115f6f6fff6), LL(0x0a438c4c46460546),
    LL(0x092645a5acac8aac), LL(0x3c970fb589891e89), LL(0xa04428b414145014), LL(0x5b42dfbae1e1a3e1),
    LL(0xb04e2ca616165816), LL(0xcdd274f73a3ae83a), LL(0x6fd0d2066969b969), LL(0x482d124109092409),
    LL(0xa7ade0d77070dd70), LL(0xd954716fb6b6e2b6), LL(0xceb7bd1ed0d067d0), LL(0x3b7ec7d6eded93ed),
    LL(0x2edb85e2cccc17cc), LL(0x2a57846842421542), LL(0xb4c22d2c98985a98), LL(0x490e55eda4a4aaa4),
    LL(0x5d8850752828a028), LL(0xda31b8865c5c6d5c), LL(0x933fed6bf8f8c7f8), LL(0x44a411c286862286),
};

static const u64 C5[256] = {
    LL(0x18c07830d8181860), LL(0x2305af462623238c), LL(0xc67ef991b8c6c63f), LL(0xe8136fcdfbe8e887),
    LL(0x874ca113cb878726), LL(0xb8a9626d11b8b8da), LL(0x0108050209010104), LL(0x4f426e9e0d4f4f21),
    LL(0x36adee6c9b3636d8), LL(0xa6590451ffa6a6a2), LL(0xd2debdb90cd2d26f), LL(0xf5fb06f70ef5f5f3),
    LL(0x79ef80f2967979f9), LL(0x6f5fcede306f6fa1), LL(0x91fcef3f6d91917e), LL(0x52aa07a4f8525255),
    LL(0x6027fdc04760609d), LL(0xbc89766535bcbcca), LL(0x9baccd2b379b9b56), LL(0x8e048c018a8e8e02),
    LL(0xa371155bd2a3a3b6), LL(0x0c603c186c0c0c30), LL(0x7bff8af6847b7bf1), LL(0x35b5e16a803535d4),
    LL(0x1de8693af51d1d74), LL(0xe05347ddb3e0e0a7), LL(0xd7f6acb321d7d77b), LL(0xc25eed999cc2c22f),
    LL(0x2e6d965c432e2eb8), LL(0x4b627a96294b4b31), LL(0xfea321e15dfefedf), LL(0x578216aed5575741),
    LL(0x15a8412abd151554), LL(0x779fb6eee87777c1), LL(0x37a5eb6e923737dc), LL(0xe57b56d79ee5e5b3),
    LL(0x9f8cd923139f9f46), LL(0xf0d317fd23f0f0e7), LL(0x4a6a7f94204a4a35), LL(0xda9e95a944dada4f),
    LL(0x58fa25b0a258587d), LL(0xc906ca8fcfc9c903), LL(0x29558d527c2929a4), LL(0x0a5022145a0a0a28),
    LL(0xb1e14f7f50b1b1fe), LL(0xa0691a5dc9a0a0ba), LL(0x6b7fdad6146b6bb1), LL(0x855cab17d985852e),
    LL(0xbd8173673cbdbdce), LL(0x5dd234ba8f5d5d69), LL(0x1080502090101040), LL(0xf4f303f507f4f4f7),
    LL(0xcb16c08bddcbcb0b), LL(0x3eedc67cd33e3ef8), LL(0x0528110a2d050514), LL(0x671fe6ce78676781),
    LL(0xe47353d597e4e4b7), LL(0x2725bb4e0227279c), LL(0x4132588273414119), LL(0x8b2c9d0ba78b8b16),
    LL(0xa7510153f6a7a7a6), LL(0x7dcf94fab27d7de9), LL(0x95dcfb374995956e), LL(0xd88e9fad56d8d847),
    LL(0xfb8b30eb70fbfbcb), LL(0xee2371c1cdeeee9f), LL(0x7cc791f8bb7c7ced), LL(0x6617e3cc71666685),
    LL(0xdda68ea77bdddd53), LL(0x17b84b2eaf17175c), LL(0x4702468e45474701), LL(0x9e84dc211a9e9e42),
    LL(0xca1ec589d4caca0f), LL(0x2d75995a582d2db4), LL(0xbf9179632ebfbfc6), LL(0x07381b0e3f07071c),
    LL(0xad012347acadad8e), LL(0x5aea2fb4b05a5a75), LL(0x836cb51bef838336), LL(0x3385ff66b63333cc),
    LL(0x633ff2c65c636391), LL(0x02100a0412020208), LL(0xaa39384993aaaa92), LL(0x71afa8e2de7171d9),
    LL(0xc80ecf8dc6c8c807), LL(0x19c87d32d1191964), LL(0x497270923b494939), LL(0xd9869aaf5fd9d943),
    LL(0xf2c31df931f2f2ef), LL(0xe34b48dba8e3e3ab), LL(0x5be22ab6b95b5b71), LL(0x8834920dbc88881a),
    LL(0x9aa4c8293e9a9a52), LL(0x262dbe4c0b262698), LL(0x328dfa64bf3232c8), LL(0xb0e94a7d59b0b0fa),
    LL(0xe91b6acff2e9e983), LL(0x0f78331e770f0f3c), LL(0xd5e6a6b733d5d573), LL(0x8074ba1df480803a),
    LL(0xbe997c6127bebec2), LL(0xcd26de87ebcdcd13), LL(0x34bde468893434d0), LL(0x487a75903248483d),
    LL(0xffab24e354ffffdb), LL(0x7af78ff48d7a7af5), LL(0x90f4ea3d6490907a), LL(0x5fc23ebe9d5f5f61),
    LL(0x201da0403d202080), LL(0x6867d5d00f6868bd), LL(0x1ad07234ca1a1a68), LL(0xae192c41b7aeae82),
    LL(0xb4c95e757db4b4ea), LL(0x549a19a8ce54544d), LL(0x93ece53b7f939376), LL(0x220daa442f222288),
    LL(0x6407e9c86364648d), LL(0xf1db12ff2af1f1e3), LL(0x73bfa2e6cc7373d1), LL(0x12905a2482121248),
    LL(0x403a5d807a40401d), LL(0x0840281048080820), LL(0xc356e89b95c3c32b), LL(0xec337bc5dfecec97),
    LL(0xdb9690ab4ddbdb4b), LL(0xa1611f5fc0a1a1be), LL(0x8d1c8307918d8d0e), LL(0x3df5c97ac83d3df4),
    LL(0x97ccf1335b979766), LL(0x0000000000000000), LL(0xcf36d483f9cfcf1b), LL(0x2b4587566e2b2bac),
    LL(0x7697b3ece17676c5), LL(0x8264b019e6828232), LL(0xd6fea9b128d6d67f), LL(0x1bd87736c31b1b6c),
    LL(0xb5c15b7774b5b5ee), LL(0xaf112943beafaf86), LL(0x6a77dfd41d6a6ab5), LL(0x50ba0da0ea50505d),
    LL(0x45124c8a57454509), LL(0xf3cb18fb38f3f3eb), LL(0x309df060ad3030c0), LL(0xef2b74c3c4efef9b),
    LL(0x3fe5c37eda3f3ffc), LL(0x55921caac7555549), LL(0xa2791059dba2a2b2), LL(0xea0365c9e9eaea8f),
    LL(0x650fecca6a656589), LL(0xbab9686903babad2), LL(0x2f65935e4a2f2fbc), LL(0xc04ee79d8ec0c027),
    LL(0xdebe81a160dede5f), LL(0x1ce06c38fc1c1c70), LL(0xfdbb2ee746fdfdd3), LL(0x4d52649a1f4d4d29),
    LL(0x92e4e03976929272), LL(0x758fbceafa7575c9), LL(0x06301e0c36060618), LL(0x8a249809ae8a8a12),
    LL(0xb2f940794bb2b2f2), LL(0xe66359d185e6e6bf), LL(0x0e70361c7e0e0e38), LL(0x1ff8633ee71f1f7c),
    LL(0x6237f7c455626295), LL(0xd4eea3b53ad4d477), LL(0xa829324d81a8a89a), LL(0x96c4f43152969662),
    LL(0xf99b3aef62f9f9c3), LL(0xc566f697a3c5c533), LL(0x2535b14a10252594), LL(0x59f220b2ab595979),
    LL(0x8454ae15d084842a), LL(0x72b7a7e4c57272d5), LL(0x39d5dd72ec3939e4), LL(0x4c5a6198164c4c2d),
    LL(0x5eca3bbc945e5e65), LL(0x78e785f09f7878fd), LL(0x38ddd870e53838e0), LL(0x8c148605988c8c0a),
    LL(0xd1c6b2bf17d1d163), LL(0xa5410b57e4a5a5ae), LL(0xe2434dd9a1e2e2af), LL(0x612ff8c24e616199),
    LL(0xb3f1457b42b3b3f6), LL(0x2115a54234212184), LL(0x9c94d625089c9c4a), LL(0x1ef0663cee1e1e78),
    LL(0x4322528661434311), LL(0xc776fc93b1c7c73b), LL(0xfcb32be54ffcfcd7), LL(0x0420140824040410),
    LL(0x51b208a2e3515159), LL(0x99bcc72f2599995e), LL(0x6d4fc4da226d6da9), LL(0x0d68391a650d0d34),
    LL(0xfa8335e979fafacf), LL(0xdfb684a369dfdf5b), LL(0x7ed79bfca97e7ee5), LL(0x243db44819242490),
    LL(0x3bc5d776fe3b3bec), LL(0xab313d4b9aabab96), LL(0xce3ed181f0cece1f), LL(0x1188552299111144),
    LL(0x8f0c8903838f8f06), LL(0x4e4a6b9c044e4e25), LL(0xb7d1517366b7b7e6), LL(0xeb0b60cbe0ebeb8b),
    LL(0x3cfdcc78c13c3cf0), LL(0x817cbf1ffd81813e), LL(0x94d4fe354094946a), LL(0xf7eb0cf31cf7f7fb),
    LL(0xb9a1676f18b9b9de), LL(0x13985f268b13134c), LL(0x2c7d9c58512c2cb0), LL(0xd3d6b8bb05d3d36b),
    LL(0xe76b5cd38ce7e7bb), LL(0x6e57cbdc396e6ea5), LL(0xc46ef395aac4c437), LL(0x03180f061b03030c),
    LL(0x568a13acdc565645), LL(0x441a49885e44440d), LL(0x7fdf9efea07f7fe1), LL(0xa921374f88a9a99e),
    LL(0x2a4d8254672a2aa8), LL(0xbbb16d6b0abbbbd6), LL(0xc146e29f87c1c123), LL(0x53a202a6f1535351),
    LL(0xdcae8ba572dcdc57), LL(0x0b582716530b0b2c), LL(0x9d9cd327019d9d4e), LL(0x6c47c1d82b6c6cad),
    LL(0x3195f562a43131c4), LL(0x7487b9e8f37474cd), LL(0xf6e309f115f6f6ff), LL(0x460a438c4c464605),
    LL(0xac092645a5acac8a), LL(0x893c970fb589891e), LL(0x14a04428b4141450), LL(0xe15b42dfbae1e1a3),
    LL(0x16b04e2ca6161658), LL(0x3acdd274f73a3ae8), LL(0x696fd0d2066969b9), LL(0x09482d1241090924),
    LL(0x70a7ade0d77070dd), LL(0xb6d954716fb6b6e2), LL(0xd0ceb7bd1ed0d067), LL(0xed3b7ec7d6eded93),
    LL(0xcc2edb85e2cccc17), LL(0x422a578468424215), LL(0x98b4c22d2c98985a), LL(0xa4490e55eda4a4aa),
    LL(0x285d8850752828a0), LL(0x5cda31b8865c5c6d), LL(0xf8933fed6bf8f8c7), LL(0x8644a411c2868622),
};

static const u64 C6[256] = {
    LL(0x6018c07830d81818), LL(0x8c2305af46262323), LL(0x3fc67ef991b8c6c6), LL(0x87e8136fcdfbe8e8),
    LL(0x26874ca113cb8787), LL(0xdab8a9626d11b8b8), LL(0x0401080502090101), LL(0x214f426e9e0d4f4f),
    LL(0xd836adee6c9b3636), LL(0xa2a6590451ffa6a6), LL(0x6fd2debdb90cd2d2), LL(0xf3f5fb06f70ef5f5),
    LL(0xf979ef80f2967979), LL(0xa16f5fcede306f6f), LL(0x7e91fcef3f6d9191), LL(0x5552aa07a4f85252),
    LL(0x9d6027fdc0476060), LL(0xcabc89766535bcbc), LL(0x569baccd2b379b9b), LL(0x028e048c018a8e8e),
    LL(0xb6a371155bd2a3a3), LL(0x300c603c186c0c0c), LL(0xf17bff8af6847b7b), LL(0xd435b5e16a803535),
    LL(0x741de8693af51d1d), LL(0xa7e05347ddb3e0e0), LL(0x7bd7f6acb321d7d7), LL(0x2fc25eed999cc2c2),
    LL(0xb82e6d965c432e2e), LL(0x314b627a96294b4b), LL(0xdffea321e15dfefe), LL(0x41578216aed55757),
    LL(0x5415a8412abd1515), LL(0xc1779fb6eee87777), LL(0xdc37a5eb6e923737), LL(0xb3e57b56d79ee5e5),
    LL(0x469f8cd923139f9f), LL(0xe7f0d317fd23f0f0), LL(0x354a6a7f94204a4a), LL(0x4fda9e95a944dada),
    LL(0x7d58fa25b0a25858), LL(0x03c906ca8fcfc9c9), LL(0xa429558d527c2929), LL(0x280a5022145a0a0a),
    LL(0xfeb1e14f7f50b1b1), LL(0xbaa0691a5dc9a0a0), LL(0xb16b7fdad6146b6b), LL(0x2e855cab17d98585),
    LL(0xcebd8173673cbdbd), LL(0x695dd234ba8f5d5d), LL(0x4010805020901010), LL(0xf7f4f303f507f4f4),
    LL(0x0bcb16c08bddcbcb), LL(0xf83eedc67cd33e3e), LL(0x140528110a2d0505), LL(0x81671fe6ce786767),
    LL(0xb7e47353d597e4e4), LL(0x9c2725bb4e022727), LL(0x1941325882734141), LL(0x168b2c9d0ba78b8b),
    LL(0xa6a7510153f6a7a7), LL(0xe97dcf94fab27d7d), LL(0x6e95dcfb37499595), LL(0x47d88e9fad56d8d8),
    LL(0xcbfb8b30eb70fbfb), LL(0x9fee2371c1cdeeee), LL(0xed7cc791f8bb7c7c), LL(0x856617e3cc716666),
    LL(0x53dda68ea77bdddd), LL(0x5c17b84b2eaf1717), LL(0x014702468e454747), LL(0x429e84dc211a9e9e),
    LL(0x0fca1ec589d4caca), LL(0xb42d75995a582d2d), LL(0xc6bf9179632ebfbf), LL(0x1c07381b0e3f0707),
    LL(0x8ead012347acadad), LL(0x755aea2fb4b05a5a), LL(0x36836cb51bef8383), LL(0xcc3385ff66b63333),
    LL(0x91633ff2c65c6363), LL(0x0802100a04120202), LL(0x92aa39384993aaaa), LL(0xd971afa8e2de7171),
    LL(0x07c80ecf8dc6c8c8), LL(0x6419c87d32d11919), LL(0x39497270923b4949), LL(0x43d9869aaf5fd9d9),
    LL(0xeff2c31df931f2f2), LL(0xabe34b48dba8e3e3), LL(0x715be22ab6b95b5b), LL(0x1a8834920dbc8888),
    LL(0x529aa4c8293e9a9a), LL(0x98262dbe4c0b2626), LL(0xc8328dfa64bf3232), LL(0xfab0e94a7d59b0b0),
    LL(0x83e91b6acff2e9e9), LL(0x3c0f78331e770f0f), LL(0x73d5e6a6b733d5d5), LL(0x3a8074ba1df48080),
    LL(0xc2be997c6127bebe), LL(0x13cd26de87ebcdcd), LL(0xd034bde468893434), LL(0x3d487a7590324848),
    LL(0xdbffab24e354ffff), LL(0xf57af78ff48d7a7a), LL(0x7a90f4ea3d649090), LL(0x615fc23ebe9d5f5f),
    LL(0x80201da0403d2020), LL(0xbd6867d5d00f6868), LL(0x681ad07234ca1a1a), LL(0x82ae192c41b7aeae),
    LL(0xeab4c95e757db4b4), LL(0x4d549a19a8ce5454), LL(0x7693ece53b7f9393), LL(0x88220daa442f2222),
    LL(0x8d6407e9c8636464), LL(0xe3f1db12ff2af1f1), LL(0xd173bfa2e6cc7373), LL(0x4812905a24821212),
    LL(0x1d403a5d807a4040), LL(0x2008402810480808), LL(0x2bc356e89b95c3c3), LL(0x97ec337bc5dfecec),
    LL(0x4bdb9690ab4ddbdb), LL(0xbea1611f5fc0a1a1), LL(0x0e8d1c8307918d8d), LL(0xf43df5c97ac83d3d),
    LL(0x6697ccf1335b9797), LL(0x0000000000000000), LL(0x1bcf36d483f9cfcf), LL(0xac2b4587566e2b2b),
    LL(0xc57697b3ece17676), LL(0x328264b019e68282), LL(0x7fd6fea9b128d6d6), LL(0x6c1bd87736c31b1b),
    LL(0xeeb5c15b7774b5b5), LL(0x86af112943beafaf), LL(0xb56a77dfd41d6a6a), LL(0x5d50ba0da0ea5050),
    LL(0x0945124c8a574545), LL(0xebf3cb18fb38f3f3), LL(0xc0309df060ad3030), LL(0x9bef2b74c3c4efef),
    LL(0xfc3fe5c37eda3f3f), LL(0x4955921caac75555), LL(0xb2a2791059dba2a2), LL(0x8fea0365c9e9eaea),
    LL(0x89650fecca6a6565), LL(0xd2bab9686903baba), LL(0xbc2f65935e4a2f2f), LL(0x27c04ee79d8ec0c0),
    LL(0x5fdebe81a160dede), LL(0x701ce06c38fc1c1c), LL(0xd3fdbb2ee746fdfd), LL(0x294d52649a1f4d4d),
    LL(0x7292e4e039769292), LL(0xc9758fbceafa7575), LL(0x1806301e0c360606), LL(0x128a249809ae8a8a),
    LL(0xf2b2f940794bb2b2), LL(0xbfe66359d185e6e6), LL(0x380e70361c7e0e0e), LL(0x7c1ff8633ee71f1f),
    LL(0x956237f7c4556262), LL(0x77d4eea3b53ad4d4), LL(0x9aa829324d81a8a8), LL(0x6296c4f431529696),
    LL(0xc3f99b3aef62f9f9), LL(0x33c566f697a3c5c5), LL(0x942535b14a102525), LL(0x7959f220b2ab5959),
    LL(0x2a8454ae15d08484), LL(0xd572b7a7e4c57272), LL(0xe439d5dd72ec3939), LL(0x2d4c5a6198164c4c),
    LL(0x655eca3bbc945e5e), LL(0xfd78e785f09f7878), LL(0xe038ddd870e53838), LL(0x0a8c148605988c8c),
    LL(0x63d1c6b2bf17d1d1), LL(0xaea5410b57e4a5a5), LL(0xafe2434dd9a1e2e2), LL(0x99612ff8c24e6161),
    LL(0xf6b3f1457b42b3b3), LL(0x842115a542342121), LL(0x4a9c94d625089c9c), LL(0x781ef0663cee1e1e),
    LL(0x1143225286614343), LL(0x3bc776fc93b1c7c7), LL(0xd7fcb32be54ffcfc), LL(0x1004201408240404),
    LL(0x5951b208a2e35151), LL(0x5e99bcc72f259999), LL(0xa96d4fc4da226d6d), LL(0x340d68391a650d0d),
    LL(0xcffa8335e979fafa), LL(0x5bdfb684a369dfdf), LL(0xe57ed79bfca97e7e), LL(0x90243db448192424),
    LL(0xec3bc5d776fe3b3b), LL(0x96ab313d4b9aabab), LL(0x1fce3ed181f0cece), LL(0x4411885522991111),
    LL(0x068f0c8903838f8f), LL(0x254e4a6b9c044e4e), LL(0xe6b7d1517366b7b7), LL(0x8beb0b60cbe0ebeb),
    LL(0xf03cfdcc78c13c3c), LL(0x3e817cbf1ffd8181), LL(0x6a94d4fe35409494), LL(0xfbf7eb0cf31cf7f7),
    LL(0xdeb9a1676f18b9b9), LL(0x4c13985f268b1313), LL(0xb02c7d9c58512c2c), LL(0x6bd3d6b8bb05d3d3),
    LL(0xbbe76b5cd38ce7e7), LL(0xa56e57cbdc396e6e), LL(0x37c46ef395aac4c4), LL(0x0c03180f061b0303),
    LL(0x45568a13acdc5656), LL(0x0d441a49885e4444), LL(0xe17fdf9efea07f7f), LL(0x9ea921374f88a9a9),
    LL(0xa82a4d8254672a2a), LL(0xd6bbb16d6b0abbbb), LL(0x23c146e29f87c1c1), LL(0x5153a202a6f15353),
    LL(0x57dcae8ba572dcdc), LL(0x2c0b582716530b0b), LL(0x4e9d9cd327019d9d), LL(0xad6c47c1d82b6c6c),
    LL(0xc43195f562a43131), LL(0xcd7487b9e8f37474), LL(0xfff6e309f115f6f6), LL(0x05460a438c4c4646),
    LL(0x8aac092645a5acac), LL(0x1e893c970fb58989), LL(0x5014a04428b41414), LL(0xa3e15b42dfbae1e1),
    LL(0x5816b04e2ca61616), LL(0xe83acdd274f73a3a), LL(0xb9696fd0d2066969), LL(0x2409482d12410909),
    LL(0xdd70a7ade0d77070), LL(0xe2b6d954716fb6b6), LL(0x67d0ceb7bd1ed0d0), LL(0x93ed3b7ec7d6eded),
    LL(0x17cc2edb85e2cccc), LL(0x15422a5784684242), LL(0x5a98b4c22d2c9898), LL(0xaaa4490e55eda4a4),
    LL(0xa0285d8850752828), LL(0x6d5cda31b8865c5c), LL(0xc7f8933fed6bf8f8), LL(0x228644a411c28686),
};

static const u64 C7[256] = {
    LL(0x186018c07830d818), LL(0x238c2305af462623), LL(0xc63fc67ef991b8c6), LL(0xe887e8136fcdfbe8),
    LL(0x8726874ca113cb87), LL(0xb8dab8a9626d11b8), LL(0x0104010805020901), LL(0x4f214f426e9e0d4f),
    LL(0x36d836adee6c9b36), LL(0xa6a2a6590451ffa6), LL(0xd26fd2debdb90cd2), LL(0xf5f3f5fb06f70ef5),
    LL(0x79f979ef80f29679), LL(0x6fa16f5fcede306f), LL(0x917e91fcef3f6d91), LL(0x525552aa07a4f852),
    LL(0x609d6027fdc04760), LL(0xbccabc89766535bc), LL(0x9b569baccd2b379b), LL(0x8e028e048c018a8e),
    LL(0xa3b6a371155bd2a3), LL(0x0c300c603c186c0c), LL(0x7bf17bff8af6847b), LL(0x35d435b5e16a8035),
    LL(0x1d741de8693af51d), LL(0xe0a7e05347ddb3e0), LL(0xd77bd7f6acb321d7), LL(0xc22fc25eed999cc2),
    LL(0x2eb82e6d965c432e), LL(0x4b314b627a96294b), LL(0xfedffea321e15dfe), LL(0x5741578216aed557),
    LL(0x155415a8412abd15), LL(0x77c1779fb6eee877), LL(0x37dc37a5eb6e9237), LL(0xe5b3e57b56d79ee5),
    LL(0x9f469f8cd923139f), LL(0xf0e7f0d317fd23f0), LL(0x4a354a6a7f94204a), LL(0xda4fda9e95a944da),
    LL(0x587d58fa25b0a258), LL(0xc903c906ca8fcfc9), LL(0x29a429558d527c29), LL(0x0a280a5022145a0a),
    LL(0xb1feb1e14f7f50b1), LL(0xa0baa0691a5dc9a0), LL(0x6bb16b7fdad6146b), LL(0x852e855cab17d985),
    LL(0xbdcebd8173673cbd), LL(0x5d695dd234ba8f5d), LL(0x1040108050209010), LL(0xf4f7f4f303f507f4),
    LL(0xcb0bcb16c08bddcb), LL(0x3ef83eedc67cd33e), LL(0x05140528110a2d05), LL(0x6781671fe6ce7867),
    LL(0xe4b7e47353d597e4), LL(0x279c2725bb4e0227), LL(0x4119413258827341), LL(0x8b168b2c9d0ba78b),
    LL(0xa7a6a7510153f6a7), LL(0x7de97dcf94fab27d), LL(0x956e95dcfb374995), LL(0xd847d88e9fad56d8),
    LL(0xfbcbfb8b30eb70fb), LL(0xee9fee2371c1cdee), LL(0x7ced7cc791f8bb7c), LL(0x66856617e3cc7166),
    LL(0xdd53dda68ea77bdd), LL(0x175c17b84b2eaf17), LL(0x47014702468e4547), LL(0x9e429e84dc211a9e),
    LL(0xca0fca1ec589d4ca), LL(0x2db42d75995a582d), LL(0xbfc6bf9179632ebf), LL(0x071c07381b0e3f07),
    LL(0xad8ead012347acad), LL(0x5a755aea2fb4b05a), LL(0x8336836cb51bef83), LL(0x33cc3385ff66b633),
    LL(0x6391633ff2c65c63), LL(0x020802100a041202), LL(0xaa92aa39384993aa), LL(0x71d971afa8e2de71),
    LL(0xc807c80ecf8dc6c8), LL(0x196419c87d32d119), LL(0x4939497270923b49), LL(0xd943d9869aaf5fd9),
    LL(0xf2eff2c31df931f2), LL(0xe3abe34b48dba8e3), LL(0x5b715be22ab6b95b), LL(0x881a8834920dbc88),
    LL(0x9a529aa4c8293e9a), LL(0x2698262dbe4c0b26), LL(0x32c8328dfa64bf32), LL(0xb0fab0e94a7d59b0),
    LL(0xe983e91b6acff2e9), LL(0x0f3c0f78331e770f), LL(0xd573d5e6a6b733d5), LL(0x803a8074ba1df480),
    LL(0xbec2be997c6127be), LL(0xcd13cd26de87ebcd), LL(0x34d034bde4688934), LL(0x483d487a75903248),
    LL(0xffdbffab24e354ff), LL(0x7af57af78ff48d7a), LL(0x907a90f4ea3d6490), LL(0x5f615fc23ebe9d5f),
    LL(0x2080201da0403d20), LL(0x68bd6867d5d00f68), LL(0x1a681ad07234ca1a), LL(0xae82ae192c41b7ae),
    LL(0xb4eab4c95e757db4), LL(0x544d549a19a8ce54), LL(0x937693ece53b7f93), LL(0x2288220daa442f22),
    LL(0x648d6407e9c86364), LL(0xf1e3f1db12ff2af1), LL(0x73d173bfa2e6cc73), LL(0x124812905a248212),
    LL(0x401d403a5d807a40), LL(0x0820084028104808), LL(0xc32bc356e89b95c3), LL(0xec97ec337bc5dfec),
    LL(0xdb4bdb9690ab4ddb), LL(0xa1bea1611f5fc0a1), LL(0x8d0e8d1c8307918d), LL(0x3df43df5c97ac83d),
    LL(0x976697ccf1335b97), LL(0x0000000000000000), LL(0xcf1bcf36d483f9cf), LL(0x2bac2b4587566e2b),
    LL(0x76c57697b3ece176), LL(0x82328264b019e682), LL(0xd67fd6fea9b128d6), LL(0x1b6c1bd87736c31b),
    LL(0xb5eeb5c15b7774b5), LL(0xaf86af112943beaf), LL(0x6ab56a77dfd41d6a), LL(0x505d50ba0da0ea50),
    LL(0x450945124c8a5745), LL(0xf3ebf3cb18fb38f3), LL(0x30c0309df060ad30), LL(0xef9bef2b74c3c4ef),
    LL(0x3ffc3fe5c37eda3f), LL(0x554955921caac755), LL(0xa2b2a2791059dba2), LL(0xea8fea0365c9e9ea),
    LL(0x6589650fecca6a65), LL(0xbad2bab9686903ba), LL(0x2fbc2f65935e4a2f), LL(0xc027c04ee79d8ec0),
    LL(0xde5fdebe81a160de), LL(0x1c701ce06c38fc1c), LL(0xfdd3fdbb2ee746fd), LL(0x4d294d52649a1f4d),
    LL(0x927292e4e0397692), LL(0x75c9758fbceafa75), LL(0x061806301e0c3606), LL(0x8a128a249809ae8a),
    LL(0xb2f2b2f940794bb2), LL(0xe6bfe66359d185e6), LL(0x0e380e70361c7e0e), LL(0x1f7c1ff8633ee71f),
    LL(0x62956237f7c45562), LL(0xd477d4eea3b53ad4), LL(0xa89aa829324d81a8), LL(0x966296c4f4315296),
    LL(0xf9c3f99b3aef62f9), LL(0xc533c566f697a3c5), LL(0x25942535b14a1025), LL(0x597959f220b2ab59),
    LL(0x842a8454ae15d084), LL(0x72d572b7a7e4c572), LL(0x39e439d5dd72ec39), LL(0x4c2d4c5a6198164c),
    LL(0x5e655eca3bbc945e), LL(0x78fd78e785f09f78), LL(0x38e038ddd870e538), LL(0x8c0a8c148605988c),
    LL(0xd163d1c6b2bf17d1), LL(0xa5aea5410b57e4a5), LL(0xe2afe2434dd9a1e2), LL(0x6199612ff8c24e61),
    LL(0xb3f6b3f1457b42b3), LL(0x21842115a5423421), LL(0x9c4a9c94d625089c), LL(0x1e781ef0663cee1e),
    LL(0x4311432252866143), LL(0xc73bc776fc93b1c7), LL(0xfcd7fcb32be54ffc), LL(0x0410042014082404),
    LL(0x515951b208a2e351), LL(0x995e99bcc72f2599), LL(0x6da96d4fc4da226d), LL(0x0d340d68391a650d),
    LL(0xfacffa8335e979fa), LL(0xdf5bdfb684a369df), LL(0x7ee57ed79bfca97e), LL(0x2490243db4481924),
    LL(0x3bec3bc5d776fe3b), LL(0xab96ab313d4b9aab), LL(0xce1fce3ed181f0ce), LL(0x1144118855229911),
    LL(0x8f068f0c8903838f), LL(0x4e254e4a6b9c044e), LL(0xb7e6b7d1517366b7), LL(0xeb8beb0b60cbe0eb),
    LL(0x3cf03cfdcc78c13c), LL(0x813e817cbf1ffd81), LL(0x946a94d4fe354094), LL(0xf7fbf7eb0cf31cf7),
    LL(0xb9deb9a1676f18b9), LL(0x134c13985f268b13), LL(0x2cb02c7d9c58512c), LL(0xd36bd3d6b8bb05d3),
    LL(0xe7bbe76b5cd38ce7), LL(0x6ea56e57cbdc396e), LL(0xc437c46ef395aac4), LL(0x030c03180f061b03),
    LL(0x5645568a13acdc56), LL(0x440d441a49885e44), LL(0x7fe17fdf9efea07f), LL(0xa99ea921374f88a9),
    LL(0x2aa82a4d8254672a), LL(0xbbd6bbb16d6b0abb), LL(0xc123c146e29f87c1), LL(0x535153a202a6f153),
    LL(0xdc57dcae8ba572dc), LL(0x0b2c0b582716530b), LL(0x9d4e9d9cd327019d), LL(0x6cad6c47c1d82b6c),
    LL(0x31c43195f562a431), LL(0x74cd7487b9e8f374), LL(0xf6fff6e309f115f6), LL(0x4605460a438c4c46),
    LL(0xac8aac092645a5ac), LL(0x891e893c970fb589), LL(0x145014a04428b414), LL(0xe1a3e15b42dfbae1),
    LL(0x165816b04e2ca616), LL(0x3ae83acdd274f73a), LL(0x69b9696fd0d20669), LL(0x092409482d124109),
    LL(0x70dd70a7ade0d770), LL(0xb6e2b6d954716fb6), LL(0xd067d0ceb7bd1ed0), LL(0xed93ed3b7ec7d6ed),
    LL(0xcc17cc2edb85e2cc), LL(0x4215422a57846842), LL(0x985a98b4c22d2c98), LL(0xa4aaa4490e55eda4),
    LL(0x28a0285d88507528), LL(0x5c6d5cda31b8865c), LL(0xf8c7f8933fed6bf8), LL(0x86228644a411c286),
};

#ifdef OBSOLETE
static const u64 C0[256] = {
    LL(0x1818281878c0d878), LL(0x23236523af0526af), LL(0xc6c657c6f97eb8f9), LL(0xe8e825e86f13fb6f),
    LL(0x87879487a14ccba1), LL(0xb8b8d5b862a91162), LL(0x0101030105080905), LL(0x4f4fd14f6e420d6e),
    LL(0x36365a36eead9bee), LL(0xa6a6f7a60459ff04), LL(0xd2d26bd2bdde0cbd), LL(0xf5f502f506fb0e06),
    LL(0x79798b7980ef9680), LL(0x6f6fb16fce5f30ce), LL(0x9191ae91effc6def), LL(0x5252f65207aaf807),
    LL(0x6060a060fd2747fd), LL(0xbcbcd9bc76893576), LL(0x9b9bb09bcdac37cd), LL(0x8e8e8f8e8c048a8c),
    LL(0xa3a3f8a31571d215), LL(0x0c0c140c3c606c3c), LL(0x7b7b8d7b8aff848a), LL(0x35355f35e1b580e1),
    LL(0x1d1d271d69e8f569), LL(0xe0e03de04753b347), LL(0xd7d764d7acf621ac), LL(0xc2c25bc2ed5e9ced),
    LL(0x2e2e722e966d4396), LL(0x4b4bdd4b7a62297a), LL(0xfefe1ffe21a35d21), LL(0x5757f9571682d516),
    LL(0x15153f1541a8bd41), LL(0x77779977b69fe8b6), LL(0x37375937eba592eb), LL(0xe5e532e5567b9e56),
    LL(0x9f9fbc9fd98c13d9), LL(0xf0f00df017d32317), LL(0x4a4ade4a7f6a207f), LL(0xdada73da959e4495),
    LL(0x5858e85825faa225), LL(0xc9c946c9ca06cfca), LL(0x29297b298d557c8d), LL(0x0a0a1e0a22505a22),
    LL(0xb1b1ceb14fe1504f), LL(0xa0a0fda01a69c91a), LL(0x6b6bbd6bda7f14da), LL(0x85859285ab5cd9ab),
    LL(0xbdbddabd73813c73), LL(0x5d5de75d34d28f34), LL(0x1010301050809050), LL(0xf4f401f403f30703),
    LL(0xcbcb40cbc016ddc0), LL(0x3e3e423ec6edd3c6), LL(0x05050f0511282d11), LL(0x6767a967e61f78e6),
    LL(0xe4e431e453739753), LL(0x27276927bb2502bb), LL(0x4141c34158327358), LL(0x8b8b808b9d2ca79d),
    LL(0xa7a7f4a70151f601), LL(0x7d7d877d94cfb294), LL(0x9595a295fbdc49fb), LL(0xd8d875d89f8e569f),
    LL(0xfbfb10fb308b7030), LL(0xeeee2fee7123cd71), LL(0x7c7c847c91c7bb91), LL(0x6666aa66e31771e3),
    LL(0xdddd7add8ea67b8e), LL(0x171739174bb8af4b), LL(0x4747c94746024546), LL(0x9e9ebf9edc841adc),
    LL(0xcaca43cac51ed4c5), LL(0x2d2d772d99755899), LL(0xbfbfdcbf79912e79), LL(0x070709071b383f1b),
    LL(0xadadeaad2301ac23), LL(0x5a5aee5a2feab02f), LL(0x83839883b56cefb5), LL(0x33335533ff85b6ff),
    LL(0x6363a563f23f5cf2), LL(0x020206020a10120a), LL(0xaaaae3aa38399338), LL(0x71719371a8afdea8),
    LL(0xc8c845c8cf0ec6cf), LL(0x19192b197dc8d17d), LL(0x4949db4970723b70), LL(0xd9d976d99a865f9a),
    LL(0xf2f20bf21dc3311d), LL(0xe3e338e3484ba848), LL(0x5b5bed5b2ae2b92a), LL(0x888885889234bc92),
    LL(0x9a9ab39ac8a43ec8), LL(0x26266a26be2d0bbe), LL(0x32325632fa8dbffa), LL(0xb0b0cdb04ae9594a),
    LL(0xe9e926e96a1bf26a), LL(0x0f0f110f33787733), LL(0xd5d562d5a6e633a6), LL(0x80809d80ba74f4ba),
    LL(0xbebedfbe7c99277c), LL(0xcdcd4acdde26ebde), LL(0x34345c34e4bd89e4), LL(0x4848d848757a3275),
    LL(0xffff1cff24ab5424), LL(0x7a7a8e7a8ff78d8f), LL(0x9090ad90eaf464ea), LL(0x5f5fe15f3ec29d3e),
    LL(0x20206020a01d3da0), LL(0x6868b868d5670fd5), LL(0x1a1a2e1a72d0ca72), LL(0xaeaeefae2c19b72c),
    LL(0xb4b4c1b45ec97d5e), LL(0x5454fc54199ace19), LL(0x9393a893e5ec7fe5), LL(0x22226622aa0d2faa),
    LL(0x6464ac64e90763e9), LL(0xf1f10ef112db2a12), LL(0x73739573a2bfcca2), LL(0x121236125a90825a),
    LL(0x4040c0405d3a7a5d), LL(0x0808180828404828), LL(0xc3c358c3e85695e8), LL(0xecec29ec7b33df7b),
    LL(0xdbdb70db90964d90), LL(0xa1a1fea11f61c01f), LL(0x8d8d8a8d831c9183), LL(0x3d3d473dc9f5c8c9),
    LL(0x9797a497f1cc5bf1), LL(0x0000000000000000), LL(0xcfcf4ccfd436f9d4), LL(0x2b2b7d2b87456e87),
    LL(0x76769a76b397e1b3), LL(0x82829b82b064e6b0), LL(0xd6d667d6a9fe28a9), LL(0x1b1b2d1b77d8c377),
    LL(0xb5b5c2b55bc1745b), LL(0xafafecaf2911be29), LL(0x6a6abe6adf771ddf), LL(0x5050f0500dbaea0d),
    LL(0x4545cf454c12574c), LL(0xf3f308f318cb3818), LL(0x30305030f09dadf0), LL(0xefef2cef742bc474),
    LL(0x3f3f413fc3e5dac3), LL(0x5555ff551c92c71c), LL(0xa2a2fba21079db10), LL(0xeaea23ea6503e965),
    LL(0x6565af65ec0f6aec), LL(0xbabad3ba68b90368), LL(0x2f2f712f93654a93), LL(0xc0c05dc0e74e8ee7),
    LL(0xdede7fde81be6081), LL(0x1c1c241c6ce0fc6c), LL(0xfdfd1afd2ebb462e), LL(0x4d4dd74d64521f64),
    LL(0x9292ab92e0e476e0), LL(0x75759f75bc8ffabc), LL(0x06060a061e30361e), LL(0x8a8a838a9824ae98),
    LL(0xb2b2cbb240f94b40), LL(0xe6e637e659638559), LL(0x0e0e120e36707e36), LL(0x1f1f211f63f8e763),
    LL(0x6262a662f73755f7), LL(0xd4d461d4a3ee3aa3), LL(0xa8a8e5a832298132), LL(0x9696a796f4c452f4),
    LL(0xf9f916f93a9b623a), LL(0xc5c552c5f666a3f6), LL(0x25256f25b13510b1), LL(0x5959eb5920f2ab20),
    LL(0x84849184ae54d0ae), LL(0x72729672a7b7c5a7), LL(0x39394b39ddd5ecdd), LL(0x4c4cd44c615a1661),
    LL(0x5e5ee25e3bca943b), LL(0x7878887885e79f85), LL(0x38384838d8dde5d8), LL(0x8c8c898c86149886),
    LL(0xd1d16ed1b2c617b2), LL(0xa5a5f2a50b41e40b), LL(0xe2e23be24d43a14d), LL(0x6161a361f82f4ef8),
    LL(0xb3b3c8b345f14245), LL(0x21216321a51534a5), LL(0x9c9cb99cd69408d6), LL(0x1e1e221e66f0ee66),
    LL(0x4343c54352226152), LL(0xc7c754c7fc76b1fc), LL(0xfcfc19fc2bb34f2b), LL(0x04040c0414202414),
    LL(0x5151f35108b2e308), LL(0x9999b699c7bc25c7), LL(0x6d6db76dc44f22c4), LL(0x0d0d170d39686539),
    LL(0xfafa13fa35837935), LL(0xdfdf7cdf84b66984), LL(0x7e7e827e9bd7a99b), LL(0x24246c24b43d19b4),
    LL(0x3b3b4d3bd7c5fed7), LL(0xababe0ab3d319a3d), LL(0xcece4fced13ef0d1), LL(0x1111331155889955),
    LL(0x8f8f8c8f890c8389), LL(0x4e4ed24e6b4a046b), LL(0xb7b7c4b751d16651), LL(0xebeb20eb600be060),
    LL(0x3c3c443cccfdc1cc), LL(0x81819e81bf7cfdbf), LL(0x9494a194fed440fe), LL(0xf7f704f70ceb1c0c),
    LL(0xb9b9d6b967a11867), LL(0x131335135f988b5f), LL(0x2c2c742c9c7d519c), LL(0xd3d368d3b8d605b8),
    LL(0xe7e734e75c6b8c5c), LL(0x6e6eb26ecb5739cb), LL(0xc4c451c4f36eaaf3), LL(0x030305030f181b0f),
    LL(0x5656fa56138adc13), LL(0x4444cc44491a5e49), LL(0x7f7f817f9edfa09e), LL(0xa9a9e6a937218837),
    LL(0x2a2a7e2a824d6782), LL(0xbbbbd0bb6db10a6d), LL(0xc1c15ec1e24687e2), LL(0x5353f55302a2f102),
    LL(0xdcdc79dc8bae728b), LL(0x0b0b1d0b27585327), LL(0x9d9dba9dd39c01d3), LL(0x6c6cb46cc1472bc1),
    LL(0x31315331f595a4f5), LL(0x74749c74b987f3b9), LL(0xf6f607f609e31509), LL(0x4646ca46430a4c43),
    LL(0xacace9ac2609a526), LL(0x89898689973cb597), LL(0x14143c1444a0b444), LL(0xe1e13ee1425bba42),
    LL(0x16163a164eb0a64e), LL(0x3a3a4e3ad2cdf7d2), LL(0x6969bb69d06f06d0), LL(0x09091b092d48412d),
    LL(0x70709070ada7d7ad), LL(0xb6b6c7b654d96f54), LL(0xd0d06dd0b7ce1eb7), LL(0xeded2aed7e3bd67e),
    LL(0xcccc49ccdb2ee2db), LL(0x4242c642572a6857), LL(0x9898b598c2b42cc2), LL(0xa4a4f1a40e49ed0e),
    LL(0x28287828885d7588), LL(0x5c5ce45c31da8631), LL(0xf8f815f83f936b3f), LL(0x86869786a444c2a4),
};

static const u64 C1[256] = {
    LL(0x781818281878c0d8), LL(0xaf23236523af0526), LL(0xf9c6c657c6f97eb8), LL(0x6fe8e825e86f13fb),
    LL(0xa187879487a14ccb), LL(0x62b8b8d5b862a911), LL(0x0501010301050809), LL(0x6e4f4fd14f6e420d),
    LL(0xee36365a36eead9b), LL(0x04a6a6f7a60459ff), LL(0xbdd2d26bd2bdde0c), LL(0x06f5f502f506fb0e),
    LL(0x8079798b7980ef96), LL(0xce6f6fb16fce5f30), LL(0xef9191ae91effc6d), LL(0x075252f65207aaf8),
    LL(0xfd6060a060fd2747), LL(0x76bcbcd9bc768935), LL(0xcd9b9bb09bcdac37), LL(0x8c8e8e8f8e8c048a),
    LL(0x15a3a3f8a31571d2), LL(0x3c0c0c140c3c606c), LL(0x8a7b7b8d7b8aff84), LL(0xe135355f35e1b580),
    LL(0x691d1d271d69e8f5), LL(0x47e0e03de04753b3), LL(0xacd7d764d7acf621), LL(0xedc2c25bc2ed5e9c),
    LL(0x962e2e722e966d43), LL(0x7a4b4bdd4b7a6229), LL(0x21fefe1ffe21a35d), LL(0x165757f9571682d5),
    LL(0x4115153f1541a8bd), LL(0xb677779977b69fe8), LL(0xeb37375937eba592), LL(0x56e5e532e5567b9e),
    LL(0xd99f9fbc9fd98c13), LL(0x17f0f00df017d323), LL(0x7f4a4ade4a7f6a20), LL(0x95dada73da959e44),
    LL(0x255858e85825faa2), LL(0xcac9c946c9ca06cf), LL(0x8d29297b298d557c), LL(0x220a0a1e0a22505a),
    LL(0x4fb1b1ceb14fe150), LL(0x1aa0a0fda01a69c9), LL(0xda6b6bbd6bda7f14), LL(0xab85859285ab5cd9),
    LL(0x73bdbddabd73813c), LL(0x345d5de75d34d28f), LL(0x5010103010508090), LL(0x03f4f401f403f307),
    LL(0xc0cbcb40cbc016dd), LL(0xc63e3e423ec6edd3), LL(0x1105050f0511282d), LL(0xe66767a967e61f78),
    LL(0x53e4e431e4537397), LL(0xbb27276927bb2502), LL(0x584141c341583273), LL(0x9d8b8b808b9d2ca7),
    LL(0x01a7a7f4a70151f6), LL(0x947d7d877d94cfb2), LL(0xfb9595a295fbdc49), LL(0x9fd8d875d89f8e56),
    LL(0x30fbfb10fb308b70), LL(0x71eeee2fee7123cd), LL(0x917c7c847c91c7bb), LL(0xe36666aa66e31771),
    LL(0x8edddd7add8ea67b), LL(0x4b171739174bb8af), LL(0x464747c947460245), LL(0xdc9e9ebf9edc841a),
    LL(0xc5caca43cac51ed4), LL(0x992d2d772d997558), LL(0x79bfbfdcbf79912e), LL(0x1b070709071b383f),
    LL(0x23adadeaad2301ac), LL(0x2f5a5aee5a2feab0), LL(0xb583839883b56cef), LL(0xff33335533ff85b6),
    LL(0xf26363a563f23f5c), LL(0x0a020206020a1012), LL(0x38aaaae3aa383993), LL(0xa871719371a8afde),
    LL(0xcfc8c845c8cf0ec6), LL(0x7d19192b197dc8d1), LL(0x704949db4970723b), LL(0x9ad9d976d99a865f),
    LL(0x1df2f20bf21dc331), LL(0x48e3e338e3484ba8), LL(0x2a5b5bed5b2ae2b9), LL(0x92888885889234bc),
    LL(0xc89a9ab39ac8a43e), LL(0xbe26266a26be2d0b), LL(0xfa32325632fa8dbf), LL(0x4ab0b0cdb04ae959),
    LL(0x6ae9e926e96a1bf2), LL(0x330f0f110f337877), LL(0xa6d5d562d5a6e633), LL(0xba80809d80ba74f4),
    LL(0x7cbebedfbe7c9927), LL(0xdecdcd4acdde26eb), LL(0xe434345c34e4bd89), LL(0x754848d848757a32),
    LL(0x24ffff1cff24ab54), LL(0x8f7a7a8e7a8ff78d), LL(0xea9090ad90eaf464), LL(0x3e5f5fe15f3ec29d),
    LL(0xa020206020a01d3d), LL(0xd56868b868d5670f), LL(0x721a1a2e1a72d0ca), LL(0x2caeaeefae2c19b7),
    LL(0x5eb4b4c1b45ec97d), LL(0x195454fc54199ace), LL(0xe59393a893e5ec7f), LL(0xaa22226622aa0d2f),
    LL(0xe96464ac64e90763), LL(0x12f1f10ef112db2a), LL(0xa273739573a2bfcc), LL(0x5a121236125a9082),
    LL(0x5d4040c0405d3a7a), LL(0x2808081808284048), LL(0xe8c3c358c3e85695), LL(0x7becec29ec7b33df),
    LL(0x90dbdb70db90964d), LL(0x1fa1a1fea11f61c0), LL(0x838d8d8a8d831c91), LL(0xc93d3d473dc9f5c8),
    LL(0xf19797a497f1cc5b), LL(0x0000000000000000), LL(0xd4cfcf4ccfd436f9), LL(0x872b2b7d2b87456e),
    LL(0xb376769a76b397e1), LL(0xb082829b82b064e6), LL(0xa9d6d667d6a9fe28), LL(0x771b1b2d1b77d8c3),
    LL(0x5bb5b5c2b55bc174), LL(0x29afafecaf2911be), LL(0xdf6a6abe6adf771d), LL(0x0d5050f0500dbaea),
    LL(0x4c4545cf454c1257), LL(0x18f3f308f318cb38), LL(0xf030305030f09dad), LL(0x74efef2cef742bc4),
    LL(0xc33f3f413fc3e5da), LL(0x1c5555ff551c92c7), LL(0x10a2a2fba21079db), LL(0x65eaea23ea6503e9),
    LL(0xec6565af65ec0f6a), LL(0x68babad3ba68b903), LL(0x932f2f712f93654a), LL(0xe7c0c05dc0e74e8e),
    LL(0x81dede7fde81be60), LL(0x6c1c1c241c6ce0fc), LL(0x2efdfd1afd2ebb46), LL(0x644d4dd74d64521f),
    LL(0xe09292ab92e0e476), LL(0xbc75759f75bc8ffa), LL(0x1e06060a061e3036), LL(0x988a8a838a9824ae),
    LL(0x40b2b2cbb240f94b), LL(0x59e6e637e6596385), LL(0x360e0e120e36707e), LL(0x631f1f211f63f8e7),
    LL(0xf76262a662f73755), LL(0xa3d4d461d4a3ee3a), LL(0x32a8a8e5a8322981), LL(0xf49696a796f4c452),
    LL(0x3af9f916f93a9b62), LL(0xf6c5c552c5f666a3), LL(0xb125256f25b13510), LL(0x205959eb5920f2ab),
    LL(0xae84849184ae54d0), LL(0xa772729672a7b7c5), LL(0xdd39394b39ddd5ec), LL(0x614c4cd44c615a16),
    LL(0x3b5e5ee25e3bca94), LL(0x857878887885e79f), LL(0xd838384838d8dde5), LL(0x868c8c898c861498),
    LL(0xb2d1d16ed1b2c617), LL(0x0ba5a5f2a50b41e4), LL(0x4de2e23be24d43a1), LL(0xf86161a361f82f4e),
    LL(0x45b3b3c8b345f142), LL(0xa521216321a51534), LL(0xd69c9cb99cd69408), LL(0x661e1e221e66f0ee),
    LL(0x524343c543522261), LL(0xfcc7c754c7fc76b1), LL(0x2bfcfc19fc2bb34f), LL(0x1404040c04142024),
    LL(0x085151f35108b2e3), LL(0xc79999b699c7bc25), LL(0xc46d6db76dc44f22), LL(0x390d0d170d396865),
    LL(0x35fafa13fa358379), LL(0x84dfdf7cdf84b669), LL(0x9b7e7e827e9bd7a9), LL(0xb424246c24b43d19),
    LL(0xd73b3b4d3bd7c5fe), LL(0x3dababe0ab3d319a), LL(0xd1cece4fced13ef0), LL(0x5511113311558899),
    LL(0x898f8f8c8f890c83), LL(0x6b4e4ed24e6b4a04), LL(0x51b7b7c4b751d166), LL(0x60ebeb20eb600be0),
    LL(0xcc3c3c443cccfdc1), LL(0xbf81819e81bf7cfd), LL(0xfe9494a194fed440), LL(0x0cf7f704f70ceb1c),
    LL(0x67b9b9d6b967a118), LL(0x5f131335135f988b), LL(0x9c2c2c742c9c7d51), LL(0xb8d3d368d3b8d605),
    LL(0x5ce7e734e75c6b8c), LL(0xcb6e6eb26ecb5739), LL(0xf3c4c451c4f36eaa), LL(0x0f030305030f181b),
    LL(0x135656fa56138adc), LL(0x494444cc44491a5e), LL(0x9e7f7f817f9edfa0), LL(0x37a9a9e6a9372188),
    LL(0x822a2a7e2a824d67), LL(0x6dbbbbd0bb6db10a), LL(0xe2c1c15ec1e24687), LL(0x025353f55302a2f1),
    LL(0x8bdcdc79dc8bae72), LL(0x270b0b1d0b275853), LL(0xd39d9dba9dd39c01), LL(0xc16c6cb46cc1472b),
    LL(0xf531315331f595a4), LL(0xb974749c74b987f3), LL(0x09f6f607f609e315), LL(0x434646ca46430a4c),
    LL(0x26acace9ac2609a5), LL(0x9789898689973cb5), LL(0x4414143c1444a0b4), LL(0x42e1e13ee1425bba),
    LL(0x4e16163a164eb0a6), LL(0xd23a3a4e3ad2cdf7), LL(0xd06969bb69d06f06), LL(0x2d09091b092d4841),
    LL(0xad70709070ada7d7), LL(0x54b6b6c7b654d96f), LL(0xb7d0d06dd0b7ce1e), LL(0x7eeded2aed7e3bd6),
    LL(0xdbcccc49ccdb2ee2), LL(0x574242c642572a68), LL(0xc29898b598c2b42c), LL(0x0ea4a4f1a40e49ed),
    LL(0x8828287828885d75), LL(0x315c5ce45c31da86), LL(0x3ff8f815f83f936b), LL(0xa486869786a444c2),
};

static const u64 C2[256] = {
    LL(0xd8781818281878c0), LL(0x26af23236523af05), LL(0xb8f9c6c657c6f97e), LL(0xfb6fe8e825e86f13),
    LL(0xcba187879487a14c), LL(0x1162b8b8d5b862a9), LL(0x0905010103010508), LL(0x0d6e4f4fd14f6e42),
    LL(0x9bee36365a36eead), LL(0xff04a6a6f7a60459), LL(0x0cbdd2d26bd2bdde), LL(0x0e06f5f502f506fb),
    LL(0x968079798b7980ef), LL(0x30ce6f6fb16fce5f), LL(0x6def9191ae91effc), LL(0xf8075252f65207aa),
    LL(0x47fd6060a060fd27), LL(0x3576bcbcd9bc7689), LL(0x37cd9b9bb09bcdac), LL(0x8a8c8e8e8f8e8c04),
    LL(0xd215a3a3f8a31571), LL(0x6c3c0c0c140c3c60), LL(0x848a7b7b8d7b8aff), LL(0x80e135355f35e1b5),
    LL(0xf5691d1d271d69e8), LL(0xb347e0e03de04753), LL(0x21acd7d764d7acf6), LL(0x9cedc2c25bc2ed5e),
    LL(0x43962e2e722e966d), LL(0x297a4b4bdd4b7a62), LL(0x5d21fefe1ffe21a3), LL(0xd5165757f9571682),
    LL(0xbd4115153f1541a8), LL(0xe8b677779977b69f), LL(0x92eb37375937eba5), LL(0x9e56e5e532e5567b),
    LL(0x13d99f9fbc9fd98c), LL(0x2317f0f00df017d3), LL(0x207f4a4ade4a7f6a), LL(0x4495dada73da959e),
    LL(0xa2255858e85825fa), LL(0xcfcac9c946c9ca06), LL(0x7c8d29297b298d55), LL(0x5a220a0a1e0a2250),
    LL(0x504fb1b1ceb14fe1), LL(0xc91aa0a0fda01a69), LL(0x14da6b6bbd6bda7f), LL(0xd9ab85859285ab5c),
    LL(0x3c73bdbddabd7381), LL(0x8f345d5de75d34d2), LL(0x9050101030105080), LL(0x0703f4f401f403f3),
    LL(0xddc0cbcb40cbc016), LL(0xd3c63e3e423ec6ed), LL(0x2d1105050f051128), LL(0x78e66767a967e61f),
    LL(0x9753e4e431e45373), LL(0x02bb27276927bb25), LL(0x73584141c3415832), LL(0xa79d8b8b808b9d2c),
    LL(0xf601a7a7f4a70151), LL(0xb2947d7d877d94cf), LL(0x49fb9595a295fbdc), LL(0x569fd8d875d89f8e),
    LL(0x7030fbfb10fb308b), LL(0xcd71eeee2fee7123), LL(0xbb917c7c847c91c7), LL(0x71e36666aa66e317),
    LL(0x7b8edddd7add8ea6), LL(0xaf4b171739174bb8), LL(0x45464747c9474602), LL(0x1adc9e9ebf9edc84),
    LL(0xd4c5caca43cac51e), LL(0x58992d2d772d9975), LL(0x2e79bfbfdcbf7991), LL(0x3f1b070709071b38),
    LL(0xac23adadeaad2301), LL(0xb02f5a5aee5a2fea), LL(0xefb583839883b56c), LL(0xb6ff33335533ff85),
    LL(0x5cf26363a563f23f), LL(0x120a020206020a10), LL(0x9338aaaae3aa3839), LL(0xdea871719371a8af),
    LL(0xc6cfc8c845c8cf0e), LL(0xd17d19192b197dc8), LL(0x3b704949db497072), LL(0x5f9ad9d976d99a86),
    LL(0x311df2f20bf21dc3), LL(0xa848e3e338e3484b), LL(0xb92a5b5bed5b2ae2), LL(0xbc92888885889234),
    LL(0x3ec89a9ab39ac8a4), LL(0x0bbe26266a26be2d), LL(0xbffa32325632fa8d), LL(0x594ab0b0cdb04ae9),
    LL(0xf26ae9e926e96a1b), LL(0x77330f0f110f3378), LL(0x33a6d5d562d5a6e6), LL(0xf4ba80809d80ba74),
    LL(0x277cbebedfbe7c99), LL(0xebdecdcd4acdde26), LL(0x89e434345c34e4bd), LL(0x32754848d848757a),
    LL(0x5424ffff1cff24ab), LL(0x8d8f7a7a8e7a8ff7), LL(0x64ea9090ad90eaf4), LL(0x9d3e5f5fe15f3ec2),
    LL(0x3da020206020a01d), LL(0x0fd56868b868d567), LL(0xca721a1a2e1a72d0), LL(0xb72caeaeefae2c19),
    LL(0x7d5eb4b4c1b45ec9), LL(0xce195454fc54199a), LL(0x7fe59393a893e5ec), LL(0x2faa22226622aa0d),
    LL(0x63e96464ac64e907), LL(0x2a12f1f10ef112db), LL(0xcca273739573a2bf), LL(0x825a121236125a90),
    LL(0x7a5d4040c0405d3a), LL(0x4828080818082840), LL(0x95e8c3c358c3e856), LL(0xdf7becec29ec7b33),
    LL(0x4d90dbdb70db9096), LL(0xc01fa1a1fea11f61), LL(0x91838d8d8a8d831c), LL(0xc8c93d3d473dc9f5),
    LL(0x5bf19797a497f1cc), LL(0x0000000000000000), LL(0xf9d4cfcf4ccfd436), LL(0x6e872b2b7d2b8745),
    LL(0xe1b376769a76b397), LL(0xe6b082829b82b064), LL(0x28a9d6d667d6a9fe), LL(0xc3771b1b2d1b77d8),
    LL(0x745bb5b5c2b55bc1), LL(0xbe29afafecaf2911), LL(0x1ddf6a6abe6adf77), LL(0xea0d5050f0500dba),
    LL(0x574c4545cf454c12), LL(0x3818f3f308f318cb), LL(0xadf030305030f09d), LL(0xc474efef2cef742b),
    LL(0xdac33f3f413fc3e5), LL(0xc71c5555ff551c92), LL(0xdb10a2a2fba21079), LL(0xe965eaea23ea6503),
    LL(0x6aec6565af65ec0f), LL(0x0368babad3ba68b9), LL(0x4a932f2f712f9365), LL(0x8ee7c0c05dc0e74e),
    LL(0x6081dede7fde81be), LL(0xfc6c1c1c241c6ce0), LL(0x462efdfd1afd2ebb), LL(0x1f644d4dd74d6452),
    LL(0x76e09292ab92e0e4), LL(0xfabc75759f75bc8f), LL(0x361e06060a061e30), LL(0xae988a8a838a9824),
    LL(0x4b40b2b2cbb240f9), LL(0x8559e6e637e65963), LL(0x7e360e0e120e3670), LL(0xe7631f1f211f63f8),
    LL(0x55f76262a662f737), LL(0x3aa3d4d461d4a3ee), LL(0x8132a8a8e5a83229), LL(0x52f49696a796f4c4),
    LL(0x623af9f916f93a9b), LL(0xa3f6c5c552c5f666), LL(0x10b125256f25b135), LL(0xab205959eb5920f2),
    LL(0xd0ae84849184ae54), LL(0xc5a772729672a7b7), LL(0xecdd39394b39ddd5), LL(0x16614c4cd44c615a),
    LL(0x943b5e5ee25e3bca), LL(0x9f857878887885e7), LL(0xe5d838384838d8dd), LL(0x98868c8c898c8614),
    LL(0x17b2d1d16ed1b2c6), LL(0xe40ba5a5f2a50b41), LL(0xa14de2e23be24d43), LL(0x4ef86161a361f82f),
    LL(0x4245b3b3c8b345f1), LL(0x34a521216321a515), LL(0x08d69c9cb99cd694), LL(0xee661e1e221e66f0),
    LL(0x61524343c5435222), LL(0xb1fcc7c754c7fc76), LL(0x4f2bfcfc19fc2bb3), LL(0x241404040c041420),
    LL(0xe3085151f35108b2), LL(0x25c79999b699c7bc), LL(0x22c46d6db76dc44f), LL(0x65390d0d170d3968),
    LL(0x7935fafa13fa3583), LL(0x6984dfdf7cdf84b6), LL(0xa99b7e7e827e9bd7), LL(0x19b424246c24b43d),
    LL(0xfed73b3b4d3bd7c5), LL(0x9a3dababe0ab3d31), LL(0xf0d1cece4fced13e), LL(0x9955111133115588),
    LL(0x83898f8f8c8f890c), LL(0x046b4e4ed24e6b4a), LL(0x6651b7b7c4b751d1), LL(0xe060ebeb20eb600b),
    LL(0xc1cc3c3c443cccfd), LL(0xfdbf81819e81bf7c), LL(0x40fe9494a194fed4), LL(0x1c0cf7f704f70ceb),
    LL(0x1867b9b9d6b967a1), LL(0x8b5f131335135f98), LL(0x519c2c2c742c9c7d), LL(0x05b8d3d368d3b8d6),
    LL(0x8c5ce7e734e75c6b), LL(0x39cb6e6eb26ecb57), LL(0xaaf3c4c451c4f36e), LL(0x1b0f030305030f18),
    LL(0xdc135656fa56138a), LL(0x5e494444cc44491a), LL(0xa09e7f7f817f9edf), LL(0x8837a9a9e6a93721),
    LL(0x67822a2a7e2a824d), LL(0x0a6dbbbbd0bb6db1), LL(0x87e2c1c15ec1e246), LL(0xf1025353f55302a2),
    LL(0x728bdcdc79dc8bae), LL(0x53270b0b1d0b2758), LL(0x01d39d9dba9dd39c), LL(0x2bc16c6cb46cc147),
    LL(0xa4f531315331f595), LL(0xf3b974749c74b987), LL(0x1509f6f607f609e3), LL(0x4c434646ca46430a),
    LL(0xa526acace9ac2609), LL(0xb59789898689973c), LL(0xb44414143c1444a0), LL(0xba42e1e13ee1425b),
    LL(0xa64e16163a164eb0), LL(0xf7d23a3a4e3ad2cd), LL(0x06d06969bb69d06f), LL(0x412d09091b092d48),
    LL(0xd7ad70709070ada7), LL(0x6f54b6b6c7b654d9), LL(0x1eb7d0d06dd0b7ce), LL(0xd67eeded2aed7e3b),
    LL(0xe2dbcccc49ccdb2e), LL(0x68574242c642572a), LL(0x2cc29898b598c2b4), LL(0xed0ea4a4f1a40e49),
    LL(0x758828287828885d), LL(0x86315c5ce45c31da), LL(0x6b3ff8f815f83f93), LL(0xc2a486869786a444),
};

static const u64 C3[256] = {
    LL(0xc0d8781818281878), LL(0x0526af23236523af), LL(0x7eb8f9c6c657c6f9), LL(0x13fb6fe8e825e86f),
    LL(0x4ccba187879487a1), LL(0xa91162b8b8d5b862), LL(0x0809050101030105), LL(0x420d6e4f4fd14f6e),
    LL(0xad9bee36365a36ee), LL(0x59ff04a6a6f7a604), LL(0xde0cbdd2d26bd2bd), LL(0xfb0e06f5f502f506),
    LL(0xef968079798b7980), LL(0x5f30ce6f6fb16fce), LL(0xfc6def9191ae91ef), LL(0xaaf8075252f65207),
    LL(0x2747fd6060a060fd), LL(0x893576bcbcd9bc76), LL(0xac37cd9b9bb09bcd), LL(0x048a8c8e8e8f8e8c),
    LL(0x71d215a3a3f8a315), LL(0x606c3c0c0c140c3c), LL(0xff848a7b7b8d7b8a), LL(0xb580e135355f35e1),
    LL(0xe8f5691d1d271d69), LL(0x53b347e0e03de047), LL(0xf621acd7d764d7ac), LL(0x5e9cedc2c25bc2ed),
    LL(0x6d43962e2e722e96), LL(0x62297a4b4bdd4b7a), LL(0xa35d21fefe1ffe21), LL(0x82d5165757f95716),
    LL(0xa8bd4115153f1541), LL(0x9fe8b677779977b6), LL(0xa592eb37375937eb), LL(0x7b9e56e5e532e556),
    LL(0x8c13d99f9fbc9fd9), LL(0xd32317f0f00df017), LL(0x6a207f4a4ade4a7f), LL(0x9e4495dada73da95),
    LL(0xfaa2255858e85825), LL(0x06cfcac9c946c9ca), LL(0x557c8d29297b298d), LL(0x505a220a0a1e0a22),
    LL(0xe1504fb1b1ceb14f), LL(0x69c91aa0a0fda01a), LL(0x7f14da6b6bbd6bda), LL(0x5cd9ab85859285ab),
    LL(0x813c73bdbddabd73), LL(0xd28f345d5de75d34), LL(0x8090501010301050), LL(0xf30703f4f401f403),
    LL(0x16ddc0cbcb40cbc0), LL(0xedd3c63e3e423ec6), LL(0x282d1105050f0511), LL(0x1f78e66767a967e6),
    LL(0x739753e4e431e453), LL(0x2502bb27276927bb), LL(0x3273584141c34158), LL(0x2ca79d8b8b808b9d),
    LL(0x51f601a7a7f4a701), LL(0xcfb2947d7d877d94), LL(0xdc49fb9595a295fb), LL(0x8e569fd8d875d89f),
    LL(0x8b7030fbfb10fb30), LL(0x23cd71eeee2fee71), LL(0xc7bb917c7c847c91), LL(0x1771e36666aa66e3),
    LL(0xa67b8edddd7add8e), LL(0xb8af4b171739174b), LL(0x0245464747c94746), LL(0x841adc9e9ebf9edc),
    LL(0x1ed4c5caca43cac5), LL(0x7558992d2d772d99), LL(0x912e79bfbfdcbf79), LL(0x383f1b070709071b),
    LL(0x01ac23adadeaad23), LL(0xeab02f5a5aee5a2f), LL(0x6cefb583839883b5), LL(0x85b6ff33335533ff),
    LL(0x3f5cf26363a563f2), LL(0x10120a020206020a), LL(0x399338aaaae3aa38), LL(0xafdea871719371a8),
    LL(0x0ec6cfc8c845c8cf), LL(0xc8d17d19192b197d), LL(0x723b704949db4970), LL(0x865f9ad9d976d99a),
    LL(0xc3311df2f20bf21d), LL(0x4ba848e3e338e348), LL(0xe2b92a5b5bed5b2a), LL(0x34bc928888858892),
    LL(0xa43ec89a9ab39ac8), LL(0x2d0bbe26266a26be), LL(0x8dbffa32325632fa), LL(0xe9594ab0b0cdb04a),
    LL(0x1bf26ae9e926e96a), LL(0x7877330f0f110f33), LL(0xe633a6d5d562d5a6), LL(0x74f4ba80809d80ba),
    LL(0x99277cbebedfbe7c), LL(0x26ebdecdcd4acdde), LL(0xbd89e434345c34e4), LL(0x7a32754848d84875),
    LL(0xab5424ffff1cff24), LL(0xf78d8f7a7a8e7a8f), LL(0xf464ea9090ad90ea), LL(0xc29d3e5f5fe15f3e),
    LL(0x1d3da020206020a0), LL(0x670fd56868b868d5), LL(0xd0ca721a1a2e1a72), LL(0x19b72caeaeefae2c),
    LL(0xc97d5eb4b4c1b45e), LL(0x9ace195454fc5419), LL(0xec7fe59393a893e5), LL(0x0d2faa22226622aa),
    LL(0x0763e96464ac64e9), LL(0xdb2a12f1f10ef112), LL(0xbfcca273739573a2), LL(0x90825a121236125a),
    LL(0x3a7a5d4040c0405d), LL(0x4048280808180828), LL(0x5695e8c3c358c3e8), LL(0x33df7becec29ec7b),
    LL(0x964d90dbdb70db90), LL(0x61c01fa1a1fea11f), LL(0x1c91838d8d8a8d83), LL(0xf5c8c93d3d473dc9),
    LL(0xcc5bf19797a497f1), LL(0x0000000000000000), LL(0x36f9d4cfcf4ccfd4), LL(0x456e872b2b7d2b87),
    LL(0x97e1b376769a76b3), LL(0x64e6b082829b82b0), LL(0xfe28a9d6d667d6a9), LL(0xd8c3771b1b2d1b77),
    LL(0xc1745bb5b5c2b55b), LL(0x11be29afafecaf29), LL(0x771ddf6a6abe6adf), LL(0xbaea0d5050f0500d),
    LL(0x12574c4545cf454c), LL(0xcb3818f3f308f318), LL(0x9dadf030305030f0), LL(0x2bc474efef2cef74),
    LL(0xe5dac33f3f413fc3), LL(0x92c71c5555ff551c), LL(0x79db10a2a2fba210), LL(0x03e965eaea23ea65),
    LL(0x0f6aec6565af65ec), LL(0xb90368babad3ba68), LL(0x654a932f2f712f93), LL(0x4e8ee7c0c05dc0e7),
    LL(0xbe6081dede7fde81), LL(0xe0fc6c1c1c241c6c), LL(0xbb462efdfd1afd2e), LL(0x521f644d4dd74d64),
    LL(0xe476e09292ab92e0), LL(0x8ffabc75759f75bc), LL(0x30361e06060a061e), LL(0x24ae988a8a838a98),
    LL(0xf94b40b2b2cbb240), LL(0x638559e6e637e659), LL(0x707e360e0e120e36), LL(0xf8e7631f1f211f63),
    LL(0x3755f76262a662f7), LL(0xee3aa3d4d461d4a3), LL(0x298132a8a8e5a832), LL(0xc452f49696a796f4),
    LL(0x9b623af9f916f93a), LL(0x66a3f6c5c552c5f6), LL(0x3510b125256f25b1), LL(0xf2ab205959eb5920),
    LL(0x54d0ae84849184ae), LL(0xb7c5a772729672a7), LL(0xd5ecdd39394b39dd), LL(0x5a16614c4cd44c61),
    LL(0xca943b5e5ee25e3b), LL(0xe79f857878887885), LL(0xdde5d838384838d8), LL(0x1498868c8c898c86),
    LL(0xc617b2d1d16ed1b2), LL(0x41e40ba5a5f2a50b), LL(0x43a14de2e23be24d), LL(0x2f4ef86161a361f8),
    LL(0xf14245b3b3c8b345), LL(0x1534a521216321a5), LL(0x9408d69c9cb99cd6), LL(0xf0ee661e1e221e66),
    LL(0x2261524343c54352), LL(0x76b1fcc7c754c7fc), LL(0xb34f2bfcfc19fc2b), LL(0x20241404040c0414),
    LL(0xb2e3085151f35108), LL(0xbc25c79999b699c7), LL(0x4f22c46d6db76dc4), LL(0x6865390d0d170d39),
    LL(0x837935fafa13fa35), LL(0xb66984dfdf7cdf84), LL(0xd7a99b7e7e827e9b), LL(0x3d19b424246c24b4),
    LL(0xc5fed73b3b4d3bd7), LL(0x319a3dababe0ab3d), LL(0x3ef0d1cece4fced1), LL(0x8899551111331155),
    LL(0x0c83898f8f8c8f89), LL(0x4a046b4e4ed24e6b), LL(0xd16651b7b7c4b751), LL(0x0be060ebeb20eb60),
    LL(0xfdc1cc3c3c443ccc), LL(0x7cfdbf81819e81bf), LL(0xd440fe9494a194fe), LL(0xeb1c0cf7f704f70c),
    LL(0xa11867b9b9d6b967), LL(0x988b5f131335135f), LL(0x7d519c2c2c742c9c), LL(0xd605b8d3d368d3b8),
    LL(0x6b8c5ce7e734e75c), LL(0x5739cb6e6eb26ecb), LL(0x6eaaf3c4c451c4f3), LL(0x181b0f030305030f),
    LL(0x8adc135656fa5613), LL(0x1a5e494444cc4449), LL(0xdfa09e7f7f817f9e), LL(0x218837a9a9e6a937),
    LL(0x4d67822a2a7e2a82), LL(0xb10a6dbbbbd0bb6d), LL(0x4687e2c1c15ec1e2), LL(0xa2f1025353f55302),
    LL(0xae728bdcdc79dc8b), LL(0x5853270b0b1d0b27), LL(0x9c01d39d9dba9dd3), LL(0x472bc16c6cb46cc1),
    LL(0x95a4f531315331f5), LL(0x87f3b974749c74b9), LL(0xe31509f6f607f609), LL(0x0a4c434646ca4643),
    LL(0x09a526acace9ac26), LL(0x3cb5978989868997), LL(0xa0b44414143c1444), LL(0x5bba42e1e13ee142),
    LL(0xb0a64e16163a164e), LL(0xcdf7d23a3a4e3ad2), LL(0x6f06d06969bb69d0), LL(0x48412d09091b092d),
    LL(0xa7d7ad70709070ad), LL(0xd96f54b6b6c7b654), LL(0xce1eb7d0d06dd0b7), LL(0x3bd67eeded2aed7e),
    LL(0x2ee2dbcccc49ccdb), LL(0x2a68574242c64257), LL(0xb42cc29898b598c2), LL(0x49ed0ea4a4f1a40e),
    LL(0x5d75882828782888), LL(0xda86315c5ce45c31), LL(0x936b3ff8f815f83f), LL(0x44c2a486869786a4),
};

static const u64 C4[256] = {
    LL(0x78c0d87818182818), LL(0xaf0526af23236523), LL(0xf97eb8f9c6c657c6), LL(0x6f13fb6fe8e825e8),
    LL(0xa14ccba187879487), LL(0x62a91162b8b8d5b8), LL(0x0508090501010301), LL(0x6e420d6e4f4fd14f),
    LL(0xeead9bee36365a36), LL(0x0459ff04a6a6f7a6), LL(0xbdde0cbdd2d26bd2), LL(0x06fb0e06f5f502f5),
    LL(0x80ef968079798b79), LL(0xce5f30ce6f6fb16f), LL(0xeffc6def9191ae91), LL(0x07aaf8075252f652),
    LL(0xfd2747fd6060a060), LL(0x76893576bcbcd9bc), LL(0xcdac37cd9b9bb09b), LL(0x8c048a8c8e8e8f8e),
    LL(0x1571d215a3a3f8a3), LL(0x3c606c3c0c0c140c), LL(0x8aff848a7b7b8d7b), LL(0xe1b580e135355f35),
    LL(0x69e8f5691d1d271d), LL(0x4753b347e0e03de0), LL(0xacf621acd7d764d7), LL(0xed5e9cedc2c25bc2),
    LL(0x966d43962e2e722e), LL(0x7a62297a4b4bdd4b), LL(0x21a35d21fefe1ffe), LL(0x1682d5165757f957),
    LL(0x41a8bd4115153f15), LL(0xb69fe8b677779977), LL(0xeba592eb37375937), LL(0x567b9e56e5e532e5),
    LL(0xd98c13d99f9fbc9f), LL(0x17d32317f0f00df0), LL(0x7f6a207f4a4ade4a), LL(0x959e4495dada73da),
    LL(0x25faa2255858e858), LL(0xca06cfcac9c946c9), LL(0x8d557c8d29297b29), LL(0x22505a220a0a1e0a),
    LL(0x4fe1504fb1b1ceb1), LL(0x1a69c91aa0a0fda0), LL(0xda7f14da6b6bbd6b), LL(0xab5cd9ab85859285),
    LL(0x73813c73bdbddabd), LL(0x34d28f345d5de75d), LL(0x5080905010103010), LL(0x03f30703f4f401f4),
    LL(0xc016ddc0cbcb40cb), LL(0xc6edd3c63e3e423e), LL(0x11282d1105050f05), LL(0xe61f78e66767a967),
    LL(0x53739753e4e431e4), LL(0xbb2502bb27276927), LL(0x583273584141c341), LL(0x9d2ca79d8b8b808b),
    LL(0x0151f601a7a7f4a7), LL(0x94cfb2947d7d877d), LL(0xfbdc49fb9595a295), LL(0x9f8e569fd8d875d8),
    LL(0x308b7030fbfb10fb), LL(0x7123cd71eeee2fee), LL(0x91c7bb917c7c847c), LL(0xe31771e36666aa66),
    LL(0x8ea67b8edddd7add), LL(0x4bb8af4b17173917), LL(0x460245464747c947), LL(0xdc841adc9e9ebf9e),
    LL(0xc51ed4c5caca43ca), LL(0x997558992d2d772d), LL(0x79912e79bfbfdcbf), LL(0x1b383f1b07070907),
    LL(0x2301ac23adadeaad), LL(0x2feab02f5a5aee5a), LL(0xb56cefb583839883), LL(0xff85b6ff33335533),
    LL(0xf23f5cf26363a563), LL(0x0a10120a02020602), LL(0x38399338aaaae3aa), LL(0xa8afdea871719371),
    LL(0xcf0ec6cfc8c845c8), LL(0x7dc8d17d19192b19), LL(0x70723b704949db49), LL(0x9a865f9ad9d976d9),
    LL(0x1dc3311df2f20bf2), LL(0x484ba848e3e338e3), LL(0x2ae2b92a5b5bed5b), LL(0x9234bc9288888588),
    LL(0xc8a43ec89a9ab39a), LL(0xbe2d0bbe26266a26), LL(0xfa8dbffa32325632), LL(0x4ae9594ab0b0cdb0),
    LL(0x6a1bf26ae9e926e9), LL(0x337877330f0f110f), LL(0xa6e633a6d5d562d5), LL(0xba74f4ba80809d80),
    LL(0x7c99277cbebedfbe), LL(0xde26ebdecdcd4acd), LL(0xe4bd89e434345c34), LL(0x757a32754848d848),
    LL(0x24ab5424ffff1cff), LL(0x8ff78d8f7a7a8e7a), LL(0xeaf464ea9090ad90), LL(0x3ec29d3e5f5fe15f),
    LL(0xa01d3da020206020), LL(0xd5670fd56868b868), LL(0x72d0ca721a1a2e1a), LL(0x2c19b72caeaeefae),
    LL(0x5ec97d5eb4b4c1b4), LL(0x199ace195454fc54), LL(0xe5ec7fe59393a893), LL(0xaa0d2faa22226622),
    LL(0xe90763e96464ac64), LL(0x12db2a12f1f10ef1), LL(0xa2bfcca273739573), LL(0x5a90825a12123612),
    LL(0x5d3a7a5d4040c040), LL(0x2840482808081808), LL(0xe85695e8c3c358c3), LL(0x7b33df7becec29ec),
    LL(0x90964d90dbdb70db), LL(0x1f61c01fa1a1fea1), LL(0x831c91838d8d8a8d), LL(0xc9f5c8c93d3d473d),
    LL(0xf1cc5bf19797a497), LL(0x0000000000000000), LL(0xd436f9d4cfcf4ccf), LL(0x87456e872b2b7d2b),
    LL(0xb397e1b376769a76), LL(0xb064e6b082829b82), LL(0xa9fe28a9d6d667d6), LL(0x77d8c3771b1b2d1b),
    LL(0x5bc1745bb5b5c2b5), LL(0x2911be29afafecaf), LL(0xdf771ddf6a6abe6a), LL(0x0dbaea0d5050f050),
    LL(0x4c12574c4545cf45), LL(0x18cb3818f3f308f3), LL(0xf09dadf030305030), LL(0x742bc474efef2cef),
    LL(0xc3e5dac33f3f413f), LL(0x1c92c71c5555ff55), LL(0x1079db10a2a2fba2), LL(0x6503e965eaea23ea),
    LL(0xec0f6aec6565af65), LL(0x68b90368babad3ba), LL(0x93654a932f2f712f), LL(0xe74e8ee7c0c05dc0),
    LL(0x81be6081dede7fde), LL(0x6ce0fc6c1c1c241c), LL(0x2ebb462efdfd1afd), LL(0x64521f644d4dd74d),
    LL(0xe0e476e09292ab92), LL(0xbc8ffabc75759f75), LL(0x1e30361e06060a06), LL(0x9824ae988a8a838a),
    LL(0x40f94b40b2b2cbb2), LL(0x59638559e6e637e6), LL(0x36707e360e0e120e), LL(0x63f8e7631f1f211f),
    LL(0xf73755f76262a662), LL(0xa3ee3aa3d4d461d4), LL(0x32298132a8a8e5a8), LL(0xf4c452f49696a796),
    LL(0x3a9b623af9f916f9), LL(0xf666a3f6c5c552c5), LL(0xb13510b125256f25), LL(0x20f2ab205959eb59),
    LL(0xae54d0ae84849184), LL(0xa7b7c5a772729672), LL(0xddd5ecdd39394b39), LL(0x615a16614c4cd44c),
    LL(0x3bca943b5e5ee25e), LL(0x85e79f8578788878), LL(0xd8dde5d838384838), LL(0x861498868c8c898c),
    LL(0xb2c617b2d1d16ed1), LL(0x0b41e40ba5a5f2a5), LL(0x4d43a14de2e23be2), LL(0xf82f4ef86161a361),
    LL(0x45f14245b3b3c8b3), LL(0xa51534a521216321), LL(0xd69408d69c9cb99c), LL(0x66f0ee661e1e221e),
    LL(0x522261524343c543), LL(0xfc76b1fcc7c754c7), LL(0x2bb34f2bfcfc19fc), LL(0x1420241404040c04),
    LL(0x08b2e3085151f351), LL(0xc7bc25c79999b699), LL(0xc44f22c46d6db76d), LL(0x396865390d0d170d),
    LL(0x35837935fafa13fa), LL(0x84b66984dfdf7cdf), LL(0x9bd7a99b7e7e827e), LL(0xb43d19b424246c24),
    LL(0xd7c5fed73b3b4d3b), LL(0x3d319a3dababe0ab), LL(0xd13ef0d1cece4fce), LL(0x5588995511113311),
    LL(0x890c83898f8f8c8f), LL(0x6b4a046b4e4ed24e), LL(0x51d16651b7b7c4b7), LL(0x600be060ebeb20eb),
    LL(0xccfdc1cc3c3c443c), LL(0xbf7cfdbf81819e81), LL(0xfed440fe9494a194), LL(0x0ceb1c0cf7f704f7),
    LL(0x67a11867b9b9d6b9), LL(0x5f988b5f13133513), LL(0x9c7d519c2c2c742c), LL(0xb8d605b8d3d368d3),
    LL(0x5c6b8c5ce7e734e7), LL(0xcb5739cb6e6eb26e), LL(0xf36eaaf3c4c451c4), LL(0x0f181b0f03030503),
    LL(0x138adc135656fa56), LL(0x491a5e494444cc44), LL(0x9edfa09e7f7f817f), LL(0x37218837a9a9e6a9),
    LL(0x824d67822a2a7e2a), LL(0x6db10a6dbbbbd0bb), LL(0xe24687e2c1c15ec1), LL(0x02a2f1025353f553),
    LL(0x8bae728bdcdc79dc), LL(0x275853270b0b1d0b), LL(0xd39c01d39d9dba9d), LL(0xc1472bc16c6cb46c),
    LL(0xf595a4f531315331), LL(0xb987f3b974749c74), LL(0x09e31509f6f607f6), LL(0x430a4c434646ca46),
    LL(0x2609a526acace9ac), LL(0x973cb59789898689), LL(0x44a0b44414143c14), LL(0x425bba42e1e13ee1),
    LL(0x4eb0a64e16163a16), LL(0xd2cdf7d23a3a4e3a), LL(0xd06f06d06969bb69), LL(0x2d48412d09091b09),
    LL(0xada7d7ad70709070), LL(0x54d96f54b6b6c7b6), LL(0xb7ce1eb7d0d06dd0), LL(0x7e3bd67eeded2aed),
    LL(0xdb2ee2dbcccc49cc), LL(0x572a68574242c642), LL(0xc2b42cc29898b598), LL(0x0e49ed0ea4a4f1a4),
    LL(0x885d758828287828), LL(0x31da86315c5ce45c), LL(0x3f936b3ff8f815f8), LL(0xa444c2a486869786),
};

static const u64 C5[256] = {
    LL(0x1878c0d878181828), LL(0x23af0526af232365), LL(0xc6f97eb8f9c6c657), LL(0xe86f13fb6fe8e825),
    LL(0x87a14ccba1878794), LL(0xb862a91162b8b8d5), LL(0x0105080905010103), LL(0x4f6e420d6e4f4fd1),
    LL(0x36eead9bee36365a), LL(0xa60459ff04a6a6f7), LL(0xd2bdde0cbdd2d26b), LL(0xf506fb0e06f5f502),
    LL(0x7980ef968079798b), LL(0x6fce5f30ce6f6fb1), LL(0x91effc6def9191ae), LL(0x5207aaf8075252f6),
    LL(0x60fd2747fd6060a0), LL(0xbc76893576bcbcd9), LL(0x9bcdac37cd9b9bb0), LL(0x8e8c048a8c8e8e8f),
    LL(0xa31571d215a3a3f8), LL(0x0c3c606c3c0c0c14), LL(0x7b8aff848a7b7b8d), LL(0x35e1b580e135355f),
    LL(0x1d69e8f5691d1d27), LL(0xe04753b347e0e03d), LL(0xd7acf621acd7d764), LL(0xc2ed5e9cedc2c25b),
    LL(0x2e966d43962e2e72), LL(0x4b7a62297a4b4bdd), LL(0xfe21a35d21fefe1f), LL(0x571682d5165757f9),
    LL(0x1541a8bd4115153f), LL(0x77b69fe8b6777799), LL(0x37eba592eb373759), LL(0xe5567b9e56e5e532),
    LL(0x9fd98c13d99f9fbc), LL(0xf017d32317f0f00d), LL(0x4a7f6a207f4a4ade), LL(0xda959e4495dada73),
    LL(0x5825faa2255858e8), LL(0xc9ca06cfcac9c946), LL(0x298d557c8d29297b), LL(0x0a22505a220a0a1e),
    LL(0xb14fe1504fb1b1ce), LL(0xa01a69c91aa0a0fd), LL(0x6bda7f14da6b6bbd), LL(0x85ab5cd9ab858592),
    LL(0xbd73813c73bdbdda), LL(0x5d34d28f345d5de7), LL(0x1050809050101030), LL(0xf403f30703f4f401),
    LL(0xcbc016ddc0cbcb40), LL(0x3ec6edd3c63e3e42), LL(0x0511282d1105050f), LL(0x67e61f78e66767a9),
    LL(0xe453739753e4e431), LL(0x27bb2502bb272769), LL(0x41583273584141c3), LL(0x8b9d2ca79d8b8b80),
    LL(0xa70151f601a7a7f4), LL(0x7d94cfb2947d7d87), LL(0x95fbdc49fb9595a2), LL(0xd89f8e569fd8d875),
    LL(0xfb308b7030fbfb10), LL(0xee7123cd71eeee2f), LL(0x7c91c7bb917c7c84), LL(0x66e31771e36666aa),
    LL(0xdd8ea67b8edddd7a), LL(0x174bb8af4b171739), LL(0x47460245464747c9), LL(0x9edc841adc9e9ebf),
    LL(0xcac51ed4c5caca43), LL(0x2d997558992d2d77), LL(0xbf79912e79bfbfdc), LL(0x071b383f1b070709),
    LL(0xad2301ac23adadea), LL(0x5a2feab02f5a5aee), LL(0x83b56cefb5838398), LL(0x33ff85b6ff333355),
    LL(0x63f23f5cf26363a5), LL(0x020a10120a020206), LL(0xaa38399338aaaae3), LL(0x71a8afdea8717193),
    LL(0xc8cf0ec6cfc8c845), LL(0x197dc8d17d19192b), LL(0x4970723b704949db), LL(0xd99a865f9ad9d976),
    LL(0xf21dc3311df2f20b), LL(0xe3484ba848e3e338), LL(0x5b2ae2b92a5b5bed), LL(0x889234bc92888885),
    LL(0x9ac8a43ec89a9ab3), LL(0x26be2d0bbe26266a), LL(0x32fa8dbffa323256), LL(0xb04ae9594ab0b0cd),
    LL(0xe96a1bf26ae9e926), LL(0x0f337877330f0f11), LL(0xd5a6e633a6d5d562), LL(0x80ba74f4ba80809d),
    LL(0xbe7c99277cbebedf), LL(0xcdde26ebdecdcd4a), LL(0x34e4bd89e434345c), LL(0x48757a32754848d8),
    LL(0xff24ab5424ffff1c), LL(0x7a8ff78d8f7a7a8e), LL(0x90eaf464ea9090ad), LL(0x5f3ec29d3e5f5fe1),
    LL(0x20a01d3da0202060), LL(0x68d5670fd56868b8), LL(0x1a72d0ca721a1a2e), LL(0xae2c19b72caeaeef),
    LL(0xb45ec97d5eb4b4c1), LL(0x54199ace195454fc), LL(0x93e5ec7fe59393a8), LL(0x22aa0d2faa222266),
    LL(0x64e90763e96464ac), LL(0xf112db2a12f1f10e), LL(0x73a2bfcca2737395), LL(0x125a90825a121236),
    LL(0x405d3a7a5d4040c0), LL(0x0828404828080818), LL(0xc3e85695e8c3c358), LL(0xec7b33df7becec29),
    LL(0xdb90964d90dbdb70), LL(0xa11f61c01fa1a1fe), LL(0x8d831c91838d8d8a), LL(0x3dc9f5c8c93d3d47),
    LL(0x97f1cc5bf19797a4), LL(0x0000000000000000), LL(0xcfd436f9d4cfcf4c), LL(0x2b87456e872b2b7d),
    LL(0x76b397e1b376769a), LL(0x82b064e6b082829b), LL(0xd6a9fe28a9d6d667), LL(0x1b77d8c3771b1b2d),
    LL(0xb55bc1745bb5b5c2), LL(0xaf2911be29afafec), LL(0x6adf771ddf6a6abe), LL(0x500dbaea0d5050f0),
    LL(0x454c12574c4545cf), LL(0xf318cb3818f3f308), LL(0x30f09dadf0303050), LL(0xef742bc474efef2c),
    LL(0x3fc3e5dac33f3f41), LL(0x551c92c71c5555ff), LL(0xa21079db10a2a2fb), LL(0xea6503e965eaea23),
    LL(0x65ec0f6aec6565af), LL(0xba68b90368babad3), LL(0x2f93654a932f2f71), LL(0xc0e74e8ee7c0c05d),
    LL(0xde81be6081dede7f), LL(0x1c6ce0fc6c1c1c24), LL(0xfd2ebb462efdfd1a), LL(0x4d64521f644d4dd7),
    LL(0x92e0e476e09292ab), LL(0x75bc8ffabc75759f), LL(0x061e30361e06060a), LL(0x8a9824ae988a8a83),
    LL(0xb240f94b40b2b2cb), LL(0xe659638559e6e637), LL(0x0e36707e360e0e12), LL(0x1f63f8e7631f1f21),
    LL(0x62f73755f76262a6), LL(0xd4a3ee3aa3d4d461), LL(0xa832298132a8a8e5), LL(0x96f4c452f49696a7),
    LL(0xf93a9b623af9f916), LL(0xc5f666a3f6c5c552), LL(0x25b13510b125256f), LL(0x5920f2ab205959eb),
    LL(0x84ae54d0ae848491), LL(0x72a7b7c5a7727296), LL(0x39ddd5ecdd39394b), LL(0x4c615a16614c4cd4),
    LL(0x5e3bca943b5e5ee2), LL(0x7885e79f85787888), LL(0x38d8dde5d8383848), LL(0x8c861498868c8c89),
    LL(0xd1b2c617b2d1d16e), LL(0xa50b41e40ba5a5f2), LL(0xe24d43a14de2e23b), LL(0x61f82f4ef86161a3),
    LL(0xb345f14245b3b3c8), LL(0x21a51534a5212163), LL(0x9cd69408d69c9cb9), LL(0x1e66f0ee661e1e22),
    LL(0x43522261524343c5), LL(0xc7fc76b1fcc7c754), LL(0xfc2bb34f2bfcfc19), LL(0x041420241404040c),
    LL(0x5108b2e3085151f3), LL(0x99c7bc25c79999b6), LL(0x6dc44f22c46d6db7), LL(0x0d396865390d0d17),
    LL(0xfa35837935fafa13), LL(0xdf84b66984dfdf7c), LL(0x7e9bd7a99b7e7e82), LL(0x24b43d19b424246c),
    LL(0x3bd7c5fed73b3b4d), LL(0xab3d319a3dababe0), LL(0xced13ef0d1cece4f), LL(0x1155889955111133),
    LL(0x8f890c83898f8f8c), LL(0x4e6b4a046b4e4ed2), LL(0xb751d16651b7b7c4), LL(0xeb600be060ebeb20),
    LL(0x3cccfdc1cc3c3c44), LL(0x81bf7cfdbf81819e), LL(0x94fed440fe9494a1), LL(0xf70ceb1c0cf7f704),
    LL(0xb967a11867b9b9d6), LL(0x135f988b5f131335), LL(0x2c9c7d519c2c2c74), LL(0xd3b8d605b8d3d368),
    LL(0xe75c6b8c5ce7e734), LL(0x6ecb5739cb6e6eb2), LL(0xc4f36eaaf3c4c451), LL(0x030f181b0f030305),
    LL(0x56138adc135656fa), LL(0x44491a5e494444cc), LL(0x7f9edfa09e7f7f81), LL(0xa937218837a9a9e6),
    LL(0x2a824d67822a2a7e), LL(0xbb6db10a6dbbbbd0), LL(0xc1e24687e2c1c15e), LL(0x5302a2f1025353f5),
    LL(0xdc8bae728bdcdc79), LL(0x0b275853270b0b1d), LL(0x9dd39c01d39d9dba), LL(0x6cc1472bc16c6cb4),
    LL(0x31f595a4f5313153), LL(0x74b987f3b974749c), LL(0xf609e31509f6f607), LL(0x46430a4c434646ca),
    LL(0xac2609a526acace9), LL(0x89973cb597898986), LL(0x1444a0b44414143c), LL(0xe1425bba42e1e13e),
    LL(0x164eb0a64e16163a), LL(0x3ad2cdf7d23a3a4e), LL(0x69d06f06d06969bb), LL(0x092d48412d09091b),
    LL(0x70ada7d7ad707090), LL(0xb654d96f54b6b6c7), LL(0xd0b7ce1eb7d0d06d), LL(0xed7e3bd67eeded2a),
    LL(0xccdb2ee2dbcccc49), LL(0x42572a68574242c6), LL(0x98c2b42cc29898b5), LL(0xa40e49ed0ea4a4f1),
    LL(0x28885d7588282878), LL(0x5c31da86315c5ce4), LL(0xf83f936b3ff8f815), LL(0x86a444c2a4868697),
};

static const u64 C6[256] = {
    LL(0x281878c0d8781818), LL(0x6523af0526af2323), LL(0x57c6f97eb8f9c6c6), LL(0x25e86f13fb6fe8e8),
    LL(0x9487a14ccba18787), LL(0xd5b862a91162b8b8), LL(0x0301050809050101), LL(0xd14f6e420d6e4f4f),
    LL(0x5a36eead9bee3636), LL(0xf7a60459ff04a6a6), LL(0x6bd2bdde0cbdd2d2), LL(0x02f506fb0e06f5f5),
    LL(0x8b7980ef96807979), LL(0xb16fce5f30ce6f6f), LL(0xae91effc6def9191), LL(0xf65207aaf8075252),
    LL(0xa060fd2747fd6060), LL(0xd9bc76893576bcbc), LL(0xb09bcdac37cd9b9b), LL(0x8f8e8c048a8c8e8e),
    LL(0xf8a31571d215a3a3), LL(0x140c3c606c3c0c0c), LL(0x8d7b8aff848a7b7b), LL(0x5f35e1b580e13535),
    LL(0x271d69e8f5691d1d), LL(0x3de04753b347e0e0), LL(0x64d7acf621acd7d7), LL(0x5bc2ed5e9cedc2c2),
    LL(0x722e966d43962e2e), LL(0xdd4b7a62297a4b4b), LL(0x1ffe21a35d21fefe), LL(0xf9571682d5165757),
    LL(0x3f1541a8bd411515), LL(0x9977b69fe8b67777), LL(0x5937eba592eb3737), LL(0x32e5567b9e56e5e5),
    LL(0xbc9fd98c13d99f9f), LL(0x0df017d32317f0f0), LL(0xde4a7f6a207f4a4a), LL(0x73da959e4495dada),
    LL(0xe85825faa2255858), LL(0x46c9ca06cfcac9c9), LL(0x7b298d557c8d2929), LL(0x1e0a22505a220a0a),
    LL(0xceb14fe1504fb1b1), LL(0xfda01a69c91aa0a0), LL(0xbd6bda7f14da6b6b), LL(0x9285ab5cd9ab8585),
    LL(0xdabd73813c73bdbd), LL(0xe75d34d28f345d5d), LL(0x3010508090501010), LL(0x01f403f30703f4f4),
    LL(0x40cbc016ddc0cbcb), LL(0x423ec6edd3c63e3e), LL(0x0f0511282d110505), LL(0xa967e61f78e66767),
    LL(0x31e453739753e4e4), LL(0x6927bb2502bb2727), LL(0xc341583273584141), LL(0x808b9d2ca79d8b8b),
    LL(0xf4a70151f601a7a7), LL(0x877d94cfb2947d7d), LL(0xa295fbdc49fb9595), LL(0x75d89f8e569fd8d8),
    LL(0x10fb308b7030fbfb), LL(0x2fee7123cd71eeee), LL(0x847c91c7bb917c7c), LL(0xaa66e31771e36666),
    LL(0x7add8ea67b8edddd), LL(0x39174bb8af4b1717), LL(0xc947460245464747), LL(0xbf9edc841adc9e9e),
    LL(0x43cac51ed4c5caca), LL(0x772d997558992d2d), LL(0xdcbf79912e79bfbf), LL(0x09071b383f1b0707),
    LL(0xeaad2301ac23adad), LL(0xee5a2feab02f5a5a), LL(0x9883b56cefb58383), LL(0x5533ff85b6ff3333),
    LL(0xa563f23f5cf26363), LL(0x06020a10120a0202), LL(0xe3aa38399338aaaa), LL(0x9371a8afdea87171),
    LL(0x45c8cf0ec6cfc8c8), LL(0x2b197dc8d17d1919), LL(0xdb4970723b704949), LL(0x76d99a865f9ad9d9),
    LL(0x0bf21dc3311df2f2), LL(0x38e3484ba848e3e3), LL(0xed5b2ae2b92a5b5b), LL(0x85889234bc928888),
    LL(0xb39ac8a43ec89a9a), LL(0x6a26be2d0bbe2626), LL(0x5632fa8dbffa3232), LL(0xcdb04ae9594ab0b0),
    LL(0x26e96a1bf26ae9e9), LL(0x110f337877330f0f), LL(0x62d5a6e633a6d5d5), LL(0x9d80ba74f4ba8080),
    LL(0xdfbe7c99277cbebe), LL(0x4acdde26ebdecdcd), LL(0x5c34e4bd89e43434), LL(0xd848757a32754848),
    LL(0x1cff24ab5424ffff), LL(0x8e7a8ff78d8f7a7a), LL(0xad90eaf464ea9090), LL(0xe15f3ec29d3e5f5f),
    LL(0x6020a01d3da02020), LL(0xb868d5670fd56868), LL(0x2e1a72d0ca721a1a), LL(0xefae2c19b72caeae),
    LL(0xc1b45ec97d5eb4b4), LL(0xfc54199ace195454), LL(0xa893e5ec7fe59393), LL(0x6622aa0d2faa2222),
    LL(0xac64e90763e96464), LL(0x0ef112db2a12f1f1), LL(0x9573a2bfcca27373), LL(0x36125a90825a1212),
    LL(0xc0405d3a7a5d4040), LL(0x1808284048280808), LL(0x58c3e85695e8c3c3), LL(0x29ec7b33df7becec),
    LL(0x70db90964d90dbdb), LL(0xfea11f61c01fa1a1), LL(0x8a8d831c91838d8d), LL(0x473dc9f5c8c93d3d),
    LL(0xa497f1cc5bf19797), LL(0x0000000000000000), LL(0x4ccfd436f9d4cfcf), LL(0x7d2b87456e872b2b),
    LL(0x9a76b397e1b37676), LL(0x9b82b064e6b08282), LL(0x67d6a9fe28a9d6d6), LL(0x2d1b77d8c3771b1b),
    LL(0xc2b55bc1745bb5b5), LL(0xecaf2911be29afaf), LL(0xbe6adf771ddf6a6a), LL(0xf0500dbaea0d5050),
    LL(0xcf454c12574c4545), LL(0x08f318cb3818f3f3), LL(0x5030f09dadf03030), LL(0x2cef742bc474efef),
    LL(0x413fc3e5dac33f3f), LL(0xff551c92c71c5555), LL(0xfba21079db10a2a2), LL(0x23ea6503e965eaea),
    LL(0xaf65ec0f6aec6565), LL(0xd3ba68b90368baba), LL(0x712f93654a932f2f), LL(0x5dc0e74e8ee7c0c0),
    LL(0x7fde81be6081dede), LL(0x241c6ce0fc6c1c1c), LL(0x1afd2ebb462efdfd), LL(0xd74d64521f644d4d),
    LL(0xab92e0e476e09292), LL(0x9f75bc8ffabc7575), LL(0x0a061e30361e0606), LL(0x838a9824ae988a8a),
    LL(0xcbb240f94b40b2b2), LL(0x37e659638559e6e6), LL(0x120e36707e360e0e), LL(0x211f63f8e7631f1f),
    LL(0xa662f73755f76262), LL(0x61d4a3ee3aa3d4d4), LL(0xe5a832298132a8a8), LL(0xa796f4c452f49696),
    LL(0x16f93a9b623af9f9), LL(0x52c5f666a3f6c5c5), LL(0x6f25b13510b12525), LL(0xeb5920f2ab205959),
    LL(0x9184ae54d0ae8484), LL(0x9672a7b7c5a77272), LL(0x4b39ddd5ecdd3939), LL(0xd44c615a16614c4c),
    LL(0xe25e3bca943b5e5e), LL(0x887885e79f857878), LL(0x4838d8dde5d83838), LL(0x898c861498868c8c),
    LL(0x6ed1b2c617b2d1d1), LL(0xf2a50b41e40ba5a5), LL(0x3be24d43a14de2e2), LL(0xa361f82f4ef86161),
    LL(0xc8b345f14245b3b3), LL(0x6321a51534a52121), LL(0xb99cd69408d69c9c), LL(0x221e66f0ee661e1e),
    LL(0xc543522261524343), LL(0x54c7fc76b1fcc7c7), LL(0x19fc2bb34f2bfcfc), LL(0x0c04142024140404),
    LL(0xf35108b2e3085151), LL(0xb699c7bc25c79999), LL(0xb76dc44f22c46d6d), LL(0x170d396865390d0d),
    LL(0x13fa35837935fafa), LL(0x7cdf84b66984dfdf), LL(0x827e9bd7a99b7e7e), LL(0x6c24b43d19b42424),
    LL(0x4d3bd7c5fed73b3b), LL(0xe0ab3d319a3dabab), LL(0x4fced13ef0d1cece), LL(0x3311558899551111),
    LL(0x8c8f890c83898f8f), LL(0xd24e6b4a046b4e4e), LL(0xc4b751d16651b7b7), LL(0x20eb600be060ebeb),
    LL(0x443cccfdc1cc3c3c), LL(0x9e81bf7cfdbf8181), LL(0xa194fed440fe9494), LL(0x04f70ceb1c0cf7f7),
    LL(0xd6b967a11867b9b9), LL(0x35135f988b5f1313), LL(0x742c9c7d519c2c2c), LL(0x68d3b8d605b8d3d3),
    LL(0x34e75c6b8c5ce7e7), LL(0xb26ecb5739cb6e6e), LL(0x51c4f36eaaf3c4c4), LL(0x05030f181b0f0303),
    LL(0xfa56138adc135656), LL(0xcc44491a5e494444), LL(0x817f9edfa09e7f7f), LL(0xe6a937218837a9a9),
    LL(0x7e2a824d67822a2a), LL(0xd0bb6db10a6dbbbb), LL(0x5ec1e24687e2c1c1), LL(0xf55302a2f1025353),
    LL(0x79dc8bae728bdcdc), LL(0x1d0b275853270b0b), LL(0xba9dd39c01d39d9d), LL(0xb46cc1472bc16c6c),
    LL(0x5331f595a4f53131), LL(0x9c74b987f3b97474), LL(0x07f609e31509f6f6), LL(0xca46430a4c434646),
    LL(0xe9ac2609a526acac), LL(0x8689973cb5978989), LL(0x3c1444a0b4441414), LL(0x3ee1425bba42e1e1),
    LL(0x3a164eb0a64e1616), LL(0x4e3ad2cdf7d23a3a), LL(0xbb69d06f06d06969), LL(0x1b092d48412d0909),
    LL(0x9070ada7d7ad7070), LL(0xc7b654d96f54b6b6), LL(0x6dd0b7ce1eb7d0d0), LL(0x2aed7e3bd67eeded),
    LL(0x49ccdb2ee2dbcccc), LL(0xc642572a68574242), LL(0xb598c2b42cc29898), LL(0xf1a40e49ed0ea4a4),
    LL(0x7828885d75882828), LL(0xe45c31da86315c5c), LL(0x15f83f936b3ff8f8), LL(0x9786a444c2a48686),
};

static const u64 C7[256] = {
    LL(0x18281878c0d87818), LL(0x236523af0526af23), LL(0xc657c6f97eb8f9c6), LL(0xe825e86f13fb6fe8),
    LL(0x879487a14ccba187), LL(0xb8d5b862a91162b8), LL(0x0103010508090501), LL(0x4fd14f6e420d6e4f),
    LL(0x365a36eead9bee36), LL(0xa6f7a60459ff04a6), LL(0xd26bd2bdde0cbdd2), LL(0xf502f506fb0e06f5),
    LL(0x798b7980ef968079), LL(0x6fb16fce5f30ce6f), LL(0x91ae91effc6def91), LL(0x52f65207aaf80752),
    LL(0x60a060fd2747fd60), LL(0xbcd9bc76893576bc), LL(0x9bb09bcdac37cd9b), LL(0x8e8f8e8c048a8c8e),
    LL(0xa3f8a31571d215a3), LL(0x0c140c3c606c3c0c), LL(0x7b8d7b8aff848a7b), LL(0x355f35e1b580e135),
    LL(0x1d271d69e8f5691d), LL(0xe03de04753b347e0), LL(0xd764d7acf621acd7), LL(0xc25bc2ed5e9cedc2),
    LL(0x2e722e966d43962e), LL(0x4bdd4b7a62297a4b), LL(0xfe1ffe21a35d21fe), LL(0x57f9571682d51657),
    LL(0x153f1541a8bd4115), LL(0x779977b69fe8b677), LL(0x375937eba592eb37), LL(0xe532e5567b9e56e5),
    LL(0x9fbc9fd98c13d99f), LL(0xf00df017d32317f0), LL(0x4ade4a7f6a207f4a), LL(0xda73da959e4495da),
    LL(0x58e85825faa22558), LL(0xc946c9ca06cfcac9), LL(0x297b298d557c8d29), LL(0x0a1e0a22505a220a),
    LL(0xb1ceb14fe1504fb1), LL(0xa0fda01a69c91aa0), LL(0x6bbd6bda7f14da6b), LL(0x859285ab5cd9ab85),
    LL(0xbddabd73813c73bd), LL(0x5de75d34d28f345d), LL(0x1030105080905010), LL(0xf401f403f30703f4),
    LL(0xcb40cbc016ddc0cb), LL(0x3e423ec6edd3c63e), LL(0x050f0511282d1105), LL(0x67a967e61f78e667),
    LL(0xe431e453739753e4), LL(0x276927bb2502bb27), LL(0x41c3415832735841), LL(0x8b808b9d2ca79d8b),
    LL(0xa7f4a70151f601a7), LL(0x7d877d94cfb2947d), LL(0x95a295fbdc49fb95), LL(0xd875d89f8e569fd8),
    LL(0xfb10fb308b7030fb), LL(0xee2fee7123cd71ee), LL(0x7c847c91c7bb917c), LL(0x66aa66e31771e366),
    LL(0xdd7add8ea67b8edd), LL(0x1739174bb8af4b17), LL(0x47c9474602454647), LL(0x9ebf9edc841adc9e),
    LL(0xca43cac51ed4c5ca), LL(0x2d772d997558992d), LL(0xbfdcbf79912e79bf), LL(0x0709071b383f1b07),
    LL(0xadeaad2301ac23ad), LL(0x5aee5a2feab02f5a), LL(0x839883b56cefb583), LL(0x335533ff85b6ff33),
    LL(0x63a563f23f5cf263), LL(0x0206020a10120a02), LL(0xaae3aa38399338aa), LL(0x719371a8afdea871),
    LL(0xc845c8cf0ec6cfc8), LL(0x192b197dc8d17d19), LL(0x49db4970723b7049), LL(0xd976d99a865f9ad9),
    LL(0xf20bf21dc3311df2), LL(0xe338e3484ba848e3), LL(0x5bed5b2ae2b92a5b), LL(0x8885889234bc9288),
    LL(0x9ab39ac8a43ec89a), LL(0x266a26be2d0bbe26), LL(0x325632fa8dbffa32), LL(0xb0cdb04ae9594ab0),
    LL(0xe926e96a1bf26ae9), LL(0x0f110f337877330f), LL(0xd562d5a6e633a6d5), LL(0x809d80ba74f4ba80),
    LL(0xbedfbe7c99277cbe), LL(0xcd4acdde26ebdecd), LL(0x345c34e4bd89e434), LL(0x48d848757a327548),
    LL(0xff1cff24ab5424ff), LL(0x7a8e7a8ff78d8f7a), LL(0x90ad90eaf464ea90), LL(0x5fe15f3ec29d3e5f),
    LL(0x206020a01d3da020), LL(0x68b868d5670fd568), LL(0x1a2e1a72d0ca721a), LL(0xaeefae2c19b72cae),
    LL(0xb4c1b45ec97d5eb4), LL(0x54fc54199ace1954), LL(0x93a893e5ec7fe593), LL(0x226622aa0d2faa22),
    LL(0x64ac64e90763e964), LL(0xf10ef112db2a12f1), LL(0x739573a2bfcca273), LL(0x1236125a90825a12),
    LL(0x40c0405d3a7a5d40), LL(0x0818082840482808), LL(0xc358c3e85695e8c3), LL(0xec29ec7b33df7bec),
    LL(0xdb70db90964d90db), LL(0xa1fea11f61c01fa1), LL(0x8d8a8d831c91838d), LL(0x3d473dc9f5c8c93d),
    LL(0x97a497f1cc5bf197), LL(0x0000000000000000), LL(0xcf4ccfd436f9d4cf), LL(0x2b7d2b87456e872b),
    LL(0x769a76b397e1b376), LL(0x829b82b064e6b082), LL(0xd667d6a9fe28a9d6), LL(0x1b2d1b77d8c3771b),
    LL(0xb5c2b55bc1745bb5), LL(0xafecaf2911be29af), LL(0x6abe6adf771ddf6a), LL(0x50f0500dbaea0d50),
    LL(0x45cf454c12574c45), LL(0xf308f318cb3818f3), LL(0x305030f09dadf030), LL(0xef2cef742bc474ef),
    LL(0x3f413fc3e5dac33f), LL(0x55ff551c92c71c55), LL(0xa2fba21079db10a2), LL(0xea23ea6503e965ea),
    LL(0x65af65ec0f6aec65), LL(0xbad3ba68b90368ba), LL(0x2f712f93654a932f), LL(0xc05dc0e74e8ee7c0),
    LL(0xde7fde81be6081de), LL(0x1c241c6ce0fc6c1c), LL(0xfd1afd2ebb462efd), LL(0x4dd74d64521f644d),
    LL(0x92ab92e0e476e092), LL(0x759f75bc8ffabc75), LL(0x060a061e30361e06), LL(0x8a838a9824ae988a),
    LL(0xb2cbb240f94b40b2), LL(0xe637e659638559e6), LL(0x0e120e36707e360e), LL(0x1f211f63f8e7631f),
    LL(0x62a662f73755f762), LL(0xd461d4a3ee3aa3d4), LL(0xa8e5a832298132a8), LL(0x96a796f4c452f496),
    LL(0xf916f93a9b623af9), LL(0xc552c5f666a3f6c5), LL(0x256f25b13510b125), LL(0x59eb5920f2ab2059),
    LL(0x849184ae54d0ae84), LL(0x729672a7b7c5a772), LL(0x394b39ddd5ecdd39), LL(0x4cd44c615a16614c),
    LL(0x5ee25e3bca943b5e), LL(0x78887885e79f8578), LL(0x384838d8dde5d838), LL(0x8c898c861498868c),
    LL(0xd16ed1b2c617b2d1), LL(0xa5f2a50b41e40ba5), LL(0xe23be24d43a14de2), LL(0x61a361f82f4ef861),
    LL(0xb3c8b345f14245b3), LL(0x216321a51534a521), LL(0x9cb99cd69408d69c), LL(0x1e221e66f0ee661e),
    LL(0x43c5435222615243), LL(0xc754c7fc76b1fcc7), LL(0xfc19fc2bb34f2bfc), LL(0x040c041420241404),
    LL(0x51f35108b2e30851), LL(0x99b699c7bc25c799), LL(0x6db76dc44f22c46d), LL(0x0d170d396865390d),
    LL(0xfa13fa35837935fa), LL(0xdf7cdf84b66984df), LL(0x7e827e9bd7a99b7e), LL(0x246c24b43d19b424),
    LL(0x3b4d3bd7c5fed73b), LL(0xabe0ab3d319a3dab), LL(0xce4fced13ef0d1ce), LL(0x1133115588995511),
    LL(0x8f8c8f890c83898f), LL(0x4ed24e6b4a046b4e), LL(0xb7c4b751d16651b7), LL(0xeb20eb600be060eb),
    LL(0x3c443cccfdc1cc3c), LL(0x819e81bf7cfdbf81), LL(0x94a194fed440fe94), LL(0xf704f70ceb1c0cf7),
    LL(0xb9d6b967a11867b9), LL(0x1335135f988b5f13), LL(0x2c742c9c7d519c2c), LL(0xd368d3b8d605b8d3),
    LL(0xe734e75c6b8c5ce7), LL(0x6eb26ecb5739cb6e), LL(0xc451c4f36eaaf3c4), LL(0x0305030f181b0f03),
    LL(0x56fa56138adc1356), LL(0x44cc44491a5e4944), LL(0x7f817f9edfa09e7f), LL(0xa9e6a937218837a9),
    LL(0x2a7e2a824d67822a), LL(0xbbd0bb6db10a6dbb), LL(0xc15ec1e24687e2c1), LL(0x53f55302a2f10253),
    LL(0xdc79dc8bae728bdc), LL(0x0b1d0b275853270b), LL(0x9dba9dd39c01d39d), LL(0x6cb46cc1472bc16c),
    LL(0x315331f595a4f531), LL(0x749c74b987f3b974), LL(0xf607f609e31509f6), LL(0x46ca46430a4c4346),
    LL(0xace9ac2609a526ac), LL(0x898689973cb59789), LL(0x143c1444a0b44414), LL(0xe13ee1425bba42e1),
    LL(0x163a164eb0a64e16), LL(0x3a4e3ad2cdf7d23a), LL(0x69bb69d06f06d069), LL(0x091b092d48412d09),
    LL(0x709070ada7d7ad70), LL(0xb6c7b654d96f54b6), LL(0xd06dd0b7ce1eb7d0), LL(0xed2aed7e3bd67eed),
    LL(0xcc49ccdb2ee2dbcc), LL(0x42c642572a685742), LL(0x98b598c2b42cc298), LL(0xa4f1a40e49ed0ea4),
    LL(0x287828885d758828), LL(0x5ce45c31da86315c), LL(0xf815f83f936b3ff8), LL(0x869786a444c2a486),
};
#endif /* OBSOLETE */

static const u64 rc[R + 1] = {
    LL(0x0000000000000000),
    LL(0x1823c6e887b8014f),
    LL(0x36a6d2f5796f9152),
    LL(0x60bc9b8ea30c7b35),
    LL(0x1de0d7c22e4bfe57),
    LL(0x157737e59ff04ada),
    LL(0x58c9290ab1a06b85),
    LL(0xbd5d10f4cb3e0567),
    LL(0xe427418ba77d95d8),
    LL(0xfbee7c66dd17479e),
    LL(0xca2dbf07ad5a8333),
};

/**
 * The core Whirlpool transform.
 */
static void processBuffer(struct NESSIEstruct * const structpointer) {
    int i, r;
    u64 K[8];        /* the round key */
    u64 block[8];    /* mu(buffer) */
    u64 state[8];    /* the cipher state */
    u64 L[8];
    u8 *buffer = structpointer->buffer;

#ifdef TRACE_INTERMEDIATE_VALUES
    printf("The 8x8 matrix Z' derived from the data-string is as follows.\n");
    for (i = 0; i < WBLOCKBYTES/8; i++) {
        printf("    %02X %02X %02X %02X %02X %02X %02X %02X\n",
            buffer[0], buffer[1], buffer[2], buffer[3],
            buffer[4], buffer[5], buffer[6], buffer[7]);
        buffer += 8;
    }
    printf("\n");
    buffer = structpointer->buffer;
#endif /* ?TRACE_INTERMEDIATE_VALUES */

    /*
     * map the buffer to a block:
     */
    for (i = 0; i < 8; i++, buffer += 8) {
        block[i] =
            (((u64)buffer[0]        ) << 56) ^
            (((u64)buffer[1] & 0xffL) << 48) ^
            (((u64)buffer[2] & 0xffL) << 40) ^
            (((u64)buffer[3] & 0xffL) << 32) ^
            (((u64)buffer[4] & 0xffL) << 24) ^
            (((u64)buffer[5] & 0xffL) << 16) ^
            (((u64)buffer[6] & 0xffL) <<  8) ^
            (((u64)buffer[7] & 0xffL)      );
    }
    /*
     * compute and apply K^0 to the cipher state:
     */
    state[0] = block[0] ^ (K[0] = structpointer->hash[0]);
    state[1] = block[1] ^ (K[1] = structpointer->hash[1]);
    state[2] = block[2] ^ (K[2] = structpointer->hash[2]);
    state[3] = block[3] ^ (K[3] = structpointer->hash[3]);
    state[4] = block[4] ^ (K[4] = structpointer->hash[4]);
    state[5] = block[5] ^ (K[5] = structpointer->hash[5]);
    state[6] = block[6] ^ (K[6] = structpointer->hash[6]);
    state[7] = block[7] ^ (K[7] = structpointer->hash[7]);
#ifdef TRACE_INTERMEDIATE_VALUES
    printf("The K_0 matrix (from the initialization value IV) and X'' matrix are as follows.\n");
    for (i = 0; i < DIGESTBYTES/8; i++) {
        printf(
            "    %02X %02X %02X %02X %02X %02X %02X %02X        %02X %02X %02X %02X %02X %02X %02X %02X\n",
            (u8)(K[i] >> 56),
            (u8)(K[i] >> 48),
            (u8)(K[i] >> 40),
            (u8)(K[i] >> 32),
            (u8)(K[i] >> 24),
            (u8)(K[i] >> 16),
            (u8)(K[i] >>  8),
            (u8)(K[i]      ),

            (u8)(state[i] >> 56),
            (u8)(state[i] >> 48),
            (u8)(state[i] >> 40),
            (u8)(state[i] >> 32),
            (u8)(state[i] >> 24),
            (u8)(state[i] >> 16),
            (u8)(state[i] >>  8),
            (u8)(state[i]      ));
    }
    printf("\n");
    printf("The following are (hexadecimal representations of) the successive values of the variables K_i for i = 1 to 10 and W'.\n");
    printf("\n");
#endif /* ?TRACE_INTERMEDIATE_VALUES */
    /*
     * iterate over all rounds:
     */
    for (r = 1; r <= R; r++) {
        /*
         * compute K^r from K^{r-1}:
         */
        L[0] =
            C0[(int)(K[0] >> 56)       ] ^
            C1[(int)(K[7] >> 48) & 0xff] ^
            C2[(int)(K[6] >> 40) & 0xff] ^
            C3[(int)(K[5] >> 32) & 0xff] ^
            C4[(int)(K[4] >> 24) & 0xff] ^
            C5[(int)(K[3] >> 16) & 0xff] ^
            C6[(int)(K[2] >>  8) & 0xff] ^
            C7[(int)(K[1]      ) & 0xff] ^
            rc[r];
        L[1] =
            C0[(int)(K[1] >> 56)       ] ^
            C1[(int)(K[0] >> 48) & 0xff] ^
            C2[(int)(K[7] >> 40) & 0xff] ^
            C3[(int)(K[6] >> 32) & 0xff] ^
            C4[(int)(K[5] >> 24) & 0xff] ^
            C5[(int)(K[4] >> 16) & 0xff] ^
            C6[(int)(K[3] >>  8) & 0xff] ^
            C7[(int)(K[2]      ) & 0xff];
        L[2] =
            C0[(int)(K[2] >> 56)       ] ^
            C1[(int)(K[1] >> 48) & 0xff] ^
            C2[(int)(K[0] >> 40) & 0xff] ^
            C3[(int)(K[7] >> 32) & 0xff] ^
            C4[(int)(K[6] >> 24) & 0xff] ^
            C5[(int)(K[5] >> 16) & 0xff] ^
            C6[(int)(K[4] >>  8) & 0xff] ^
            C7[(int)(K[3]      ) & 0xff];
        L[3] =
            C0[(int)(K[3] >> 56)       ] ^
            C1[(int)(K[2] >> 48) & 0xff] ^
            C2[(int)(K[1] >> 40) & 0xff] ^
            C3[(int)(K[0] >> 32) & 0xff] ^
            C4[(int)(K[7] >> 24) & 0xff] ^
            C5[(int)(K[6] >> 16) & 0xff] ^
            C6[(int)(K[5] >>  8) & 0xff] ^
            C7[(int)(K[4]      ) & 0xff];
        L[4] =
            C0[(int)(K[4] >> 56)       ] ^
            C1[(int)(K[3] >> 48) & 0xff] ^
            C2[(int)(K[2] >> 40) & 0xff] ^
            C3[(int)(K[1] >> 32) & 0xff] ^
            C4[(int)(K[0] >> 24) & 0xff] ^
            C5[(int)(K[7] >> 16) & 0xff] ^
            C6[(int)(K[6] >>  8) & 0xff] ^
            C7[(int)(K[5]      ) & 0xff];
        L[5] =
            C0[(int)(K[5] >> 56)       ] ^
            C1[(int)(K[4] >> 48) & 0xff] ^
            C2[(int)(K[3] >> 40) & 0xff] ^
            C3[(int)(K[2] >> 32) & 0xff] ^
            C4[(int)(K[1] >> 24) & 0xff] ^
            C5[(int)(K[0] >> 16) & 0xff] ^
            C6[(int)(K[7] >>  8) & 0xff] ^
            C7[(int)(K[6]      ) & 0xff];
        L[6] =
            C0[(int)(K[6] >> 56)       ] ^
            C1[(int)(K[5] >> 48) & 0xff] ^
            C2[(int)(K[4] >> 40) & 0xff] ^
            C3[(int)(K[3] >> 32) & 0xff] ^
            C4[(int)(K[2] >> 24) & 0xff] ^
            C5[(int)(K[1] >> 16) & 0xff] ^
            C6[(int)(K[0] >>  8) & 0xff] ^
            C7[(int)(K[7]      ) & 0xff];
        L[7] =
            C0[(int)(K[7] >> 56)       ] ^
            C1[(int)(K[6] >> 48) & 0xff] ^
            C2[(int)(K[5] >> 40) & 0xff] ^
            C3[(int)(K[4] >> 32) & 0xff] ^
            C4[(int)(K[3] >> 24) & 0xff] ^
            C5[(int)(K[2] >> 16) & 0xff] ^
            C6[(int)(K[1] >>  8) & 0xff] ^
            C7[(int)(K[0]      ) & 0xff];
        K[0] = L[0];
        K[1] = L[1];
        K[2] = L[2];
        K[3] = L[3];
        K[4] = L[4];
        K[5] = L[5];
        K[6] = L[6];
        K[7] = L[7];
        /*
         * apply the r-th round transformation:
         */
        L[0] =
            C0[(int)(state[0] >> 56)       ] ^
            C1[(int)(state[7] >> 48) & 0xff] ^
            C2[(int)(state[6] >> 40) & 0xff] ^
            C3[(int)(state[5] >> 32) & 0xff] ^
            C4[(int)(state[4] >> 24) & 0xff] ^
            C5[(int)(state[3] >> 16) & 0xff] ^
            C6[(int)(state[2] >>  8) & 0xff] ^
            C7[(int)(state[1]      ) & 0xff] ^
            K[0];
        L[1] =
            C0[(int)(state[1] >> 56)       ] ^
            C1[(int)(state[0] >> 48) & 0xff] ^
            C2[(int)(state[7] >> 40) & 0xff] ^
            C3[(int)(state[6] >> 32) & 0xff] ^
            C4[(int)(state[5] >> 24) & 0xff] ^
            C5[(int)(state[4] >> 16) & 0xff] ^
            C6[(int)(state[3] >>  8) & 0xff] ^
            C7[(int)(state[2]      ) & 0xff] ^
            K[1];
        L[2] =
            C0[(int)(state[2] >> 56)       ] ^
            C1[(int)(state[1] >> 48) & 0xff] ^
            C2[(int)(state[0] >> 40) & 0xff] ^
            C3[(int)(state[7] >> 32) & 0xff] ^
            C4[(int)(state[6] >> 24) & 0xff] ^
            C5[(int)(state[5] >> 16) & 0xff] ^
            C6[(int)(state[4] >>  8) & 0xff] ^
            C7[(int)(state[3]      ) & 0xff] ^
            K[2];
        L[3] =
            C0[(int)(state[3] >> 56)       ] ^
            C1[(int)(state[2] >> 48) & 0xff] ^
            C2[(int)(state[1] >> 40) & 0xff] ^
            C3[(int)(state[0] >> 32) & 0xff] ^
            C4[(int)(state[7] >> 24) & 0xff] ^
            C5[(int)(state[6] >> 16) & 0xff] ^
            C6[(int)(state[5] >>  8) & 0xff] ^
            C7[(int)(state[4]      ) & 0xff] ^
            K[3];
        L[4] =
            C0[(int)(state[4] >> 56)       ] ^
            C1[(int)(state[3] >> 48) & 0xff] ^
            C2[(int)(state[2] >> 40) & 0xff] ^
            C3[(int)(state[1] >> 32) & 0xff] ^
            C4[(int)(state[0] >> 24) & 0xff] ^
            C5[(int)(state[7] >> 16) & 0xff] ^
            C6[(int)(state[6] >>  8) & 0xff] ^
            C7[(int)(state[5]      ) & 0xff] ^
            K[4];
        L[5] =
            C0[(int)(state[5] >> 56)       ] ^
            C1[(int)(state[4] >> 48) & 0xff] ^
            C2[(int)(state[3] >> 40) & 0xff] ^
            C3[(int)(state[2] >> 32) & 0xff] ^
            C4[(int)(state[1] >> 24) & 0xff] ^
            C5[(int)(state[0] >> 16) & 0xff] ^
            C6[(int)(state[7] >>  8) & 0xff] ^
            C7[(int)(state[6]      ) & 0xff] ^
            K[5];
        L[6] =
            C0[(int)(state[6] >> 56)       ] ^
            C1[(int)(state[5] >> 48) & 0xff] ^
            C2[(int)(state[4] >> 40) & 0xff] ^
            C3[(int)(state[3] >> 32) & 0xff] ^
            C4[(int)(state[2] >> 24) & 0xff] ^
            C5[(int)(state[1] >> 16) & 0xff] ^
            C6[(int)(state[0] >>  8) & 0xff] ^
            C7[(int)(state[7]      ) & 0xff] ^
            K[6];
        L[7] =
            C0[(int)(state[7] >> 56)       ] ^
            C1[(int)(state[6] >> 48) & 0xff] ^
            C2[(int)(state[5] >> 40) & 0xff] ^
            C3[(int)(state[4] >> 32) & 0xff] ^
            C4[(int)(state[3] >> 24) & 0xff] ^
            C5[(int)(state[2] >> 16) & 0xff] ^
            C6[(int)(state[1] >>  8) & 0xff] ^
            C7[(int)(state[0]      ) & 0xff] ^
            K[7];
        state[0] = L[0];
        state[1] = L[1];
        state[2] = L[2];
        state[3] = L[3];
        state[4] = L[4];
        state[5] = L[5];
        state[6] = L[6];
        state[7] = L[7];
#ifdef TRACE_INTERMEDIATE_VALUES
    printf("i = %d:\n", r);
    for (i = 0; i < DIGESTBYTES/8; i++) {
        printf(
            "    %02X %02X %02X %02X %02X %02X %02X %02X        %02X %02X %02X %02X %02X %02X %02X %02X\n",
            (u8)(K[i] >> 56),
            (u8)(K[i] >> 48),
            (u8)(K[i] >> 40),
            (u8)(K[i] >> 32),
            (u8)(K[i] >> 24),
            (u8)(K[i] >> 16),
            (u8)(K[i] >>  8),
            (u8)(K[i]      ),

            (u8)(state[i] >> 56),
            (u8)(state[i] >> 48),
            (u8)(state[i] >> 40),
            (u8)(state[i] >> 32),
            (u8)(state[i] >> 24),
            (u8)(state[i] >> 16),
            (u8)(state[i] >>  8),
            (u8)(state[i]      ));
    }
    printf("\n");
#endif /* ?TRACE_INTERMEDIATE_VALUES */
    }
    /*
     * apply the Miyaguchi-Preneel compression function:
     */
    structpointer->hash[0] ^= state[0] ^ block[0];
    structpointer->hash[1] ^= state[1] ^ block[1];
    structpointer->hash[2] ^= state[2] ^ block[2];
    structpointer->hash[3] ^= state[3] ^ block[3];
    structpointer->hash[4] ^= state[4] ^ block[4];
    structpointer->hash[5] ^= state[5] ^ block[5];
    structpointer->hash[6] ^= state[6] ^ block[6];
    structpointer->hash[7] ^= state[7] ^ block[7];
#ifdef TRACE_INTERMEDIATE_VALUES
    //printf("Intermediate hash value (after Miyaguchi-Preneel):\n");
    printf("The value of Y' output from the round-function is as follows.\n");
    for (i = 0; i < DIGESTBYTES/8; i++) {
        printf("    %02X %02X %02X %02X %02X %02X %02X %02X\n",
            (u8)(structpointer->hash[i] >> 56),
            (u8)(structpointer->hash[i] >> 48),
            (u8)(structpointer->hash[i] >> 40),
            (u8)(structpointer->hash[i] >> 32),
            (u8)(structpointer->hash[i] >> 24),
            (u8)(structpointer->hash[i] >> 16),
            (u8)(structpointer->hash[i] >>  8),
            (u8)(structpointer->hash[i]      ));
    }
    printf("\n");
#endif /* ?TRACE_INTERMEDIATE_VALUES */
}

/**
 * Initialize the hashing state.
 */
void NESSIEinit(struct NESSIEstruct * const structpointer) {
    int i;

    memset(structpointer->bitLength, 0, 32);
    structpointer->bufferBits = structpointer->bufferPos = 0;
    structpointer->buffer[0] = 0; /* it's only necessary to cleanup buffer[bufferPos] */
    for (i = 0; i < 8; i++) {
        structpointer->hash[i] = 0L; /* initial value */
    }
#ifdef TRACE_INTERMEDIATE_VALUES
    /*
    printf("Initial hash value:\n");
    for (i = 0; i < DIGESTBYTES/8; i++) {
        printf("    %02X %02X %02X %02X %02X %02X %02X %02X\n",
            (u8)(structpointer->hash[i] >> 56),
            (u8)(structpointer->hash[i] >> 48),
            (u8)(structpointer->hash[i] >> 40),
            (u8)(structpointer->hash[i] >> 32),
            (u8)(structpointer->hash[i] >> 24),
            (u8)(structpointer->hash[i] >> 16),
            (u8)(structpointer->hash[i] >>  8),
            (u8)(structpointer->hash[i]      ));
    }
    printf("\n");
    */
#endif /* ?TRACE_INTERMEDIATE_VALUES */
}

/**
 * Delivers input data to the hashing algorithm.
 *
 * @param    source        plaintext data to hash.
 * @param    sourceBits    how many bits of plaintext to process.
 *
 * This method maintains the invariant: bufferBits < DIGESTBITS
 */
void NESSIEadd(const unsigned char * const source,
               unsigned long sourceBits,
               struct NESSIEstruct * const structpointer) {
    /*
                       sourcePos
                       |
                       +-------+-------+-------
                          ||||||||||||||||||||| source
                       +-------+-------+-------
    +-------+-------+-------+-------+-------+-------
    ||||||||||||||||||||||                           buffer
    +-------+-------+-------+-------+-------+-------
                    |
                    bufferPos
    */
    int sourcePos    = 0; /* index of leftmost source u8 containing data (1 to 8 bits). */
    int sourceGap    = (8 - ((int)sourceBits & 7)) & 7; /* space on source[sourcePos]. */
    int bufferRem    = structpointer->bufferBits & 7; /* occupied bits on buffer[bufferPos]. */
    int i;
    u32 b, carry;
    u8 *buffer       = structpointer->buffer;
    u8 *bitLength    = structpointer->bitLength;
    int bufferBits   = structpointer->bufferBits;
    int bufferPos    = structpointer->bufferPos;

    /*
     * tally the length of the added data:
     */
    u64 value = sourceBits;
    for (i = 31, carry = 0; i >= 0 && (carry != 0 || value != LL(0)); i--) {
        carry += bitLength[i] + ((u32)value & 0xff);
        bitLength[i] = (u8)carry;
        carry >>= 8;
        value >>= 8;
    }
    /*
     * process data in chunks of 8 bits (a more efficient approach would be to take whole-word chunks):
     */
    while (sourceBits > 8) {
        /* N.B. at least source[sourcePos] and source[sourcePos+1] contain data. */
        /*
         * take a byte from the source:
         */
        b = ((source[sourcePos] << sourceGap) & 0xff) |
            ((source[sourcePos + 1] & 0xff) >> (8 - sourceGap));
        /*
         * process this byte:
         */
        buffer[bufferPos++] |= (u8)(b >> bufferRem);
        bufferBits += 8 - bufferRem; /* bufferBits = 8*bufferPos; */
        if (bufferBits == DIGESTBITS) {
            /*
             * process data block:
             */
            processBuffer(structpointer);
            /*
             * reset buffer:
             */
            bufferBits = bufferPos = 0;
        }
        buffer[bufferPos] = b << (8 - bufferRem);
        bufferBits += bufferRem;
        /*
         * proceed to remaining data:
         */
        sourceBits -= 8;
        sourcePos++;
    }
    /* now 0 <= sourceBits <= 8;
     * furthermore, all data (if any is left) is in source[sourcePos].
     */
    if (sourceBits > 0) {
        b = (source[sourcePos] << sourceGap) & 0xff; /* bits are left-justified on b. */
        /*
         * process the remaining bits:
         */
        buffer[bufferPos] |= b >> bufferRem;
    } else {
        b = 0;
    }
    if (bufferRem + sourceBits < 8) {
        /*
         * all remaining data fits on buffer[bufferPos],
         * and there still remains some space.
         */
        bufferBits += sourceBits;
    } else {
        /*
         * buffer[bufferPos] is full:
         */
        bufferPos++;
        bufferBits += 8 - bufferRem; /* bufferBits = 8*bufferPos; */
        sourceBits -= 8 - bufferRem;
        /* now 0 <= sourceBits < 8;
         * furthermore, all data (if any is left) is in source[sourcePos].
         */
        if (bufferBits == DIGESTBITS) {
            /*
             * process data block:
             */
            processBuffer(structpointer);
            /*
             * reset buffer:
             */
            bufferBits = bufferPos = 0;
        }
        buffer[bufferPos] = b << (8 - bufferRem);
        bufferBits += (int)sourceBits;
    }
    structpointer->bufferBits   = bufferBits;
    structpointer->bufferPos    = bufferPos;
}

/**
 * Get the hash value from the hashing state.
 * 
 * This method uses the invariant: bufferBits < DIGESTBITS
 */
void NESSIEfinalize(struct NESSIEstruct * const structpointer,
                    unsigned char * const result) {
    int i;
    u8 *buffer      = structpointer->buffer;
    u8 *bitLength   = structpointer->bitLength;
    int bufferBits  = structpointer->bufferBits;
    int bufferPos   = structpointer->bufferPos;
    u8 *digest      = result;

    /*
     * append a '1'-bit:
     */
    buffer[bufferPos] |= 0x80U >> (bufferBits & 7);
    bufferPos++; /* all remaining bits on the current u8 are set to zero. */
    /*
     * pad with zero bits to complete (N*WBLOCKBITS - LENGTHBITS) bits:
     */
    if (bufferPos > WBLOCKBYTES - LENGTHBYTES) {
        if (bufferPos < WBLOCKBYTES) {
            memset(&buffer[bufferPos], 0, WBLOCKBYTES - bufferPos);
        }
        /*
         * process data block:
         */
        processBuffer(structpointer);
        /*
         * reset buffer:
         */
        bufferPos = 0;
    }
    if (bufferPos < WBLOCKBYTES - LENGTHBYTES) {
        memset(&buffer[bufferPos], 0, (WBLOCKBYTES - LENGTHBYTES) - bufferPos);
    }
    bufferPos = WBLOCKBYTES - LENGTHBYTES;
    /*
     * append bit length of hashed data:
     */
    memcpy(&buffer[WBLOCKBYTES - LENGTHBYTES], bitLength, LENGTHBYTES);
    /*
     * process data block:
     */
    processBuffer(structpointer);
    /*
     * return the completed message digest:
     */
    for (i = 0; i < DIGESTBYTES/8; i++) {
        digest[0] = (u8)(structpointer->hash[i] >> 56);
        digest[1] = (u8)(structpointer->hash[i] >> 48);
        digest[2] = (u8)(structpointer->hash[i] >> 40);
        digest[3] = (u8)(structpointer->hash[i] >> 32);
        digest[4] = (u8)(structpointer->hash[i] >> 24);
        digest[5] = (u8)(structpointer->hash[i] >> 16);
        digest[6] = (u8)(structpointer->hash[i] >>  8);
        digest[7] = (u8)(structpointer->hash[i]      );
        digest += 8;
    }
    structpointer->bufferBits   = bufferBits;
    structpointer->bufferPos    = bufferPos;
}

static void display(const u8 array[], int length) {
    int i;
    for (i = 0; i < length; i++) {
        if (i%32 == 0) {
            printf("\n");
        }
        if (i%8 == 0) {
            printf(" ");
        }
        printf("%02X", array[i]);
    }
}

#define LONG_ITERATION 100000000

/**
 * Generate the test vector set for Whirlpool.
 *
 * The test consists of:
 * 1. hashing all bit strings containing only zero bits
 *    for all lengths from 0 to 1023;
 * 2. hashing all 512-bit strings containing a single set bit;
 * 3. the iterated hashing of the 512-bit string of zero bits a large number of times.
 */
void makeNESSIETestVectors() {
    int i;
    struct NESSIEstruct w;
    u8 digest[DIGESTBYTES];
    u8 data[128];

    memset(data, 0, sizeof(data));
    printf("Message digests of strings of 0-bits and length L:\n");
    for (i = 0; i < 1024; i++) {
        NESSIEinit(&w);
        NESSIEadd(data, i, &w);
        NESSIEfinalize(&w, digest);
        printf("    L = %4d: ", i); display(digest, DIGESTBYTES); printf("\n");
    }
    printf("Message digests of all 512-bit strings S containing a single 1-bit:\n");
    memset(data, 0, sizeof(data));
    for (i = 0; i < 512; i++) {
        /* set bit i: */
        data[i/8] |= 0x80U >> (i % 8);
        NESSIEinit(&w);
        NESSIEadd(data, 512, &w);
        NESSIEfinalize(&w, digest);
        printf("    S = "); display(data, 512/8); printf(": ");
        display(digest, DIGESTBYTES); printf("\n");
        /* reset bit i: */
        data[i/8] = 0;
    }
    memset(digest, 0, sizeof(digest));
    for (i = 0; i < LONG_ITERATION; i++) {
        NESSIEinit(&w);
        NESSIEadd(digest, 512, &w);
        NESSIEfinalize(&w, digest);
    }
    fflush(stdout);
    printf("Iterated message digest computation (%d times): ", LONG_ITERATION);
    display(digest, DIGESTBYTES); printf("\n");
}



/*
#define TIMING_ITERATIONS 100000

static void timing() {
    int i;
    NESSIEstruct w;
    u8 digest[DIGESTBYTES];
    u8 data[1024];
    clock_t elapsed;
    float sec;
    
    memset(data, 0, sizeof(data));

    printf("Overall timing...");
    elapsed = -clock();
    for (i = 0; i < TIMING_ITERATIONS; i++) {
        NESSIEinit(&w);
        NESSIEadd(data, 8*sizeof(data), &w);
        NESSIEfinalize(&w, digest);
    }
    elapsed += clock();
    sec = (float)elapsed/CLOCKS_PER_SEC;
    printf(" %.1f s, %.1f Mbit/s, %.1f cycles/byte.\n",
        sec,
        (float)8*sizeof(data)*TIMING_ITERATIONS/sec/1000000,
        (float)550e6*sec/(sizeof(data)*TIMING_ITERATIONS));

    printf("Compression function timing...");
    NESSIEinit(&w);
    elapsed = -clock();
    for (i = 0; i < TIMING_ITERATIONS; i++) {
        processBuffer(&w);
    }
    elapsed += clock();
    NESSIEfinalize(&w, digest);
    sec = (float)elapsed/CLOCKS_PER_SEC;
    printf(" %.1f s, %.1f Mbit/s, %.1f cycles/byte.\n",
        sec,
        (float)512*TIMING_ITERATIONS/sec/1000000,
        (float)550e6*sec/(64*TIMING_ITERATIONS));
}
*/

void testAPI(void) {
    u32 pieceLen, totalLen, dataLen;
    NESSIEstruct w;
    u8 dataBuf[512], expectedDigest[DIGESTBYTES], computedDigest[DIGESTBYTES];

    for (dataLen = 0; dataLen <= sizeof(dataBuf); dataLen++) {
        if ((dataLen & 0xff) == 0) {
            fprintf(stderr, "."); fflush(stderr);
        }
        /*
         * do the hashing in pieces of variable length:
         */
        NESSIEinit(&w);
        NESSIEadd(dataBuf, 8*dataLen, &w);
        NESSIEfinalize(&w, expectedDigest);
        if (dataLen > 0) {
            for (pieceLen = 1; pieceLen <= dataLen; pieceLen++) {
                NESSIEinit(&w);
                for (totalLen = 0; totalLen + pieceLen <= dataLen; totalLen += pieceLen) {
                    NESSIEadd(dataBuf + totalLen, 8*pieceLen, &w);
                }
                if (totalLen < dataLen) {
                    NESSIEadd(dataBuf + totalLen, 8*(dataLen - totalLen), &w);
                }
                NESSIEfinalize(&w, computedDigest);
                if (memcmp(computedDigest, expectedDigest, DIGESTBYTES) != 0) {
                    fprintf(stderr, "API error @ pieceLen = %lu\n", pieceLen);
                    display(computedDigest, DIGESTBYTES); printf("\n\n");
                    display(expectedDigest, DIGESTBYTES); printf("\n\n");
                    return;
                }
            }
        } else {
            NESSIEinit(&w);
            NESSIEfinalize(&w, computedDigest);
            if (memcmp(computedDigest, expectedDigest, DIGESTBYTES) != 0) {
                fprintf(stderr, "API error @ pieceLen = 0\n");
                return;
            }
        }
    }
    printf("No error detected.\n");
}


// those are the Whirlpool test vectors
static char *whirlpool512_msg[] = 
{
  "",
	"a",
	"abc",
	"message digest",
	"abcdefghijklmnopqrstuvwxyz",
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
	"12345678901234567890123456789012345678901234567890123456789012345678901234567890",
	"abcdbcdecdefdefgefghfghighijhijk",
	NULL
};

static char *whirlpool512_val[] =
{
   "19fa61d75522a4669b44e39c1d2e1726c530232130d407f89afee0964997f7a7" \
	"3e83be698b288febcf88e3e03c4f0757ea8964e59b63d93708b138cc42a66eb3",

	"8aca2602792aec6f11a67206531fb7d7f0dff59413145e6973c45001d0087b42" \
	"d11bc645413aeff63a42391a39145a591a92200d560195e53b478584fdae231a",

	"4e2448a4c6f486bb16b6562c73b4020bf3043e3a731bce721ae1b303d97e6d4c" \
	"7181eebdb6c57e277d0e34957114cbd6c797fc9d95d8b582d225292076d4eef5",

	"378c84a4126e2dc6e56dcc7458377aac838d00032230f53ce1f5700c0ffb4d3b" \
	"8421557659ef55c106b4b52ac5a4aaa692ed920052838f3362e86dbd37a8903e",

	"f1d754662636ffe92c82ebb9212a484a8d38631ead4238f5442ee13b8054e41b" \
	"08bf2a9251c30b6a0b8aae86177ab4a6f68f673e7207865d5d9819a3dba4eb3b",
	
	"dc37e008cf9ee69bf11f00ed9aba26901dd7c28cdec066cc6af42e40f82f3a1e" \
	"08eba26629129d8fb7cb57211b9281a65517cc879d7b962142c65f5a7af01467",

	"466ef18babb0154d25b9d38a6414f5c08784372bccb204d6549c4afadb601429" \
	"4d5bd8df2a6c44e538cd047b2681a51a2c60481e88c5a20b2c2a80cf3a9a083b",

	"2a987ea40f917061f5d6f0a0e4644f488a7a5a52deee656207c562f988e95c69" \
	"16bdc8031bc5be1b7b947639fe050b56939baaa0adff9ae6745b7b181c3be3fd",
	
	"0c99005beb57eff50a7cf005560ddf5d29057fd86b20bfd62deca0f1ccea4af5" \
	"1fc15490eddc47af32bb2b66c34ff9ad8c6008ad677f77126953b226e4ed8b01"
};


bool whirlpool_test(void) {
  struct NESSIEstruct w;
  u8 digest[DIGESTBYTES];
	int i, j;
	char output[140];
	unsigned char buf[1000];
			
	for( i = 0; i < 9; i++ ) {
    NESSIEinit(&w);
		if (i<8) {
			NESSIEadd((const unsigned char *) whirlpool512_msg[i], 8*(int)strlen( whirlpool512_msg[i] ), &w);
		}
		else {
			memset( buf, 'a', 1000 );
				
			for( j = 0; j < 1000; j++ ) {
				NESSIEadd((const unsigned char *)buf, 8*1000, &w);
			}
		}

    NESSIEfinalize(&w, digest);

		for( j = 0; j < 64; j++ ) {
			sprintf_s( output + j * 2,3, "%02x", digest[j] );
		}
			
		if( memcmp( output, whirlpool512_val[i], 128 ) ) {
			return false;
		}
	}
	return true;
};

#undef R




















//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
/* This is an independent implementation of the encryption algorithm:   */
/*                                                                      */
/*         Twofish by Bruce Schneier and colleagues                     */
/*                                                                      */
/* which is a candidate algorithm in the Advanced Encryption Standard   */
/* programme of the US National Institute of Standards and Technology.  */
/*                                                                      */
/* Copyright in this implementation is held by Dr B R Gladman but I     */
/* hereby give permission for its free direct or derivative use subject */
/* to acknowledgment of its origin and compliance with any conditions   */
/* that the originators of t he algorithm place on its exploitation.    */
/*                                                                      */
/* My thanks to Doug Whiting and Niels Ferguson for comments that led   */
/* to improvements in this implementation.                              */
/*                                                                      */
/* Dr Brian Gladman (gladman@seven77.demon.co.uk) 14th January 1999     */

/* modified in order to use the libmcrypt API by Nikos Mavroyanopoulos 
 * All modifications are placed under the license of libmcrypt.
 */

/* $Id: twofish.c,v 1.16 2003/01/19 17:48:27 nmav Exp $ */

/* Timing data for Twofish (twofish.c)

   128 bit key:
   Key Setup:    8414 cycles
   Encrypt:       376 cycles =    68.1 mbits/sec
   Decrypt:       374 cycles =    68.4 mbits/sec
   Mean:          375 cycles =    68.3 mbits/sec

   192 bit key:
   Key Setup:   11628 cycles
   Encrypt:       376 cycles =    68.1 mbits/sec
   Decrypt:       374 cycles =    68.4 mbits/sec
   Mean:          375 cycles =    68.3 mbits/sec

   256 bit key:
   Key Setup:   15457 cycles
   Encrypt:       381 cycles =    67.2 mbits/sec
   Decrypt:       374 cycles =    68.4 mbits/sec
   Mean:          378 cycles =    67.8 mbits/sec

 */


	/* Extract byte from a 32 bit quantity (little endian notation)     */
	#define byte(x,n)   ((byte)((x) >> (8 * n)))

	/* finite field arithmetic for GF(2**8) with the modular    */
	/* polynomial x^8 + x^6 + x^5 + x^3 + 1 (0x169)             */

	#define G_M 0x0169

	byte tab_5b[4] = { 0, G_M >> 2, G_M >> 1, (G_M >> 1) ^ (G_M >> 2) };
	byte tab_ef[4] = { 0, (G_M >> 1) ^ (G_M >> 2), G_M >> 1, G_M >> 2 };

	#define ffm_01(x)    (x)
	#define ffm_5b(x)   ((x) ^ ((x) >> 2) ^ tab_5b[(x) & 3])
	#define ffm_ef(x)   ((x) ^ ((x) >> 1) ^ ((x) >> 2) ^ tab_ef[(x) & 3])

	byte ror4[16] = { 0, 8, 1, 9, 2, 10, 3, 11, 4, 12, 5, 13, 6, 14, 7, 15 };
	byte ashx[16] = { 0, 9, 2, 11, 4, 13, 6, 15, 8, 1, 10, 3, 12, 5, 14, 7 };

	byte qt0[2][16] = {
		{8, 1, 7, 13, 6, 15, 3, 2, 0, 11, 5, 9, 14, 12, 10, 4}
		,
		{2, 8, 11, 13, 15, 7, 6, 14, 3, 1, 9, 4, 0, 10, 12, 5}
	};

	byte qt1[2][16] = {
		{14, 12, 11, 8, 1, 2, 3, 5, 15, 4, 10, 6, 7, 0, 9, 13}
		,
		{1, 14, 2, 11, 4, 12, 3, 7, 6, 13, 10, 5, 15, 9, 0, 8}
	};

	byte qt2[2][16] = {
		{11, 10, 5, 14, 6, 13, 9, 0, 12, 8, 15, 3, 2, 4, 7, 1}
		,
		{4, 12, 7, 5, 1, 6, 9, 10, 0, 14, 13, 8, 2, 11, 3, 15}
	};

	byte qt3[2][16] = {
		{13, 7, 15, 4, 1, 2, 6, 14, 9, 11, 3, 0, 8, 5, 12, 10}
		,
		{11, 9, 5, 1, 12, 3, 13, 14, 6, 4, 7, 15, 2, 0, 8, 10}
	};

	byte qp(const word32 n, const byte x)
	{
		byte a0, a1, a2, a3, a4, b0, b1, b2, b3, b4;

		a0 = x >> 4;
		b0 = x & 15;
		a1 = a0 ^ b0;
		b1 = ror4[b0] ^ ashx[a0];
		a2 = qt0[n][a1];
		b2 = qt1[n][b1];
		a3 = a2 ^ b2;
		b3 = ror4[b2] ^ ashx[a2];
		a4 = qt2[n][a3];
		b4 = qt3[n][b3];
		return (b4 << 4) | a4;
	}

	#ifdef  Q_TABLES

	#define q(n,x)  pkey->q_tab[n][x]

	void gen_qtab(TWI * pkey)
	{
		word32 i;

		for (i = 0; i < 256; ++i) {
			q(0, i) = qp(0, (byte) i);
			q(1, i) = qp(1, (byte) i);
		}
	}

	#else

	#define q(n,x)  qp(n, x)

	#endif

	#ifdef  M_TABLE

	void gen_mtab(TWI * pkey)
	{
		word32 i, f01, f5b, fef;

		for (i = 0; i < 256; ++i) {
			f01 = q(1, i);
			f5b = ffm_5b(f01);
			fef = ffm_ef(f01);
			pkey->m_tab[0][i] =
					f01 + (f5b << 8) + (fef << 16) + (fef << 24);
			pkey->m_tab[2][i] =
					f5b + (fef << 8) + (f01 << 16) + (fef << 24);

			f01 = q(0, i);
			f5b = ffm_5b(f01);
			fef = ffm_ef(f01);
			pkey->m_tab[1][i] =
					fef + (fef << 8) + (f5b << 16) + (f01 << 24);
			pkey->m_tab[3][i] =
					f5b + (f01 << 8) + (fef << 16) + (f5b << 24);
		}
	}

	#define mds(n,x)    pkey->m_tab[n][x]

	#else

	#define fm_00   ffm_01
	#define fm_10   ffm_5b
	#define fm_20   ffm_ef
	#define fm_30   ffm_ef
	#define q_0(x)  q(1,x)

	#define fm_01   ffm_ef
	#define fm_11   ffm_ef
	#define fm_21   ffm_5b
	#define fm_31   ffm_01
	#define q_1(x)  q(0,x)

	#define fm_02   ffm_5b
	#define fm_12   ffm_ef
	#define fm_22   ffm_01
	#define fm_32   ffm_ef
	#define q_2(x)  q(1,x)

	#define fm_03   ffm_5b
	#define fm_13   ffm_01
	#define fm_23   ffm_ef
	#define fm_33   ffm_5b
	#define q_3(x)  q(0,x)

	#define f_0(n,x)    ((word32)fm_0##n(x))
	#define f_1(n,x)    ((word32)fm_1##n(x) << 8)
	#define f_2(n,x)    ((word32)fm_2##n(x) << 16)
	#define f_3(n,x)    ((word32)fm_3##n(x) << 24)

	#define mds(n,x)    f_0(n,q_##n(x)) ^ f_1(n,q_##n(x)) ^ f_2(n,q_##n(x)) ^ f_3(n,q_##n(x))

	#endif

	word32 h_fun(TWI * pkey, const word32 x, const word32 key[])
	{
		word32 b0, b1, b2, b3;

	#ifndef M_TABLE
		word32 m5b_b0, m5b_b1, m5b_b2, m5b_b3;
		word32 mef_b0, mef_b1, mef_b2, mef_b3;
	#endif

		b0 = byte(x, 0);
		b1 = byte(x, 1);
		b2 = byte(x, 2);
		b3 = byte(x, 3);

		switch (pkey->k_len) {
		case 4:
			b0 = q(1, b0) ^ byte(key[3], 0);
			b1 = q(0, b1) ^ byte(key[3], 1);
			b2 = q(0, b2) ^ byte(key[3], 2);
			b3 = q(1, b3) ^ byte(key[3], 3);
		case 3:
			b0 = q(1, b0) ^ byte(key[2], 0);
			b1 = q(1, b1) ^ byte(key[2], 1);
			b2 = q(0, b2) ^ byte(key[2], 2);
			b3 = q(0, b3) ^ byte(key[2], 3);
		case 2:
			b0 = q(0, q(0, b0) ^ byte(key[1], 0)) ^ byte(key[0], 0);
			b1 = q(0, q(1, b1) ^ byte(key[1], 1)) ^ byte(key[0], 1);
			b2 = q(1, q(0, b2) ^ byte(key[1], 2)) ^ byte(key[0], 2);
			b3 = q(1, q(1, b3) ^ byte(key[1], 3)) ^ byte(key[0], 3);
		}
	#ifdef  M_TABLE

		return mds(0, b0) ^ mds(1, b1) ^ mds(2, b2) ^ mds(3, b3);

	#else

		b0 = q(1, b0);
		b1 = q(0, b1);
		b2 = q(1, b2);
		b3 = q(0, b3);
		m5b_b0 = ffm_5b(b0);
		m5b_b1 = ffm_5b(b1);
		m5b_b2 = ffm_5b(b2);
		m5b_b3 = ffm_5b(b3);
		mef_b0 = ffm_ef(b0);
		mef_b1 = ffm_ef(b1);
		mef_b2 = ffm_ef(b2);
		mef_b3 = ffm_ef(b3);
		b0 ^= mef_b1 ^ m5b_b2 ^ m5b_b3;
		b3 ^= m5b_b0 ^ mef_b1 ^ mef_b2;
		b2 ^= mef_b0 ^ m5b_b1 ^ mef_b3;
		b1 ^= mef_b0 ^ mef_b2 ^ m5b_b3;

		return b0 | (b3 << 8) | (b2 << 16) | (b1 << 24);

	#endif
	}

	#ifdef  MK_TABLE

	#define q20(x)  q(0,q(0,x) ^ byte(key[1],0)) ^ byte(key[0],0)
	#define q21(x)  q(0,q(1,x) ^ byte(key[1],1)) ^ byte(key[0],1)
	#define q22(x)  q(1,q(0,x) ^ byte(key[1],2)) ^ byte(key[0],2)
	#define q23(x)  q(1,q(1,x) ^ byte(key[1],3)) ^ byte(key[0],3)

	#define q30(x)  q(0,q(0,q(1, x) ^ byte(key[2],0)) ^ byte(key[1],0)) ^ byte(key[0],0)
	#define q31(x)  q(0,q(1,q(1, x) ^ byte(key[2],1)) ^ byte(key[1],1)) ^ byte(key[0],1)
	#define q32(x)  q(1,q(0,q(0, x) ^ byte(key[2],2)) ^ byte(key[1],2)) ^ byte(key[0],2)
	#define q33(x)  q(1,q(1,q(0, x) ^ byte(key[2],3)) ^ byte(key[1],3)) ^ byte(key[0],3)

	#define q40(x)  q(0,q(0,q(1, q(1, x) ^ byte(key[3],0)) ^ byte(key[2],0)) ^ byte(key[1],0)) ^ byte(key[0],0)
	#define q41(x)  q(0,q(1,q(1, q(0, x) ^ byte(key[3],1)) ^ byte(key[2],1)) ^ byte(key[1],1)) ^ byte(key[0],1)
	#define q42(x)  q(1,q(0,q(0, q(0, x) ^ byte(key[3],2)) ^ byte(key[2],2)) ^ byte(key[1],2)) ^ byte(key[0],2)
	#define q43(x)  q(1,q(1,q(0, q(1, x) ^ byte(key[3],3)) ^ byte(key[2],3)) ^ byte(key[1],3)) ^ byte(key[0],3)

	void gen_mk_tab(TWI * pkey, word32 key[])
	{
		word32 i;
		byte by;

		switch (pkey->k_len) {
		case 2:
			for (i = 0; i < 256; ++i) {
				by = (byte) i;
	#ifdef ONE_STEP
				pkey->mk_tab[0][i] = mds(0, q20(by));
				pkey->mk_tab[1][i] = mds(1, q21(by));
				pkey->mk_tab[2][i] = mds(2, q22(by));
				pkey->mk_tab[3][i] = mds(3, q23(by));
	#else
				pkey->sb[0][i] = q20(by);
				pkey->sb[1][i] = q21(by);
				pkey->sb[2][i] = q22(by);
				pkey->sb[3][i] = q23(by);
	#endif
			}
			break;

		case 3:
			for (i = 0; i < 256; ++i) {
				by = (byte) i;
	#ifdef ONE_STEP
				pkey->mk_tab[0][i] = mds(0, q30(by));
				pkey->mk_tab[1][i] = mds(1, q31(by));
				pkey->mk_tab[2][i] = mds(2, q32(by));
				pkey->mk_tab[3][i] = mds(3, q33(by));
	#else
				pkey->sb[0][i] = q30(by);
				pkey->sb[1][i] = q31(by);
				pkey->sb[2][i] = q32(by);
				pkey->sb[3][i] = q33(by);
	#endif
			}
			break;

		case 4:
			for (i = 0; i < 256; ++i) {
				by = (byte) i;
	#ifdef ONE_STEP
				pkey->mk_tab[0][i] = mds(0, q40(by));
				pkey->mk_tab[1][i] = mds(1, q41(by));
				pkey->mk_tab[2][i] = mds(2, q42(by));
				pkey->mk_tab[3][i] = mds(3, q43(by));
	#else
				pkey->sb[0][i] = q40(by);
				pkey->sb[1][i] = q41(by);
				pkey->sb[2][i] = q42(by);
				pkey->sb[3][i] = q43(by);
	#endif
			}
		}
	}

	#ifdef ONE_STEP
	#define g0_fun(x) ( pkey->mk_tab[0][byte(x,0)] ^ pkey->mk_tab[1][byte(x,1)] \
												^ pkey->mk_tab[2][byte(x,2)] ^ pkey->mk_tab[3][byte(x,3)] )
	#define g1_fun(x) ( pkey->mk_tab[0][byte(x,3)] ^ pkey->mk_tab[1][byte(x,0)] \
												^ pkey->mk_tab[2][byte(x,1)] ^ pkey->mk_tab[3][byte(x,2)] )
	#else
	#define g0_fun(x) ( mds(0, pkey->sb[0][byte(x,0)]) ^ mds(1, pkey->sb[1][byte(x,1)]) \
												^ mds(2, pkey->sb[2][byte(x,2)]) ^ mds(3, pkey->sb[3][byte(x,3)]) )
	#define g1_fun(x) ( mds(0, pkey->sb[0][byte(x,3)]) ^ mds(1, pkey->sb[1][byte(x,0)]) \
												^ mds(2, pkey->sb[2][byte(x,1)]) ^ mds(3, pkey->sb[3][byte(x,2)]) )
	#endif

	#else

	#define g0_fun(x)   h_fun(pkey, x,pkey->s_key)
	#define g1_fun(x)   h_fun(pkey, rotl32(x,8),pkey->s_key)

	#endif

	/* The (12,8) Reed Soloman code has the generator polynomial

		 g(x) = x^4 + (a + 1/a) * x^3 + a * x^2 + (a + 1/a) * x + 1

		 where the coefficients are in the finite field GF(2^8) with a
		 modular polynomial a^8 + a^6 + a^3 + a^2 + 1. To generate the
		 remainder we have to start with a 12th order polynomial with our
		 eight input bytes as the coefficients of the 4th to 11th terms. 
		 That is:

		 m[7] * x^11 + m[6] * x^10 ... + m[0] * x^4 + 0 * x^3 +... + 0

		 We then multiply the generator polynomial by m[7] * x^7 and subtract
		 it - xor in GF(2^8) - from the above to eliminate the x^7 term (the 
		 artihmetic on the coefficients is done in GF(2^8). We then multiply 
		 the generator polynomial by x^6 * coeff(x^10) and use this to remove
		 the x^10 term. We carry on in this way until the x^4 term is removed
		 so that we are left with:

		 r[3] * x^3 + r[2] * x^2 + r[1] 8 x^1 + r[0]

		 which give the resulting 4 bytes of the remainder. This is equivalent 
		 to the matrix multiplication in the Twofish description but much faster 
		 to implement.

	 */

	#define G_MOD   0x0000014d

	word32 mds_rem(word32 p0, word32 p1)
	{
		word32 i, t, u;

		for (i = 0; i < 8; ++i) {
			t = p1 >> 24;	/* get most significant coefficient */

			p1 = (p1 << 8) | (p0 >> 24);
			p0 <<= 8;	/* shift others up */

			/* multiply t by a (the primitive element - i.e. left shift) */

			u = (t << 1);

			if (t & 0x80)
				/* subtract modular polynomial on overflow */
				u ^= G_MOD;

			p1 ^= t ^ (u << 16);	/* remove t * (a * x^2 + 1)  */

			u ^= (t >> 1);	/* form u = a * t + t / a = t * (a + 1 / a); */

			if (t & 0x01)
				/* add the modular polynomial on underflow */
				u ^= G_MOD >> 1;

			p1 ^= (u << 24) | (u << 8);	/* remove t * (a + 1/a) * (x^3 + x) */

		}

		return p1;
	}




	int twofish_set_key(TWI * pkey, const word32 in_key[],const word32 key_len)
	{
		word32 i, a, b, me_key[4], mo_key[4];

	#ifdef Q_TABLES
		pkey->qt_gen = 0;

		if (!pkey->qt_gen) {
			gen_qtab(pkey);
			pkey->qt_gen = 1;
		}
	#endif

	#ifdef M_TABLE
		pkey->mt_gen = 0;
		if (!pkey->mt_gen) {
			gen_mtab(pkey);
			pkey->mt_gen = 1;
		}
	#endif

		pkey->k_len = (key_len * 8) / 64;	/* 2, 3 or 4 */

		for (i = 0; i < pkey->k_len; ++i) {
	#ifdef WORDS_BIGENDIAN
			a = byteswap32(in_key[i + i]);
			me_key[i] = a;
			b = byteswap32(in_key[i + i + 1]);
	#else
			a = in_key[i + i];
			me_key[i] = a;
			b = in_key[i + i + 1];
	#endif
			mo_key[i] = b;
			pkey->s_key[pkey->k_len - i - 1] = mds_rem(a, b);
		}

		for (i = 0; i < 40; i += 2) {
			a = 0x01010101 * i;
			b = a + 0x01010101;
			a = h_fun(pkey, a, me_key);
			b = rotl32(h_fun(pkey, b, mo_key), 8);
			pkey->l_key[i] = a + b;
			pkey->l_key[i + 1] = rotl32(a + 2 * b, 9);
		}

	#ifdef MK_TABLE
		gen_mk_tab(pkey, pkey->s_key);
	#endif
		return 0;
	}

	/* This macro was moved to an inline function, because it
	 * was breaking some compilers.
	 */
	inline 
	static void f_rnd(int i, word32* blk, TWI* pkey, word32 t0, word32 t1)
	{
			t1 = g1_fun(blk[1]); 
			t0 = g0_fun(blk[0]);
	    
			blk[2] = rotr32(blk[2] ^ (t0 + t1 + pkey->l_key[4 * (i) + 8]), 1);

			blk[3] = rotl32(blk[3], 1) ^ (t0 + 2 * t1 + pkey->l_key[4 * (i) + 9]);
			t1 = g1_fun(blk[3]); 
			t0 = g0_fun(blk[2]);
	    
			blk[0] = rotr32(blk[0] ^ (t0 + t1 + pkey->l_key[4 * (i) + 10]), 1);
			blk[1] = rotl32(blk[1], 1) ^ (t0 + 2 * t1 + pkey->l_key[4 * (i) + 11]);
	}

	/* encrypt a block of text  */
	void twofish_encrypt(TWI * pkey, word32 * in_blk)
	{
		word32 t0=0, t1=0, blk[4];
	#ifdef WORDS_BIGENDIAN
		blk[0] = byteswap32(in_blk[0]) ^ pkey->l_key[0];
		blk[1] = byteswap32(in_blk[1]) ^ pkey->l_key[1];
		blk[2] = byteswap32(in_blk[2]) ^ pkey->l_key[2];
		blk[3] = byteswap32(in_blk[3]) ^ pkey->l_key[3];
	#else
		blk[0] = in_blk[0] ^ pkey->l_key[0];
		blk[1] = in_blk[1] ^ pkey->l_key[1];
		blk[2] = in_blk[2] ^ pkey->l_key[2];
		blk[3] = in_blk[3] ^ pkey->l_key[3];
	#endif

		f_rnd(0, blk, pkey, t0, t1);
		f_rnd(1, blk, pkey, t0, t1);
		f_rnd(2, blk, pkey, t0, t1);
		f_rnd(3, blk, pkey, t0, t1);
		f_rnd(4, blk, pkey, t0, t1);
		f_rnd(5, blk, pkey, t0, t1);
		f_rnd(6, blk, pkey, t0, t1);
		f_rnd(7, blk, pkey, t0, t1);

	#ifdef WORDS_BIGENDIAN
		in_blk[0] = byteswap32(blk[2] ^ pkey->l_key[4]);
		in_blk[1] = byteswap32(blk[3] ^ pkey->l_key[5]);
		in_blk[2] = byteswap32(blk[0] ^ pkey->l_key[6]);
		in_blk[3] = byteswap32(blk[1] ^ pkey->l_key[7]);
	#else
		in_blk[0] = blk[2] ^ pkey->l_key[4];
		in_blk[1] = blk[3] ^ pkey->l_key[5];
		in_blk[2] = blk[0] ^ pkey->l_key[6];
		in_blk[3] = blk[1] ^ pkey->l_key[7];
	#endif
	}

	/* decrypt a block of text  */

	#define i_rnd(i)                                                        \
					t1 = g1_fun(blk[1]); t0 = g0_fun(blk[0]);                       \
					blk[2] = rotl32(blk[2], 1) ^ (t0 + t1 + pkey->l_key[4 * (i) + 10]);     \
					blk[3] = rotr32(blk[3] ^ (t0 + 2 * t1 + pkey->l_key[4 * (i) + 11]), 1); \
					t1 = g1_fun(blk[3]); t0 = g0_fun(blk[2]);                       \
					blk[0] = rotl32(blk[0], 1) ^ (t0 + t1 + pkey->l_key[4 * (i) +  8]);     \
					blk[1] = rotr32(blk[1] ^ (t0 + 2 * t1 + pkey->l_key[4 * (i) +  9]), 1)

	void twofish_decrypt(TWI * pkey, word32 * in_blk)
	{
		word32 t0, t1, blk[4];

	#ifdef WORDS_BIGENDIAN
		blk[0] = byteswap32(in_blk[0]) ^ pkey->l_key[4];
		blk[1] = byteswap32(in_blk[1]) ^ pkey->l_key[5];
		blk[2] = byteswap32(in_blk[2]) ^ pkey->l_key[6];
		blk[3] = byteswap32(in_blk[3]) ^ pkey->l_key[7];
	#else
		blk[0] = in_blk[0] ^ pkey->l_key[4];
		blk[1] = in_blk[1] ^ pkey->l_key[5];
		blk[2] = in_blk[2] ^ pkey->l_key[6];
		blk[3] = in_blk[3] ^ pkey->l_key[7];
	#endif

		i_rnd(7);
		i_rnd(6);
		i_rnd(5);
		i_rnd(4);
		i_rnd(3);
		i_rnd(2);
		i_rnd(1);
		i_rnd(0);

	#ifdef WORDS_BIGENDIAN
		in_blk[0] = byteswap32(blk[2] ^ pkey->l_key[0]);
		in_blk[1] = byteswap32(blk[3] ^ pkey->l_key[1]);
		in_blk[2] = byteswap32(blk[0] ^ pkey->l_key[2]);
		in_blk[3] = byteswap32(blk[1] ^ pkey->l_key[3]);
	#else
		in_blk[0] = blk[2] ^ pkey->l_key[0];
		in_blk[1] = blk[3] ^ pkey->l_key[1];
		in_blk[2] = blk[0] ^ pkey->l_key[2];
		in_blk[3] = blk[1] ^ pkey->l_key[3];
	#endif

	}



	#define TWOFISH_CIPHER "019f9809de1711858faac3a3ba20fbc3"
	#define TWOFISH_PT "\xD4\x91\xDB\x16\xE7\xB1\xC3\x9E\x86\xCB\x08\x6B\x78\x9F\x54\x19"
	#define TWOFISH_KEY "\x9F\x58\x9F\x5C\xF6\x12\x2C\x32\xB6\xBF\xEC\x2F\x2A\xE8\xC3\x5A"

	bool twofish_self_test()
	{
		unsigned char keyword[16];
		unsigned char plaintext[16];
		unsigned char ciphertext[16];
		int blocksize = 16; // 16 bytes = 128 bit (it's a 128 bit block cipher)
		int j;
		twofish_context ctx;
		unsigned char cipher_tmp[220];

		memcpy( keyword, TWOFISH_KEY, 16);
		memcpy( plaintext, TWOFISH_PT, 16);

		memcpy(ciphertext, plaintext, 16);
		
		
		twofish_set_key(&ctx,(const unsigned __int32 *) keyword, 16); // 16:128 bit, 24: 192 bit, 32: 256 bit key 

		twofish_encrypt(&ctx, (unsigned __int32 *) ciphertext);

		for (j = 0; j < blocksize; j++) {
			sprintf_s(&((char *) cipher_tmp)[2 * j],3, "%.2x",
				ciphertext[j]);
		}

		if (strcmp((char *) cipher_tmp, TWOFISH_CIPHER) != 0) {
			printf("failed compatibility\n");
			printf("Expected: %s\nGot: %s\n", TWOFISH_CIPHER,
						 (char *) cipher_tmp);
			return false;
		}
		twofish_decrypt(&ctx, (unsigned __int32 *) ciphertext);

		if (memcmp(ciphertext, plaintext, 16) != 0) {
			printf("failed internally\n");
			return false;
		}
		return true;
	}

	#undef i_rnd


















//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
	#define TIGER_PASSES 3

	extern const word32 tiger_table[4*256][2];

	#define h0init 0x89ABCDEF;
	#define h1init 0x01234567;
	#define h2init 0x76543210;
	#define h3init 0xFEDCBA98;
	#define h4init 0xC3B2E187;
	#define h5init 0xF096A5B4;

	#define t1 (tiger_table)
	#define t2 (tiger_table+256)
	#define t3 (tiger_table+256*2)
	#define t4 (tiger_table+256*3)

	#define sub64(s0, s1, p0, p1) \
				temps0 = (p0); \
				tcarry = s0 < temps0; \
				s0 -= temps0; \
				s1 -= (p1) + tcarry;

	#define add64(s0, s1, p0, p1) \
				temps0 = (p0); \
				s0 += temps0; \
				tcarry = s0 < temps0; \
				s1 += (p1) + tcarry;

	#define xor64(s0, s1, p0, p1) \
				s0 ^= (p0); \
				s1 ^= (p1);

	#define mul5(s0, s1) \
				tempt0 = s0<<2; \
				tempt1 = (s1<<2)|(s0>>30); \
				add64(s0, s1, tempt0, tempt1);

	#define mul7(s0, s1) \
				tempt0 = s0<<3; \
				tempt1 = (s1<<3)|(s0>>29); \
				sub64(tempt0, tempt1, s0, s1); \
				s0 = tempt0; \
				s1 = tempt1;

	#define mul9(s0, s1) \
				tempt0 = s0<<3; \
				tempt1 = (s1<<3)|(s0>>29); \
				add64(s0, s1, tempt0, tempt1);

	#define save_abc \
				aa0 = a0; \
				aa1 = a1; \
				bb0 = b0; \
				bb1 = b1; \
				cc0 = c0; \
				cc1 = c1;

	#define round(a0,a1,b0,b1,c0,c1,x0,x1,mul) \
				xor64(c0, c1, x0, x1); \
				temp0  = t1[((c0)>>(0*8))&0xFF][0] ; \
				temp1  = t1[((c0)>>(0*8))&0xFF][1] ; \
				temp0 ^= t2[((c0)>>(2*8))&0xFF][0] ; \
				temp1 ^= t2[((c0)>>(2*8))&0xFF][1] ; \
				temp0 ^= t3[((c1)>>(0*8))&0xFF][0] ; \
				temp1 ^= t3[((c1)>>(0*8))&0xFF][1] ; \
				temp0 ^= t4[((c1)>>(2*8))&0xFF][0] ; \
				temp1 ^= t4[((c1)>>(2*8))&0xFF][1] ; \
				sub64(a0, a1, temp0, temp1); \
				temp0  = t4[((c0)>>(1*8))&0xFF][0] ; \
				temp1  = t4[((c0)>>(1*8))&0xFF][1] ; \
				temp0 ^= t3[((c0)>>(3*8))&0xFF][0] ; \
				temp1 ^= t3[((c0)>>(3*8))&0xFF][1] ; \
				temp0 ^= t2[((c1)>>(1*8))&0xFF][0] ; \
				temp1 ^= t2[((c1)>>(1*8))&0xFF][1] ; \
				temp0 ^= t1[((c1)>>(3*8))&0xFF][0] ; \
				temp1 ^= t1[((c1)>>(3*8))&0xFF][1] ; \
				add64(b0, b1, temp0, temp1); \
				if((mul)==5) \
		{mul5(b0, b1);} \
				else \
		if((mul)==7) \
			{mul7(b0, b1);} \
		else \
			{mul9(b0, b1)};

	#define pass(a0,a1,b0,b1,c0,c1,mul) \
				round(a0,a1,b0,b1,c0,c1,x00,x01,mul); \
				round(b0,b1,c0,c1,a0,a1,x10,x11,mul); \
				round(c0,c1,a0,a1,b0,b1,x20,x21,mul); \
				round(a0,a1,b0,b1,c0,c1,x30,x31,mul); \
				round(b0,b1,c0,c1,a0,a1,x40,x41,mul); \
				round(c0,c1,a0,a1,b0,b1,x50,x51,mul); \
				round(a0,a1,b0,b1,c0,c1,x60,x61,mul); \
				round(b0,b1,c0,c1,a0,a1,x70,x71,mul);

	#define key_schedule \
				sub64(x00, x01, x70^0xA5A5A5A5, x71^0xA5A5A5A5); \
				xor64(x10, x11, x00, x01); \
				add64(x20, x21, x10, x11); \
				sub64(x30, x31, x20^((~x10)<<19), ~x21^(((x11)<<19)|((x10)>>13))); \
				xor64(x40, x41, x30, x31); \
				add64(x50, x51, x40, x41); \
				sub64(x60, x61, ~x50^(((x40)>>23)|((x41)<<9)), x51^((~x41)>>23)); \
				xor64(x70, x71, x60, x61); \
				add64(x00, x01, x70, x71); \
				sub64(x10, x11, x00^((~x70)<<19), ~x01^(((x71)<<19)|((x70)>>13))); \
				xor64(x20, x21, x10, x11); \
				add64(x30, x31, x20, x21); \
				sub64(x40, x41, ~x30^(((x20)>>23)|((x21)<<9)), x31^((~x21)>>23)); \
				xor64(x50, x51, x40, x41); \
				add64(x60, x61, x50, x51); \
				sub64(x70, x71, x60^0x89ABCDEF, x61^0x01234567);

	#define feedforward \
				xor64(a0, a1, aa0, aa1); \
				sub64(b0, b1, bb0, bb1); \
				add64(c0, c1, cc0, cc1);

	#ifdef UNROLL_COMPRESS
	#define compress \
				save_abc \
				pass(a0,a1,b0,b1,c0,c1,5); \
				key_schedule; \
				pass(c0,c1,a0,a1,b0,b1,7); \
				key_schedule; \
				pass(b0,b1,c0,c1,a0,a1,9); \
				for(pass_no=3; pass_no<TIGER_PASSES; pass_no++) { \
					key_schedule \
		pass(a0,a1,b0,b1,c0,c1,9); \
		tmpa=a0; a0=c0; c0=b0; b0=tmpa; \
		tmpa=a1; a1=c1; c1=b1; b1=tmpa;} \
				feedforward
	#else
	#define compress \
				save_abc \
				for(pass_no=0; pass_no<TIGER_PASSES; pass_no++) { \
					if(pass_no != 0) {key_schedule} \
		pass(a0,a1,b0,b1,c0,c1,(pass_no==0?5:pass_no==1?7:9)) \
		tmpa=a0; a0=c0; c0=b0; b0=tmpa; \
		tmpa=a1; a1=c1; c1=b1; b1=tmpa;} \
				feedforward
	#endif

	#define tiger_compress_macro(str, state) \
	{ \
		register word32 a0, a1, b0, b1, c0, c1, tmpa; \
		word32 aa0, aa1, bb0, bb1, cc0, cc1; \
		word32 x00, x01, x10, x11, x20, x21, x30, x31, \
					 x40, x41, x50, x51, x60, x61, x70, x71; \
		register word32 temp0, temp1, tempt0, tempt1, temps0, tcarry; \
		int pass_no; \
	\
		a0 = state[0]; \
		a1 = state[1]; \
		b0 = state[2]; \
		b1 = state[3]; \
		c0 = state[4]; \
		c1 = state[5]; \
	\
		x00=str[0*2]; x01=str[0*2+1]; x10=str[1*2]; x11=str[1*2+1]; \
		x20=str[2*2]; x21=str[2*2+1]; x30=str[3*2]; x31=str[3*2+1]; \
		x40=str[4*2]; x41=str[4*2+1]; x50=str[5*2]; x51=str[5*2+1]; \
		x60=str[6*2]; x61=str[6*2+1]; x70=str[7*2]; x71=str[7*2+1]; \
	\
		compress; \
	\
		state[0] = a0; \
		state[1] = a1; \
		state[2] = b0; \
		state[3] = b1; \
		state[4] = c0; \
		state[5] = c1; \
	}

	#define UNROLL_COMPRESS
	#ifdef UNROLL_COMPRESS
	/* The compress function is inlined */
	#define tiger_compress(str, state) \
		tiger_compress_macro(((word32*)str), ((word32*)state))
	#else
	/* The compress function is a function */
	tiger_compress(const word32 *str, word32 state[6])
	{
		tiger_compress_macro(((word32*)str), ((word32*)state));
	}
	#endif


	#ifndef EXTRACT_UCHAR
	#define EXTRACT_UCHAR(p)  (*(word8 *)(p))
	#endif

	#define STRING2INT(s) ((((((EXTRACT_UCHAR(s+3) << 8)    \
				 | EXTRACT_UCHAR(s+2)) << 8)  \
				 | EXTRACT_UCHAR(s+1)) << 8)  \
				 | EXTRACT_UCHAR(s))

	static void tiger_block(struct tiger_ctx *ctx, word8 * block)
	{
		word32 data[TIGER_DATALEN];
		int i;
		
		/* Update block count */
		if (!++ctx->count_l)
			++ctx->count_h;

		/* Endian independent conversion */
		for (i = 0; i < TIGER_DATALEN; i++, block += sizeof(word32)) data[i] = STRING2INT(block);

		tiger_compress(data, ctx->digest);
	}

	void tiger_init(struct tiger_ctx *ctx)
	{
		/* Set the h-vars to their initial values */
		ctx->digest[0] = h0init;
		ctx->digest[1] = h1init;
		ctx->digest[2] = h2init;
		ctx->digest[3] = h3init;
		ctx->digest[4] = h4init;
		ctx->digest[5] = h5init;

		/* Initialize bit count */
		ctx->count_l = ctx->count_h = 0;

		/* Initialize buffer */
		ctx->index = 0;
	}

	void tiger_update(struct tiger_ctx *ctx, word8 * buffer, word32 len)
	{
		if (ctx->index) {	/* Try to fill partial block */
			unsigned left = TIGER_DATASIZE - ctx->index;
			if (len < left) {
				memcpy(ctx->block + ctx->index, buffer, len);
				ctx->index += len;
				return;	/* Finished */
			} else {
				memcpy(ctx->block + ctx->index, buffer, left);
					 tiger_block(ctx, ctx->block);
				buffer += left;
				len -= left;
			}
		}
		while (len >= TIGER_DATASIZE) {
				 tiger_block(ctx, buffer);
			buffer += TIGER_DATASIZE;
			len -= TIGER_DATASIZE;
		}
		if ((ctx->index = len))
			/* This assignment is intended */
			/* Buffer leftovers */
			memcpy(ctx->block, buffer, len);
	}

	void tiger_final(struct tiger_ctx *ctx)
	{
		word32 data[TIGER_DATALEN];
		int i;
		int words;

		i = ctx->index;
		/* Set the first char of padding to 0x01.  This is safe since there is
			 always at least one byte free */
		ctx->block[i++] = 0x01;

		/* Fill rest of word */
		for (; i & 3; i++)
			ctx->block[i] = 0;

		/* i is now a multiple of the word size 4 */
		words = i >> 2;
		for (i = 0; i < words; i++)
			data[i] = STRING2INT(ctx->block + 4 * i);

		if (words > (TIGER_DATALEN - 2)) {	/* No room for length in this block. Process it and
							 * pad with another one */
			for (i = words; i < TIGER_DATALEN; i++)
				data[i] = 0;
			tiger_compress( data, ctx->digest);
			for (i = 0; i < (TIGER_DATALEN - 2); i++)
				data[i] = 0;
		} else
			for (i = words; i < TIGER_DATALEN - 2; i++)
				data[i] = 0;
		/* Theres 512 = 2^9 bits in one block */
		data[TIGER_DATALEN - 1] =
				(ctx->count_h << 9) | (ctx->count_l >> 23);
		data[TIGER_DATALEN - 2] = (ctx->count_l << 9) | (ctx->index << 3);
		tiger_compress( data, ctx->digest);
	}

	void tiger192_digest(struct tiger_ctx *ctx, word8 * s)
	{
		int i;

		if (s!=NULL)
		for (i = 0; i < TIGER192_DIGESTLEN; i+=2) { /* 64 bit and LITTLE ENDIAN -that's cool! */
			s[7] = (unsigned char)ctx->digest[i];
			s[6] = 0xff & (unsigned char)(ctx->digest[i] >> 8);
			s[5] = 0xff & (unsigned char)(ctx->digest[i] >> 16);
			s[4] = 0xff & (unsigned char)(ctx->digest[i] >> 24);
			s[3] = (unsigned char)ctx->digest[i+1];
			s[2] = 0xff & (unsigned char)(ctx->digest[i+1] >> 8);
			s[1] = 0xff & (unsigned char)(ctx->digest[i+1] >> 16);
			s[0] = 0xff & (unsigned char)(ctx->digest[i+1] >> 24);
			s+=8;
		}

	}

	void tiger128_digest(struct tiger_ctx *ctx, word8 * s)
	{
		int i;

		if (s!=NULL)
		for (i = 0; i < TIGER128_DIGESTLEN; i+=2) { /* 64 bit and LITTLE ENDIAN -that's cool! */
			s[7] = (unsigned char)ctx->digest[i];
			s[6] = 0xff & (unsigned char)(ctx->digest[i] >> 8);
			s[5] = 0xff & (unsigned char)(ctx->digest[i] >> 16);
			s[4] = 0xff & (unsigned char)(ctx->digest[i] >> 24);
			s[3] = (unsigned char)ctx->digest[i+1];
			s[2] = 0xff & (unsigned char)(ctx->digest[i+1] >> 8);
			s[1] = 0xff & (unsigned char)(ctx->digest[i+1] >> 16);
			s[0] = 0xff & (unsigned char)(ctx->digest[i+1] >> 24);
			s+=8;
		}

	}


	void tiger160_digest(struct tiger_ctx *ctx, word8 * s)
	{
		int i;

		for (i = 0; i < TIGER160_DIGESTLEN-1; i+=2) { /* 64 bit and LITTLE ENDIAN -that's cool! */
			s[7] = (unsigned char)ctx->digest[i];
			s[6] = 0xff & (unsigned char)(ctx->digest[i] >> 8);
			s[5] = 0xff & (unsigned char)(ctx->digest[i] >> 16);
			s[4] = 0xff & (unsigned char)(ctx->digest[i] >> 24);
			s[3] = (unsigned char)ctx->digest[i+1];
			s[2] = 0xff & (unsigned char)(ctx->digest[i+1] >> 8);
			s[1] = 0xff & (unsigned char)(ctx->digest[i+1] >> 16);
			s[0] = 0xff & (unsigned char)(ctx->digest[i+1] >> 24);
			s+=8;
		}
		/* last block */
		s[3] = (unsigned char)ctx->digest[i+1];
		s[2] = 0xff & (unsigned char)(ctx->digest[i+1] >> 8);
		s[1] = 0xff & (unsigned char)(ctx->digest[i+1] >> 16);
		s[0] = 0xff & (unsigned char)(ctx->digest[i+1] >> 24);
	}






	// those are the Tiger192 test vectors (also for Tiger128 and Tiger160)
	static char *tiger192_msg[] = 
	{
			"",
			"abc",
			"Tiger",
			"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-",
			"ABCDEFGHIJKLMNOPQRSTUVWXYZ=abcdefghijklmnopqrstuvwxyz+0123456789",
			"Tiger - A Fast New Hash Function, by Ross Anderson and Eli Biham",
			"Tiger - A Fast New Hash Function, by Ross Anderson and Eli Biham, proceedings of Fast Software Encryption 3, Cambridge.",
			"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-"
			
	};

	static char *tiger192_val[] =
	{
			"24f0130c63ac933216166e76b1bb925ff373de2d49584e7a",
			"f258c1e88414ab2a527ab541ffc5b8bf935f7b951c132951",
			"9f00f599072300dd276abb38c8eb6dec37790c116f9d2bdf",
			"87fb2a9083851cf7470d2cf810e6df9eb586445034a5a386",
			"467db80863ebce488df1cd1261655de957896565975f9197",
			"0c410a042968868a1671da5a3fd29a725ec1e457d3cdb303",
			"ebf591d5afa655ce7f22894ff87f54ac89c811b6b0da3193",
			"00b83eb4e53440c576ac6aaee0a7485825fd15e70a59ffe4"
	};



		// tests with standard test vectors and returns true if everything is OK
	bool tiger_128_test(void) {
		int i, j;
		char output[36];
		tiger_ctx ctx;
		unsigned char tiger_sum[17];
				
		for( i = 0; i < 3; i++ ) {
			tiger_init( &ctx );
			tiger_update( &ctx, (word8 *) tiger192_msg[i],(int)strlen( tiger192_msg[i] ) );
				
			tiger_final( &ctx );
			tiger128_digest( &ctx, tiger_sum );


			for( j = 0; j < 16; j++ ) {
				sprintf_s( output + j * 2,3, "%02x", tiger_sum[j] );
			}
				
			if( memcmp( output, tiger192_val[i], 32 ) ) {
				return false;
			}
		}
		return true;
	};


		// tests with standard test vectors and returns true if everything is OK
	bool tiger_160_test(void) {
		int i, j;
		char output[45];
		tiger_ctx ctx;
		unsigned char tiger_sum[25];
				
		for( i = 0; i < 3; i++ ) {
			tiger_init( &ctx );
			tiger_update( &ctx, (word8 *) tiger192_msg[i],(int)strlen( tiger192_msg[i] ) );
				
			tiger_final( &ctx );
			tiger160_digest( &ctx, tiger_sum );


			for( j = 0; j < 20; j++ ) {
				sprintf_s( output + j * 2,3, "%02x", tiger_sum[j] );
			}
				
			if( memcmp( output, tiger192_val[i], 40 ) ) {
				return false;
			}
		}
		return true;
	};




		// tests with standard test vectors and returns true if everything is OK
	bool tiger_192_test(void) {
		int i, j;
		char output[55];
		tiger_ctx ctx;
		unsigned char tiger_sum[25];
				
		for( i = 0; i < 8; i++ ) {
			tiger_init( &ctx );
			tiger_update( &ctx, (word8 *) tiger192_msg[i],(int)strlen( tiger192_msg[i] ) );
				
			tiger_final( &ctx );
			tiger192_digest( &ctx, tiger_sum );
			
			for( j = 0; j < 24; j++ ) {
				sprintf_s( output + j * 2,3, "%02x", tiger_sum[j] );
			}
				
			if( memcmp( output, tiger192_val[i], 48 ) ) {
				return false;
			}
		}
		return true;
	};



	#undef t1
	#undef t2
	#undef t3
	#undef t4

	#undef h0init
	#undef h1init
	#undef h2init
	#undef h3init
	#undef h4init
	#undef h5init

















//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
/*
 *  FIPS-180-2 compliant SHA-256 implementation
 *
 *  Copyright (C) 2001-2003  Christophe Devine
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
	// start of SHA-2 functions
	#define GET_UINT32(n,b,i)                       \
	{                                               \
			(n) = ( (uint32) (b)[(i)    ] << 24 )       \
					| ( (uint32) (b)[(i) + 1] << 16 )       \
					| ( (uint32) (b)[(i) + 2] <<  8 )       \
					| ( (uint32) (b)[(i) + 3]       );      \
	}

	#define PUT_UINT32(n,b,i)                       \
	{                                               \
			(b)[(i)    ] = (uint8) ( (n) >> 24 );       \
			(b)[(i) + 1] = (uint8) ( (n) >> 16 );       \
			(b)[(i) + 2] = (uint8) ( (n) >>  8 );       \
			(b)[(i) + 3] = (uint8) ( (n)       );       \
	}

	void sha256_starts( sha256_context *ctx )
	{
			ctx->total[0] = 0;
			ctx->total[1] = 0;

			ctx->state[0] = 0x6A09E667;
			ctx->state[1] = 0xBB67AE85;
			ctx->state[2] = 0x3C6EF372;
			ctx->state[3] = 0xA54FF53A;
			ctx->state[4] = 0x510E527F;
			ctx->state[5] = 0x9B05688C;
			ctx->state[6] = 0x1F83D9AB;
			ctx->state[7] = 0x5BE0CD19;
	}

	void sha256_process( sha256_context *ctx, uint8 data[64] )
	{
			uint32 temp1, temp2, W[64];
			uint32 A, B, C, D, E, F, G, H;

			GET_UINT32( W[0],  data,  0 );
			GET_UINT32( W[1],  data,  4 );
			GET_UINT32( W[2],  data,  8 );
			GET_UINT32( W[3],  data, 12 );
			GET_UINT32( W[4],  data, 16 );
			GET_UINT32( W[5],  data, 20 );
			GET_UINT32( W[6],  data, 24 );
			GET_UINT32( W[7],  data, 28 );
			GET_UINT32( W[8],  data, 32 );
			GET_UINT32( W[9],  data, 36 );
			GET_UINT32( W[10], data, 40 );
			GET_UINT32( W[11], data, 44 );
			GET_UINT32( W[12], data, 48 );
			GET_UINT32( W[13], data, 52 );
			GET_UINT32( W[14], data, 56 );
			GET_UINT32( W[15], data, 60 );

	#define  SHR(x,n) ((x & 0xFFFFFFFF) >> n)
	#define ROTR(x,n) (SHR(x,n) | (x << (32 - n)))

	#define S0(x) (ROTR(x, 7) ^ ROTR(x,18) ^  SHR(x, 3))
	#define S1(x) (ROTR(x,17) ^ ROTR(x,19) ^  SHR(x,10))

	#define S2(x) (ROTR(x, 2) ^ ROTR(x,13) ^ ROTR(x,22))
	#define S3(x) (ROTR(x, 6) ^ ROTR(x,11) ^ ROTR(x,25))

	#define F0(x,y,z) ((x & y) | (z & (x | y)))
	#define F1(x,y,z) (z ^ (x & (y ^ z)))

	#define R(t)                                    \
	(                                               \
			W[t] = S1(W[t -  2]) + W[t -  7] +          \
						 S0(W[t - 15]) + W[t - 16]            \
	)

	#define P(a,b,c,d,e,f,g,h,x,K)                  \
	{                                               \
			temp1 = h + S3(e) + F1(e,f,g) + K + x;      \
			temp2 = S2(a) + F0(a,b,c);                  \
			d += temp1; h = temp1 + temp2;              \
	}

			A = ctx->state[0];
			B = ctx->state[1];
			C = ctx->state[2];
			D = ctx->state[3];
			E = ctx->state[4];
			F = ctx->state[5];
			G = ctx->state[6];
			H = ctx->state[7];

			P( A, B, C, D, E, F, G, H, W[ 0], 0x428A2F98 );
			P( H, A, B, C, D, E, F, G, W[ 1], 0x71374491 );
			P( G, H, A, B, C, D, E, F, W[ 2], 0xB5C0FBCF );
			P( F, G, H, A, B, C, D, E, W[ 3], 0xE9B5DBA5 );
			P( E, F, G, H, A, B, C, D, W[ 4], 0x3956C25B );
			P( D, E, F, G, H, A, B, C, W[ 5], 0x59F111F1 );
			P( C, D, E, F, G, H, A, B, W[ 6], 0x923F82A4 );
			P( B, C, D, E, F, G, H, A, W[ 7], 0xAB1C5ED5 );
			P( A, B, C, D, E, F, G, H, W[ 8], 0xD807AA98 );
			P( H, A, B, C, D, E, F, G, W[ 9], 0x12835B01 );
			P( G, H, A, B, C, D, E, F, W[10], 0x243185BE );
			P( F, G, H, A, B, C, D, E, W[11], 0x550C7DC3 );
			P( E, F, G, H, A, B, C, D, W[12], 0x72BE5D74 );
			P( D, E, F, G, H, A, B, C, W[13], 0x80DEB1FE );
			P( C, D, E, F, G, H, A, B, W[14], 0x9BDC06A7 );
			P( B, C, D, E, F, G, H, A, W[15], 0xC19BF174 );
			P( A, B, C, D, E, F, G, H, R(16), 0xE49B69C1 );
			P( H, A, B, C, D, E, F, G, R(17), 0xEFBE4786 );
			P( G, H, A, B, C, D, E, F, R(18), 0x0FC19DC6 );
			P( F, G, H, A, B, C, D, E, R(19), 0x240CA1CC );
			P( E, F, G, H, A, B, C, D, R(20), 0x2DE92C6F );
			P( D, E, F, G, H, A, B, C, R(21), 0x4A7484AA );
			P( C, D, E, F, G, H, A, B, R(22), 0x5CB0A9DC );
			P( B, C, D, E, F, G, H, A, R(23), 0x76F988DA );
			P( A, B, C, D, E, F, G, H, R(24), 0x983E5152 );
			P( H, A, B, C, D, E, F, G, R(25), 0xA831C66D );
			P( G, H, A, B, C, D, E, F, R(26), 0xB00327C8 );
			P( F, G, H, A, B, C, D, E, R(27), 0xBF597FC7 );
			P( E, F, G, H, A, B, C, D, R(28), 0xC6E00BF3 );
			P( D, E, F, G, H, A, B, C, R(29), 0xD5A79147 );
			P( C, D, E, F, G, H, A, B, R(30), 0x06CA6351 );
			P( B, C, D, E, F, G, H, A, R(31), 0x14292967 );
			P( A, B, C, D, E, F, G, H, R(32), 0x27B70A85 );
			P( H, A, B, C, D, E, F, G, R(33), 0x2E1B2138 );
			P( G, H, A, B, C, D, E, F, R(34), 0x4D2C6DFC );
			P( F, G, H, A, B, C, D, E, R(35), 0x53380D13 );
			P( E, F, G, H, A, B, C, D, R(36), 0x650A7354 );
			P( D, E, F, G, H, A, B, C, R(37), 0x766A0ABB );
			P( C, D, E, F, G, H, A, B, R(38), 0x81C2C92E );
			P( B, C, D, E, F, G, H, A, R(39), 0x92722C85 );
			P( A, B, C, D, E, F, G, H, R(40), 0xA2BFE8A1 );
			P( H, A, B, C, D, E, F, G, R(41), 0xA81A664B );
			P( G, H, A, B, C, D, E, F, R(42), 0xC24B8B70 );
			P( F, G, H, A, B, C, D, E, R(43), 0xC76C51A3 );
			P( E, F, G, H, A, B, C, D, R(44), 0xD192E819 );
			P( D, E, F, G, H, A, B, C, R(45), 0xD6990624 );
			P( C, D, E, F, G, H, A, B, R(46), 0xF40E3585 );
			P( B, C, D, E, F, G, H, A, R(47), 0x106AA070 );
			P( A, B, C, D, E, F, G, H, R(48), 0x19A4C116 );
			P( H, A, B, C, D, E, F, G, R(49), 0x1E376C08 );
			P( G, H, A, B, C, D, E, F, R(50), 0x2748774C );
			P( F, G, H, A, B, C, D, E, R(51), 0x34B0BCB5 );
			P( E, F, G, H, A, B, C, D, R(52), 0x391C0CB3 );
			P( D, E, F, G, H, A, B, C, R(53), 0x4ED8AA4A );
			P( C, D, E, F, G, H, A, B, R(54), 0x5B9CCA4F );
			P( B, C, D, E, F, G, H, A, R(55), 0x682E6FF3 );
			P( A, B, C, D, E, F, G, H, R(56), 0x748F82EE );
			P( H, A, B, C, D, E, F, G, R(57), 0x78A5636F );
			P( G, H, A, B, C, D, E, F, R(58), 0x84C87814 );
			P( F, G, H, A, B, C, D, E, R(59), 0x8CC70208 );
			P( E, F, G, H, A, B, C, D, R(60), 0x90BEFFFA );
			P( D, E, F, G, H, A, B, C, R(61), 0xA4506CEB );
			P( C, D, E, F, G, H, A, B, R(62), 0xBEF9A3F7 );
			P( B, C, D, E, F, G, H, A, R(63), 0xC67178F2 );

			ctx->state[0] += A;
			ctx->state[1] += B;
			ctx->state[2] += C;
			ctx->state[3] += D;
			ctx->state[4] += E;
			ctx->state[5] += F;
			ctx->state[6] += G;
			ctx->state[7] += H;
	}

	void sha256_update( sha256_context *ctx, uint8 *input, uint32 length )
	{
			uint32 left, fill;

			if( ! length ) return;

			left = ctx->total[0] & 0x3F;
			fill = 64 - left;

			ctx->total[0] += length;
			ctx->total[0] &= 0xFFFFFFFF;

			if( ctx->total[0] < length )
					ctx->total[1]++;

			if( left && length >= fill )
			{
					memcpy( (void *) (ctx->buffer + left),
									(void *) input, fill );
					sha256_process( ctx, ctx->buffer );
					length -= fill;
					input  += fill;
					left = 0;
			}

			while( length >= 64 )
			{
					sha256_process( ctx, input );
					length -= 64;
					input  += 64;
			}

			if( length )
			{
					memcpy( (void *) (ctx->buffer + left),
									(void *) input, length );
			}
	}

	static uint8 sha256_padding[64] =
	{
	 0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};

	void sha256_finish( sha256_context *ctx, uint8 digest[32] )
	{
			uint32 last, padn;
			uint32 high, low;
			uint8 msglen[8];

			high = ( ctx->total[0] >> 29 )
					 | ( ctx->total[1] <<  3 );
			low  = ( ctx->total[0] <<  3 );

			PUT_UINT32( high, msglen, 0 );
			PUT_UINT32( low,  msglen, 4 );

			last = ctx->total[0] & 0x3F;
			padn = ( last < 56 ) ? ( 56 - last ) : ( 120 - last );

			sha256_update( ctx, sha256_padding, padn );
			sha256_update( ctx, msglen, 8 );

			PUT_UINT32( ctx->state[0], digest,  0 );
			PUT_UINT32( ctx->state[1], digest,  4 );
			PUT_UINT32( ctx->state[2], digest,  8 );
			PUT_UINT32( ctx->state[3], digest, 12 );
			PUT_UINT32( ctx->state[4], digest, 16 );
			PUT_UINT32( ctx->state[5], digest, 20 );
			PUT_UINT32( ctx->state[6], digest, 24 );
			PUT_UINT32( ctx->state[7], digest, 28 );
	}




	// those are the standard FIPS-180-2 test vectors
	 
	static char *sha256_msg[] = 
	{
			"abc",
			"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
			NULL
	};

	static char *sha256_val[] =
	{
			"ba7816bf8f01cfea414140de5dae2223" \
			"b00361a396177a9cb410ff61f20015ad",
			"248d6a61d20638b8e5c026930c3e6039" \
			"a33ce45964ff2167f6ecedd419db06c1",
			"cdc76e5c9914fb9281a1c7e284d73e67" \
			"f1809a48a497200e046d39ccc7112cd0"
	};





		// tests with standard test vectors and returns true if everything is OK
	bool sha256_test(void) {
		int i, j;
		char output[69];
		sha256_context ctx;
		unsigned char buf[1000];
		unsigned char sha256sum[32];

				
		for( i = 0; i < 3; i++ ) {
			sha256_starts( &ctx );
			if( i < 2 ) {
				sha256_update( &ctx, (uint8 *) sha256_msg[i],(int)strlen( sha256_msg[i] ) );
			}
			else {
				memset( buf, 'a', 1000 );
					
				for( j = 0; j < 1000; j++ ) {
					sha256_update( &ctx, (uint8 *) buf, 1000 );
				}
			}
				
			sha256_finish( &ctx, sha256sum );


			for( j = 0; j < 32; j++ ) {
				sprintf_s( output + j * 2,3, "%02x", sha256sum[j] );
			}
				
			if( memcmp( output, sha256_val[i], 64 ) ) {
				return false;
			}
		}
		return true;
	}



	#undef R
	#undef P
	#undef F0
	#undef F1
	#undef ROTR
	#undef SHR
	#undef S0
	#undef S1
	#undef S2
	#undef S3
	#undef GET_UINT32
	#undef PUT_UINT32
















//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
























//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
/*
 *  FIPS-180-1 compliant SHA-1 implementation
 *
 *  Copyright (C) 2001-2003  Christophe Devine
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
	#define GET_UINT32(n,b,i)                       \
	{                                               \
			(n) = ( (uint32) (b)[(i)    ] << 24 )       \
					| ( (uint32) (b)[(i) + 1] << 16 )       \
					| ( (uint32) (b)[(i) + 2] <<  8 )       \
					| ( (uint32) (b)[(i) + 3]       );      \
	}

	#define PUT_UINT32(n,b,i)                       \
	{                                               \
			(b)[(i)    ] = (uint8) ( (n) >> 24 );       \
			(b)[(i) + 1] = (uint8) ( (n) >> 16 );       \
			(b)[(i) + 2] = (uint8) ( (n) >>  8 );       \
			(b)[(i) + 3] = (uint8) ( (n)       );       \
	}

	void sha1_starts( sha1_context *ctx )
	{
			ctx->total[0] = 0;
			ctx->total[1] = 0;

			ctx->state[0] = 0x67452301;
			ctx->state[1] = 0xEFCDAB89;
			ctx->state[2] = 0x98BADCFE;
			ctx->state[3] = 0x10325476;
			ctx->state[4] = 0xC3D2E1F0;
	}

	void sha1_process( sha1_context *ctx, uint8 data[64] )
	{
			uint32 temp, W[16], A, B, C, D, E;

			GET_UINT32( W[0],  data,  0 );
			GET_UINT32( W[1],  data,  4 );
			GET_UINT32( W[2],  data,  8 );
			GET_UINT32( W[3],  data, 12 );
			GET_UINT32( W[4],  data, 16 );
			GET_UINT32( W[5],  data, 20 );
			GET_UINT32( W[6],  data, 24 );
			GET_UINT32( W[7],  data, 28 );
			GET_UINT32( W[8],  data, 32 );
			GET_UINT32( W[9],  data, 36 );
			GET_UINT32( W[10], data, 40 );
			GET_UINT32( W[11], data, 44 );
			GET_UINT32( W[12], data, 48 );
			GET_UINT32( W[13], data, 52 );
			GET_UINT32( W[14], data, 56 );
			GET_UINT32( W[15], data, 60 );

	#define S(x,n) ((x << n) | ((x & 0xFFFFFFFF) >> (32 - n)))

	#define R(t)                                            \
	(                                                       \
			temp = W[(t -  3) & 0x0F] ^ W[(t - 8) & 0x0F] ^     \
						 W[(t - 14) & 0x0F] ^ W[ t      & 0x0F],      \
			( W[t & 0x0F] = S(temp,1) )                         \
	)

	#define P(a,b,c,d,e,x)                                  \
	{                                                       \
			e += S(a,5) + F(b,c,d) + K + x; b = S(b,30);        \
	}

			A = ctx->state[0];
			B = ctx->state[1];
			C = ctx->state[2];
			D = ctx->state[3];
			E = ctx->state[4];

	#define F(x,y,z) (z ^ (x & (y ^ z)))
	#define K 0x5A827999

			P( A, B, C, D, E, W[0]  );
			P( E, A, B, C, D, W[1]  );
			P( D, E, A, B, C, W[2]  );
			P( C, D, E, A, B, W[3]  );
			P( B, C, D, E, A, W[4]  );
			P( A, B, C, D, E, W[5]  );
			P( E, A, B, C, D, W[6]  );
			P( D, E, A, B, C, W[7]  );
			P( C, D, E, A, B, W[8]  );
			P( B, C, D, E, A, W[9]  );
			P( A, B, C, D, E, W[10] );
			P( E, A, B, C, D, W[11] );
			P( D, E, A, B, C, W[12] );
			P( C, D, E, A, B, W[13] );
			P( B, C, D, E, A, W[14] );
			P( A, B, C, D, E, W[15] );
			P( E, A, B, C, D, R(16) );
			P( D, E, A, B, C, R(17) );
			P( C, D, E, A, B, R(18) );
			P( B, C, D, E, A, R(19) );

	#undef K
	#undef F

	#define F(x,y,z) (x ^ y ^ z)
	#define K 0x6ED9EBA1

			P( A, B, C, D, E, R(20) );
			P( E, A, B, C, D, R(21) );
			P( D, E, A, B, C, R(22) );
			P( C, D, E, A, B, R(23) );
			P( B, C, D, E, A, R(24) );
			P( A, B, C, D, E, R(25) );
			P( E, A, B, C, D, R(26) );
			P( D, E, A, B, C, R(27) );
			P( C, D, E, A, B, R(28) );
			P( B, C, D, E, A, R(29) );
			P( A, B, C, D, E, R(30) );
			P( E, A, B, C, D, R(31) );
			P( D, E, A, B, C, R(32) );
			P( C, D, E, A, B, R(33) );
			P( B, C, D, E, A, R(34) );
			P( A, B, C, D, E, R(35) );
			P( E, A, B, C, D, R(36) );
			P( D, E, A, B, C, R(37) );
			P( C, D, E, A, B, R(38) );
			P( B, C, D, E, A, R(39) );

	#undef K
	#undef F

	#define F(x,y,z) ((x & y) | (z & (x | y)))
	#define K 0x8F1BBCDC

			P( A, B, C, D, E, R(40) );
			P( E, A, B, C, D, R(41) );
			P( D, E, A, B, C, R(42) );
			P( C, D, E, A, B, R(43) );
			P( B, C, D, E, A, R(44) );
			P( A, B, C, D, E, R(45) );
			P( E, A, B, C, D, R(46) );
			P( D, E, A, B, C, R(47) );
			P( C, D, E, A, B, R(48) );
			P( B, C, D, E, A, R(49) );
			P( A, B, C, D, E, R(50) );
			P( E, A, B, C, D, R(51) );
			P( D, E, A, B, C, R(52) );
			P( C, D, E, A, B, R(53) );
			P( B, C, D, E, A, R(54) );
			P( A, B, C, D, E, R(55) );
			P( E, A, B, C, D, R(56) );
			P( D, E, A, B, C, R(57) );
			P( C, D, E, A, B, R(58) );
			P( B, C, D, E, A, R(59) );

	#undef K
	#undef F

	#define F(x,y,z) (x ^ y ^ z)
	#define K 0xCA62C1D6

			P( A, B, C, D, E, R(60) );
			P( E, A, B, C, D, R(61) );
			P( D, E, A, B, C, R(62) );
			P( C, D, E, A, B, R(63) );
			P( B, C, D, E, A, R(64) );
			P( A, B, C, D, E, R(65) );
			P( E, A, B, C, D, R(66) );
			P( D, E, A, B, C, R(67) );
			P( C, D, E, A, B, R(68) );
			P( B, C, D, E, A, R(69) );
			P( A, B, C, D, E, R(70) );
			P( E, A, B, C, D, R(71) );
			P( D, E, A, B, C, R(72) );
			P( C, D, E, A, B, R(73) );
			P( B, C, D, E, A, R(74) );
			P( A, B, C, D, E, R(75) );
			P( E, A, B, C, D, R(76) );
			P( D, E, A, B, C, R(77) );
			P( C, D, E, A, B, R(78) );
			P( B, C, D, E, A, R(79) );

	#undef K
	#undef F

			ctx->state[0] += A;
			ctx->state[1] += B;
			ctx->state[2] += C;
			ctx->state[3] += D;
			ctx->state[4] += E;
	}

	void sha1_update( sha1_context *ctx, uint8 *input, uint32 length )
	{
			uint32 left, fill;

			if( ! length ) return;

			left = ctx->total[0] & 0x3F;
			fill = 64 - left;

			ctx->total[0] += length;
			ctx->total[0] &= 0xFFFFFFFF;

			if( ctx->total[0] < length )
					ctx->total[1]++;

			if( left && length >= fill )
			{
					memcpy( (void *) (ctx->buffer + left),
									(void *) input, fill );
					sha1_process( ctx, ctx->buffer );
					length -= fill;
					input  += fill;
					left = 0;
			}

			while( length >= 64 )
			{
					sha1_process( ctx, input );
					length -= 64;
					input  += 64;
			}

			if( length )
			{
					memcpy( (void *) (ctx->buffer + left),
									(void *) input, length );
			}
	}

	static uint8 sha1_padding[64] =
	{
	 0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};

	void sha1_finish( sha1_context *ctx, uint8 digest[20] )
	{
			uint32 last, padn;
			uint32 high, low;
			uint8 msglen[8];

			high = ( ctx->total[0] >> 29 )
					 | ( ctx->total[1] <<  3 );
			low  = ( ctx->total[0] <<  3 );

			PUT_UINT32( high, msglen, 0 );
			PUT_UINT32( low,  msglen, 4 );

			last = ctx->total[0] & 0x3F;
			padn = ( last < 56 ) ? ( 56 - last ) : ( 120 - last );

			sha1_update( ctx, sha1_padding, padn );
			sha1_update( ctx, msglen, 8 );

			PUT_UINT32( ctx->state[0], digest,  0 );
			PUT_UINT32( ctx->state[1], digest,  4 );
			PUT_UINT32( ctx->state[2], digest,  8 );
			PUT_UINT32( ctx->state[3], digest, 12 );
			PUT_UINT32( ctx->state[4], digest, 16 );
	}






	// those are the standard FIPS 180-1 test vectors

	static char *sha1_msg[] = 
	{
			"abc",
			"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
			NULL
	};

	static char *sha1_val[] =
	{
			"a9993e364706816aba3e25717850c26c9cd0d89d",
			"84983e441c3bd26ebaae4aa1f95129e5e54670f1",
			"34aa973cd4c4daa4f61eeb2bdbad27316534016f"
	};



		// tests with standard test vectors and returns true if everything is OK
	bool test_SHA_1(void) {
			bool result=true; // false=fail, true=pass
			int i, j;
			char output[44];
			struct sha1_context ctx;
			unsigned char sha1sum[20], buffer[1000];

			for( i = 0; i < 3; i++ )  {
					sha1_starts( &ctx );

					if( i < 2 ) {
							sha1_update( &ctx, (uint8 *) sha1_msg[i], (int)strlen( sha1_msg[i] ) );
					}
					else  {
							memset( buffer, 'a', 1000 );

							for( j = 0; j < 1000; j++ ) {
									sha1_update( &ctx, (uint8 *) buffer, 1000 );
							}
					}

					sha1_finish( &ctx, sha1sum );

					for( j = 0; j < 20; j++ ) {
							sprintf_s( output + j * 2,3, "%02x", sha1sum[j] );
					}

					if( ! memcmp( output, sha1_val[i], 40 ) ) {
							result=true;
					}
					else {
							return(false);
					}
			}
			return(result);
	}










//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
/* This is an independent implementation of the encryption algorithm:
 *
 * Serpent by Ross Anderson, Eli Biham and Lars Knudsen 
 *
 * which is a candidate algorithm in the Advanced Encryption Standard
 * programme of the US National Institute of Standards and Technology
 *
 * Copyright in this implementation is held by Dr B R Gladman but I
 * hereby give permission for its free direct or derivative use subject
 * to acknowledgment of its origin and compliance with any conditions
 * that the originators of the algorithm place on its exploitation.
 *
 * Dr Brian Gladman (gladman@seven77.demon.co.uk) 14th January 1999
 */

/* modified in order to use the libmcrypt API by Nikos Mavroyanopoulos 
 * All modifications are placed under the license of libmcrypt.
 */

/*
Algorithm serpent (serpent.c)

128 bit key:
Key Setup:    2366 cycles
Encrypt:       954 cycles =    26.8 mbits/sec
Decrypt:       907 cycles =    28.2 mbits/sec
Mean:          931 cycles =    27.5 mbits/sec

192 bit key:
Key Setup:    2382 cycles
Encrypt:       967 cycles =    26.5 mbits/sec
Decrypt:       915 cycles =    28.0 mbits/sec
Mean:          941 cycles =    27.2 mbits/sec

256 bit key:
Key Setup:    2360 cycles
Encrypt:       967 cycles =    26.5 mbits/sec
Decrypt:       915 cycles =    28.0 mbits/sec
Mean:          941 cycles =    27.2 mbits/sec
*/
	/* initialise the key schedule from the user supplied key   */
	int serpent_set_key(SERPENT_KEY * spkey, const word32 * in_key,word32 key_len)
	{
		word32 i, lk, a, b, c, d, e, f, g, h;
		word32 t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14,t15, t16;

		key_len *= 8;
		if (key_len > 256)

			return -1;

		i = 0;
		lk = (key_len + 31) / 32;

		while (i < lk) {
	#ifdef WORDS_BIGENDIAN
			spkey->l_key[i] = byteswap32(in_key[i]);
	#else
			spkey->l_key[i] = (in_key[i]);
	#endif
			i++;
		}

		if (key_len < 256) {
			while (i < 8)

				spkey->l_key[i++] = 0;

			i = key_len / 32;
			lk = 1 << key_len % 32;
			spkey->l_key[i] = (spkey->l_key[i] & (lk - 1)) | lk;
		}

		for (i = 0; i < 132; ++i) {
			lk = spkey->l_key[i] ^ spkey->l_key[i +
									3] ^ spkey->l_key[i +
												5] ^
					spkey->l_key[i + 7] ^ 0x9e3779b9 ^ i;

			spkey->l_key[i + 8] = (lk << 11) | (lk >> 21);
		}

		k_set(0, a, b, c, d);
		sb3(a, b, c, d, e, f, g, h);
		k_get(0, e, f, g, h);
		k_set(1, a, b, c, d);
		sb2(a, b, c, d, e, f, g, h);
		k_get(1, e, f, g, h);
		k_set(2, a, b, c, d);
		sb1(a, b, c, d, e, f, g, h);
		k_get(2, e, f, g, h);
		k_set(3, a, b, c, d);
		sb0(a, b, c, d, e, f, g, h);
		k_get(3, e, f, g, h);
		k_set(4, a, b, c, d);
		sb7(a, b, c, d, e, f, g, h);
		k_get(4, e, f, g, h);
		k_set(5, a, b, c, d);
		sb6(a, b, c, d, e, f, g, h);
		k_get(5, e, f, g, h);
		k_set(6, a, b, c, d);
		sb5(a, b, c, d, e, f, g, h);
		k_get(6, e, f, g, h);
		k_set(7, a, b, c, d);
		sb4(a, b, c, d, e, f, g, h);
		k_get(7, e, f, g, h);
		k_set(8, a, b, c, d);
		sb3(a, b, c, d, e, f, g, h);
		k_get(8, e, f, g, h);
		k_set(9, a, b, c, d);
		sb2(a, b, c, d, e, f, g, h);
		k_get(9, e, f, g, h);
		k_set(10, a, b, c, d);
		sb1(a, b, c, d, e, f, g, h);
		k_get(10, e, f, g, h);
		k_set(11, a, b, c, d);
		sb0(a, b, c, d, e, f, g, h);
		k_get(11, e, f, g, h);
		k_set(12, a, b, c, d);
		sb7(a, b, c, d, e, f, g, h);
		k_get(12, e, f, g, h);
		k_set(13, a, b, c, d);
		sb6(a, b, c, d, e, f, g, h);
		k_get(13, e, f, g, h);
		k_set(14, a, b, c, d);
		sb5(a, b, c, d, e, f, g, h);
		k_get(14, e, f, g, h);
		k_set(15, a, b, c, d);
		sb4(a, b, c, d, e, f, g, h);
		k_get(15, e, f, g, h);
		k_set(16, a, b, c, d);
		sb3(a, b, c, d, e, f, g, h);
		k_get(16, e, f, g, h);
		k_set(17, a, b, c, d);
		sb2(a, b, c, d, e, f, g, h);
		k_get(17, e, f, g, h);
		k_set(18, a, b, c, d);
		sb1(a, b, c, d, e, f, g, h);
		k_get(18, e, f, g, h);
		k_set(19, a, b, c, d);
		sb0(a, b, c, d, e, f, g, h);
		k_get(19, e, f, g, h);
		k_set(20, a, b, c, d);
		sb7(a, b, c, d, e, f, g, h);
		k_get(20, e, f, g, h);
		k_set(21, a, b, c, d);
		sb6(a, b, c, d, e, f, g, h);
		k_get(21, e, f, g, h);
		k_set(22, a, b, c, d);
		sb5(a, b, c, d, e, f, g, h);
		k_get(22, e, f, g, h);
		k_set(23, a, b, c, d);
		sb4(a, b, c, d, e, f, g, h);
		k_get(23, e, f, g, h);
		k_set(24, a, b, c, d);
		sb3(a, b, c, d, e, f, g, h);
		k_get(24, e, f, g, h);
		k_set(25, a, b, c, d);
		sb2(a, b, c, d, e, f, g, h);
		k_get(25, e, f, g, h);
		k_set(26, a, b, c, d);
		sb1(a, b, c, d, e, f, g, h);
		k_get(26, e, f, g, h);
		k_set(27, a, b, c, d);
		sb0(a, b, c, d, e, f, g, h);
		k_get(27, e, f, g, h);
		k_set(28, a, b, c, d);
		sb7(a, b, c, d, e, f, g, h);
		k_get(28, e, f, g, h);
		k_set(29, a, b, c, d);
		sb6(a, b, c, d, e, f, g, h);
		k_get(29, e, f, g, h);
		k_set(30, a, b, c, d);
		sb5(a, b, c, d, e, f, g, h);
		k_get(30, e, f, g, h);
		k_set(31, a, b, c, d);
		sb4(a, b, c, d, e, f, g, h);
		k_get(31, e, f, g, h);
		k_set(32, a, b, c, d);
		sb3(a, b, c, d, e, f, g, h);
		k_get(32, e, f, g, h);

		return 0;
	}



	/* encrypt a block of text  */
	void serpent_encrypt(SERPENT_KEY * spkey, word32 * in_blk)
	{
		word32 a, b, c, d, e, f, g, h;
		word32 t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14,
				t15, t16;

	#ifdef WORDS_BIGENDIAN
		a = byteswap32(in_blk[0]);
		b = byteswap32(in_blk[1]);
		c = byteswap32(in_blk[2]);
		d = byteswap32(in_blk[3]);
	#else
		a = in_blk[0];
		b = in_blk[1];
		c = in_blk[2];
		d = in_blk[3];
	#endif

		k_xor(0, a, b, c, d);
		sb0(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(1, e, f, g, h);
		sb1(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(2, a, b, c, d);
		sb2(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(3, e, f, g, h);
		sb3(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(4, a, b, c, d);
		sb4(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(5, e, f, g, h);
		sb5(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(6, a, b, c, d);
		sb6(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(7, e, f, g, h);
		sb7(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(8, a, b, c, d);
		sb0(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(9, e, f, g, h);
		sb1(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(10, a, b, c, d);
		sb2(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(11, e, f, g, h);
		sb3(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(12, a, b, c, d);
		sb4(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(13, e, f, g, h);
		sb5(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(14, a, b, c, d);
		sb6(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(15, e, f, g, h);
		sb7(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(16, a, b, c, d);
		sb0(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(17, e, f, g, h);
		sb1(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(18, a, b, c, d);
		sb2(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(19, e, f, g, h);
		sb3(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(20, a, b, c, d);
		sb4(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(21, e, f, g, h);
		sb5(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(22, a, b, c, d);
		sb6(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(23, e, f, g, h);
		sb7(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(24, a, b, c, d);
		sb0(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(25, e, f, g, h);
		sb1(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(26, a, b, c, d);
		sb2(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(27, e, f, g, h);
		sb3(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(28, a, b, c, d);
		sb4(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(29, e, f, g, h);
		sb5(e, f, g, h, a, b, c, d);
		rot(a, b, c, d);
		k_xor(30, a, b, c, d);
		sb6(a, b, c, d, e, f, g, h);
		rot(e, f, g, h);
		k_xor(31, e, f, g, h);
		sb7(e, f, g, h, a, b, c, d);
		k_xor(32, a, b, c, d);

	#ifdef WORDS_BIGENDIAN
		in_blk[0] = byteswap32(a);
		in_blk[1] = byteswap32(b);
		in_blk[2] = byteswap32(c);
		in_blk[3] = byteswap32(d);
	#else
		in_blk[0] = a;
		in_blk[1] = b;
		in_blk[2] = c;
		in_blk[3] = d;
	#endif
	}




	/* decrypt a block of text  */
	void serpent_decrypt(SERPENT_KEY * spkey, word32 * in_blk)
	{
		word32 a, b, c, d, e, f, g, h;
		word32 t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14,
				t15, t16;

	#ifdef WORDS_BIGENDIAN
		a = byteswap32(in_blk[0]);
		b = byteswap32(in_blk[1]);
		c = byteswap32(in_blk[2]);
		d = byteswap32(in_blk[3]);
	#else
		a = in_blk[0];
		b = in_blk[1];
		c = in_blk[2];
		d = in_blk[3];
	#endif

		k_xor(32, a, b, c, d);
		ib7(a, b, c, d, e, f, g, h);
		k_xor(31, e, f, g, h);
		irot(e, f, g, h);
		ib6(e, f, g, h, a, b, c, d);
		k_xor(30, a, b, c, d);
		irot(a, b, c, d);
		ib5(a, b, c, d, e, f, g, h);
		k_xor(29, e, f, g, h);
		irot(e, f, g, h);
		ib4(e, f, g, h, a, b, c, d);
		k_xor(28, a, b, c, d);
		irot(a, b, c, d);
		ib3(a, b, c, d, e, f, g, h);
		k_xor(27, e, f, g, h);
		irot(e, f, g, h);
		ib2(e, f, g, h, a, b, c, d);
		k_xor(26, a, b, c, d);
		irot(a, b, c, d);
		ib1(a, b, c, d, e, f, g, h);
		k_xor(25, e, f, g, h);
		irot(e, f, g, h);
		ib0(e, f, g, h, a, b, c, d);
		k_xor(24, a, b, c, d);
		irot(a, b, c, d);
		ib7(a, b, c, d, e, f, g, h);
		k_xor(23, e, f, g, h);
		irot(e, f, g, h);
		ib6(e, f, g, h, a, b, c, d);
		k_xor(22, a, b, c, d);
		irot(a, b, c, d);
		ib5(a, b, c, d, e, f, g, h);
		k_xor(21, e, f, g, h);
		irot(e, f, g, h);
		ib4(e, f, g, h, a, b, c, d);
		k_xor(20, a, b, c, d);
		irot(a, b, c, d);
		ib3(a, b, c, d, e, f, g, h);
		k_xor(19, e, f, g, h);
		irot(e, f, g, h);
		ib2(e, f, g, h, a, b, c, d);
		k_xor(18, a, b, c, d);
		irot(a, b, c, d);
		ib1(a, b, c, d, e, f, g, h);
		k_xor(17, e, f, g, h);
		irot(e, f, g, h);
		ib0(e, f, g, h, a, b, c, d);
		k_xor(16, a, b, c, d);
		irot(a, b, c, d);
		ib7(a, b, c, d, e, f, g, h);
		k_xor(15, e, f, g, h);
		irot(e, f, g, h);
		ib6(e, f, g, h, a, b, c, d);
		k_xor(14, a, b, c, d);
		irot(a, b, c, d);
		ib5(a, b, c, d, e, f, g, h);
		k_xor(13, e, f, g, h);
		irot(e, f, g, h);
		ib4(e, f, g, h, a, b, c, d);
		k_xor(12, a, b, c, d);
		irot(a, b, c, d);
		ib3(a, b, c, d, e, f, g, h);
		k_xor(11, e, f, g, h);
		irot(e, f, g, h);
		ib2(e, f, g, h, a, b, c, d);
		k_xor(10, a, b, c, d);
		irot(a, b, c, d);
		ib1(a, b, c, d, e, f, g, h);
		k_xor(9, e, f, g, h);
		irot(e, f, g, h);
		ib0(e, f, g, h, a, b, c, d);
		k_xor(8, a, b, c, d);
		irot(a, b, c, d);
		ib7(a, b, c, d, e, f, g, h);
		k_xor(7, e, f, g, h);
		irot(e, f, g, h);
		ib6(e, f, g, h, a, b, c, d);
		k_xor(6, a, b, c, d);
		irot(a, b, c, d);
		ib5(a, b, c, d, e, f, g, h);
		k_xor(5, e, f, g, h);
		irot(e, f, g, h);
		ib4(e, f, g, h, a, b, c, d);
		k_xor(4, a, b, c, d);
		irot(a, b, c, d);
		ib3(a, b, c, d, e, f, g, h);
		k_xor(3, e, f, g, h);
		irot(e, f, g, h);
		ib2(e, f, g, h, a, b, c, d);
		k_xor(2, a, b, c, d);
		irot(a, b, c, d);
		ib1(a, b, c, d, e, f, g, h);
		k_xor(1, e, f, g, h);
		irot(e, f, g, h);
		ib0(e, f, g, h, a, b, c, d);
		k_xor(0, a, b, c, d);

	#ifdef WORDS_BIGENDIAN
		in_blk[0] = byteswap32(a);
		in_blk[1] = byteswap32(b);
		in_blk[2] = byteswap32(c);
		in_blk[3] = byteswap32(d);
	#else
		in_blk[0] = a;
		in_blk[1] = b;
		in_blk[2] = c;
		in_blk[3] = d;
	#endif
	}





	int serpent_get_size()
	{
		return sizeof(SERPENT_KEY);
	}





	#define CIPHER_SERPENT "9a99455df5080bfccadf049b5aaf7d61"

	bool serpent_self_test()
	{
		char keyword[33];
		unsigned char plaintext[16];
		unsigned char ciphertext[16];
		int blocksize = 16;
		int  j;
		SERPENT_KEY key;
		unsigned char cipher_tmp[210];


		for (j = 0; j < 32; j++) {
			keyword[j] = ((j * 2 + 10) % 256);
		}


		for (j = 0; j < blocksize; j++) {
			plaintext[j] = j % 256;
		}

		memcpy(ciphertext, plaintext, blocksize);

		serpent_set_key(&key, (const unsigned __int32 *) &(keyword[0]), 32);

		serpent_encrypt(&key, (unsigned __int32 *) &(ciphertext[0]));

		for (j = 0; j < blocksize; j++) {
			sprintf_s(&((char *) cipher_tmp)[2 * j],3, "%.2x",
				ciphertext[j]);
		}

		if (strcmp((char *) cipher_tmp, CIPHER_SERPENT) != 0) {
			printf("failed compatibility\n");
			printf("Expected: %s\nGot: %s\n", CIPHER_SERPENT,
						 (char *) cipher_tmp);
			return false;
		}
		serpent_decrypt(&key, (unsigned __int32 *) &(ciphertext[0]));


		if (memcmp(ciphertext, plaintext, 16) != 0) {
			printf("failed internally\n");
			return false;
		}
		return true;
	}





















//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
// ripemd.h

// RIPEMD-160 is a 160-bit cryptographic hash function, designed by Hans
// Dobbertin, Antoon Bosselaers, and Bart Preneel. It is intended to be
// used as a secure replacement for the 128-bit hash functions MD4, MD5,
// and RIPEMD
// See also: http://www.esat.kuleuven.ac.be/~bosselae/ripemd160.html

// RIPEMD-128 is a plug-in substitute for RIPEMD (or MD4 and MD5, for
// that matter) with a 128-bit result. 128-bit hash results do not
// offer sufficient protection for the next ten years, and
// applications using 128-bit hash functions should consider upgrading
// to a 160-bit hash function.

// RIPEMD-256 and RIPEMD-320 are optional extensions of, respectively,
// RIPEMD-128 and RIPEMD-160, and are intended for applications of
// hash functions that require a longer hash result without needing a
// larger security level.
 

// The following code was written by Nikos Mavroyanopoulos and B. Poettering for the mhash library.
//
	// 32-bit rotate left - kludged with shifts
	#define ROTL(n,X)  (((X)<<(n))|((X)>>(32-(n))))

	#define f0(x,y,z)  (x^y^z)
	//#define f16(x,y,z) ((x&y)|(~(x) & z))
	#define f16(x,y,z) ((z)^((x)&((y)^(z))))
	#define f32(x,y,z) ((x|~(y))^z)
	//#define f48(x,y,z) ((x&z)|(y&~(z)))
	#define f48(x,y,z) ((y)^((z)&((x)^(y))))
	#define f64(x,y,z) (x^(y|(~z)))

	#define K0  0x00000000
	#define K1  0x5A827999     // 2^30 * sqrt(2) 
	#define K2  0x6ED9EBA1     // 2^30 * sqrt(3) 
	#define K3  0x8F1BBCDC     // 2^30 * sqrt(5) 
	#define K4  0xA953FD4E     // 2^30 * sqrt(7) 

	#define KK0 0x50A28BE6     // 2^30 * cbrt(2) 
	#define KK1 0x5C4DD124     // 2^30 * cbrt(3) 
	#define KK2 0x6D703EF3     // 2^30 * cbrt(5) 
	#define KK3 0x7A6D76E9     // 2^30 * cbrt(7) 
	#define KK4 0x00000000

	#define h0init 0x67452301
	#define h1init 0xEFCDAB89
	#define h2init 0x98BADCFE
	#define h3init 0x10325476
	#define h4init 0xC3D2E1F0
	#define h5init 0x76543210
	#define h6init 0xFEDCBA98
	#define h7init 0x89ABCDEF
	#define h8init 0x01234567
	#define h9init 0x3C2D1E0F


	// Initialize the RIPEMD values 

	void ripemd_init(struct ripemd_ctx *ctx)
	{
		// Set the h-vars to their initial values 
		ctx->digest[0] = h0init;
		ctx->digest[1] = h1init;
		ctx->digest[2] = h2init;
		ctx->digest[3] = h3init;

		// Initialize bit count 
		ctx->bitcount = 0;

		// Initialize buffer 
		ctx->index = 0;
	}

	void ripemd128_init(struct ripemd_ctx *ctx)
	{
		ripemd_init(ctx);
		ctx->digest_len = 8 * RIPEMD128_DIGESTSIZE;
	}

	void ripemd160_init(struct ripemd_ctx *ctx)
	{
		ripemd_init(ctx);
		ctx->digest[4] = h4init;
		ctx->digest_len = 8 * RIPEMD160_DIGESTSIZE;
	}

	void ripemd256_init(struct ripemd_ctx *ctx)
	{
		ripemd_init(ctx);
		ctx->digest[4] = h5init;
		ctx->digest[5] = h6init;
		ctx->digest[6] = h7init;
		ctx->digest[7] = h8init;
		ctx->digest_len = 8 * RIPEMD256_DIGESTSIZE;
	}

	void ripemd320_init(struct ripemd_ctx *ctx)
	{
		ripemd_init(ctx);
		ctx->digest[4] = h4init;
		ctx->digest[5] = h5init;
		ctx->digest[6] = h6init;
		ctx->digest[7] = h7init;
		ctx->digest[8] = h8init;
		ctx->digest[9] = h9init;
		ctx->digest_len = 8 * RIPEMD320_DIGESTSIZE;
	}


	// The RIPEMD core functions 

	#define subRound128(a, b, c, d, f, k, r, data) \
			( a = ROTL( r, a + f(b,c,d) + data + k))

	#define subRound160(a, b, c, d, e, f, k, r, data) \
			( a = ROTL( r, a + f(b,c,d) + data + k) + e, c = ROTL(10, c) )

	static void ripemd128_transform(struct ripemd_ctx *ctx, word32 * data)
	{
		word32 A, B, C, D;	// Local vars 
		word32 AA, BB, CC, DD;	// Local vars 
		word32 T;

		// Set up first buffer and local data buffer 
		A = AA = ctx->digest[0];
		B = BB = ctx->digest[1];
		C = CC = ctx->digest[2];
		D = DD = ctx->digest[3];

	// j=0...15 
		subRound128(A, B, C, D, f0, K0, 11, data[0]);
		subRound128(D, A, B, C, f0, K0, 14, data[1]);
		subRound128(C, D, A, B, f0, K0, 15, data[2]);
		subRound128(B, C, D, A, f0, K0, 12, data[3]);
		subRound128(A, B, C, D, f0, K0, 5, data[4]);
		subRound128(D, A, B, C, f0, K0, 8, data[5]);
		subRound128(C, D, A, B, f0, K0, 7, data[6]);
		subRound128(B, C, D, A, f0, K0, 9, data[7]);
		subRound128(A, B, C, D, f0, K0, 11, data[8]);
		subRound128(D, A, B, C, f0, K0, 13, data[9]);
		subRound128(C, D, A, B, f0, K0, 14, data[10]);
		subRound128(B, C, D, A, f0, K0, 15, data[11]);
		subRound128(A, B, C, D, f0, K0, 6, data[12]);
		subRound128(D, A, B, C, f0, K0, 7, data[13]);
		subRound128(C, D, A, B, f0, K0, 9, data[14]);
		subRound128(B, C, D, A, f0, K0, 8, data[15]);

		subRound128(AA, BB, CC, DD, f48, KK0, 8, data[5]);
		subRound128(DD, AA, BB, CC, f48, KK0, 9, data[14]);
		subRound128(CC, DD, AA, BB, f48, KK0, 9, data[7]);
		subRound128(BB, CC, DD, AA, f48, KK0, 11, data[0]);
		subRound128(AA, BB, CC, DD, f48, KK0, 13, data[9]);
		subRound128(DD, AA, BB, CC, f48, KK0, 15, data[2]);
		subRound128(CC, DD, AA, BB, f48, KK0, 15, data[11]);
		subRound128(BB, CC, DD, AA, f48, KK0, 5, data[4]);
		subRound128(AA, BB, CC, DD, f48, KK0, 7, data[13]);
		subRound128(DD, AA, BB, CC, f48, KK0, 7, data[6]);
		subRound128(CC, DD, AA, BB, f48, KK0, 8, data[15]);
		subRound128(BB, CC, DD, AA, f48, KK0, 11, data[8]);
		subRound128(AA, BB, CC, DD, f48, KK0, 14, data[1]);
		subRound128(DD, AA, BB, CC, f48, KK0, 14, data[10]);
		subRound128(CC, DD, AA, BB, f48, KK0, 12, data[3]);
		subRound128(BB, CC, DD, AA, f48, KK0, 6, data[12]);

	// j=16...31 
		subRound128(A, B, C, D, f16, K1, 7, data[7]);
		subRound128(D, A, B, C, f16, K1, 6, data[4]);
		subRound128(C, D, A, B, f16, K1, 8, data[13]);
		subRound128(B, C, D, A, f16, K1, 13, data[1]);
		subRound128(A, B, C, D, f16, K1, 11, data[10]);
		subRound128(D, A, B, C, f16, K1, 9, data[6]);
		subRound128(C, D, A, B, f16, K1, 7, data[15]);
		subRound128(B, C, D, A, f16, K1, 15, data[3]);
		subRound128(A, B, C, D, f16, K1, 7, data[12]);
		subRound128(D, A, B, C, f16, K1, 12, data[0]);
		subRound128(C, D, A, B, f16, K1, 15, data[9]);
		subRound128(B, C, D, A, f16, K1, 9, data[5]);
		subRound128(A, B, C, D, f16, K1, 11, data[2]);
		subRound128(D, A, B, C, f16, K1, 7, data[14]);
		subRound128(C, D, A, B, f16, K1, 13, data[11]);
		subRound128(B, C, D, A, f16, K1, 12, data[8]);

		subRound128(AA, BB, CC, DD, f32, KK1, 9, data[6]);
		subRound128(DD, AA, BB, CC, f32, KK1, 13, data[11]);
		subRound128(CC, DD, AA, BB, f32, KK1, 15, data[3]);
		subRound128(BB, CC, DD, AA, f32, KK1, 7, data[7]);
		subRound128(AA, BB, CC, DD, f32, KK1, 12, data[0]);
		subRound128(DD, AA, BB, CC, f32, KK1, 8, data[13]);
		subRound128(CC, DD, AA, BB, f32, KK1, 9, data[5]);
		subRound128(BB, CC, DD, AA, f32, KK1, 11, data[10]);
		subRound128(AA, BB, CC, DD, f32, KK1, 7, data[14]);
		subRound128(DD, AA, BB, CC, f32, KK1, 7, data[15]);
		subRound128(CC, DD, AA, BB, f32, KK1, 12, data[8]);
		subRound128(BB, CC, DD, AA, f32, KK1, 7, data[12]);
		subRound128(AA, BB, CC, DD, f32, KK1, 6, data[4]);
		subRound128(DD, AA, BB, CC, f32, KK1, 15, data[9]);
		subRound128(CC, DD, AA, BB, f32, KK1, 13, data[1]);
		subRound128(BB, CC, DD, AA, f32, KK1, 11, data[2]);

	// j=32...47 
		subRound128(A, B, C, D, f32, K2, 11, data[3]);
		subRound128(D, A, B, C, f32, K2, 13, data[10]);
		subRound128(C, D, A, B, f32, K2, 6, data[14]);
		subRound128(B, C, D, A, f32, K2, 7, data[4]);
		subRound128(A, B, C, D, f32, K2, 14, data[9]);
		subRound128(D, A, B, C, f32, K2, 9, data[15]);
		subRound128(C, D, A, B, f32, K2, 13, data[8]);
		subRound128(B, C, D, A, f32, K2, 15, data[1]);
		subRound128(A, B, C, D, f32, K2, 14, data[2]);
		subRound128(D, A, B, C, f32, K2, 8, data[7]);
		subRound128(C, D, A, B, f32, K2, 13, data[0]);
		subRound128(B, C, D, A, f32, K2, 6, data[6]);
		subRound128(A, B, C, D, f32, K2, 5, data[13]);
		subRound128(D, A, B, C, f32, K2, 12, data[11]);
		subRound128(C, D, A, B, f32, K2, 7, data[5]);
		subRound128(B, C, D, A, f32, K2, 5, data[12]);

		subRound128(AA, BB, CC, DD, f16, KK2, 9, data[15]);
		subRound128(DD, AA, BB, CC, f16, KK2, 7, data[5]);
		subRound128(CC, DD, AA, BB, f16, KK2, 15, data[1]);
		subRound128(BB, CC, DD, AA, f16, KK2, 11, data[3]);
		subRound128(AA, BB, CC, DD, f16, KK2, 8, data[7]);
		subRound128(DD, AA, BB, CC, f16, KK2, 6, data[14]);
		subRound128(CC, DD, AA, BB, f16, KK2, 6, data[6]);
		subRound128(BB, CC, DD, AA, f16, KK2, 14, data[9]);
		subRound128(AA, BB, CC, DD, f16, KK2, 12, data[11]);
		subRound128(DD, AA, BB, CC, f16, KK2, 13, data[8]);
		subRound128(CC, DD, AA, BB, f16, KK2, 5, data[12]);
		subRound128(BB, CC, DD, AA, f16, KK2, 14, data[2]);
		subRound128(AA, BB, CC, DD, f16, KK2, 13, data[10]);
		subRound128(DD, AA, BB, CC, f16, KK2, 13, data[0]);
		subRound128(CC, DD, AA, BB, f16, KK2, 7, data[4]);
		subRound128(BB, CC, DD, AA, f16, KK2, 5, data[13]);

	// j=48...63 
		subRound128(A, B, C, D, f48, K3, 11, data[1]);
		subRound128(D, A, B, C, f48, K3, 12, data[9]);
		subRound128(C, D, A, B, f48, K3, 14, data[11]);
		subRound128(B, C, D, A, f48, K3, 15, data[10]);
		subRound128(A, B, C, D, f48, K3, 14, data[0]);
		subRound128(D, A, B, C, f48, K3, 15, data[8]);
		subRound128(C, D, A, B, f48, K3, 9, data[12]);
		subRound128(B, C, D, A, f48, K3, 8, data[4]);
		subRound128(A, B, C, D, f48, K3, 9, data[13]);
		subRound128(D, A, B, C, f48, K3, 14, data[3]);
		subRound128(C, D, A, B, f48, K3, 5, data[7]);
		subRound128(B, C, D, A, f48, K3, 6, data[15]);
		subRound128(A, B, C, D, f48, K3, 8, data[14]);
		subRound128(D, A, B, C, f48, K3, 6, data[5]);
		subRound128(C, D, A, B, f48, K3, 5, data[6]);
		subRound128(B, C, D, A, f48, K3, 12, data[2]);

		subRound128(AA, BB, CC, DD, f0, KK4, 15, data[8]);
		subRound128(DD, AA, BB, CC, f0, KK4, 5, data[6]);
		subRound128(CC, DD, AA, BB, f0, KK4, 8, data[4]);
		subRound128(BB, CC, DD, AA, f0, KK4, 11, data[1]);
		subRound128(AA, BB, CC, DD, f0, KK4, 14, data[3]);
		subRound128(DD, AA, BB, CC, f0, KK4, 14, data[11]);
		subRound128(CC, DD, AA, BB, f0, KK4, 6, data[15]);
		subRound128(BB, CC, DD, AA, f0, KK4, 14, data[0]);
		subRound128(AA, BB, CC, DD, f0, KK4, 6, data[5]);
		subRound128(DD, AA, BB, CC, f0, KK4, 9, data[12]);
		subRound128(CC, DD, AA, BB, f0, KK4, 12, data[2]);
		subRound128(BB, CC, DD, AA, f0, KK4, 9, data[13]);
		subRound128(AA, BB, CC, DD, f0, KK4, 12, data[9]);
		subRound128(DD, AA, BB, CC, f0, KK4, 5, data[7]);
		subRound128(CC, DD, AA, BB, f0, KK4, 15, data[10]);
		subRound128(BB, CC, DD, AA, f0, KK4, 8, data[14]);

		T = ctx->digest[1] + DD + C;
					ctx->digest[1] = ctx->digest[2] + AA + D;
		ctx->digest[2] = ctx->digest[3] + BB + A;
		ctx->digest[3] = ctx->digest[0] + CC + B;
		ctx->digest[0] = T;
	}

	static void ripemd160_transform(struct ripemd_ctx *ctx, word32 * data)
	{
		word32 A, B, C, D, E;   	// Local vars 
		word32 AA, BB, CC, DD, EE;      // Local vars 
		word32 T;

		// Set up first buffer and local data buffer 
		A = AA = ctx->digest[0];
		B = BB = ctx->digest[1];
		C = CC = ctx->digest[2];
		D = DD = ctx->digest[3];
		E = EE = ctx->digest[4];

	// j=0...15 
		subRound160(A, B, C, D, E, f0, K0, 11, data[0]);
		subRound160(E, A, B, C, D, f0, K0, 14, data[1]);
		subRound160(D, E, A, B, C, f0, K0, 15, data[2]);
		subRound160(C, D, E, A, B, f0, K0, 12, data[3]);
		subRound160(B, C, D, E, A, f0, K0, 5, data[4]);
		subRound160(A, B, C, D, E, f0, K0, 8, data[5]);
		subRound160(E, A, B, C, D, f0, K0, 7, data[6]);
		subRound160(D, E, A, B, C, f0, K0, 9, data[7]);
		subRound160(C, D, E, A, B, f0, K0, 11, data[8]);
		subRound160(B, C, D, E, A, f0, K0, 13, data[9]);
		subRound160(A, B, C, D, E, f0, K0, 14, data[10]);
		subRound160(E, A, B, C, D, f0, K0, 15, data[11]);
		subRound160(D, E, A, B, C, f0, K0, 6, data[12]);
		subRound160(C, D, E, A, B, f0, K0, 7, data[13]);
		subRound160(B, C, D, E, A, f0, K0, 9, data[14]);
		subRound160(A, B, C, D, E, f0, K0, 8, data[15]);

		subRound160(AA, BB, CC, DD, EE, f64, KK0, 8, data[5]);
		subRound160(EE, AA, BB, CC, DD, f64, KK0, 9, data[14]);
		subRound160(DD, EE, AA, BB, CC, f64, KK0, 9, data[7]);
		subRound160(CC, DD, EE, AA, BB, f64, KK0, 11, data[0]);
		subRound160(BB, CC, DD, EE, AA, f64, KK0, 13, data[9]);
		subRound160(AA, BB, CC, DD, EE, f64, KK0, 15, data[2]);
		subRound160(EE, AA, BB, CC, DD, f64, KK0, 15, data[11]);
		subRound160(DD, EE, AA, BB, CC, f64, KK0, 5, data[4]);
		subRound160(CC, DD, EE, AA, BB, f64, KK0, 7, data[13]);
		subRound160(BB, CC, DD, EE, AA, f64, KK0, 7, data[6]);
		subRound160(AA, BB, CC, DD, EE, f64, KK0, 8, data[15]);
		subRound160(EE, AA, BB, CC, DD, f64, KK0, 11, data[8]);
		subRound160(DD, EE, AA, BB, CC, f64, KK0, 14, data[1]);
		subRound160(CC, DD, EE, AA, BB, f64, KK0, 14, data[10]);
		subRound160(BB, CC, DD, EE, AA, f64, KK0, 12, data[3]);
		subRound160(AA, BB, CC, DD, EE, f64, KK0, 6, data[12]);

	// j=16...31 
		subRound160(E, A, B, C, D, f16, K1, 7, data[7]);
		subRound160(D, E, A, B, C, f16, K1, 6, data[4]);
		subRound160(C, D, E, A, B, f16, K1, 8, data[13]);
		subRound160(B, C, D, E, A, f16, K1, 13, data[1]);
		subRound160(A, B, C, D, E, f16, K1, 11, data[10]);
		subRound160(E, A, B, C, D, f16, K1, 9, data[6]);
		subRound160(D, E, A, B, C, f16, K1, 7, data[15]);
		subRound160(C, D, E, A, B, f16, K1, 15, data[3]);
		subRound160(B, C, D, E, A, f16, K1, 7, data[12]);
		subRound160(A, B, C, D, E, f16, K1, 12, data[0]);
		subRound160(E, A, B, C, D, f16, K1, 15, data[9]);
		subRound160(D, E, A, B, C, f16, K1, 9, data[5]);
		subRound160(C, D, E, A, B, f16, K1, 11, data[2]);
		subRound160(B, C, D, E, A, f16, K1, 7, data[14]);
		subRound160(A, B, C, D, E, f16, K1, 13, data[11]);
		subRound160(E, A, B, C, D, f16, K1, 12, data[8]);

		subRound160(EE, AA, BB, CC, DD, f48, KK1, 9, data[6]);
		subRound160(DD, EE, AA, BB, CC, f48, KK1, 13, data[11]);
		subRound160(CC, DD, EE, AA, BB, f48, KK1, 15, data[3]);
		subRound160(BB, CC, DD, EE, AA, f48, KK1, 7, data[7]);
		subRound160(AA, BB, CC, DD, EE, f48, KK1, 12, data[0]);
		subRound160(EE, AA, BB, CC, DD, f48, KK1, 8, data[13]);
		subRound160(DD, EE, AA, BB, CC, f48, KK1, 9, data[5]);
		subRound160(CC, DD, EE, AA, BB, f48, KK1, 11, data[10]);
		subRound160(BB, CC, DD, EE, AA, f48, KK1, 7, data[14]);
		subRound160(AA, BB, CC, DD, EE, f48, KK1, 7, data[15]);
		subRound160(EE, AA, BB, CC, DD, f48, KK1, 12, data[8]);
		subRound160(DD, EE, AA, BB, CC, f48, KK1, 7, data[12]);
		subRound160(CC, DD, EE, AA, BB, f48, KK1, 6, data[4]);
		subRound160(BB, CC, DD, EE, AA, f48, KK1, 15, data[9]);
		subRound160(AA, BB, CC, DD, EE, f48, KK1, 13, data[1]);
		subRound160(EE, AA, BB, CC, DD, f48, KK1, 11, data[2]);

	// j=32...47 
		subRound160(D, E, A, B, C, f32, K2, 11, data[3]);
		subRound160(C, D, E, A, B, f32, K2, 13, data[10]);
		subRound160(B, C, D, E, A, f32, K2, 6, data[14]);
		subRound160(A, B, C, D, E, f32, K2, 7, data[4]);
		subRound160(E, A, B, C, D, f32, K2, 14, data[9]);
		subRound160(D, E, A, B, C, f32, K2, 9, data[15]);
		subRound160(C, D, E, A, B, f32, K2, 13, data[8]);
		subRound160(B, C, D, E, A, f32, K2, 15, data[1]);
		subRound160(A, B, C, D, E, f32, K2, 14, data[2]);
		subRound160(E, A, B, C, D, f32, K2, 8, data[7]);
		subRound160(D, E, A, B, C, f32, K2, 13, data[0]);
		subRound160(C, D, E, A, B, f32, K2, 6, data[6]);
		subRound160(B, C, D, E, A, f32, K2, 5, data[13]);
		subRound160(A, B, C, D, E, f32, K2, 12, data[11]);
		subRound160(E, A, B, C, D, f32, K2, 7, data[5]);
		subRound160(D, E, A, B, C, f32, K2, 5, data[12]);

		subRound160(DD, EE, AA, BB, CC, f32, KK2, 9, data[15]);
		subRound160(CC, DD, EE, AA, BB, f32, KK2, 7, data[5]);
		subRound160(BB, CC, DD, EE, AA, f32, KK2, 15, data[1]);
		subRound160(AA, BB, CC, DD, EE, f32, KK2, 11, data[3]);
		subRound160(EE, AA, BB, CC, DD, f32, KK2, 8, data[7]);
		subRound160(DD, EE, AA, BB, CC, f32, KK2, 6, data[14]);
		subRound160(CC, DD, EE, AA, BB, f32, KK2, 6, data[6]);
		subRound160(BB, CC, DD, EE, AA, f32, KK2, 14, data[9]);
		subRound160(AA, BB, CC, DD, EE, f32, KK2, 12, data[11]);
		subRound160(EE, AA, BB, CC, DD, f32, KK2, 13, data[8]);
		subRound160(DD, EE, AA, BB, CC, f32, KK2, 5, data[12]);
		subRound160(CC, DD, EE, AA, BB, f32, KK2, 14, data[2]);
		subRound160(BB, CC, DD, EE, AA, f32, KK2, 13, data[10]);
		subRound160(AA, BB, CC, DD, EE, f32, KK2, 13, data[0]);
		subRound160(EE, AA, BB, CC, DD, f32, KK2, 7, data[4]);
		subRound160(DD, EE, AA, BB, CC, f32, KK2, 5, data[13]);

	// j=48...63 
		subRound160(C, D, E, A, B, f48, K3, 11, data[1]);
		subRound160(B, C, D, E, A, f48, K3, 12, data[9]);
		subRound160(A, B, C, D, E, f48, K3, 14, data[11]);
		subRound160(E, A, B, C, D, f48, K3, 15, data[10]);
		subRound160(D, E, A, B, C, f48, K3, 14, data[0]);
		subRound160(C, D, E, A, B, f48, K3, 15, data[8]);
		subRound160(B, C, D, E, A, f48, K3, 9, data[12]);
		subRound160(A, B, C, D, E, f48, K3, 8, data[4]);
		subRound160(E, A, B, C, D, f48, K3, 9, data[13]);
		subRound160(D, E, A, B, C, f48, K3, 14, data[3]);
		subRound160(C, D, E, A, B, f48, K3, 5, data[7]);
		subRound160(B, C, D, E, A, f48, K3, 6, data[15]);
		subRound160(A, B, C, D, E, f48, K3, 8, data[14]);
		subRound160(E, A, B, C, D, f48, K3, 6, data[5]);
		subRound160(D, E, A, B, C, f48, K3, 5, data[6]);
		subRound160(C, D, E, A, B, f48, K3, 12, data[2]);

		subRound160(CC, DD, EE, AA, BB, f16, KK3, 15, data[8]);
		subRound160(BB, CC, DD, EE, AA, f16, KK3, 5, data[6]);
		subRound160(AA, BB, CC, DD, EE, f16, KK3, 8, data[4]);
		subRound160(EE, AA, BB, CC, DD, f16, KK3, 11, data[1]);
		subRound160(DD, EE, AA, BB, CC, f16, KK3, 14, data[3]);
		subRound160(CC, DD, EE, AA, BB, f16, KK3, 14, data[11]);
		subRound160(BB, CC, DD, EE, AA, f16, KK3, 6, data[15]);
		subRound160(AA, BB, CC, DD, EE, f16, KK3, 14, data[0]);
		subRound160(EE, AA, BB, CC, DD, f16, KK3, 6, data[5]);
		subRound160(DD, EE, AA, BB, CC, f16, KK3, 9, data[12]);
		subRound160(CC, DD, EE, AA, BB, f16, KK3, 12, data[2]);
		subRound160(BB, CC, DD, EE, AA, f16, KK3, 9, data[13]);
		subRound160(AA, BB, CC, DD, EE, f16, KK3, 12, data[9]);
		subRound160(EE, AA, BB, CC, DD, f16, KK3, 5, data[7]);
		subRound160(DD, EE, AA, BB, CC, f16, KK3, 15, data[10]);
		subRound160(CC, DD, EE, AA, BB, f16, KK3, 8, data[14]);

	// j=64...79 
		subRound160(B, C, D, E, A, f64, K4, 9, data[4]);
		subRound160(A, B, C, D, E, f64, K4, 15, data[0]);
		subRound160(E, A, B, C, D, f64, K4, 5, data[5]);
		subRound160(D, E, A, B, C, f64, K4, 11, data[9]);
		subRound160(C, D, E, A, B, f64, K4, 6, data[7]);
		subRound160(B, C, D, E, A, f64, K4, 8, data[12]);
		subRound160(A, B, C, D, E, f64, K4, 13, data[2]);
		subRound160(E, A, B, C, D, f64, K4, 12, data[10]);
		subRound160(D, E, A, B, C, f64, K4, 5, data[14]);
		subRound160(C, D, E, A, B, f64, K4, 12, data[1]);
		subRound160(B, C, D, E, A, f64, K4, 13, data[3]);
		subRound160(A, B, C, D, E, f64, K4, 14, data[8]);
		subRound160(E, A, B, C, D, f64, K4, 11, data[11]);
		subRound160(D, E, A, B, C, f64, K4, 8, data[6]);
		subRound160(C, D, E, A, B, f64, K4, 5, data[15]);
		subRound160(B, C, D, E, A, f64, K4, 6, data[13]);

		subRound160(BB, CC, DD, EE, AA, f0, KK4, 8, data[12]);
		subRound160(AA, BB, CC, DD, EE, f0, KK4, 5, data[15]);
		subRound160(EE, AA, BB, CC, DD, f0, KK4, 12, data[10]);
		subRound160(DD, EE, AA, BB, CC, f0, KK4, 9, data[4]);
		subRound160(CC, DD, EE, AA, BB, f0, KK4, 12, data[1]);
		subRound160(BB, CC, DD, EE, AA, f0, KK4, 5, data[5]);
		subRound160(AA, BB, CC, DD, EE, f0, KK4, 14, data[8]);
		subRound160(EE, AA, BB, CC, DD, f0, KK4, 6, data[7]);
		subRound160(DD, EE, AA, BB, CC, f0, KK4, 8, data[6]);
		subRound160(CC, DD, EE, AA, BB, f0, KK4, 13, data[2]);
		subRound160(BB, CC, DD, EE, AA, f0, KK4, 6, data[13]);
		subRound160(AA, BB, CC, DD, EE, f0, KK4, 5, data[14]);
		subRound160(EE, AA, BB, CC, DD, f0, KK4, 15, data[0]);
		subRound160(DD, EE, AA, BB, CC, f0, KK4, 13, data[3]);
		subRound160(CC, DD, EE, AA, BB, f0, KK4, 11, data[9]);
		subRound160(BB, CC, DD, EE, AA, f0, KK4, 11, data[11]);


		T = ctx->digest[1] + DD + C;
		ctx->digest[1] = ctx->digest[2] + EE + D;
		ctx->digest[2] = ctx->digest[3] + AA + E;
		ctx->digest[3] = ctx->digest[4] + BB + A;
		ctx->digest[4] = ctx->digest[0] + CC + B;
		ctx->digest[0] = T;
	}


	static void ripemd256_transform(struct ripemd_ctx *ctx, word32 * data)
	{
		word32 A, B, C, D;	// Local vars 
		word32 AA, BB, CC, DD;	// Local vars
		word32 T;

		// Set up first buffer and local data buffer 
		A = ctx->digest[0];
		B = ctx->digest[1];
		C = ctx->digest[2];
		D = ctx->digest[3];
		AA = ctx->digest[4];
		BB = ctx->digest[5];
		CC = ctx->digest[6];
		DD = ctx->digest[7];

	// j=0...15 
		subRound128(A, B, C, D, f0, K0, 11, data[0]);
		subRound128(D, A, B, C, f0, K0, 14, data[1]);
		subRound128(C, D, A, B, f0, K0, 15, data[2]);
		subRound128(B, C, D, A, f0, K0, 12, data[3]);
		subRound128(A, B, C, D, f0, K0, 5, data[4]);
		subRound128(D, A, B, C, f0, K0, 8, data[5]);
		subRound128(C, D, A, B, f0, K0, 7, data[6]);
		subRound128(B, C, D, A, f0, K0, 9, data[7]);
		subRound128(A, B, C, D, f0, K0, 11, data[8]);
		subRound128(D, A, B, C, f0, K0, 13, data[9]);
		subRound128(C, D, A, B, f0, K0, 14, data[10]);
		subRound128(B, C, D, A, f0, K0, 15, data[11]);
		subRound128(A, B, C, D, f0, K0, 6, data[12]);
		subRound128(D, A, B, C, f0, K0, 7, data[13]);
		subRound128(C, D, A, B, f0, K0, 9, data[14]);
		subRound128(B, C, D, A, f0, K0, 8, data[15]);

		subRound128(AA, BB, CC, DD, f48, KK0, 8, data[5]);
		subRound128(DD, AA, BB, CC, f48, KK0, 9, data[14]);
		subRound128(CC, DD, AA, BB, f48, KK0, 9, data[7]);
		subRound128(BB, CC, DD, AA, f48, KK0, 11, data[0]);
		subRound128(AA, BB, CC, DD, f48, KK0, 13, data[9]);
		subRound128(DD, AA, BB, CC, f48, KK0, 15, data[2]);
		subRound128(CC, DD, AA, BB, f48, KK0, 15, data[11]);
		subRound128(BB, CC, DD, AA, f48, KK0, 5, data[4]);
		subRound128(AA, BB, CC, DD, f48, KK0, 7, data[13]);
		subRound128(DD, AA, BB, CC, f48, KK0, 7, data[6]);
		subRound128(CC, DD, AA, BB, f48, KK0, 8, data[15]);
		subRound128(BB, CC, DD, AA, f48, KK0, 11, data[8]);
		subRound128(AA, BB, CC, DD, f48, KK0, 14, data[1]);
		subRound128(DD, AA, BB, CC, f48, KK0, 14, data[10]);
		subRound128(CC, DD, AA, BB, f48, KK0, 12, data[3]);
		subRound128(BB, CC, DD, AA, f48, KK0, 6, data[12]);

		T = A; A = AA; AA = T;

	// j=16...31 
		subRound128(A, B, C, D, f16, K1, 7, data[7]);
		subRound128(D, A, B, C, f16, K1, 6, data[4]);
		subRound128(C, D, A, B, f16, K1, 8, data[13]);
		subRound128(B, C, D, A, f16, K1, 13, data[1]);
		subRound128(A, B, C, D, f16, K1, 11, data[10]);
		subRound128(D, A, B, C, f16, K1, 9, data[6]);
		subRound128(C, D, A, B, f16, K1, 7, data[15]);
		subRound128(B, C, D, A, f16, K1, 15, data[3]);
		subRound128(A, B, C, D, f16, K1, 7, data[12]);
		subRound128(D, A, B, C, f16, K1, 12, data[0]);
		subRound128(C, D, A, B, f16, K1, 15, data[9]);
		subRound128(B, C, D, A, f16, K1, 9, data[5]);
		subRound128(A, B, C, D, f16, K1, 11, data[2]);
		subRound128(D, A, B, C, f16, K1, 7, data[14]);
		subRound128(C, D, A, B, f16, K1, 13, data[11]);
		subRound128(B, C, D, A, f16, K1, 12, data[8]);

		subRound128(AA, BB, CC, DD, f32, KK1, 9, data[6]);
		subRound128(DD, AA, BB, CC, f32, KK1, 13, data[11]);
		subRound128(CC, DD, AA, BB, f32, KK1, 15, data[3]);
		subRound128(BB, CC, DD, AA, f32, KK1, 7, data[7]);
		subRound128(AA, BB, CC, DD, f32, KK1, 12, data[0]);
		subRound128(DD, AA, BB, CC, f32, KK1, 8, data[13]);
		subRound128(CC, DD, AA, BB, f32, KK1, 9, data[5]);
		subRound128(BB, CC, DD, AA, f32, KK1, 11, data[10]);
		subRound128(AA, BB, CC, DD, f32, KK1, 7, data[14]);
		subRound128(DD, AA, BB, CC, f32, KK1, 7, data[15]);
		subRound128(CC, DD, AA, BB, f32, KK1, 12, data[8]);
		subRound128(BB, CC, DD, AA, f32, KK1, 7, data[12]);
		subRound128(AA, BB, CC, DD, f32, KK1, 6, data[4]);
		subRound128(DD, AA, BB, CC, f32, KK1, 15, data[9]);
		subRound128(CC, DD, AA, BB, f32, KK1, 13, data[1]);
		subRound128(BB, CC, DD, AA, f32, KK1, 11, data[2]);

		T = B; B = BB; BB = T;

	// j=32...47 
		subRound128(A, B, C, D, f32, K2, 11, data[3]);
		subRound128(D, A, B, C, f32, K2, 13, data[10]);
		subRound128(C, D, A, B, f32, K2, 6, data[14]);
		subRound128(B, C, D, A, f32, K2, 7, data[4]);
		subRound128(A, B, C, D, f32, K2, 14, data[9]);
		subRound128(D, A, B, C, f32, K2, 9, data[15]);
		subRound128(C, D, A, B, f32, K2, 13, data[8]);
		subRound128(B, C, D, A, f32, K2, 15, data[1]);
		subRound128(A, B, C, D, f32, K2, 14, data[2]);
		subRound128(D, A, B, C, f32, K2, 8, data[7]);
		subRound128(C, D, A, B, f32, K2, 13, data[0]);
		subRound128(B, C, D, A, f32, K2, 6, data[6]);
		subRound128(A, B, C, D, f32, K2, 5, data[13]);
		subRound128(D, A, B, C, f32, K2, 12, data[11]);
		subRound128(C, D, A, B, f32, K2, 7, data[5]);
		subRound128(B, C, D, A, f32, K2, 5, data[12]);

		subRound128(AA, BB, CC, DD, f16, KK2, 9, data[15]);
		subRound128(DD, AA, BB, CC, f16, KK2, 7, data[5]);
		subRound128(CC, DD, AA, BB, f16, KK2, 15, data[1]);
		subRound128(BB, CC, DD, AA, f16, KK2, 11, data[3]);
		subRound128(AA, BB, CC, DD, f16, KK2, 8, data[7]);
		subRound128(DD, AA, BB, CC, f16, KK2, 6, data[14]);
		subRound128(CC, DD, AA, BB, f16, KK2, 6, data[6]);
		subRound128(BB, CC, DD, AA, f16, KK2, 14, data[9]);
		subRound128(AA, BB, CC, DD, f16, KK2, 12, data[11]);
		subRound128(DD, AA, BB, CC, f16, KK2, 13, data[8]);
		subRound128(CC, DD, AA, BB, f16, KK2, 5, data[12]);
		subRound128(BB, CC, DD, AA, f16, KK2, 14, data[2]);
		subRound128(AA, BB, CC, DD, f16, KK2, 13, data[10]);
		subRound128(DD, AA, BB, CC, f16, KK2, 13, data[0]);
		subRound128(CC, DD, AA, BB, f16, KK2, 7, data[4]);
		subRound128(BB, CC, DD, AA, f16, KK2, 5, data[13]);

		T = C; C = CC; CC = T;

	// j=48...63 
		subRound128(A, B, C, D, f48, K3, 11, data[1]);
		subRound128(D, A, B, C, f48, K3, 12, data[9]);
		subRound128(C, D, A, B, f48, K3, 14, data[11]);
		subRound128(B, C, D, A, f48, K3, 15, data[10]);
		subRound128(A, B, C, D, f48, K3, 14, data[0]);
		subRound128(D, A, B, C, f48, K3, 15, data[8]);
		subRound128(C, D, A, B, f48, K3, 9, data[12]);
		subRound128(B, C, D, A, f48, K3, 8, data[4]);
		subRound128(A, B, C, D, f48, K3, 9, data[13]);
		subRound128(D, A, B, C, f48, K3, 14, data[3]);
		subRound128(C, D, A, B, f48, K3, 5, data[7]);
		subRound128(B, C, D, A, f48, K3, 6, data[15]);
		subRound128(A, B, C, D, f48, K3, 8, data[14]);
		subRound128(D, A, B, C, f48, K3, 6, data[5]);
		subRound128(C, D, A, B, f48, K3, 5, data[6]);
		subRound128(B, C, D, A, f48, K3, 12, data[2]);

		subRound128(AA, BB, CC, DD, f0, KK4, 15, data[8]);
		subRound128(DD, AA, BB, CC, f0, KK4, 5, data[6]);
		subRound128(CC, DD, AA, BB, f0, KK4, 8, data[4]);
		subRound128(BB, CC, DD, AA, f0, KK4, 11, data[1]);
		subRound128(AA, BB, CC, DD, f0, KK4, 14, data[3]);
		subRound128(DD, AA, BB, CC, f0, KK4, 14, data[11]);
		subRound128(CC, DD, AA, BB, f0, KK4, 6, data[15]);
		subRound128(BB, CC, DD, AA, f0, KK4, 14, data[0]);
		subRound128(AA, BB, CC, DD, f0, KK4, 6, data[5]);
		subRound128(DD, AA, BB, CC, f0, KK4, 9, data[12]);
		subRound128(CC, DD, AA, BB, f0, KK4, 12, data[2]);
		subRound128(BB, CC, DD, AA, f0, KK4, 9, data[13]);
		subRound128(AA, BB, CC, DD, f0, KK4, 12, data[9]);
		subRound128(DD, AA, BB, CC, f0, KK4, 5, data[7]);
		subRound128(CC, DD, AA, BB, f0, KK4, 15, data[10]);
		subRound128(BB, CC, DD, AA, f0, KK4, 8, data[14]);

		// T = D; D = DD; DD = T; 

					ctx->digest[0] += A;
					ctx->digest[1] += B;
					ctx->digest[2] += C;
					ctx->digest[3] += DD;
					ctx->digest[4] += AA;
					ctx->digest[5] += BB;
					ctx->digest[6] += CC;
					ctx->digest[7] += D;
	}

	static void ripemd320_transform(struct ripemd_ctx *ctx, word32 * data)
	{
		word32 A, B, C, D, E;   	// Local vars 
		word32 AA, BB, CC, DD, EE;      // Local vars 
		word32 T;

		// Set up first buffer and local data buffer
		A = ctx->digest[0];
		B = ctx->digest[1];
		C = ctx->digest[2];
		D = ctx->digest[3];
		E = ctx->digest[4];
		AA = ctx->digest[5];
		BB = ctx->digest[6];
		CC = ctx->digest[7];
		DD = ctx->digest[8];
		EE = ctx->digest[9];

	// j=0...15 
		subRound160(A, B, C, D, E, f0, K0, 11, data[0]);
		subRound160(E, A, B, C, D, f0, K0, 14, data[1]);
		subRound160(D, E, A, B, C, f0, K0, 15, data[2]);
		subRound160(C, D, E, A, B, f0, K0, 12, data[3]);
		subRound160(B, C, D, E, A, f0, K0, 5, data[4]);
		subRound160(A, B, C, D, E, f0, K0, 8, data[5]);
		subRound160(E, A, B, C, D, f0, K0, 7, data[6]);
		subRound160(D, E, A, B, C, f0, K0, 9, data[7]);
		subRound160(C, D, E, A, B, f0, K0, 11, data[8]);
		subRound160(B, C, D, E, A, f0, K0, 13, data[9]);
		subRound160(A, B, C, D, E, f0, K0, 14, data[10]);
		subRound160(E, A, B, C, D, f0, K0, 15, data[11]);
		subRound160(D, E, A, B, C, f0, K0, 6, data[12]);
		subRound160(C, D, E, A, B, f0, K0, 7, data[13]);
		subRound160(B, C, D, E, A, f0, K0, 9, data[14]);
		subRound160(A, B, C, D, E, f0, K0, 8, data[15]);

		subRound160(AA, BB, CC, DD, EE, f64, KK0, 8, data[5]);
		subRound160(EE, AA, BB, CC, DD, f64, KK0, 9, data[14]);
		subRound160(DD, EE, AA, BB, CC, f64, KK0, 9, data[7]);
		subRound160(CC, DD, EE, AA, BB, f64, KK0, 11, data[0]);
		subRound160(BB, CC, DD, EE, AA, f64, KK0, 13, data[9]);
		subRound160(AA, BB, CC, DD, EE, f64, KK0, 15, data[2]);
		subRound160(EE, AA, BB, CC, DD, f64, KK0, 15, data[11]);
		subRound160(DD, EE, AA, BB, CC, f64, KK0, 5, data[4]);
		subRound160(CC, DD, EE, AA, BB, f64, KK0, 7, data[13]);
		subRound160(BB, CC, DD, EE, AA, f64, KK0, 7, data[6]);
		subRound160(AA, BB, CC, DD, EE, f64, KK0, 8, data[15]);
		subRound160(EE, AA, BB, CC, DD, f64, KK0, 11, data[8]);
		subRound160(DD, EE, AA, BB, CC, f64, KK0, 14, data[1]);
		subRound160(CC, DD, EE, AA, BB, f64, KK0, 14, data[10]);
		subRound160(BB, CC, DD, EE, AA, f64, KK0, 12, data[3]);
		subRound160(AA, BB, CC, DD, EE, f64, KK0, 6, data[12]);

		T = A; A = AA; AA = T;

	// j=16...31 
		subRound160(E, A, B, C, D, f16, K1, 7, data[7]);
		subRound160(D, E, A, B, C, f16, K1, 6, data[4]);
		subRound160(C, D, E, A, B, f16, K1, 8, data[13]);
		subRound160(B, C, D, E, A, f16, K1, 13, data[1]);
		subRound160(A, B, C, D, E, f16, K1, 11, data[10]);
		subRound160(E, A, B, C, D, f16, K1, 9, data[6]);
		subRound160(D, E, A, B, C, f16, K1, 7, data[15]);
		subRound160(C, D, E, A, B, f16, K1, 15, data[3]);
		subRound160(B, C, D, E, A, f16, K1, 7, data[12]);
		subRound160(A, B, C, D, E, f16, K1, 12, data[0]);
		subRound160(E, A, B, C, D, f16, K1, 15, data[9]);
		subRound160(D, E, A, B, C, f16, K1, 9, data[5]);
		subRound160(C, D, E, A, B, f16, K1, 11, data[2]);
		subRound160(B, C, D, E, A, f16, K1, 7, data[14]);
		subRound160(A, B, C, D, E, f16, K1, 13, data[11]);
		subRound160(E, A, B, C, D, f16, K1, 12, data[8]);

		subRound160(EE, AA, BB, CC, DD, f48, KK1, 9, data[6]);
		subRound160(DD, EE, AA, BB, CC, f48, KK1, 13, data[11]);
		subRound160(CC, DD, EE, AA, BB, f48, KK1, 15, data[3]);
		subRound160(BB, CC, DD, EE, AA, f48, KK1, 7, data[7]);
		subRound160(AA, BB, CC, DD, EE, f48, KK1, 12, data[0]);
		subRound160(EE, AA, BB, CC, DD, f48, KK1, 8, data[13]);
		subRound160(DD, EE, AA, BB, CC, f48, KK1, 9, data[5]);
		subRound160(CC, DD, EE, AA, BB, f48, KK1, 11, data[10]);
		subRound160(BB, CC, DD, EE, AA, f48, KK1, 7, data[14]);
		subRound160(AA, BB, CC, DD, EE, f48, KK1, 7, data[15]);
		subRound160(EE, AA, BB, CC, DD, f48, KK1, 12, data[8]);
		subRound160(DD, EE, AA, BB, CC, f48, KK1, 7, data[12]);
		subRound160(CC, DD, EE, AA, BB, f48, KK1, 6, data[4]);
		subRound160(BB, CC, DD, EE, AA, f48, KK1, 15, data[9]);
		subRound160(AA, BB, CC, DD, EE, f48, KK1, 13, data[1]);
		subRound160(EE, AA, BB, CC, DD, f48, KK1, 11, data[2]);

		T = B; B = BB; BB = T;

	// j=32...47 
		subRound160(D, E, A, B, C, f32, K2, 11, data[3]);
		subRound160(C, D, E, A, B, f32, K2, 13, data[10]);
		subRound160(B, C, D, E, A, f32, K2, 6, data[14]);
		subRound160(A, B, C, D, E, f32, K2, 7, data[4]);
		subRound160(E, A, B, C, D, f32, K2, 14, data[9]);
		subRound160(D, E, A, B, C, f32, K2, 9, data[15]);
		subRound160(C, D, E, A, B, f32, K2, 13, data[8]);
		subRound160(B, C, D, E, A, f32, K2, 15, data[1]);
		subRound160(A, B, C, D, E, f32, K2, 14, data[2]);
		subRound160(E, A, B, C, D, f32, K2, 8, data[7]);
		subRound160(D, E, A, B, C, f32, K2, 13, data[0]);
		subRound160(C, D, E, A, B, f32, K2, 6, data[6]);
		subRound160(B, C, D, E, A, f32, K2, 5, data[13]);
		subRound160(A, B, C, D, E, f32, K2, 12, data[11]);
		subRound160(E, A, B, C, D, f32, K2, 7, data[5]);
		subRound160(D, E, A, B, C, f32, K2, 5, data[12]);

		subRound160(DD, EE, AA, BB, CC, f32, KK2, 9, data[15]);
		subRound160(CC, DD, EE, AA, BB, f32, KK2, 7, data[5]);
		subRound160(BB, CC, DD, EE, AA, f32, KK2, 15, data[1]);
		subRound160(AA, BB, CC, DD, EE, f32, KK2, 11, data[3]);
		subRound160(EE, AA, BB, CC, DD, f32, KK2, 8, data[7]);
		subRound160(DD, EE, AA, BB, CC, f32, KK2, 6, data[14]);
		subRound160(CC, DD, EE, AA, BB, f32, KK2, 6, data[6]);
		subRound160(BB, CC, DD, EE, AA, f32, KK2, 14, data[9]);
		subRound160(AA, BB, CC, DD, EE, f32, KK2, 12, data[11]);
		subRound160(EE, AA, BB, CC, DD, f32, KK2, 13, data[8]);
		subRound160(DD, EE, AA, BB, CC, f32, KK2, 5, data[12]);
		subRound160(CC, DD, EE, AA, BB, f32, KK2, 14, data[2]);
		subRound160(BB, CC, DD, EE, AA, f32, KK2, 13, data[10]);
		subRound160(AA, BB, CC, DD, EE, f32, KK2, 13, data[0]);
		subRound160(EE, AA, BB, CC, DD, f32, KK2, 7, data[4]);
		subRound160(DD, EE, AA, BB, CC, f32, KK2, 5, data[13]);

		T = C; C = CC; CC = T;

	// j=48...63 
		subRound160(C, D, E, A, B, f48, K3, 11, data[1]);
		subRound160(B, C, D, E, A, f48, K3, 12, data[9]);
		subRound160(A, B, C, D, E, f48, K3, 14, data[11]);
		subRound160(E, A, B, C, D, f48, K3, 15, data[10]);
		subRound160(D, E, A, B, C, f48, K3, 14, data[0]);
		subRound160(C, D, E, A, B, f48, K3, 15, data[8]);
		subRound160(B, C, D, E, A, f48, K3, 9, data[12]);
		subRound160(A, B, C, D, E, f48, K3, 8, data[4]);
		subRound160(E, A, B, C, D, f48, K3, 9, data[13]);
		subRound160(D, E, A, B, C, f48, K3, 14, data[3]);
		subRound160(C, D, E, A, B, f48, K3, 5, data[7]);
		subRound160(B, C, D, E, A, f48, K3, 6, data[15]);
		subRound160(A, B, C, D, E, f48, K3, 8, data[14]);
		subRound160(E, A, B, C, D, f48, K3, 6, data[5]);
		subRound160(D, E, A, B, C, f48, K3, 5, data[6]);
		subRound160(C, D, E, A, B, f48, K3, 12, data[2]);

		subRound160(CC, DD, EE, AA, BB, f16, KK3, 15, data[8]);
		subRound160(BB, CC, DD, EE, AA, f16, KK3, 5, data[6]);
		subRound160(AA, BB, CC, DD, EE, f16, KK3, 8, data[4]);
		subRound160(EE, AA, BB, CC, DD, f16, KK3, 11, data[1]);
		subRound160(DD, EE, AA, BB, CC, f16, KK3, 14, data[3]);
		subRound160(CC, DD, EE, AA, BB, f16, KK3, 14, data[11]);
		subRound160(BB, CC, DD, EE, AA, f16, KK3, 6, data[15]);
		subRound160(AA, BB, CC, DD, EE, f16, KK3, 14, data[0]);
		subRound160(EE, AA, BB, CC, DD, f16, KK3, 6, data[5]);
		subRound160(DD, EE, AA, BB, CC, f16, KK3, 9, data[12]);
		subRound160(CC, DD, EE, AA, BB, f16, KK3, 12, data[2]);
		subRound160(BB, CC, DD, EE, AA, f16, KK3, 9, data[13]);
		subRound160(AA, BB, CC, DD, EE, f16, KK3, 12, data[9]);
		subRound160(EE, AA, BB, CC, DD, f16, KK3, 5, data[7]);
		subRound160(DD, EE, AA, BB, CC, f16, KK3, 15, data[10]);
		subRound160(CC, DD, EE, AA, BB, f16, KK3, 8, data[14]);

		T = D; D = DD; DD = T;

	// j=64...79 
		subRound160(B, C, D, E, A, f64, K4, 9, data[4]);
		subRound160(A, B, C, D, E, f64, K4, 15, data[0]);
		subRound160(E, A, B, C, D, f64, K4, 5, data[5]);
		subRound160(D, E, A, B, C, f64, K4, 11, data[9]);
		subRound160(C, D, E, A, B, f64, K4, 6, data[7]);
		subRound160(B, C, D, E, A, f64, K4, 8, data[12]);
		subRound160(A, B, C, D, E, f64, K4, 13, data[2]);
		subRound160(E, A, B, C, D, f64, K4, 12, data[10]);
		subRound160(D, E, A, B, C, f64, K4, 5, data[14]);
		subRound160(C, D, E, A, B, f64, K4, 12, data[1]);
		subRound160(B, C, D, E, A, f64, K4, 13, data[3]);
		subRound160(A, B, C, D, E, f64, K4, 14, data[8]);
		subRound160(E, A, B, C, D, f64, K4, 11, data[11]);
		subRound160(D, E, A, B, C, f64, K4, 8, data[6]);
		subRound160(C, D, E, A, B, f64, K4, 5, data[15]);
		subRound160(B, C, D, E, A, f64, K4, 6, data[13]);

		subRound160(BB, CC, DD, EE, AA, f0, KK4, 8, data[12]);
		subRound160(AA, BB, CC, DD, EE, f0, KK4, 5, data[15]);
		subRound160(EE, AA, BB, CC, DD, f0, KK4, 12, data[10]);
		subRound160(DD, EE, AA, BB, CC, f0, KK4, 9, data[4]);
		subRound160(CC, DD, EE, AA, BB, f0, KK4, 12, data[1]);
		subRound160(BB, CC, DD, EE, AA, f0, KK4, 5, data[5]);
		subRound160(AA, BB, CC, DD, EE, f0, KK4, 14, data[8]);
		subRound160(EE, AA, BB, CC, DD, f0, KK4, 6, data[7]);
		subRound160(DD, EE, AA, BB, CC, f0, KK4, 8, data[6]);
		subRound160(CC, DD, EE, AA, BB, f0, KK4, 13, data[2]);
		subRound160(BB, CC, DD, EE, AA, f0, KK4, 6, data[13]);
		subRound160(AA, BB, CC, DD, EE, f0, KK4, 5, data[14]);
		subRound160(EE, AA, BB, CC, DD, f0, KK4, 15, data[0]);
		subRound160(DD, EE, AA, BB, CC, f0, KK4, 13, data[3]);
		subRound160(CC, DD, EE, AA, BB, f0, KK4, 11, data[9]);
		subRound160(BB, CC, DD, EE, AA, f0, KK4, 11, data[11]);

		// T = E; E = EE; EE = T; 

					ctx->digest[0] += A;
					ctx->digest[1] += B;
					ctx->digest[2] += C;
					ctx->digest[3] += D;
					ctx->digest[4] += EE;
					ctx->digest[5] += AA;
					ctx->digest[6] += BB;
					ctx->digest[7] += CC;
					ctx->digest[8] += DD;
					ctx->digest[9] += E;
	}

	static void ripemd_transform(struct ripemd_ctx *ctx, word32 * data)
	{
		switch(ctx->digest_len)  // select the right compression function
			{
			case 8 * RIPEMD128_DIGESTSIZE: 
				ripemd128_transform(ctx, data); break;
			case 8 * RIPEMD160_DIGESTSIZE: 
				ripemd160_transform(ctx, data); break;
			case 8 * RIPEMD256_DIGESTSIZE: 
				ripemd256_transform(ctx, data); break;
			case 8 * RIPEMD320_DIGESTSIZE: 
				ripemd320_transform(ctx, data); break;
			}
	}

	#if 1

	#ifndef EXTRACT_UCHAR
	#define EXTRACT_UCHAR(p)  (*(word8 *)(p))
	#endif

	#define STRING2INT(s) ((((((EXTRACT_UCHAR(s+3) << 8)    \
				 | EXTRACT_UCHAR(s+2)) << 8)  \
				 | EXTRACT_UCHAR(s+1)) << 8)  \
				 | EXTRACT_UCHAR(s))
	#else
	word32 STRING2INT(word8 * s)
	{
		word32 r;
		int i;

		for (i = 0, r = 0; i < 4; i++)
			r = (r << 8) | s[3-i];
		return r;
	}
	#endif

	static void ripemd_block(struct ripemd_ctx *ctx, word8 * block)
	{
		word32 data[RIPEMD_DATALEN];
		int i;

		// Update bit count
		ctx->bitcount += RIPEMD_DATASIZE * 8;

		// Endian independent conversion 
		for (i = 0; i < RIPEMD_DATALEN; i++, block += 4)
			data[i] = STRING2INT(block);

		ripemd_transform(ctx, data);
	}

	void ripemd_update(struct ripemd_ctx *ctx, word8 * buffer, word32 len)
	{
		if (ctx->index) {	// Try to fill partial block 
			unsigned left = RIPEMD_DATASIZE - ctx->index;
			if (len < left) {
				memcpy(ctx->block + ctx->index, buffer, len);
				ctx->index += len;
				return;	// Finished 
			} else {
				memcpy(ctx->block + ctx->index, buffer, left);
				ripemd_block(ctx, ctx->block);
				buffer += left;
				len -= left;
			}
		}
		while (len >= RIPEMD_DATASIZE) {
			ripemd_block(ctx, buffer);
			buffer += RIPEMD_DATASIZE;
			len -= RIPEMD_DATASIZE;
		}
		if ((ctx->index = len))
			// This assignment is intended 
			// Buffer leftovers 
			memcpy(ctx->block, buffer, len);
	}

	// Final wrapup - pad to RIPEMD_DATASIZE-byte boundary with the bit pattern
	//   1 0* (64-bit count of bits processed, LSB-first)

	void ripemd_final(struct ripemd_ctx *ctx)
	{
		word32 data[RIPEMD_DATALEN];
		int i;
		int words;

		i = ctx->index;
		// Set the first char of padding to 0x80.  This is safe since there is
		// always at least one byte free
		ctx->block[i++] = 0x80;

		// Fill rest of word
		for (; i & 3; i++)
			ctx->block[i] = 0;

		// i is now a multiple of the word size 4 
		words = i >> 2;
		for (i = 0; i < words; i++)
			data[i] = STRING2INT(ctx->block + 4 * i);

		if (words > (RIPEMD_DATALEN - 2)) { // No room for length in this block. 
					 // Process it and pad with another one 
			for (i = words; i < RIPEMD_DATALEN; i++)
				data[i] = 0;
			ripemd_transform(ctx, data);
			for (i = 0; i < (RIPEMD_DATALEN - 2); i++)
				data[i] = 0;
		} else
			for (i = words; i < RIPEMD_DATALEN - 2; i++)
				data[i] = 0;

		// add length padding
		ctx->bitcount += 8 * ctx->index;
		data[RIPEMD_DATALEN - 2] = (unsigned __int32) (ctx->bitcount & (((unsigned __int64)1 << 32) - 1));
		data[RIPEMD_DATALEN - 1] = (unsigned __int32) (ctx->bitcount >> 32);

		ripemd_transform(ctx, data);
	}

	void ripemd_digest(struct ripemd_ctx *ctx, word8 * s)
	{
		int i;

		if (s!=NULL)
		for (i = 0; i < ctx->digest_len / 32; i++) {
			*s++ = 0xff & (unsigned char)ctx->digest[i];
			*s++ = 0xff & (unsigned char)(ctx->digest[i] >> 8);
			*s++ = 0xff & (unsigned char)(ctx->digest[i] >> 16);
			*s++ = 0xff & (unsigned char)(ctx->digest[i] >> 24);
		}
	}



	// those are the standard RIPEMD128 test vectors
	 
	static char *ripemd128_msg[] = 
	{
			"",
			"a",
			"abc",
			"message digest",
			"abcdefghijklmnopqrstuvwxyz",
			"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
			"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
			"12345678901234567890123456789012345678901234567890123456789012345678901234567890",
			NULL
	};

	static char *ripemd128_val[] =
	{
			"cdf26213a150dc3ecb610f18f6b38b46",
			"86be7afa339d0fc7cfc785e72f578d33",
			"c14a12199c66e4ba84636b0f69144c77",
			"9e327b3d6e523062afc1132d7df9d1b8",
			"fd2aa607f71dc8f510714922b371834e",
			"a1aa0689d0fafa2ddc22e88b49133a06",
			"d1e959eb179c911faea4624c60c5c702",
			"3f45ef194732c2dbb2c4a2c769795fa3",
			"4a7f5723f954eba1216c9d8f6320431f"
	};






		// tests with standard test vectors and returns true if everything is OK
	bool ripemd128_test(void) {
		int i, j;
		char output[66];
		ripemd_ctx ctx;
		unsigned char buf[1000];
		unsigned char ripemd_sum[32];

				
		for( i = 0; i < 9; i++ ) {
			ripemd128_init( &ctx );
			if( i < 8 ) {
				ripemd_update( &ctx, (word8 *) ripemd128_msg[i],(int)strlen( ripemd128_msg[i] ) );
			}
			else {
				memset( buf, 'a', 1000 );
					
				for( j = 0; j < 1000; j++ ) {
					ripemd_update( &ctx, (word8 *) buf, 1000 );
				}
			}
				
			ripemd_final( &ctx );
			ripemd_digest( &ctx, ripemd_sum );


			for( j = 0; j < 16; j++ ) {
				sprintf_s( output + j * 2,3, "%02x", ripemd_sum[j] );
			}
				
			if( memcmp( output, ripemd128_val[i], 32 ) ) {
				return false;
			}
		}
		return true;
	}








	// those are the standard RIPEMD160 test vectors
	 
	static char *ripemd160_msg[] = 
	{
			"",
			"a",
			"abc",
			"message digest",
			"abcdefghijklmnopqrstuvwxyz",
			"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
			"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
			"12345678901234567890123456789012345678901234567890123456789012345678901234567890",
			NULL
	};

	static char *ripemd160_val[] =
	{
			"9c1185a5c5e9fc54612808977ee8f548b2258d31",
			"0bdc9d2d256b3ee9daae347be6f4dc835a467ffe",
			"8eb208f7e05d987a9b044a8e98c6b087f15a0bfc",
			"5d0689ef49d2fae572b881b123a85ffa21595f36",
			"f71c27109c692c1b56bbdceb5b9d2865b3708dbc",
			"12a053384a9c0c88e405a06c27dcf49ada62eb2b",
			"b0e20b6e3116640286ed3a87a5713079b21f5189",
			"9b752e45573d4b39f4dbd3323cab82bf63326bfb",
			"52783243c1697bdbe16d37f97f68f08325dc1528"
	};



		// tests with standard test vectors and returns true if everything is OK
	bool ripemd160_test(void) {
		int i, j;
		char output[66];
		ripemd_ctx ctx;
		unsigned char buf[1000];
		unsigned char ripemd_sum[20];

				
		for( i = 0; i < 9; i++ ) {
			ripemd160_init( &ctx );
			if( i < 8 ) {
				ripemd_update( &ctx, (word8 *) ripemd160_msg[i],(int)strlen( ripemd160_msg[i] ) );
			}
			else {
				memset( buf, 'a', 1000 );
					
				for( j = 0; j < 1000; j++ ) {
					ripemd_update( &ctx, (word8 *) buf, 1000 );
				}
			}
				
			ripemd_final( &ctx );
			ripemd_digest( &ctx, ripemd_sum );


			for( j = 0; j < 20; j++ ) {
				sprintf_s( output + j * 2,3, "%02x", ripemd_sum[j] );
			}
				
			if( memcmp( output, ripemd160_val[i], 40 ) ) {
				return false;
			}
		}
		return true;
	}





	// those are the standard RIPEMD256 test vectors
	 
	static char *ripemd256_msg[] = 
	{
			"",
			"a",
			"abc",
			"message digest",
			"abcdefghijklmnopqrstuvwxyz",
			"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
			"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
			"12345678901234567890123456789012345678901234567890123456789012345678901234567890",
			NULL
	};

	static char *ripemd256_val[] =
	{
			"02ba4c4e5f8ecd1877fc52d64d30e37a2d9774fb1e5d026380ae0168e3c5522d",
			"f9333e45d857f5d90a91bab70a1eba0cfb1be4b0783c9acfcd883a9134692925",
			"afbd6e228b9d8cbbcef5ca2d03e6dba10ac0bc7dcbe4680e1e42d2e975459b65",
			"87e971759a1ce47a514d5c914c392c9018c7c46bc14465554afcdf54a5070c0e",
			"649d3034751ea216776bf9a18acc81bc7896118a5197968782dd1fd97d8d5133",
			"3843045583aac6c8c8d9128573e7a9809afb2a0f34ccc36ea9e72f16f6368e3f",
			"5740a408ac16b720b84424ae931cbb1fe363d1d0bf4017f1a89f7ea6de77a0b8",
			"06fdcc7a409548aaf91368c06a6275b553e3f099bf0ea4edfd6778df89a890dd",
			"ac953744e10e31514c150d4d8d7b677342e33399788296e43ae4850ce4f97978"
	};



		// tests with standard test vectors and returns true if everything is OK
	bool ripemd256_test(void) {
		int i, j;
		char output[66];
		ripemd_ctx ctx;
		unsigned char buf[1000];
		unsigned char ripemd_sum[32];

				
		for( i = 0; i < 9; i++ ) {
			ripemd256_init( &ctx );
			if( i < 8 ) {
				ripemd_update( &ctx, (word8 *) ripemd256_msg[i],(int)strlen( ripemd256_msg[i] ) );
			}
			else {
				memset( buf, 'a', 1000 );
					
				for( j = 0; j < 1000; j++ ) {
					ripemd_update( &ctx, (word8 *) buf, 1000 );
				}
			}
				
			ripemd_final( &ctx );
			ripemd_digest( &ctx, ripemd_sum );


			for( j = 0; j < 32; j++ ) {
				sprintf_s( output + j * 2,3, "%02x", ripemd_sum[j] );
			}
				
			if( memcmp( output, ripemd256_val[i], 64 ) ) {
				return false;
			}
		}
		return true;
	}






	// those are the standard RIPEMD256 test vectors
	 
	static char *ripemd320_msg[] = 
	{
			"",
			"a",
			"abc",
			"message digest",
			"abcdefghijklmnopqrstuvwxyz",
			"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
			"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
			"12345678901234567890123456789012345678901234567890123456789012345678901234567890",
			NULL
	};

	static char *ripemd320_val[] =
	{
			"22d65d5661536cdc75c1fdf5c6de7b41b9f27325ebc61e8557177d705a0ec880151c3a32a00899b8",
			"ce78850638f92658a5a585097579926dda667a5716562cfcf6fbe77f63542f99b04705d6970dff5d",
			"de4c01b3054f8930a79d09ae738e92301e5a17085beffdc1b8d116713e74f82fa942d64cdbc4682d",
			"3a8e28502ed45d422f68844f9dd316e7b98533fa3f2a91d29f84d425c88d6b4eff727df66a7c0197",
			"cabdb1810b92470a2093aa6bce05952c28348cf43ff60841975166bb40ed234004b8824463e6b009",
			"d034a7950cf722021ba4b84df769a5de2060e259df4c9bb4a4268c0e935bbc7470a969c9d072a1ac",
			"ed544940c86d67f250d232c30b7b3e5770e0c60c8cb9a4cafe3b11388af9920e1b99230b843c86a4",
			"557888af5f6d8ed62ab66945c6d2a0a47ecd5341e915eb8fea1d0524955f825dc717e4a008ab2d42",
			"bdee37f4371e20646b8b0d862dda16292ae36f40965e8c8509e63d1dbddecc503e2b63eb9245bb66"
	};



		// tests with standard test vectors and returns true if everything is OK
	bool ripemd320_test(void) {
		int i, j;
		char output[84];
		ripemd_ctx ctx;
		unsigned char buf[1000];
		unsigned char ripemd_sum[40];

				
		for( i = 0; i < 9; i++ ) {
			ripemd320_init( &ctx );
			if( i < 8 ) {
				ripemd_update( &ctx, (word8 *) ripemd320_msg[i],(int)strlen( ripemd256_msg[i] ) );
			}
			else {
				memset( buf, 'a', 1000 );
					
				for( j = 0; j < 1000; j++ ) {
					ripemd_update( &ctx, (word8 *) buf, 1000 );
				}
			}
				
			ripemd_final( &ctx );
			ripemd_digest( &ctx, ripemd_sum );


			for( j = 0; j < 40; j++ ) {
				sprintf_s( output + j * 2,3, "%02x", ripemd_sum[j] );
			}
				
			if( memcmp( output, ripemd320_val[i], 80 ) ) {
				return false;
			}
		}
		return true;
	}




















//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
// RFC 1321 compliant MD5 implementation,
// by Christophe Devine <devine@cr0.net>;
// this is freeware
	#define GET_UINT32_MD5(n,b,i)                   \
	{                                               \
			(n) = ( (uint32) (b)[(i) + 3] << 24 )       \
					| ( (uint32) (b)[(i) + 2] << 16 )       \
					| ( (uint32) (b)[(i) + 1] <<  8 )       \
					| ( (uint32) (b)[(i)    ]       );      \
	}

	#define PUT_UINT32_MD5(n,b,i)                   \
	{                                               \
			(b)[(i)    ] = (uint8) ( (n)       );       \
			(b)[(i) + 1] = (uint8) ( (n) >>  8 );       \
			(b)[(i) + 2] = (uint8) ( (n) >> 16 );       \
			(b)[(i) + 3] = (uint8) ( (n) >> 24 );       \
	}

	void md5_starts( md5_context *ctx )
	{
			ctx->total[0] = 0;
			ctx->total[1] = 0;
			ctx->state[0] = 0x67452301;
			ctx->state[1] = 0xEFCDAB89;
			ctx->state[2] = 0x98BADCFE;
			ctx->state[3] = 0x10325476;
	}

	void md5_process( md5_context *ctx, uint8 data[64] )
	{
			uint32 A, B, C, D, X[16];

			GET_UINT32_MD5( X[0],  data,  0 );
			GET_UINT32_MD5( X[1],  data,  4 );
			GET_UINT32_MD5( X[2],  data,  8 );
			GET_UINT32_MD5( X[3],  data, 12 );
			GET_UINT32_MD5( X[4],  data, 16 );
			GET_UINT32_MD5( X[5],  data, 20 );
			GET_UINT32_MD5( X[6],  data, 24 );
			GET_UINT32_MD5( X[7],  data, 28 );
			GET_UINT32_MD5( X[8],  data, 32 );
			GET_UINT32_MD5( X[9],  data, 36 );
			GET_UINT32_MD5( X[10], data, 40 );
			GET_UINT32_MD5( X[11], data, 44 );
			GET_UINT32_MD5( X[12], data, 48 );
			GET_UINT32_MD5( X[13], data, 52 );
			GET_UINT32_MD5( X[14], data, 56 );
			GET_UINT32_MD5( X[15], data, 60 );

	#define S(x,n) ((x << n) | ((x & 0xFFFFFFFF) >> (32 - n)))

	#define P_MD5(a,b,c,d,k,s,t)                            \
	{                                                       \
			a += F(b,c,d) + X[k] + t; a = S(a,s) + b;           \
	}

			A = ctx->state[0];
			B = ctx->state[1];
			C = ctx->state[2];
			D = ctx->state[3];

	#define F(x,y,z) (z ^ (x & (y ^ z)))

			P_MD5( A, B, C, D,  0,  7, 0xD76AA478 );
			P_MD5( D, A, B, C,  1, 12, 0xE8C7B756 );
			P_MD5( C, D, A, B,  2, 17, 0x242070DB );
			P_MD5( B, C, D, A,  3, 22, 0xC1BDCEEE );
			P_MD5( A, B, C, D,  4,  7, 0xF57C0FAF );
			P_MD5( D, A, B, C,  5, 12, 0x4787C62A );
			P_MD5( C, D, A, B,  6, 17, 0xA8304613 );
			P_MD5( B, C, D, A,  7, 22, 0xFD469501 );
			P_MD5( A, B, C, D,  8,  7, 0x698098D8 );
			P_MD5( D, A, B, C,  9, 12, 0x8B44F7AF );
			P_MD5( C, D, A, B, 10, 17, 0xFFFF5BB1 );
			P_MD5( B, C, D, A, 11, 22, 0x895CD7BE );
			P_MD5( A, B, C, D, 12,  7, 0x6B901122 );
			P_MD5( D, A, B, C, 13, 12, 0xFD987193 );
			P_MD5( C, D, A, B, 14, 17, 0xA679438E );
			P_MD5( B, C, D, A, 15, 22, 0x49B40821 );

	#undef F

	#define F(x,y,z) (y ^ (z & (x ^ y)))

			P_MD5( A, B, C, D,  1,  5, 0xF61E2562 );
			P_MD5( D, A, B, C,  6,  9, 0xC040B340 );
			P_MD5( C, D, A, B, 11, 14, 0x265E5A51 );
			P_MD5( B, C, D, A,  0, 20, 0xE9B6C7AA );
			P_MD5( A, B, C, D,  5,  5, 0xD62F105D );
			P_MD5( D, A, B, C, 10,  9, 0x02441453 );
			P_MD5( C, D, A, B, 15, 14, 0xD8A1E681 );
			P_MD5( B, C, D, A,  4, 20, 0xE7D3FBC8 );
			P_MD5( A, B, C, D,  9,  5, 0x21E1CDE6 );
			P_MD5( D, A, B, C, 14,  9, 0xC33707D6 );
			P_MD5( C, D, A, B,  3, 14, 0xF4D50D87 );
			P_MD5( B, C, D, A,  8, 20, 0x455A14ED );
			P_MD5( A, B, C, D, 13,  5, 0xA9E3E905 );
			P_MD5( D, A, B, C,  2,  9, 0xFCEFA3F8 );
			P_MD5( C, D, A, B,  7, 14, 0x676F02D9 );
			P_MD5( B, C, D, A, 12, 20, 0x8D2A4C8A );

	#undef F
	    
	#define F(x,y,z) (x ^ y ^ z)

			P_MD5( A, B, C, D,  5,  4, 0xFFFA3942 );
			P_MD5( D, A, B, C,  8, 11, 0x8771F681 );
			P_MD5( C, D, A, B, 11, 16, 0x6D9D6122 );
			P_MD5( B, C, D, A, 14, 23, 0xFDE5380C );
			P_MD5( A, B, C, D,  1,  4, 0xA4BEEA44 );
			P_MD5( D, A, B, C,  4, 11, 0x4BDECFA9 );
			P_MD5( C, D, A, B,  7, 16, 0xF6BB4B60 );
			P_MD5( B, C, D, A, 10, 23, 0xBEBFBC70 );
			P_MD5( A, B, C, D, 13,  4, 0x289B7EC6 );
			P_MD5( D, A, B, C,  0, 11, 0xEAA127FA );
			P_MD5( C, D, A, B,  3, 16, 0xD4EF3085 );
			P_MD5( B, C, D, A,  6, 23, 0x04881D05 );
			P_MD5( A, B, C, D,  9,  4, 0xD9D4D039 );
			P_MD5( D, A, B, C, 12, 11, 0xE6DB99E5 );
			P_MD5( C, D, A, B, 15, 16, 0x1FA27CF8 );
			P_MD5( B, C, D, A,  2, 23, 0xC4AC5665 );

	#undef F

	#define F(x,y,z) (y ^ (x | ~z))

			P_MD5( A, B, C, D,  0,  6, 0xF4292244 );
			P_MD5( D, A, B, C,  7, 10, 0x432AFF97 );
			P_MD5( C, D, A, B, 14, 15, 0xAB9423A7 );
			P_MD5( B, C, D, A,  5, 21, 0xFC93A039 );
			P_MD5( A, B, C, D, 12,  6, 0x655B59C3 );
			P_MD5( D, A, B, C,  3, 10, 0x8F0CCC92 );
			P_MD5( C, D, A, B, 10, 15, 0xFFEFF47D );
			P_MD5( B, C, D, A,  1, 21, 0x85845DD1 );
			P_MD5( A, B, C, D,  8,  6, 0x6FA87E4F );
			P_MD5( D, A, B, C, 15, 10, 0xFE2CE6E0 );
			P_MD5( C, D, A, B,  6, 15, 0xA3014314 );
			P_MD5( B, C, D, A, 13, 21, 0x4E0811A1 );
			P_MD5( A, B, C, D,  4,  6, 0xF7537E82 );
			P_MD5( D, A, B, C, 11, 10, 0xBD3AF235 );
			P_MD5( C, D, A, B,  2, 15, 0x2AD7D2BB );
			P_MD5( B, C, D, A,  9, 21, 0xEB86D391 );

	#undef F

			ctx->state[0] += A;
			ctx->state[1] += B;
			ctx->state[2] += C;
			ctx->state[3] += D;
	}

	void md5_update( md5_context *ctx, uint8 *input, uint32 length )
	{
			uint32 left, fill;

			if( ! length ) return;

			left = ( ctx->total[0] >> 3 ) & 0x3F;
			fill = 64 - left;

			ctx->total[0] += length <<  3;
			ctx->total[1] += length >> 29;

			ctx->total[0] &= 0xFFFFFFFF;
			ctx->total[1] += ctx->total[0] < ( length << 3 );

			if( left && length >= fill )
			{
					memcpy( (void *) (ctx->buffer + left), (void *) input, fill );
					md5_process( ctx, ctx->buffer );
					length -= fill;
					input  += fill;
					left = 0;
			}

			while( length >= 64 )
			{
					md5_process( ctx, input );
					length -= 64;
					input  += 64;
			}

			if( length )
			{
					memcpy( (void *) (ctx->buffer + left), (void *) input, length );
			}
	}

	static uint8 md5_padding[64] =
	{
	 0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};

	void md5_finish( md5_context *ctx, uint8 digest[16] )
	{
			uint32 last, padn;
			uint8 msglen[8];
			int i;

			for (i=0;i<8;i++) msglen[i]=0;
			PUT_UINT32_MD5( ctx->total[0], msglen, 0 );
			PUT_UINT32_MD5( ctx->total[1], msglen, 4 );

			last = ( ctx->total[0] >> 3 ) & 0x3F;
			padn = ( last < 56 ) ? ( 56 - last ) : ( 120 - last );

			md5_update( ctx, md5_padding, padn );
			md5_update( ctx, msglen, 8 );

			PUT_UINT32_MD5( ctx->state[0], digest,  0 );
			PUT_UINT32_MD5( ctx->state[1], digest,  4 );
			PUT_UINT32_MD5( ctx->state[2], digest,  8 );
			PUT_UINT32_MD5( ctx->state[3], digest, 12 );
	}



	// those are the standard RFC 1321 test vectors

	static char *msg_md5[] =
	{
			"",
			"a",
			"abc",
			"message digest",
			"abcdefghijklmnopqrstuvwxyz",
			"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
			"12345678901234567890123456789012345678901234567890123456789012" \
					"345678901234567890"
	};

	static char *val_md5[] =
	{
			"d41d8cd98f00b204e9800998ecf8427e",
			"0cc175b9c0f1b6a831c399e269772661",
			"900150983cd24fb0d6963f7d28e17f72",
			"f96b697d7cb7938d525a2f31aaf161d0",
			"c3fcd3d76192e4007dfb496cca67e13b",
			"d174ab98d277d9f5a5611c2c9f419d9f",
			"57edf4a22be3c955ac49da2e2107b67a"
	};




		// tests with standard test vectors and returns true if everything is OK
	bool test_MD5(void) {
			bool result=true; // false=fail, true=pass
			int i, j;
			char output[36];
			md5_context ctx;
			unsigned char md5sum[16];
			
			
			for( i = 0; i < 7; i++ )
			{
				md5_starts( &ctx );
				md5_update( &ctx, (uint8 *) msg_md5[i], (int)strlen( msg_md5[i] ) );
				md5_finish( &ctx, md5sum );
				
				for( j = 0; j < 16; j++ )
				{
					sprintf_s(output + j * 2,3, "%02x", md5sum[j] );
				}
				
				if( ! memcmp( output, val_md5[i], 32 ) )
				{
					result=true;
				}
				else
				{
					result=false;
				}
			}
			
			return(result);
	}














//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
/*
 * The HAVAL hashing function
 *
 * Enhanced Public domain implementation by Paulo S.L.M. Barreto <pbarreto@nw.com.br>
 * Enhancements by C. B. Roellgen: Release version didn't compile with VC6: Problem fixed
 *
 *
 * -----------------------------------------------------------------------------
 *
 * Reference: Zheng, Y., Pieprzyk, J., Seberry, J.:
 * "HAVAL - a one-way hashing algorithm with variable length of output",
 * Advances in Cryptology (AusCrypt'92), LNCS 718 (1993), 83-104, Springer-Verlag.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



	void mhash_bzero(void *s, int n)
	{
		memset(s, '\0', n);
	}


	#define HAVAL_VERSION	1

	/* Platform-specific settings */

	/* #define HARDWARE_ROTATIONS */
	/*#pragma optimize( "", off ) */
	/*#pragma option -Od */


	/* Nonlinear F functions */

	/*#define F1(X6, X5, X4, X3, X2, X1, X0) \
		((X1) & (X4) ^ (X2) & (X5) ^ (X3) & (X6) ^ (X0) & (X1) ^ (X0))*/
	//#define F1(X6, X5, X4, X3, X2, X1, X0) \
	//	(((X1) & ((X4) ^ (X0))) ^ ((X2) & (X5)) ^ ((X3) & (X6)) ^ (X0))

	// compiler-friendly version using stack
	word32 F1(word32 X6,word32 X5,word32 X4,word32 X3,word32 X2,word32 X1,word32 X0) {
		word32 result;
		result=(((X1) & ((X4) ^ (X0))) ^ ((X2) & (X5)) ^ ((X3) & (X6)) ^ (X0));
		return result;
	}

	/*#define F2(X6, X5, X4, X3, X2, X1, X0) \
		((X1) & (X2) & (X3) ^ (X2) & (X4) & (X5) ^ \
		(X1) & (X2) ^ (X1) & (X4) ^ (X2) & (X6) ^ (X3) & (X5) ^ \
		(X4) & (X5) ^ (X0) & (X2) ^ (X0))*/
	//#define F2(X6, X5, X4, X3, X2, X1, X0) \
	//	(((X2) & (((X1) & (~(X3))) ^ ((X4) & (X5)) ^ (X6) ^ (X0))) ^ \
	//	(((X4) & ((X1) ^ (X5))) ^ ((X3) & (X5)) ^ (X0)))

	// compiler-friendly version using stack
	word32 F2(word32 X6,word32 X5,word32 X4,word32 X3,word32 X2,word32 X1,word32 X0) {
		word32 result;
		result=(((X2) & (((X1) & (~(X3))) ^ ((X4) & (X5)) ^ (X6) ^ (X0))) ^ (((X4) & ((X1) ^ (X5))) ^ ((X3) & (X5)) ^ (X0)));
		return result;
	}


	/*#define F3(X6, X5, X4, X3, X2, X1, X0) \
		((X1) & (X2) & (X3) ^ (X1) & (X4) ^ (X2) & (X5) ^ (X3) & (X6) ^ (X0) &
	(X3) ^ (X0))*/
	//#define F3(X6, X5, X4, X3, X2, X1, X0) \
	//	(((X3) & (((X1) & (X2)) ^ (X6) ^ (X0))) ^ ((X1) & (X4)) ^ \
	//	((X2) & (X5)) ^ (X0))

	// compiler-friendly version using stack
	word32 F3(word32 X6,word32 X5,word32 X4,word32 X3,word32 X2,word32 X1,word32 X0) {
		word32 result;
		result=	(((X3) & (((X1) & (X2)) ^ (X6) ^ (X0))) ^ ((X1) & (X4)) ^ ((X2) & (X5)) ^ (X0));
		return result;
	}



	/*#define F4(X6, X5, X4, X3, X2, X1, X0) \
		((X1) & (X2) & (X3) ^ (X2) & (X4) & (X5) ^ (X3) & (X4) & (X6) ^ \
		(X1) & (X4) ^ (X2) & (X6) ^ (X3) & (X4) ^ (X3) & (X5) ^ \
		(X3) & (X6) ^ (X4) & (X5) ^ (X4) & (X6) ^ (X0) & (X4) ^(X0))*/
	//#define F4(X6, X5, X4, X3, X2, X1, X0) \
	//	(((X4) & (((~(X2)) & (X5)) ^ ((X3) | (X6)) ^ (X1) ^ (X0))) ^ \
	//	((X3) & (((X1) & (X2)) ^ (X5) ^ (X6))) ^ ((X2) & (X6)) ^ (X0))

	// compiler-friendly version using stack
	word32 F4(word32 X6,word32 X5,word32 X4,word32 X3,word32 X2,word32 X1,word32 X0) {
		word32 result;
		result=	(((X4) & (((~(X2)) & (X5)) ^ ((X3) | (X6)) ^ (X1) ^ (X0))) ^ ((X3) & (((X1) & (X2)) ^ (X5) ^ (X6))) ^ ((X2) & (X6)) ^ (X0));
		return result;
	}


	/*#define F5(X6, X5, X4, X3, X2, X1, X0) \
		((X1) & (X4) ^ (X2) & (X5) ^ (X3) & (X6) ^ \
		(X0) & (X1) & (X2) & (X3) ^ (X0) & (X5) ^ (X0))*/
	//#define F5(X6, X5, X4, X3, X2, X1, X0) \
	//	(((X1) & ((X4) ^ ((X0) & (X2) & (X3)))) ^ \
	//	(((X2) ^ (X0)) & (X5)) ^ ((X3) & (X6)) ^ (X0))

	// compiler-friendly version using stack
	word32 F5(word32 X6,word32 X5,word32 X4,word32 X3,word32 X2,word32 X1,word32 X0) {
		word32 result;
		result=	(((X1) & ((X4) ^ ((X0) & (X2) & (X3)))) ^ (((X2) ^ (X0)) & (X5)) ^ ((X3) & (X6)) ^ (X0));
		return result;
	}


	#define ROTR(v,n)   (( (v) >> ((word32)(n)) ) | ( (v) << (32 - ((word32)(n)) )))
	/*
		// alternative version if stack is required
	word32 ROTR(word32 v, word32 n) {
	 word32 result;
	 result=( ( v >> n ) | ( v << (32 - n) ) );
	 return result;
	}
	*/


	#define BYTE_REVERSE(out, in, n) \
	{ \
		register int i; \
		for (i = 0; i < n; i++) { \
									((byte*)out)[4*i+3] = ((byte*)in)[4*i+0];\
									((byte*)out)[4*i+2] = ((byte*)in)[4*i+1];\
									((byte*)out)[4*i+1] = ((byte*)in)[4*i+2];\
									((byte*)out)[4*i+0] = ((byte*)in)[4*i+3];\
		} \
	}


	static void havalTransform3 (word32 E[8], const byte D[128], word32 T[8])
	{

	#ifndef WORDS_BIGENDIAN
		word32 *W = (word32 *)D;
	#else  /* !LITTLE_ENDIAN */
		word32 W[32];
		BYTE_REVERSE(W, D, 32);
	#endif /* ?LITTLE_ENDIAN */

		/* PASS 1: */

		T[7] = ROTR (F1 (E[1], E[0], E[3], E[5], E[6], E[2], E[4]), 7) + ROTR (E[7], 11) + W[ 0];
		T[6] = ROTR (F1 (E[0], T[7], E[2], E[4], E[5], E[1], E[3]), 7) + ROTR (E[6], 11) + W[ 1];
		T[5] = ROTR (F1 (T[7], T[6], E[1], E[3], E[4], E[0], E[2]), 7) + ROTR (E[5], 11) + W[ 2];
		T[4] = ROTR (F1 (T[6], T[5], E[0], E[2], E[3], T[7], E[1]), 7) + ROTR (E[4], 11) + W[ 3];
		T[3] = ROTR (F1 (T[5], T[4], T[7], E[1], E[2], T[6], E[0]), 7) + ROTR (E[3], 11) + W[ 4];
		T[2] = ROTR (F1 (T[4], T[3], T[6], E[0], E[1], T[5], T[7]), 7) + ROTR (E[2], 11) + W[ 5];
		T[1] = ROTR (F1 (T[3], T[2], T[5], T[7], E[0], T[4], T[6]), 7) + ROTR (E[1], 11) + W[ 6];
		T[0] = ROTR (F1 (T[2], T[1], T[4], T[6], T[7], T[3], T[5]), 7) + ROTR (E[0], 11) + W[ 7];


		T[7] = ROTR (F1 (T[1], T[0], T[3], T[5], T[6], T[2], T[4]), 7) + ROTR (T[7], 11) + W[ 8];
		T[6] = ROTR (F1 (T[0], T[7], T[2], T[4], T[5], T[1], T[3]), 7) + ROTR (T[6], 11) + W[ 9];
		T[5] = ROTR (F1 (T[7], T[6], T[1], T[3], T[4], T[0], T[2]), 7) + ROTR (T[5], 11) + W[10];
		T[4] = ROTR (F1 (T[6], T[5], T[0], T[2], T[3], T[7], T[1]), 7) + ROTR (T[4], 11) + W[11];
		T[3] = ROTR (F1 (T[5], T[4], T[7], T[1], T[2], T[6], T[0]), 7) + ROTR (T[3], 11) + W[12];
		T[2] = ROTR (F1 (T[4], T[3], T[6], T[0], T[1], T[5], T[7]), 7) + ROTR (T[2], 11) + W[13];
		T[1] = ROTR (F1 (T[3], T[2], T[5], T[7], T[0], T[4], T[6]), 7) + ROTR (T[1], 11) + W[14];
		T[0] = ROTR (F1 (T[2], T[1], T[4], T[6], T[7], T[3], T[5]), 7) + ROTR (T[0], 11) + W[15];

		T[7] = ROTR (F1 (T[1], T[0], T[3], T[5], T[6], T[2], T[4]), 7) + ROTR (T[7], 11) + W[16];
		T[6] = ROTR (F1 (T[0], T[7], T[2], T[4], T[5], T[1], T[3]), 7) + ROTR (T[6], 11) + W[17];
		T[5] = ROTR (F1 (T[7], T[6], T[1], T[3], T[4], T[0], T[2]), 7) + ROTR (T[5], 11) + W[18];
		T[4] = ROTR (F1 (T[6], T[5], T[0], T[2], T[3], T[7], T[1]), 7) + ROTR (T[4], 11) + W[19];
		T[3] = ROTR (F1 (T[5], T[4], T[7], T[1], T[2], T[6], T[0]), 7) + ROTR (T[3], 11) + W[20];
		T[2] = ROTR (F1 (T[4], T[3], T[6], T[0], T[1], T[5], T[7]), 7) + ROTR (T[2], 11) + W[21];
		T[1] = ROTR (F1 (T[3], T[2], T[5], T[7], T[0], T[4], T[6]), 7) + ROTR (T[1], 11) + W[22];
		T[0] = ROTR (F1 (T[2], T[1], T[4], T[6], T[7], T[3], T[5]), 7) + ROTR (T[0], 11) + W[23];

		T[7] = ROTR (F1 (T[1], T[0], T[3], T[5], T[6], T[2], T[4]), 7) + ROTR (T[7], 11) + W[24];
		T[6] = ROTR (F1 (T[0], T[7], T[2], T[4], T[5], T[1], T[3]), 7) + ROTR (T[6], 11) + W[25];
		T[5] = ROTR (F1 (T[7], T[6], T[1], T[3], T[4], T[0], T[2]), 7) + ROTR (T[5], 11) + W[26];
		T[4] = ROTR (F1 (T[6], T[5], T[0], T[2], T[3], T[7], T[1]), 7) + ROTR (T[4], 11) + W[27];
		T[3] = ROTR (F1 (T[5], T[4], T[7], T[1], T[2], T[6], T[0]), 7) + ROTR (T[3], 11) + W[28];
		T[2] = ROTR (F1 (T[4], T[3], T[6], T[0], T[1], T[5], T[7]), 7) + ROTR (T[2], 11) + W[29];
		T[1] = ROTR (F1 (T[3], T[2], T[5], T[7], T[0], T[4], T[6]), 7) + ROTR (T[1], 11) + W[30];
		T[0] = ROTR (F1 (T[2], T[1], T[4], T[6], T[7], T[3], T[5]), 7) + ROTR (T[0], 11) + W[31];

		/* PASS 2: */

		T[7] = ROTR (F2 (T[4], T[2], T[1], T[0], T[5], T[3], T[6]), 7) + ROTR (T[7], 11) + W[ 5] + 0x452821E6UL;
		T[6] = ROTR (F2 (T[3], T[1], T[0], T[7], T[4], T[2], T[5]), 7) + ROTR (T[6], 11) + W[14] + 0x38D01377UL;
		T[5] = ROTR (F2 (T[2], T[0], T[7], T[6], T[3], T[1], T[4]), 7) + ROTR (T[5], 11) + W[26] + 0xBE5466CFUL;
		T[4] = ROTR (F2 (T[1], T[7], T[6], T[5], T[2], T[0], T[3]), 7) + ROTR (T[4], 11) + W[18] + 0x34E90C6CUL;
		T[3] = ROTR (F2 (T[0], T[6], T[5], T[4], T[1], T[7], T[2]), 7) + ROTR (T[3], 11) + W[11] + 0xC0AC29B7UL;
		T[2] = ROTR (F2 (T[7], T[5], T[4], T[3], T[0], T[6], T[1]), 7) + ROTR (T[2], 11) + W[28] + 0xC97C50DDUL;
		T[1] = ROTR (F2 (T[6], T[4], T[3], T[2], T[7], T[5], T[0]), 7) + ROTR (T[1], 11) + W[ 7] + 0x3F84D5B5UL;
		T[0] = ROTR (F2 (T[5], T[3], T[2], T[1], T[6], T[4], T[7]), 7) + ROTR (T[0], 11) + W[16] + 0xB5470917UL;

		T[7] = ROTR (F2 (T[4], T[2], T[1], T[0], T[5], T[3], T[6]), 7) + ROTR (T[7], 11) + W[ 0] + 0x9216D5D9UL;
		T[6] = ROTR (F2 (T[3], T[1], T[0], T[7], T[4], T[2], T[5]), 7) + ROTR (T[6], 11) + W[23] + 0x8979FB1BUL;
		T[5] = ROTR (F2 (T[2], T[0], T[7], T[6], T[3], T[1], T[4]), 7) + ROTR (T[5], 11) + W[20] + 0xD1310BA6UL;
		T[4] = ROTR (F2 (T[1], T[7], T[6], T[5], T[2], T[0], T[3]), 7) + ROTR (T[4], 11) + W[22] + 0x98DFB5ACUL;
		T[3] = ROTR (F2 (T[0], T[6], T[5], T[4], T[1], T[7], T[2]), 7) + ROTR (T[3], 11) + W[ 1] + 0x2FFD72DBUL;
		T[2] = ROTR (F2 (T[7], T[5], T[4], T[3], T[0], T[6], T[1]), 7) + ROTR (T[2], 11) + W[10] + 0xD01ADFB7UL;
		T[1] = ROTR (F2 (T[6], T[4], T[3], T[2], T[7], T[5], T[0]), 7) + ROTR (T[1], 11) + W[ 4] + 0xB8E1AFEDUL;
		T[0] = ROTR (F2 (T[5], T[3], T[2], T[1], T[6], T[4], T[7]), 7) + ROTR (T[0], 11) + W[ 8] + 0x6A267E96UL;

		T[7] = ROTR (F2 (T[4], T[2], T[1], T[0], T[5], T[3], T[6]), 7) + ROTR (T[7], 11) + W[30] + 0xBA7C9045UL;
		T[6] = ROTR (F2 (T[3], T[1], T[0], T[7], T[4], T[2], T[5]), 7) + ROTR (T[6], 11) + W[ 3] + 0xF12C7F99UL;
		T[5] = ROTR (F2 (T[2], T[0], T[7], T[6], T[3], T[1], T[4]), 7) + ROTR (T[5], 11) + W[21] + 0x24A19947UL;
		T[4] = ROTR (F2 (T[1], T[7], T[6], T[5], T[2], T[0], T[3]), 7) + ROTR (T[4], 11) + W[ 9] + 0xB3916CF7UL;
		T[3] = ROTR (F2 (T[0], T[6], T[5], T[4], T[1], T[7], T[2]), 7) + ROTR (T[3], 11) + W[17] + 0x0801F2E2UL;
		T[2] = ROTR (F2 (T[7], T[5], T[4], T[3], T[0], T[6], T[1]), 7) + ROTR (T[2], 11) + W[24] + 0x858EFC16UL;
		T[1] = ROTR (F2 (T[6], T[4], T[3], T[2], T[7], T[5], T[0]), 7) + ROTR (T[1], 11) + W[29] + 0x636920D8UL;
		T[0] = ROTR (F2 (T[5], T[3], T[2], T[1], T[6], T[4], T[7]), 7) + ROTR (T[0], 11) + W[ 6] + 0x71574E69UL;

		T[7] = ROTR (F2 (T[4], T[2], T[1], T[0], T[5], T[3], T[6]), 7) + ROTR (T[7], 11) + W[19] + 0xA458FEA3UL;
		T[6] = ROTR (F2 (T[3], T[1], T[0], T[7], T[4], T[2], T[5]), 7) + ROTR (T[6], 11) + W[12] + 0xF4933D7EUL;
		T[5] = ROTR (F2 (T[2], T[0], T[7], T[6], T[3], T[1], T[4]), 7) + ROTR (T[5], 11) + W[15] + 0x0D95748FUL;
		T[4] = ROTR (F2 (T[1], T[7], T[6], T[5], T[2], T[0], T[3]), 7) + ROTR (T[4], 11) + W[13] + 0x728EB658UL;
		T[3] = ROTR (F2 (T[0], T[6], T[5], T[4], T[1], T[7], T[2]), 7) + ROTR (T[3], 11) + W[ 2] + 0x718BCD58UL;
		T[2] = ROTR (F2 (T[7], T[5], T[4], T[3], T[0], T[6], T[1]), 7) + ROTR (T[2], 11) + W[25] + 0x82154AEEUL;
		T[1] = ROTR (F2 (T[6], T[4], T[3], T[2], T[7], T[5], T[0]), 7) + ROTR (T[1], 11) + W[31] + 0x7B54A41DUL;
		T[0] = ROTR (F2 (T[5], T[3], T[2], T[1], T[6], T[4], T[7]), 7) + ROTR (T[0], 11) + W[27] + 0xC25A59B5UL;

		/* PASS 3: */

		T[7] = ROTR (F3 (T[6], T[1], T[2], T[3], T[4], T[5], T[0]), 7) + ROTR (T[7], 11) + W[19] + 0x9C30D539UL;
		T[6] = ROTR (F3 (T[5], T[0], T[1], T[2], T[3], T[4], T[7]), 7) + ROTR (T[6], 11) + W[ 9] + 0x2AF26013UL;
		T[5] = ROTR (F3 (T[4], T[7], T[0], T[1], T[2], T[3], T[6]), 7) + ROTR (T[5], 11) + W[ 4] + 0xC5D1B023UL;
		T[4] = ROTR (F3 (T[3], T[6], T[7], T[0], T[1], T[2], T[5]), 7) + ROTR (T[4], 11) + W[20] + 0x286085F0UL;
		T[3] = ROTR (F3 (T[2], T[5], T[6], T[7], T[0], T[1], T[4]), 7) + ROTR (T[3], 11) + W[28] + 0xCA417918UL;
		T[2] = ROTR (F3 (T[1], T[4], T[5], T[6], T[7], T[0], T[3]), 7) + ROTR (T[2], 11) + W[17] + 0xB8DB38EFUL;
		T[1] = ROTR (F3 (T[0], T[3], T[4], T[5], T[6], T[7], T[2]), 7) + ROTR (T[1], 11) + W[ 8] + 0x8E79DCB0UL;
		T[0] = ROTR (F3 (T[7], T[2], T[3], T[4], T[5], T[6], T[1]), 7) + ROTR (T[0], 11) + W[22] + 0x603A180EUL;

		T[7] = ROTR (F3 (T[6], T[1], T[2], T[3], T[4], T[5], T[0]), 7) + ROTR (T[7], 11) + W[29] + 0x6C9E0E8BUL;
		T[6] = ROTR (F3 (T[5], T[0], T[1], T[2], T[3], T[4], T[7]), 7) + ROTR (T[6], 11) + W[14] + 0xB01E8A3EUL;
		T[5] = ROTR (F3 (T[4], T[7], T[0], T[1], T[2], T[3], T[6]), 7) + ROTR (T[5], 11) + W[25] + 0xD71577C1UL;
		T[4] = ROTR (F3 (T[3], T[6], T[7], T[0], T[1], T[2], T[5]), 7) + ROTR (T[4], 11) + W[12] + 0xBD314B27UL;
		T[3] = ROTR (F3 (T[2], T[5], T[6], T[7], T[0], T[1], T[4]), 7) + ROTR (T[3], 11) + W[24] + 0x78AF2FDAUL;
		T[2] = ROTR (F3 (T[1], T[4], T[5], T[6], T[7], T[0], T[3]), 7) + ROTR (T[2], 11) + W[30] + 0x55605C60UL;
		T[1] = ROTR (F3 (T[0], T[3], T[4], T[5], T[6], T[7], T[2]), 7) + ROTR (T[1], 11) + W[16] + 0xE65525F3UL;
		T[0] = ROTR (F3 (T[7], T[2], T[3], T[4], T[5], T[6], T[1]), 7) + ROTR (T[0], 11) + W[26] + 0xAA55AB94UL;

		T[7] = ROTR (F3 (T[6], T[1], T[2], T[3], T[4], T[5], T[0]), 7) + ROTR (T[7], 11) + W[31] + 0x57489862UL;
		T[6] = ROTR (F3 (T[5], T[0], T[1], T[2], T[3], T[4], T[7]), 7) + ROTR (T[6], 11) + W[15] + 0x63E81440UL;
		T[5] = ROTR (F3 (T[4], T[7], T[0], T[1], T[2], T[3], T[6]), 7) + ROTR (T[5], 11) + W[ 7] + 0x55CA396AUL;
		T[4] = ROTR (F3 (T[3], T[6], T[7], T[0], T[1], T[2], T[5]), 7) + ROTR (T[4], 11) + W[ 3] + 0x2AAB10B6UL;
		T[3] = ROTR (F3 (T[2], T[5], T[6], T[7], T[0], T[1], T[4]), 7) + ROTR (T[3], 11) + W[ 1] + 0xB4CC5C34UL;
		T[2] = ROTR (F3 (T[1], T[4], T[5], T[6], T[7], T[0], T[3]), 7) + ROTR (T[2], 11) + W[ 0] + 0x1141E8CEUL;
		T[1] = ROTR (F3 (T[0], T[3], T[4], T[5], T[6], T[7], T[2]), 7) + ROTR (T[1], 11) + W[18] + 0xA15486AFUL;
		T[0] = ROTR (F3 (T[7], T[2], T[3], T[4], T[5], T[6], T[1]), 7) + ROTR (T[0], 11) + W[27] + 0x7C72E993UL;

		E[7] += T[7] = ROTR (F3 (T[6], T[1], T[2], T[3], T[4], T[5], T[0]), 7) + ROTR (T[7], 11) + W[13] + 0xB3EE1411UL;
		E[6] += T[6] = ROTR (F3 (T[5], T[0], T[1], T[2], T[3], T[4], T[7]), 7) + ROTR (T[6], 11) + W[ 6] + 0x636FBC2AUL;
		E[5] += T[5] = ROTR (F3 (T[4], T[7], T[0], T[1], T[2], T[3], T[6]), 7) + ROTR (T[5], 11) + W[21] + 0x2BA9C55DUL;
		E[4] += T[4] = ROTR (F3 (T[3], T[6], T[7], T[0], T[1], T[2], T[5]), 7) + ROTR (T[4], 11) + W[10] + 0x741831F6UL;
		E[3] += T[3] = ROTR (F3 (T[2], T[5], T[6], T[7], T[0], T[1], T[4]), 7) + ROTR (T[3], 11) + W[23] + 0xCE5C3E16UL;
		E[2] += T[2] = ROTR (F3 (T[1], T[4], T[5], T[6], T[7], T[0], T[3]), 7) + ROTR (T[2], 11) + W[11] + 0x9B87931EUL;
		E[1] += T[1] = ROTR (F3 (T[0], T[3], T[4], T[5], T[6], T[7], T[2]), 7) + ROTR (T[1], 11) + W[ 5] + 0xAFD6BA33UL;
		E[0] += T[0] = ROTR (F3 (T[7], T[2], T[3], T[4], T[5], T[6], T[1]), 7) + ROTR (T[0], 11) + W[ 2] + 0x6C24CF5CUL;

	} /* havalTransform3 */


	static void havalTransform4 (word32 E[8], const byte D[128], word32 T[8])
	{
	#ifndef WORDS_BIGENDIAN
		word32 *W = (word32 *)D;
	#else  /* !LITTLE_ENDIAN */
		word32 W[32];
		BYTE_REVERSE(W, D, 32);
	#endif /* ?LITTLE_ENDIAN */

		/* PASS 1: */

		T[7] = ROTR (F1 (E[2], E[6], E[1], E[4], E[5], E[3], E[0]), 7) + ROTR (E[7], 11) + W[ 0];
		T[6] = ROTR (F1 (E[1], E[5], E[0], E[3], E[4], E[2], T[7]), 7) + ROTR (E[6], 11) + W[ 1];
		T[5] = ROTR (F1 (E[0], E[4], T[7], E[2], E[3], E[1], T[6]), 7) + ROTR (E[5], 11) + W[ 2];
		T[4] = ROTR (F1 (T[7], E[3], T[6], E[1], E[2], E[0], T[5]), 7) + ROTR (E[4], 11) + W[ 3];
		T[3] = ROTR (F1 (T[6], E[2], T[5], E[0], E[1], T[7], T[4]), 7) + ROTR (E[3], 11) + W[ 4];
		T[2] = ROTR (F1 (T[5], E[1], T[4], T[7], E[0], T[6], T[3]), 7) + ROTR (E[2], 11) + W[ 5];
		T[1] = ROTR (F1 (T[4], E[0], T[3], T[6], T[7], T[5], T[2]), 7) + ROTR (E[1], 11) + W[ 6];
		T[0] = ROTR (F1 (T[3], T[7], T[2], T[5], T[6], T[4], T[1]), 7) + ROTR (E[0], 11) + W[ 7];

		T[7] = ROTR (F1 (T[2], T[6], T[1], T[4], T[5], T[3], T[0]), 7) + ROTR (T[7], 11) + W[ 8];
		T[6] = ROTR (F1 (T[1], T[5], T[0], T[3], T[4], T[2], T[7]), 7) + ROTR (T[6], 11) + W[ 9];
		T[5] = ROTR (F1 (T[0], T[4], T[7], T[2], T[3], T[1], T[6]), 7) + ROTR (T[5], 11) + W[10];
		T[4] = ROTR (F1 (T[7], T[3], T[6], T[1], T[2], T[0], T[5]), 7) + ROTR (T[4], 11) + W[11];
		T[3] = ROTR (F1 (T[6], T[2], T[5], T[0], T[1], T[7], T[4]), 7) + ROTR (T[3], 11) + W[12];
		T[2] = ROTR (F1 (T[5], T[1], T[4], T[7], T[0], T[6], T[3]), 7) + ROTR (T[2], 11) + W[13];
		T[1] = ROTR (F1 (T[4], T[0], T[3], T[6], T[7], T[5], T[2]), 7) + ROTR (T[1], 11) + W[14];
		T[0] = ROTR (F1 (T[3], T[7], T[2], T[5], T[6], T[4], T[1]), 7) + ROTR (T[0], 11) + W[15];

		T[7] = ROTR (F1 (T[2], T[6], T[1], T[4], T[5], T[3], T[0]), 7) + ROTR (T[7], 11) + W[16];
		T[6] = ROTR (F1 (T[1], T[5], T[0], T[3], T[4], T[2], T[7]), 7) + ROTR (T[6], 11) + W[17];
		T[5] = ROTR (F1 (T[0], T[4], T[7], T[2], T[3], T[1], T[6]), 7) + ROTR (T[5], 11) + W[18];
		T[4] = ROTR (F1 (T[7], T[3], T[6], T[1], T[2], T[0], T[5]), 7) + ROTR (T[4], 11) + W[19];
		T[3] = ROTR (F1 (T[6], T[2], T[5], T[0], T[1], T[7], T[4]), 7) + ROTR (T[3], 11) + W[20];
		T[2] = ROTR (F1 (T[5], T[1], T[4], T[7], T[0], T[6], T[3]), 7) + ROTR (T[2], 11) + W[21];
		T[1] = ROTR (F1 (T[4], T[0], T[3], T[6], T[7], T[5], T[2]), 7) + ROTR (T[1], 11) + W[22];
		T[0] = ROTR (F1 (T[3], T[7], T[2], T[5], T[6], T[4], T[1]), 7) + ROTR (T[0], 11) + W[23];

		T[7] = ROTR (F1 (T[2], T[6], T[1], T[4], T[5], T[3], T[0]), 7) + ROTR (T[7], 11) + W[24];
		T[6] = ROTR (F1 (T[1], T[5], T[0], T[3], T[4], T[2], T[7]), 7) + ROTR (T[6], 11) + W[25];
		T[5] = ROTR (F1 (T[0], T[4], T[7], T[2], T[3], T[1], T[6]), 7) + ROTR (T[5], 11) + W[26];
		T[4] = ROTR (F1 (T[7], T[3], T[6], T[1], T[2], T[0], T[5]), 7) + ROTR (T[4], 11) + W[27];
		T[3] = ROTR (F1 (T[6], T[2], T[5], T[0], T[1], T[7], T[4]), 7) + ROTR (T[3], 11) + W[28];
		T[2] = ROTR (F1 (T[5], T[1], T[4], T[7], T[0], T[6], T[3]), 7) + ROTR (T[2], 11) + W[29];
		T[1] = ROTR (F1 (T[4], T[0], T[3], T[6], T[7], T[5], T[2]), 7) + ROTR (T[1], 11) + W[30];
		T[0] = ROTR (F1 (T[3], T[7], T[2], T[5], T[6], T[4], T[1]), 7) + ROTR (T[0], 11) + W[31];

		/* PASS 2: */

		T[7] = ROTR (F2 (T[3], T[5], T[2], T[0], T[1], T[6], T[4]), 7) + ROTR (T[7], 11) + W[ 5] + 0x452821E6UL;
		T[6] = ROTR (F2 (T[2], T[4], T[1], T[7], T[0], T[5], T[3]), 7) + ROTR (T[6], 11) + W[14] + 0x38D01377UL;
		T[5] = ROTR (F2 (T[1], T[3], T[0], T[6], T[7], T[4], T[2]), 7) + ROTR (T[5], 11) + W[26] + 0xBE5466CFUL;
		T[4] = ROTR (F2 (T[0], T[2], T[7], T[5], T[6], T[3], T[1]), 7) + ROTR (T[4], 11) + W[18] + 0x34E90C6CUL;
		T[3] = ROTR (F2 (T[7], T[1], T[6], T[4], T[5], T[2], T[0]), 7) + ROTR (T[3], 11) + W[11] + 0xC0AC29B7UL;
		T[2] = ROTR (F2 (T[6], T[0], T[5], T[3], T[4], T[1], T[7]), 7) + ROTR (T[2], 11) + W[28] + 0xC97C50DDUL;
		T[1] = ROTR (F2 (T[5], T[7], T[4], T[2], T[3], T[0], T[6]), 7) + ROTR (T[1], 11) + W[ 7] + 0x3F84D5B5UL;
		T[0] = ROTR (F2 (T[4], T[6], T[3], T[1], T[2], T[7], T[5]), 7) + ROTR (T[0], 11) + W[16] + 0xB5470917UL;

		T[7] = ROTR (F2 (T[3], T[5], T[2], T[0], T[1], T[6], T[4]), 7) + ROTR (T[7], 11) + W[ 0] + 0x9216D5D9UL;
		T[6] = ROTR (F2 (T[2], T[4], T[1], T[7], T[0], T[5], T[3]), 7) + ROTR (T[6], 11) + W[23] + 0x8979FB1BUL;
		T[5] = ROTR (F2 (T[1], T[3], T[0], T[6], T[7], T[4], T[2]), 7) + ROTR (T[5], 11) + W[20] + 0xD1310BA6UL;
		T[4] = ROTR (F2 (T[0], T[2], T[7], T[5], T[6], T[3], T[1]), 7) + ROTR (T[4], 11) + W[22] + 0x98DFB5ACUL;
		T[3] = ROTR (F2 (T[7], T[1], T[6], T[4], T[5], T[2], T[0]), 7) + ROTR (T[3], 11) + W[ 1] + 0x2FFD72DBUL;
		T[2] = ROTR (F2 (T[6], T[0], T[5], T[3], T[4], T[1], T[7]), 7) + ROTR (T[2], 11) + W[10] + 0xD01ADFB7UL;
		T[1] = ROTR (F2 (T[5], T[7], T[4], T[2], T[3], T[0], T[6]), 7) + ROTR (T[1], 11) + W[ 4] + 0xB8E1AFEDUL;
		T[0] = ROTR (F2 (T[4], T[6], T[3], T[1], T[2], T[7], T[5]), 7) + ROTR (T[0], 11) + W[ 8] + 0x6A267E96UL;

		T[7] = ROTR (F2 (T[3], T[5], T[2], T[0], T[1], T[6], T[4]), 7) + ROTR (T[7], 11) + W[30] + 0xBA7C9045UL;
		T[6] = ROTR (F2 (T[2], T[4], T[1], T[7], T[0], T[5], T[3]), 7) + ROTR (T[6], 11) + W[ 3] + 0xF12C7F99UL;
		T[5] = ROTR (F2 (T[1], T[3], T[0], T[6], T[7], T[4], T[2]), 7) + ROTR (T[5], 11) + W[21] + 0x24A19947UL;
		T[4] = ROTR (F2 (T[0], T[2], T[7], T[5], T[6], T[3], T[1]), 7) + ROTR (T[4], 11) + W[ 9] + 0xB3916CF7UL;
		T[3] = ROTR (F2 (T[7], T[1], T[6], T[4], T[5], T[2], T[0]), 7) + ROTR (T[3], 11) + W[17] + 0x0801F2E2UL;
		T[2] = ROTR (F2 (T[6], T[0], T[5], T[3], T[4], T[1], T[7]), 7) + ROTR (T[2], 11) + W[24] + 0x858EFC16UL;
		T[1] = ROTR (F2 (T[5], T[7], T[4], T[2], T[3], T[0], T[6]), 7) + ROTR (T[1], 11) + W[29] + 0x636920D8UL;
		T[0] = ROTR (F2 (T[4], T[6], T[3], T[1], T[2], T[7], T[5]), 7) + ROTR (T[0], 11) + W[ 6] + 0x71574E69UL;

		T[7] = ROTR (F2 (T[3], T[5], T[2], T[0], T[1], T[6], T[4]), 7) + ROTR (T[7], 11) + W[19] + 0xA458FEA3UL;
		T[6] = ROTR (F2 (T[2], T[4], T[1], T[7], T[0], T[5], T[3]), 7) + ROTR (T[6], 11) + W[12] + 0xF4933D7EUL;
		T[5] = ROTR (F2 (T[1], T[3], T[0], T[6], T[7], T[4], T[2]), 7) + ROTR (T[5], 11) + W[15] + 0x0D95748FUL;
		T[4] = ROTR (F2 (T[0], T[2], T[7], T[5], T[6], T[3], T[1]), 7) + ROTR (T[4], 11) + W[13] + 0x728EB658UL;
		T[3] = ROTR (F2 (T[7], T[1], T[6], T[4], T[5], T[2], T[0]), 7) + ROTR (T[3], 11) + W[ 2] + 0x718BCD58UL;
		T[2] = ROTR (F2 (T[6], T[0], T[5], T[3], T[4], T[1], T[7]), 7) + ROTR (T[2], 11) + W[25] + 0x82154AEEUL;
		T[1] = ROTR (F2 (T[5], T[7], T[4], T[2], T[3], T[0], T[6]), 7) + ROTR (T[1], 11) + W[31] + 0x7B54A41DUL;
		T[0] = ROTR (F2 (T[4], T[6], T[3], T[1], T[2], T[7], T[5]), 7) + ROTR (T[0], 11) + W[27] + 0xC25A59B5UL;

		/* PASS 3: */

		T[7] = ROTR (F3 (T[1], T[4], T[3], T[6], T[0], T[2], T[5]), 7) + ROTR (T[7], 11) + W[19] + 0x9C30D539UL;
		T[6] = ROTR (F3 (T[0], T[3], T[2], T[5], T[7], T[1], T[4]), 7) + ROTR (T[6], 11) + W[ 9] + 0x2AF26013UL;
		T[5] = ROTR (F3 (T[7], T[2], T[1], T[4], T[6], T[0], T[3]), 7) + ROTR (T[5], 11) + W[ 4] + 0xC5D1B023UL;
		T[4] = ROTR (F3 (T[6], T[1], T[0], T[3], T[5], T[7], T[2]), 7) + ROTR (T[4], 11) + W[20] + 0x286085F0UL;
		T[3] = ROTR (F3 (T[5], T[0], T[7], T[2], T[4], T[6], T[1]), 7) + ROTR (T[3], 11) + W[28] + 0xCA417918UL;
		T[2] = ROTR (F3 (T[4], T[7], T[6], T[1], T[3], T[5], T[0]), 7) + ROTR (T[2], 11) + W[17] + 0xB8DB38EFUL;
		T[1] = ROTR (F3 (T[3], T[6], T[5], T[0], T[2], T[4], T[7]), 7) + ROTR (T[1], 11) + W[ 8] + 0x8E79DCB0UL;
		T[0] = ROTR (F3 (T[2], T[5], T[4], T[7], T[1], T[3], T[6]), 7) + ROTR (T[0], 11) + W[22] + 0x603A180EUL;

		T[7] = ROTR (F3 (T[1], T[4], T[3], T[6], T[0], T[2], T[5]), 7) + ROTR (T[7], 11) + W[29] + 0x6C9E0E8BUL;
		T[6] = ROTR (F3 (T[0], T[3], T[2], T[5], T[7], T[1], T[4]), 7) + ROTR (T[6], 11) + W[14] + 0xB01E8A3EUL;
		T[5] = ROTR (F3 (T[7], T[2], T[1], T[4], T[6], T[0], T[3]), 7) + ROTR (T[5], 11) + W[25] + 0xD71577C1UL;
		T[4] = ROTR (F3 (T[6], T[1], T[0], T[3], T[5], T[7], T[2]), 7) + ROTR (T[4], 11) + W[12] + 0xBD314B27UL;
		T[3] = ROTR (F3 (T[5], T[0], T[7], T[2], T[4], T[6], T[1]), 7) + ROTR (T[3], 11) + W[24] + 0x78AF2FDAUL;
		T[2] = ROTR (F3 (T[4], T[7], T[6], T[1], T[3], T[5], T[0]), 7) + ROTR (T[2], 11) + W[30] + 0x55605C60UL;
		T[1] = ROTR (F3 (T[3], T[6], T[5], T[0], T[2], T[4], T[7]), 7) + ROTR (T[1], 11) + W[16] + 0xE65525F3UL;
		T[0] = ROTR (F3 (T[2], T[5], T[4], T[7], T[1], T[3], T[6]), 7) + ROTR (T[0], 11) + W[26] + 0xAA55AB94UL;

		T[7] = ROTR (F3 (T[1], T[4], T[3], T[6], T[0], T[2], T[5]), 7) + ROTR (T[7], 11) + W[31] + 0x57489862UL;
		T[6] = ROTR (F3 (T[0], T[3], T[2], T[5], T[7], T[1], T[4]), 7) + ROTR (T[6], 11) + W[15] + 0x63E81440UL;
		T[5] = ROTR (F3 (T[7], T[2], T[1], T[4], T[6], T[0], T[3]), 7) + ROTR (T[5], 11) + W[ 7] + 0x55CA396AUL;
		T[4] = ROTR (F3 (T[6], T[1], T[0], T[3], T[5], T[7], T[2]), 7) + ROTR (T[4], 11) + W[ 3] + 0x2AAB10B6UL;
		T[3] = ROTR (F3 (T[5], T[0], T[7], T[2], T[4], T[6], T[1]), 7) + ROTR (T[3], 11) + W[ 1] + 0xB4CC5C34UL;
		T[2] = ROTR (F3 (T[4], T[7], T[6], T[1], T[3], T[5], T[0]), 7) + ROTR (T[2], 11) + W[ 0] + 0x1141E8CEUL;
		T[1] = ROTR (F3 (T[3], T[6], T[5], T[0], T[2], T[4], T[7]), 7) + ROTR (T[1], 11) + W[18] + 0xA15486AFUL;
		T[0] = ROTR (F3 (T[2], T[5], T[4], T[7], T[1], T[3], T[6]), 7) + ROTR (T[0], 11) + W[27] + 0x7C72E993UL;

		T[7] = ROTR (F3 (T[1], T[4], T[3], T[6], T[0], T[2], T[5]), 7) + ROTR (T[7], 11) + W[13] + 0xB3EE1411UL;
		T[6] = ROTR (F3 (T[0], T[3], T[2], T[5], T[7], T[1], T[4]), 7) + ROTR (T[6], 11) + W[ 6] + 0x636FBC2AUL;
		T[5] = ROTR (F3 (T[7], T[2], T[1], T[4], T[6], T[0], T[3]), 7) + ROTR (T[5], 11) + W[21] + 0x2BA9C55DUL;
		T[4] = ROTR (F3 (T[6], T[1], T[0], T[3], T[5], T[7], T[2]), 7) + ROTR (T[4], 11) + W[10] + 0x741831F6UL;
		T[3] = ROTR (F3 (T[5], T[0], T[7], T[2], T[4], T[6], T[1]), 7) + ROTR (T[3], 11) + W[23] + 0xCE5C3E16UL;
		T[2] = ROTR (F3 (T[4], T[7], T[6], T[1], T[3], T[5], T[0]), 7) + ROTR (T[2], 11) + W[11] + 0x9B87931EUL;
		T[1] = ROTR (F3 (T[3], T[6], T[5], T[0], T[2], T[4], T[7]), 7) + ROTR (T[1], 11) + W[ 5] + 0xAFD6BA33UL;
		T[0] = ROTR (F3 (T[2], T[5], T[4], T[7], T[1], T[3], T[6]), 7) + ROTR (T[0], 11) + W[ 2] + 0x6C24CF5CUL;

		/* PASS 4: */

		T[7] = ROTR (F4 (T[6], T[4], T[0], T[5], T[2], T[1], T[3]), 7) + ROTR (T[7], 11) + W[24] + 0x7A325381UL;
		T[6] = ROTR (F4 (T[5], T[3], T[7], T[4], T[1], T[0], T[2]), 7) + ROTR (T[6], 11) + W[ 4] + 0x28958677UL;
		T[5] = ROTR (F4 (T[4], T[2], T[6], T[3], T[0], T[7], T[1]), 7) + ROTR (T[5], 11) + W[ 0] + 0x3B8F4898UL;
		T[4] = ROTR (F4 (T[3], T[1], T[5], T[2], T[7], T[6], T[0]), 7) + ROTR (T[4], 11) + W[14] + 0x6B4BB9AFUL;
		T[3] = ROTR (F4 (T[2], T[0], T[4], T[1], T[6], T[5], T[7]), 7) + ROTR (T[3], 11) + W[ 2] + 0xC4BFE81BUL;
		T[2] = ROTR (F4 (T[1], T[7], T[3], T[0], T[5], T[4], T[6]), 7) + ROTR (T[2], 11) + W[ 7] + 0x66282193UL;
		T[1] = ROTR (F4 (T[0], T[6], T[2], T[7], T[4], T[3], T[5]), 7) + ROTR (T[1], 11) + W[28] + 0x61D809CCUL;
		T[0] = ROTR (F4 (T[7], T[5], T[1], T[6], T[3], T[2], T[4]), 7) + ROTR (T[0], 11) + W[23] + 0xFB21A991UL;

		T[7] = ROTR (F4 (T[6], T[4], T[0], T[5], T[2], T[1], T[3]), 7) + ROTR (T[7], 11) + W[26] + 0x487CAC60UL;
		T[6] = ROTR (F4 (T[5], T[3], T[7], T[4], T[1], T[0], T[2]), 7) + ROTR (T[6], 11) + W[ 6] + 0x5DEC8032UL;
		T[5] = ROTR (F4 (T[4], T[2], T[6], T[3], T[0], T[7], T[1]), 7) + ROTR (T[5], 11) + W[30] + 0xEF845D5DUL;
		T[4] = ROTR (F4 (T[3], T[1], T[5], T[2], T[7], T[6], T[0]), 7) + ROTR (T[4], 11) + W[20] + 0xE98575B1UL;
		T[3] = ROTR (F4 (T[2], T[0], T[4], T[1], T[6], T[5], T[7]), 7) + ROTR (T[3], 11) + W[18] + 0xDC262302UL;
		T[2] = ROTR (F4 (T[1], T[7], T[3], T[0], T[5], T[4], T[6]), 7) + ROTR (T[2], 11) + W[25] + 0xEB651B88UL;
		T[1] = ROTR (F4 (T[0], T[6], T[2], T[7], T[4], T[3], T[5]), 7) + ROTR (T[1], 11) + W[19] + 0x23893E81UL;
		T[0] = ROTR (F4 (T[7], T[5], T[1], T[6], T[3], T[2], T[4]), 7) + ROTR (T[0], 11) + W[ 3] + 0xD396ACC5UL;

		T[7] = ROTR (F4 (T[6], T[4], T[0], T[5], T[2], T[1], T[3]), 7) + ROTR (T[7], 11) + W[22] + 0x0F6D6FF3UL;
		T[6] = ROTR (F4 (T[5], T[3], T[7], T[4], T[1], T[0], T[2]), 7) + ROTR (T[6], 11) + W[11] + 0x83F44239UL;
		T[5] = ROTR (F4 (T[4], T[2], T[6], T[3], T[0], T[7], T[1]), 7) + ROTR (T[5], 11) + W[31] + 0x2E0B4482UL;
		T[4] = ROTR (F4 (T[3], T[1], T[5], T[2], T[7], T[6], T[0]), 7) + ROTR (T[4], 11) + W[21] + 0xA4842004UL;
		T[3] = ROTR (F4 (T[2], T[0], T[4], T[1], T[6], T[5], T[7]), 7) + ROTR (T[3], 11) + W[ 8] + 0x69C8F04AUL;
		T[2] = ROTR (F4 (T[1], T[7], T[3], T[0], T[5], T[4], T[6]), 7) + ROTR (T[2], 11) + W[27] + 0x9E1F9B5EUL;
		T[1] = ROTR (F4 (T[0], T[6], T[2], T[7], T[4], T[3], T[5]), 7) + ROTR (T[1], 11) + W[12] + 0x21C66842UL;
		T[0] = ROTR (F4 (T[7], T[5], T[1], T[6], T[3], T[2], T[4]), 7) + ROTR (T[0], 11) + W[ 9] + 0xF6E96C9AUL;

		E[7] += T[7] = ROTR (F4 (T[6], T[4], T[0], T[5], T[2], T[1], T[3]), 7) + ROTR (T[7], 11) + W[ 1] + 0x670C9C61UL;
		E[6] += T[6] = ROTR (F4 (T[5], T[3], T[7], T[4], T[1], T[0], T[2]), 7) + ROTR (T[6], 11) + W[29] + 0xABD388F0UL;
		E[5] += T[5] = ROTR (F4 (T[4], T[2], T[6], T[3], T[0], T[7], T[1]), 7) + ROTR (T[5], 11) + W[ 5] + 0x6A51A0D2UL;
		E[4] += T[4] = ROTR (F4 (T[3], T[1], T[5], T[2], T[7], T[6], T[0]), 7) + ROTR (T[4], 11) + W[15] + 0xD8542F68UL;
		E[3] += T[3] = ROTR (F4 (T[2], T[0], T[4], T[1], T[6], T[5], T[7]), 7) + ROTR (T[3], 11) + W[17] + 0x960FA728UL;
		E[2] += T[2] = ROTR (F4 (T[1], T[7], T[3], T[0], T[5], T[4], T[6]), 7) + ROTR (T[2], 11) + W[10] + 0xAB5133A3UL;
		E[1] += T[1] = ROTR (F4 (T[0], T[6], T[2], T[7], T[4], T[3], T[5]), 7) + ROTR (T[1], 11) + W[16] + 0x6EEF0B6CUL;
		E[0] += T[0] = ROTR (F4 (T[7], T[5], T[1], T[6], T[3], T[2], T[4]), 7) + ROTR (T[0], 11) + W[13] + 0x137A3BE4UL;

	} /* havalTransform4 */


	static void havalTransform5 (word32 E[8], const byte D[128], word32 T[8])
	{
	#ifndef WORDS_BIGENDIAN
		word32 *W = (word32 *)D;
	#else  /* !LITTLE_ENDIAN */
		word32 W[32];
		BYTE_REVERSE(W, D, 32);
	#endif /* ?LITTLE_ENDIAN */

		/* PASS 1: */

		T[7] = ROTR (F1 (E[3], E[4], E[1], E[0], E[5], E[2], E[6]), 7) + ROTR (E[7], 11) + W[ 0];
		T[6] = ROTR (F1 (E[2], E[3], E[0], T[7], E[4], E[1], E[5]), 7) + ROTR (E[6], 11) + W[ 1];
		T[5] = ROTR (F1 (E[1], E[2], T[7], T[6], E[3], E[0], E[4]), 7) + ROTR (E[5], 11) + W[ 2];
		T[4] = ROTR (F1 (E[0], E[1], T[6], T[5], E[2], T[7], E[3]), 7) + ROTR (E[4], 11) + W[ 3];
		T[3] = ROTR (F1 (T[7], E[0], T[5], T[4], E[1], T[6], E[2]), 7) + ROTR (E[3], 11) + W[ 4];
		T[2] = ROTR (F1 (T[6], T[7], T[4], T[3], E[0], T[5], E[1]), 7) + ROTR (E[2], 11) + W[ 5];
		T[1] = ROTR (F1 (T[5], T[6], T[3], T[2], T[7], T[4], E[0]), 7) + ROTR (E[1], 11) + W[ 6];
		T[0] = ROTR (F1 (T[4], T[5], T[2], T[1], T[6], T[3], T[7]), 7) + ROTR (E[0], 11) + W[ 7];

		T[7] = ROTR (F1 (T[3], T[4], T[1], T[0], T[5], T[2], T[6]), 7) + ROTR (T[7], 11) + W[ 8];
		T[6] = ROTR (F1 (T[2], T[3], T[0], T[7], T[4], T[1], T[5]), 7) + ROTR (T[6], 11) + W[ 9];
		T[5] = ROTR (F1 (T[1], T[2], T[7], T[6], T[3], T[0], T[4]), 7) + ROTR (T[5], 11) + W[10];
		T[4] = ROTR (F1 (T[0], T[1], T[6], T[5], T[2], T[7], T[3]), 7) + ROTR (T[4], 11) + W[11];
		T[3] = ROTR (F1 (T[7], T[0], T[5], T[4], T[1], T[6], T[2]), 7) + ROTR (T[3], 11) + W[12];
		T[2] = ROTR (F1 (T[6], T[7], T[4], T[3], T[0], T[5], T[1]), 7) + ROTR (T[2], 11) + W[13];
		T[1] = ROTR (F1 (T[5], T[6], T[3], T[2], T[7], T[4], T[0]), 7) + ROTR (T[1], 11) + W[14];
		T[0] = ROTR (F1 (T[4], T[5], T[2], T[1], T[6], T[3], T[7]), 7) + ROTR (T[0], 11) + W[15];

		T[7] = ROTR (F1 (T[3], T[4], T[1], T[0], T[5], T[2], T[6]), 7) + ROTR (T[7], 11) + W[16];
		T[6] = ROTR (F1 (T[2], T[3], T[0], T[7], T[4], T[1], T[5]), 7) + ROTR (T[6], 11) + W[17];
		T[5] = ROTR (F1 (T[1], T[2], T[7], T[6], T[3], T[0], T[4]), 7) + ROTR (T[5], 11) + W[18];
		T[4] = ROTR (F1 (T[0], T[1], T[6], T[5], T[2], T[7], T[3]), 7) + ROTR (T[4], 11) + W[19];
		T[3] = ROTR (F1 (T[7], T[0], T[5], T[4], T[1], T[6], T[2]), 7) + ROTR (T[3], 11) + W[20];
		T[2] = ROTR (F1 (T[6], T[7], T[4], T[3], T[0], T[5], T[1]), 7) + ROTR (T[2], 11) + W[21];
		T[1] = ROTR (F1 (T[5], T[6], T[3], T[2], T[7], T[4], T[0]), 7) + ROTR (T[1], 11) + W[22];
		T[0] = ROTR (F1 (T[4], T[5], T[2], T[1], T[6], T[3], T[7]), 7) + ROTR (T[0], 11) + W[23];

		T[7] = ROTR (F1 (T[3], T[4], T[1], T[0], T[5], T[2], T[6]), 7) + ROTR (T[7], 11) + W[24];
		T[6] = ROTR (F1 (T[2], T[3], T[0], T[7], T[4], T[1], T[5]), 7) + ROTR (T[6], 11) + W[25];
		T[5] = ROTR (F1 (T[1], T[2], T[7], T[6], T[3], T[0], T[4]), 7) + ROTR (T[5], 11) + W[26];
		T[4] = ROTR (F1 (T[0], T[1], T[6], T[5], T[2], T[7], T[3]), 7) + ROTR (T[4], 11) + W[27];
		T[3] = ROTR (F1 (T[7], T[0], T[5], T[4], T[1], T[6], T[2]), 7) + ROTR (T[3], 11) + W[28];
		T[2] = ROTR (F1 (T[6], T[7], T[4], T[3], T[0], T[5], T[1]), 7) + ROTR (T[2], 11) + W[29];
		T[1] = ROTR (F1 (T[5], T[6], T[3], T[2], T[7], T[4], T[0]), 7) + ROTR (T[1], 11) + W[30];
		T[0] = ROTR (F1 (T[4], T[5], T[2], T[1], T[6], T[3], T[7]), 7) + ROTR (T[0], 11) + W[31];

		/* PASS 2: */

		T[7] = ROTR (F2 (T[6], T[2], T[1], T[0], T[3], T[4], T[5]), 7) + ROTR (T[7], 11) + W[ 5] + 0x452821E6UL;
		T[6] = ROTR (F2 (T[5], T[1], T[0], T[7], T[2], T[3], T[4]), 7) + ROTR (T[6], 11) + W[14] + 0x38D01377UL;
		T[5] = ROTR (F2 (T[4], T[0], T[7], T[6], T[1], T[2], T[3]), 7) + ROTR (T[5], 11) + W[26] + 0xBE5466CFUL;
		T[4] = ROTR (F2 (T[3], T[7], T[6], T[5], T[0], T[1], T[2]), 7) + ROTR (T[4], 11) + W[18] + 0x34E90C6CUL;
		T[3] = ROTR (F2 (T[2], T[6], T[5], T[4], T[7], T[0], T[1]), 7) + ROTR (T[3], 11) + W[11] + 0xC0AC29B7UL;
		T[2] = ROTR (F2 (T[1], T[5], T[4], T[3], T[6], T[7], T[0]), 7) + ROTR (T[2], 11) + W[28] + 0xC97C50DDUL;
		T[1] = ROTR (F2 (T[0], T[4], T[3], T[2], T[5], T[6], T[7]), 7) + ROTR (T[1], 11) + W[ 7] + 0x3F84D5B5UL;
		T[0] = ROTR (F2 (T[7], T[3], T[2], T[1], T[4], T[5], T[6]), 7) + ROTR (T[0], 11) + W[16] + 0xB5470917UL;

		T[7] = ROTR (F2 (T[6], T[2], T[1], T[0], T[3], T[4], T[5]), 7) + ROTR (T[7], 11) + W[ 0] + 0x9216D5D9UL;
		T[6] = ROTR (F2 (T[5], T[1], T[0], T[7], T[2], T[3], T[4]), 7) + ROTR (T[6], 11) + W[23] + 0x8979FB1BUL;
		T[5] = ROTR (F2 (T[4], T[0], T[7], T[6], T[1], T[2], T[3]), 7) + ROTR (T[5], 11) + W[20] + 0xD1310BA6UL;
		T[4] = ROTR (F2 (T[3], T[7], T[6], T[5], T[0], T[1], T[2]), 7) + ROTR (T[4], 11) + W[22] + 0x98DFB5ACUL;
		T[3] = ROTR (F2 (T[2], T[6], T[5], T[4], T[7], T[0], T[1]), 7) + ROTR (T[3], 11) + W[ 1] + 0x2FFD72DBUL;
		T[2] = ROTR (F2 (T[1], T[5], T[4], T[3], T[6], T[7], T[0]), 7) + ROTR (T[2], 11) + W[10] + 0xD01ADFB7UL;
		T[1] = ROTR (F2 (T[0], T[4], T[3], T[2], T[5], T[6], T[7]), 7) + ROTR (T[1], 11) + W[ 4] + 0xB8E1AFEDUL;
		T[0] = ROTR (F2 (T[7], T[3], T[2], T[1], T[4], T[5], T[6]), 7) + ROTR (T[0], 11) + W[ 8] + 0x6A267E96UL;

		T[7] = ROTR (F2 (T[6], T[2], T[1], T[0], T[3], T[4], T[5]), 7) + ROTR (T[7], 11) + W[30] + 0xBA7C9045UL;
		T[6] = ROTR (F2 (T[5], T[1], T[0], T[7], T[2], T[3], T[4]), 7) + ROTR (T[6], 11) + W[ 3] + 0xF12C7F99UL;
		T[5] = ROTR (F2 (T[4], T[0], T[7], T[6], T[1], T[2], T[3]), 7) + ROTR (T[5], 11) + W[21] + 0x24A19947UL;
		T[4] = ROTR (F2 (T[3], T[7], T[6], T[5], T[0], T[1], T[2]), 7) + ROTR (T[4], 11) + W[ 9] + 0xB3916CF7UL;
		T[3] = ROTR (F2 (T[2], T[6], T[5], T[4], T[7], T[0], T[1]), 7) + ROTR (T[3], 11) + W[17] + 0x0801F2E2UL;
		T[2] = ROTR (F2 (T[1], T[5], T[4], T[3], T[6], T[7], T[0]), 7) + ROTR (T[2], 11) + W[24] + 0x858EFC16UL;
		T[1] = ROTR (F2 (T[0], T[4], T[3], T[2], T[5], T[6], T[7]), 7) + ROTR (T[1], 11) + W[29] + 0x636920D8UL;
		T[0] = ROTR (F2 (T[7], T[3], T[2], T[1], T[4], T[5], T[6]), 7) + ROTR (T[0], 11) + W[ 6] + 0x71574E69UL;

		T[7] = ROTR (F2 (T[6], T[2], T[1], T[0], T[3], T[4], T[5]), 7) + ROTR (T[7], 11) + W[19] + 0xA458FEA3UL;
		T[6] = ROTR (F2 (T[5], T[1], T[0], T[7], T[2], T[3], T[4]), 7) + ROTR (T[6], 11) + W[12] + 0xF4933D7EUL;
		T[5] = ROTR (F2 (T[4], T[0], T[7], T[6], T[1], T[2], T[3]), 7) + ROTR (T[5], 11) + W[15] + 0x0D95748FUL;
		T[4] = ROTR (F2 (T[3], T[7], T[6], T[5], T[0], T[1], T[2]), 7) + ROTR (T[4], 11) + W[13] + 0x728EB658UL;
		T[3] = ROTR (F2 (T[2], T[6], T[5], T[4], T[7], T[0], T[1]), 7) + ROTR (T[3], 11) + W[ 2] + 0x718BCD58UL;
		T[2] = ROTR (F2 (T[1], T[5], T[4], T[3], T[6], T[7], T[0]), 7) + ROTR (T[2], 11) + W[25] + 0x82154AEEUL;
		T[1] = ROTR (F2 (T[0], T[4], T[3], T[2], T[5], T[6], T[7]), 7) + ROTR (T[1], 11) + W[31] + 0x7B54A41DUL;
		T[0] = ROTR (F2 (T[7], T[3], T[2], T[1], T[4], T[5], T[6]), 7) + ROTR (T[0], 11) + W[27] + 0xC25A59B5UL;

		/* PASS 3: */

		T[7] = ROTR (F3 (T[2], T[6], T[0], T[4], T[3], T[1], T[5]), 7) + ROTR (T[7], 11) + W[19] + 0x9C30D539UL;
		T[6] = ROTR (F3 (T[1], T[5], T[7], T[3], T[2], T[0], T[4]), 7) + ROTR (T[6], 11) + W[ 9] + 0x2AF26013UL;
		T[5] = ROTR (F3 (T[0], T[4], T[6], T[2], T[1], T[7], T[3]), 7) + ROTR (T[5], 11) + W[ 4] + 0xC5D1B023UL;
		T[4] = ROTR (F3 (T[7], T[3], T[5], T[1], T[0], T[6], T[2]), 7) + ROTR (T[4], 11) + W[20] + 0x286085F0UL;
		T[3] = ROTR (F3 (T[6], T[2], T[4], T[0], T[7], T[5], T[1]), 7) + ROTR (T[3], 11) + W[28] + 0xCA417918UL;
		T[2] = ROTR (F3 (T[5], T[1], T[3], T[7], T[6], T[4], T[0]), 7) + ROTR (T[2], 11) + W[17] + 0xB8DB38EFUL;
		T[1] = ROTR (F3 (T[4], T[0], T[2], T[6], T[5], T[3], T[7]), 7) + ROTR (T[1], 11) + W[ 8] + 0x8E79DCB0UL;
		T[0] = ROTR (F3 (T[3], T[7], T[1], T[5], T[4], T[2], T[6]), 7) + ROTR (T[0], 11) + W[22] + 0x603A180EUL;

		T[7] = ROTR (F3 (T[2], T[6], T[0], T[4], T[3], T[1], T[5]), 7) + ROTR (T[7], 11) + W[29] + 0x6C9E0E8BUL;
		T[6] = ROTR (F3 (T[1], T[5], T[7], T[3], T[2], T[0], T[4]), 7) + ROTR (T[6], 11) + W[14] + 0xB01E8A3EUL;
		T[5] = ROTR (F3 (T[0], T[4], T[6], T[2], T[1], T[7], T[3]), 7) + ROTR (T[5], 11) + W[25] + 0xD71577C1UL;
		T[4] = ROTR (F3 (T[7], T[3], T[5], T[1], T[0], T[6], T[2]), 7) + ROTR (T[4], 11) + W[12] + 0xBD314B27UL;
		T[3] = ROTR (F3 (T[6], T[2], T[4], T[0], T[7], T[5], T[1]), 7) + ROTR (T[3], 11) + W[24] + 0x78AF2FDAUL;
		T[2] = ROTR (F3 (T[5], T[1], T[3], T[7], T[6], T[4], T[0]), 7) + ROTR (T[2], 11) + W[30] + 0x55605C60UL;
		T[1] = ROTR (F3 (T[4], T[0], T[2], T[6], T[5], T[3], T[7]), 7) + ROTR (T[1], 11) + W[16] + 0xE65525F3UL;
		T[0] = ROTR (F3 (T[3], T[7], T[1], T[5], T[4], T[2], T[6]), 7) + ROTR (T[0], 11) + W[26] + 0xAA55AB94UL;

		T[7] = ROTR (F3 (T[2], T[6], T[0], T[4], T[3], T[1], T[5]), 7) + ROTR (T[7], 11) + W[31] + 0x57489862UL;
		T[6] = ROTR (F3 (T[1], T[5], T[7], T[3], T[2], T[0], T[4]), 7) + ROTR (T[6], 11) + W[15] + 0x63E81440UL;
		T[5] = ROTR (F3 (T[0], T[4], T[6], T[2], T[1], T[7], T[3]), 7) + ROTR (T[5], 11) + W[ 7] + 0x55CA396AUL;
		T[4] = ROTR (F3 (T[7], T[3], T[5], T[1], T[0], T[6], T[2]), 7) + ROTR (T[4], 11) + W[ 3] + 0x2AAB10B6UL;
		T[3] = ROTR (F3 (T[6], T[2], T[4], T[0], T[7], T[5], T[1]), 7) + ROTR (T[3], 11) + W[ 1] + 0xB4CC5C34UL;
		T[2] = ROTR (F3 (T[5], T[1], T[3], T[7], T[6], T[4], T[0]), 7) + ROTR (T[2], 11) + W[ 0] + 0x1141E8CEUL;
		T[1] = ROTR (F3 (T[4], T[0], T[2], T[6], T[5], T[3], T[7]), 7) + ROTR (T[1], 11) + W[18] + 0xA15486AFUL;
		T[0] = ROTR (F3 (T[3], T[7], T[1], T[5], T[4], T[2], T[6]), 7) + ROTR (T[0], 11) + W[27] + 0x7C72E993UL;

		T[7] = ROTR (F3 (T[2], T[6], T[0], T[4], T[3], T[1], T[5]), 7) + ROTR (T[7], 11) + W[13] + 0xB3EE1411UL;
		T[6] = ROTR (F3 (T[1], T[5], T[7], T[3], T[2], T[0], T[4]), 7) + ROTR (T[6], 11) + W[ 6] + 0x636FBC2AUL;
		T[5] = ROTR (F3 (T[0], T[4], T[6], T[2], T[1], T[7], T[3]), 7) + ROTR (T[5], 11) + W[21] + 0x2BA9C55DUL;
		T[4] = ROTR (F3 (T[7], T[3], T[5], T[1], T[0], T[6], T[2]), 7) + ROTR (T[4], 11) + W[10] + 0x741831F6UL;
		T[3] = ROTR (F3 (T[6], T[2], T[4], T[0], T[7], T[5], T[1]), 7) + ROTR (T[3], 11) + W[23] + 0xCE5C3E16UL;
		T[2] = ROTR (F3 (T[5], T[1], T[3], T[7], T[6], T[4], T[0]), 7) + ROTR (T[2], 11) + W[11] + 0x9B87931EUL;
		T[1] = ROTR (F3 (T[4], T[0], T[2], T[6], T[5], T[3], T[7]), 7) + ROTR (T[1], 11) + W[ 5] + 0xAFD6BA33UL;
		T[0] = ROTR (F3 (T[3], T[7], T[1], T[5], T[4], T[2], T[6]), 7) + ROTR (T[0], 11) + W[ 2] + 0x6C24CF5CUL;

		/* PASS 4: */

		T[7] = ROTR (F4 (T[1], T[5], T[3], T[2], T[0], T[4], T[6]), 7) + ROTR (T[7], 11) + W[24] + 0x7A325381UL;
		T[6] = ROTR (F4 (T[0], T[4], T[2], T[1], T[7], T[3], T[5]), 7) + ROTR (T[6], 11) + W[ 4] + 0x28958677UL;
		T[5] = ROTR (F4 (T[7], T[3], T[1], T[0], T[6], T[2], T[4]), 7) + ROTR (T[5], 11) + W[ 0] + 0x3B8F4898UL;
		T[4] = ROTR (F4 (T[6], T[2], T[0], T[7], T[5], T[1], T[3]), 7) + ROTR (T[4], 11) + W[14] + 0x6B4BB9AFUL;
		T[3] = ROTR (F4 (T[5], T[1], T[7], T[6], T[4], T[0], T[2]), 7) + ROTR (T[3], 11) + W[ 2] + 0xC4BFE81BUL;
		T[2] = ROTR (F4 (T[4], T[0], T[6], T[5], T[3], T[7], T[1]), 7) + ROTR (T[2], 11) + W[ 7] + 0x66282193UL;
		T[1] = ROTR (F4 (T[3], T[7], T[5], T[4], T[2], T[6], T[0]), 7) + ROTR (T[1], 11) + W[28] + 0x61D809CCUL;
		T[0] = ROTR (F4 (T[2], T[6], T[4], T[3], T[1], T[5], T[7]), 7) + ROTR (T[0], 11) + W[23] + 0xFB21A991UL;

		T[7] = ROTR (F4 (T[1], T[5], T[3], T[2], T[0], T[4], T[6]), 7) + ROTR (T[7], 11) + W[26] + 0x487CAC60UL;
		T[6] = ROTR (F4 (T[0], T[4], T[2], T[1], T[7], T[3], T[5]), 7) + ROTR (T[6], 11) + W[ 6] + 0x5DEC8032UL;
		T[5] = ROTR (F4 (T[7], T[3], T[1], T[0], T[6], T[2], T[4]), 7) + ROTR (T[5], 11) + W[30] + 0xEF845D5DUL;
		T[4] = ROTR (F4 (T[6], T[2], T[0], T[7], T[5], T[1], T[3]), 7) + ROTR (T[4], 11) + W[20] + 0xE98575B1UL;
		T[3] = ROTR (F4 (T[5], T[1], T[7], T[6], T[4], T[0], T[2]), 7) + ROTR (T[3], 11) + W[18] + 0xDC262302UL;
		T[2] = ROTR (F4 (T[4], T[0], T[6], T[5], T[3], T[7], T[1]), 7) + ROTR (T[2], 11) + W[25] + 0xEB651B88UL;
		T[1] = ROTR (F4 (T[3], T[7], T[5], T[4], T[2], T[6], T[0]), 7) + ROTR (T[1], 11) + W[19] + 0x23893E81UL;
		T[0] = ROTR (F4 (T[2], T[6], T[4], T[3], T[1], T[5], T[7]), 7) + ROTR (T[0], 11) + W[ 3] + 0xD396ACC5UL;

		T[7] = ROTR (F4 (T[1], T[5], T[3], T[2], T[0], T[4], T[6]), 7) + ROTR (T[7], 11) + W[22] + 0x0F6D6FF3UL;
		T[6] = ROTR (F4 (T[0], T[4], T[2], T[1], T[7], T[3], T[5]), 7) + ROTR (T[6], 11) + W[11] + 0x83F44239UL;
		T[5] = ROTR (F4 (T[7], T[3], T[1], T[0], T[6], T[2], T[4]), 7) + ROTR (T[5], 11) + W[31] + 0x2E0B4482UL;
		T[4] = ROTR (F4 (T[6], T[2], T[0], T[7], T[5], T[1], T[3]), 7) + ROTR (T[4], 11) + W[21] + 0xA4842004UL;
		T[3] = ROTR (F4 (T[5], T[1], T[7], T[6], T[4], T[0], T[2]), 7) + ROTR (T[3], 11) + W[ 8] + 0x69C8F04AUL;
		T[2] = ROTR (F4 (T[4], T[0], T[6], T[5], T[3], T[7], T[1]), 7) + ROTR (T[2], 11) + W[27] + 0x9E1F9B5EUL;
		T[1] = ROTR (F4 (T[3], T[7], T[5], T[4], T[2], T[6], T[0]), 7) + ROTR (T[1], 11) + W[12] + 0x21C66842UL;
		T[0] = ROTR (F4 (T[2], T[6], T[4], T[3], T[1], T[5], T[7]), 7) + ROTR (T[0], 11) + W[ 9] + 0xF6E96C9AUL;

		T[7] = ROTR (F4 (T[1], T[5], T[3], T[2], T[0], T[4], T[6]), 7) + ROTR (T[7], 11) + W[ 1] + 0x670C9C61UL;
		T[6] = ROTR (F4 (T[0], T[4], T[2], T[1], T[7], T[3], T[5]), 7) + ROTR (T[6], 11) + W[29] + 0xABD388F0UL;
		T[5] = ROTR (F4 (T[7], T[3], T[1], T[0], T[6], T[2], T[4]), 7) + ROTR (T[5], 11) + W[ 5] + 0x6A51A0D2UL;
		T[4] = ROTR (F4 (T[6], T[2], T[0], T[7], T[5], T[1], T[3]), 7) + ROTR (T[4], 11) + W[15] + 0xD8542F68UL;
		T[3] = ROTR (F4 (T[5], T[1], T[7], T[6], T[4], T[0], T[2]), 7) + ROTR (T[3], 11) + W[17] + 0x960FA728UL;
		T[2] = ROTR (F4 (T[4], T[0], T[6], T[5], T[3], T[7], T[1]), 7) + ROTR (T[2], 11) + W[10] + 0xAB5133A3UL;
		T[1] = ROTR (F4 (T[3], T[7], T[5], T[4], T[2], T[6], T[0]), 7) + ROTR (T[1], 11) + W[16] + 0x6EEF0B6CUL;
		T[0] = ROTR (F4 (T[2], T[6], T[4], T[3], T[1], T[5], T[7]), 7) + ROTR (T[0], 11) + W[13] + 0x137A3BE4UL;

		/* PASS 5: */

		T[7] = ROTR (F5 (T[2], T[5], T[0], T[6], T[4], T[3], T[1]), 7) + ROTR (T[7], 11) + W[27] + 0xBA3BF050UL;
		T[6] = ROTR (F5 (T[1], T[4], T[7], T[5], T[3], T[2], T[0]), 7) + ROTR (T[6], 11) + W[ 3] + 0x7EFB2A98UL;
		T[5] = ROTR (F5 (T[0], T[3], T[6], T[4], T[2], T[1], T[7]), 7) + ROTR (T[5], 11) + W[21] + 0xA1F1651DUL;
		T[4] = ROTR (F5 (T[7], T[2], T[5], T[3], T[1], T[0], T[6]), 7) + ROTR (T[4], 11) + W[26] + 0x39AF0176UL;
		T[3] = ROTR (F5 (T[6], T[1], T[4], T[2], T[0], T[7], T[5]), 7) + ROTR (T[3], 11) + W[17] + 0x66CA593EUL;
		T[2] = ROTR (F5 (T[5], T[0], T[3], T[1], T[7], T[6], T[4]), 7) + ROTR (T[2], 11) + W[11] + 0x82430E88UL;
		T[1] = ROTR (F5 (T[4], T[7], T[2], T[0], T[6], T[5], T[3]), 7) + ROTR (T[1], 11) + W[20] + 0x8CEE8619UL;
		T[0] = ROTR (F5 (T[3], T[6], T[1], T[7], T[5], T[4], T[2]), 7) + ROTR (T[0], 11) + W[29] + 0x456F9FB4UL;

		T[7] = ROTR (F5 (T[2], T[5], T[0], T[6], T[4], T[3], T[1]), 7) + ROTR (T[7], 11) + W[19] + 0x7D84A5C3UL;
		T[6] = ROTR (F5 (T[1], T[4], T[7], T[5], T[3], T[2], T[0]), 7) + ROTR (T[6], 11) + W[ 0] + 0x3B8B5EBEUL;
		T[5] = ROTR (F5 (T[0], T[3], T[6], T[4], T[2], T[1], T[7]), 7) + ROTR (T[5], 11) + W[12] + 0xE06F75D8UL;
		T[4] = ROTR (F5 (T[7], T[2], T[5], T[3], T[1], T[0], T[6]), 7) + ROTR (T[4], 11) + W[ 7] + 0x85C12073UL;
		T[3] = ROTR (F5 (T[6], T[1], T[4], T[2], T[0], T[7], T[5]), 7) + ROTR (T[3], 11) + W[13] + 0x401A449FUL;
		T[2] = ROTR (F5 (T[5], T[0], T[3], T[1], T[7], T[6], T[4]), 7) + ROTR (T[2], 11) + W[ 8] + 0x56C16AA6UL;
		T[1] = ROTR (F5 (T[4], T[7], T[2], T[0], T[6], T[5], T[3]), 7) + ROTR (T[1], 11) + W[31] + 0x4ED3AA62UL;
		T[0] = ROTR (F5 (T[3], T[6], T[1], T[7], T[5], T[4], T[2]), 7) + ROTR (T[0], 11) + W[10] + 0x363F7706UL;

		T[7] = ROTR (F5 (T[2], T[5], T[0], T[6], T[4], T[3], T[1]), 7) + ROTR (T[7], 11) + W[ 5] + 0x1BFEDF72UL;
		T[6] = ROTR (F5 (T[1], T[4], T[7], T[5], T[3], T[2], T[0]), 7) + ROTR (T[6], 11) + W[ 9] + 0x429B023DUL;
		T[5] = ROTR (F5 (T[0], T[3], T[6], T[4], T[2], T[1], T[7]), 7) + ROTR (T[5], 11) + W[14] + 0x37D0D724UL;
		T[4] = ROTR (F5 (T[7], T[2], T[5], T[3], T[1], T[0], T[6]), 7) + ROTR (T[4], 11) + W[30] + 0xD00A1248UL;
		T[3] = ROTR (F5 (T[6], T[1], T[4], T[2], T[0], T[7], T[5]), 7) + ROTR (T[3], 11) + W[18] + 0xDB0FEAD3UL;
		T[2] = ROTR (F5 (T[5], T[0], T[3], T[1], T[7], T[6], T[4]), 7) + ROTR (T[2], 11) + W[ 6] + 0x49F1C09BUL;
		T[1] = ROTR (F5 (T[4], T[7], T[2], T[0], T[6], T[5], T[3]), 7) + ROTR (T[1], 11) + W[28] + 0x075372C9UL;
		T[0] = ROTR (F5 (T[3], T[6], T[1], T[7], T[5], T[4], T[2]), 7) + ROTR (T[0], 11) + W[24] + 0x80991B7BUL;

		E[7] += T[7] = ROTR (F5 (T[2], T[5], T[0], T[6], T[4], T[3], T[1]), 7) + ROTR (T[7], 11) + W[ 2] + 0x25D479D8UL;
		E[6] += T[6] = ROTR (F5 (T[1], T[4], T[7], T[5], T[3], T[2], T[0]), 7) + ROTR (T[6], 11) + W[23] + 0xF6E8DEF7UL;
		E[5] += T[5] = ROTR (F5 (T[0], T[3], T[6], T[4], T[2], T[1], T[7]), 7) + ROTR (T[5], 11) + W[16] + 0xE3FE501AUL;
		E[4] += T[4] = ROTR (F5 (T[7], T[2], T[5], T[3], T[1], T[0], T[6]), 7) + ROTR (T[4], 11) + W[22] + 0xB6794C3BUL;
		E[3] += T[3] = ROTR (F5 (T[6], T[1], T[4], T[2], T[0], T[7], T[5]), 7) + ROTR (T[3], 11) + W[ 4] + 0x976CE0BDUL;
		E[2] += T[2] = ROTR (F5 (T[5], T[0], T[3], T[1], T[7], T[6], T[4]), 7) + ROTR (T[2], 11) + W[ 1] + 0x04C006BAUL;
		E[1] += T[1] = ROTR (F5 (T[4], T[7], T[2], T[0], T[6], T[5], T[3]), 7) + ROTR (T[1], 11) + W[25] + 0xC1A94FB6UL;
		E[0] += T[0] = ROTR (F5 (T[3], T[6], T[1], T[7], T[5], T[4], T[2]), 7) + ROTR (T[0], 11) + W[15] + 0x409F60C4UL;

	} /* havalTransform5 */

	int havalInit256( havalContext *hcp) {
		return havalInit( hcp, 3, 256);
	}
	int havalInit224( havalContext *hcp) {
		return havalInit( hcp, 3, 224);
	}
	int havalInit192( havalContext *hcp) {
		return havalInit( hcp, 3, 192);
	}
	int havalInit160( havalContext *hcp) {
		return havalInit( hcp, 3, 160);
	}
	int havalInit128( havalContext *hcp) {
		return havalInit( hcp, 3, 128);
	}

	int havalInit (havalContext *hcp, int passes, int hashLength)
	{
		if (hcp == NULL) {
			return 1; /* bad context */
		}
		/* check number of passes: */
		if (passes != 3 && passes != 4 && passes != 5) {
			return 2; /* invalid number of passes */
		}
		/* check hash length: */
		if (hashLength != 128 &&
			hashLength != 160 &&
			hashLength != 192 &&
			hashLength != 224 &&
			hashLength != 256) {
			return 3; /* invalid hash length */
		}
		/* properly initialize HAVAL context: */
		mhash_bzero(hcp, sizeof (havalContext));
		hcp->passes = (word16) passes;
		hcp->hashLength = (word16) hashLength;
			hcp->digest[0] = 0x243F6A88UL;
			hcp->digest[1] = 0x85A308D3UL;
			hcp->digest[2] = 0x13198A2EUL;
			hcp->digest[3] = 0x03707344UL;
			hcp->digest[4] = 0xA4093822UL;
			hcp->digest[5] = 0x299F31D0UL;
			hcp->digest[6] = 0x082EFA98UL;
			hcp->digest[7] = 0xEC4E6C89UL;
		return 0; /* OK */
	} /* havalInit */


	int havalUpdate (havalContext *hcp, const byte *dataBuffer, size_t dataLength)
	{
		if (hcp == NULL) {
			return 1; /* bad context */
		}
		if (dataBuffer == NULL || dataLength == 0) {
			return 0; /* nothing to do */
		}

	/*	assert (hcp->occupied < 128);  invariant */

		/* update bit count: Extra parentheses for Borland C++ --Tines*/
		if ( ((word32)dataLength << 3) > (0xFFFFFFFFUL - hcp->bitCount[0]) ) {
			hcp->bitCount[1]++;
		}
		hcp->bitCount[0] += (word32)dataLength << 3;
		
		/* if the data buffer is not enough to complete */
		/* the context data block, just append it: */
		if (hcp->occupied + (word32)dataLength < 128) { /* caveat: typecast avoids 16-bit overflow */
			memcpy (&hcp->block[hcp->occupied], dataBuffer, dataLength);
			hcp->occupied += dataLength;
	/*		assert (hcp->occupied < 128); */
			return 0; /* delay processing */
		}

		/* complete the context data block: */
		memcpy (&hcp->block[hcp->occupied], dataBuffer, 128 - hcp->occupied);
		dataBuffer += 128 - hcp->occupied;
		dataLength -= 128 - hcp->occupied;

		switch (hcp->passes) {
		case 3:
			/* process the completed context data block: */
			havalTransform3 (hcp->digest, hcp->block, hcp->temp);
			/* process data in chunks of 128 bytes: */
			while (dataLength >= 128) {
				havalTransform3 (hcp->digest, dataBuffer, hcp->temp);
				dataBuffer += 128;
				dataLength -= 128;
			}
			break;
		case 4:
			/* process the completed context data block: */
			havalTransform4 (hcp->digest, hcp->block, hcp->temp);
			/* process data in chunks of 128 bytes: */
			while (dataLength >= 128) {
				havalTransform4 (hcp->digest, dataBuffer, hcp->temp);
				dataBuffer += 128;
				dataLength -= 128;
			}
			break;
		case 5:
			/* process the completed context data block: */
			havalTransform5 (hcp->digest, hcp->block, hcp->temp);
			/* process data in chunks of 128 bytes: */
			while (dataLength >= 128) {
				havalTransform5 (hcp->digest, dataBuffer, hcp->temp);
				dataBuffer += 128;
				dataLength -= 128;
			}
			break;
		}

		/* delay processing of remaining data: */
		memcpy (hcp->block, dataBuffer, dataLength);
		hcp->occupied = dataLength; /* < 128 */
		
	/*	assert (hcp->occupied < 128); */
		return 0; /* OK */
	} /* havalUpdate */


	int havalFinal (havalContext *hcp, byte *digest)
	{
		word32 w;

		if (hcp == NULL) {
			return 1; /* bad context */
		}
		if (digest == NULL) {
			return 2; /* bad digest buffer */
		}

	/*	assert (hcp->occupied < 128);  invariant */

		/* append toggle to the context data block: */
		hcp->block[hcp->occupied] = 0x01; /* corrected from 0x80 */

		/* pad the message with null bytes to make it 944 (mod 1024) bits long: */
		if (hcp->occupied++ >= 118) {
			/* no room for tail data on the current context block */
			mhash_bzero (&hcp->block[hcp->occupied], 128 - (int)hcp->occupied);
			/* process the completed context data block: */
			switch (hcp->passes) {
			case 3:
				havalTransform3 (hcp->digest, hcp->block, hcp->temp);
				break;
			case 4:
				havalTransform4 (hcp->digest, hcp->block, hcp->temp);
				break;
			case 5:
				havalTransform5 (hcp->digest, hcp->block, hcp->temp);
				break;
			}
			mhash_bzero (hcp->block, 118);
		} else {
			mhash_bzero (&hcp->block[hcp->occupied], 118 - (int)hcp->occupied);
		}
		/* append tail data and process last (padded) message block: */
		hcp->block[118] = (byte)(
			((hcp->hashLength & 0x03U) << 6) |
			((hcp->passes     & 0x07U) << 3) |
			(HAVAL_VERSION           & 0x07U));
		hcp->block[119] = (byte) (hcp->hashLength >> 2);
		w = hcp->bitCount[0];
		hcp->block[120] = (byte)(w);
		hcp->block[121] = (byte)(w >>  8);
		hcp->block[122] = (byte)(w >> 16);
		hcp->block[123] = (byte)(w >> 24);
		w = hcp->bitCount[1];
		hcp->block[124] = (byte)(w);
		hcp->block[125] = (byte)(w >>  8);
		hcp->block[126] = (byte)(w >> 16);
		hcp->block[127] = (byte)(w >> 24);
		switch (hcp->passes) {
		case 3:
			havalTransform3 (hcp->digest, hcp->block, hcp->temp);
			break;
		case 4:
			havalTransform4 (hcp->digest, hcp->block, hcp->temp);
			break;
		case 5:
			havalTransform5 (hcp->digest, hcp->block, hcp->temp);
			break;
		}

		/* fold 256-bit digest to fit the desired hash length (blaargh!): */
			/* Byte reverse each 32-bit section while outputting if big-endian -- Tines*/
		switch (hcp->hashLength) {
		case 128:
			hcp->digest[3] +=
				( (hcp->digest[7] & 0xFF000000UL)
				| (hcp->digest[6] & 0x00FF0000UL)
				| (hcp->digest[5] & 0x0000FF00UL)
				| (hcp->digest[4] & 0x000000FFUL)
				);
			hcp->digest[2] +=
				(((hcp->digest[7] & 0x00FF0000UL)
				| (hcp->digest[6] & 0x0000FF00UL)
				| (hcp->digest[5] & 0x000000FFUL)
				) << 8) |
				( (hcp->digest[4] & 0xFF000000UL) >> 24);
			hcp->digest[1] +=
				(((hcp->digest[7] & 0x0000FF00UL)
				| (hcp->digest[6] & 0x000000FFUL)) << 16) |
				(((hcp->digest[5] & 0xFF000000UL)
				| (hcp->digest[4] & 0x00FF0000UL)) >> 16);
			hcp->digest[0] +=
				(((hcp->digest[6] & 0xFF000000UL)
				| (hcp->digest[5] & 0x00FF0000UL)
				| (hcp->digest[4] & 0x0000FF00UL)
				) >> 8) |
				( (hcp->digest[7] & 0x000000FFUL) << 24);
	#ifndef WORDS_BIGENDIAN
			memcpy (digest, hcp->digest, 128/8);
	#else
				BYTE_REVERSE(digest, hcp->digest, 128/32);
	#endif /* ?LITTLE_ENDIAN */
			break;
		case 160:
			hcp->digest[4] +=
				((hcp->digest[7] & 0xFE000000UL) | (hcp->digest[6] & 0x01F80000UL) | (hcp->digest[5] & 0x0007F000UL)) >> 12;
			hcp->digest[3] +=
				((hcp->digest[7] & 0x01F80000UL) | (hcp->digest[6] & 0x0007F000UL) | (hcp->digest[5] & 0x00000FC0UL)) >> 6;
			hcp->digest[2] +=
				((hcp->digest[7] & 0x0007F000UL) | (hcp->digest[6] & 0x00000FC0UL) | (hcp->digest[5] & 0x0000003FUL));
			hcp->digest[1] +=
				ROTR
				((hcp->digest[7] & 0x00000FC0UL) | (hcp->digest[6] & 0x0000003FUL) | (hcp->digest[5] & 0xFE000000UL), 25);
			hcp->digest[0] +=
				ROTR
				((hcp->digest[7] & 0x0000003FUL) | (hcp->digest[6] & 0xFE000000UL) | (hcp->digest[5] & 0x01F80000UL), 19);
	#ifndef WORDS_BIGENDIAN
			memcpy (digest, hcp->digest, 160/8);
	#else
				BYTE_REVERSE(digest, hcp->digest, 160/32);
	#endif /* ?LITTLE_ENDIAN */
			break;
		case 192:
			hcp->digest[5] +=
				((hcp->digest[7] & 0xFC000000UL) | (hcp->digest[6] & 0x03E00000UL)) >> 21;
			hcp->digest[4] +=
				((hcp->digest[7] & 0x03E00000UL) | (hcp->digest[6] & 0x001F0000UL)) >> 16;
			hcp->digest[3] +=
				((hcp->digest[7] & 0x001F0000UL) | (hcp->digest[6] & 0x0000FC00UL)) >> 10;
			hcp->digest[2] +=
				((hcp->digest[7] & 0x0000FC00UL) | (hcp->digest[6] & 0x000003E0UL)) >>  5;
			hcp->digest[1] +=
				((hcp->digest[7] & 0x000003E0UL) | (hcp->digest[6] & 0x0000001FUL));
			hcp->digest[0] +=
				ROTR
				((hcp->digest[7] & 0x0000001FUL) | (hcp->digest[6] & 0xFC000000UL), 26);
	#ifndef WORDS_BIGENDIAN
			memcpy (digest, hcp->digest, 192/8);
	#else
				BYTE_REVERSE(digest, hcp->digest, 192/32);
	#endif /* ?LITTLE_ENDIAN */
			break;
		case 224:
			hcp->digest[6] += (hcp->digest[7]      ) & 0x0000000FUL;
			hcp->digest[5] += (hcp->digest[7] >>  4) & 0x0000001FUL;
			hcp->digest[4] += (hcp->digest[7] >>  9) & 0x0000000FUL;
			hcp->digest[3] += (hcp->digest[7] >> 13) & 0x0000001FUL;
			hcp->digest[2] += (hcp->digest[7] >> 18) & 0x0000000FUL;
			hcp->digest[1] += (hcp->digest[7] >> 22) & 0x0000001FUL;
			hcp->digest[0] += (hcp->digest[7] >> 27) & 0x0000001FUL;
	#ifndef WORDS_BIGENDIAN
			memcpy (digest, hcp->digest, 224/8);
	#else
				BYTE_REVERSE(digest, hcp->digest, 224/32);
	#endif /* ?LITTLE_ENDIAN */
			break;
		case 256:
	#ifndef WORDS_BIGENDIAN
			memcpy (digest, hcp->digest, 256/8);
	#else
				BYTE_REVERSE(digest, hcp->digest, 256/32);
	#endif /* ?LITTLE_ENDIAN */
			break;
		}

		/* destroy sensitive information: */
		mhash_bzero (hcp, sizeof (havalContext));
		return 0; /* OK */
	} /* havalFinal */




		// HAVAL 160 test vectors
	static char *HAVAL160_msg[] = 
	{
			"a"
	};

	static char *HAVAL160_val[] =
	{
			"4da08f514a7275dbc4cece4a347385983983a830",
	};


		// tests with standard test vectors and returns true if everything is OK
	bool haval160_test(void) {
		int i, j;
		char output[44];
		havalContext ctx;
		unsigned char haval160sum[21];

				
		for( i = 0; i < 1; i++ ) {
			havalInit160( &ctx );
			havalUpdate( &ctx, (uint8 *) HAVAL160_msg[i],strlen( HAVAL160_msg[i] ) );
			havalFinal( &ctx, haval160sum );

			for( j = 0; j < 20; j++ ) {
				sprintf_s( output + j * 2,3, "%02x", haval160sum[j] );
			}
				
			if( memcmp( output, HAVAL160_val[i], 20 ) ) {
				return false;
			}
		}
		return true;
	}





		// HAVAL 256 test vectors
	static char *HAVAL256_msg[] = 
	{
			"abcdefghijklmnopqrstuvwxyz",
			"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
	};

	static char *HAVAL256_val[] =
	{
			"72fad4bde1da8c8332fb60561a780e7f504f21547b98686824fc33fc796afa76",
			"899397d96489281e9e76d5e65abab751f312e06c06c07c9c1d42abd31bb6a404",
	};


		// tests with standard test vectors and returns true if everything is OK
	bool haval256_test(void) {
		int i, j;
		char output[68];
		havalContext ctx;
		unsigned char haval256sum[33];

				
		for( i = 0; i < 2; i++ ) {
			havalInit256( &ctx );
			havalUpdate( &ctx, (uint8 *) HAVAL256_msg[i],strlen( HAVAL256_msg[i] ) );
			havalFinal( &ctx, haval256sum );

			for( j = 0; j < 32; j++ ) {
				sprintf_s( output + j * 2,3, "%02x", haval256sum[j] );
			}
				
			if( memcmp( output, HAVAL256_val[i], 64 ) ) {
				return false;
			}
		}
		return true;
	}

	
	
	


























//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
/* This is an independent implementation of the encryption algorithm:   */
/*                                                                      */
/*         CAST-256 by Carlisle Adams of Entrust Tecnhologies           */
/*                                                                      */
/* which is a candidate algorithm in the Advanced Encryption Standard   */
/* programme of the US National Institute of Standards and Technology.  */
/*                                                                      */
/* Copyright in this implementation is held by Dr B R Gladman but I     */
/* hereby give permission for its free direct or derivative use subject */
/* to acknowledgment of its origin and compliance with any conditions   */
/* that the originators of the algorithm place on its exploitation.     */
/*                                                                      */
/* Dr Brian Gladman (gladman@seven77.demon.co.uk) 14th January 1999     */

/* modified in order to use the libmcrypt API by Nikos Mavroyanopoulos 
 * All modifications are placed under the license of libmcrypt.
 */

/* $Id: cast-256.c,v 1.13 2003/01/19 17:48:27 nmav Exp $ */

/* modified for mcrypt */

/* Timing data for CAST-256 (cast.c)

Core timing without I/O endian conversion:

128 bit key:
Key Setup:    4333 cycles
Encrypt:       633 cycles =    40.4 mbits/sec
Decrypt:       634 cycles =    40.4 mbits/sec
Mean:          634 cycles =    40.4 mbits/sec

192 bit key:
Key Setup:    4342 cycles
Encrypt:       633 cycles =    40.4 mbits/sec
Decrypt:       633 cycles =    40.4 mbits/sec
Mean:          633 cycles =    40.4 mbits/sec

256 bit key:
Key Setup:    4325 cycles
Encrypt:       639 cycles =    40.1 mbits/sec
Decrypt:       638 cycles =    40.1 mbits/sec
Mean:          639 cycles =    40.1 mbits/sec

Full timing with I/O endian conversion:

128 bit key:
Key Setup:    4294 cycles
Encrypt:       678 cycles =    37.8 mbits/sec
Decrypt:       669 cycles =    38.3 mbits/sec
Mean:          674 cycles =    38.0 mbits/sec

192 bit key:
Key Setup:    4314 cycles
Encrypt:       678 cycles =    37.8 mbits/sec
Decrypt:       670 cycles =    38.2 mbits/sec
Mean:          674 cycles =    38.0 mbits/sec

256 bit key:
Key Setup:    4313 cycles
Encrypt:       678 cycles =    37.8 mbits/sec
Decrypt:       669 cycles =    38.3 mbits/sec
Mean:          674 cycles =    38.0 mbits/sec

*/
	#define _mcrypt_set_key cast_256_LTX__mcrypt_set_key
	#define _mcrypt_encrypt cast_256_LTX__mcrypt_encrypt
	#define _mcrypt_decrypt cast_256_LTX__mcrypt_decrypt
	#define _mcrypt_get_size cast_256_LTX__mcrypt_get_size
	#define _mcrypt_get_block_size cast_256_LTX__mcrypt_get_block_size
	#define _is_block_algorithm cast_256_LTX__is_block_algorithm
	#define _mcrypt_get_key_size cast_256_LTX__mcrypt_get_key_size
	#define _mcrypt_get_supported_key_sizes cast_256_LTX__mcrypt_get_supported_key_sizes
	#define _mcrypt_get_algorithms_name cast_256_LTX__mcrypt_get_algorithms_name
	#define _mcrypt_self_test cast_256_LTX__mcrypt_self_test
	#define _mcrypt_algorithm_version cast_256_LTX__mcrypt_algorithm_version

	#define byte(x,n)   ((byte)((x) >> (8 * n)))

	word32 cast256_sbox[4][256] = { {
					 0x30fb40d4, 0x9fa0ff0b, 0x6beccd2f,
					 0x3f258c7a, 0x1e213f2f, 0x9C004dd3,
					 0x6003e540, 0xcf9fc949, 0xbfd4af27,
					 0x88bbbdb5, 0xe2034090, 0x98d09675,
					 0x6e63a0e0, 0x15c361d2, 0xc2e7661d,
					 0x22d4ff8e, 0x28683b6f, 0xc07fd059,
					 0xff2379c8, 0x775f50e2, 0x43c340d3,
					 0xdf2f8656, 0x887ca41a, 0xa2d2bd2d,
					 0xa1c9e0d6, 0x346c4819, 0x61b76d87,
					 0x22540f2f, 0x2abe32e1, 0xaa54166b,
					 0x22568e3a, 0xa2d341d0, 0x66db40c8,
					 0xa784392f, 0x004dff2f, 0x2db9d2de,
					 0x97943fac, 0x4a97c1d8, 0x527644b7,
					 0xb5f437a7, 0xb82cbaef, 0xd751d159,
					 0x6ff7f0ed, 0x5a097a1f, 0x827b68d0,
					 0x90ecf52e, 0x22b0c054, 0xbc8e5935,
					 0x4b6d2f7f, 0x50bb64a2, 0xd2664910,
					 0xbee5812d, 0xb7332290, 0xe93b159f,
					 0xb48ee411, 0x4bff345d, 0xfd45c240,
					 0xad31973f, 0xc4f6d02e, 0x55fc8165,
					 0xd5b1caad, 0xa1ac2dae, 0xa2d4b76d,
					 0xc19b0C50, 0x882240f2, 0x0c6e4f38,
					 0xa4e4bfd7, 0x4f5ba272, 0x564c1d2f,
					 0xc59c5319, 0xb949e354, 0xb04669fe,
					 0xb1b6ab8a, 0xc71358dd, 0x6385c545,
					 0x110f935d, 0x57538ad5, 0x6a390493,
					 0xe63d37e0, 0x2a54f6b3, 0x3a787d5f,
					 0x6276a0b5, 0x19a6fcdf, 0x7a42206a,
					 0x29f9d4d5, 0xf61b1891, 0xbb72275e,
					 0xaa508167, 0x38901091, 0xc6b505eb,
					 0x84c7cb8c, 0x2ad75a0f, 0x874a1427,
					 0xa2d1936b, 0x2ad286af, 0xaa56d291,
					 0xd7894360, 0x425c750d, 0x93b39e26,
					 0x187184c9, 0x6c00b32d, 0x73e2bb14,
					 0xa0bebc3c, 0x54623779, 0x64459eab,
					 0x3f328b82, 0x7718cf82, 0x59a2cea6,
					 0x04ee002e, 0x89fe78e6, 0x3fab0950,
					 0x325ff6C2, 0x81383f05, 0x6963c5c8,
					 0x76cb5ad6, 0xd49974c9, 0xca180dcf,
					 0x380782d5, 0xc7fa5cf6, 0x8ac31511,
					 0x35e79e13, 0x47da91d0, 0xf40f9086,
					 0xa7e2419e, 0x31366241, 0x051ef495,
					 0xaa573b04, 0x4a805d8d, 0x548300d0,
					 0x00322a3c, 0xbf64cddf, 0xba57a68e,
					 0x75c6372b, 0x50afd341, 0xa7c13275,
					 0x915a0bf5, 0x6b54bfab, 0x2b0b1426,
					 0xab4cc9d7, 0x449ccd82, 0xf7fbf265,
					 0xab85c5f3, 0x1b55db94, 0xaad4e324,
					 0xcfa4bd3f, 0x2deaa3e2, 0x9e204d02,
					 0xc8bd25ac, 0xeadf55b3, 0xd5bd9e98,
					 0xe31231b2, 0x2ad5ad6c, 0x954329de,
					 0xadbe4528, 0xd8710f69, 0xaa51c90f,
					 0xaa786bf6, 0x22513f1e, 0xaa51a79b,
					 0x2ad344cc, 0x7b5a41f0, 0xd37cfbad,
					 0x1b069505, 0x41ece491, 0xb4c332e6,
					 0x032268d4, 0xc9600acc, 0xce387e6d,
					 0xbf6bb16c, 0x6a70fb78, 0x0d03d9c9,
					 0xd4df39de, 0xe01063da, 0x4736f464,
					 0x5ad328d8, 0xb347cc96, 0x75bb0fc3,
					 0x98511bfb, 0x4ffbcc35, 0xb58bcf6a,
					 0xe11f0abc, 0xbfc5fe4a, 0xa70aec10,
					 0xac39570a, 0x3f04442f, 0x6188b153,
					 0xe0397a2e, 0x5727cb79, 0x9ceb418f,
					 0x1cacd68d, 0x2ad37c96, 0x0175cb9d,
					 0xc69dff09, 0xc75b65f0, 0xd9db40d8,
					 0xec0e7779, 0x4744ead4, 0xb11c3274,
					 0xdd24cb9e, 0x7e1c54bd, 0xf01144f9,
					 0xd2240eb1, 0x9675b3fd, 0xa3ac3755,
					 0xd47c27af, 0x51c85f4d, 0x56907596,
					 0xa5bb15e6, 0x580304f0, 0xca042cf1,
					 0x011a37ea, 0x8dbfaadb, 0x35ba3e4a,
					 0x3526ffa0, 0xc37b4d09, 0xbc306ed9,
					 0x98a52666, 0x5648f725, 0xff5e569d,
					 0x0ced63d0, 0x7c63b2cf, 0x700b45e1,
					 0xd5ea50f1, 0x85a92872, 0xaf1fbda7,
					 0xd4234870, 0xa7870bf3, 0x2d3b4d79,
					 0x42e04198, 0x0cd0ede7, 0x26470db8,
					 0xf881814C, 0x474d6ad7, 0x7c0c5e5c,
					 0xd1231959, 0x381b7298, 0xf5d2f4db,
					 0xab838653, 0x6e2f1e23, 0x83719c9e,
					 0xbd91e046, 0x9a56456e, 0xdc39200c,
					 0x20c8c571, 0x962bda1c, 0xe1e696ff,
					 0xb141ab08, 0x7cca89b9, 0x1a69e783,
					 0x02cc4843, 0xa2f7c579, 0x429ef47d,
					 0x427b169c, 0x5ac9f049, 0xdd8f0f00,
					 0x5c8165bf}
	,
	{
	 0x1f201094, 0xef0ba75b, 0x69e3cf7e, 0x393f4380, 0xfe61cf7a, 0xeec5207a,
	 0x55889c94, 0x72fc0651, 0xada7ef79, 0x4e1d7235, 0xd55a63ce, 0xde0436ba,
	 0x99c430ef, 0x5f0c0794, 0x18dcdb7d, 0xa1d6eff3, 0xa0b52f7b, 0x59e83605,
	 0xee15b094, 0xe9ffd909, 0xdc440086, 0xef944459, 0xba83ccb3, 0xe0c3cdfb,
	 0xd1da4181, 0x3b092ab1, 0xf997f1c1, 0xa5e6cf7b, 0x01420ddb, 0xe4e7ef5b,
	 0x25a1ff41, 0xe180f806, 0x1fc41080, 0x179bee7a, 0xd37ac6a9, 0xfe5830a4,
	 0x98de8b7f, 0x77e83f4e, 0x79929269, 0x24fa9f7b, 0xe113c85b, 0xacc40083,
	 0xd7503525, 0xf7ea615f, 0x62143154, 0x0d554b63, 0x5d681121, 0xc866c359,
	 0x3d63cf73, 0xcee234c0, 0xd4d87e87, 0x5c672b21, 0x071f6181, 0x39f7627f,
	 0x361e3084, 0xe4eb573b, 0x602f64a4, 0xd63acd9c, 0x1bbc4635, 0x9e81032d,
	 0x2701f50c, 0x99847ab4, 0xa0e3df79, 0xba6cf38c, 0x10843094, 0x2537a95e,
	 0xf46f6ffe, 0xa1ff3b1f, 0x208cfb6a, 0x8f458c74, 0xd9e0a227, 0x4ec73a34,
	 0xfc884f69, 0x3e4de8df, 0xef0e0088, 0x3559648d, 0x8a45388c, 0x1d804366,
	 0x721d9bfd, 0xa58684bb, 0xe8256333, 0x844e8212, 0x128d8098, 0xfed33fb4,
	 0xce280ae1, 0x27e19ba5, 0xd5a6c252, 0xe49754bd, 0xc5d655dd, 0xeb667064,
	 0x77840b4d, 0xa1b6a801, 0x84db26a9, 0xe0b56714, 0x21f043b7, 0xe5d05860,
	 0x54f03084, 0x066ff472, 0xa31aa153, 0xdadc4755, 0xb5625dbf, 0x68561be6,
	 0x83ca6b94, 0x2d6ed23b, 0xeccf01db, 0xa6d3d0ba, 0xb6803d5c, 0xaf77a709,
	 0x33b4a34c, 0x397bc8d6, 0x5ee22b95, 0x5f0e5304, 0x81ed6f61, 0x20e74364,
	 0xb45e1378, 0xde18639b, 0x881ca122, 0xb96726d1, 0x8049a7e8, 0x22b7da7b,
	 0x5e552d25, 0x5272d237, 0x79d2951c, 0xc60d894c, 0x488cb402, 0x1ba4fe5b,
	 0xa4b09f6b, 0x1ca815cf, 0xa20c3005, 0x8871df63, 0xb9de2fcb, 0x0cc6c9e9,
	 0x0beeff53, 0xe3214517, 0xb4542835, 0x9f63293c, 0xee41e729, 0x6e1d2d7c,
	 0x50045286, 0x1e6685f3, 0xf33401c6, 0x30a22c95, 0x31a70850, 0x60930f13,
	 0x73f98417, 0xa1269859, 0xec645c44, 0x52c877a9, 0xcdff33a6, 0xa02b1741,
	 0x7cbad9a2, 0x2180036f, 0x50d99c08, 0xcb3f4861, 0xc26bd765, 0x64a3f6ab,
	 0x80342676, 0x25a75e7b, 0xe4e6d1fc, 0x20c710e6, 0xcdf0b680, 0x17844d3b,
	 0x31eef84d, 0x7e0824e4, 0x2ccb49eb, 0x846a3bae, 0x8ff77888, 0xee5d60f6,
	 0x7af75673, 0x2fdd5cdb, 0xa11631c1, 0x30f66f43, 0xb3faec54, 0x157fd7fa,
	 0xef8579cc, 0xd152de58, 0xdb2ffd5e, 0x8f32ce19, 0x306af97a, 0x02f03ef8,
	 0x99319ad5, 0xc242fa0f, 0xa7e3ebb0, 0xc68e4906, 0xb8da230c, 0x80823028,
	 0xdcdef3c8, 0xd35fb171, 0x088a1bc8, 0xbec0c560, 0x61a3c9e8, 0xbca8f54d,
	 0xc72feffa, 0x22822e99, 0x82c570b4, 0xd8d94e89, 0x8b1c34bc, 0x301e16e6,
	 0x273be979, 0xb0ffeaa6, 0x61d9b8c6, 0x00b24869, 0xb7ffce3f, 0x08dc283b,
	 0x43daf65a, 0xf7e19798, 0x7619b72f, 0x8f1c9ba4, 0xdc8637a0, 0x16a7d3b1,
	 0x9fc393b7, 0xa7136eeb, 0xc6bcc63e, 0x1a513742, 0xef6828bc, 0x520365d6,
	 0x2d6a77ab, 0x3527ed4b, 0x821fd216, 0x095c6e2e, 0xdb92f2fb, 0x5eea29cb,
	 0x145892f5, 0x91584f7f, 0x5483697b, 0x2667a8cc, 0x85196048, 0x8c4bacea,
	 0x833860d4, 0x0d23e0f9, 0x6c387e8a, 0x0ae6d249, 0xb284600c, 0xd835731d,
	 0xdcb1c647, 0xac4c56ea, 0x3ebd81b3, 0x230eabb0, 0x6438bc87, 0xf0b5b1fa,
	 0x8f5ea2b3, 0xfc184642, 0x0a036b7a, 0x4fb089bd, 0x649da589, 0xa345415e,
	 0x5c038323, 0x3e5d3bb9, 0x43d79572, 0x7e6dd07c, 0x06dfdf1e, 0x6c6cc4ef,
	 0x7160a539, 0x73bfbe70, 0x83877605, 0x4523ecf1}
	,
	{
	 0x8defc240, 0x25fa5d9f, 0xeb903dbf, 0xe810c907, 0x47607fff, 0x369fe44b,
	 0x8c1fc644, 0xaececa90, 0xbeb1f9bf, 0xeefbcaea, 0xe8cf1950, 0x51df07ae,
	 0x920e8806, 0xf0ad0548, 0xe13c8d83, 0x927010d5, 0x11107d9f, 0x07647db9,
	 0xb2e3e4d4, 0x3d4f285e, 0xb9afa820, 0xfade82e0, 0xa067268b, 0x8272792e,
	 0x553fb2c0, 0x489ae22b, 0xd4ef9794, 0x125e3fbc, 0x21fffcee, 0x825b1bfd,
	 0x9255c5ed, 0x1257a240, 0x4e1a8302, 0xbae07fff, 0x528246e7, 0x8e57140e,
	 0x3373f7bf, 0x8c9f8188, 0xa6fc4ee8, 0xc982b5a5, 0xa8c01db7, 0x579fc264,
	 0x67094f31, 0xf2bd3f5f, 0x40fff7c1, 0x1fb78dfc, 0x8e6bd2c1, 0x437be59b,
	 0x99b03dbf, 0xb5dbc64b, 0x638dc0e6, 0x55819d99, 0xa197c81c, 0x4a012d6e,
	 0xc5884a28, 0xccc36f71, 0xb843c213, 0x6c0743f1, 0x8309893c, 0x0feddd5f,
	 0x2f7fe850, 0xd7c07f7e, 0x02507fbf, 0x5afb9a04, 0xa747d2d0, 0x1651192e,
	 0xaf70bf3e, 0x58c31380, 0x5f98302e, 0x727cc3c4, 0x0a0fb402, 0x0f7fef82,
	 0x8c96fdad, 0x5d2c2aae, 0x8ee99a49, 0x50da88b8, 0x8427f4a0, 0x1eac5790,
	 0x796fb449, 0x8252dc15, 0xefbd7d9b, 0xa672597d, 0xada840d8, 0x45f54504,
	 0xfa5d7403, 0xe83ec305, 0x4f91751a, 0x925669c2, 0x23efe941, 0xa903f12e,
	 0x60270df2, 0x0276e4b6, 0x94fd6574, 0x927985b2, 0x8276dbcb, 0x02778176,
	 0xf8af918d, 0x4e48f79e, 0x8f616ddf, 0xe29d840e, 0x842f7d83, 0x340ce5c8,
	 0x96bbb682, 0x93b4b148, 0xef303cab, 0x984faf28, 0x779faf9b, 0x92dc560d,
	 0x224d1e20, 0x8437aa88, 0x7d29dc96, 0x2756d3dc, 0x8b907cee, 0xb51fd240,
	 0xe7c07ce3, 0xe566b4a1, 0xc3e9615e, 0x3cf8209d, 0x6094d1e3, 0xcd9ca341,
	 0x5c76460e, 0x00ea983b, 0xd4d67881, 0xfd47572c, 0xf76cedd9, 0xbda8229c,
	 0x127dadaa, 0x438a074e, 0x1f97c090, 0x081bdb8a, 0x93a07ebe, 0xb938ca15,
	 0x97b03cff, 0x3dc2c0f8, 0x8d1ab2ec, 0x64380e51, 0x68cc7bfb, 0xd90f2788,
	 0x12490181, 0x5de5ffd4, 0xdd7ef86a, 0x76a2e214, 0xb9a40368, 0x925d958f,
	 0x4b39fffa, 0xba39aee9, 0xa4ffd30b, 0xfaf7933b, 0x6d498623, 0x193cbcfa,
	 0x27627545, 0x825cf47a, 0x61bd8ba0, 0xd11e42d1, 0xcead04f4, 0x127ea392,
	 0x10428db7, 0x8272a972, 0x9270c4a8, 0x127de50b, 0x285ba1c8, 0x3c62f44f,
	 0x35c0eaa5, 0xe805d231, 0x428929fb, 0xb4fcdf82, 0x4fb66a53, 0x0e7dc15b,
	 0x1f081fab, 0x108618ae, 0xfcfd086d, 0xf9ff2889, 0x694bcc11, 0x236a5cae,
	 0x12deca4d, 0x2c3f8cc5, 0xd2d02dfe, 0xf8ef5896, 0xe4cf52da, 0x95155b67,
	 0x494a488c, 0xb9b6a80c, 0x5c8f82bc, 0x89d36b45, 0x3a609437, 0xec00c9a9,
	 0x44715253, 0x0a874b49, 0xd773bc40, 0x7c34671c, 0x02717ef6, 0x4feb5536,
	 0xa2d02fff, 0xd2bf60c4, 0xd43f03c0, 0x50b4ef6d, 0x07478cd1, 0x006e1888,
	 0xa2e53f55, 0xb9e6d4bc, 0xa2048016, 0x97573833, 0xd7207d67, 0xde0f8f3d,
	 0x72f87b33, 0xabcc4f33, 0x7688c55d, 0x7b00a6b0, 0x947b0001, 0x570075d2,
	 0xf9bb88f8, 0x8942019e, 0x4264a5ff, 0x856302e0, 0x72dbd92b, 0xee971b69,
	 0x6ea22fde, 0x5f08ae2b, 0xaf7a616d, 0xe5c98767, 0xcf1febd2, 0x61efc8c2,
	 0xf1ac2571, 0xcc8239c2, 0x67214cb8, 0xb1e583d1, 0xb7dc3e62, 0x7f10bdce,
	 0xf90a5c38, 0x0ff0443d, 0x606e6dc6, 0x60543a49, 0x5727c148, 0x2be98a1d,
	 0x8ab41738, 0x20e1be24, 0xaf96da0f, 0x68458425, 0x99833be5, 0x600d457d,
	 0x282f9350, 0x8334b362, 0xd91d1120, 0x2b6d8da0, 0x642b1e31, 0x9c305a00,
	 0x52bce688, 0x1b03588a, 0xf7baefd5, 0x4142ed9c, 0xa4315c11, 0x83323ec5,
	 0xdfef4636, 0xa133c501, 0xe9d3531c, 0xee353783}
	,
	{
	 0x9db30420, 0x1fb6e9de, 0xa7be7bef, 0xd273a298, 0x4a4f7bdb, 0x64ad8c57,
	 0x85510443, 0xfa020ed1, 0x7e287aff, 0xe60fb663, 0x095f35a1, 0x79ebf120,
	 0xfd059d43, 0x6497b7b1, 0xf3641f63, 0x241e4adf, 0x28147f5f, 0x4fa2b8cd,
	 0xc9430040, 0x0cc32220, 0xfdd30b30, 0xc0a5374f, 0x1d2d00d9, 0x24147b15,
	 0xee4d111a, 0x0fca5167, 0x71ff904c, 0x2d195ffe, 0x1a05645f, 0x0c13fefe,
	 0x081b08ca, 0x05170121, 0x80530100, 0xe83e5efe, 0xac9af4f8, 0x7fe72701,
	 0xd2b8ee5f, 0x06df4261, 0xbb9e9b8a, 0x7293ea25, 0xce84ffdf, 0xf5718801,
	 0x3dd64b04, 0xa26f263b, 0x7ed48400, 0x547eebe6, 0x446d4ca0, 0x6cf3d6f5,
	 0x2649abdf, 0xaea0c7f5, 0x36338cc1, 0x503f7e93, 0xd3772061, 0x11b638e1,
	 0x72500e03, 0xf80eb2bb, 0xabe0502e, 0xec8d77de, 0x57971e81, 0xe14f6746,
	 0xc9335400, 0x6920318f, 0x081dbb99, 0xffc304a5, 0x4d351805, 0x7f3d5ce3,
	 0xa6c866c6, 0x5d5bcca9, 0xdaec6fea, 0x9f926f91, 0x9f46222f, 0x3991467d,
	 0xa5bf6d8e, 0x1143c44f, 0x43958302, 0xd0214eeb, 0x022083b8, 0x3fb6180c,
	 0x18f8931e, 0x281658e6, 0x26486e3e, 0x8bd78a70, 0x7477e4c1, 0xb506e07c,
	 0xf32d0a25, 0x79098b02, 0xe4eabb81, 0x28123b23, 0x69dead38, 0x1574ca16,
	 0xdf871b62, 0x211c40b7, 0xa51a9ef9, 0x0014377b, 0x041e8ac8, 0x09114003,
	 0xbd59e4d2, 0xe3d156d5, 0x4fe876d5, 0x2f91a340, 0x557be8de, 0x00eae4a7,
	 0x0ce5c2ec, 0x4db4bba6, 0xe756bdff, 0xdd3369ac, 0xec17b035, 0x06572327,
	 0x99afc8b0, 0x56c8c391, 0x6b65811c, 0x5e146119, 0x6e85cb75, 0xbe07c002,
	 0xc2325577, 0x893ff4ec, 0x5bbfc92d, 0xd0ec3b25, 0xb7801ab7, 0x8d6d3b24,
	 0x20c763ef, 0xc366a5fc, 0x9c382880, 0x0ace3205, 0xaac9548a, 0xeca1d7c7,
	 0x041afa32, 0x1d16625a, 0x6701902c, 0x9b757a54, 0x31d477f7, 0x9126b031,
	 0x36cc6fdb, 0xc70b8b46, 0xd9e66a48, 0x56e55a79, 0x026a4ceb, 0x52437eff,
	 0x2f8f76b4, 0x0df980a5, 0x8674cde3, 0xedda04eb, 0x17a9be04, 0x2c18f4df,
	 0xb7747f9d, 0xab2af7b4, 0xefc34d20, 0x2e096b7c, 0x1741a254, 0xe5b6a035,
	 0x213d42f6, 0x2c1c7c26, 0x61c2f50f, 0x6552daf9, 0xd2c231f8, 0x25130f69,
	 0xd8167fa2, 0x0418f2c8, 0x001a96a6, 0x0d1526ab, 0x63315c21, 0x5e0a72ec,
	 0x49bafefd, 0x187908d9, 0x8d0dbd86, 0x311170a7, 0x3e9b640c, 0xcc3e10d7,
	 0xd5cad3b6, 0x0caec388, 0xf73001e1, 0x6c728aff, 0x71eae2a1, 0x1f9af36e,
	 0xcfcbd12f, 0xc1de8417, 0xac07be6b, 0xcb44a1d8, 0x8b9b0f56, 0x013988c3,
	 0xb1c52fca, 0xb4be31cd, 0xd8782806, 0x12a3a4e2, 0x6f7de532, 0x58fd7eb6,
	 0xd01ee900, 0x24adffc2, 0xf4990fc5, 0x9711aac5, 0x001d7b95, 0x82e5e7d2,
	 0x109873f6, 0x00613096, 0xc32d9521, 0xada121ff, 0x29908415, 0x7fbb977f,
	 0xaf9eb3db, 0x29c9ed2a, 0x5ce2a465, 0xa730f32c, 0xd0aa3fe8, 0x8a5cc091,
	 0xd49e2ce7, 0x0ce454a9, 0xd60acd86, 0x015f1919, 0x77079103, 0xdea03af6,
	 0x78a8565e, 0xdee356df, 0x21f05cbe, 0x8b75e387, 0xb3c50651, 0xb8a5c3ef,
	 0xd8eeb6d2, 0xe523be77, 0xc2154529, 0x2f69efdf, 0xafe67afb, 0xf470c4b2,
	 0xf3e0eb5b, 0xd6cc9876, 0x39e4460c, 0x1fda8538, 0x1987832f, 0xca007367,
	 0xa99144f8, 0x296b299e, 0x492fc295, 0x9266beab, 0xb5676e69, 0x9bd3ddda,
	 0xdf7e052f, 0xdb25701c, 0x1b5e51ee, 0xf65324e6, 0x6afce36c, 0x0316cc04,
	 0x8644213e, 0xb7dc59d0, 0x7965291f, 0xccd6fd43, 0x41823979, 0x932bcdf6,
	 0xb657c34d, 0x4edfd282, 0x7ae5290c, 0x3cb9536b, 0x851e20fe, 0x9833557e,
	 0x13ecf0b0, 0xd3ffb372, 0x3f85c5c1, 0x0aef7ed2}
	};


	#define f1(y,x,kr,km)           \
			t  = rotl32(km + x, kr);      \
			u  = cast256_sbox[0][byte(t,3)];   \
			u ^= cast256_sbox[1][byte(t,2)];   \
			u -= cast256_sbox[2][byte(t,1)];   \
			u += cast256_sbox[3][byte(t,0)];   \
			y ^= u

	#define f2(y,x,kr,km)           \
			t  = rotl32(km ^ x, kr);      \
			u  = cast256_sbox[0][byte(t,3)];   \
			u -= cast256_sbox[1][byte(t,2)];   \
			u += cast256_sbox[2][byte(t,1)];   \
			u ^= cast256_sbox[3][byte(t,0)];   \
			y ^= u

	#define f3(y,x,kr,km)           \
			t  = rotl32(km - x, kr);      \
			u  = cast256_sbox[0][byte(t,3)];   \
			u += cast256_sbox[1][byte(t,2)];   \
			u ^= cast256_sbox[2][byte(t,1)];   \
			u -= cast256_sbox[3][byte(t,0)];   \
			y ^= u

	#define f_rnd(x,n)                              \
			f1(x[2],x[3],key->l_key[n],    key->l_key[n + 4]);    \
			f2(x[1],x[2],key->l_key[n + 1],key->l_key[n + 5]);    \
			f3(x[0],x[1],key->l_key[n + 2],key->l_key[n + 6]);    \
			f1(x[3],x[0],key->l_key[n + 3],key->l_key[n + 7])

	#define i_rnd(x, n)                             \
			f1(x[3],x[0],key->l_key[n + 3],key->l_key[n + 7]);    \
			f3(x[0],x[1],key->l_key[n + 2],key->l_key[n + 6]);    \
			f2(x[1],x[2],key->l_key[n + 1],key->l_key[n + 5]);    \
			f1(x[2],x[3],key->l_key[n],    key->l_key[n + 4])

	#define k_rnd(k,tr,tm)          \
			f1(k[6],k[7],tr[0],tm[0]);  \
			f2(k[5],k[6],tr[1],tm[1]);  \
			f3(k[4],k[5],tr[2],tm[2]);  \
			f1(k[3],k[4],tr[3],tm[3]);  \
			f2(k[2],k[3],tr[4],tm[4]);  \
			f3(k[1],k[2],tr[5],tm[5]);  \
			f1(k[0],k[1],tr[6],tm[6]);  \
			f2(k[7],k[0],tr[7],tm[7])


	int cast256_set_key(cast256_key * key, const word32 * in_key, const int key_len)
	{
		word32 i, j, t, u, cm, cr, lk[8], tm[8], tr[8];


		for (i = 0; i < key_len / sizeof(word32); ++i)
	#ifdef WORDS_BIGENDIAN
			lk[i] = byteswap32(in_key[i]);
	#else
			lk[i] = in_key[i];
	#endif

		for (; i < 8; ++i)

			lk[i] = 0;

		cm = 0x5a827999;
		cr = 19;

		for (i = 0; i < 96; i += 8) {
			for (j = 0; j < 8; ++j) {
				tm[j] = cm;
				cm += 0x6ed9eba1;
				tr[j] = cr;
				cr += 17;
			}

			k_rnd(lk, tr, tm);

			for (j = 0; j < 8; ++j) {
				tm[j] = cm;
				cm += 0x6ed9eba1;
				tr[j] = cr;
				cr += 17;
			}

			k_rnd(lk, tr, tm);

			key->l_key[i + 0] = lk[0];
			key->l_key[i + 1] = lk[2];
			key->l_key[i + 2] = lk[4];
			key->l_key[i + 3] = lk[6];
			key->l_key[i + 4] = lk[7];
			key->l_key[i + 5] = lk[5];
			key->l_key[i + 6] = lk[3];
			key->l_key[i + 7] = lk[1];
		}

		return 0;
	}

	/* encrypt a block of text  */
	/* 16 bytes */
	void cast256_encrypt(cast256_key * key, word32 * blk)
	{
		word32 t, u;

	#ifdef WORDS_BIGENDIAN
		blk[0] = byteswap32(blk[0]);
		blk[1] = byteswap32(blk[1]);
		blk[2] = byteswap32(blk[2]);
		blk[3] = byteswap32(blk[3]);
	#endif

		f_rnd(blk, 0);
		f_rnd(blk, 8);
		f_rnd(blk, 16);
		f_rnd(blk, 24);
		f_rnd(blk, 32);
		f_rnd(blk, 40);
		i_rnd(blk, 48);
		i_rnd(blk, 56);
		i_rnd(blk, 64);
		i_rnd(blk, 72);
		i_rnd(blk, 80);
		i_rnd(blk, 88);

	#ifdef WORDS_BIGENDIAN
		blk[0] = byteswap32(blk[0]);
		blk[1] = byteswap32(blk[1]);
		blk[2] = byteswap32(blk[2]);
		blk[3] = byteswap32(blk[3]);
	#endif
	}

	/* decrypt a block of text  */

	void cast256_decrypt(cast256_key * key, word32 * blk)
	{
		word32 t, u;

	#ifdef WORDS_BIGENDIAN
		blk[0] = byteswap32(blk[0]);
		blk[1] = byteswap32(blk[1]);
		blk[2] = byteswap32(blk[2]);
		blk[3] = byteswap32(blk[3]);
	#endif
		f_rnd(blk, 88);
		f_rnd(blk, 80);
		f_rnd(blk, 72);
		f_rnd(blk, 64);
		f_rnd(blk, 56);
		f_rnd(blk, 48);
		i_rnd(blk, 40);
		i_rnd(blk, 32);
		i_rnd(blk, 24);
		i_rnd(blk, 16);
		i_rnd(blk, 8);
		i_rnd(blk, 0);

	#ifdef WORDS_BIGENDIAN
		blk[0] = byteswap32(blk[0]);
		blk[1] = byteswap32(blk[1]);
		blk[2] = byteswap32(blk[2]);
		blk[3] = byteswap32(blk[3]);
	#endif

	}


	int cast256_get_size()
	{
		return sizeof(cast256_key);
	}



	bool cast256_self_test()
	{
		char keyword[33];
		unsigned char plaintext[16];
		unsigned char ciphertext[16];
		int blocksize = 16, j;
		cast256_instance key;
		unsigned char cipher_tmp[210];


		for (j = 0; j < 32; j++) {
			keyword[j] = ((j * 2 + 10) % 256);
		}

		for (j = 0; j < blocksize; j++) {
			plaintext[j] = j % 256;
		}

		memcpy(ciphertext, plaintext, blocksize);

		cast256_set_key(&key, (const unsigned __int32 *) keyword, 32);
		cast256_encrypt(&key, (unsigned __int32 *) ciphertext);

		for (j = 0; j < blocksize; j++) {
			sprintf_s(&((char *) cipher_tmp)[2 * j],3, "%.2x",
				ciphertext[j]);
		}

		cast256_decrypt(&key, (unsigned __int32 *) ciphertext);


		if (memcmp(ciphertext, plaintext, 16) != 0) {
			printf("failed internally\n");
			return false;
		}

		return true;
	}


















//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------------
// FIPS-197 compliant AES implementation
// by Christophe Devine <devine@cr0.net>.
// AES is royalty-free

		// forward S-box 
	static uint32 FSb[256] =
	{
			0x63, 0x7C, 0x77, 0x7B, 0xF2, 0x6B, 0x6F, 0xC5,
			0x30, 0x01, 0x67, 0x2B, 0xFE, 0xD7, 0xAB, 0x76,
			0xCA, 0x82, 0xC9, 0x7D, 0xFA, 0x59, 0x47, 0xF0,
			0xAD, 0xD4, 0xA2, 0xAF, 0x9C, 0xA4, 0x72, 0xC0,
			0xB7, 0xFD, 0x93, 0x26, 0x36, 0x3F, 0xF7, 0xCC,
			0x34, 0xA5, 0xE5, 0xF1, 0x71, 0xD8, 0x31, 0x15,
			0x04, 0xC7, 0x23, 0xC3, 0x18, 0x96, 0x05, 0x9A,
			0x07, 0x12, 0x80, 0xE2, 0xEB, 0x27, 0xB2, 0x75,
			0x09, 0x83, 0x2C, 0x1A, 0x1B, 0x6E, 0x5A, 0xA0,
			0x52, 0x3B, 0xD6, 0xB3, 0x29, 0xE3, 0x2F, 0x84,
			0x53, 0xD1, 0x00, 0xED, 0x20, 0xFC, 0xB1, 0x5B,
			0x6A, 0xCB, 0xBE, 0x39, 0x4A, 0x4C, 0x58, 0xCF,
			0xD0, 0xEF, 0xAA, 0xFB, 0x43, 0x4D, 0x33, 0x85,
			0x45, 0xF9, 0x02, 0x7F, 0x50, 0x3C, 0x9F, 0xA8,
			0x51, 0xA3, 0x40, 0x8F, 0x92, 0x9D, 0x38, 0xF5,
			0xBC, 0xB6, 0xDA, 0x21, 0x10, 0xFF, 0xF3, 0xD2,
			0xCD, 0x0C, 0x13, 0xEC, 0x5F, 0x97, 0x44, 0x17,
			0xC4, 0xA7, 0x7E, 0x3D, 0x64, 0x5D, 0x19, 0x73,
			0x60, 0x81, 0x4F, 0xDC, 0x22, 0x2A, 0x90, 0x88,
			0x46, 0xEE, 0xB8, 0x14, 0xDE, 0x5E, 0x0B, 0xDB,
			0xE0, 0x32, 0x3A, 0x0A, 0x49, 0x06, 0x24, 0x5C,
			0xC2, 0xD3, 0xAC, 0x62, 0x91, 0x95, 0xE4, 0x79,
			0xE7, 0xC8, 0x37, 0x6D, 0x8D, 0xD5, 0x4E, 0xA9,
			0x6C, 0x56, 0xF4, 0xEA, 0x65, 0x7A, 0xAE, 0x08,
			0xBA, 0x78, 0x25, 0x2E, 0x1C, 0xA6, 0xB4, 0xC6,
			0xE8, 0xDD, 0x74, 0x1F, 0x4B, 0xBD, 0x8B, 0x8A,
			0x70, 0x3E, 0xB5, 0x66, 0x48, 0x03, 0xF6, 0x0E,
			0x61, 0x35, 0x57, 0xB9, 0x86, 0xC1, 0x1D, 0x9E,
			0xE1, 0xF8, 0x98, 0x11, 0x69, 0xD9, 0x8E, 0x94,
			0x9B, 0x1E, 0x87, 0xE9, 0xCE, 0x55, 0x28, 0xDF,
			0x8C, 0xA1, 0x89, 0x0D, 0xBF, 0xE6, 0x42, 0x68,
			0x41, 0x99, 0x2D, 0x0F, 0xB0, 0x54, 0xBB, 0x16
	};

	// forward table
	#define FT \
	\
			V(C6,63,63,A5), V(F8,7C,7C,84), V(EE,77,77,99), V(F6,7B,7B,8D), \
			V(FF,F2,F2,0D), V(D6,6B,6B,BD), V(DE,6F,6F,B1), V(91,C5,C5,54), \
			V(60,30,30,50), V(02,01,01,03), V(CE,67,67,A9), V(56,2B,2B,7D), \
			V(E7,FE,FE,19), V(B5,D7,D7,62), V(4D,AB,AB,E6), V(EC,76,76,9A), \
			V(8F,CA,CA,45), V(1F,82,82,9D), V(89,C9,C9,40), V(FA,7D,7D,87), \
			V(EF,FA,FA,15), V(B2,59,59,EB), V(8E,47,47,C9), V(FB,F0,F0,0B), \
			V(41,AD,AD,EC), V(B3,D4,D4,67), V(5F,A2,A2,FD), V(45,AF,AF,EA), \
			V(23,9C,9C,BF), V(53,A4,A4,F7), V(E4,72,72,96), V(9B,C0,C0,5B), \
			V(75,B7,B7,C2), V(E1,FD,FD,1C), V(3D,93,93,AE), V(4C,26,26,6A), \
			V(6C,36,36,5A), V(7E,3F,3F,41), V(F5,F7,F7,02), V(83,CC,CC,4F), \
			V(68,34,34,5C), V(51,A5,A5,F4), V(D1,E5,E5,34), V(F9,F1,F1,08), \
			V(E2,71,71,93), V(AB,D8,D8,73), V(62,31,31,53), V(2A,15,15,3F), \
			V(08,04,04,0C), V(95,C7,C7,52), V(46,23,23,65), V(9D,C3,C3,5E), \
			V(30,18,18,28), V(37,96,96,A1), V(0A,05,05,0F), V(2F,9A,9A,B5), \
			V(0E,07,07,09), V(24,12,12,36), V(1B,80,80,9B), V(DF,E2,E2,3D), \
			V(CD,EB,EB,26), V(4E,27,27,69), V(7F,B2,B2,CD), V(EA,75,75,9F), \
			V(12,09,09,1B), V(1D,83,83,9E), V(58,2C,2C,74), V(34,1A,1A,2E), \
			V(36,1B,1B,2D), V(DC,6E,6E,B2), V(B4,5A,5A,EE), V(5B,A0,A0,FB), \
			V(A4,52,52,F6), V(76,3B,3B,4D), V(B7,D6,D6,61), V(7D,B3,B3,CE), \
			V(52,29,29,7B), V(DD,E3,E3,3E), V(5E,2F,2F,71), V(13,84,84,97), \
			V(A6,53,53,F5), V(B9,D1,D1,68), V(00,00,00,00), V(C1,ED,ED,2C), \
			V(40,20,20,60), V(E3,FC,FC,1F), V(79,B1,B1,C8), V(B6,5B,5B,ED), \
			V(D4,6A,6A,BE), V(8D,CB,CB,46), V(67,BE,BE,D9), V(72,39,39,4B), \
			V(94,4A,4A,DE), V(98,4C,4C,D4), V(B0,58,58,E8), V(85,CF,CF,4A), \
			V(BB,D0,D0,6B), V(C5,EF,EF,2A), V(4F,AA,AA,E5), V(ED,FB,FB,16), \
			V(86,43,43,C5), V(9A,4D,4D,D7), V(66,33,33,55), V(11,85,85,94), \
			V(8A,45,45,CF), V(E9,F9,F9,10), V(04,02,02,06), V(FE,7F,7F,81), \
			V(A0,50,50,F0), V(78,3C,3C,44), V(25,9F,9F,BA), V(4B,A8,A8,E3), \
			V(A2,51,51,F3), V(5D,A3,A3,FE), V(80,40,40,C0), V(05,8F,8F,8A), \
			V(3F,92,92,AD), V(21,9D,9D,BC), V(70,38,38,48), V(F1,F5,F5,04), \
			V(63,BC,BC,DF), V(77,B6,B6,C1), V(AF,DA,DA,75), V(42,21,21,63), \
			V(20,10,10,30), V(E5,FF,FF,1A), V(FD,F3,F3,0E), V(BF,D2,D2,6D), \
			V(81,CD,CD,4C), V(18,0C,0C,14), V(26,13,13,35), V(C3,EC,EC,2F), \
			V(BE,5F,5F,E1), V(35,97,97,A2), V(88,44,44,CC), V(2E,17,17,39), \
			V(93,C4,C4,57), V(55,A7,A7,F2), V(FC,7E,7E,82), V(7A,3D,3D,47), \
			V(C8,64,64,AC), V(BA,5D,5D,E7), V(32,19,19,2B), V(E6,73,73,95), \
			V(C0,60,60,A0), V(19,81,81,98), V(9E,4F,4F,D1), V(A3,DC,DC,7F), \
			V(44,22,22,66), V(54,2A,2A,7E), V(3B,90,90,AB), V(0B,88,88,83), \
			V(8C,46,46,CA), V(C7,EE,EE,29), V(6B,B8,B8,D3), V(28,14,14,3C), \
			V(A7,DE,DE,79), V(BC,5E,5E,E2), V(16,0B,0B,1D), V(AD,DB,DB,76), \
			V(DB,E0,E0,3B), V(64,32,32,56), V(74,3A,3A,4E), V(14,0A,0A,1E), \
			V(92,49,49,DB), V(0C,06,06,0A), V(48,24,24,6C), V(B8,5C,5C,E4), \
			V(9F,C2,C2,5D), V(BD,D3,D3,6E), V(43,AC,AC,EF), V(C4,62,62,A6), \
			V(39,91,91,A8), V(31,95,95,A4), V(D3,E4,E4,37), V(F2,79,79,8B), \
			V(D5,E7,E7,32), V(8B,C8,C8,43), V(6E,37,37,59), V(DA,6D,6D,B7), \
			V(01,8D,8D,8C), V(B1,D5,D5,64), V(9C,4E,4E,D2), V(49,A9,A9,E0), \
			V(D8,6C,6C,B4), V(AC,56,56,FA), V(F3,F4,F4,07), V(CF,EA,EA,25), \
			V(CA,65,65,AF), V(F4,7A,7A,8E), V(47,AE,AE,E9), V(10,08,08,18), \
			V(6F,BA,BA,D5), V(F0,78,78,88), V(4A,25,25,6F), V(5C,2E,2E,72), \
			V(38,1C,1C,24), V(57,A6,A6,F1), V(73,B4,B4,C7), V(97,C6,C6,51), \
			V(CB,E8,E8,23), V(A1,DD,DD,7C), V(E8,74,74,9C), V(3E,1F,1F,21), \
			V(96,4B,4B,DD), V(61,BD,BD,DC), V(0D,8B,8B,86), V(0F,8A,8A,85), \
			V(E0,70,70,90), V(7C,3E,3E,42), V(71,B5,B5,C4), V(CC,66,66,AA), \
			V(90,48,48,D8), V(06,03,03,05), V(F7,F6,F6,01), V(1C,0E,0E,12), \
			V(C2,61,61,A3), V(6A,35,35,5F), V(AE,57,57,F9), V(69,B9,B9,D0), \
			V(17,86,86,91), V(99,C1,C1,58), V(3A,1D,1D,27), V(27,9E,9E,B9), \
			V(D9,E1,E1,38), V(EB,F8,F8,13), V(2B,98,98,B3), V(22,11,11,33), \
			V(D2,69,69,BB), V(A9,D9,D9,70), V(07,8E,8E,89), V(33,94,94,A7), \
			V(2D,9B,9B,B6), V(3C,1E,1E,22), V(15,87,87,92), V(C9,E9,E9,20), \
			V(87,CE,CE,49), V(AA,55,55,FF), V(50,28,28,78), V(A5,DF,DF,7A), \
			V(03,8C,8C,8F), V(59,A1,A1,F8), V(09,89,89,80), V(1A,0D,0D,17), \
			V(65,BF,BF,DA), V(D7,E6,E6,31), V(84,42,42,C6), V(D0,68,68,B8), \
			V(82,41,41,C3), V(29,99,99,B0), V(5A,2D,2D,77), V(1E,0F,0F,11), \
			V(7B,B0,B0,CB), V(A8,54,54,FC), V(6D,BB,BB,D6), V(2C,16,16,3A)

	#define V(a,b,c,d) 0x##a##b##c##d
	static uint32 FT0[256] = { FT };
	#undef V

	#define V(a,b,c,d) 0x##d##a##b##c
	static uint32 FT1[256] = { FT };
	#undef V

	#define V(a,b,c,d) 0x##c##d##a##b
	static uint32 FT2[256] = { FT };
	#undef V

	#define V(a,b,c,d) 0x##b##c##d##a
	static uint32 FT3[256] = { FT };
	#undef V

	// reverse S-box

	static uint32 RSb[256] =
	{
			0x52, 0x09, 0x6A, 0xD5, 0x30, 0x36, 0xA5, 0x38,
			0xBF, 0x40, 0xA3, 0x9E, 0x81, 0xF3, 0xD7, 0xFB,
			0x7C, 0xE3, 0x39, 0x82, 0x9B, 0x2F, 0xFF, 0x87,
			0x34, 0x8E, 0x43, 0x44, 0xC4, 0xDE, 0xE9, 0xCB,
			0x54, 0x7B, 0x94, 0x32, 0xA6, 0xC2, 0x23, 0x3D,
			0xEE, 0x4C, 0x95, 0x0B, 0x42, 0xFA, 0xC3, 0x4E,
			0x08, 0x2E, 0xA1, 0x66, 0x28, 0xD9, 0x24, 0xB2,
			0x76, 0x5B, 0xA2, 0x49, 0x6D, 0x8B, 0xD1, 0x25,
			0x72, 0xF8, 0xF6, 0x64, 0x86, 0x68, 0x98, 0x16,
			0xD4, 0xA4, 0x5C, 0xCC, 0x5D, 0x65, 0xB6, 0x92,
			0x6C, 0x70, 0x48, 0x50, 0xFD, 0xED, 0xB9, 0xDA,
			0x5E, 0x15, 0x46, 0x57, 0xA7, 0x8D, 0x9D, 0x84,
			0x90, 0xD8, 0xAB, 0x00, 0x8C, 0xBC, 0xD3, 0x0A,
			0xF7, 0xE4, 0x58, 0x05, 0xB8, 0xB3, 0x45, 0x06,
			0xD0, 0x2C, 0x1E, 0x8F, 0xCA, 0x3F, 0x0F, 0x02,
			0xC1, 0xAF, 0xBD, 0x03, 0x01, 0x13, 0x8A, 0x6B,
			0x3A, 0x91, 0x11, 0x41, 0x4F, 0x67, 0xDC, 0xEA,
			0x97, 0xF2, 0xCF, 0xCE, 0xF0, 0xB4, 0xE6, 0x73,
			0x96, 0xAC, 0x74, 0x22, 0xE7, 0xAD, 0x35, 0x85,
			0xE2, 0xF9, 0x37, 0xE8, 0x1C, 0x75, 0xDF, 0x6E,
			0x47, 0xF1, 0x1A, 0x71, 0x1D, 0x29, 0xC5, 0x89,
			0x6F, 0xB7, 0x62, 0x0E, 0xAA, 0x18, 0xBE, 0x1B,
			0xFC, 0x56, 0x3E, 0x4B, 0xC6, 0xD2, 0x79, 0x20,
			0x9A, 0xDB, 0xC0, 0xFE, 0x78, 0xCD, 0x5A, 0xF4,
			0x1F, 0xDD, 0xA8, 0x33, 0x88, 0x07, 0xC7, 0x31,
			0xB1, 0x12, 0x10, 0x59, 0x27, 0x80, 0xEC, 0x5F,
			0x60, 0x51, 0x7F, 0xA9, 0x19, 0xB5, 0x4A, 0x0D,
			0x2D, 0xE5, 0x7A, 0x9F, 0x93, 0xC9, 0x9C, 0xEF,
			0xA0, 0xE0, 0x3B, 0x4D, 0xAE, 0x2A, 0xF5, 0xB0,
			0xC8, 0xEB, 0xBB, 0x3C, 0x83, 0x53, 0x99, 0x61,
			0x17, 0x2B, 0x04, 0x7E, 0xBA, 0x77, 0xD6, 0x26,
			0xE1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0C, 0x7D
	};

	// reverse table

	#define RT \
	\
			V(51,F4,A7,50), V(7E,41,65,53), V(1A,17,A4,C3), V(3A,27,5E,96), \
			V(3B,AB,6B,CB), V(1F,9D,45,F1), V(AC,FA,58,AB), V(4B,E3,03,93), \
			V(20,30,FA,55), V(AD,76,6D,F6), V(88,CC,76,91), V(F5,02,4C,25), \
			V(4F,E5,D7,FC), V(C5,2A,CB,D7), V(26,35,44,80), V(B5,62,A3,8F), \
			V(DE,B1,5A,49), V(25,BA,1B,67), V(45,EA,0E,98), V(5D,FE,C0,E1), \
			V(C3,2F,75,02), V(81,4C,F0,12), V(8D,46,97,A3), V(6B,D3,F9,C6), \
			V(03,8F,5F,E7), V(15,92,9C,95), V(BF,6D,7A,EB), V(95,52,59,DA), \
			V(D4,BE,83,2D), V(58,74,21,D3), V(49,E0,69,29), V(8E,C9,C8,44), \
			V(75,C2,89,6A), V(F4,8E,79,78), V(99,58,3E,6B), V(27,B9,71,DD), \
			V(BE,E1,4F,B6), V(F0,88,AD,17), V(C9,20,AC,66), V(7D,CE,3A,B4), \
			V(63,DF,4A,18), V(E5,1A,31,82), V(97,51,33,60), V(62,53,7F,45), \
			V(B1,64,77,E0), V(BB,6B,AE,84), V(FE,81,A0,1C), V(F9,08,2B,94), \
			V(70,48,68,58), V(8F,45,FD,19), V(94,DE,6C,87), V(52,7B,F8,B7), \
			V(AB,73,D3,23), V(72,4B,02,E2), V(E3,1F,8F,57), V(66,55,AB,2A), \
			V(B2,EB,28,07), V(2F,B5,C2,03), V(86,C5,7B,9A), V(D3,37,08,A5), \
			V(30,28,87,F2), V(23,BF,A5,B2), V(02,03,6A,BA), V(ED,16,82,5C), \
			V(8A,CF,1C,2B), V(A7,79,B4,92), V(F3,07,F2,F0), V(4E,69,E2,A1), \
			V(65,DA,F4,CD), V(06,05,BE,D5), V(D1,34,62,1F), V(C4,A6,FE,8A), \
			V(34,2E,53,9D), V(A2,F3,55,A0), V(05,8A,E1,32), V(A4,F6,EB,75), \
			V(0B,83,EC,39), V(40,60,EF,AA), V(5E,71,9F,06), V(BD,6E,10,51), \
			V(3E,21,8A,F9), V(96,DD,06,3D), V(DD,3E,05,AE), V(4D,E6,BD,46), \
			V(91,54,8D,B5), V(71,C4,5D,05), V(04,06,D4,6F), V(60,50,15,FF), \
			V(19,98,FB,24), V(D6,BD,E9,97), V(89,40,43,CC), V(67,D9,9E,77), \
			V(B0,E8,42,BD), V(07,89,8B,88), V(E7,19,5B,38), V(79,C8,EE,DB), \
			V(A1,7C,0A,47), V(7C,42,0F,E9), V(F8,84,1E,C9), V(00,00,00,00), \
			V(09,80,86,83), V(32,2B,ED,48), V(1E,11,70,AC), V(6C,5A,72,4E), \
			V(FD,0E,FF,FB), V(0F,85,38,56), V(3D,AE,D5,1E), V(36,2D,39,27), \
			V(0A,0F,D9,64), V(68,5C,A6,21), V(9B,5B,54,D1), V(24,36,2E,3A), \
			V(0C,0A,67,B1), V(93,57,E7,0F), V(B4,EE,96,D2), V(1B,9B,91,9E), \
			V(80,C0,C5,4F), V(61,DC,20,A2), V(5A,77,4B,69), V(1C,12,1A,16), \
			V(E2,93,BA,0A), V(C0,A0,2A,E5), V(3C,22,E0,43), V(12,1B,17,1D), \
			V(0E,09,0D,0B), V(F2,8B,C7,AD), V(2D,B6,A8,B9), V(14,1E,A9,C8), \
			V(57,F1,19,85), V(AF,75,07,4C), V(EE,99,DD,BB), V(A3,7F,60,FD), \
			V(F7,01,26,9F), V(5C,72,F5,BC), V(44,66,3B,C5), V(5B,FB,7E,34), \
			V(8B,43,29,76), V(CB,23,C6,DC), V(B6,ED,FC,68), V(B8,E4,F1,63), \
			V(D7,31,DC,CA), V(42,63,85,10), V(13,97,22,40), V(84,C6,11,20), \
			V(85,4A,24,7D), V(D2,BB,3D,F8), V(AE,F9,32,11), V(C7,29,A1,6D), \
			V(1D,9E,2F,4B), V(DC,B2,30,F3), V(0D,86,52,EC), V(77,C1,E3,D0), \
			V(2B,B3,16,6C), V(A9,70,B9,99), V(11,94,48,FA), V(47,E9,64,22), \
			V(A8,FC,8C,C4), V(A0,F0,3F,1A), V(56,7D,2C,D8), V(22,33,90,EF), \
			V(87,49,4E,C7), V(D9,38,D1,C1), V(8C,CA,A2,FE), V(98,D4,0B,36), \
			V(A6,F5,81,CF), V(A5,7A,DE,28), V(DA,B7,8E,26), V(3F,AD,BF,A4), \
			V(2C,3A,9D,E4), V(50,78,92,0D), V(6A,5F,CC,9B), V(54,7E,46,62), \
			V(F6,8D,13,C2), V(90,D8,B8,E8), V(2E,39,F7,5E), V(82,C3,AF,F5), \
			V(9F,5D,80,BE), V(69,D0,93,7C), V(6F,D5,2D,A9), V(CF,25,12,B3), \
			V(C8,AC,99,3B), V(10,18,7D,A7), V(E8,9C,63,6E), V(DB,3B,BB,7B), \
			V(CD,26,78,09), V(6E,59,18,F4), V(EC,9A,B7,01), V(83,4F,9A,A8), \
			V(E6,95,6E,65), V(AA,FF,E6,7E), V(21,BC,CF,08), V(EF,15,E8,E6), \
			V(BA,E7,9B,D9), V(4A,6F,36,CE), V(EA,9F,09,D4), V(29,B0,7C,D6), \
			V(31,A4,B2,AF), V(2A,3F,23,31), V(C6,A5,94,30), V(35,A2,66,C0), \
			V(74,4E,BC,37), V(FC,82,CA,A6), V(E0,90,D0,B0), V(33,A7,D8,15), \
			V(F1,04,98,4A), V(41,EC,DA,F7), V(7F,CD,50,0E), V(17,91,F6,2F), \
			V(76,4D,D6,8D), V(43,EF,B0,4D), V(CC,AA,4D,54), V(E4,96,04,DF), \
			V(9E,D1,B5,E3), V(4C,6A,88,1B), V(C1,2C,1F,B8), V(46,65,51,7F), \
			V(9D,5E,EA,04), V(01,8C,35,5D), V(FA,87,74,73), V(FB,0B,41,2E), \
			V(B3,67,1D,5A), V(92,DB,D2,52), V(E9,10,56,33), V(6D,D6,47,13), \
			V(9A,D7,61,8C), V(37,A1,0C,7A), V(59,F8,14,8E), V(EB,13,3C,89), \
			V(CE,A9,27,EE), V(B7,61,C9,35), V(E1,1C,E5,ED), V(7A,47,B1,3C), \
			V(9C,D2,DF,59), V(55,F2,73,3F), V(18,14,CE,79), V(73,C7,37,BF), \
			V(53,F7,CD,EA), V(5F,FD,AA,5B), V(DF,3D,6F,14), V(78,44,DB,86), \
			V(CA,AF,F3,81), V(B9,68,C4,3E), V(38,24,34,2C), V(C2,A3,40,5F), \
			V(16,1D,C3,72), V(BC,E2,25,0C), V(28,3C,49,8B), V(FF,0D,95,41), \
			V(39,A8,01,71), V(08,0C,B3,DE), V(D8,B4,E4,9C), V(64,56,C1,90), \
			V(7B,CB,84,61), V(D5,32,B6,70), V(48,6C,5C,74), V(D0,B8,57,42)

	#define V(a,b,c,d) 0x##a##b##c##d
	static uint32 RT0[256] = { RT };
	#undef V

	#define V(a,b,c,d) 0x##d##a##b##c
	static uint32 RT1[256] = { RT };
	#undef V

	#define V(a,b,c,d) 0x##c##d##a##b
	static uint32 RT2[256] = { RT };
	#undef V

	#define V(a,b,c,d) 0x##b##c##d##a
	static uint32 RT3[256] = { RT };
	#undef V

	// round constants

	static uint32 RCON[10] =
	{
			0x01000000, 0x02000000, 0x04000000, 0x08000000,
			0x10000000, 0x20000000, 0x40000000, 0x80000000,
			0x1B000000, 0x36000000
	};

	// platform-independant 32-bit integer manipulation macros

	#define GET_UINT32(n,b,i)                       \
	{                                               \
			(n) = ( (uint32) (b)[(i)    ] << 24 )       \
					| ( (uint32) (b)[(i) + 1] << 16 )       \
					| ( (uint32) (b)[(i) + 2] <<  8 )       \
					| ( (uint32) (b)[(i) + 3]       );      \
	}

	#define PUT_UINT32(n,b,i)                       \
	{                                               \
			(b)[(i)    ] = (uint8) ( (n) >> 24 );       \
			(b)[(i) + 1] = (uint8) ( (n) >> 16 );       \
			(b)[(i) + 2] = (uint8) ( (n) >>  8 );       \
			(b)[(i) + 3] = (uint8) ( (n)       );       \
	}


	// key scheduling routine

	int aes_set_key( struct aes_context *ctx, uint8 *key, int nbits )
	{
			int i;
			uint32 *RK;

			switch( nbits )
			{
					case 128: ctx->nr = 10; break;
					case 192: ctx->nr = 12; break;
					case 256: ctx->nr = 14; break;
					default : return( 1 );
			}

			RK = ctx->erk;

			for( i = 0; i < (nbits >> 5); i++ )
			{
					GET_UINT32( RK[i], key, i * 4 );
			}

			// setup encryption round keys

			switch( nbits )
			{
			case 128:

					for( i = 0; i < 10; i++, RK += 4 )
					{
							RK[4]  = RK[0] ^ RCON[i] ^
													( FSb[ (uint8) ( RK[3] >> 16 ) ] << 24 ) ^
													( FSb[ (uint8) ( RK[3] >>  8 ) ] << 16 ) ^
													( FSb[ (uint8) ( RK[3]       ) ] <<  8 ) ^
													( FSb[ (uint8) ( RK[3] >> 24 ) ]       );

							RK[5]  = RK[1] ^ RK[4];
							RK[6]  = RK[2] ^ RK[5];
							RK[7]  = RK[3] ^ RK[6];
					}
					break;

			case 192:

					for( i = 0; i < 8; i++, RK += 6 )
					{
							RK[6]  = RK[0] ^ RCON[i] ^
													( FSb[ (uint8) ( RK[5] >> 16 ) ] << 24 ) ^
													( FSb[ (uint8) ( RK[5] >>  8 ) ] << 16 ) ^
													( FSb[ (uint8) ( RK[5]       ) ] <<  8 ) ^
													( FSb[ (uint8) ( RK[5] >> 24 ) ]       );

							RK[7]  = RK[1] ^ RK[6];
							RK[8]  = RK[2] ^ RK[7];
							RK[9]  = RK[3] ^ RK[8];
							RK[10] = RK[4] ^ RK[9];
							RK[11] = RK[5] ^ RK[10];
					}
					break;

			case 256:

					for( i = 0; i < 7; i++, RK += 8 )
					{
							RK[8]  = RK[0] ^ RCON[i] ^
													( FSb[ (uint8) ( RK[7] >> 16 ) ] << 24 ) ^
													( FSb[ (uint8) ( RK[7] >>  8 ) ] << 16 ) ^
													( FSb[ (uint8) ( RK[7]       ) ] <<  8 ) ^
													( FSb[ (uint8) ( RK[7] >> 24 ) ]       );

							RK[9]  = RK[1] ^ RK[8];
							RK[10] = RK[2] ^ RK[9];
							RK[11] = RK[3] ^ RK[10];

							RK[12] = RK[4] ^
													( FSb[ (uint8) ( RK[11] >> 24 ) ] << 24 ) ^
													( FSb[ (uint8) ( RK[11] >> 16 ) ] << 16 ) ^
													( FSb[ (uint8) ( RK[11] >>  8 ) ] <<  8 ) ^
													( FSb[ (uint8) ( RK[11]       ) ]       );

							RK[13] = RK[5] ^ RK[12];
							RK[14] = RK[6] ^ RK[13];
							RK[15] = RK[7] ^ RK[14];
					}
					break;
			}

			// setup decryption round keys

			for( i = 0; i <= ctx->nr; i++ )
			{
					ctx->drk[i * 4    ] = ctx->erk[( ctx->nr - i ) * 4    ];
					ctx->drk[i * 4 + 1] = ctx->erk[( ctx->nr - i ) * 4 + 1];
					ctx->drk[i * 4 + 2] = ctx->erk[( ctx->nr - i ) * 4 + 2];
					ctx->drk[i * 4 + 3] = ctx->erk[( ctx->nr - i ) * 4 + 3];
			}

			for( i = 1, RK = ctx->drk + 4; i < ctx->nr; i++, RK += 4 )
			{
					RK[0] = RT0[ FSb[ (uint8) ( RK[0] >> 24 ) ] ] ^
									RT1[ FSb[ (uint8) ( RK[0] >> 16 ) ] ] ^
									RT2[ FSb[ (uint8) ( RK[0] >>  8 ) ] ] ^
									RT3[ FSb[ (uint8) ( RK[0]       ) ] ];

					RK[1] = RT0[ FSb[ (uint8) ( RK[1] >> 24 ) ] ] ^
									RT1[ FSb[ (uint8) ( RK[1] >> 16 ) ] ] ^
									RT2[ FSb[ (uint8) ( RK[1] >>  8 ) ] ] ^
									RT3[ FSb[ (uint8) ( RK[1]       ) ] ];

					RK[2] = RT0[ FSb[ (uint8) ( RK[2] >> 24 ) ] ] ^
									RT1[ FSb[ (uint8) ( RK[2] >> 16 ) ] ] ^
									RT2[ FSb[ (uint8) ( RK[2] >>  8 ) ] ] ^
									RT3[ FSb[ (uint8) ( RK[2]       ) ] ];

					RK[3] = RT0[ FSb[ (uint8) ( RK[3] >> 24 ) ] ] ^
									RT1[ FSb[ (uint8) ( RK[3] >> 16 ) ] ] ^
									RT2[ FSb[ (uint8) ( RK[3] >>  8 ) ] ] ^
									RT3[ FSb[ (uint8) ( RK[3]       ) ] ];
			}

			return( 0 );
	}

	// 128-bit block encryption routine
	void aes_encrypt( struct aes_context *ctx, uint8 data[16] )
	{
			uint32 *RK, X0, X1, X2, X3, Y0, Y1, Y2, Y3;

			RK = ctx->erk;

			GET_UINT32( X0, data,  0 ); X0 ^= RK[0];
			GET_UINT32( X1, data,  4 ); X1 ^= RK[1];
			GET_UINT32( X2, data,  8 ); X2 ^= RK[2];
			GET_UINT32( X3, data, 12 ); X3 ^= RK[3];

	#define FROUND(X0,X1,X2,X3,Y0,Y1,Y2,Y3)         \
	{                                               \
			RK += 4;                                    \
																									\
			X0 = RK[0] ^ FT0[ (uint8) ( Y0 >> 24 ) ] ^  \
									 FT1[ (uint8) ( Y1 >> 16 ) ] ^  \
									 FT2[ (uint8) ( Y2 >>  8 ) ] ^  \
									 FT3[ (uint8) ( Y3       ) ];   \
																									\
			X1 = RK[1] ^ FT0[ (uint8) ( Y1 >> 24 ) ] ^  \
									 FT1[ (uint8) ( Y2 >> 16 ) ] ^  \
									 FT2[ (uint8) ( Y3 >>  8 ) ] ^  \
									 FT3[ (uint8) ( Y0       ) ];   \
																									\
			X2 = RK[2] ^ FT0[ (uint8) ( Y2 >> 24 ) ] ^  \
									 FT1[ (uint8) ( Y3 >> 16 ) ] ^  \
									 FT2[ (uint8) ( Y0 >>  8 ) ] ^  \
									 FT3[ (uint8) ( Y1       ) ];   \
																									\
			X3 = RK[3] ^ FT0[ (uint8) ( Y3 >> 24 ) ] ^  \
									 FT1[ (uint8) ( Y0 >> 16 ) ] ^  \
									 FT2[ (uint8) ( Y1 >>  8 ) ] ^  \
									 FT3[ (uint8) ( Y2       ) ];   \
	}

			FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );           // round 1
			FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );           // round 2
			FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );           // round 3
			FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );           // round 4
			FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );           // round 5
			FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );           // round 6
			FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );           // round 7
			FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );           // round 8
			FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );           // round 9

			if( ctx->nr > 10 )
			{
					FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       // round 10
					FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       // round 11
			}

			if( ctx->nr > 12 )
			{
					FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       // round 12
					FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       // round 13
			}

			// last round

			RK += 4;

			X0 = RK[0] ^ ( FSb[ (uint8) ( Y0 >> 24 ) ] << 24 ) ^
									 ( FSb[ (uint8) ( Y1 >> 16 ) ] << 16 ) ^
									 ( FSb[ (uint8) ( Y2 >>  8 ) ] <<  8 ) ^
									 ( FSb[ (uint8) ( Y3       ) ]       );

			X1 = RK[1] ^ ( FSb[ (uint8) ( Y1 >> 24 ) ] << 24 ) ^
									 ( FSb[ (uint8) ( Y2 >> 16 ) ] << 16 ) ^
									 ( FSb[ (uint8) ( Y3 >>  8 ) ] <<  8 ) ^
									 ( FSb[ (uint8) ( Y0       ) ]       );

			X2 = RK[2] ^ ( FSb[ (uint8) ( Y2 >> 24 ) ] << 24 ) ^
									 ( FSb[ (uint8) ( Y3 >> 16 ) ] << 16 ) ^
									 ( FSb[ (uint8) ( Y0 >>  8 ) ] <<  8 ) ^
									 ( FSb[ (uint8) ( Y1       ) ]       );

			X3 = RK[3] ^ ( FSb[ (uint8) ( Y3 >> 24 ) ] << 24 ) ^
									 ( FSb[ (uint8) ( Y0 >> 16 ) ] << 16 ) ^
									 ( FSb[ (uint8) ( Y1 >>  8 ) ] <<  8 ) ^
									 ( FSb[ (uint8) ( Y2       ) ]       );

			PUT_UINT32( X0, data,  0 );
			PUT_UINT32( X1, data,  4 );
			PUT_UINT32( X2, data,  8 );
			PUT_UINT32( X3, data, 12 );
	}

	// 128-bit block decryption routine

	void aes_decrypt( struct aes_context *ctx, uint8 data[16] )
	{
			uint32 *RK, X0, X1, X2, X3, Y0, Y1, Y2, Y3;

			RK = ctx->drk;

			GET_UINT32( X0, data,  0 ); X0 ^= RK[0];
			GET_UINT32( X1, data,  4 ); X1 ^= RK[1];
			GET_UINT32( X2, data,  8 ); X2 ^= RK[2];
			GET_UINT32( X3, data, 12 ); X3 ^= RK[3];

	#define RROUND(X0,X1,X2,X3,Y0,Y1,Y2,Y3)         \
	{                                               \
			RK += 4;                                    \
																									\
			X0 = RK[0] ^ RT0[ (uint8) ( Y0 >> 24 ) ] ^  \
									 RT1[ (uint8) ( Y3 >> 16 ) ] ^  \
									 RT2[ (uint8) ( Y2 >>  8 ) ] ^  \
									 RT3[ (uint8) ( Y1       ) ];   \
																									\
			X1 = RK[1] ^ RT0[ (uint8) ( Y1 >> 24 ) ] ^  \
									 RT1[ (uint8) ( Y0 >> 16 ) ] ^  \
									 RT2[ (uint8) ( Y3 >>  8 ) ] ^  \
									 RT3[ (uint8) ( Y2       ) ];   \
																									\
			X2 = RK[2] ^ RT0[ (uint8) ( Y2 >> 24 ) ] ^  \
									 RT1[ (uint8) ( Y1 >> 16 ) ] ^  \
									 RT2[ (uint8) ( Y0 >>  8 ) ] ^  \
									 RT3[ (uint8) ( Y3       ) ];   \
																									\
			X3 = RK[3] ^ RT0[ (uint8) ( Y3 >> 24 ) ] ^  \
									 RT1[ (uint8) ( Y2 >> 16 ) ] ^  \
									 RT2[ (uint8) ( Y1 >>  8 ) ] ^  \
									 RT3[ (uint8) ( Y0       ) ];   \
	}

			RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );           // round 1
			RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );           // round 2
			RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );           // round 3
			RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );           // round 4
			RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );           // round 5
			RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );           // round 6
			RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );           // round 7
			RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );           // round 8
			RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );           // round 9

			if( ctx->nr > 10 )
			{
					RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       // round 10
					RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       // round 11
			}

			if( ctx->nr > 12 )
			{
					RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       // round 12
					RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       // round 13
			}

			// last round

			RK += 4;

			X0 = RK[0] ^ ( RSb[ (uint8) ( Y0 >> 24 ) ] << 24 ) ^
									 ( RSb[ (uint8) ( Y3 >> 16 ) ] << 16 ) ^
									 ( RSb[ (uint8) ( Y2 >>  8 ) ] <<  8 ) ^
									 ( RSb[ (uint8) ( Y1       ) ]       );

			X1 = RK[1] ^ ( RSb[ (uint8) ( Y1 >> 24 ) ] << 24 ) ^
									 ( RSb[ (uint8) ( Y0 >> 16 ) ] << 16 ) ^
									 ( RSb[ (uint8) ( Y3 >>  8 ) ] <<  8 ) ^
									 ( RSb[ (uint8) ( Y2       ) ]       );

			X2 = RK[2] ^ ( RSb[ (uint8) ( Y2 >> 24 ) ] << 24 ) ^
									 ( RSb[ (uint8) ( Y1 >> 16 ) ] << 16 ) ^
									 ( RSb[ (uint8) ( Y0 >>  8 ) ] <<  8 ) ^
									 ( RSb[ (uint8) ( Y3       ) ]       );

			X3 = RK[3] ^ ( RSb[ (uint8) ( Y3 >> 24 ) ] << 24 ) ^
									 ( RSb[ (uint8) ( Y2 >> 16 ) ] << 16 ) ^
									 ( RSb[ (uint8) ( Y1 >>  8 ) ] <<  8 ) ^
									 ( RSb[ (uint8) ( Y0       ) ]       );

			PUT_UINT32( X0, data,  0 );
			PUT_UINT32( X1, data,  4 );
			PUT_UINT32( X2, data,  8 );
			PUT_UINT32( X3, data, 12 );
	}






		// AES test routine:

		// those are the standard FIPS 197 test vectors
	static unsigned char KEYs[3][33] =
	{
			"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F",

			"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F" \
					"\x10\x11\x12\x13\x14\x15\x16\x17",

			"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F" \
					"\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F"
	};

	static unsigned char PT[17] =
			"\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xAA\xBB\xCC\xDD\xEE\xFF";

	static unsigned char CTs[3][17] =
	{
			"\x69\xC4\xE0\xD8\x6A\x7B\x04\x30\xD8\xCD\xB7\x80\x70\xB4\xC5\x5A",
			"\xDD\xA9\x7C\xA4\x86\x4C\xDF\xE0\x6E\xAF\x70\xA0\xEC\x0D\x71\x91",
			"\x8E\xA2\xB7\xCA\x51\x67\x45\xBF\xEA\xFC\x49\x90\x4B\x49\x60\x89"
	};


		// returns true for pass
	bool test_AES(void) {
			bool result=true; // false=fail, true=pass
			int i;
			struct aes_context ctx;
			unsigned char data[16];
			
			for( i = 0; i < 3; i++ )
			{
				memcpy( data, PT, 16 );
				
				aes_set_key( &ctx, KEYs[i], 128 + i * 64 );
				aes_encrypt( &ctx, data );
				
				if (memcmp( data, CTs[i], 16 ) )
				{
					result=0;
					return(result);
				}
			}
			
			for( i = 0; i < 3; i++ )
			{
				memcpy( data, CTs[i], 16 );
				
				aes_set_key( &ctx, KEYs[i], 128 + i * 64 );
				aes_decrypt( &ctx, data );
				
				if (memcmp( data, PT, 16 ) )
				{
					result=false;
					return(result);
				}
			}
			
			return(result);
	}


} // namespace crypto_primitives

